/*
 *	bsyscalls.h
 *	Blocking sys-call interface definition
 *	Copyright (C) 2000 F.R.M. Barnes, J. Moores and P.H. Welch
 *
 *	This program is free software; you can redistribute it and/or modify
 *	it under the terms of the GNU General Public License as published by
 *	the Free Software Foundation; either version 2 of the License, or
 *	(at your option) any later version.
 *
 *	This program is distributed in the hope that it will be useful,
 *	but WITHOUT ANY WARRANTY; without even the implied warranty of
 *	MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *	GNU General Public License for more details.
 *
 *	You should have received a copy of the GNU General Public License
 *	along with this program; if not, write to the Free Software
 *	Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 */

#ifndef __BSYSCALLS_H
#define __BSYSCALLS_H

#define NUM_CLONES		(3)
#define MAX_NUM_CLONES		(4096)
#define CLONE_STACK_SIZE	(32 * 4096)

typedef struct TAG_bsc_thread {
	int		pid;				/* process ID */
	int		thr_num;			/* thread number */
	int		par_sync;			/* pipe for synchronisation */
	int		*ws_ptr;			/* Wptr of process executing (already adjusted for return) */
	int		*ws_arg;			/* argument to func */
	void		(*func)(word *);		/* function to execute */
	char		*raddr;				/* return address */
	int		terminated;			/* set when terminated */

	int		cancel;				/* set when cancelling */
	int		adjustment;			/* are we terminateable ? (or offset of workspace params) */
	sigjmp_buf	*jbuf;				/* ptr to jump buffer - (allocated just beyond this structure) */
	void		*user_ptr;			/* ptr to where the user can store state (just beyond `jbuf') */
	void		(*clone_cleanup)(void *);	/* ptr to clean-up function */
	char		padding[12];

	#ifdef BLOCKING_PROFILE
		char		bp_padding[16];		/* cache line snuglies */
		long long	time_dispatch;		/* time when dispatched */
		long long	time_kill;		/* time when signalled to kill */
	#endif
} bsc_thread;

#ifdef BLOCKING_PROFILE
	#define ll_lo(X)	((int *)&(X))[0]
	#define ll_hi(X)	((int *)&(X))[1]

	#define LOADTIME(time) \
		__asm__ __volatile__ ("\n\t" \
			".byte 0x0f,0x31\n\t" \
			: "=a" (ll_lo(time)), "=d" (ll_hi(time)));
#endif

#ifndef __BSYSCALLS_C
extern int		bsyscalls_create_clones (void);
extern void		bsyscalls_dispatch (word *ws, char *raddr, word *ws_arg, void (*func)(word *), int adjustment);
extern void		bsyscalls_destroy_clones (void);
extern int		bsyscalls_pending (void);
#endif	/* !__BSYSCALLS_C */

#endif	/* !__BSYSCALLS_H */

