/*
 *	Occam entry point
 *	Copyright (C) 1996-1999 Jim Moores
 *	Based on the KRoC/sparc kernel Copyright (C) 1994-2000 D.C. Wood and P.H. Welch
 *
 *	This program is free software; you can redistribute it and/or modify
 *	it under the terms of the GNU General Public License as published by
 *	the Free Software Foundation; either version 2 of the License, or
 *	(at your option) any later version.
 *
 *	This program is distributed in the hope that it will be useful,
 *	but WITHOUT ANY WARRANTY; without even the implied warranty of
 *	MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *	GNU General Public License for more details.
 *
 *	You should have received a copy of the GNU General Public License
 *	along with this program; if not, write to the Free Software
 *	Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 */

#include <stdlib.h>
#include <stdio.h>

#include "sched_types.h"
#include "sched_consts.h"
#include "sched.h"
#include "sched_asm_inserts.h"
#include "occam_io.h"
#include "queue_defs.h"

#include "deadlock.h"

#ifdef BLOCKING_SYSCALLS
	#include <setjmp.h>
	#include "bsyscalls.h"
#endif


extern void _occam_start(); /* occam code entry point */
extern word _wsbytes; /* occam workspace size */
extern word _vsbytes; /* occam vectorspace size */
extern void *kernel(); /* set up jump table */

extern FILE *kroc_out;
extern FILE *kroc_err;

static int saved_stack_ptr;

int		_enter () {
	char		*ws, *vs;
	void		*hard_print = kernel(0);	/* returns pointer to hard print code to pass to */
							/* init_occam_io - need to change this... */
#ifdef DEBUG_KERNEL
	fprintf( stderr, "in _enter()\n" );
#endif
	init_occam_io(hard_print);
	/*Wptr = (word *)malloc(_wsbytes+(Frame*4));*/
	ws = (char *)malloc( _wsbytes + 24 + _vsbytes + 256 );
	Wptr = (word *)ws;
#ifdef DEBUG_KERNEL
	if( !Wptr ) {
	  	fprintf( stderr,  "malloc(%d) failed.\n", _wsbytes + 24 + _vsbytes + 256 );
	} else {
	  	fprintf( stderr, "allocated %d bytes for workspace and vectorspace.\n", _wsbytes+24+_vsbytes+256 );
	}
#else
	if( !Wptr ) {
		fprintf( stderr, "KRoC: fatal.  Unable to allocate workspace. (%d bytes)\n", _wsbytes + 24 + _vsbytes + 256 );
		user_bad_exit ();
	}
#endif
	(byte *)Wptr += _wsbytes;
	Wptr[1] = (word)&kbd_chan;
	Wptr[2] = (word)&scr_chan;
	Wptr[3] = (word)&err_chan;
	/*Wptr[4] = (word)malloc(_vsbytes);*/
	vs = ws + (_wsbytes + 24 + 128);
	Wptr[4] = (word)vs;
#ifdef DEBUG_KERNEL
	fprintf( stderr, "Wptr=%p  ws=%p  vs=%p  _wsbytes=%d  _vsbytes=%d\n", Wptr, ws, vs, _wsbytes, _vsbytes );
#endif
	/* give deadlock stuff a handle on ws/vs */
	give_wsvs (ws, vs);

	Fptr = Bptr = Tptr = NotProcess_p;
#ifdef BLOCKING_SYSCALLS
	if (bsyscalls_create_clones ()) {
		fprintf (stderr, "KRoC: Unable to create clones for blocking syscalls.\n");
		user_bad_exit ();
	}
#endif	/* BLOCKING_SYSCALLS */
	STATUS = 0;
	scr_workspace[Device] = (word)kroc_out;
	err_workspace[Device] = (word)kroc_err;
	/* save stack-ptr */
	__asm__ __volatile__ ("\n" \
		"	movl %%esp, %0\n" \
		: "=g" (saved_stack_ptr) : : "memory");
	RESTORE;
	_occam_start();
	__asm__ __volatile__ ("\n" \
		"	movl %0, %%esp\n" \
		: : "g" (saved_stack_ptr) : "memory");
#ifdef BLOCKING_DEBUG
	fprintf (stderr, "KRoC: OUT OF OCCAM WORLD!\n");
#endif
	STATUS &= 1;
	return STATUS;
}


