/*
 *	occam I/O workspace setup
 *	Copyright (C) 1996-1999 Jim Moores
 *	Based on the KRoC/sparc kernel Copyright (C) 1994-2000 D.C. Wood and P.H. Welch
 *
 *	This program is free software; you can redistribute it and/or modify
 *	it under the terms of the GNU General Public License as published by
 *	the Free Software Foundation; either version 2 of the License, or
 *	(at your option) any later version.
 *
 *	This program is distributed in the hope that it will be useful,
 *	but WITHOUT ANY WARRANTY; without even the implied warranty of
 *	MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *	GNU General Public License for more details.
 *
 *	You should have received a copy of the GNU General Public License
 *	along with this program; if not, write to the Free Software
 *	Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 */

#include <sys/types.h>

#include "sched_consts.h"
#include "sched_types.h"
#include "sched_asm_inserts.h"
#include "sched.h"

word _kbd_workspace_bottom[4];
word _scr_workspace_bottom[6];
word _err_workspace_bottom[6];

/* pointers to the right offsets into each workspace for convenience */
/* this is slightly different to an assembler label. */
word *kbd_workspace;
word *scr_workspace;
word *err_workspace;

/* the actual locations holding the channel words.  Obviously if the */
/* channel address is needed then the address of (&) operator must be */
/* used */
word *kbd_chan;
word *scr_chan;
word *err_chan;

/* need to pass in pointers to hard_print at run time */
/* with a serious cast! */
void init_occam_io(void *hard_print) {
  int i;
  kbd_workspace = &(_kbd_workspace_bottom[3]);
  for (i=0; i<4; i++) {
    _kbd_workspace_bottom[i] = NotProcess_p;
  }
  scr_workspace = &(_scr_workspace_bottom[3]);
  _scr_workspace_bottom[0] = (word)scr_workspace;
  _scr_workspace_bottom[1] = (word)NotProcess_p;
  _scr_workspace_bottom[2] = (word)hard_print;
  /* this is where scr_workspace points to */
  _scr_workspace_bottom[3] = 0;
  _scr_workspace_bottom[4] = (word)&scr_chan;
  _scr_workspace_bottom[5] = 0;

  err_workspace = &(_err_workspace_bottom[3]);
  _err_workspace_bottom[0] = (word)err_workspace;
  _err_workspace_bottom[1] = NotProcess_p;
  _err_workspace_bottom[2] = (word)hard_print;
  /* this is where scr_workspace points to */
  _err_workspace_bottom[3] = 0;
  _err_workspace_bottom[4] = (word)&err_chan;
  _err_workspace_bottom[5] = 0;

  /* now initialise channel words.  This was not possible before because the */
  /* err and scr workspace addresses were not known until just now... */
  /* of course this could all be done statically in assembler... */
  kbd_chan = NotProcess_p;
  scr_chan = scr_workspace;
  err_chan = err_workspace;
}


