/*
 *	Transputer style kernel in mixed C/asm
 *	Copyright (C) 1998 Jim Moores
 *	Based on the KRoC/sparc kernel Copyright (C) 1994-2000 D.C. Wood and P.H. Welch
 *	Portions copyright (C) 1999, 2000 Fred Barnes
 *
 *	This program is free software; you can redistribute it and/or modify
 *	it under the terms of the GNU General Public License as published by
 *	the Free Software Foundation; either version 2 of the License, or
 *	(at your option) any later version.
 *
 *	This program is distributed in the hope that it will be useful,
 *	but WITHOUT ANY WARRANTY; without even the implied warranty of
 *	MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *	GNU General Public License for more details.
 *
 *	You should have received a copy of the GNU General Public License
 *	along with this program; if not, write to the Free Software
 *	Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 */

#include <stdio.h>
#include <sys/types.h>

#include "sched_types.h"
#include "sched_consts.h"
#include "sched_asm_inserts.h"
#include "occam_io.h"
#include "timer.h"
#include "csplib_consts.h"
#include "csplib_types.h"

#ifndef RTS_H
	#define RTS_H
	#include "rts.h"
#endif

#include "deadlock.h"
#ifdef BLOCKING_SYSCALLS
	#include <setjmp.h>
	#include "bsyscalls.h"
#endif

#include "dynproc.h"
#include "atomics.h"

typedef void (*FUNCPTR)();

#include "debug.h"

/* variables for global spin lock SMP support */
#ifdef GSL_SMP
	word global_spin_lock = 0;
	word wait_lock = 0;
#endif


/* PUBLIC variables that are used elsewhere in the run time system. */

/* this has to go here, because it refers to current_max_priority */
#include "queues.h"

#ifdef ENABLE_CPU_TIMERS
unsigned long long glob_cpumhz;
#endif

extern T_norm;

/*{{{  global static variables.*/

sync_type sf;
word STATUS=0;

word* Wptr;
word* Tptr;

/* paramter passing variables */
word** channel_address;
word return_address;
byte* source_address;
byte* destination_address;
word count;
Time wait_time;
int test_time;
word start_address;
word start_offset;
word guard;
word process_address;
word code_offset;
word precondition;
word* usage;

Time timeout;
word result;

/* pseudo private variables (there names are externally visible, but they */
/* should not be used. */
Time now; /* must be int to work (signed) */
word temp1;
word *temp_ptr;
word* other_workspace;
word *current;
word *previous;
word fired;
bool isexternal;
word trapval_A, trapval_B, trapval_C;
word overflow_info2, overflow_info, filename_addr, procedure_addr;
word zerodiv_info2, zerodiv_info;
word range_info1, range_info2;
word seterr_info1, seterr_info2;
word b_param, b_func;
word res_ptr;
word kr_param, *kr_wptr, *kr_func;
void (*kr_rfunc)(void);

word *sp_aptr;
word *sp_bptr;
word *sp_cptr;

#ifdef ENABLE_KTRACES
	#define ENTRY_TRACE(X, FMT, args...) \
		fprintf (stderr, "<enter>   "#X" (" FMT ")\n", ##args)
#else
	#define ENTRY_TRACE(X, FMT, args...)
#endif

#ifdef BLOCKING_SYSCALLS
	word			b_return_address;
	extern spl_t		queue_lock;
	extern volatile word	*volatile q_fptr, *volatile q_bptr;
	extern int		num_spare;
	extern int		num_queued;
#endif

/*}}}*/

void X_stopp (); /* forward declaration */

/*{{{  print out timer queue for debugging*/
void print_tq()
{
	word *tqptr = Tptr;
	fprintf(stderr, "Timer Queue dump.\n");
	while ((word)tqptr != NotProcess_p) {
		fprintf(stderr, "(Wptr = 0x%x, Wptr[Time_f] = %d, Wptr[State] = 0x%x)\n", tqptr, tqptr[Time_f], tqptr[State]);
		tqptr = (word*)tqptr[TLink];
	}
	fprintf(stderr, "--- end Timer queue dump.\n");
}
/*}}}*/

/*{{{  kernel panic function*/
void kernel_panic()
{
	fprintf(stderr,"Kernel Panic: called kernel function not implemented.\n");
	exit(1);
}
/*}}}*/

/*{{{  dump_trap_info for dumping on TRAP*/
void dump_trap_info (word a_val, word b_val, word c_val)
{
	int i;

	fprintf (stderr, "** KRoC: TRAP **\n");
	fprintf (stderr, "\tWptr  0x%8.8x    raddr 0x%8.8x\n", (unsigned int)Wptr, (unsigned int)return_address);
	fprintf (stderr, "\tAreg  0x%8.8x    Iptr  0x%8.8x\n", (unsigned int)a_val, (unsigned int)Wptr[Iptr]);
	fprintf (stderr, "\tBreg  0x%8.8x    Creg  0x%8.8x\n", (unsigned int)b_val, (unsigned int)c_val);
	for (i=6; i >= -5; i-=2) {
		fprintf (stderr, "\tWptr[%-2d] @ (0x%8.8x) = 0x%8.8x", i, (unsigned int)&(Wptr[i]), (unsigned int)Wptr[i]);
		fprintf (stderr, "\tWptr[%-2d] @ (0x%8.8x) = 0x%8.8x\n", i-1, (unsigned int)&(Wptr[i-1]), (unsigned int)Wptr[i-1]);
	}
	return;
}
/*}}}  */

/*{{{  divide-by-zero handling function*/
void zerodiv_happened (void)
{
	int line_num, file_num, proc_num;
	char *div_filename, *div_procname;
	int *div_fc, *div_off;
	int *pvr_fc, *pvr_off;

	line_num = zerodiv_info & 0xffff;
	file_num = (zerodiv_info2 >> 16) & 0xffff;
	proc_num = zerodiv_info2 & 0xffff;
	div_fc = (int *)filename_addr;
	if (file_num > *div_fc) {
		fprintf (stderr, "KRoC: filename table inconsistent!\n");
		file_num = -1;
	} else {
		div_off = div_fc + (file_num + 1);
		div_filename = (char *)filename_addr + *div_off;
	}
	pvr_fc = (int *)procedure_addr;
	if (proc_num > *pvr_fc) {
		fprintf (stderr, "KRoC: procedure table inconsistent!\n");
		proc_num = -1;
	} else {
		pvr_off = pvr_fc + (proc_num + 1);
		div_procname = (char *)procedure_addr + *pvr_off;
	}
	fprintf (stderr, "KRoC: divide-by-zero in PROC \"%s\" in file \"%s\" near line %d\n", (proc_num >= 0) ? div_procname : "<unknown>",
		(file_num >= 0) ? div_filename : "<unknown>", line_num);
	return;
}
/*}}}  */

/*{{{  overflow handling function*/
void overflow_happened (void)
{
	static char *overflow_ops[] = { "INVALID", "ADD", "SUB", "MUL", "DIV", "REM", "LADD", "LSUB", "ADC", "STOP" };
	static int num_overflow_ops = 10;
	int ovr_idx, line_num, file_num, proc_num;
	char *ovr_filename, *ovr_procname;
	int *ovr_fc, *ovr_off;
	int *pvr_fc, *pvr_off;

	ovr_idx = (overflow_info >> 24) & 0xff;
	line_num = overflow_info & 0xffff;
	file_num = (overflow_info2 >> 16) & 0xffff;
	proc_num = overflow_info2 & 0xffff;
	if ((ovr_idx >= num_overflow_ops) || (ovr_idx < 0)) {
		ovr_idx = 0;
	}
	ovr_fc = (int *)filename_addr;
	if (file_num > *ovr_fc) {
		fprintf (stderr, "KRoC: filename table inconsistent!\n");
		file_num = -1;
	} else {
		ovr_off = ovr_fc + (file_num + 1);
		ovr_filename = (char *)filename_addr + *ovr_off;
	}
	pvr_fc = (int *)procedure_addr;
	if (proc_num > *pvr_fc) {
		fprintf (stderr, "KRoC: procedure table inconsistent!\n");
		proc_num = -1;
	} else {
		pvr_off = pvr_fc + (proc_num + 1);
		ovr_procname = (char *)procedure_addr + *pvr_off;
	}
	fprintf (stderr, "KRoC: overflow on operation %s in PROC \"%s\" in file \"%s\" near line %d\n",
		overflow_ops[ovr_idx], (proc_num >= 0) ? ovr_procname : "<unknown>", (file_num >= 0) ? ovr_filename : "<unknown>", line_num);
	return;
}
/*}}}*/


/*{{{  range error handling */
void handle_range_error (void)
{
	static char *rangerr_ops[] = { "INVALID", "SHIFT", "CSNGL", "CSUB0", "CCNT1" };
	static int num_rangerr_ops = 5;
	int range_op, line_num, file_num, proc_num;
	char *range_file, *range_proc;
	int rt_bits;

	range_op = (range_info1 >> 24) & 0xff;
	rt_bits = (range_info1 >> 16) & 0xff;
	line_num = range_info1 & 0xffff;
	if (rt_bits != 0xff) {
		fprintf (stderr, "KRoC: Range error (debug data incorrect - rt_bits=%x)\n", rt_bits);
		return;
	}
	if ((range_op >= num_rangerr_ops) || (range_op < 0)) {
		fprintf (stderr, "KRoC: Range error (debug data incorrect - range_op=%d)\n", range_op);
		return;
	}
	file_num = (range_info2 >> 16) & 0xffff;
	proc_num = range_info2 & 0xffff;
	range_file = (char *)filename_addr;
	range_proc = (char *)procedure_addr;
	if ((file_num >= *(int *)(range_file)) || (file_num < 0)) {
		fprintf (stderr, "KRoC: Range error (debug data incorrect - file_num=%d)\n", file_num);
		return;
	}
	if ((proc_num >= *(int *)(range_proc)) || (proc_num < 0)) {
		fprintf (stderr, "KRoC: Range error (debug data incorrect - proc_num=%d)\n", proc_num);
		return;
	}
	range_file += *(int *)(range_file + (4 * (file_num+1)));
	range_proc += *(int *)(range_proc + (4 * (proc_num+1)));
	fprintf (stderr, "KRoC: Range error on operation %s in PROC \"%s\" in file \"%s\" near line %d\n",
		rangerr_ops[range_op], range_proc, range_file, line_num);
	return;
}
/*}}}*/


/*{{{  seterr error handling */
void handle_seterr (void)
{
	int line_num, file_num, proc_num;
	char *seterr_file, *seterr_proc;
	int rt_bits;

	rt_bits = (seterr_info1 >> 16) & 0xffff;
	line_num = seterr_info1 & 0xffff;
	if (rt_bits != 0xfb00) {
		fprintf (stderr, "KRoC: SetErr: application level error (debug data incorrect - rt_bits=%4.4x)\n", rt_bits);
		return;
	}
	file_num = (seterr_info2 >> 16) & 0xffff;
	proc_num = seterr_info2 & 0xffff;
	seterr_file = (char *)filename_addr;
	seterr_proc = (char *)procedure_addr;
	if ((file_num >= *(int *)(seterr_file)) || (file_num < 0)) {
		fprintf (stderr, "KRoC: SetErr: application level error (debug data incorrect - file_num=%d)\n", file_num);
		return;
	}
	if ((proc_num >= *(int *)(seterr_proc)) || (proc_num < 0)) {
		fprintf (stderr, "KRoC: SetErr: application level error (debug data incorrect - proc_num=%d)\n", proc_num);
		return;
	}
	seterr_file += *(int *)(seterr_file + (4 * (file_num+1)));
	seterr_proc += *(int *)(seterr_proc + (4 * (proc_num+1)));
	fprintf (stderr, "KRoC: SetErr: application level error in PROC \"%s\", in file \"%s\" near line %d\n",
		seterr_proc, seterr_file, line_num);
	return;
}


/*{{{  err_no_resources*/
void err_no_resources (word *wptr, int ra)
{
	fprintf (stderr, "KRoC: RESOURCE support not enabled in this kernel. (Wptr=%8lx)\n", wptr);
	occam_exit (1, ra);
	return;
}
/*}}}*/

/*{{{  err_no_bsyscalls*/
void err_no_bsyscalls (word *wptr, int ra)
{
	fprintf (stderr, "KRoC: blocking system calls not enabled in this kernel. (Wptr=%8lx)\n", wptr);
	occam_exit (1, ra);
	return;
}
/*}}}*/


word jump_table[40];
word *occam_table;
word occam_taddr;


/*
 *	void *kernel (int mostly_zero)
 *	setup jump table
 */
void *kernel (int mostly_zero)
{
	int num_entries = 38;

	word err_stop = (word)kernel_panic;
	/* dynamic process loading */
	#ifdef DYNAMIC_PROCS
		K_LOADLABADDR (X_kernel_run, occam_taddr);
		jump_table[num_entries-37] = occam_taddr;
	#else
		K_LOADLABADDR (X_no_dynprocs, occam_taddr);
		jump_table[num_entries-37] = occam_taddr;
	#endif
	/* division by zero */
	K_LOADLABADDR (X_zero_div, occam_taddr);
	jump_table[num_entries-36] = occam_taddr;
	#ifdef RESOURCES
		K_LOADLABADDR (X_res_claim, occam_taddr);
		jump_table[num_entries-35] = occam_taddr;
		K_LOADLABADDR (X_res_release, occam_taddr);
		jump_table[num_entries-34] = occam_taddr;
	#else
		K_LOADLABADDR (X_no_resources, occam_taddr);
		jump_table[num_entries-35] = occam_taddr;
		K_LOADLABADDR (X_no_resources, occam_taddr);
		jump_table[num_entries-34] = occam_taddr;
	#endif
	#ifdef BLOCKING_SYSCALLS
		K_LOADLABADDR (X_bx_dispatch, occam_taddr);
		jump_table[num_entries-33] = occam_taddr;
		K_LOADLABADDR (X_b_dispatch, occam_taddr);
		jump_table[num_entries-32] = occam_taddr;
	#else
		K_LOADLABADDR (X_no_bsyscalls, occam_taddr);
		jump_table[num_entries-33] = occam_taddr;
		K_LOADLABADDR (X_no_bsyscalls, occam_taddr);
		jump_table[num_entries-32] = occam_taddr;
	#endif
	/* without CCSP, we free up some spaces */
	K_LOADLABADDR (X_BSeterr, occam_taddr);
	jump_table[num_entries-31] = occam_taddr;
	K_LOADLABADDR (X_BasicRangeError, occam_taddr);
	jump_table[num_entries-30] = occam_taddr;
	K_LOADLABADDR (X_trap, occam_taddr);
	jump_table[num_entries-29] = occam_taddr;
	K_LOADLABADDR (X_overflow, occam_taddr);
	jump_table[num_entries-28] = occam_taddr;
	/* default occam table continues */
	K_LOADLABADDR (X_stlf, occam_taddr);
	jump_table[num_entries-27] = occam_taddr;
	K_LOADLABADDR (X_stlb, occam_taddr);
	jump_table[num_entries-26] = occam_taddr;
	K_LOADLABADDR (X_savel, occam_taddr);
	jump_table[num_entries-25] = occam_taddr;
	K_LOADLABADDR (X_scheduler, occam_taddr);
	jump_table[num_entries-24] = occam_taddr;
/*	jump_table[num_entries-23] = (word)&&X_C_startp;*/	/* startp for C */
/*	jump_table[num_entries-22] = (word)&&X_C_endp;	*/	/* stopp for C */
	jump_table[num_entries-23] = 0;
	jump_table[num_entries-22] = 0;
	K_LOADLABADDR (X_norm, occam_taddr);
	jump_table[num_entries-21] = occam_taddr;
	K_LOADLABADDR (X_Seterr, occam_taddr);
	jump_table[num_entries-20] = occam_taddr;
	K_LOADLABADDR (X_RangeCheckError, occam_taddr);
	jump_table[num_entries-19] = occam_taddr;
	K_LOADLABADDR (X_dist, occam_taddr);
	jump_table[num_entries-18] = occam_taddr;
	K_LOADLABADDR (X_disc, occam_taddr);
	jump_table[num_entries-17] = occam_taddr;
	K_LOADLABADDR (X_diss, occam_taddr);
	jump_table[num_entries-16] = occam_taddr;
	K_LOADLABADDR (X_enbt, occam_taddr);
	jump_table[num_entries-15] = occam_taddr;
	K_LOADLABADDR (X_enbc, occam_taddr);
	jump_table[num_entries-14] = occam_taddr;
	K_LOADLABADDR (X_taltwt, occam_taddr);
	jump_table[num_entries-13] = occam_taddr;
	K_LOADLABADDR (X_altwt, occam_taddr);
	jump_table[num_entries-12] = occam_taddr;
	K_LOADLABADDR (X_tin, occam_taddr);
	jump_table[num_entries-11] = occam_taddr;
	K_LOADLABADDR (X_ldtimer, occam_taddr);
	jump_table[num_entries-10] = occam_taddr;
	K_LOADLABADDR (X_pause, occam_taddr);
	jump_table[num_entries-9] = occam_taddr;
	K_LOADLABADDR (X_stopp, occam_taddr);
	jump_table[num_entries-8] = occam_taddr;
	K_LOADLABADDR (X_endp, occam_taddr);
	jump_table[num_entries-7] = occam_taddr;
	K_LOADLABADDR (X_startp, occam_taddr);
	jump_table[num_entries-6] = occam_taddr;
	K_LOADLABADDR (X_runp, occam_taddr);
	jump_table[num_entries-5] = occam_taddr;
	K_LOADLABADDR (X_outbyte, occam_taddr);
	jump_table[num_entries-4] = occam_taddr;
	K_LOADLABADDR (X_outword, occam_taddr);
	jump_table[num_entries-3] = occam_taddr;
	K_LOADLABADDR (X_out, occam_taddr);
	jump_table[num_entries-2] = occam_taddr;
	K_LOADLABADDR (X_in, occam_taddr);
	jump_table[num_entries-1] = occam_taddr;
	jump_table[num_entries] = num_entries;			/* number of entries */
	occam_table = &(jump_table[num_entries]);
	K_LOADLABADDR (hard_print, occam_taddr);
	return (void *)occam_taddr;
}


/*
 *	void kernel_X_zero_div (void)
 *	entry point for integer division-by-zero
 */
void kernel_X_zero_div (void)
{
	K_SETLABEL (X_zero_div);
	LOADWPTR;
	LOADRETURN;
	FOURPARAMS(zerodiv_info2, zerodiv_info, procedure_addr, filename_addr);
	ENTRY_TRACE (X_zero_div, "%x, %x, %p, %p", zerodiv_info2, zerodiv_info, procedure_addr, filename_addr);
	zerodiv_happened();
	occam_exit (1, return_address);
	RESTORE;
	EJUMP(return_address);
}


/*
 * Here are the necessary kernel calls for RESOURCE/CLAIM in the occam world
 * This is *NOT* SMP safe, but could be made SMP safe reasonably easily.
 */
#ifdef RESOURCES
	/*
	 *	void kernel_X_res_claim (void)
	 *	claims a resource
	 */
	void kernel_X_res_claim (void)
	{
		K_SETLABEL (X_res_claim);
		LOADWPTR;
		LOADRETURN;
		PARAM(res_ptr);
		ENTRY_TRACE (X_res_claim, "%p", res_ptr);
		/* Structure of *res_ptr is { int Fptr, int Bptr, int held } */
		{
			int *tptr = (int *)res_ptr;
			if (tptr[2] & 1) {
				/* resource already held -- block */
				if (tptr[1]) {
					((word *)(tptr[1]))[Link] = (word)Wptr;
					tptr[1] = (int)Wptr;
				} else {
					tptr[0] = tptr[1] = (int)Wptr;
				}
				K_JUMP (X_scheduler);
			}
			tptr[2] |= 1;
		}
		RESTORE;
		EJUMP(return_address);
	}


	/*
	 *	void kernel_X_res_release (void)
	 *	releases a resource
	 */
	void kernel_X_res_release (void)
	{
		K_SETLABEL (X_res_release);
		LOADWPTR;
		LOADRETURN;
		PARAM(res_ptr);
		ENTRY_TRACE (X_res_release, "%p", res_ptr);
		QUEUE(Wptr);
		Wptr[Iptr] = return_address;
		{
			int *tptr = (int *)res_ptr;
			if (tptr[0]) {
				QUEUE((word *)(tptr[0]));
				if (tptr[0] == tptr[1]) {
					tptr[0] = 0;
				} else {
					tptr[0] = (int)(((word *)(tptr[0]))[Link]);
				}
			} else {
				tptr[2] &= ~1;
			}
		}
		K_JUMP(X_scheduler);
	}
#else
	/*
	 *	void kernel_X_no_resources (void)
	 *	no-resources entry point
	 */
	void kernel_X_no_resources (void)
	{
		K_SETLABEL (X_no_resources);
		LOADWPTR;
		LOADRETURN;
		ENTRY_TRACE (X_no_resources, "");
		err_no_resources (Wptr, return_address);
	}
#endif

/*
 * Below are the two interfaces for handling blocking system calls
 */
#ifdef BLOCKING_SYSCALLS
	/*
	 *	void kernel_X_b_dispatch (void)
	 *	dispatches a blocking call
	 */
	void kernel_X_b_dispatch (void)
	{
		K_SETLABEL (X_b_dispatch);
		LOADWPTR;
		LOADRETURN;
		TWOPARAMS(b_func, b_param);
		ENTRY_TRACE (X_b_dispatch, "%p, %p", b_func, b_param);
		/* adjust workspace for return */
		#ifdef BLOCKING_DEBUG
			fprintf (stderr, "X_b_dispatch: entry: return_address = %p, b_param = %p, b_func = %p, WPtr = %p\n",
				return_address, b_param, b_func, Wptr);
		#endif
		/* (char *)Wptr = (char *)Wptr + 16; */
		#ifdef BLOCKING_DEBUG
		{
			int dummy, i;
			&dummy;
			fprintf (stderr, "X_b_dispatch: return_address = %p, b_param = %p, b_func = %p, Wptr = %p\n",
				return_address, b_param, b_func, Wptr);
			for (i=6; i >= -5; i-=2) {
				fprintf (stderr, "\tWptr[%-2d] @ (0x%8.8x) = 0x%8.8x", i, (unsigned int)&(Wptr[i]), (unsigned int)Wptr[i]);
				fprintf (stderr, "\tWptr[%-2d] @ (0x%8.8x) = 0x%8.8x\n", i-1, (unsigned int)&(Wptr[i-1]), (unsigned int)Wptr[i-1]);
			}
		}
		#endif
		bsyscalls_dispatch ((word *)Wptr, (char *)return_address, (word *)b_param, (void (*)(word *))b_func, 0);
		K_JUMP(X_scheduler);
	}


	/*
	 *	void kernel_X_bx_dispatch (void)
	 *	dispatches a terminateable blocking call
	 */
	void kernel_X_bx_dispatch (void)
	{
		K_SETLABEL(X_bx_dispatch);
		LOADWPTR;
		LOADRETURN;
		/* (char *)Wptr = (char *)Wptr + 16; */
		TWOPARAMS(b_func, b_param);
		ENTRY_TRACE (X_bx_dispatch, "%p, %p", b_func, b_param);
		#ifdef BLOCKING_DEBUG
		{
			int dummy;
			&dummy;
			fprintf (stderr, "X_bx_dispatch: return_address = %p, Wptr[Iptr] = %p, b_param = %p, b_func = %p, Wptr = %p\n",
				return_address, Wptr[Iptr], b_param, b_func, Wptr);
		}
		#endif
		bsyscalls_dispatch ((word *)Wptr, (char *)return_address, (word *)b_param, (void (*)(word *))b_func, 4);
		K_JUMP(X_scheduler);
	}
#else
	/*
	 *	void kernel_X_no_bsyscalls (void)
	 *	no blocking-system calls functionality entry-point
	 */
	void kernel_X_no_bsyscalls (void)
	{
		LBL(X_no_bsyscalls);
		LOADWPTR;
		LOADRETURN;
		ENTRY_TRACE (X_no_bsyscalls, "");
		err_no_bsyscalls (Wptr, return_address);
	}
#endif	/* BLOCKING_SYSCALLS */

/*
 * for the dynamic proc stuff
 */
#ifdef	DYNAMIC_PROCS
	/*
	 *	void kernel_X_kernel_run (void)
	 *	dynamic kernel run entry point
	 */
	void kernel_X_kernel_run (void)
	{
		K_SETLABEL(X_kernel_run);
		LOADWPTR;
		LOADRETURN;
		PARAM(kr_param);
		ENTRY_TRACE (X_kernel_run, "%p", kr_param);
		kr_wptr = load_dynamic_process ((int *)kr_param, &kr_func);
		if (!kr_wptr) {
			fprintf (stderr, "X_kernel_run: error loading dynamic process\n");
			RESTORE;
			EJUMP(return_address);
		}
		/* assume for now */
		kr_wptr[8] = (word)Wptr;
		kr_wptr[9] = return_address;
		Wptr = kr_wptr;
		#ifdef DEBUG_DYNAMIC_PROCS
			fprintf (stderr, "X_kernel_run: set Wptr to %p, jump to %p\n", Wptr, kr_func);
		#endif
		RESTORE;
		kr_rfunc = (void (*)(void))kr_func;
		kr_rfunc();
		/* schedule for a bit */
		#ifdef DEBUG_DYNAMIC_PROCS
			fprintf (stderr, "X_kernel_run: back in occam world!\n");
		#endif
		Wptr = (word *)((char *)Wptr + 16);		/* re-adjust Wptr */
		return_address = Wptr[9];
		kr_wptr = (word *)(Wptr[7]);
		Wptr = (word *)(Wptr[8]);			/* restore old Wptr */
		free (kr_wptr);					/* free memory used by call */
		RESTORE;
		EJUMP(return_address);
	}
#else	/* !DYNAMIC_PROCS */

	/*
	 *	void kernel_no_dynprocs (void)
	 *	no dynamic process support entry point
	 */
	void kernel_no_dynprocs (void)
	{
		K_SETLABEL (X_no_dynprocs);
		LOADWPTR;
		LOADRETURN;
		PARAM (kr_param);
		ENTRY_TRACE (X_no_dynprocs, "%p", kr_param);
		no_dynamic_process_support ();
		RESTORE;
		EJUMP (return_address);
	}
#endif	/* !DYNAMIC_PROCS */

/* This function is used by the new synchronisation primitives in occam
 * (BUCKETs, EVENTs etc).
 */


/*
 *	void kernel_X_savel (void)
 *	retreive queue pointers
 */
void kernel_X_savel (void)
{
	K_SETLABEL (X_savel);
	#ifdef ENABLE_EXPKI
		ONE_PARAM_IN (temp_ptr);
		LOADWPTR;
		LOADRETURN;
	#else
		LOADWPTR;
		LOADRETURN;
		PARAM(temp_ptr);
	#endif
	ENTRY_TRACE (X_savel, "%p", temp_ptr);
	/*
	 * so that MINT, DIFF and CJ work properly
	 */
	temp_ptr[1] = (word)Bptr;
	temp_ptr[0] = Fptr ? (word)Fptr : 0x80000000;
	DBGMSG( "savel: temp_ptr[0] (Fptr) = 0x%lx, temp_ptr[1] (Bptr) = 0x%lx\n", temp_ptr[0], temp_ptr[1] );
	RESTORE;
	EJUMP(return_address);
}


/*
 *	void kernel_X_trap (void)
 *	trap entry-point
 */
void kernel_X_trap (void)
{
	K_SETLABEL (X_trap);
	#ifdef ENABLE_EXPKI
		THREE_PARAM_IN (trapval_A, trapval_B, trapval_C);
		LOADWPTR;
		LOADRETURN;
	#else
		LOADWPTR;
		LOADRETURN;
		THREEPARAMS(trapval_A,trapval_B,trapval_C);
	#endif
	ENTRY_TRACE (X_trap, "0x%x, 0x%x, 0x%x", trapval_A, trapval_B, trapval_C);
	dump_trap_info (trapval_A, trapval_B, trapval_C);
	#ifdef ENABLE_EXPKI
		RESTORE;
		THREE_PARAM_OUT (trapval_A, trapval_B, trapval_C);
	#else
		RETURNVAL(trapval_C);
		RETURNVAL(trapval_B);
		RETURNVAL(trapval_A);
		RESTORE;
	#endif
	EJUMP(return_address);
}


/*
 *	void kernel_X_overflow (void)
 *	arithmetic overflow entry point
 */
void kernel_X_overflow (void)
{
	K_SETLABEL (X_overflow);
	#ifdef ENABLE_EXPKI
		FOUR_PARAM_IN(overflow_info2, overflow_info, procedure_addr, filename_addr);
		LOADWPTR;
		LOADRETURN;
	#else
		LOADWPTR;
		LOADRETURN;
		FOURPARAMS(overflow_info2, overflow_info, procedure_addr, filename_addr);
	#endif
	ENTRY_TRACE (X_overflow, "%x, %x, %p, %p", overflow_info2, overflow_info, procedure_addr, filename_addr);
	overflow_happened();		/* Parameters dealt with (they're visible throughout) */
	occam_exit (1, return_address);
	RESTORE;
	EJUMP(return_address);
}


/*
 *	void kernel_X_stlf (void)
 *	store Fptr for low-pri queue
 */
void kernel_X_stlf (void)
{
	K_SETLABEL (X_stlf);
	#ifdef ENABLE_EXPKI
		ONE_PARAM_IN(Fptr);
		LOADWPTR;
		LOADRETURN;
	#else
		LOADWPTR;
		LOADRETURN;
		PARAM(Fptr);
	#endif
	ENTRY_TRACE (X_stlf, "%p", Fptr);
	/*
	 * See above.
	 */
	if( (word)Fptr == (word)0x80000000 ) {
		Fptr = 0;
	}
	DBGMSG( "stlf: Fptr set to 0x%lx\n", Fptr );
	RESTORE;
	EJUMP(return_address);
}


/*
 *	void kernel_X_stlb (void)
 *	stoare Bptr for low-pri queue
 */
void kernel_X_stlb (void)
{
	K_SETLABEL (X_stlb);
	#ifdef ENABLE_EXPKI
		ONE_PARAM(Bptr);
		LOADWPTR;
		LOADRETURN;
	#else
		LOADWPTR;
		LOADRETURN;
		PARAM(Bptr);
	#endif
	ENTRY_TRACE (X_stlb, "%p", Bptr);
	DBGMSG( "stlb: Bptr set to 0x%lx\n", Bptr );
	RESTORE;
	EJUMP(return_address);
}


/*
 *	void kernel_X_norm (void)
 *	normalises 64-bit floating point ?
 */
void kernel_X_norm (void)
{
	K_SETLABEL (X_norm);
	ENTRY_TRACE (X_norm, "");
	asm("
		.globl _T_norm
	_T_Norm:
		pop	%edi
		pop	%eax
		pop	%edx
		movl	$0x1F, %ecx
		bsrl	%edx, %ebx
		jne	shiftem
		movl	$0x3F, %ecx
		bsrl	%eax, %ebx
		jne	longshift
		movl	$0x40, %ecx
		jmp	pushem
	longshift:
		movl	%eax, %edx
		xorl	%eax, %eax
		subl	%ebx, %ecx
		shll	%cl, %edx
		jmp	pushem
	shiftem:
		subl	%ebx, %ecx
		shldl	%cl, %eax, %edx
		shll	%cl, %eax
	pushem:
		push	%ecx
		push	%edx
		push	%eax
		jmp	*%edi
	");
	/* The implementation of normalise
	LOADWPTR;
	LOADRETURN;
	*/
	fprintf(stderr, "Norm: not implemented, exiting.\n");
	user_bad_exit();
}


/*
 *	void kernel_X_Seterr (void)
 *	SETERR entry-point
 */
void kernel_X_Seterr (void)
{
	K_SETLABEL (X_Seterr);
	LOADWPTR;
	LOADRETURN;
	FOURPARAMS (procedure_addr, filename_addr, seterr_info2, seterr_info1);
	ENTRY_TRACE (X_Seterr, "%p, %p, %x, %x", procedure_addr, filename_addr, seterr_info2, seterr_info1);
	handle_seterr ();
	occam_exit (1, return_address);
	user_bad_exit();
}


/*
 *	void kernel_X_BSeterr (void)
 *	basic SETERR entry-point
 */
void kernel_X_BSeterr (void)
{
	K_SETLABEL (X_BSeterr);
	LOADWPTR;
	LOADRETURN;
	ENTRY_TRACE (X_BSeterr, "");
	fprintf(stderr, "SetErr: Application level error at 0x%x, Wptr = 0x%x.\n", return_address, Wptr);
	user_bad_exit();
}


/*
 *	void kernel_X_RangeCheckError (void)
 *	Range error entry-point
 */
void kernel_X_RangeCheckError (void)
{
	K_SETLABEL (X_RangeCheckError);
	LOADWPTR;
	LOADRETURN;
	FOURPARAMS (procedure_addr, filename_addr, range_info2, range_info1);
	ENTRY_TRACE (X_RangeCheckError, "%p, %p, %x, %x", procedure_addr, filename_addr, range_info2, range_info1);
	handle_range_error ();
	occam_exit (1, return_address);
	RESTORE;
	EJUMP (return_address);
}


/*
 *	void kernel_X_BasicRangeError (void)
 *	basic range error entry point
 */
void kernel_X_BasicRangeError (void)
{
	K_SETLABEL (X_BasicRangeError);
	LOADWPTR;
	LOADRETURN;
	ENTRY_TRACE (X_BasicRangeError, "");
	fprintf (stderr, "KRoC: Range error.\n");
	occam_exit (1, Wptr[Iptr]);
	raise (SIGSEGV);
	RESTORE;
	EJUMP (return_address);
}


/*
 *	void kernel_hard_print (void)
 *	does "hard-printing" -- bit gross
 */
void kernel_hard_print (void)
{
	K_SETLABEL (hard_print);
	while (true) {
		LOADWPTR;

		if (Wptr[0]  == Flush) {
			fflush((FILE *)Wptr[Device]);
		} else {
			fputc((int)(Wptr[0]), (FILE *)Wptr[Device]);
		}
		/* calling a kernel function like this is a fudge */
		PUSH(&(Wptr[0]));
		PUSH(Wptr[IOChannel]);
		PUSH(1);
		RESTORE;
		((FUNCPTR)(occam_table[Table_In])) ();
	}
}


/*
 *	void kernel_kbd_process (void)
 *	the keyboard process
 */
void kernel_kbd_process (void)
{
	K_SETLABEL (kbd_process);
	DBGMSG("kbd_process: started.  Calling read_the_kbd.\n");
	temp1 = read_the_kbd();
	Wptr = kbd_workspace;
	DBGMSG("kbd_process: received keypress as 0x%x. Calling OutByte.\n", temp1);
	PUSH(&kbd_chan);
	PUSH(temp1);
	RESTORE; /* get Wptr into ebp */
	((FUNCPTR)(occam_table[Table_OutByte]))();
	DBGMSG("kbd_process: returned from OutByte, acknowledging keypress.\n");
	acknowledge();
	DBGMSG("kbd_process: calling scheduler.\n");
	K_JUMP(X_scheduler);
}


/*
 *	void kernel_X_outbyte (void)
 *	output a byte
 */
void kernel_X_outbyte (void)
{
	K_SETLABEL (X_outbyte);
	#ifdef ENABLE_EXPKI
		TWO_PARAM_IN(Wptr[0], channel_address);
		LOADWPTR;
		LOADRETURN;
	#else
		LOADWPTR;
		LOADRETURN;
		TWOPARAMS(Wptr[0], channel_address);
	#endif
	ENTRY_TRACE (X_outbyte, "%x, %p", Wptr[0], channel_address);
	source_address = (byte *)&(Wptr[0]);
	count = 1;
	/*}}}*/
	/*{{{  out (copy of)*/
	/*{{{  debug*/
	DBGMSG("out: channel address = 0x%x\n" \
		"     count = %d\n" \
		"     source = 0x%x\n" \
		"     return_address = 0x%x\n", \
		channel_address, count, source_address, \
		return_address);
	/*}}}*/
	if (*channel_address == NotProcess_p) {
		*channel_address = Wptr;
		Wptr[Pointer] = (word)source_address;
		Wptr[Iptr] = return_address; /* save return address in workspace. */
		K_JUMP(X_scheduler);
	} else {
		temp_ptr = (word *)(((word)(*channel_address)) & (~1));
		DBGMSG("out: other_process[State] = 0x%x", temp_ptr[State]);

		/* mask off priority bit */
		if (temp_ptr[State] <= Ready_p) {

			/*{{{  debug*/
			DBGMSG( "out: channel state is something ALTy.\n");
			/*}}}*/
			if (temp_ptr[State] == Waiting_p) {
				QUEUE(temp_ptr);
			}
			temp_ptr[State] = Ready_p;
			*channel_address = Wptr;
			Wptr[Pointer] = (word)source_address;
			Wptr[Iptr] = return_address; /* save return address in workspace. */
			K_JUMP(X_scheduler);
		} else {
			*channel_address = (word *)(((word)(*channel_address)) & (~1));
			destination_address = (byte *)((*channel_address)[Pointer]);
			/* copy... */
			/* put first process on back of queue */
			DBGMSG("in: queuing this process and other process.\n");
			/*{{{  queue both processes and get a new one from run queue*/
			QUEUE(*channel_address);
			Wptr[Iptr] = return_address; /* save this processes Iptr */
			QUEUE(Wptr);
			Wptr = GetNextProcess_NonEmptyQ();
			return_address = Wptr[Iptr]; /* get new processes old Iptr */
			/*}}}*/

			/*{{{  do copy and exit.  (copied above for optimisation)*/
			*channel_address = NotProcess_p; /* same as NULL */
			/* BLOCK COPY */
			#ifdef DONT_USE_MEMCPY
				/*{{{  do the copy with a while loop*/
				while (count--) {
					*(destination_address++) = *(source_address++);
				}
				/*}}}*/
			#else
				/*{{{  use memcopy (custom version)*/
				MEMCPY(source_address, destination_address, count);
				/*}}}*/
			#endif
			/* set up ebp = Wptr and esi = occam_table for occam */
			RESTORE;
			EJUMP(return_address);
			/*}}}*/
		}
	}
}


/*
 *	void kernel_X_outword (void)
 *	output a word
 */
void kernel_X_outword (void)
{
	K_SETLABEL (X_outword);
	#ifdef ENABLE_EXPKI
		TWO_PARAM_IN (Wptr[0], channel_address);
		LOADWPTR;
		LOADRETURN;
	#else
		LOADWPTR;
		LOADRETURN;
		TWOPARAMS(Wptr[0], channel_address);
	#endif
	ENTRY_TRACE (X_outword, "%x, %p", Wptr[0], channel_address);
	source_address = (byte *)&(Wptr[0]);
	count = 4;
	/*}}}*/
	/*{{{  out (copy of)*/
	/*{{{  debug*/
	DBGMSG("out: channel address = 0x%x\n" \
		"     count = %d\n" \
		"     source = 0x%x\n" \
		"     return_address = 0x%x\n", \
		channel_address, count, source_address, \
		return_address);
	/*}}}*/
	if (*channel_address == NotProcess_p) {
		*channel_address = Wptr;
		Wptr[Pointer] = (word)source_address;
		Wptr[Iptr] = return_address; /* save return address in workspace. */
		K_JUMP(X_scheduler);
	} else {
		temp_ptr = (word *)(((word)(*channel_address)) & (~1));
		DBGMSG("out: other_process[State] = 0x%x", temp_ptr[State]);
		/* mask off priority bit */
		if (temp_ptr[State] <= Ready_p) {
			/*{{{  debug*/
			DBGMSG( "out: channel state is something ALTy.\n");
			/*}}}*/
			if (temp_ptr[State] == Waiting_p) {
				QUEUE(temp_ptr);
			}
			temp_ptr[State] = Ready_p;
			*channel_address = Wptr;
			Wptr[Pointer] = (word)source_address;
			Wptr[Iptr] = return_address; /* save return address in workspace. */
			K_JUMP(X_scheduler);
		} else {
			*channel_address = (word *)(((word)(*channel_address)) & (~1));
			destination_address = (byte *)((*channel_address)[Pointer]);
			/* copy... */
			/* put first process on back of queue */
			DBGMSG("in: queuing this process and other process.\n");
			/*{{{  queue both processes and get a new one from run queue*/
			QUEUE(*channel_address);
			Wptr[Iptr] = return_address; /* save this processes Iptr */
			QUEUE(Wptr);
			Wptr = GetNextProcess_NonEmptyQ();
			return_address = Wptr[Iptr]; /* get new processes old Iptr */
			/*}}}*/

			/*{{{  do copy and exit.  (copied above for optimisation)*/
			*channel_address = NotProcess_p; /* same as NULL */
			/* BLOCK COPY */
			#ifdef DONT_USE_MEMCPY
				/*{{{  do the copy with a while loop*/
				while (count--) {
					*(destination_address++) = *(source_address++);
				}
				/*}}}*/
			#else
				/*{{{  use memcopy (custom version)*/
				MEMCPY(source_address, destination_address, count);
				/*}}}*/
			#endif
			/* set up ebp = Wptr and esi = occam_table for occam */
			RESTORE;
			EJUMP(return_address);
			/*}}}*/
		}
	}
}


/*
 *	void kernel_X_out (void)
 *	output
 */
void kernel_X_out (void)
{
	K_SETLABEL (X_out);
	#ifdef ENABLE_EXPKI
		THREE_PARAM_IN(count, channel_address, source_address);
		LOADWPTR;
		LOADRETURN;
	#else
		LOADWPTR;
		LOADRETURN;
		THREEPARAMS(count, channel_address, source_address);
	#endif
	ENTRY_TRACE (X_out, "%d, %p, %p", count, channel_address, source_address);
	/*{{{  debug*/
	DBGMSG("out: channel address = 0x%x\n" \
		"     count = %d\n" \
		"     source = 0x%x\n" \
		"     return_address = 0x%x\n", \
		channel_address, count, source_address, \
		return_address);
	/*}}}*/
	if (*channel_address == NotProcess_p) {
		*channel_address = Wptr;
		Wptr[Pointer] = (word)source_address;
		Wptr[Iptr] = return_address; /* save return address in workspace. */
		K_JUMP(X_scheduler);
	} else {
		temp_ptr = (word *)(((word)(*channel_address)) & (~1));
		DBGMSG("out: other_process[State] = 0x%x", temp_ptr[State]);
		/* mask off priority bit */
		if (temp_ptr[State] <= Ready_p) {
			/*{{{  debug*/
			DBGMSG( "out: channel state is something ALTy.\n");
			/*}}}*/
			if (temp_ptr[State] == Waiting_p) {
				QUEUE(temp_ptr);
			}
			temp_ptr[State] = Ready_p;
			*channel_address = Wptr;
			Wptr[Pointer] = (word)source_address;
			Wptr[Iptr] = return_address; /* save return address in workspace. */
			K_JUMP(X_scheduler);
		} else {
			*channel_address = (word *)(((word)(*channel_address)) & (~1));
			destination_address = (byte *)((*channel_address)[Pointer]);
			/* copy... */
			/* put first process on back of queue */
			DBGMSG("in: queuing this process and other process.\n");
			/*{{{  queue both processes and get a new one from run queue*/
			QUEUE(*channel_address);
			Wptr[Iptr] = return_address; /* save this processes Iptr */
			QUEUE(Wptr);
			Wptr = GetNextProcess_NonEmptyQ();
			return_address = Wptr[Iptr]; /* get new processes old Iptr */
			/*}}}*/

			/*{{{  do copy and exit.  (copied above for optimisation)*/
			*channel_address = NotProcess_p; /* same as NULL */
			/* BLOCK COPY */
			#ifdef DONT_USE_MEMCPY
				/*{{{  do the copy with a while loop*/
				while (count--) {
					*(destination_address++) = *(source_address++);
				}
				/*}}}*/
			#else
				/*{{{  use memcopy (custom version)*/
				MEMCPY(source_address, destination_address, count);
				/*}}}*/
			#endif
			/* set up ebp = Wptr and esi = occam_table for occam */
			RESTORE;
			EJUMP(return_address);
			/*}}}*/
		}
	}
}


/*
 *	void kernel_X_in (void)
 *	input
 */
void kernel_X_in (void)
{
	K_SETLABEL (X_in);
	#ifdef ENABLE_EXPKI
		THREE_PARAM_IN(count, channel_address, destination_address);
		LOADWPTR;
		LOADRETURN;
	#else
		LOADWPTR;
		LOADRETURN;
		THREEPARAMS(count, channel_address, destination_address);
	#endif
	ENTRY_TRACE (X_in, "%d, %p, %p", count, channel_address, destination_address);
	/*{{{  debug*/
	DBGMSG("in: return address = 0x%x\n" \
		"    count = %d\n" \
		"    channel address = 0x%x\n" \
		"    destination address = 0x%x\n", \
		return_address, count, channel_address, \
		destination_address);
	/*}}}*/
	if (*channel_address == NotProcess_p) {
		*channel_address = Wptr;
		Wptr[Pointer] = (word)destination_address;
		Wptr[Iptr] = return_address; /* save return address in workspace. */
		K_JUMP(X_scheduler);
	} else {
		source_address = (byte*)(*channel_address)[Pointer];
		/* copy ... */
		/* put first process on back of queue */
		DBGMSG("in: queuing this process and other process.\n");
		/*{{{  queue both processes and get a new one from the front of the run q*/
		QUEUE(*channel_address);
		Wptr[Iptr] = return_address;
		QUEUE(Wptr);
		Wptr = GetNextProcess_NonEmptyQ();
		return_address = Wptr[Iptr];
		/*}}}*/
		/*{{{  copy data and return (rememeber there are copies above)*/
		/*{{{  Comment*/
		/* BUG FIX - this didn't used to be done here, but in the queue juggling
		 * stuff above - I therefore accidentally left it out if the channel was
		 * external, which meant that the channel was never reset.
		 * NB: need to be careful of this if doing multiprocessor SMP version
		 */
		/*}}}*/
		*channel_address = NotProcess_p;
		/* BLOCK COPY */
		/*{{{  debug*/
		DBGMSG("in: Source address = 0x%x, Destination address = 0x%x\n", \
			source_address, destination_address);
		/*}}}*/
		#ifdef DONT_USE_MEMCPY
			/*{{{  copy using a while loop*/
			while (count) {
				*(destination_address++) = *(source_address++);
				count--;
			}
			/*}}}*/
		#else
			MEMCPY(source_address, destination_address, count);
		#endif
		/* set up ebp = Wptr and esi = occam_table for occam */
		RESTORE;
		EJUMP(return_address);
		/*}}}*/
	}
}


/*
 *	void kernel_X_pause (void)
 *	called on loop-ends (optionally) to reschedule
 */
void kernel_X_pause (void)
{
	K_SETLABEL (X_pause); /* called on loopend */
	K_JUMP (X_resched);
}


/*
 *	void kernel_X_resched (void)
 *	reschedule
 */
void kernel_X_resched (void)
{
	K_SETLABEL (X_resched);
	LOADWPTR;
	LOADRETURN;
	ENTRY_TRACE (X_resched, "");
	QUEUE(Wptr);
	/* Inlined stopp more efficient */
	Wptr[Iptr] = return_address;
	DBGMSG("X_resched: Wptr = 0x%x.  Entering scheduler.",Wptr);
	/* Jump on */
	K_JUMP(X_scheduler);
}


/*
 *	void kernel_X_stopp
 *	stop process
 */
void kernel_X_stopp (void)
{
	K_SETLABEL (X_stopp);
	LOADWPTR;
	LOADRETURN;
	ENTRY_TRACE (X_stopp, "");
	Wptr[Iptr] = return_address; /* save return address in workspace. */
	K_JUMP(X_scheduler);
}


/*
 *	void kernel_X_scheduler (void)
 *	scheduler proper
 *
 *	This routine cannot be "call'ed" because it doesn't deal with the stack, so
 *	it should never be called from C directly, only from assembler using a jump.
 *	It should not be called externally anyway...
 *	FOOTNOTE: But it is jumped to (now!) from CSPlib in ProcPar().
 *	[CSPlib disabled in this scheduler]
 */
void kernel_X_scheduler (void)
{
	K_SETLABEL (X_scheduler);
	ENTRY_TRACE (X_scheduler, "");
	DBGMSG("scheduler: called with Wptr = 0x%x.\n", Wptr);
	/* current_max_priority may be changed by a subsequent call to
	 * External_HandleExternals() which may reschedule processes of higher
	 * priority than the current processes priority.
	 */
	if (sf.combined_sync!=0) {
		DBGMSG("scheduler: one of the sync flags has been set.\n");
		/* one of the sync flags has been set */
		if (sf.s.kbd_sync) {
			DBGMSG( "scheduler: it was the keyboard sync.\n");
			sf.s.kbd_sync = 0;
			K_JUMP (kbd_process);
		} else {
			if (sf.s.tim_sync) {
				sf.s.tim_sync = 0;
				now = Time_GetTime();
				/* possible problem in following while loop in 2nd expr */
				while ((Tptr != NotProcess_p) && (!(Time_AFTER(GetTimeField(Tptr),now)))) {
					/* move first process on timer queue to the run queue */
					temp_ptr = Tptr;
					Tptr = (word *)Tptr[TLink];
					temp_ptr[TLink] = TimeSet_p;
					GetTimeField(temp_ptr) = now;
					if (temp_ptr[State] != Ready_p) {
						temp_ptr[State] = Ready_p;
						QUEUE(temp_ptr);
					}
				}
				if (Tptr != NotProcess_p) {
					Time_SetAlarm( Time_MINUS(GetTimeField(Tptr), now) );
				} else {
					Time_ResetAlarm();
				}
			} else {
				#ifdef BLOCKING_SYSCALLS
					if (sf.s.block_sync) {
						sf.s.block_sync = 0;
						#ifdef BLOCKING_DEBUG
							fprintf (stderr, "** %s:%d block_sync flag has been set (num_queued=%d, &queue_lock=0x%x).\n", __FILE__, __LINE__, num_queued, (int)&queue_lock);
						#endif
						/* Join in */
						barrier ();
						spl_lock (&queue_lock);
						barrier ();
						if (q_fptr) {
							#ifdef BLOCKING_DEBUG
								fprintf (stderr, "** %s:%d In lock q_fptr=0x%8.8x, q_bptr=0x%8.8x, num_queued=%d\n", __FILE__, __LINE__, (int)q_fptr, (int)q_bptr, num_queued);
							#endif
							QUEUE_VQ (q_fptr, q_bptr);
							q_fptr = NotProcess_p;
							num_spare += num_queued;
							num_queued = 0;
						} else {
							fprintf (stderr, "KRoC: fatal! -- block_sync set, but no processes!\n");
							kernel_panic ();
						}
						barrier ();
						spl_unlock (&queue_lock);
						barrier ();
						#ifdef BLOCKING_DEBUG
							fprintf (stderr, "** %s:%d going to continue running now...\n", __FILE__, __LINE__);
							fprintf (stderr, "** first process is 0x%lx (Iptr = 0x%lx)\n", Fptr, Fptr[Iptr]);
						#endif
					} else {
				#endif
					kernel_panic();
				#ifdef BLOCKING_SYSCALLS
					}
				#endif
			}
		}
	}
	/*{{{  */
	DBGMSG("scheduler: there are processes waiting.\n");
	#ifdef BLOCKING_DEBUG
		fprintf (stderr, "%s:%d: run queue state: [Fptr=0x%x, Bptr=0x%x]\n", __FILE__, __LINE__, (int)Fptr, (int)Bptr);
	#endif
	/*{{{  if (more processes) {*/
	if (NoProcesses()) {
		DBGMSG( "scheduler: Run queue empty.\n");
		/* unfortunately we cannot have deadlock detection with external comms
		 * because we have no way to detect processes waiting on external
		 * channels.
		 */
		if (kbd_chan==NotProcess_p) {
			if (Tptr == NotProcess_p) {
				#ifdef BLOCKING_SYSCALLS
					if (!bsyscalls_pending ()) {
				#endif
					STATUS &= 1;
					kernel_dead(STATUS);
				#ifdef BLOCKING_SYSCALLS
					}
				#endif
			}
		}
		safe_pause();
		K_JUMP (X_scheduler);
	} else {
		/*}}}*/
		Wptr = GetNextProcess();
	}
	return_address = Wptr[Iptr];
	DBGMSG("scheduler: next process continues from 0x%x.\n", return_address);
	#ifdef BLOCKING_DEBUG
		fprintf (stderr, "%s:%d: next process 0x%8x continues from 0x%8x [Fptr=0x%x, Bptr=0x%x]\n", __FILE__, __LINE__, (unsigned int)Wptr, (unsigned int)return_address,
			(unsigned int)Fptr, (unsigned int)Bptr);
	#endif
	RESTORE;
	EJUMP(return_address);
	/*}}}*/
	DBGMSG("scheduler: fallen off end of function!\n");
	exit(1);
	/*}}}*/
}


/*
 *	void kernel_X_endp (void)
 *	end process
 */
void kernel_X_endp (void)
{
	K_SETLABEL (X_endp);
	#ifdef ENABLE_EXPKI
		ONE_PARAM_IN(usage); /* Wptr becomes successor_workspace */
		LOADRETURN; /* junk the return address */
		LOADWPTR;
	#else
		LOADRETURN; /* junk the return address */
		LOADWPTR;
		PARAM(usage); /* Wptr becomes successor_workspace */
	#endif
	ENTRY_TRACE (X_endp, "%p", usage);
	usage[Count]--;
	if (usage[Count]) {
		Wptr = usage;
		K_JUMP(X_scheduler); /* Wptr is workspace so we don't need to set up ebp */
	} else {
		Wptr = usage; /* added to correctly update WPtr 091198 */
		RESTORE;
		EJUMP(usage[IptrSucc]);
	}
}


/*
 *	void kernel_X_startp (void)
 *	start process
 */
void kernel_X_startp (void)
{
	K_SETLABEL (X_startp);
	#ifdef ENABLE_EXPKI
		TWO_PARAM_IN(other_workspace, start_offset);
		LOADWPTR;
		LOADRETURN;
	#else
		LOADWPTR;
		LOADRETURN;
		TWOPARAMS(other_workspace, start_offset);
	#endif
	ENTRY_TRACE (X_startp, "%p, %d", other_workspace, start_offset);
	DBGMSG("startp: called with other_workspace = 0x%x, offset = 0x%x\n", other_workspace, start_offset);
	other_workspace[Iptr] = start_offset+return_address;
	QUEUE(other_workspace);
	RESTORE; /* restore Wptr and esi */
	EJUMP(return_address);
}


/*
 *	void kernel_X_runp (void)
 *	run process
 */
void kernel_X_runp (void)
{
	K_SETLABEL (X_runp);
	#ifdef ENABLE_EXPKI
		ONE_PARAM_IN(other_workspace);
		LOADWPTR;
		LOADRETURN;
	#else
		LOADWPTR;
		LOADRETURN;
		PARAM(other_workspace);
	#endif
	ENTRY_TRACE (X_runp, "%p", other_workspace);
	(word)other_workspace &= 0xfffffffe; /* clear bit zero (priority bit) */
	QUEUE(other_workspace);
	RESTORE; /* restore Wptr and esi */
	EJUMP(return_address);
}


/*
 *	void kernel_X_ldtimer (void)
 *	load timer
 */
void kernel_X_ldtimer (void)
{
	K_SETLABEL (X_ldtimer);
	LOADWPTR;
	LOADRETURN;
	ENTRY_TRACE (X_ldtimer, "");
	now = Time_GetTime();
	DBGMSG( "ldtimer: now = %d.\n", now);
	#ifdef ENABLE_EXPKI
		RESTORE;
		ONE_PARAM_OUT (now);
	#else
		RETURNVAL(now);
		RESTORE;
	#endif
	EJUMP(return_address);
}


/*
 *	void kernel_X_tin (void)
 *	timer input
 */
void kernel_X_tin (void)
{
	K_SETLABEL (X_tin);
	#ifdef ENABLE_EXPKI
		ONE_PARAM_IN(wait_time);
		LOADWPTR;
		LOADRETURN;
	#else
		LOADWPTR;
		LOADRETURN;
		PARAM(wait_time);
	#endif
	ENTRY_TRACE (X_tin, "%d", wait_time);
	now = Time_GetTime();
	DBGMSG( "tin: now= %d  ", now );
	DBGMSG( "wait_time= %d\n", wait_time );
	test_time = now - wait_time;
	&test_time;
	&now;
	&wait_time;
	DBGMSG( "test_time= %d\n", test_time );
	#if 1
		if (Time_AFTER(now,wait_time)) {
	#else
		/* if ((now-wait_time)>0) { */
		if ( test_time > 0 ) {
	#endif
	/* Indenting looks a little peculiar here */
		DBGMSG( "tin: timer expired, returning to caller.\n");
		DBGMSG( "tin: wait time is %d, ", wait_time );
		DBGMSG( "now is %d, ", now );
		DBGMSG( "now-wt = %d,", (now-wait_time) );
		DBGMSG( "now-wt > 0 = %d\n", ((now-wait_time)>0) );
		#if 0
			fprintf(stderr, "tin: now-wait_time >0.\n");
		#endif
		RESTORE;
		EJUMP(return_address);
	} else {
		asm("X_tin_no:");
		DBGMSG( "tin: wait time not expired.");
		DBGMSG( "tin: wait time is %d, now is %d, now-wt = %d," \
			" now-wt > 0 = %d\n", \
			wait_time, now, (now-wait_time), \
			((now-wait_time)>0));
		Wptr[Iptr] = return_address;
		Wptr[State] = Waiting_p; /* T9000 book says enabling */
		wait_time++; /* from T9000 book... */
		GetTimeField(Wptr) = wait_time;
		/* start of pseudo code in SPARC code */
		current = Tptr;
		previous = NotProcess_p;
		while ((current != NotProcess_p) && (Time_AFTER(wait_time,GetTimeField(current)))) {
			previous = current;
			current = (word *)current[TLink];
		}
		Wptr[TLink] = (word)current;
		if (previous != NotProcess_p) {
			previous[TLink] = (word)Wptr;
		} else {
			Tptr = Wptr;
			DBGMSG( "tin: setting alarm with delay %d.", (GetTimeField(Tptr)) - now);
			Time_SetAlarm(Time_MINUS(GetTimeField(Tptr), now));
		}
		K_JUMP(X_scheduler);
	}
}


/*
 *	void kernel_X_altstart (void)
 *	ALT start
 *	NOT USED AT PRESENT  [inlined in tranpc]
 */
void kernel_X_altstart (void)
{
	K_SETLABEL (X_altstart);
	LOADRETURN;
	LOADWPTR;
	ENTRY_TRACE (X_altstart, "");
	Wptr[State] = Enabling_p;
	RESTORE;
	EJUMP(return_address);
}


/*
 *	void kernel_X_altend (void)
 *	ALT end
 *	NOT USED AT PRESENT  [inlined in tranpc]
 */
void kernel_X_altend (void)
{
	K_SETLABEL (X_altend);
	LOADWPTR;
	LOADRETURN;
	ENTRY_TRACE (X_altend, "");
	DBGMSG("altend: called.\n");
	return_address += Wptr[Temp];
	RESTORE;
	EJUMP(return_address);
}


/*
 *	void kernel_X_altwt (void)
 *	ALT wait
 */
void kernel_X_altwt (void)
{
	K_SETLABEL (X_altwt);
	LOADWPTR;
	LOADRETURN;
	ENTRY_TRACE (X_altwt, "");
	DBGMSG("altwt: called.\n");
	Wptr[Temp] = NoneSelected_o;
	if (Wptr[State] == Ready_p) {
		DBGMSG("altwt: guard is Ready_p.  Returning to caller.\n");
		RESTORE;
		EJUMP(return_address);
	} else {
		DBGMSG( "altwt: guard is not Ready_p.   Decheduling current process.\n");
		Wptr[State] = Waiting_p;
		SETRETURN(return_address);
		RESTORE;
		K_JUMP (X_stopp);
	}
}


/*
 *	void kernel_X_taltwt (void)
 *	timer ALT wait
 */
void kernel_X_taltwt (void)
{
	K_SETLABEL (X_taltwt);
	LOADWPTR;
	LOADRETURN;
	ENTRY_TRACE(X_taltwt, "");
	DBGMSG( "taltwt: called.\n");
	Wptr[Temp] = NoneSelected_o;
	if (Wptr[State] == Ready_p) {
		DBGMSG("taltwt: State is Ready_p so Wptr[Time_f] = now.\n");
		now = Time_GetTime();
		GetTimeField(Wptr) = now;
	} else {
		if (Wptr[TLink] == TimeNotSet_p) {
			DBGMSG("taltwt: TLink is TimeNotSet_p.  Descheduling.\n");
			Wptr[State] = Waiting_p;
			/* ... wait until Wptr^[State] = Ready_p (execute stopp) */
			SETRETURN(return_address);
			RESTORE;
			K_JUMP (X_stopp);
		} else {
			DBGMSG("taltwt: TLink is TimeSet_p.\n");
			now = Time_GetTime();
			if (Time_AFTER(now, GetTimeField(Wptr))) { /* now AFTER Wptr^[Time_f] */
				DBGMSG("taltwt: now AFTER Wptr[Time_f].\n");
				Wptr[State] = Ready_p;                     /* from T9000 Instruction */
				GetTimeField(Wptr) = now;                     /* Set p.436 */
			} else {
				DBGMSG("taltwt: now is not AFTER Wptr[Time_f].");
				DBGMSG("Inserting this process into timer queue.\n");
				GetTimeField(Wptr)++;
				/* ... insert this process in timer queue */
				current = Tptr;
				previous = (word *)NotProcess_p;
				while ( (current != NotProcess_p) &&
					(Time_AFTER(GetTimeField(Wptr), GetTimeField(current)))
					) {
					previous = current; /* deduce: previous <> NotProcess_p */
					current = (word *)current[TLink];
				}
				Wptr[TLink] = (word)current;
				if (previous != NotProcess_p) { /* iff: current <> Tptr */
					DBGMSG("taltwt: Inserting in middle of queue.\n");
					previous[TLink] = (word)Wptr;       /* insert in middle of queue */
				} else {
					DBGMSG("taltwt: Inserting at head of queue (and setting alarm).\n");
					DBGMSG("taltwt: Wptr[Time_f] = %d, now = %d", Wptr[Time_f], now);
					Tptr = Wptr;                  /* insert at head of queue */
					Time_SetAlarm(Time_MINUS(GetTimeField(Wptr), now));
				}
				/* ... end insert */
				if (Wptr[State] != Ready_p) { /* redundant ? */
					#ifdef DEBUG_KERNEL
						DBGMSG("taltwt: Wptr[State] is not Ready_p, set to Waiting_p");
						DBGMSG("and deschedule.\n");
					#endif
					Wptr[State] = Waiting_p; /* see altwt */
					/* ... wait until Wptr[State] = Ready_p (execute stopp) */
					SETRETURN(return_address);
					RESTORE;
					K_JUMP (X_stopp);
				} else {
					impossible(1, return_address);
				}
			}
		}
	}
	DBGMSG( "taltwt:exiting to caller.\n");
	RESTORE;
	EJUMP(return_address);
}


/*
 *	void kernel_X_dist (void)
 *	disable timer
 */
void kernel_X_dist (void)
{
	K_SETLABEL (X_dist);
	#ifdef ENABLE_EXPKI
		THREE_PARAM_IN(code_offset, precondition, timeout);
		LOADWPTR;
		LOADRETURN;
	#else
		LOADWPTR;
		LOADRETURN;
		THREEPARAMS(code_offset, precondition, timeout);
	#endif
	ENTRY_TRACE (X_dist, "%d, %d, %d", code_offset, precondition, timeout);
	DBGMSG( "dist: entered.\n");
	if (!precondition) {
		DBGMSG( "dist: precondition is false.\n");
		result = false;
	} else {
		DBGMSG( "dist: precondition is true.\n");
		if (Wptr[TLink] == TimeNotSet_p) {
			DBGMSG( "dist: no timer is ready.\n");
			result = false;
		} else {
			if (Wptr[TLink] == TimeSet_p) {
				DBGMSG( "dist: some timer is ready...\n");
				/* some timer has gone off */
				if (!(Time_AFTER(GetTimeField(Wptr), (Time)timeout))) {
					DBGMSG( "dist: ... but not this one.");
					result = false;
				} else {
					if (Wptr[Temp] != NoneSelected_o) {
						DBGMSG("dist: a previous guard was selected.\n");
						result = false;
					} else {
						DBGMSG("dist: This is the first ready guard.\n");
						Wptr[Temp] = code_offset;
						result = true;
					}
				}
			} else {
				/*{{{  cancel timer*/
				DBGMSG("dist: strage else condition - remove process from tq.\n");
				result = false;
				/* remove process from timer queue */
				if (Tptr == Wptr) {
					DBGMSG( "dist: Process is at head of timer Q.\n");
					Tptr = (word *)Tptr[TLink];
					if (Tptr != NotProcess_p) {
						DBGMSG("dist: there is something else on the timer queue.\n");
						/* Tptr[Link] = NotProcess_p -- commented out in comment! */
						now = Time_GetTime();
						if (!(Time_AFTER(GetTimeField(Wptr), now))) { /*should be same as */
							/*if (((int)Tptr[Time_f] - now) <= 0) { */
							DBGMSG("dist: time out has happened.  Setting timer sync.\n");
							sf.s.tim_sync = 1;
						} else {
							DBGMSG("dist: timeout not happened yet, set alarm.\n");
							Time_SetAlarm(Time_MINUS(GetTimeField(Tptr), now));
						}
					}
				} else {
					if (Tptr == NotProcess_p) {
						impossible(3, return_address);
					} else {
						DBGMSG( "dist: searching timer queue from 2nd element.\n");
						previous = Tptr; /* search from 2nd item on timer queue */
						current = (word *)Tptr[TLink];
						while ((current != NotProcess_p) && (current != Wptr)) {
							DBGMSG( "dist: onto next item...\n");
							previous = current;
							current = (word *)current[TLink];
						}
						if (current == NotProcess_p) {
							impossible(4, return_address);
						}
						/* deduce: current == Wptr */
						previous[TLink] = Wptr[TLink];
					}
				}
				DBGMSG("dist: Timer cancelled.\n");
				Wptr[TLink] = TimeNotSet_p; /* i.e. timer cancelled */
				/*}}}*/
			}
		}
	}
	#ifdef ENABLE_EXPKI
		RESTORE;
		ONE_PARAM_OUT (result);
	#else
		RETURNVAL(result);
		RESTORE;
	#endif
	EJUMP(return_address);
}


/*
 *	void kernel_X_enbs (void)
 *	enable skip guard
 *
 *	NOT USED AT PRESENT  [inlined by tranpc]
 */
void kernel_X_enbs (void)
{
	K_SETLABEL (X_enbs);
	#ifdef ENABLE_EXPKI
		ONE_PARAM_IN(guard);
		LOADWPTR;
		LOADRETURN;
	#else
		LOADWPTR;
		LOADRETURN;
		PARAM(guard);
	#endif
	ENTRY_TRACE (X_enbs, "%d", guard);
	if (guard) {
		Wptr[State] = Ready_p;
	}
	#ifdef ENABLE_EXPKI
		RESTORE;
		ONE_PARAM_OUT (guard);
	#else
		RETURNVAL(guard);
		RESTORE;
	#endif
	EJUMP(return_address);
}


/*
 *	void kernel_X_enbc (void)
 *	enable channel
 */
void kernel_X_enbc (void)
{
	K_SETLABEL (X_enbc);
	#ifdef ENABLE_EXPKI
		TWO_PARAM_IN(guard, channel_address);
		LOADWPTR;
		LOADRETURN;
	#else
		LOADWPTR;
		LOADRETURN;
		TWOPARAMS(guard, channel_address);
	#endif
	ENTRY_TRACE (X_enbc, "%d, %p", guard, channel_address);
	DBGMSG("enbc: guard = %d, channel address = 0x%x\n",
	guard, channel_address);
	if (guard) {
		if (*channel_address == NotProcess_p) {
			*channel_address = (word *)((word)Wptr | 1);
		} else {
			if (*channel_address != (word *)((word)Wptr | 1)) {
				Wptr[State] = Ready_p;
			}
		}
	}
	#ifdef ENABLE_EXPKI
		RESTORE;
		ONE_PARAM_OUT(guard);
	#else
		RETURNVAL(guard);
		RESTORE;
	#endif
	EJUMP(return_address);
}


/*
 *	void kernel_X_enbt (void)
 *	enable timer
 */
void kernel_X_enbt (void)
{
	K_SETLABEL (X_enbt);
	#ifdef ENABLE_EXPKI
		TWO_PARAM_IN(guard, timeout);
		LOADWPTR;
		LOADRETURN;
	#else
		LOADWPTR;
		LOADRETURN;
		TWOPARAMS(guard, timeout);
	#endif
	ENTRY_TRACE (X_enbt, "%d, %d", guard, timeout);
	DBGMSG("enbt: entered. timeout = %d\n", timeout);
	if (guard) {
		if (Wptr[TLink] == TimeNotSet_p) {
			DBGMSG( "enbt: This is the first enbt.\n");
			Wptr[TLink] = TimeSet_p;
			GetTimeField(Wptr) = (Time)timeout;
		} else {
			DBGMSG( "enbt: This is not the first enbt.\n");
			if (Wptr[TLink] == TimeSet_p) {
				if (Time_AFTER(GetTimeField(Wptr), (Time)timeout)) {
					#ifdef DEBUG_KERNEL
						DBGMSG( "enbt: This enbt has an earlier time.\n");
					#endif
					GetTimeField(Wptr) = (Time)timeout;
				}
			}
		}
	}
	#ifdef ENABLE_EXPKI
		RESTORE;
		ONE_PARAM_OUT (guard);
	#else
		RESTORE;
		RETURNVAL(guard);
	#endif
	EJUMP(return_address);
}


/*
 *	void kernel_X_disc (void)
 *	disable channel
 */
void kernel_X_disc (void)
{
	K_SETLABEL (X_disc);
	#ifdef ENABLE_EXPKI
		THREE_PARAM_IN(process_address, guard, channel_address);
		LOADWPTR;
		LOADRETURN;
	#else
		LOADWPTR;
		LOADRETURN;
		THREEPARAMS(process_address, guard, channel_address);
	#endif
	ENTRY_TRACE (X_disc, "%p, %d, %p", process_address, guard, channel_address);
	DBGMSG( "disc: process_offset = 0x%x, guard = %d, channel address = 0x%x\n",
	process_address, guard, channel_address);
	if (!guard) {
		fired = false;
	} else {
		if (*channel_address == NotProcess_p) {
			fired = false;
		} else {
			if (*channel_address == (word *)((word)Wptr | 1)) {
				*channel_address = NotProcess_p;
				fired = false;
			} else {
				if (Wptr[Temp] == NoneSelected_o) {
					Wptr[Temp] = process_address; /* actually offset */
					fired = true;
				} else {
					fired = false;
				}
			}
		}
	}
	#ifdef DEBUG_KERNEL
		DBGMSG( "disc: fired = %d.\n", fired);
	#endif
	#ifdef ENABLE_EXPKI
		RESTORE;
		ONE_PARAM_OUT (fired);
	#else
		RETURNVAL(fired);
		RESTORE;
	#endif
	EJUMP(return_address);
}


/*
 *	void kernel_X_diss (void)
 *	disable SKIP guard
 */
void kernel_X_diss (void)
{
	K_SETLABEL (X_diss);
	#ifdef ENABLE_EXPKI
		TWO_PARAM_IN(process_address, guard);
		LOADWPTR;
		LOADRETURN;
	#else
		LOADWPTR;
		LOADRETURN;
		TWOPARAMS(process_address, guard);
	#endif
	ENTRY_TRACE (X_diss, "%p, %d", process_address, guard);
	#ifdef DEBUG_KERNEL
		DBGMSG( "diss: entered.\n");
	#endif
	if (!guard) {
		fired = false;
	} else {
		if (Wptr[Temp] != NoneSelected_o) {
			fired = false;
		} else {
			Wptr[Temp] = process_address;
			fired = true;
		}
	}
	#ifdef ENABLE_EXPKI
		RESTORE;
		ONE_PARAM_OUT (fired);
	#else
		RETURNVAL(fired);
		RESTORE;
	#endif
	EJUMP(return_address);
}


