/*
 *	Pentium/386 and GCC specific assembler inserts
 *	Copyright (C) 1998 Jim Moores
 *
 *	This program is free software; you can redistribute it and/or modify
 *	it under the terms of the GNU General Public License as published by
 *	the Free Software Foundation; either version 2 of the License, or
 *	(at your option) any later version.
 *
 *	This program is distributed in the hope that it will be useful,
 *	but WITHOUT ANY WARRANTY; without even the implied warranty of
 *	MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *	GNU General Public License for more details.
 *
 *	You should have received a copy of the GNU General Public License
 *	along with this program; if not, write to the Free Software
 *	Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 */

/*
 * This file attempts to isolate as many as possible of the platform 
 * specific features of this package as possible.  The macros have been
 * designed so that it should be easy to port this kernel to a more
 * register based machine fairly easily (eg Alpha, ARM etc).  Note that
 * there are numerous functions in this file labelled such that are no
 * longer used.
 */

#ifndef SCHED_ASM_INSERTS_H
#define SCHED_ASM_INSERTS_H
/* Load the Workspace pointer register into a variable.
 */
#define LOADWPTR \
  __asm__ __volatile__ ("movl   %%ebp, %0\n" \
                        : "=m" (Wptr) \
                       )

/* Restore the workspace pointer and kernel entry table pointers to their
 * values before this function call.  This should be combined with a EJUMP
 * really, because it always happens.  Note that the jump table pointer is
 * not necessary in C.
 */
#define RESTORE \
  __asm__ __volatile__ ("movl   %0, %%ebp\n \
			 movl	%1, %%esi\n" \
                        : /* no outputs */ \
                        : "g" (Wptr), "m" (occam_table) \
                       )

/* Load the parameter value passed to the kernel into a variable.  
 */ 
#define PARAM(A) \
  __asm__ __volatile__ ("popl   %0\n" \
                        : "=g" (A) \
                       )
/* 
 * Load the two parameter values passed to the kernel into variables.  
 */ 
#define TWOPARAMS(A,B) \
  __asm__ __volatile__ ("popl   %0\n \
                         popl   %1\n" \
                        : "=g" (A), "=g" (B) \
                       )

/* Load the three parameter values passed to the kernel into variables.  
 */ 
#define THREEPARAMS(A,B,C) \
  __asm__ __volatile__ ("popl   %0\n \
                         popl   %1\n \
                         popl   %2\n" \
                        : "=g" (A), "=g" (B), "=g" (C) \
                       )

/* Load the four parameter values passed to the kernel into variables.  
 * this is in fact used to pass the time as a 64bit value, in two 32bit
 * chunks.
 */ 
 /* Mainly used for 64-bit quantities, but I use it for occam debugging. [fred] */
#define FOURPARAMS(A,B,C,D) \
  __asm__ __volatile__ ("popl   %0\n" \
                        "popl   %1\n" \
                        "popl   %2\n" \
                        "popl   %3\n" \
                        : "=g" (A), "=g" (B), "=g" (C), "=g" (D) \
                       )

/* experimental kernel interface stuff
 */

#define ONE_PARAM_IN(A) \
	__asm__ __volatile__ ( \
		"\n" \
		: "=a" (A))
#define TWO_PARAM_IN(A,B) \
	__asm__ __volatile__ ( \
		"\n" \
		: "=b" (A), "=a" (B))
#define THREE_PARAM_IN(A,B,C) \
	__asm__ __volatile__ ( \
		"\n" \
		: "=c" (A), "=b" (B), "=a" (C));
#define FOUR_PARAM_IN(A,B,C,D) \
	__asm__ __volatile__ ( \
		"\n" \
		: "=d" (A), "=c" (B), "=b" (C), "=a" (D));
#define ONE_PARAM_OUT(A) \
	__asm__ __volatile__ ( \
		"\n" \
		: : "a" (A));
#define TWO_PARAM_OUT(A,B) \
	__asm__ __volatile__ ( \
		"\n" \
		: : "b" (A), "a" (B));
#define THREE_PARAM_OUT(A,B,C) \
	__asm__ __volatile__ ( \
		"\n" \
		: : "c" (A), "b" (B), "a" (C));


/* loads the return address on entry to the kernel function into a global 
 * variable called 'return_address'.  This could be paramterised, but we
 * never use any other variable, so this seemed pointless.
 */
#define LOADRETURN \
  __asm__ __volatile__ ("popl   %0\n" \
                        : "=g" (return_address) \
                       )

/* Return a value from the current kernel function.  In tranpc this is done
 * by pushing on the system stack, but the convension is to use EAX on the
 * 386.  We use tranpc's convention for compatablilty with occam.
 */
#define RETURNVAL(X) \
  __asm__ __volatile__ ("pushl  %0\n" \
                        : /* no outputs */ \
                        : "g" (X) \
                       )

/* NO LONGER USED - USED TO SET THE RETURN ADDRESS */
/* KERNEL FUNCTIONS NOW EXIT WITH EJUMP's */
#define SETRETURN(X) \
  __asm__ __volatile__ ("pushl  %0\n" \
                        : /* no outputs */ \
                        : "g" (X) \
                       )

/* NO LONGER USED */
#define JUMP(X) \
  __asm__ __volatile__ ("jmp    %0 \n" \
                        : /* no outputs */ \
                        : "r" (X) \
                        : "memory" \
                       )
/* NO LONGER USED - REPLACED WITH PARAMETER LOADING */
#define PUSH(X) \
  __asm__ __volatile__ ("pushl  %0\n" \
                        : /* no outputs */ \
                        : "g" (X) \
                       )

/*
 * scheduler label functions
 */
#define K_LOADLABADDR(X,V) \
	__asm__ __volatile__ ("\n" \
		"	leal "#X", %%eax\n" \
		"	movl %%eax, (%0)\n" \
		: "=g" (V) : : "eax" \
		)
#define K_JUMP(X) \
	__asm__ __volatile__ ("\n" \
		"	jmp "#X"\n" \
		)
#define K_SETLABEL(X) \
	__asm__ __volatile__ ("\n" \
		"	"#X":\n" \
		)


/* an EJUMP is a jump out of the kernel into user code, usually the return
   address 
   Note that only GCC supports goto *(X), may need to be replaced with an
   asm jump if porting. */
#define EJUMP(X) goto *(X)

/* Restore the stack pointer before returning to C */
#define RESTORESP(X) \
  __asm__ __volatile__ ("movl   %0, %%esp\n" \
                        : /* no outputs */ \
                        : "g" (X) \
                       ) 

/* Some macros for stacking/unstacking the static context so that kernel
 * calls can be called back from calls out to the communications subsystem
 * (is this 're-entrancy'?).  The current use of these assumes that they 
 * have LIFO stacking semantics.
 */
#define SAVESTATIC(X) \
    __asm__ __volatile__ ("pushl  %0\n" \
                          : /* no ouputs */ \
                          : "q" (X) \
                         );

#define RESTORESTATIC(X) \
    __asm__ __volatile__ ("popl  %0\n" \
                          : "=q" (X) \
                          : /* no inputs */ \
                         )
#define SAVESTATICS \
/*  __asm__ __volatile__ ("pushl  %0\n" \
                        "pushl  %1\n" \
                        : \
                        : "q" (Wptr), "q" (return_address) \
                        : "eax", "ebx" \
                       )*/

#define RESTORESTATICS \
/*  __asm__ __volatile__ ("popl %0\n" \
                        "popl %1\n" \
                        : "=q" (return_address), "=q" (Wptr) \
                        : \
                        : "eax", "ebx" \
                       )*/

/* MACRO LBL(A) -   declared a kernel function's label on two levels: 
 *                  the C level, as in
 *                    label: 
 *                  and as a global asm label so that functions can be
 *                  called without going through the jump table.  
 *                    asm(".globl label\n label:");
 *
 * IMPLEMENTATION NOTE:
 * this is really wierd - we rely here on ANSI C string concatenation: 
 * "Jim" "John" "Jan" == "JimJohnJan" therefore if B is a string we get 
 * ".globl" "<label>" concating to ".globl <label>" followed by 
 * "<label>" ":" concatting to "<label>:".  Items that appear quoted
 * are actually outside the quotes, this is the confusing thing.
 * I also use 'stringification' as documented in the GNU cpp 
 * documentation.  If '#define A foo' then '#A' is '"foo"'.  This is
 * needed to insert the name into the asm block.  Before I new how to
 * do this, this macro took two paramters, one 'stringified' and one
 * not.  This is nicer. 
 */ 

#define MEMCPY(SRC,DEST,COUNT) \
  __asm__ __volatile__ ( \
    "cld\n" \
    "rep\n" \
    "movsb\n" \
    : /* no outputs */ \
    : "S" (SRC), "D" (DEST), "c" (COUNT) \
  )
#ifdef USE_ASM_BITOPS
#define SETBIT(WORD, BITNUM) \
  __asm__ __volatile__ ( \
    "btsl %1, %0" \
    : "&=g" (WORD) \
    : "g" (WORD), "i" (BITNUM) \
  )

#define CLEARBIT(WORD, BITNUM) \
  __asm__ __volatile__ ( \
    "btrl %1, %0" \
    : "&=g" (WORD) \
    : "g" (WORD), "i" (BITNUM) \
  )

#define SEARCHWORD(WORD, BITNUM) \
  __asm__ __volatile__ ( \
    "bsfl %0, %1" \
    : "=r" (BITNUM) \
    : "g" (WORD) \
  )
#endif
#ifdef GSL_SMP
#define CLAIM(SEMA) \
  __asm__ __volatile__ ( \
    "   movl    $1, %%eax\n" \
    "1: xchg    %%eax, %0\n" \
    "   cmpl    $0, %%eax\n" \
    "   jne     1b\n" \
    : /* no outputs */ \
    : "m" (SEMA) \
    : "eax" \
  )

#define RELEASE(SEMA) \
  __asm__ __volatile__ ( \
    "   movl   $0, %0" \
    : /* no outputs */ \
    : "m" (SEMA) \
  )
#endif
#endif /* sched_asm_inserts.h */
