/*
 *	Portable timer implementation
 *	Copyright (C) 1996-1999 Jim Moores
 *	Modifications (C) 2000 Fred Barnes <frmb2@ukc.ac.uk>
 *
 *	This program is free software; you can redistribute it and/or modify
 *	it under the terms of the GNU General Public License as published by
 *	the Free Software Foundation; either version 2 of the License, or
 *	(at your option) any later version.
 *
 *	This program is distributed in the hope that it will be useful,
 *	but WITHOUT ANY WARRANTY; without even the implied warranty of
 *	MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *	GNU General Public License for more details.
 *
 *	You should have received a copy of the GNU General Public License
 *	along with this program; if not, write to the Free Software
 *	Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 */

/*
 * only include these definitions if not using the faster 
 * macro definitions in the header file.
 * There are some extra functions defined at the end of this
 * block - these are the functions used by the kernel, but 
 * are derived from the CERN/UKC spec so that only those need
 * be provided.
 */

/*
 *	Note: Only the UNIX timer implementation exists here now
 */

#include "timer.h"

/* BEGIN UNIX TIMER IMPLEMENTATION */

long long Time_hz = 1000000LL;
long long Time_ticks_per_hz = 1LL;
long long Time_ticks_per_sec = 1000000LL;
long long Time_ticks_per_us = 1LL;
double Time_ticks_per_us_d = 1.0;

#ifdef ENABLE_CPU_TIMERS
	extern unsigned long long glob_cpumhz;
#endif

#ifndef INLINE_TIMERS
#warning Not inlining timer code

Time Time_GetTime(void) {
	#ifdef ENABLE_CPU_TIMERS
		Time time_in_us;
		unsigned long long time_now;

		__asm__ __volatile__ ("\n" \
			"	.byte 0x0f, 0x31\n" \
			: "=a" (((int *)&time_now)[0]), "=d" (((int *)&time_now)[1]));
		time_now /= glob_cpumhz;
		time_in_us = (Time)(time_now & 0xffffffffLL);
		return time_in_us;
	#else
		return rtime();
	#endif
}

Time Time_PLUS(Time t1, Time t2) {
	return (t1+t2);
}

Time Time_MINUS(Time t1, Time t2) {
	return (t1-t2);
}

bool Time_AFTER(Time t1, Time t2) {
	Time temp_time;

	temp_time = t2 - t1;
	&temp_time;		/* Not registerised please */
	if (temp_time & 0x80000000) {
		return 1;
	}
	return 0;
}

bool Time_PollAlarm(void) {
	return (sf.s.tim_sync);
}

void Time_SetAlarm(Time t) {
	ualarm(t, 0);
}

void Time_ResetAlarm(void) {
	ualarm(0,0);
}
#endif

