/*
 *	KRoC scheduler startup code for occam
 *	Copyright (C) 1995, 1996, 1997 D.C.Wood, P.H.Welch
 *
 *	This program is free software; you can redistribute it and/or modify
 *	it under the terms of the GNU General Public License as published by
 *	the Free Software Foundation; either version 2 of the License, or
 *	(at your option) any later version.
 *
 *	This program is distributed in the hope that it will be useful,
 *	but WITHOUT ANY WARRANTY; without even the implied warranty of
 *	MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *	GNU General Public License for more details.
 *
 *	You should have received a copy of the GNU General Public License
 *	along with this program; if not, write to the Free Software
 *	Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 */

/*
 * $Source: /proj/ofa/rts/RCS/rtsmain.c,v $
 *
 * $Id: rtsmain.c,v 1.17 1997/09/03 15:42:33 djb1 Exp $
 *
 * (C) Copyright 1995,1996,1997 D.C. Wood <D.C.Wood@ukc.ac.uk> and 
 *                              P.H. Welch <P.H.Welch@ukc.ac.uk>
 * University of Kent at Canterbury
 *
 * Based on work by: Peter Shephard @ QMC Sept. 1984 for
 * UNIX OPS 2.0 (Release 2.0.1 : 24th. July, 1986)
 */

#define RTS_MAIN

#include <unistd.h>
#include <fcntl.h>
#include <string.h>

#include "rts.h"
#include "sched_types.h"

extern sync_type sf;

void keyboard_process (int user_pid, int to_user_fd, int from_user_fd, bool is_a_tty);
void user_process (int kbd_pid, int to_kbd_fd, int from_kbd_fd, bool is_a_tty);

/* RTS common variables */

/* Two pipes (socketpairs) */
int kbd_chars[2];
int kbd_ack[2];

#ifdef DEBUG_RTS
bool trace;
#endif

#ifdef ENABLE_CPU_TIMERS
extern unsigned long long glob_cpumhz;
#endif

char *short_cmdline;
char *long_cmdline;

int kroc_argc;
char **kroc_argv;


static int my_fork (void);
static void savecmdline (int argc, char **argv);


static int my_fork (void) {
	int n = 5;
	int cpid;

	do {
		if ((cpid = fork ()) != ERR) {
			return (cpid);
		} else {
			sleep (2);
		}
	} while (--n >= 0);

	fputs ("KROC: fork() failed\n", stderr);
	exit (-1);
}  /* my_fork */


static void savecmdline (int argc, char **argv)
{
	static char cmdline[512];
	int i = 0;

	cmdline[0] = '\0';
	long_cmdline = &cmdline[0];
	short_cmdline = &cmdline[0];
	while (i < argc) {
		strcat (cmdline, argv[i]);
		i++;
		if (i < argc) {
			strcat (cmdline, " ");  /* separate arguments except the last one */
		}
		if (i == 1) {
			short_cmdline = &cmdline[strlen(cmdline)];
		}
	}
	return;
}


int main (int argc, char **argv)
{
	int cpid;
	int is_a_tty;
  
	kroc_argc = argc;
	kroc_argv = argv;
  
  
	savecmdline (argc, argv);
	/*
	 * create pipes down which :-
	 *       (a) parent will send keyboard chars to occam;
	 *       (b) child will acknowledge with same char
	 *               (only when an Occam process inputs it);
	 */
	if (pipe (kbd_chars) == ERR || pipe (kbd_ack) == ERR) {
		fputs ("failed to create pipe (s)\n", stderr);
		exit (-3);
	}

	#ifdef DEBUG_RTS
		trace =  TRUE;
	#endif /* DEBUG_RTS */
  
	is_a_tty = isatty (0);

	if (is_a_tty) {
		save_tty_state ();
	}

	#ifdef ENABLE_CPU_TIMERS
	{
		FILE *fp;

		/* better read clock value from someplace */
		fp = fopen (".kroc_clock", "r");
		if (!fp) {
			fp = fopen ("/etc/kroc_clock", "r");
		}
		if (!fp) {
			fprintf (stderr, "KRoC: unable to find clock file (.kroc_clock or /etc/kroc_clock)\n");
			exit (1);
		}
		if (!fscanf (fp, "%Lu", &glob_cpumhz)) {
			fprintf (stderr, "KRoC: error reading clock speed\n");
			exit (1);
		}
		fclose (fp);
	}
	#endif

	sf.combined_sync = 0;
	/* fork off process to run occam kernel */
	if ((cpid=my_fork ())) {
		/* parent */
		close (kbd_chars[1]);
		close (kbd_ack[0]);
    
		/* Close stdin = fd 0 = keyboard */
		/* fclose(stdin); */
    
		user_process (cpid, kbd_ack[1], kbd_chars[0], is_a_tty);
	} else {
		/* child */
		close (kbd_chars[0]);
		close (kbd_ack[1]);
    
		fclose(stdout);
    
		keyboard_process (getppid(), kbd_chars[1], kbd_ack[0], is_a_tty);
	}
	/* keep compiler happy */
	return 0;
}  /* main */


