/*
 *	CSPLib header file
 *	Copyright (C) 1998 Jim Moores
 *
 *	This program is free software; you can redistribute it and/or modify
 *	it under the terms of the GNU General Public License as published by
 *	the Free Software Foundation; either version 2 of the License, or
 *	(at your option) any later version.
 *
 *	This program is distributed in the hope that it will be useful,
 *	but WITHOUT ANY WARRANTY; without even the implied warranty of
 *	MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *	GNU General Public License for more details.
 *
 *	You should have received a copy of the GNU General Public License
 *	along with this program; if not, write to the Free Software
 *	Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 */

#ifndef CSPLIB_H
#define CSPLIB_H

#include "external.h"

#include "ukcthreads_types.h"

#include "csptime.h"

#include "csplib_consts.h"

#ifdef CPU_TIMERS
  #ifndef TIMER_H
    #define lo(X) ((word *)&(X))[0]
    #define hi(X) ((word *)&(X))[1]
  #endif
#endif

#ifdef PRIORITIES
  #define highest(A,B) ((A)>(B) ? (A) : (B))
#endif

#include "csplib_types.h"
#include "sched_types.h"
extern void X_out();
extern void X_in();
extern void CSPInit();
extern sync_type sf;
void EndProcess(Process *p);

#include "i386.h"

/* All kernel functions should be defined in the following format, note
 * the use of a EBNF style notation - it's not part of the syntax!
 *
 *   #define <NAME>(<params> [, <output paramter>]) \
 *     __asm__ __volatile__ (" \
 *       "SaveStackFrame[RV]" \
 *       "<no of params>ParamCall[RV](<kernel function>)" \
 *      ["GetResult"] \
 *       "RestoreStackFrame[RV]" \
 *     "
 *     : ["=&q"] 
 *     : FirstReg (p->wptr) {, SecondReg (<param>) } 
 *     : AllRegisters
 *
 * Where <no of params> should be the string "One", "Two", or "Three"
 * (not with extra quotes).  <kernel function> is the name of the 
 * function in the kernel to be called, eg "X_outbyte".  This should
 * be double quoted.
 * The reason everything has to be in a register is that if anything
 * is referred to from memory then the stack pointer relative reference 
 * for a local variable will be wrong because the stack pointer has 
 * changed and the C compiler doesn't know (because it can't actually 
 * parse the asm blocks, they just get passed on to the assembler.
 * The result paramter must have the contraint =&q which firstly 
 * prevents the allocation of a register used by an input operand (&)
 * and also restricts the allocation of registers for the result to
 * eax, ebx, ecx or edx.  This is to prevent to occasional allocation
 * of ebp or esp as the result register (which would be corrupted by
 * RestorestackFrame[RV].
 * The optional RV suffix must be included on the SaveStackFrame and 
 * <x>ParamCall macros if the function returns a macro.
 */

/* C_STARTP */
#define PriStartP(other_workspace,entry_point) \
  __asm__ __volatile__ (" \
    "SaveStackFrame" \
    "TwoParamCall("X_C_pri_startp")" \
    "RestoreStackFrame" \
    " \
    : /* no outputs */ \
    : FirstReg (p->wptr), SecondReg (other_workspace), ThirdReg (entry_point) \
    : AllRegisters_123 /* corrupted */ \
  )

#define StartP(other_workspace,entry_point) \
  __asm__ __volatile__ (" \
    "SaveStackFrame" \
    "TwoParamCall("X_C_startp")" \
    "RestoreStackFrame" \
    " \
    : /* no outputs */ \
    : FirstReg (p->wptr), SecondReg (other_workspace), ThirdReg (entry_point) \
    : AllRegisters_123 /* corrupted */ \
  )

/* RESCHEDULE */
#define IndirectMacroProcReschedule(p) \
  __asm__ __volatile__ (" \
    "SaveStackFrame" \
    "NoParamCall("X_resched")" \
    "RestoreStackFrame" \
    " \
    : /* no outputs */ \
    : FirstReg (p->wptr) /* important that its in a register */ \
    : AllRegisters_1 /* corrupted */ \
  )

#define PriSearchReschedule() \
  __asm__ __volatile__ (" \
    "SaveStackFrame" \
    "NoParamCall("X_C_pri_search_reschedule")" \
    "RestoreStackFrame" \
    " \
    : /* no outputs */ \
    : FirstReg (p->wptr) /* important that its in a register */ \
    : AllRegisters_1 /* corrupted */ \
  )

#define PriReschedule(i) \
  __asm__ __volatile__ (" \
    "SaveStackFrame" \
    "OneParamCall("X_C_pri_reschedule")" \
    "RestoreStackFrame" \
    " \
    : /* no outputs */ \
    : FirstReg (p->wptr), SecondReg (i) \
    : AllRegisters_12 /* corrupted */ \
  )

#define PriSchedule(i) \
  __asm__ __volatile__ (" \
    "SaveStackFrame" \
    "OneParamCall("X_C_pri_schedule")" \
    "RestoreStackFrame" \
    " \
    : /* no outputs */ \
    : FirstReg (p->wptr), SecondReg (i) \
    : AllRegisters_12 /* corrupted */ \
  )

#define SetArbitraryPriority(process, level) \
  __asm__ __volatile__ (" \
    "SaveStackFrame" \
    "TwoParamCall("X_C_arbitrary_priority_change")" \
    "RestoreStackFrame" \
    " \
    : /* no outputs */ \
    : FirstReg (p->wptr), SecondReg (process->wptr), ThirdReg (level) \
    : AllRegisters_123 /* corrupted */ \
  )

#define SetOwnPriority(level) \
  __asm__ __volatile__ (" \
    "SaveStackFrame" \
    "OneParamCall("X_C_own_priority_change")" \
    "RestoreStackFrame" \
    " \
    : /* no outputs */ \
    : FirstReg (p->wptr), SecondReg (level) \
    : AllRegisters_12 /* corrupted */ \
  )

/* STOPP */
#define IndirectMacroProcStop(p) \
  __asm__ __volatile__ (" \
    "SaveStackFrame" \
    "NoParamCall("X_stopp")" \
    "RestoreStackFrame" \
    " \
    : /* no outputs */ \
    : FirstReg (p->wptr) /* important that its in a register */ \
    : AllRegisters_1 /* corrupted */ \
  )

/* ------------------ SIMPLE CHANNEL COMMUNICATIONS ------------------*/
#define ChanOutInt(C,V) \
  __asm__ __volatile__("\
    "SaveStackFrame" \
    "TwoParamCall("X_outword")" \
    "RestoreStackFrame" \
    " \
    : /* no outputs */ \
    : FirstReg (p->wptr), SecondReg (V), ThirdReg (C) \
    : AllRegisters_123 /* corrupted */  \
  )

#define ChanOutChar(C,V) \
  __asm__ __volatile__("\
    "SaveStackFrame" \
    "TwoParamCall("X_outbyte")" \
    "RestoreStackFrame" \
    " \
    : /* no outputs */ \
    : FirstReg (p->wptr), SecondReg (V), ThirdReg (C) \
    : AllRegisters_123 /* corrupted */  \
  )

#define ChanOut(CHAN,SOURCE,COUNT) \
  __asm__ __volatile__("\
    "SaveStackFrame" \
    "ThreeParamCall("X_out")" \
    "RestoreStackFrame" \
    " \
    : /* no outputs */ \
    : FirstReg (p->wptr), SecondReg (COUNT), ThirdReg (CHAN), ForthReg (SOURCE) \
    : AllRegisters_1234 /* corrupted */  \
  )

#define ChanIn(C,D,N) \
  __asm__ __volatile__("\
    "SaveStackFrame" \
    "ThreeParamCall("X_in")" \
    "RestoreStackFrame" \
    " \
    : /* no outputs */ \
    : FirstReg (p->wptr), SecondReg (N), ThirdReg (C), ForthReg (D) \
    : AllRegisters_1234 /* corrupted */ \
  )

/* EXTERNAL COMMUNICATIONS MACRO DEFINITIONS */
/* NOTE: these macros use different macros to save and restore the stack 
 * frame.
 * Special macro for external communications - allows caller to specify what
 * workspace pointer to provide.  It calls a modified version of X_out that
 * does not block (ie just reschedules the other
 * process by queuing it).
 */
#define ExternChanOutNoBlock(WPTR,CHAN,SRC,COUNT) \
  __asm__ __volatile__("\
    "ExternalSaveStackFrame" \
    "ThreeParamCall("X_nb_out")" \
    "ExternalRestoreStackFrame" \
    " \
    : /* no outputs */ \
    : FirstReg (WPTR), SecondReg (COUNT), ThirdReg (CHAN), ForthReg (SRC) \
    : AllRegisters_1234 /* corrupted */  \
  )

/* Special macro for external communications - allows caller to specify what
 * workspace pointer to provide.  It calls a modified version of X_in that
 * does no block copy, and does not block (ie just reschedules the other 
 * process by queuing it).
 */
#define ExternChanInNoTrans(WPTR,CHAN) \
  __asm__ __volatile__("\
    "ExternalSaveStackFrame" \
    "OneParamCall("X_zlnb_in")" \
    "ExternalRestoreStackFrame" \
    " \
    : /* no outputs */ \
    : FirstReg (WPTR), ThirdReg (CHAN)  \
    : AllRegisters_13 /* corrupted */ \
  )

/* This macro calls the kernel function X_release_alt.  This releases any
 * processes alting over this channel, but performs no data transfer.  It
 * is called when the start of the data from an outputter has arrived.
 */
#define ExternPerformedAlt(WPTR,CHAN) \
  __asm__ __volatile__("\
    "ExternalSaveStackFrame" \
    "OneParamCall("X_release_alt")" \
    "ExternalRestoreStackFrame" \
    " \
    : /* no outputs */ \
    : FirstReg (WPTR), ThirdReg (CHAN)  \
    : AllRegisters_13 /* corrupted */ \
  )
/* END OF EXTERNAL COMMUNICATIONS MACROS */

/* these need function wrappers around them to make them the same as the
 * Inmos library (ie returning values rather than supplying addresses. 
 * The IndirectMacro prefix indicates macros that are included in these 
 * wrapper functions.
 */

#define IndirectMacroChanInInt(C,D) \
  __asm__ __volatile__("\
    "SaveStackFrame" \
    "ThreeParamCall("X_in")" \
    "RestoreStackFrame" \
    " \
    : /* no outputs */ \
    : FirstReg (p->wptr), "i" (4), SecondReg (C), ThirdReg (D) \
    : AllRegisters_123 /* Corrupted */ \
  )

#define IndirectMacroChanInChar(C,D) \
  __asm__ __volatile__("\
    "SaveStackFrame" \
    "ThreeParamCall("X_in")" \
    "RestoreStackFrame" \
    " \
    : /* no outputs */ \
    : FirstReg (p->wptr), "i" (1), SecondReg (C), ThirdReg (D) \
    : AllRegisters_123 /* corrupted */ \
  )
/* ------------------ ALT CALLS (CHANNEL, TIMEOUT, SKIP) ------------------ */
/* SETUP/SHUTDOWN CALLS */
#if 0
#define AltStart \
  __asm__ __volatile__("\
    "SaveStackFrame" \
    "NoParamCall("X_altstart")" \
    "RestoreStackFrame" \
    " \
    : /* no outputs */ \
    : FirstReg (p->wptr) \
    : AllRegisters_1 /* corrupted */ \
  )
#else
#define Enabling_p NotProcess_p+1
#define TimeNotSet_p NotProcess_p+2
#define State -3
#define TLink -4

#define AltStart p->wptr[State] = Enabling_p

#define TAltStart \
  p->wptr[State] = Enabling_p; \
  p->wptr[TLink] = TimeNotSet_p
#endif


#define AltEnd \
  __asm__ __volatile__("\
    "SaveStackFrame" \
    "NoParamCall("X_altend")" \
    "RestoreStackFrame" \
    " \
    : /* no outputs */ \
    : FirstReg (p->wptr) \
    : AllRegisters_1 /* corrupted */ \
  )
/* WAITING CALLS */
#define AltWait \
  __asm__ __volatile__("\
    "SaveStackFrame" \
    "NoParamCall("X_altwt")" \
    "RestoreStackFrame" \
    " \
    : /* no outputs */ \
    : FirstReg (p->wptr) \
    : AllRegisters_1 /* corrupted */ \
  )

#define TAltWait \
  __asm__ __volatile__("\
    "SaveStackFrame" \
    "NoParamCall("X_taltwt")" \
    "RestoreStackFrame" \
    " \
    : /* no outputs */ \
    : FirstReg (p->wptr) \
    : AllRegisters_1 /* corrupted */ \
  )
/* ENABLE CALLS */
/* paramters are guard, channel and result address */
#define EnbC(G,C,R) \
  __asm__ __volatile__("\
    "SaveStackFrameRV" \
    "TwoParamCallRV("X_enbc")" \
    "GetResult" \
    "RestoreStackFrameRV" \
    " \
    : FirstReturnReg (R) \
    : SecondReg (p->wptr), ThirdReg (G), ForthReg (C) \
    : AllRegisters_1234 /* corrupted */ \
  )

/* paramters are guard, timeout and result address */
#ifdef CPU_TIMERS
#define EnbT(G,T,R) \
  __asm__ __volatile__("\
    "SaveStackFrameRV" \
    "ThreeParamCallRV("X_enbt")" \
    "GetResult" \
    "RestoreStackFrameRV" \
    " \
    : FirstReturnReg (R) \
    : SecondReg (p->wptr), ThirdReg (G), ForthReg (lo(T)), FifthReg (hi(T))\
    : AllRegisters_12345 /* corrupted */ \
  )
#else
#define EnbT(G,T,R) \
  __asm__ __volatile__("\
    "SaveStackFrameRV" \
    "TwoParamCallRV("X_enbt")" \
    "GetResult" \
    "RestoreStackFrameRV" \
    " \
    : FirstReturnReg (R) \
    : SecondReg (p->wptr), ThirdReg (G), ForthReg (T) \
    : AllRegisters_1234 /* corrupted */ \
  )
#endif

/* paramters are guard, and result address */
#define EnbS(G,R) \
  __asm__ __volatile__("\
    "SaveStackFrameRV" \
    "OneParamCallRV("X_enbs")" \
    "GetResult" \
    "RestoreStackFrameRV" \
    " \
    : FirstReturnReg (R) \
    : SecondReg (p->wptr), ThirdReg (G) \
    : AllRegisters_123 /* corrupted */ \
  )
/* DISABLE CALLS*/
/* paramters are Process offset, Guard, Channel, Result address. */
#define DisC(P,G,C,R) \
  __asm__ __volatile__("\
    "SaveStackFrameRV" \
    "ThreeParamCallRV("X_disc")" \
    "GetResult" \
    "RestoreStackFrameRV" \
    " \
    : FirstReturnReg (R) \
    : SecondReg (p->wptr), ThirdReg (P), ForthReg (G), FifthReg (C) \
    : AllRegisters_12345 /* corrupted */ \
  )

/* paramters are Process offset, Guard, Result address. */
#define DisS(P,G,R) \
  __asm__ __volatile__("\
    "SaveStackFrameRV" \
    "TwoParamCallRV("X_diss")" \
    "GetResult" \
    "RestoreStackFrameRV" \
    " \
    : FirstReturnReg (R) \
    : SecondReg (p->wptr), ThirdReg (P), ForthReg (G) \
    : AllRegisters_1234 /* corrupted */ \
  )

/* paramters are Process offset, Guard, Timeout, Result address. */
#ifdef CPU_TIMERS
#define DisT(P,G,T,R) \
  __asm__ __volatile__("\
    "SaveStackFrameRV" \
    "FourParamCallRV("X_dist")" \
    "GetResult" \
    "RestoreStackFrameRV" \
    " \
    : FirstReturnReg (R) \
    : SecondReg (p->wptr), ThirdReg (P), ForthReg (G), FifthReg (lo(T)), SixthReg (hi(T)) \
    : AllRegisters_123456 /* corrupted */ \
  )
#else
#define DisT(P,G,T,R) \
  __asm__ __volatile__("\
    "SaveStackFrameRV" \
    "ThreeParamCallRV("X_dist")" \
    "GetResult" \
    "RestoreStackFrameRV" \
    " \
    : FirstReturnReg (R) \
    : SecondReg (p->wptr), ThirdReg (P), ForthReg (G), FifthReg (T) \
    : AllRegisters_12345 /* corrupted */ \
  )
#endif
/* ------------------ SIMPLE TIMER CALLS ------------------ */
/* paramter is result address */ 
#ifdef CPU_TIMERS
#define LdTimer(R) \
  __asm__ __volatile__("\
    "SaveStackFrameRV64" \
    "NoParamCallRV("X_ldtimer")" \
    "Get64Result" \
    "RestoreStackFrameRV64" \
    " \
    : FirstReturnReg (lo(R)), SecondReturnReg (hi(R)) \
    : ThirdReg (p->wptr) \
    : AllRegisters_123 /* corrupted */ \
  )
#else
#define LdTimer(R) \
  __asm__ __volatile__("\
    "SaveStackFrameRV" \
    "NoParamCallRV("X_ldtimer")" \
    "GetResult" \
    "RestoreStackFrameRV" \
    " \
    : FirstReturnReg (R) \
    : SecondReg (p->wptr) \
    : AllRegisters_12 /* corrupted */ \
  )
#endif
/* parameter is Timeout */
#ifdef CPU_TIMERS
/* 64 bit version */
#define Tin(T) \
  __asm__ __volatile__("\
    "SaveStackFrame" \
    "TwoParamCall("X_tin")" \
    "RestoreStackFrame" \
    " \
    : /* no outputs */ \
    : FirstReg (p->wptr), SecondReg (lo(T)), ThirdReg (hi(T)) \
    : AllRegisters_123 /* corrupted */ \
  )
#else
#define Tin(T) \
  __asm__ __volatile__("\
    "SaveStackFrame" \
    "OneParamCall("X_tin")" \
    "RestoreStackFrame" \
    " \
    : /* no outputs */ \
    : FirstReg (p->wptr), SecondReg (T) \
    : AllRegisters_12 /* corrupted */ \
  )
#endif

/* --------------------- Simple Semaphores -------------------*/
#define ClaimSemaphore(S) \
  __asm__ __volatile__("\
    "SaveStackFrame" \
    "OneParamCall("X_C_claim_semaphore")" \
    "RestoreStackFrame" \
    " \
    : /* no outputs */ \
    : FirstReg (p->wptr), ThirdReg (S)  \
    : AllRegisters_13 /* corrupted */ \
  )

#define ReleaseSemaphore(S) \
  __asm__ __volatile__("\
    "SaveStackFrame" \
    "OneParamCall("X_C_release_semaphore")" \
    "RestoreStackFrame" \
    " \
    : /* no outputs */ \
    : FirstReg (p->wptr), ThirdReg (S)  \
    : AllRegisters_13 /* corrupted */ \
  )

#define ReleaseAllSemaphore(S) \
  __asm__ __volatile__("\
    "SaveStackFrame" \
    "OneParamCall("X_C_release_all_semaphore")" \
    "RestoreStackFrame" \
    " \
    : /* no outputs */ \
    : FirstReg (p->wptr), ThirdReg (S)  \
    : AllRegisters_13 /* corrupted */ \
  )

/* ------------------------ Priorities -----------------------*/
#ifdef PRIORITIES
/* Used inside ProcPriRun, ProcPriPar to set the priority level, not a user
 * macro.
 */
#define SetPriorityNoReschedule(P,V) (P)->workspace[CPriority]=(V)

#define NotOnRunQ(process) ((process)->workspace[COnRunQ])

#define IndirectMacroProcSetPriority(P,V) (P)->workspace[CPriority]=(V)
#define IndirectMacroProcGetPriority(P) ((P)->workspace[CPriority])
#endif

/* The following macro definitions use the GCC specific variable
 * number of arguments in macros feature.  See the documentation
 * (type 'info gcc' ) and go to the section:
 * C Extensions: Macro Varargs
 * I tried to avoid using this feature but it proved to expensive
 * in that it would have made scaling to multiprocessor machines 
 * difficult.
 */
#define ProcAlt(args...) MProcAlt(p, ## args)

#define ProcPrecondAlt(args...) MProcPrecondAlt(p, ## args)

#define ProcTimerAlt(args...) MProcTimerAlt(p, ## args)

#define ProcPrecondTimerAlt(args...) MProcPrecondTimerAlt(p, ## args)

#define ProcJoin(args...) MProcJoin(p, ## args)

#define ProcPar(args...) MProcPar(p, ## args) 

#define ProcPriPar(args...) MProcPriPar(p, ## args) 

#define ProcRun(args...) MProcRun(p, ## args)
 
#define ProcPriRun(args...) MProcPriRun(p, ## args)

#define ProcReschedule() IndirectMacroProcReschedule(p)

#define ProcStop() IndirectMacroProcStop(p)

#define ProcServiceExternals() if (TestExternal_Poll())  ProcReschedule()
/*MProcServiceExternals(p)*/

#ifdef PRIORITIES
  #if 0
    #define ProcSetPriority(V) IndirectMacroProcSetPriority(p, V)
  #else
    #define ProcSetPriority(V) SetOwnPriority(V)
  #endif
  #define ProcGetPriority() IndirectMacroProcGetPriority(p)

  /* These functions are not safe to use on running processes */
  #if 0
    #define ProcSetProcPriority(P,V) IndirectMacroProcSetPriority(P, V)
  #else
    #define ProcSetProcPriority(P,V) SetArbitraryPriority(P,V)
  #endif
  #define ProcGetProcPriority(P) IndirectMacroProcGetPriority(P)
#endif

#ifdef DONT_INLINE_INPUT
#define ChanInChar(C) MChanInChar(p, C)
#else
#define ChanInChar(C) ( \
  { \
    int temp; \
    IndirectMacroChanInChar(C, &temp); \
    temp; \
  } \
)
#endif

#ifdef DONT_INLINE_INPUT    
#define ChanInInt(C) MChanInInt(p, C)
#else
#define ChanInInt(C) ( \
  { \
    int temp; \
    IndirectMacroChanInInt(C, &temp); \
    temp; \
  } \
)
#endif

#define SemClaim(S) MSemClaim(p, S)
#define SemRelease(S) MSemRelease(p, S)
#define SemReleaseAll(S) MSemReleaseAll(p, S)
void MProcPar (Process *,Process *, ...); 

#ifdef PRIORITIES
void MProcPriPar (Process *,Process *, int, ...); 
#endif

Process *ProcAlloc (void (*func)(), int, int, ...);

Channel *ChanAlloc (void);
void ChanAllocClean(Channel *);
void ChanInit(Channel *);
int MChanInInt(Process *, Channel *);
unsigned char MChanInChar(Process *, Channel *);
#ifdef PRIORITIES
void MProcPriRun(Process *, Process *, int);
#endif
void MProcRun(Process *, Process *);
void MProcJoin(Process *, Process *, ...);
int MProcAlt(Process *, Channel *, ...);
int MProcPrecondAlt(Process *, Channel *, bool, ...);
int MProcTimerAlt(Process *, Time, Channel *, ...);
int MProcPrecondTimerAlt(Process *, Time, Channel *, bool, ...);
Process *ProcInit(void (*)(), word *, int, int, ...);
void ProcParam(Process *, ...);
Process *ProcAlloc(void (*)(), int, int, ...);
void ProcAllocClean(Process *);
void ProcInitClean(Process *);
#ifdef EXTERNAL_COMMS
void MProcServiceExternals(Process *p);
#endif

void SemInit(Semaphore *, int);
Semaphore *SemAlloc(int);
void SemAllocClean(Semaphore *);
void MSemClaim(Process *, Semaphore *);
void MSemRelease(Process *, Semaphore *);
void MSemReleaseAll(Process *, Semaphore *); 
Time FProcTime(Process *);

#endif /* csplib.h */
