/*
 *	External channel interface
 *	Copyright (C) 1998 Jim Moores
 *
 *	This program is free software; you can redistribute it and/or modify
 *	it under the terms of the GNU General Public License as published by
 *	the Free Software Foundation; either version 2 of the License, or
 *	(at your option) any later version.
 *
 *	This program is distributed in the hope that it will be useful,
 *	but WITHOUT ANY WARRANTY; without even the implied warranty of
 *	MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *	GNU General Public License for more details.
 *
 *	You should have received a copy of the GNU General Public License
 *	along with this program; if not, write to the Free Software
 *	Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 */

/* the normal_channel should contain the value 1 to indicate that it is an 
 * external channel.  The internal_channel word is then used to hold a pointer
 * to a workspace that appears to be that of the ouputting process.  This
 * workspace's State field is used to hold the ALT state.
 */
#ifndef EXTERNAL_H
#define EXTERNAL_H

#include "ukcthreads_types.h"

typedef struct {
  Channel normal_channel;
  Channel internal_channel;
  #ifdef CPU_TIMERS
  /* an extra word for the time (Wptr[-6]) */
  word workspace[17]; /* pseudo process workspace for channel */
  #else
  word workspace[16]; /* pseudo process workspace for channel */
  #endif
  char *p; /* point to the data */
  int len; /* length of data */
  /* ... */
} External;

/* as we use a reduced workspace descriptor in an external channel structure the
 * stack and frame pointer save fields are in different places
 */
#define ExternalStackPtrRegSave "4"
#define ExternalFramePtrRegSave "8"

/* The following definition is just for testing purposes, I needed to pass
 * some extra parameters in to simplify the implementation of the test rig
 */
External *ExternalAlloc(int chan_num, int direction, char *specification);

void ExternalInit(External* ex, char* specification); 

void ExternalInitClean(External* ex); 

void ExternalAllocClean(External *ex);

void External_DoIn(External* ex); 

void External_DoneIn(External* ex);

void External_PerformedIn(External* ex); 

void External_DoOut(External* ex); 

void External_PerformedOut(External* ex); 

bool External_Poll();

#define TestExternal_Poll() (sf.combined_sync) 

void External_HandleExternals(); 

void External_Terminate();

#endif /* external.h */
