/*
 *	Machine dependant definitions
 *	Copyright (C) 1998 Jim Moores
 *
 *	This program is free software; you can redistribute it and/or modify
 *	it under the terms of the GNU General Public License as published by
 *	the Free Software Foundation; either version 2 of the License, or
 *	(at your option) any later version.
 *
 *	This program is distributed in the hope that it will be useful,
 *	but WITHOUT ANY WARRANTY; without even the implied warranty of
 *	MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *	GNU General Public License for more details.
 *
 *	You should have received a copy of the GNU General Public License
 *	along with this program; if not, write to the Free Software
 *	Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 */

#ifndef I386_H
#define I386_H

/* START OF MACHINE DEPENDENT DEFINITIONS
 * --------------------------------------
 * It should be possible to only change the following definitions to port
 * this library to a different architecture.  SaveStackFrame and
 * RestoreStackFrame should save any registers that GCC cannot mark as
 * corrupt by itself.  This will usually be the system stack and frame
 * pointers which should be saved in the process descriptor.
 * The other macros describe the kernel calling convention.  The kernel
 * does not use the standard 386 calling convention of the caller adjusting
 * the stack after the call, instead the called code deals with this.
 * On, for instance a RISC machine, the paramters can be loaded into
 * registers instead of pushed onto the stack.  That is the reason there
 * is a separate call for each number of paramters.  Return paramters can
 * be dealt with as required too.  This work would of course need to be
 * closely coordinated with a kernel port.
 */

/* CLAIM and RELEASE of binary global spin-lock */
#ifdef GSL_SMP
extern word global_spin_lock;
#define CLAIM(SEMA) \
  __asm__ __volatile__ ( \
    "   movl    $1, %%eax\n" \
    "1: xchg    %%eax, %0\n" \
    "   cmpl    $0, %%eax\n" \
    "   jne     1b\n" \
    : /* no outputs */ \
    : "m" (SEMA) \
    : "eax" \
  )

#define RELEASE(SEMA) \
  __asm__ __volatile__ ( \
    "   movl   $0, %0" \
    : /* no outputs */ \
    : "m" (SEMA) \
  )

#define KCLAIM "\
     movl    $1, %%eax\n \
  1: xchg    %%eax, global_spin_lock\n \
     cmpl    $0, %%eax\n \
     jne     1b\n \
     "

#define KRELEASE "\
     movl    $0, global_spin_lock\n \
     " 
#else
#define KCLAIM
#define KRELEASE
#endif
/* If the stack frame save/restore macros are to be used, they assume that
 * the workspace pointer value is passed in reg %0 (ie the first thing on the
 * inputs list) if there is no return value, and %1 if there is a return
 * value.
 */

/* Each macro has two versions - one to be used if the kernel function does
 * not return a value and one that does.  This is unfortunately necessary
 * because the list of output operands comes before the list of input
 * operands, and operand numbers start from the first one parsed. The macros
 * postfixed RV get Returned Values.  Some have identical bodies for both
 * versions, but are included so the macro can be written with all calls
 * either postfixed or not, rather than a mix that needs to be remembered.
 */
#define SaveStackFrame "\
    movl %%ebp, "FramePtrRegSave"(%0)\n \
    movl %%esp, "StackPtrRegSave"(%0)\n \
    movl %0, %%ebp\n \
    "
#define SaveStackFrameRV "\
    movl %%ebp, "FramePtrRegSave"(%1)\n \
    movl %%esp, "StackPtrRegSave"(%1)\n \
    movl %1, %%ebp\n \
    "
#define SaveStackFrameRV64 "\
    movl %%ebp, "FramePtrRegSave"(%2)\n \
    movl %%esp, "StackPtrRegSave"(%2)\n \
    movl %2, %%ebp\n \
    "
#define RestoreStackFrame "\
    movl "StackPtrRegSave"(%%ebp), %%esp\n \
    movl "FramePtrRegSave"(%%ebp), %%ebp\n \
    "
#define RestoreStackFrameRV "\
    movl "StackPtrRegSave"(%%ebp), %%esp\n \
    movl "FramePtrRegSave"(%%ebp), %%ebp\n \
    "
#define RestoreStackFrameRV64 "\
    movl "StackPtrRegSave"(%%ebp), %%esp\n \
    movl "FramePtrRegSave"(%%ebp), %%ebp\n \
    "

#define NoParamCall(X) "\
    "KCLAIM"
    call "X"\n \
    "KRELEASE"
    "
#define NoParamCallRV(X) "\
    "KCLAIM"
    call "X"\n \
    "KRELEASE"
    "
#define OneParamCall(X) "\
    pushl %1\n \
    "KCLAIM"
    call "X"\n \
    "KRELEASE"
    "
#define OneParamCallRV(X) "\
    pushl %2\n \
    "KCLAIM"
    call "X"\n \
    "KRELEASE"
    "
#define TwoParamCall(X) "\
    pushl %2\n \
    pushl %1\n \
    "KCLAIM"
    call "X"\n \
    "KRELEASE"
    "
#define TwoParamCallRV(X) "\
    pushl %3\n \
    pushl %2\n \
    "KCLAIM"
    call "X"\n \
    "KRELEASE"
    "
#define ThreeParamCall(X) "\
    pushl %3\n \
    pushl %2\n \
    pushl %1\n \
    "KCLAIM"
    call "X"\n \
    "KRELEASE"
    "
#define ThreeParamCallRV(X) "\
    pushl %4\n \
    pushl %3\n \
    pushl %2\n \
    "KCLAIM"
    call "X"\n \
    "KRELEASE"
    "

#ifdef CPU_TIMERS
/* needed for 64bit interface */
#define FourParamCall(X) "\
    pushl %4\n \
    pushl %3\n \
    pushl %2\n \
    pushl %1\n \
    "KCLAIM"
    call "X"\n \
    "
#define FourParamCallRV(X) "\
    pushl %5\n \
    pushl %4\n \
    pushl %3\n \
    pushl %2\n \
    "KCLAIM"
    call "X"\n \
    "
#endif

/* This macro is for extracting the result from kernel functions */
#define GetResult "\
    popl %0\n \
    "

#ifdef CPU_TIMERS
/* This macro is for extracting a 64bit result from a kernel function
 * It should not be used with a mulitple parameter call (*ParamCall[RV]).
 */
#define Get64Result "\
    popl %0\n \
    popl %1\n \
    "
#endif

/* For calls from the communications subsystem, you don't need to do
 * a context save as such because none of the kernel calls block.  We
 * could just use the normal mechanism, but this means every external
 * channel must have a full process descriptor instead of just a small
 * dummy workspace.  I've therefore added these extra macros to support
 * a 'cut-down' context save on the stack.
 */
#define ExternalSaveStackFrame "\
    pushl %%ebp\n \
    movl %0, %%ebp\n \
    "
#define ExternalRestoreStackFrame "\
    pop %%ebp\n \
    movl %0, %%ebp\n \
    "

/* List of registers to tell compiler are corrupted over a context switch
 * Note that references to esp and ebp (stack and frame pointers) appear
 * to be ignored (even with -fomit-frame-pointer), and so need to be saved
 * The st and st(x) registers are the floating point stack.  st is referred
 * to as st(0) in some GNU documentation.  These names were gleaned by looking
 * at the 386 specific files in the GCC compiler sources.
 */
#define AllRegistersThree "0","1","2","dx","si","di","memory","cc", \
 "st", "st(1)", "st(2)", "st(3)", "st(4)", "st(5)", "st(6)", "st(7)"
#define AllRegisters "ax","bx","cx","dx","si","di","memory","cc", \
 "st", "st(1)", "st(2)", "st(3)", "st(4)", "st(5)", "st(6)", "st(7)"
#define AllRegisters_1 "bx","cx","dx","si","di","memory","cc", \
 "st", "st(1)", "st(2)", "st(3)", "st(4)", "st(5)", "st(6)", "st(7)"
#define AllRegisters_12 "cx","dx","si","di","memory","cc", \
 "st", "st(1)", "st(2)", "st(3)", "st(4)", "st(5)", "st(6)", "st(7)"
#define AllRegisters_13 "bx","dx","si","di","memory","cc", \
 "st", "st(1)", "st(2)", "st(3)", "st(4)", "st(5)", "st(6)", "st(7)"
#define AllRegisters_123 "dx","si","di","memory","cc", \
 "st", "st(1)", "st(2)", "st(3)", "st(4)", "st(5)", "st(6)", "st(7)"
#define AllRegisters_1234 "si","di","memory","cc", \
 "st", "st(1)", "st(2)", "st(3)", "st(4)", "st(5)", "st(6)", "st(7)"
#define AllRegisters_12345 "si","memory","cc", \
 "st", "st(1)", "st(2)", "st(3)", "st(4)", "st(5)", "st(6)", "st(7)"
#define AllRegisters_2345 "ax","si","memory","cc", \
 "st", "st(1)", "st(2)", "st(3)", "st(4)", "st(5)", "st(6)", "st(7)"
#define AllRegisters_123456 "memory","cc", \
 "st", "st(1)", "st(2)", "st(3)", "st(4)", "st(5)", "st(6)", "st(7)"

#define FirstReg "ax"
#define SecondReg "bx"
#define ThirdReg "cx"
#define ForthReg "dx"
#define FifthReg "DI"
#define SixthReg "SI"

#define FirstReturnReg "=eax"
#define SecondReturnReg "=ebx"

/* END OF MACHINE DEPENDENT DEFINITIONS */


#endif /* i386.h */
