/*
 *	Kernel macros associated with external channels
 *	Copyright (C) 1998 Jim Moores
 *
 *	This program is free software; you can redistribute it and/or modify
 *	it under the terms of the GNU General Public License as published by
 *	the Free Software Foundation; either version 2 of the License, or
 *	(at your option) any later version.
 *
 *	This program is distributed in the hope that it will be useful,
 *	but WITHOUT ANY WARRANTY; without even the implied warranty of
 *	MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *	GNU General Public License for more details.
 *
 *	You should have received a copy of the GNU General Public License
 *	along with this program; if not, write to the Free Software
 *	Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 */

/* Macros for use in the kernel associated with external channels */
#ifndef K_EXTERN_MACROS_H
#define K_EXTERN_MACROS_H
/* define magic value that indicates channel is external */
#define External_Tag ((word *)1)

/* MACRO IsExternal(Channel *) returns true if the channel is external 
 */
#define IsExternal(c) (*((word *)(c)) == 1)

/* MACRO GetAddrInternal(External *) gets the address of the internal channel 
 *       part of the external channel structure.  Must only be used on
 *       external channels (ie where IsExternal(c) == true)
 */
#define GetAddrInternal(c) (&((External *)(c))->internal_channel)

/* MACRO GetInternal(External *) gets the value in the internal channel 
 *       part of the external channel structure.  Must only be used on
 *       external channels (ie where IsExternal(c) == true)
 */
#define GetInternal(c) (((External *)(c))->internal_channel)

/* MACRO GetAddrWorkspace(External *) gets the address of the _top_ of
 *       workspace field in the external channel structure.  This means that
 *       its negative offsets can be indexed directly.
 */ 
#ifdef CPU_TIMERS
  #define GetAddrWorkspace(c) (&((External *)(c))->workspace[6])
#else
  #define GetAddrWorkspace(c) (&((External *)(c))->workspace[5])
#endif
#endif /* k_extern_macros.h */
