/*
 * KROC translator: Pseudo-directives for debugging etc.
 *
 * $Source: /extra/play/cvs/kroc/src/host/f_pseudo.c,v $
 *
 * $Id: f_pseudo.c,v 1.1 1999/09/27 21:10:12 rivimey Exp $
 *
 *
 * (c) Copyright 1999 R. Ivimey-Cook <ruthc@sharra.demon.co.uk>
 *
 * (C) Copyright 1995,1996,1997 D.C. Wood <D.C.Wood@ukc.ac.uk>
 * University of Kent at Canterbury
 *   /proj/kroc/develop/src/RCS/f_pseudo.c,v
 *   f_pseudo.c,v 1.18 1997/01/08 17:23:31 djb1 Exp
 *
 * Mostly independent of target.
 */

static char f_pseudo_c_rcsid[] = "$Id: f_pseudo.c,v 1.1 1999/09/27 21:10:12 rivimey Exp $";

#include "octran.h"

void dcheck (int max) 
{
    if (n_operand < 0 || n_operand > max) {
        fputs ("\n*** Invalid operand", stderr);
        error ();
    }
}

void message (const char *s, int m, int n)
{
    fprintf (out, "%c\t%s\t%d -> %d", CCHAR, s, m, n);
    NL;
}

void _CHECK (void)
{
    message ("checks", o_checks, n_operand);
    dcheck (3);
    o_checks = n_operand;
}

void _COND (void)
{
    message ("condopt", o_opt, n_operand);
    dcheck (1);
    o_opt = n_operand;
}

void _ECHO (void)
{
    message ("echo", o_echo, n_operand);
    dcheck (1);
    o_echo = n_operand;
}

void _STACK (void)
{
    message ("istack", istackp, n_operand);
    dcheck (3);
    istackp = n_operand;
}

void _STATE (void)
{
    int i;
    fprintf (stderr, "\n... state = %d, [", istackp);
#if defined(FLOAT)
    for (i = fstackp; i > 0; --i)
        fputc ((ftype[i] == F_D) ? 'D' : 'S', stderr);
#else
    fputc ('S', stderr);
#endif
    fprintf (stderr, "] %s\n", b_operand);
    where ();
}

void _STIME (void)
{
    message ("time", o_time, n_operand);
    dcheck (1);
    o_time = n_operand;
}


#if defined(FLOAT)

static void fcheck (void);

static void fcheck (void)
{
    if (fstackp == 0) {
        fputs ("\n*** Floating-point stack empty", stderr);
        error ();
    }
}

void _DOUBLE (void)
{
    fcheck ();
    message ("ftype", FAtype, F_D);
    FAtype = F_D;
}

void _FSTACK (void)
{
    message ("fstack", fstackp, n_operand);
    dcheck (3);
    fstackp = n_operand;
}

void _FTYPE (void)
{
    fcheck ();
    message ("ftype", FAtype, n_operand);
    dcheck (3);
    FAtype = n_operand;
}

void _SINGLE (void)
{
    fcheck ();
    message ("ftype", FAtype, F_S);
    FAtype = F_S;
}

void _FDUMP (void)
{
    int i, temp = 0;
    for (i = fstackp; i > 0; --i)
        temp = (temp << 1) | ftype[i];
    temp = (temp << 2) | fstackp;

#if defined(TARGET_CPU_SPARC)
    extcall(X_FDUMP); NL;
    mnem("mov"); dec(temp); P0; DELAY; nl;
    
#elif defined(TARGET_CPU_ALPHA)
    mnem("or"); Z; dec(temp); P0; nl;
    CALL_RTN("$fdump");
        
#endif
}

#endif /* FLOAT */


void _DUMP (void)
{       
#if defined(TARGET_CPU_SPARC)
    extcall(X_DUMP); NL;
    mnem("mov"); dec(istackp); P0; DELAY; nl;
#elif defined(TARGET_CPU_ALPHA)
    mnem("or"); Z; dec(istackp); P0; nl;
    CALL_RTN("$dump");
#elif defined(TARGET_CPU_ARM)
    fprintf (stderr, "\n*** DUMP inimplemented\n");
    error();
#endif
}

void _ERROR (void)
{       
    fprintf (stderr, "\n*** Unrecognized instruction \"%s\"\n", b_symbol);

    if (o_un) {
#if defined(TARGET_CPU_SPARC)
        mnem("ta"); dec(T_ZERO + T_UNREC); COMMENT("unrecognized instruction"); NL;
#elif defined(TARGET_CPU_ALPHA)
        GEN_TRAP(T_ZERO + T_UNREC);
#endif
    } else
        error ();
}

void _ABORT (void)
{
#if defined(TARGET_CPU_SPARC)
    mnem("ld"); str("[%g0+1]"); Z; COMMENT("force core dump"); NL;
#elif defined(TARGET_CPU_ALPHA)
    CLR_Reg(T0)
        mnem("ldq"); Z; index1(Temp0, 1); COMMENT("force core dump"); nl;
#endif
}

void unimp (void)
{       
    fprintf (stderr, "\n*** Unimplemented instruction \"%s\"\n", b_symbol);
    if (o_un) {
#if defined(TARGET_CPU_SPARC)
        mnem("ta"); dec(T_ZERO + T_UNIMP);
#elif defined(TARGET_CPU_ALPHA) || defined(TARGET_CPU_ARM)
        GEN_TRAP(T_ZERO + T_UNIMP);
#endif
        COMMENT("unimplemented instruction"); NL;
    } else
        error ();
}
