
D                       [0-9]
L                       [a-zA-Z_]
H                       [a-fA-F0-9]
E                       [Ee][+-]?{D}+
FS                      (f|F|l|L)
IS                      (u|U|l|L)*

%{
#include "config.h"
#include "ocinf.h"
#include <stdio.h>
#include "y.tab.h"

void count (void);
int yywrap (void);
void comment (void);
%}

%%
"/*"                    { comment(); }

"auto"                  { count(); return(AUTO); }
"break"                 { count(); return(BREAK); }
"case"                  { count(); return(CASE); }
"char"                  { count(); return(CHAR); }
"const"                 { count(); return(CONST); }
"continue"              { count(); return(CONTINUE); }
"default"               { count(); return(DEFAULT); }
"do"                    { count(); return(DO); }
"double"                { count(); return(DOUBLE); }
"else"                  { count(); return(ELSE); }
"enum"                  { count(); return(ENUM); }
"extern"                { count(); return(EXTERN); }
"float"                 { count(); return(FLOAT); }
"for"                   { count(); return(FOR); }
"goto"                  { count(); return(GOTO); }
"if"                    { count(); return(IF); }
"int"                   { count(); return(INT); }
"long"                  { count(); return(LONG); }
"register"              { count(); return(REGISTER); }
"return"                { count(); return(RETURN); }
"short"                 { count(); return(SHORT); }
"signed"                { count(); return(SIGNED); }
"sizeof"                { count(); return(SIZEOF); }
"static"                { count(); return(STATIC); }
"struct"                { count(); return(STRUCT); }
"switch"                { count(); return(SWITCH); }
"typedef"               { count(); return(TYPEDEF); }
"union"                 { count(); return(UNION); }
"unsigned"              { count(); return(UNSIGNED); }
"void"                  { count(); return(VOID); }
"volatile"              { count(); return(VOLATILE); }
"while"                 { count(); return(WHILE); }

{L}({L}|{D})*           { count(); return(check_type()); }

0[xX]{H}+{IS}?          { count(); yylval.con.C_type = INT; yylval.con.C_val.ival = atoi(yytext);
                          return(CONSTANT); }
0{D}+{IS}?              { count(); yylval.con.C_type = INT; yylval.con.C_val.ival = atoi(yytext);
                          return(CONSTANT); }
{D}+{IS}?               { count(); yylval.con.C_type = INT; yylval.con.C_val.ival = atoi(yytext);
                          return(CONSTANT); }
'(\\.|[^\\'])+'         { count(); yylval.con.C_type = INT; yylval.con.C_val.ival = atoi(yytext);
                          return(CONSTANT); }

{D}+{E}{FS}?            { count(); yylval.con.C_type = FLOAT; yylval.con.C_val.fval = atof(yytext); 
                          return(CONSTANT); }
{D}*"."{D}+({E})?{FS}?  { count(); yylval.con.C_type = FLOAT; yylval.con.C_val.fval = atof(yytext);
                          return(CONSTANT); }
{D}+"."{D}*({E})?{FS}?  { count(); yylval.con.C_type = FLOAT; yylval.con.C_val.fval = atof(yytext);
                          return(CONSTANT); }

\"(\\.|[^\\"])*\"       { count(); return(STRING_LITERAL); }

">>="                   { count(); return(RIGHT_ASSIGN); }
"<<="                   { count(); return(LEFT_ASSIGN); }
"+="                    { count(); return(ADD_ASSIGN); }
"-="                    { count(); return(SUB_ASSIGN); }
"*="                    { count(); return(MUL_ASSIGN); }
"/="                    { count(); return(DIV_ASSIGN); }
"%="                    { count(); return(MOD_ASSIGN); }
"&="                    { count(); return(AND_ASSIGN); }
"^="                    { count(); return(XOR_ASSIGN); }
"|="                    { count(); return(OR_ASSIGN); }
">>"                    { count(); return(RIGHT_OP); }
"<<"                    { count(); return(LEFT_OP); }
"++"                    { count(); return(INC_OP); }
"--"                    { count(); return(DEC_OP); }
"->"                    { count(); return(PTR_OP); }
"&&"                    { count(); return(AND_OP); }
"||"                    { count(); return(OR_OP); }
"<="                    { count(); return(LE_OP); }
">="                    { count(); return(GE_OP); }
"=="                    { count(); return(EQ_OP); }
"!="                    { count(); return(NE_OP); }
";"                     { count(); return(';'); }
"{"                     { count(); return('{'); }
"}"                     { count(); return('}'); }
","                     { count(); return(','); }
":"                     { count(); return(':'); }
"="                     { count(); return('='); }
"("                     { count(); return('('); }
")"                     { count(); return(')'); }
"["                     { count(); return('['); }
"]"                     { count(); return(']'); }
"."                     { count(); return('.'); }
"&"                     { count(); return('&'); }
"!"                     { count(); return('!'); }
"~"                     { count(); return('~'); }
"-"                     { count(); return('-'); }
"+"                     { count(); return('+'); }
"*"                     { count(); return('*'); }
"/"                     { count(); return('/'); }
"%"                     { count(); return('%'); }
"<"                     { count(); return('<'); }
">"                     { count(); return('>'); }
"^"                     { count(); return('^'); }
"|"                     { count(); return('|'); }
"?"                     { count(); return('?'); }

[ \t\v\n\f]             { count(); }
.                       { /* ignore bad characters */ }

%%

extern BOOLEAN  ECHO_SOURCE;

int
yywrap (void)
{
        return(1);
}

void
comment (void)
{
        char c, c1;

loop:
        while ((c = input()) != '*' && c != 0)
        if (ECHO_SOURCE) putchar(c);  

        if ((c1 = input()) != '/' && c != 0)
        {
                if (ECHO_SOURCE) unput(c1);
                goto loop;
        }

        if (ECHO_SOURCE) { if (c != 0) putchar(c1); }
}


int column = 0;

int sourceline = 0;

void count (void)
{
  int i;

  for (i = 0; yytext[i] != '\0'; i++)
    if (yytext[i] == '\n') {
      sourceline++;
      column = 0;
    }
    else if (yytext[i] == '\t')
      column += 8 - (column % 8);
    else
      column++;

  if (ECHO_SOURCE) ECHO;
}

extern BOOLEAN typedef_symbol_table_find (char *);
extern int     typedef_symbol_table_index (char *);
extern int     identifier_symbol_table_add (char *);
extern char   *identifier_symbol_table_ptr (int index);


int check_type (void)
{
   int ret_val;

   if (typedef_symbol_table_find(yytext)) {

      /* TYPE_NAME */
      ret_val = TYPE_NAME;
      yylval.tname.tindex = typedef_symbol_table_index(yytext);
     }

   else

     {
      /* IDENTIFIER */
      ret_val = IDENTIFIER;
      yylval.iname.index = identifier_symbol_table_add(yytext);
     }

   return (ret_val);
}

