/*
 * Kent Retargeted occam Compiler
 *
 * Transputer to native assembly-language translator
 * Main program - process command-line arguments and run translation.
 *
 * $Source: /extra/play/cvs/kroc/src/host/octran.c,v $
 *
 * $Id: octran.c,v 1.1 1999/09/27 21:10:12 rivimey Exp $
 *
 * (C) Copyright 1995,1996,1997 D.C. Wood <D.C.Wood@ukc.ac.uk>
 * University of Kent at Canterbury
 *
 * Some options may depend on the target.
 *
 */

static char octran_c_rcsid[] = "$Id: octran.c,v 1.1 1999/09/27 21:10:12 rivimey Exp $";

#include "octran.h"
#include "common.h"
#include <string.h>

/* Defined in version.c */
extern char kroc_version_message[];     
extern char kroc_build_number[];     
extern char host_string[];

const char *program_name;

/* Prototypes */
static int param (const char *arg[], int i, int j);


/* check parameter to option */
static int param (const char *arg[], int i, int j)
{
  int k;
  k = arg[i][2] - '0';
  if (k < 0 || k > j) {
    fprintf (stderr, "%s: Illegal option '%s'\n", program_name, arg[i]);
    exit (1);
  }
  return k;
}


int cflag;                              /* print comment? */
int n_operand;                          /* numerical operand */
int nflag;                              /* "n_operand" occupied */
int posn;                               /* position on output line */
int cond;                               /* pending condition for CJ */
int lcount;                             /* count of LX labels */

char buffer[BUF_SIZE];                  /* input buffer */
char b_label[PROC_SIZE];                /* components of source line */
char b_symbol[SYM_SIZE];
char b_operand[PROC_SIZE];              /* string operand */
char b_proc[PROC_SIZE];
char b_entry[PROC_SIZE];
char b_comment[BUF_SIZE];

char l_label[PROC_SIZE];                /* last label */
char l_operand[SYM_SIZE];               /* last operand */
char l_left[SYM_SIZE];                  /* L<left> - L<right> */
char l_right[SYM_SIZE];
char l_comment[BUF_SIZE];

char filename[FILE_SIZE];
char lastname[FILE_SIZE];

int main (int argc, char *argv[]) /* read command-line argments */
{
  int i, flag;
  char ch;

  /* initialise the I/O streams */
  in = stdin;
  out = stdout;
  
  program_name = basename ((char*)argv[0]);
  for (i = 1; i < argc; ++i) {
    ch = argv[i][0];

    if (ch == '+' || ch == '-') { /* turn on or off */
      flag = ch == '+';

      switch (argv[i][1]) {
        case 'k':               /* use jump table to enter kernel */
          o_kjumptable =flag;
          break;
        case 'c':               /* error checking level */
          o_checks = param ((const char **)argv, i, 3);
          break;
        case 'o':               /* optimize conditionals */
          o_opt = flag;
          break;
        case 'h':               /* hardware/software arithmetic */
          o_hardware = flag;    /* True => hardware */
          break;
        case 't':               /* real/user time */
          o_time = flag;        /* True => real time */
          break;
        case 'v':               /* verbose mode */
          o_verbose = flag;
          break;
        case 'l':               /* translate as a library */
          o_lib = flag;
          break;
        case 'g':               /* make PROC names global */
          o_global = flag;
          break;
        case 'u':               /* translate unrecognized/unimplemented */
          o_un = flag;          /* instructions as traps */
          break;
        case 'n':               /* line numbers */
          o_lines = flag;
          break;
        case 'e':               /* echo source as comments */
          o_echo = flag;
          break;
        case 'r':               /* reverse endianism */
          o_reverse = flag;
          break;
        case 's':               /* occam workspace on C stack */
          o_stackws = flag;
          break;
        case 'a':               /* abort translation on error */
          o_abort = flag;
          break;
        case '%':               /* truncate PROC names after % */
          o_trunc = flag;
          break;
        case 'd':               /* generate debugging directives */
          o_debug = flag;
          break;
        case 'M':               /* Michael Poole's special comments */
          o_mdp = flag;
          break;
        case 'R':		/* big-endian REAL64s */
          o_real64 = flag;
          break;
        case 'L':		/* optimize label differences */
          o_label = flag;
          break;
        case 'V':		/* FPtemp etc. relative to Vptr */
          o_vptr = flag;
          break;
        case 'S':		/* Reschedule on jumps */
          o_resched = flag;
          break;
        default:
          fprintf (stderr, "%s: Unknown flag '%s'\n",
                   program_name, argv[i]);
          exit (1);
      }
    } else { /* isn't on or off option - expect filenames */
      if (in == stdin) {
        source = (char*)argv[i];
        in = fopen (source, "r");
        if (in == NULL) {
          fprintf (stderr, "%s: Cannot open file '%s' - ",
                   program_name, source);
          perror (NULL);
          exit (1);
        }
      } else if (out == stdout) {
        dest = (char*)argv[i];
        if (strcmp (source, dest) == 0) {
          fprintf (stderr, "%s: Input and output files the same\n",
                   program_name);
          exit (1);
        }
        out = fopen (dest, "w");
        if (out == NULL) {
          fprintf (stderr, "%s: Cannot open file '%s' - ",
                   program_name, dest);
          perror (NULL);
          exit (1);
        }
      } else {
        fprintf (stderr, "Files '%s' and '%s' already opened\n",
                 source, dest);
        exit (1);
      }
    }
  }
  o_echo |= o_lines;      /* implies echo */
  /* if (o_verbose) */
#if defined(RELEASE_VERSION)
  fprintf (stderr, "%s for %s\n", kroc_version_message, host_string);
#else
  fprintf (stderr, "%s build %s for %s\n", kroc_version_message, kroc_build_number, host_string);
#endif
  parse ();
  exit (0);
}
