/*
 * KROC translator: Function definitions and prototypes
 *
 * $Source: /extra/play/cvs/kroc/src/host/octran.h,v $
 *
 * $Id: octran.h,v 1.1 1999/09/27 21:10:12 rivimey Exp $
 *
 * (C) Copyright 1995,1996,1997 D.C. Wood <D.C.Wood@ukc.ac.uk>
 * University of Kent at Canterbury
 */

#include <stdio.h>
#include <stdlib.h>
#include <unistd.h>

#include "config.h"

#undef FLOAT                           /* don't include floating-point */


/* **************************************************************** */

#include "f_proto.h"

/* **************************************************************** */

/*      General-purpose definitions.            */
/*      Should be independent of target.        */

#define True            (1==1)          /* C Booleans */
#define False           (!True)

#define PUSH            push ()
#define POP             pop (1)
#define nl              fputc ('\n', out); oline += 1
#define NL              newline (); nl;
#define CHECK           check (C_VS, T_OFLOW)
#if defined(TARGET_CPU_SPARC) || defined(TARGET_CPU_ARM)
#define BREAK           fprintf (out, "%c\t------------------------\n", CCHAR)
#elif defined(TARGET_CPU_ALPHA)
#define BREAK           fprintf (out, "%c\t----------------\n", CCHAR1)
#endif
#define LABEL           label ()

/*      Checking levels                 */

#define CH_NONE         0               /* typedef enum? */
#define CH_SETFLAG      1
#define CH_TRAP         2
#define CH_HALT         3

/*      Conditionals                            */
/*      Must be in true/false even/odd pairs    */

#include "cond.h"


#define INV(C)          ((C)^1)         /* toggle l.s. bit */

/*	Flags in instruction look-up table	*/

#define F_OPERAND	1		/* instruction has an operand */
#define F_CJ		2
#define F_EQC		4
#define F_O_LABEL	8

/*      Input buffer                    */

#define BUF_SIZE        500             /* generous! */
#define SYM_SIZE        20              /* longest should be MOVE2DNONZERO */
#define PROC_SIZE       50              /* "abbrs.of.function.results.and.params"!! */
#define FILE_SIZE       100             /* could be very/long/unix/path/name */

#define B_LABEL         0               /* items in buffer */
#define B_INSTRUCTION   8
#define B_OPERAND       16
#define B_PROC          32
#define B_BYTE          13              /* .BYTE/byte operand starts early */
#define B_FILENAME      12

#define N_PROC          8               /* "-- PROC " */
#define N_FUNC          12              /* "-- FUNCTION " */
#define N_STUB          25              /* "-- Stub for library call " */
#define N_SECLAB        16              /* "-- section label" */
#define N_LEX           15              /* ", lexlevel: 0, " */

/*      Array sizes                     */

#define LAB_SIZE        8
#define PT_SIZE         1000
#define BYTE_SIZE       16
#define LT_SIZE         10000

/*      Types of PROC/FUNCTION          */

#define P_PROC          0
#define P_FUNCTION      1
#define P_EXTERNAL      2
#define P_C_EXTERNAL    3

/*      Transputer definitions.                 */

/*      Many depend on machine word size -      */
/*      32 bits (T800 etc.) or 16 bits (T200)   */

#define BitsPerWord     32              /* size-dependent */

#define BitsPerByte     8

#define BytesPerWord    (BitsPerWord/BitsPerByte)
#define MostNeg         (1 << (BitsPerWord - 1))

#define ByteSelectMask  (BytesPerWord - 1)
					/* for inverting byte addresses */
#define ShortSelectMask (BytesPerWord - 2)
					/* for inverting short addresses */

/*      Offsets in local workspace      */

#define pw_Temp         0
#define pw_Iptr         (-1)
#define pw_Link         (-2)
#define pw_Pointer      (-3)
#define pw_State        (-3)
#define pw_TLink        (-4)
#define pw_Length       (-4)
#define pw_Time         (-5)

/*      Loop end data structure         */

#define le_Index        0
#define le_Count        1

/*      Parallel process data structure */

#define pp_IptrSucc     0
#define pp_Count        1

/*      Magic values                    */

#define Enabling_d      1               /* differences from MostNeg */
#define Waiting_d       2
#define Ready_d         3

#define TimeSet_d       1
#define TimeNotSet_d    2

#define NotProcess_p    MostNeg

#define Enabling_p      (MostNeg + Enabling_d)
#define Waiting_p       (MostNeg + Waiting_d)
#define Ready_p         (MostNeg + Ready_d)

#define TimeSet_p       (MostNeg + TimeSet_d)
#define TimeNotSet_p    (MostNeg + TimeNotSet_d)

#define NoneSelected_o  (-1)

#define LengthError_p   (-1)


#include "regs.h"

/*      Miscellaneous                   */

/* Badly named */
#undef TRUE
#define TRUE            dec(1)          /* Transputer Booleans */
#undef FALSE
#define FALSE           dec(0)

#define DELAY           delay ("")
#define ANNUL           annul ("")
#define ADJUST          mnem("inc"); dec(8); L; delay("adjust PC")

#if defined(TARGET_CPU_SPARC)
#define MAXO            4095            /* 13-bit operand field (*BytesPerWord) */
#elif defined(TARGET_CPU_ALPHA)
#define MAXO            4095*8          /* 16-bit operand field (*BytesPerWord) */
#endif
#define MINO            (-MAXO)         /* (to avoid special cases) */

#define EF              dec(1)          /* error flag */
#define HOEF            dec(2)          /* halt-on-error flag */


/*      Operands for trap numbers.      */

/*      The Sun-4 Assembly Language Reference Manual says:      */
/*      "Trap numbers 16-31 are available for use by the user,  */
/*      and will not be usurped by Sun".                        */

#define T_ZERO          16              /* Unix-dependent */

#define T_SETERR        0               /* unconditional (SETERR) */
#define T_OFLOW         1               /* arithmetic overflow */
#define T_RANGE         2               /* range check (CSUB0, CCNT1, CSNGL, CWORD) */
#define T_FLOAT         3               /* floating-point (FPCHKERR, FPSETERR) */
#define T_UNREC         4               /* unrecognized instruction */
#define T_UNIMP         5               /* unimplemented instruction */

/*      External functions called from translated occam         */
/*      Must be in the same order as jump table in scheduler    */

#define X_FPCONST       0               /* in kernel */
#define X_OUTBYTE       1
#define X_OUTWORD       2
#define X_OUT           3
#define X_IN            4
#define X_MOVE          5
#define X_STOPP         6
#define X_SCHEDULER     7
#define X_RESCHED       8
#define X_TIN           9
#define X_TALTWT        10
#define X_DIST          11
#define X_VIN           12
#define X_VOUT          13

#define X_DUMP          14
#define X_FDUMP         15

#define X_MUL           16              /* in system */
#define X_DIV           17
#define X_REM           18
#define X_UMUL          19

#define X_RTIME         20              /* in RTS */
#define X_UTIME         21

#define X_REMAINDER     22              /* in math library */

#define X_SIZE          23              /* number of entries */

/*      Data declarations.              */

extern const char *program_name;

/*      Should be independent of target.        */

extern int istackp;                     /* evaluation stack pointer */
extern int fstackp;                     /* f.p. stack pointer */
extern int comma;                       /* between operands */
extern int cflag;                       /* print comment? */
extern int n_operand;                   /* numerical operand */
extern int nflag;                       /* "n_operand" occupied */
extern int iline;                       /* input line number */
extern int oline;                       /* output line number */
extern int fline;                       /* occam source file line number */
extern int posn;                        /* position on output line */
extern int ishuff;                      /* expect stack overflow following ENBC, DISC, etc. */
extern int shufflecount;                /* number of shuffles */
extern int lshufflecount;               /* number of label shuffles */
extern int cond;                        /* pending condition for CJ */
extern int sround;                      /* target rounding mode */
extern int lcount;                      /* count of LX labels */
extern int started;			/* generated any code yet? */

extern FILE *in;                        /* input file */
extern FILE *out;                       /* output file */

extern char *source;
extern char *dest;

extern int ws;                          /* workspace */
extern int vs;                          /* vector space */

extern char buffer[BUF_SIZE];           /* input buffer */
extern char b_label[PROC_SIZE];         /* components of source line */
extern char b_symbol[SYM_SIZE];
extern char b_operand[PROC_SIZE];       /* string operand */
extern char b_proc[PROC_SIZE];
extern char b_entry[PROC_SIZE];
extern char b_comment[BUF_SIZE];

extern char l_label[PROC_SIZE];         /* last label */
extern char l_operand[SYM_SIZE];	/* last operand */
extern char l_left[SYM_SIZE];		/* <left> - <right> */
extern char l_right[SYM_SIZE];
extern char l_comment[BUF_SIZE];

extern char filename[FILE_SIZE];
extern char lastname[FILE_SIZE];

/*      options         */

extern int o_kjumptable;                /* use jump table to enter kernel */
extern int o_checks;                    /* check level */
extern int o_hardware;                  /* hardware/software arithmetic */
extern int o_stackws;                   /* occam workspace on C stack */
extern int o_time;                      /* real time or user time? */
extern int o_opt;                       /* optimize conditions? */
extern int o_verbose;                   /* messages? */
extern int o_echo;                      /* echo source? */
extern int o_lines;                     /* line numbers? */
extern int o_global;                    /* make PROC names global? */
extern int o_un;                        /* trap on unimplemented/unrecognized instructions */
extern int o_reverse;                   /* reverse endianism */
extern int o_lib;                       /* library */
extern int o_abort;                     /* abort on error */
extern int o_trunc;                     /* truncate "%" names */
extern int o_mdp;                       /* Micheal Poole's extra comments */
extern int o_real64;			/* big-endian REAL64s */
extern int o_debug;			/* generate debugging directives */
extern int o_label;			/* optimize label differences */
extern int o_vptr;			/* FPtemp etc. relative to Vptr */
extern int o_resched;			/* Reschedule on jumps */

extern int stcount;                     /* entries in symbol table */

extern const char *ptype[4];            /* PROC/FUNCTION type names */


/*      Type definitions                */

/*      "proctab" and "labtab" could be merged  */

typedef union {                          /* for translator state */
  int i;
  struct {
    unsigned flag   : 1;    /* not S_EMPTY */
    unsigned def    : 1;    /* label defined */
    unsigned istack : 2;
    unsigned ishuff : 2;
    unsigned itypes : 3;    /* for 68000 */
    unsigned fstack : 2;
    unsigned fshuff : 2;    /* not used (yet) */
    unsigned ftypes : 3;
    unsigned ptype  : 2;    /* for pt_type */
  } bits;
} state_type;


typedef struct {                          /* for procedure table */
  char       label[LAB_SIZE];
  char       proc[PROC_SIZE];
  state_type state;
} pt_type;

typedef void (*pfv) (void);

typedef struct {                          /* for instruction table */
  const char *mnem;
  pfv         func;
  char        flags;
} it_type;

typedef struct {                          /* for conditional instructions */
  const char *name;
  const char *branch;
  const char *trap;
} cond_type;

extern pt_type proctab[PT_SIZE];        /* procedure table */

extern cond_type condtab[C_SIZE];       /* conditional instructions */

/*      Members of structures           */

#define B_FLAG          bits.flag
#define B_DEF		bits.def
#define B_ISTACK        bits.istack
#define B_ITYPES        bits.itypes
#define B_ISHUFF        bits.ishuff
#define B_FSTACK        bits.fstack
#define B_FTYPES        bits.ftypes
#define B_FSHUFF        bits.fshuff
#define B_PTYPE         bits.ptype

#define S_FLAG          state.B_FLAG
#define S_DEF		state.B_DEF
#define S_ISTACK        state.B_ISTACK
#define S_ITYPES        state.B_ITYPES
#define S_ISHUFF        state.B_ISHUFF
#define S_FSTACK        state.B_FSTACK
#define S_FTYPES        state.B_FTYPES
#define S_FSHUFF        state.B_FSHUFF
#define S_PTYPE         state.B_PTYPE

#define S_EMPTY         0

/*      Floating-Point definitions.             */

/*      Mostly dependent on target, but         */
/*      something similar will be required.     */

/*      FLOAT is only a guideline;              */
/*      undefining it is unlikely to work.      */

#if defined(TARGET_CPU_ALPHA)
extern int rround;                      /* ALPHA: remembered rounding mode */
#endif

#if defined(FLOAT)

/*      Floating-point registers        */
/*      Now defined in regs.h */

#define LO              0               /* little-endian */
#define HI              1

/*      Floating-point exceptions       */

#define NV_OF_DZ        0x1a            /* Invalid/Overflow/Divide-by-Zero */
#define AEXC            (NV_OF_DZ<<5)   /* Accrued Exception field */
#define TEM             (NV_OF_DZ<<23)  /* Trap Enable Mask */

/*      Floating-point types            */

#define F_S             0
#define F_D             1

#define FAtype          ftype[fstackp]
#define FBtype          ftype[fstackp - 1]
#define FCtype          ftype[fstackp - 2]

#define SNGL            (FAtype == F_S)
#define DBLE            (FAtype == F_D)

#define FS              (SNGL ? 's' : 'd')

#define HALF            (SNGL ? HALFS : HALFD)
#define TWO32           (SNGL ? TWO32S : TWO32D)
#define TWOM32          (SNGL ? TWOM32S : TWOM32D)
#define MAGIC           (SNGL ? TWO23S : TWO52D)

/*      Floating-point rounding modes           */
/*      For SPARC, must correspond to fsr RD field      */

#define R_N             0
#define R_Z             1
#define R_P             2
#define R_M             3

/*      Floating-point function prototypes -> fproto.h  */

extern int ftype[4];    /* Types of f.p. registers - f_float.c */

#endif /* FLOAT */

/* **************************************************************** */

/* translator Function prototypes (was proto.h) */

/* cond.c */

void defcond (int c);
void check (int c, int chk);

/* float.c */

#if defined(FLOAT)
#if defined(TARGET_CPU_SPARC)
void fpush (int t);
const char *fname (int r, int s);
#elif defined(TARGET_CPU_ALPHA)
const char *fname (int r);
void fround_modifier (void);
#endif
void fmnem (const char *op);
void fround (int tround);
void fscheck (int s);
#endif /* FLOAT */

/* label.c */

void init_labtab (void);
int shuffle (int s, int t);
void deflab (void);
state_type reflab (int s);

/* lookup.c */

int lookup (const char *string);
void translate (int token);

/* main.c */

int main (int argc, char *argv[]);

/* parse.c */

char *stenter (const char *l, const char *p, int t);
void parse (void);

/* util.c */

void printst (void);
void where (void);
void error (void);
void lineup (void);
void newline (void);
void COMMENT (const char *s);
const char *rgname (int r);
void mnem (const char *op);
void dummy (const char *op);
void sep (void);
void dec (int op);
void hex (int op);
void str (const char *op);
void qstr (const char *op);
void quotedstr (const char *op);
#if defined(TARGET_CPU_SPARC)
void sched (const char *op);
void extcall (int op);
#endif
void offset (int c);
#if defined(TARGET_CPU_ALPHA)
void tline (const char *op);
void offset0 (int c);
#endif
#if defined(TARGET_CPU_SPARC)
void sindex (const char *r, int c);
void vindex (const char *r);
#endif
#if defined(TARGET_CPU_ALPHA)
void index0 (const char *r, int c);      /* uindex */
void index1 (const char *r, int c);      /* sindex */
#endif
#if defined(TARGET_CPU_ARM)
void immdec (int op);
void index0 (const char *r, int c);      /* uindex */
void index1 (const char *r, int c);      /* sindex */
int big (int n, int bits);
#endif
void index2 (const char *r1, const char *r2);
#if defined(TARGET_CPU_ALPHA)
void indirect (const char *r);
void indirect0 (const char *r);
#endif
void debug (void);
void debugp (const char *name);

/*      Not used */

void label (void);
void LAB (int n);
void here (int c);

/*      Details will depend on target, but      */
/*      something similar will often be needed. */

void lab (int n);
void f (int n);
void b (int n);
void equate (const char *s);
#if defined(TARGET_CPU_SPARC)
const char *extname (int op);
#endif
void push (void);
void pop (int n);
void printglob (void);
void entry (void);
void initial (void);
void final (void);
void align (void);

/*      SPARC only      */

void mnema (const char *op);
void delay (const char *s);
void annul (const char *s);

/*      Should depend on target assembler       */

const char map (char ch);

void LCOMMENT (void);

void defldc (void);

#if defined(TARGET_CPU_SPARC)
void vindex (const char *op);
#endif
