/*
 * KROC translator: Parse Transputer assembly-language line.
 *
 * $Source: /extra/play/cvs/kroc/src/host/parse.c,v $
 *
 * $Id: parse.c,v 1.1 1999/09/27 21:10:12 rivimey Exp $
 *
 * (C) Copyright 1995,1996,1997 D.C. Wood <D.C.Wood@ukc.ac.uk>
 * University of Kent at Canterbury
 *
 * Should be independent of the target machine.
 */

static char parse_c_rcsid[] = "$Id: parse.c,v 1.1 1999/09/27 21:10:12 rivimey Exp $";

#include <stdlib.h>
#include <string.h>
#include "octran.h"

extern it_type itable[];                /* instruction table (lookup.c) */

static void stackdepth (char *p);
#ifdef FLOAT
static void fpstacktype (char *p);
#endif
#if defined(TARGET_CPU_ALPHA)
static void fold (void);
static void note (void);
#endif
static void comment (char *cp);
static void number (char *string);


/* Michael Poole's special comments (o_mdp) don't work yet */

static void stackdepth (char *p)              /* read integer stack depth */
{
  int sd;

  if (strncmp (p, " overflow !!!", 13) == 0) {
    if (o_verbose)
      fputs ("\n... stack overflow", stderr);
  } else {
    sd = atoi (p);
    if (sd != istackp) {
      if (o_verbose) {
        fprintf (stderr, "\n... stack depth = %d (expected %d)\n",
                 sd, istackp);
        where ();
      }
      istackp = sd;
    } else if (o_verbose)
      fprintf (stderr, "\n... stack depth = %d\n", sd);
    
    ishuff = 0;
  }
}


#ifdef FLOAT
/* read floating-point stack type */
static void fpstacktype (char *p)
{
  int fs;

  if (*p == 's')
    fs = F_S;
  else /* (*p == 'd') */
    fs = F_D;
  fstackp = 1;            /* should check this, too */

  if (fs != FAtype) {
    if (o_verbose) {
      fprintf (stderr, "\n... f.p. stack type = %c (expected %c)\n",
               *p, FAtype == F_S ? 's' : 'd');
      where();
    }
    FAtype = fs;
  } else if (o_verbose)
    fprintf (stderr, "\n... f.p. stack type = %c\n", *p);
}
#endif


#if defined(TARGET_CPU_ALPHA)
/*      Not implemented         */

/* process fold comment */
static void fold (void)
{
  if (o_verbose) {
    fprintf (stderr, "\n... %s\n", &buffer[15]);
    where ();
  }
}


/* process note comment */
static void note (void)
{   
  if (o_verbose) {
    fprintf (stderr, "\n... %s\n", &buffer[14]);
    where ();
  }
}
#endif


/* enter PROC/FUNCTION in symbol table */
char *stenter (const char *l, const char *p, int t)
{
  int j, k;
  
  j = p - buffer;
  for (k = 0; buffer[j] != ',' && buffer[j] != ' '; ++j, ++k)
    b_proc[k] = map(buffer[j]);
  
  b_proc[k] = 0;
  stcount += 1;
  if (stcount >= PT_SIZE) {
    fprintf (stderr, "%s:%s:%d: Symbol table overflow",
             program_name, source, iline);
    error ();
  }
  
  if (t == P_EXTERNAL) {     
    if (b_proc[1] == 'C' && b_proc[2] == '_') { 
#ifdef HAVE_C_NO_LEADING_UNDERSCORE
      strcpy (&b_proc[1], &b_proc[2]);        /* C.name => _name  */
#else
      b_proc[1] = '_';                        /* C.name => __name */
#endif
      t = P_C_EXTERNAL;
    }

    k = o_mdp ? 0 : 1;          /* compromise */
    proctab[stcount].S_ISTACK = k;
    proctab[stcount].S_FSTACK = 0;      /* ??? */
    proctab[stcount].S_ISHUFF = k;      /* fshuff? */
    proctab[stcount].S_FTYPES = 0;      /* unnecessary? */
  }

  strcpy (proctab[stcount].label, l);
  k = (t == P_C_EXTERNAL) ? 1 : 0;
  strcpy (proctab[stcount].proc, &b_proc[k]);
  proctab[stcount].S_PTYPE = t;
  fstackp = 0;                            /* (ORTMS p.9) */

  if (o_verbose) {       
    fprintf (stderr, "\n... Start of %s %s %s",
             l, ptype[proctab[stcount].S_PTYPE], b_proc);
    where ();
  }

  return &buffer[j];
}


/* deal with special comments */
static void comment (char *cp)
{
  int j, k;
  char *p;
  
#if defined(TARGET_CPU_ALPHA)
  if (o_mdp) {
    if (strncmp (cp, "--{{{  ", 7) == 0)
      fold ();
    else if (strncmp (cp, "--... ", 6) == 0)
      note ();
  }
#endif
  
  if (strncmp (cp, "-- section label", N_SECLAB) == 0)
    align ();
  
  p = strstr (cp, "-- PROC ");
  if (p != NULL && b_label[0] != 0) {
    p = stenter (b_label, p + N_PROC, P_PROC);

    if (!o_lib && strncmp (p, ", lexlevel: 0, ", N_LEX) == 0) {
      ws = strtol (p + N_LEX + 3, &p, 10);
      vs = strtol (p + 5, &p, 10);
    strcpy (b_entry, b_proc);
    
#if defined(TARGET_CPU_ARM)
    entry ();
#endif
    }

    return;
  }
  
  p = strstr (cp, "-- FUNCTION ");
  if (p != NULL && b_label[0] != 0) {
    p = stenter (b_label, p + N_FUNC, P_FUNCTION);
    return;
  }
  
  p = strstr (cp, "-- Stub for library call ");
  if (p != NULL)
    p = stenter (b_label, p + N_STUB, P_EXTERNAL);
  else if (strncmp (cp, "-- \"", 4) == 0) {
    for (j = B_FILENAME, k = 0; buffer[j] != '"'; ++j, ++k)
      filename[k] = buffer[j];
    
    filename[k] = 0;
    fline = atoi (&buffer[j + 2]);
    /*      This is NOT necessarily between occam statements,       */
    /*      so the evaluation stacks may not be empty.              */
  } else if (strcmp (cp, "-- REAL64 words swapped and quadaligned") == 0)
    o_real64 = True;

}


/* (try to) translate string operand to number */
static void number (char *string)
{
  char *p;
  long n = strtol (string, &p, 10);
  
  if (p == string) {
    nflag = False;          /* not a number */
    return;
  }
  
  nflag = True;
  n_operand = (int)n;
}


/* parse input line - structured spaghetti! */
void parse (void)
{
  init_labtab();
  b_entry[0] = 0;
  b_proc[0] = 0;
  filename[0] = '\0';

  initial ();
  for (iline = 1; ; ++iline) {            /* main loop */
    int len;          /* Length of line (excluding comments) */
    int actuallen;    /* Actual length of line (buffer) */
    char *cp;         /* Comment pointer */
    int cpix;         /* Comment index in buffer */
    int inst = 0;     /* Instruction number found */
    
    /* initialize buffer */
    {
      int i;

      for (i = 0; i < BUF_SIZE; ++i)
        buffer[i] = ' ';
    }

    /* read line into buffer */
    {
      int i, ch;
      
      for (i = 0; i < BUF_SIZE - 8; ++i) {      /* allow for a tab */
        ch = getc (in);
        switch (ch) {
          case EOF:
          case '\n':
            break;                /* leave inner loop */
          case '\t':      /* expand tabs */
            i = i | 7;
            continue;     /* leave switch */
          default:
            buffer[i] = ch;
            continue;
        }
        break;
      } /* for */
      len = actuallen = i;

      buffer[len] = 0;
      if (ch == EOF)
        break;                        /* leave outer loop */
    }

    /* parse contents of buffer */
    b_label[0] = 0;           /* null strings */
    b_operand[0] = 0;
    b_symbol[0] = 0;
    b_comment[0] = 0;
    cflag = True;
    if (len <= 1) {             /* blank line */
      nl;
      continue;
    }

    /* Find comments */
    {
      int j;
      
      /* cp = strstr (buffer, "--");  */
      /* quicker to do it by hand */
      cp = NULL; cpix = -1;
      for (j = 0; j < len; ++j) {       /* look for comment */
        if (buffer[j] == '-' && buffer[j+1] == '-') {
          cpix = j; cp = buffer + cpix;
          len = j;
          break;
        }
      }
    }

    /* Handle comments if found */
    if (cp) {
      if (o_mdp) {
        if (strncmp (cp, "-- stackdepth = ", 16) == 0)
          stackdepth (cp+16);
#ifdef FLOAT
        else if (strncmp (cp, "-- fpstack =  ", 14) == 0)
          fpstacktype (cp+14);
#endif
        else if (strncmp (cp, "-- stackdepth -= 1", 18) == 0)
          shuffle (istackp, istackp - 1);
      }
      
      if (!cpix) {              /* whole-line comment */
        *cp = CCHAR; fputs (buffer, out); nl;
        continue;
      }
      else                    /* comment */
        strcpy(b_comment, &buffer[cpix+2]);
    }


    if (buffer[B_LABEL] != ' ') {      /* label */
      int j, k;
      
      for (j = 0, k = 0; buffer[j] != ':'; ++j, ++k) {
        if (k >= PROC_SIZE) {
          fputs ("\n*** Syntax error", stderr);
          error ();
        }
        
        b_label[k] = map(buffer[j]);
      }
      
      b_label[k] = 0;
      if (o_opt && cond)
        defcond (cond);
    } else if (o_echo) {  /* don't echo labels */
      fputc (CCHAR, out);
      if (o_lines)
        fprintf (out, "%d:", iline);
      fprintf (out, "\t%s", &buffer[B_INSTRUCTION]); nl;
    }

    if (cp != NULL) {
      if (strstr (cp, " bytes for alignment") != NULL)
      {
        /*        align ();       */
        dummy ("ignore"); NL;
        continue;
      }
      else
        comment (cp);
    }

    if (len > B_INSTRUCTION && buffer[B_INSTRUCTION] != ' ' &&
        cpix != B_INSTRUCTION) {        /* instruction */
      int j, k;

      for (j = B_INSTRUCTION, k = 0; buffer[j] != ' '; ++j, ++k)
        b_symbol[k] = buffer[j];

      b_symbol[k] = 0;

      inst = lookup (b_symbol);
      if (itable[inst].flags != 0) {     /* operand */
        for (k = 0; j < len; j++)
          if (buffer[j] != ' ')   /* compress */
            b_operand[k++] = map(buffer[j]);

        b_operand[k] = 0;
        number (b_operand);
      }
    }

    if (b_label[0] != 0)
      deflab ();
    else if (b_symbol[0] != 0)
      translate (inst);
    else if (b_comment[0] != '\0') {
#if defined(_3L) || defined(TARGET_CPU_ARM)
      fprintf (out, "\t%c-%s", CCHAR, b_comment); nl;   /* for 3L acc must be '!' */
#else
      fprintf (out, "\t%c-%s", CCHAR2, b_comment); nl;  /* for gcc must be '#' */
#endif
    }
    
    strcpy (l_label, b_label);
  } /* for (iline = 1 ... ) */

  
  printst ();

#if defined(TARGET_CPU_SPARC)
  if (o_global)
    printglob ();
#endif

  final ();
}
