/*
 * KRoC kernel ASM scheduler definitions
 *
 * $Source: /extra/play/cvs/kroc/src/host/sched.h,v $
 *
 * $Id: sched.h,v 1.1 1999/09/27 21:10:12 rivimey Exp $
 *
 * (C) Copyright 1995,1996,1997 D.J. Beckett <D.J.Beckett@ukc.ac.uk>
 *                              D.C. Wood <D.C.Wood@ukc.ac.uk>
 * University of Kent at Canterbury
 */

#include "config.h"

/* This file is included in assembler files so may not always be able
   to use the new ANSI C ## operator in the C pre-processor */

#ifdef __STDC__
#define CAT(a,b) a ## b
#else
#define _IDENT(a) a
#define CAT(a,b) _IDENT(a)b
#endif

#ifdef HAVE_C_NO_LEADING_UNDERSCORE
#define MAKENAME(n) n
#else
#define MAKENAME(n) CAT(_,n)
#endif

/*       Constants      */

False           = 0
True            = 1
Flush           = 255                   /*  Code character */

/*      Constants used in transputer instructions */

MostNeg         = 0x80000000

NotProcess_d    = 0                     /*  Differences from MostNeg */
Enabling_d      = 1
Waiting_d       = 2
Ready_d         = 3
TimeSet_d       = 1
TimeNotSet_d    = 2

NoneSelected_o  = -1

R_N             = 0                     /*  Floating-point rounding modes */
R_Z             = 1
R_P             = 2
R_M             = 3


Temp            = 0                     /*  Process workspace offsets */
Iptr            = -1*4
Link            = -2*4

#ifdef T9000

Count           = Link
Pointer         = -4*4
Length          = Pointer
TLink           = -5*4
Time            = -6*4

#else

Pointer         = -3*4
TLink           = -4*4
Time            = -5*4

#endif

State           = Pointer

#if defined(TARGET_CPU_SPARC)
/*{{{  register naming*/
!       Register names

#define P0      %o0                     /* Parameters/results for C functions */
#define P1      %o1

#define L       %o7                     /* Link (return address) */

#define R       %l0                     /* C for full stack */
#define Q       %l1                     /* B for full stack */
#define P       %l2                     /* A for full stack */
#define Wptr    %l3                     /* Workspace pointer */
#define Fptr    %l4                     /* Front queue pointer */
#define Bptr    %l5                     /* Back queue pointer */
#define Tptr    %l6                     /* Time queue pointer */
#define Vptr    %l7                     /* Pointer to variables */

#define T0      %i0                     /* Temporary registers */
#define T1      %i1
#define T2      %i2
#define T3      %i3
#define MIN     %i4                     /* MostNeg = MinInt = NotProcess.p */
#define STATUS  %i5                     /* Contains ErrorFlag */

#define Z       %g0                     /* Zero */

/*}}}  */
!       Options

#define KERNEL_RUN                      /* Allow position-independent calls */
#define IMPOSSIBLE			/* Check for impossible errors */
#undef  QUICK                           /* Sometimes much slower! */
#undef  T9000                           /* Not fully implemented */
#undef  VPTR                            /* Under development */
#define LITTLE_ENDIAN			/* Currently required */


/*       (The 4 below is BytesPerWord) */

Frame           = 14*4                  /*  Iptr, in, out, err, vsp, aexc(2), */
					/*  fsr(4), FPtemp(2); rounded up     */
System          = 24*4                  /*  Register dump etc.; rounded up    */
Aexc            = -8*4                  /*  Rounding modes etc.               */
Zero            = -7*4
RN              = -6*4                  /*  ) ? Reverse order for */
RZ              = -5*4                  /*  )   IEEE32OP etc      */
RP              = -4*4
RM              = -3*4
FPtemp          = -2*4                  /*  Temporary for floating-point */

NV_OF_DZ        = 0x1a                  /*  iNValid, OverFlow, Divide by Zero */
AEXC            = NV_OF_DZ<<5
TEM             = NV_OF_DZ<<23

Channel         = 1*4                   /*  Extra locals for hard channels */
Device          = 2*4


#endif
#if defined(TARGET_CPU_ALPHA)
/*{{{  register naming*/
/* assumes regdef.h already included */

#define R0      v0      /* Result from C functions */
#define P0      a0      /* Parameter for C functions */
#define P1      a1      /* Parameter for C functions */
#define P2      a2      /* Parameter for C functions */
#define P3      a3      /* Parameter for C functions */
#define P4      a4      /* Parameter for C functions */
#define P5      a5      /* Parameter for C functions + scratch reg */


#define REG_0   t0                     /* Local registers */
#define REG_1   t1
#define REG_2   t2


#define Save1   t3                    /* NOT USED */
/* #define Vptr    t7            NOT USED Pointer to variables */

#define T0      t8                     /* Also treated as locals */
#define T1      t9
#define T2      t10
#define T3      t11

/* the next group are those which will survive calls to C */
#define MIN     s0                    /* constant MOSTNEG INT32 */
#define Front   s1                    /* front of process queue */
#define Back    s2                    /* back of process queue */
#define TimeQ   s3                    /* front of ordered time queue */
#define Wptr    s4                    /* workspace pointer */
#define L       s5                    /* return address to occam */
#define ETAB    s6                    /* base address of entry point table */
/*}}}  */
#endif
#if defined(TARGET_CPU_ARM)
/* no register naming needed ATM */
#endif
