#ifndef SCHED_H
#define SCHED_H

#include "sched_types.h"

#include "timer.h"

/* PUBLIC variables that are used elsewhere in the run time system. */
extern sync_type sf;
extern word STATUS;

#if 0
#ifdef PRIORITIES
extern word* Fptr[NUM_PRIORITIES];
extern word* Bptr[NUM_PRIORITIES];
#else
extern word* Fptr;
extern word* Bptr;
#endif
#endif

extern word* Wptr;
extern word* Tptr;

/* paramter passing variables */
extern word** channel_address;
extern word return_address;
extern byte* source_address;
extern byte* destination_address;
extern word count;
extern Time wait_time;
extern word start_address;
extern word guard;
extern word process_address;

/* pseudo private variables (there names are externally visible, but they */
/* should not be used. */
extern Time now;
extern word temp1;
extern word *temp_ptr;
extern word *other_workspace;
extern word *current;
extern word *previous;

extern word *occam_table;

void dead(int status);
void kernel_panic();
int c_enter(void);

typedef void KrnVoidProc(void);
typedef void KrnWordProc(word);
typedef void KrnTwoWordProc(word, word);
typedef void KrnThreeWordProc(word, word, word);
typedef void KrnWordPtrProc(word *);

#define K_VER 1
#define K_REV 1

typedef struct {
    
    int sz : 8;  	/* sizeof() procvec */
    int fl_ver: 4; 	/* version # of procvec */
    int fl_rev: 4; 	/* minor version -- changed functions, same procvec */
    int fl_occ: 1; 	/* occam supported */
    int fl_c: 1; 	/* 'C' supported */
    int fl_ext: 1;	/* external comms supported */
    int fl_pri: 1; 	/* many priorities */
    int fl_unused: 4;
    int prios: 8;	/* number of priorities supported >= 1 */

    KrnVoidProc	*krnSavel;
    KrnWordProc	*krnStlf;
    KrnWordProc	*krnStlb;
    KrnVoidProc	*krnNorm;
    KrnVoidProc	*krnSeterr;
    KrnVoidProc	*krnRangecheckerror;
    KrnTwoWordProc	*krnOutbyte;
    KrnVoidProc	*krnOutword;
    KrnThreeWordProc	*krnOut;
    KrnVoidProc	*krnReleaseAlt;
    KrnVoidProc	*krnZlnbIn;
    KrnVoidProc	*krnNbOut;
    KrnVoidProc	*krnIn;
    KrnVoidProc	*krnPause;
    KrnVoidProc	*krnResched;
    KrnVoidProc	*krnStopp;
    KrnVoidProc	*krnScheduler;
    KrnVoidProc	*krnCanyPriChange;
    KrnVoidProc	*krnCownPriChange;
    KrnVoidProc	*krnCpriSchedule;
    KrnVoidProc	*krnCpriReschedule;
    KrnVoidProc	*krnCpriStartp;
    KrnVoidProc	*krnCstartp;
    KrnVoidProc	*krnCendp;
    KrnVoidProc	*krnEndp;
    KrnVoidProc	*krnStartp;
    KrnVoidProc	*krnRunp;
    KrnVoidProc	*krnLdtimer;
    KrnVoidProc	*krnDisc;
    KrnVoidProc	*krnDiss;
    KrnVoidProc	*krnDist;
    KrnVoidProc	*krnEnbt;
    KrnVoidProc	*krnEnbc;
    KrnVoidProc	*krnEnbs;
    KrnVoidProc	*krnTaltwt;
    KrnVoidProc	*krnAltStart;
    KrnVoidProc	*krnAltWt;
    KrnVoidProc	*krnAltEnd;
    KrnVoidProc	*krnTin;
    KrnVoidProc	*krnCclaimSem;
    KrnVoidProc	*krnCreleaseSem;
    KrnVoidProc	*krnCreleaseAllSem;
    
} KernelProcVec;


#endif /* sched.h */

