/*
 * KRoC kernel C macro scheduler definitions
 *
 * (C) Copyright 1995,1996,1997 D.J. Beckett <D.J.Beckett@ukc.ac.uk>
 *                              D.C. Wood <D.C.Wood@ukc.ac.uk>
 * Completely rewritten for C kernel:
 * (C) Copyright 1998           Jim Moores <jm40@ukc.ac.uk>
 * University of Kent at Canterbury
 */

#ifndef SCHED_CONSTS_H
#define SCHED_CONSTS_H

#ifdef HAVE_CONFIG_H
#  include <config.h>
#endif

/* This file is included in assembler files so may not always be able
   to use the new ANSI C ## operator in the C pre-processor */

#ifdef __STDC__
#  define CAT(a,b) a ## b
#else
#  define _IDENT(a) a
#  define CAT(a,b) _IDENT(a)b
#endif

/* #define HAVE_C_LEADING_UNDERSCORE */

#ifdef HAVE_C_LEADING_UNDERSCORE
#  define MAKENAME(n) CAT(_,n)
#else
#  define MAKENAME(n) n
#endif

/*       Constants      */

#define false 0
#define true  1
#define Flush 255

/*      Constants used in transputer instructions */

#define MostNeg 0x80000000

/* That all important constant on the transputer (was MinInt) */
#define NotProcess_p 0

/* ALT flag states */
#define Enabling_p NotProcess_p+1
#define Waiting_p NotProcess_p+2
#define Ready_p NotProcess_p+3
#define TimeSet_p NotProcess_p+1
#define TimeNotSet_p NotProcess_p+2

#ifdef EXTERNAL_COMMS
  /* Flag used in ALT over external channel in Pointer field */
  #define BlockIn_p NotProcess_p+1
#endif

/* some offsets into the process descriptor for fields supporting priorities */
#ifdef PRIORITIES
  #define Priority_f 5
  #define SuspendQ 6
  #define NUM_PRIORITIES 32
#endif

/* Workspace offsets used by kernel */
/* note that in the assembler version these were all multiplied by the */
/* wordsize, we are using the array access operators so we don't need to */
#define Temp 0
#define Iptr -1
#define Link -2
#define Pointer -3
#define TLink -4

#ifdef CPU_TIMERS
  #define Time_f -3 
  /* note that this is 7 _double_ words because the array must be accessed
   * as an array of 'Time', which is 64-bit, so it really equates to fields
   * at offsets of 14 and 15.
   */
#else
  #define Time_f -5
#endif
/* wanted to prevent clash between field and the type 'Time' */

#define State -3

#define Count 1
#define IptrSucc 0
#define NoneSelected_o MostNeg /* should be 0 really */

#define Frame 14 /* words! */
#define System 24*4 /* dunno where this is used */

#define IOChannel 1 /* Extra locals for hard channels */
#define Device 2


#define KERNEL_RUN                      /* Allow position-independent calls */
#define IMPOSSIBLE			/* Check for impossible errors */
#undef  T9000                           /* Not fully implemented */
#undef  VPTR                            /* Under development */
#define LITTLE_ENDIAN			/* Currently required */

/* Indexes into kernel jump table */

#define Table_In -1
#define Table_Out -2
#define Table_OutWord -3
#define Table_OutByte -4
#define Table_RunP -5
#define Table_StartP -6
#define Table_EndP -7
#define Table_StopP -8
#define Table_Pause -9
#define Table_LdTimer -10
#define Table_TIn -11
#define Table_AltWt -12
#define Table_TAltWt -13
#define Table_EnbC -14
#define Table_EnbT -15
#define Table_DisS -16
#define Table_DisC -17
#define Table_DisT -18
#define Table_RangeCheckError -19
#define Table_T_Seterr -20
#define Table_Norm -21
#define Table_C_EndP -22
#define Table_C_StartP -23
#define Table_Scheduler -24

#endif /* sched_consts.h */
