#ifndef TIMER_HEADER
#define TIMER_HEADER

#include "csptime.h"

#include "rts.h"

#include "sched.h"

#ifdef CPU_TIMERS
#include <stdio.h>
#include <unistd.h>

#define abs(X) ((X)>0?(X):-(X))

#define lo(X) ((word *)&(X))[0]
#define hi(X) ((word *)&(X))[1]

#define GetTimeField(X) (((Time *)(X))[Time_f])
extern Time timeout_alarm;
#define INLINE_TIMERS
#ifdef INLINE_TIMERS
#ifdef DONT_INLINE_WITH_MACROS
/* This form of inlining just doesn't work as static */
static inline Time Time_GetTime(void)
{
  Time now;
  __asm__ __volatile__(".byte 0x0f,0x31"
	  :"=a" (lo(now)),
	   "=d" (hi(now))
         );
  return now;
}

static inline bool Time_PollAlarm()
{
  Time now = Time_GetTime();
  return now >= timeout_alarm;
}

static inline void Time_SetAlarm(Time t)
{
  timeout_alarm = t;
}

static inline void Time_ResetAlarm(void) {
  timeout_alarm = 0x7FFFFFFFFFFFFFFFLL;
}

static inline Time Time_PLUS(Time t1, Time t2) {
  return (t1+t2);
}

static inline Time Time_MINUS(Time t1, Time t2) {
  return (t1-t2);
}

static inline bool Time_AFTER(Time t1, Time t2) {
  return (t1-t2)>0LL;
}
#else
#warning Inlining from macros
#define Time_SetAlarm(t) timeout_alarm = (t)
#define Time_ResetAlarm() timeout_alarm = 0x7FFFFFFFFFFFFFFFLL
#define Time_PLUS(t1,t2) ((t1)+(t2))
#define Time_MINUS(t1,t2) ((t1)-(t2))
#define Time_AFTER(t1,t2) ((t1)-(t2)>0LL)
/*
#define Time_GetTime() ( \
  { \
    Time now; \
    __asm__ (".byte 0x0f,0x31" \
            :"=a" (lo(now)), \
             "=d" (hi(now)) \
           ); \
    now; \
  } \
)
*/
Time Time_GetTime(void);
#define Time_PollAlarm() Time_AFTER(Time_GetTime(), timeout_alarm)
#endif
#else
Time Time_GetTime(void);
bool Time_PollAlarm();
void Time_SetAlarm(Time);
void Time_ResetAlarm(void);
Time Time_PLUS(Time, Time);
Time Time_MINUS(Time, Time);
bool Time_AFTER(Time, Time);
#endif

int Time_CPUSpeed();
void Time_Init();
void Time_SetCPUSpeed(int cpu_speed);

/* END NATIVE IMPLEMENTATION */
#else
/* BEGIN UNIX IMPLEMENTATION */
#warning UNIX Timer Implmentation
#define GetTimeField(X) ((Time)((X)[Time_f]))

extern sync_type sf;

#ifdef INLINE_TIMERS
#define Time_GetTime() rtime()
#define Time_SetAlarm(t) ualarm(t,0)
#define Time_ResetAlarm() ualarm(0,0)
#define Time_PLUS(t1,t2) ((t1)+(t2))
#define Time_MINUS(t1,t2) ((t1)-(t2))
#define Time_AFTER(t1,t2) ((t1)-(t2)>0)
#define Time_PollAlarm() (sf.s.tim_sync)
#else
Time Time_GetTime(void);
bool Time_PollAlarm();
void Time_SetAlarm(Time);
void Time_ResetAlarm(void);
Time Time_PLUS(Time, Time);
Time Time_MINUS(Time, Time);
bool Time_AFTER(Time, Time);
#endif
/* END UNIX TIMER IMPLEMENTATION */
#endif

#endif /* timer.h */
