/* $Id: bind3.c,v 1.3 1997/03/25 14:52:40 djb1 Exp $ */

/*
 *	workspace allocation
 *	Copyright (C) 1987 Inmos Limited
 *
 *	This program is free software; you can redistribute it and/or modify
 *	it under the terms of the GNU General Public License as published by
 *	the Free Software Foundation; either version 2 of the License, or
 *	(at your option) any later version.
 *
 *	This program is distributed in the hope that it will be useful,
 *	but WITHOUT ANY WARRANTY; without even the implied warranty of
 *	MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *	GNU General Public License for more details.
 *
 *	You should have received a copy of the GNU General Public License
 *	along with this program; if not, write to the Free Software
 *	Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 */

/*#define DEBUG*/

/*
 *  modified for alpha to align anything bigger than one slot on quad word
 *  boundary, and to round up wssize to even number of slots
 */

/*{{{  include files*/
#ifdef HAVE_CONFIG_H
#  include <config.h>
#endif
#include <stdio.h>
#include "suplib.h" /* IMPORTED */

#include "includes.h"
#include "generror.h"
#include "genhdr.h"
#include "bind3def.h"
#include "gen1def.h"
#include "gen2def.h"
/*}}}*/

/*{{{  #defines*/
/*#define MAXVAR 2048*/ /* changed for bug 1075 13/12/90 */
#define MAXVAR 64 /* start value - now no fixed limit */ /* changed for bug 1075 24/9/91 */

#define VARROW (MAXVAR/(sizeof(int) * 8))

#define   SORT_USAGE_CUTOFF 16 /* first 16 vars are sorted by usage, rest by scope */
#define RESORT_SLOTS_CUTOFF 16 /* first 16 slots are sorted by usage, rest by scope */
/*}}}*/

/*{{{  global variables*/
PUBLIC BIT32 datasize, vsp, maxvsp;
PUBLIC treenode *enclosing_vsp_nptr; /* nptr of next enclosing vectorpsace variable */
PUBLIC int staticlinkused;
PUBLIC int var_base, num_var;
PUBLIC INT32 loop_weight;

/*PUBLIC treenode *current_proc;*/ /* no longer needed bug 1075 24/9/91 */

PUBLIC int alloc_strategy = ALLOC_DEFAULT;
/*}}}*/

/*{{{  local variables*/
PRIVATE int maxvar = MAXVAR; /* Current max number of variables in bitmap */
PRIVATE int varrow = VARROW; /* Current size of bitmap */
PRIVATE int nextlabel;
/*PRIVATE int live_var[VARROW];*/ /* bug 1075 24/9/91 */
PRIVATE int *live_var = NULL; /* bug 1075 24/9/91 */
PRIVATE int *var_map = NULL;
/*PRIVATE treenode *var_list[MAXVAR];*/ /* bug 1075 24/9/91 */
PRIVATE treenode **var_list = NULL; /* bug 1075 24/9/91 */
PRIVATE treenode *fixedws_list;

PRIVATE treenode *vars_mapped_onto_vsptr;
/*}}}*/

/*{{{  vectorspace allocation*/
/*{{{  PUBLIC void initvsp (n)*/
/*****************************************************************************
 *
 *  initvsp initialises the vector space and maximum vector space usage to
 *          n slots.
 *
 *****************************************************************************/
PUBLIC void initvsp ( INT32 n )
{
  if (needs_quadalign && ((n&1) == 1))
    n++;
  vsp = n;
  maxvsp = n;
  enclosing_vsp_nptr = NULL; /* INSdi01511 */
}
/*}}}*/
/*{{{  PRIVATE void setvsp(n)*/
/*****************************************************************************
 *
 *  setvsp sets the vector space to be n words, adjusting maxvsp if neccessary.
 *
 *****************************************************************************/
PRIVATE void setvsp ( INT32 n )
{
  vsp = n;
  maxvsp = max_INT32(maxvsp, vsp);
}
/*}}}*/
/*{{{  PUBLIC BIT32 newvs (n)*/
/*****************************************************************************
 *
 *  newvs reserves n words of vector space and adjusts maxvs if neccessary.
 *        Returns the starting slot number of the reserved space.
 *
 *****************************************************************************/
PUBLIC INT32 newvs ( INT32 n )
{
  BIT32 result;
  result = (needs_quadalign && ((n&1) == 1)) ? vsp+1 : vsp;
  setvsp(result + (BIT32)n);
  return(result);
}
/*}}}*/
/*}}}*/
/*{{{  workspace allocation*/
/*{{{  comment*/
/* allocation:
 *
 * A bit map showing the interference between variables is built up
 * during the register mapping phase of this pass.
 * The map is an  n x n bit var showing which are 'live'
 * at the same time:
 *    For example the map for
 *
 *    a1:                            a1 a2 a3 a4
 *    SEQ                         a1 1  1  1  1
 *      : :                       a2 1  1  1  0
 *      a2:           =           a3 1  1  1  0
 *      SEQ                       a4 1  0  0  0
 *        : :
 *        a3:
 *        SEQ
 *          : :
 *      a4:
 *      SEQ
 *        : :
 *
 * After the map has been built it is sorted by usecount.
 * Those variables requiring fixed positions are allocated first.
 * The allocation is done by scanning the list of variables looking for the
 * first slot large enough to take the variable which is not used by any
 * variable live at the same time as it.
 *
 * Variables which overlay each other are chained onto the original varible.
 */
/*}}}*/
/*{{{  PUBLIC INT32 allocsize(nptr)*/
PUBLIC INT32 allocsize ( treenode *const nptr )
{
  treenode *const type = NTypeOf(nptr);
  INT32 size = wordsin(type);
  if (nodetypeoftag(TagOf(nptr)) != NAMENODE) /* bug 1254 8/5/91 */
    badtag(genlocn, TagOf(nptr), "allocsize");
  if (size < 0)
    geninternal_is(GEN_ERROR_IN_ROUTINE, size, "allocsize");
  if (chanaspointer && (TagOf(nptr) == N_DECL) && channel_basetype(type))
    {
      if (!is_channel_constructor(nptr)) /* bug TS/2038 15/01/93 */
        {
          /* it must not be a constructor */
        #if OPTIMISE_LOCAL_CHANS
          if ((TagOf(type) != S_CHAN) || NChanMarkOf(nptr))
            size *= 2;
          /* this leaves (unmarked) scalar channels as using a single word */
        #else
          size *= 2;
        #endif
        }
    }
  return size;
}
/*}}}*/
/*{{{  support*/
/*{{{  magic macros*/
/* These macros are used for setting and testing bits in arrays of words */
#define WORDSHIFT 5     /* such that 2^WORDSHIFT = (MAXVAR/VARROW) */
#define BYTEMASK  ((1 << WORDSHIFT) - 1)  /* (2^WORDSHIFT) - 1 */

#define wordnum(n) ((n) >> WORDSHIFT)
#define bitnum(n)  ((n) &  BYTEMASK)

#define setbit(array,n)   ((array)[wordnum(n)] |=  (1 << (bitnum(n))))
#define clearbit(array,n) ((array)[wordnum(n)] &= ~(1 << (bitnum(n))))
#define testbit(array,n)  ((array)[wordnum(n)] &   (1 << (bitnum(n))))

/* Because the bit map is symetrical (var_map[i][j] == var_map[j][i])
 * we can store it in half the space.
 *
 * This macro is used to map a row number to its position in a linear array
 */

#define startofrow(n)    ((((n)+1)*(n)) >> 1) /* 1/2 n(n+1) */
/*}}}*/

/*{{{  PRIVATE void setbitmap (map, i, j)*/
PRIVATE void setbitmap ( int *map , int i , int j )
{
  int bitpsn;
  if (i > j)
    bitpsn = startofrow(i) + j;
  else
    bitpsn = startofrow(j) + i;
  setbit(map, bitpsn);
}
/*}}}*/
/*{{{  PRIVATE void clearbitmap (map, i, j)*/
#if 0 /* clearbitmap is never used */
PRIVATE void clearbitmap ( int *map , int i , int j )
{
  int bitpsn;
  if (i > j)
    bitpsn = startofrow(i) + j;
  else
    bitpsn = startofrow(j) + i;
  clearbit(map, bitpsn);
}
#endif /* 0 */
/*}}}*/
/*{{{  PRIVATE int testbitmap (map, i, j)*/
PRIVATE int testbitmap ( int *map , int i , int j )
{
  int bitpsn;
  if (i > j)
    bitpsn = startofrow(i) + j;
  else
    bitpsn = startofrow(j) + i;
  return (testbit(map, bitpsn) != 0);
}
/*}}}*/
/*{{{  PRIVATE int markbitmap(map, mask, from, to, n)*/

/* Allocation map             */
/* map of live variables      */
/* range of variables to mark */
/* bit to be set              */
PRIVATE int markbitmap (int *const map, int *const mask,
                        const int from, const int to, const int n)
{
  register int i;
  int count = 0;
  for (i = from; i<=to; i++)
    if (testbit(mask, i))
      {
        setbitmap(map, i, n);
        count++;
      }
  return(count);
}
/*}}}*/
/*{{{  PRIVATE BOOL vars_overlap (b1, l1, b2, l2)*/
/* Test for overlap of two vars,
* a variable of length 0 never overlaps
*/
PRIVATE BOOL vars_overlap (const INT32 b1, const INT32 l1, const INT32 b2, const INT32 l2 )
{
  return ((l1 != 0 && l2 != 0) &&
          (((b1 >= b2) && (b1 < b2+l2)) || ((b2 >= b1) && (b2 < b1+l1))));
}
/*}}}*/
/*{{{  PRIVATE INT32 numslots (nptr)*/
/* Give number of slots needed to store name nptr */
PRIVATE INT32 numslots ( treenode *nptr )
{
  INT32 length = 0; /* initialised to shut up gcc's optimiser */
  switch (TagOf(nptr))
    {
      /*{{{  T_RESERVEDWS*/
      case T_RESERVEDWS:
        length = NVRSizeOf(nptr);
        break;
      /*}}}*/
      /*{{{  T_TEMP, N_DECL, T_PREEVALTEMP*/
      case T_TEMP:
      case N_DECL:
      case T_PREEVALTEMP:
        if (isplaced(nptr))
          length = 0;
        else if (isinvectorspace(nptr) || (NModeOf(nptr) == NM_POINTER))
          length = 1;
        else
          length = allocsize(nptr);
        break;
      /*}}}*/
      /*{{{  N_ABBR N_VALABBR N_RETYPE N_VALRETYPE*/
      case N_VALABBR: case N_ABBR:
      case N_VALRETYPE: case N_RETYPE:
        {
          const abbrevmode_t am = be_abbrevmode(NDeclOf(nptr));
          if ((am == AM_CONST) || (am == AM_ISRHS))
            length = 0;
          else if ((am == AM_PTR) || isinvectorspace(nptr))
            length = 1;
          else
            length = allocsize(nptr);
        }
        break;
      /*}}}*/
      /*{{{  N_REPL*/
      case N_REPL:
        /* Replicator occupies two words, one for base, one for count */
        {
          /* This was used before the change in the way replicated ALT was done: */
          /*
            int t = TagOf(NDeclOf(nptr));
            length = (t == S_REPLALT || t == S_PRIREPLALT) ? 3 : 2;
          */
          length = 2;
        }
        break;
      /*}}}*/
      default:
        badtag(genlocn, TagOf(nptr), "numslots");
    }
  return (length);
}
/*}}}*/
/*{{{  PRIVATE void check_bitmap_relationships*/
PRIVATE void check_bitmap_relationships(void)
{
  /* Here we have some code to check the assumptions made in the
     live variable maps. Eg that sizeof(int) == 32

     It also checks the consistency of updating maxvar and varrow.
  */

  if (((1 << WORDSHIFT) != (sizeof(int) * 8)) /* check WORDSHIFT is correct */
   || ((maxvar & BYTEMASK) != 0) /* check we have a whole number of varrows */
   || ((maxvar/varrow) != (1 << WORDSHIFT)))
    err_abort();
}
/*}}}*/
/*{{{  PRIVATE void initrow (row)*/
/* Set all elements of row zero */
PRIVATE void initrow ( int *row )
{
#if 0
  int i;
  for (i = 0; i < varrow; i++)
    row[i] = 0;
#else
  /* ANSI C guarantees that (int)0 is all-bits-zero */
  memset(row, '\0', varrow * sizeof(*row));
#endif
}
/*}}}*/
/*{{{  PRIVATE int numwords_of()*/
PRIVATE int numwords_of(const int n)
{
  const int i = startofrow(n);
  int numwords = wordnum(i);
  if (bitnum(i) > 0)
    numwords++;
  return numwords;
}
/*}}}*/
/*{{{  PRIVATE void init_varmap(map)*/
PRIVATE void init_varmap (int map[])
{
  const int numwords = numwords_of(maxvar);
#if 0
  int i;
  for (i = 0; i < numwords; i++)
    map[i] = 0;
#else
  /* ANSI C guarantees that (int)0 is all-bits-zero */
  memset(map, '\0', numwords * sizeof(*map));
#endif
}
/*}}}*/
/*{{{  PRIVATE int *alloc_var_map ()*/
/* allocate new bitmap */
PRIVATE int *alloc_var_map (void)
{
  /* maxvar is never small enough for numwords_of(maxvar) to be zero */
  return memalloc(sizeof(int) * numwords_of(maxvar));
}
/*}}}*/
/*{{{  PRIVATE int *alloc_live_var*/
PRIVATE int *alloc_live_var(void)
{
  return memalloc(sizeof(*live_var) * varrow); /* bug 1075 24/9/91 */
}
/*}}}*/
/*{{{  PRIVATE treenode **alloc_var_list*/
PRIVATE treenode **alloc_var_list(void)
{
  return memalloc(sizeof(*var_list) * maxvar); /* bug 1075 24/9/91 */
}
/*}}}*/
/*{{{  PRIVATE int nextvarnum()*/
PRIVATE int nextvarnum ( void )
{
  if (num_var >= maxvar)
#if 0 /* bug 1075 24/9/91 */
    msg_out_is(SEV_FATAL, GEN, GEN_TOO_MANY_VARS, genlocn,
               MAXVAR, WNameOf(NNameOf(current_proc)));
#else
    /*{{{  stretch all the relevant arrays*/
    {
      const int old_varrow = varrow;
      const int old_maxvar = maxvar;

      varrow *= 2;
      maxvar *= 2;
      DEBUG_MSG(("nextvarnum: stretching maxvar from %d to %d\n", old_maxvar, maxvar));
      /*printf("nextvarnum: stretching maxvar from %d to %d\n", old_maxvar, maxvar);*/
      {
        int *const new_var_map = alloc_var_map();
        init_varmap(new_var_map);
        memcpy(new_var_map, var_map, numwords_of(old_maxvar) * sizeof(*var_map));
        memfree(var_map);
        var_map = new_var_map;
      }
      {
        treenode **const new_var_list = alloc_var_list();
        memcpy(new_var_list, var_list, old_maxvar * sizeof(*var_list));
        memfree(var_list);
        var_list = new_var_list;
      }
      {
        int *const new_live_var = alloc_live_var();
        initrow(new_live_var);
        memcpy(new_live_var, live_var, old_varrow * sizeof(*live_var));
        memfree(live_var);
        live_var = new_live_var;
      }
      check_bitmap_relationships();
    }
    /*}}}*/
#endif
  return num_var++;
}
/*}}}*/
/*{{{  PRIVATE int clashcount(first_var, n)*/
#if 0  /* clashcount is never called */
PRIVATE int clashcount ( int first_var , int n )
{
  int i;
  int total = 0;
  for (i=first_var; i < num_var; i++)
    if (testbitmap(var_map, n, i))
      total++;
  return (total);
}
#endif /* 0 */
/*}}}*/

/*{{{  compare routines*/
/*{{{  PRIVATEPARAM int compare_scope (v1, v2) ** passed as a parameter*/
/*{{{  comment*/
/* Compare two variables c1 and c2,
   The ordering used to compare vars is :

   T_RESERVEDWS <  all other vars

   all other vars are then compared on scope
*/
/*}}}*/
PRIVATEPARAM int compare_scope ( const void *const v1 , const void *const v2 )
{
  const treenode *const var1 = var_list[*(const int *)v1];
  const treenode *const var2 = var_list[*(const int *)v2];
  const int tag1 = TagOf(var1);
  const int tag2 = TagOf(var2);
  if (tag1 == T_RESERVEDWS || tag2 == T_RESERVEDWS)
    /*{{{  */
    {
      if (tag1 == tag2)
        return 0;
      else if (tag1 == T_RESERVEDWS)
        return -1;
      else
        return 1;
    }
    /*}}}*/
  else if (tag1 == T_TEMP || tag2 == T_TEMP)
    /*{{{  */
    {
      if (tag1 == tag2)
        return 0;
      else if (tag1 == T_TEMP)
        return -1;
      else
        return 1;
    }
    /*}}}*/
  else if (NScopeOf(var1) > NScopeOf(var2))
    return -1;
  else if (NScopeOf(var1) == NScopeOf(var2))
    return 0;
  else
    return 1;
}
/*}}}*/
/*{{{  PRIVATEPARAM int compare_use (v1, v2) ** passed as a parameter*/
/*{{{  comment*/
/* Compare two variables c1 and c2,
   The ordering used to compare vars is :

   T_RESERVEDWS <  all other vars

   all other vars are then compared on use count
*/
/*}}}*/
PRIVATEPARAM int compare_use (const void *const v1, const void *const v2 )
{
  const treenode *const var1 = var_list[*(const int *)v1];
  const treenode *const var2 = var_list[*(const int *)v2];
  const int tag1 = TagOf(var1);
  const int tag2 = TagOf(var2);
  if (tag1 == T_RESERVEDWS || tag2 == T_RESERVEDWS)
    /*{{{  */
    {
      if (tag1 == tag2)
        return 0;
      else if (tag1 == T_RESERVEDWS)
        return -1;
      else
        return 1;
    }
    /*}}}*/
  else if (NVUseCountOf(var1) > NVUseCountOf(var2))
    return -1;
  else if (NVUseCountOf(var1) == NVUseCountOf(var2))
    #if 0
    /*{{{  compare clashes*/
    {
      if (NVClashCountOf(var1) < NVClashCountOf(var2))
        return -1;
      else if (NVClashCountOf(var1) == NVClashCountOf(var2))
        return 0;
      else
        return 1;
    }
    /*}}}*/
    #endif
    /*return 0;*/
    return compare_scope(v1, v2);
  else
    return 1;
}
/*}}}*/
#if 0  /* never used */
/*{{{  PRIVATEPARAM int compare_clash (v1, v2) ** passed as a parameter*/
/*{{{  comment*/
/* Compare two variables c1 and c2,
   The ordering used to compare vars is :

   T_RESERVEDWS <  all other vars

   all other vars are then compared on use count
*/
/*}}}*/
PRIVATE int compare_clash ( int *const v1 , int *const v2 )
{
  const treenode *const var1 = var_list[*v1];
  const treenode *const var2 = var_list[*v2];
  const int tag1 = TagOf(var1);
  const int tag2 = TagOf(var2);
  if (tag1 == T_RESERVEDWS || tag2 == T_RESERVEDWS)
    /*{{{  */
    {
      if (tag1 == tag2)
        return 0;
      else if (tag1 == T_RESERVEDWS)
        return -1;
      else
        return 1;
    }
    /*}}}*/
  else if (NVClashCountOf(var1) < NVClashCountOf(var2))
    return -1;
  else if (NVClashCountOf(var1) == NVClashCountOf(var2))
    /*{{{  compare use count*/
    {
      if (NVUseCountOf(var1) > NVUseCountOf(var2))
        return -1;
      else if (NVUseCountOf(var1) == NVUseCountOf(var2))
        return 0;
      else
        return 1;
    }
    /*}}}*/
  else
    return 1;
}
/*}}}*/
/*{{{  PRIVATEPARAM int compare_offset (v1, v2) ** passed as a parameter*/
/*{{{  comment*/
/* Compare two variables c1 and c2,
   The ordering used to compare vars is :

   T_RESERVEDWS <  all other vars

   all other vars are then compared on use count
*/
/*}}}*/
PRIVATEPARAM int compare_offset ( int *const v1 , int *const v2 )
{
  const treenode *const var1 = var_list[*v1];
  const treenode *const var2 = var_list[*v2];
  if (NVOffsetOf(var1) < NVOffsetOf(var2))
    return -1;
  else if (NVOffsetOf(var1) == NVOffsetOf(var2))
    return 0;
  else
    return 1;
}
/*}}}*/
#endif /* 0 */
/*}}}*/
/*}}}*/
/*{{{  PRIVATE void printname(nptr)*/
PRIVATE void printname ( treenode *nptr )
{
  switch(TagOf(nptr))
    {
      case T_RESERVEDWS:
        fprintf(outfile, "$Reserved%-6d", NVVarNumOf(nptr)); break;
      case T_TEMP: case T_PREEVALTEMP:
        fprintf(outfile, "$temp%-10d",    NVVarNumOf(nptr)); break;
      default:
        fprintf(outfile, "%-15s", WNameOf(NNameOf(nptr)));   break;
    }
}
/*}}}*/

/*{{{  PRIVATE void copyrow (dest, source)   NEVER USED*/
#if 0 /* copyrow is never used */
/* Copy the contents of source to dest */
PRIVATE void copyrow ( int *dest , int *source )
{
  memcpy (dest, source, sizeof(int) * varrow);
}
#endif /* copyrow is never used */

/*}}}*/
/*{{{  PUBLIC int *save_scope()*/
/* Copy the contents of source to dest */
PUBLIC int *save_scope ( void )
{
  const size_t size = sizeof(int) * varrow;
  int *const scope = (int *)memalloc(size + sizeof(int)); /* one extra for varrow */
  scope[0] = varrow; /* save current value of varrow */
  memcpy (&scope[1], live_var, size);
  DEBUG_MSG(("save_scope: varrow is: %d\n", varrow));
  return scope;
}
/*}}}*/
/*{{{  PUBLIC void restore_scope(scope)*/
/* Copy the contents of source to dest */
PUBLIC void restore_scope ( int *scope )
{
  const int old_varrow = scope[0];
  memcpy (live_var, &scope[1], sizeof(int) * old_varrow);
  if (old_varrow != varrow) /* bug 1075 24/9/91 */
    {
      memset (&live_var[old_varrow], '\0', (varrow - old_varrow) * sizeof(int));
      /*printf("restore_scope: varrow is: %d, old_varrow is %d\n", varrow, old_varrow);*/
    }
  DEBUG_MSG(("restore_scope: varrow is: %d, old_varrow is %d\n", varrow, old_varrow));
  memfree(scope);
}
/*}}}*/

/*{{{  PUBLIC void init_mapping()*/
PUBLIC void init_mapping ( void )
{
  initrow(live_var);
  init_varmap(var_map);
}
/*}}}*/

/*{{{  PRIVATE void printbitmap()*/
PRIVATE void printbitmap ( int sorted_var[] )
{
  int i, j;
  fputs ("********** Variable allocation map **********\n", outfile);
  fputs ("      WS SIZE USAGE                 ", outfile);
  for (i = var_base; i < num_var; i++) fputc('0'+(i%10), outfile);
  fputc('\n',outfile);
  for (i = var_base; i < num_var; i++)
    {
      const int varnum = sorted_var[i];
      treenode *nptr = var_list[varnum];
      if (NVOffsetOf(nptr) == NO_SLOT)
        fprintf(outfile,"%-3d:NONE %3ld %5ld : ", i,
                numslots(nptr), NVUseCountOf(nptr));
      else
        fprintf(outfile,"%-3d:%4ld %3ld %5ld : ", i,
                NVOffsetOf(nptr), numslots(nptr), NVUseCountOf(nptr));
      printname(nptr);
      for (j=var_base; j < num_var; j++)
        fputs(testbitmap(var_map, varnum, sorted_var[j]) ? "*" : ".", outfile);
      fputc('\n', outfile);
    }
  fprintf(outfile, "*********************************************\n");
}
/*}}}*/
/*{{{  PUBLIC void create_var (nptr)*/
/* Make a new variable, and mark all live variables */
PUBLIC void create_var ( treenode *nptr )
{
  const int n = nextvarnum();
  setbit (live_var, n);
  markbitmap(var_map, live_var, var_base, num_var-1, n);
  SetNVVarNum(nptr, n);
  SetNVOffset(nptr, NO_SLOT); /* bug 1135 31/1/91 */
  if (TagOf(nptr) != T_RESERVEDWS)
    SetNVNext(nptr, NULL); /* Clear list of overlayed vars */
  var_list [n] = nptr;
  DEBUG_MSG(("create_var: %s, n=%d, RESERVED?=%d\n",
             WNameOf(NNameOf(nptr)), n, (TagOf(nptr) == T_RESERVEDWS)));
}
/*}}}*/
/*{{{  PUBLIC void create_var_subsidiary*/
PUBLIC void create_var_subsidiary (treenode *nptr)
{
  /* This creates the vectorspace and other temporaries required for this var */
  if (isinvectorspace(nptr))
    {
      SetNVVSOffset(nptr, newvs(allocsize(nptr)));
      SetNVNextTemp(nptr, enclosing_vsp_nptr); /* INSdi01515 */
      enclosing_vsp_nptr = nptr;
    }
  if (complexinitialise(NTypeOf(nptr)))
    /* Pointer and count held in workspace */
    reservelowworkspace(2);
}
/*}}}*/
/*{{{  PUBLIC treenode *get_var()*/
PUBLIC treenode *get_var(const int n)
{
  return var_list[n];
}
/*}}}*/
/*{{{  PUBLIC void kill_var (n)*/
/* variable no longer in use */
PUBLIC void kill_var ( treenode *nptr )
{
  const int n = (int)NVVarNumOf(nptr);
  DEBUG_MSG(("kill_var: n=%d\n", n));
  clearbit(live_var, n);
}
/*}}}*/
/*{{{  PUBLIC void resurrect_var(nptr)*/
/* Bring a variable back from the dead */
PUBLIC void resurrect_var ( treenode *nptr )
{
  const int n = (int)NVVarNumOf(nptr);
  DEBUG_MSG(("resurrect_var: n=%d, RESERVED?=%d ", n, (TagOf(nptr) == T_RESERVEDWS)));
  if (!testbit(live_var, n))
    {
      DEBUG_MSG(("(not already live)\n"));
      setbit(live_var, n);
      markbitmap(var_map, live_var, var_base, num_var-1, n);
      if (TagOf(nptr) == N_DECL)
        create_var_subsidiary(nptr);  /* resurrect the vectorspace etc */
    }
#ifdef DEBUG
  else
    DEBUG_MSG(("(already live)\n"));
#endif
}
/*}}}*/
/*{{{  PUBLIC treenode *resurrect_specs(tptr)*/
/* ressurect list of specs returning a pointer to the process
* following the specs.
*/
PUBLIC treenode *resurrect_specs ( treenode *tptr )
{
  for (; isspecification(tptr); tptr = DBodyOf(tptr))
    /*{{{  resurect vars*/
    {
      switch (TagOf(tptr))
        {
          case S_DECL: case S_ABBR: case S_VALABBR:
          case S_RETYPE: case S_VALRETYPE :
            {
              treenode *n = DNameOf(tptr);
              if (TagOf(n) == S_LIST)
                while (!EndOfList(n))
                  {
                    resurrect_var(ThisItem(n));
                    n = NextItem(n);
                  }
              else
                resurrect_var(n);
            }
            break;
        }
    }
    /*}}}*/
  return tptr;
}
/*}}}*/
/*{{{  PUBLIC int create_immortalvar (nptr)*/
/* Create a new variable which is live for all varibles */
PUBLIC int create_immortalvar ( treenode *nptr )
{
  const int n = nextvarnum();
  int i;
  /* mark all those in scope as interfering with the new var */
  for (i = var_base; i<=n; i++)
    setbitmap(var_map, i, n);
  var_list [n] = nptr;
  SetNVVarNum(nptr, n);
  SetNVNext(nptr, NULL); /* Clear list of overlayed vars */
  return n;
}
/*}}}*/

/*{{{  PUBLIC treenode *alloc_fixedws(slot, size)*/
/*{{{  comment*/
/* Reserve workspace from slot to (slot+size) - 1.
* Mark all live variables as interfering with this area of workspace
*
* We keep a list of nodes already created in the past.
*/
/*}}}*/
PUBLIC treenode *alloc_fixedws (const INT32 slot , const INT32 size )
{
  /* Look down the list of slots for one which is
     the same start, the same size, not allocated at all or
     is allocated within our block.
  */
  treenode *nptr = fixedws_list;
  DEBUG_MSG(("alloc_fixedws; slot:%d, size:%d\n", slot, size));
  {
    int i;
    for (i = var_base; i < num_var; i++)
      {
        treenode *ptr = var_list[i];
        if (testbit(live_var, i) && (TagOf(ptr) == T_RESERVEDWS))
          {
            const BOOL ov = vars_overlap(slot, size, NVOffsetOf(ptr), NVRSizeOf(ptr));
            DEBUG_MSG(("alloc_fixedws: Live with reserved %d, clashes? %d\n", i, ov));
            if (ov)
              genwarn_i(GEN_WORKSPACE_CLASH, NVOffsetOf(ptr));
          }
      }
  }
  while ((nptr!=NULL) &&
         !((NVOffsetOf(nptr)==slot && NVRSizeOf(nptr)==size) &&
           ((NVVarNumOf(nptr) == -1) ||
            (NVVarNumOf(nptr) >= var_base && NVVarNumOf(nptr) < num_var))))
    nptr = NVNextOf(nptr);
  if (nptr == NULL)
    /*{{{  Make new fixed workspace node*/
    {
      switch_to_real_workspace();
      nptr = newnamenode(T_RESERVEDWS, genlocn, tempname_p, NULL,
                         NULL, be_lexlevel, (BIT16)(-1), NM_DEFAULT);
      SetNVRSize(nptr, size);
      SetNVNext(nptr, fixedws_list);  fixedws_list = nptr;
      create_var(nptr);  /* Make a new var */
      DEBUG_MSG(("alloc_fixedws; creating new node\n"));
    }
    /*}}}*/
  else
    /*{{{  Use existing record*/
    {
      DEBUG_MSG(("alloc_fixedws; resurrecting?:%d\n", NVVarNumOf(nptr) != -1));
      if (NVVarNumOf(nptr) == -1)
        create_var(nptr); /* Not used in this block yet */
      else
        resurrect_var(nptr);    /* re-use this node */
    }
    /*}}}*/
  SetNVOffset(nptr, slot); /* leave this set correctly */
  return nptr;
}
/*}}}*/
/*{{{  PUBLIC void free_fixedwslist()*/
PUBLIC void free_fixedwslist ( void )
{
  /* Mark fixed workspace node as being unused */
  int i;
  DEBUG_MSG(("free_fixedwslist:\n"));
  for (i = var_base; i < num_var; i++)
    if (TagOf(var_list[i]) == T_RESERVEDWS)
      SetNVVarNum(var_list[i], -1);
}
/*}}}*/

/*{{{  PUBLIC void offsetvars (first_var, offset)  NEVER USED*/
#if 0 /* offsetvars is never used */
/* Add offset to all vars in the var list from first_var onwards */
PUBLIC void offsetvars ( int first_var , int offset )
{
  int i;
  for (i = first_var; i < num_var; i++)
    {
      treenode *var = var_list[i];
      if (TagOf(var) != T_RESERVEDWS)
        while (var != NULL)
          {
            SetNVOffset(var, NVOffsetOf(var) + offset);
            if (diagnostics)
              /*{{{  print some diagnostics*/
              {
                printname(var);
                fprintf (outfile, " moved to workspace offset %ld\n", NVOffsetOf(var));
              }
              /*}}}*/
            var = NVNextOf(var);
          }
    }
}
#endif /* offsetvars is never used */
/*}}}*/
/*{{{  PUBLIC void addnamechain (chain, item)*/
/* Add a new name onto a names namechain */
PUBLIC void addnamechain ( treenode *const chain , treenode *const item )
{
  SetNVNext(item, NVNextOf(chain));
  SetNVNext(chain, item);
  SetNVVarNum(item, NVVarNumOf(chain));
}
/*}}}*/
/*{{{  PUBLIC void reservelowworkspace (n)*/
/*****************************************************************************
 *
 *  reservelowworkspace reserves the bottom n slots of workspace.
 *  only those variable currently live will be effected.
 *
 *****************************************************************************/
PUBLIC void reservelowworkspace ( INT32 n )
{
  kill_var(alloc_fixedws(0, n));
}
/*}}}*/

/*{{{  PRIVATE void printdiagnostic*/
PRIVATE void printdiagnostic(treenode *const var, const BOOL show_ws, const BOOL mapped_onto_vsptr)
{
  printname(var);
  if (show_ws)
    fprintf (outfile, " at workspace offset %ld", NVOffsetOf(var));
  else if (mapped_onto_vsptr)
    fputs(" mapped onto vsptr", outfile);
  if (isinvectorspace(var))
    {
      INT32 vsoffset = NVVSOffsetOf(var);
      fprintf(outfile, "  vector space [%ld..%ld]",
              vsoffset, vsoffset + allocsize(var) - 1);
    }
  fputc('\n', outfile);
}
/*}}}*/
/*{{{  PRIVATE void allocate_overlays*/
PRIVATE void allocate_overlays(treenode *nptr, const INT32 base)
{
  for (; nptr != NULL; nptr = NVNextOf(nptr))
    {
      SetNVOffset(nptr, (INT32)((BIT32)base + NVOffsetOf(nptr)));
      if (diagnostics && (TagOf(nptr) != T_TEMP) && (TagOf(nptr) != T_PREEVALTEMP))
        {
          fputs(" - ", outfile);
          printdiagnostic(nptr, TRUE, FALSE);
        }
    }
}
/*}}}*/
/*{{{  PUBLIC void initallocvars*/
PUBLIC void initallocvars(void)
{
  vars_mapped_onto_vsptr = NULL; /* added for bug 1061 30/11/90 */
}
/*}}}*/
/*{{{  PUBLIC INT32 allocvars*/
/* Allocate the vars using bit map built up during first pass.
* Only first_var and above are allocated.
* Returns size of workspace allocated.
* If free_paramslots > 0 then they will be used for least used vars
* which will fit.
*/
PUBLIC INT32 allocvars (const int first_var, const int free_paramslots, const BOOL inside_par)
{
  /* If we have more than a small number of variables, we allocate them
     using the heap; if just a small number, use the stack - bug 1075 24/9/91 */
  int local_sorted_var[MAXVAR];
  int *const sorted_var = /* sorted order of vars */
    (num_var > MAXVAR) ? memalloc(sizeof(*sorted_var) * num_var)
                       : &local_sorted_var[0];
  INT32 wssize = 0;
  /*{{{  debug message*/
  if (diagnostics)
    fprintf(outfile, "Allocating vars from %d upto %d\n", var_base, num_var);
  /*}}}*/
  if (num_var > first_var)
    {
      int i;
      treenode *alloclist = NULL;
      /*{{{  initialise names*/
      for (i=first_var; i < num_var; i++)
        {
          if (TagOf(var_list[i]) != T_RESERVEDWS)
            {
               treenode *next;
               INT32 total = 0;
               const INT32 size = numslots(var_list[i]);
               for (next = var_list[i]; next != NULL; next = NVNextOf(next))
                 total += NVUseCountOf(next);
               /* IF the object is greater than a word long (changed to 2 words)
                  then we scale down the usage count:
               */
               if ((total != 0) && (size > 2) && /* bug TS/1506 02/12/91 */
                   ((alloc_strategy & ALLOC_NODIVBYSIZE) == 0))
                 {
                   /* bug 1153 - 11/2/91 - take the sizes of objects into account */
                   total /= size;
                   if (total == 0) total = 1; /* in case of `underflow' */
                 }
               SetNVUseCount(var_list[i], total);
               SetNVOffset(var_list[i], NO_SLOT);
               /* SetNVClashCount(var_list[i], clashcount(first_var, i)); */
            }
          sorted_var[i] = i;
        }
      /*}}}*/
      /*{{{  sort names*/
      {
        const int size = num_var - first_var;
        if (alloc_strategy & ALLOC_BYSCOPE)
          sup_qsort(&sorted_var[first_var], size, sizeof(int), compare_scope);
        else
          {
            sup_qsort(&sorted_var[first_var],size,sizeof(int), compare_use);
          #if 0 /* now done after allocating the first 'n' variables */
            if (size > (SORT_USAGE_CUTOFF+1))
              sup_qsort(&sorted_var[first_var+SORT_USAGE_CUTOFF],size-SORT_USAGE_CUTOFF,sizeof(int),
                        compare_scope);
          #endif
          }
      }
      /*}}}*/
      /*{{{  allocate all the variables*/
      {
        /* alloclist is the list of allocated variables kept in the order:
         *     (NVOffsetOf(n) < NVOffset(NVAllocNextOf(n))) ||
         *     ((NVOffsetOf(n) == NVOffset(NVAllocNextOf(n)) &&
         *       ((numslots(n) >= numslots(NVAllocNextOf(n)))
         */
        BOOL re_sorted = FALSE;
        for (i=first_var; i < num_var; i++)
          /*{{{  allocate next var*/
          {
            const int varnum = sorted_var[i];
            treenode *const var = var_list[varnum];
            const INT32 length = numslots(var);
            if (length == 0 || NVUseCountOf(var) == 0)
              /*{{{  Don't allocate*/
              {
                if (diagnostics)
                  { printname(var); fputs(" not allocated\n", outfile); }
              }
              /*}}}*/
            else if (!inside_par && ((alloc_strategy & ALLOC_NOVSOPT) == 0) &&
                     isinvectorspace(var) && (NVVSOffsetOf(var) == 0))
              /*{{{  Map onto vectorspace pointer*/
              {
                SetNVAllocNext(var, vars_mapped_onto_vsptr);
                vars_mapped_onto_vsptr = var;
                if (diagnostics)
                  printdiagnostic(var, FALSE, TRUE);
              }
              /*}}}*/
            else
              {
                INT32 base = 0;
                treenode *this = alloclist, *last = NULL;
                if (TagOf(var) != T_RESERVEDWS && NVOffsetOf(var) == NO_SLOT)
                  /*{{{  allocate var*/
                  {
                    BOOL quadneeded;
                    BOOL slotfound = FALSE;
                    if (needs_quadalign)
                      quadneeded = (length >= 2);
                    /* Only look for overlap with vars already allocated */
                    while ((this != NULL) && !slotfound)
                      /*{{{  look for ws slot*/
                      {
                        /* Base is the next guess at the position to allocate variable */
                        BOOL baseisquad ;
                        INT32 end = base + length;
                        if (needs_quadalign)
                        {
                          baseisquad = (base & 1) == 0;
                          if (quadneeded && !baseisquad)
                          {
                            base++; end++;
                            if (diagnostics)
                              fprintf (outfile, "align, length = %d, base = %d\n", length, base);
                          }
                        }
                        if (end <= NVOffsetOf(this)) /* bug 709 changed to '<=' from '<' 14/9/90 */
                          slotfound = TRUE; /* position found */
                        else
                          {
                            treenode *firstclash = this;
                            /*{{{  look for end of gap*/
                            while ((firstclash != NULL) &&
                                    !testbitmap(var_map, varnum, NVVarNumOf(firstclash)))
                            {
                              firstclash = NVAllocNextOf(firstclash);
                            }
                            /*}}}*/
                            if ((firstclash == NULL) ||
                                (end <= NVOffsetOf(firstclash))) /* bug 709 changed to '<=' from '<' 14/9/90 */
                              slotfound = TRUE;
                            else
                              /*{{{  look for start of next gap*/
                              {
                                base = NVOffsetOf(firstclash) + numslots(firstclash);
                                last = firstclash;
                                this = NVAllocNextOf(firstclash);
                                while ((this != NULL) && (NVOffsetOf(this) < base))
                                  {
                                    INT32 thisend = NVOffsetOf(this) + numslots(this);
                                    if (testbitmap(var_map, varnum, NVVarNumOf(this)) &&
                                        (base < thisend))
                                    {
                                      base = thisend;
                                    }
                                    last = this;
                                    this = NVAllocNextOf(this);
                                  }
                                if (needs_quadalign)
                                {
                                  baseisquad = (base & 1) == 0;
                                  if ((this == NULL) && quadneeded && !baseisquad)
                                  {
                                    base++; end++;
                                    if (diagnostics)
                                      fprintf (outfile, "align, length = %d, base = %d\n", length, base);
                                  }
                                }
                              }
                              /*}}}*/
                          }
                      }
                      /*}}}*/
                    /*{{{  Allocate Offset for this var*/
                    SetNVOffset(var, base);
                    if (diagnostics)
                      printdiagnostic(var, TRUE, FALSE);
                    allocate_overlays(NVNextOf(var), base);
                    /*}}}*/
                  }   
                  /*}}}*/
                else
                  /*{{{  find position in alloc list where this should go*/
                  {
                    while ((this != NULL) && (NVOffsetOf(this) < NVOffsetOf(var)))
                      {
                        last = this;
                        this = NVAllocNextOf(this);
                      }
                    base = NVOffsetOf(var);
                    if (diagnostics)
                      printdiagnostic(var, TRUE, FALSE);
                  }
                  /*}}}*/
                wssize = max_INT32(wssize, base+length);
                /*{{{  Add var to alloc list*/
                /* Assume :
                 *   (last == NULL || NVOffsetOf(last) < base)
                 *   (this == NULL || NVOffsetOf(this) >= base)
                 */
                while ((this != NULL) &&
                       ((NVOffsetOf(this) < base) ||
                         ((NVOffsetOf(this) == base) && (numslots(this) > length))))
                  {
                    last = this;
                    this = NVAllocNextOf(this);
                  }
                if (last == NULL)
                  {                                           /* insert on front    */
                    SetNVAllocNext(var, alloclist);
                    alloclist = var;
                  }
                else
                  {
                    SetNVAllocNext(last, var);
                    SetNVAllocNext(var, this);                /* insert after last  */
                  }
                /*}}}*/
                /*{{{  debugging */
                #if 0 /*#ifdef DEBUG*/
                {
                  treenode *tptr = alloclist;
                  fputs("alloclist: ", outfile);
                  while (tptr != NULL)
                    {
                      fprintf(outfile, "%s @:%ld-%ld, ",
                       (TagOf(tptr) == T_RESERVEDWS) ? "reserved" : WNameOf(NNameOf(tptr)),
                       NVOffsetOf(tptr), NVOffsetOf(tptr) + numslots(tptr) - 1);
                      tptr = NVAllocNextOf(tptr);
                    }
                  fputc('\n', outfile);
                }
                #endif
                /*}}}*/
                if (!re_sorted && (wssize >= RESORT_SLOTS_CUTOFF) && (i < num_var - 2))
                  /*{{{  re-sort by scope*/
                  {
                    if (diagnostics)
                      fprintf(outfile, "re-sorting %d variables by scope\n", num_var-(i+1));
                    if ((alloc_strategy & ALLOC_BYSCOPE) == 0)
                      sup_qsort(&sorted_var[i+1],num_var-(i+1),sizeof(int),
                                compare_scope);
                    re_sorted = TRUE;
                  }
                  /*}}}*/
              }   
          }       
          /*}}}*/
      }   
      /*}}}*/
      /*{{{  try and squeeze top variables into parameter slots*/
      if (free_paramslots > 0)
        {
          treenode *movelist[MAXREGS];
          int  free_slot = free_paramslots - 1;
          BOOL move_vars = TRUE;
          /* bug TS/2019 04/01/93: this used to say `free_slot > 0' */
          while (free_slot >= 0 && move_vars && wssize > 0)
            {
              int j;
              movelist[free_slot] = NULL;
              for (j = first_var; j < num_var && move_vars; j++)
                /*{{{  check we can move all vars at top of workspace*/
                {
                  treenode *const var = var_list[j];
                  const INT32 length = numslots(var);
                  if ((NVOffsetOf(var) + length) == wssize)
                    {
                      if ((TagOf(var) == T_RESERVEDWS) ||
                          (length > (int)(free_slot+1)))
                        move_vars = FALSE; /* No point in doing more moves */
                      else
                        {
                          SetNVAllocNext(var, movelist[free_slot]);
                          movelist[free_slot] = var;
                        }
                    }
                }
                /*}}}*/
              if (move_vars) /* We can move top variables into unused param slot */
                {
                  free_slot--;
                  wssize--;
                }
            }
          /*{{{  Allocate to vars to their new slots*/
          {
            INT32 top;
            if (needs_quadalign)
              wssize = (wssize + 1) & (~1); /* round up to an even number */
            free_slot++;
            top = wssize + INS_EXTRA + (int)(REG_PARAMS-free_paramslots) + free_slot;
            while (free_slot < free_paramslots)
              {
                treenode *var;
                for (var = movelist[free_slot]; var != NULL; var = NVAllocNextOf(var))
                  {
                    const INT32 base = top - (numslots(var)-1);
                    const INT32 original_base = NVOffsetOf(var); /* bug INSdi01968 */
                    /*{{{  Allocate Offset for this var*/
                    SetNVOffset(var, base);
                    if (diagnostics)
                      /*{{{  print some diagnostics*/
                      {
                        printname(var);
                        fprintf (outfile, " moved to param slot, ws offset = %ld\n", NVOffsetOf(var));
                      }
                      /*}}}*/
                    /*}}}*/
                    /*{{{  allocate offsets for overlayed vars*/
                    {
                      treenode *nextvar;
                      for (nextvar = NVNextOf(var); nextvar != NULL; nextvar = NVNextOf(nextvar))
                        {
                          const INT32 old_diff = NVOffsetOf(nextvar) - original_base;
                          SetNVOffset(nextvar, base + old_diff);
                          if (diagnostics)
                            /*{{{  print some diagnostics*/
                            {
                              fputs(" + ", outfile);
                              printname(nextvar);
                              if (old_diff != 0)
                                fprintf(outfile, " moved to overlay, ws offset = %ld", NVOffsetOf(nextvar));
                              fputc('\n', outfile);
                            }
                            /*}}}*/
                        }
                    }
                    /*}}}*/
                  }
                top++;
                free_slot++;
              }
          }
          /*}}}*/
        }
      /*}}}*/
      /*{{{  debug message*/
      if (diagnostics)
        printbitmap(sorted_var);
      /*}}}*/
      /*{{{  check allocation -- commented out*/
      #if 0
      {
        int i,j;
        for (i=first_var; i < num_var; i++)
        {
          treenode *v1 = var_list[sorted_var[i]];
          for (j=first_var; j < num_var; j++)
            if (i != j)
            {
              treenode *v2 = var_list[sorted_var[j]];
              if (TagOf(v1) == T_RESERVEDWS && TagOf(v2) == T_RESERVEDWS)
                ;
              else if (NVOffsetOf(v1) == NO_SLOT || NVOffsetOf(v2) == NO_SLOT)
                ;
              else if (testbitmap(var_map, sorted_var[i], sorted_var[j]) &&
                  vars_overlap (NVOffsetOf(v1), numslots(v1),
                                NVOffsetOf(v2), numslots(v2)))
              {
                fprintf(outfile, "(%d)%s, (%d)%s\n",
                        i, WNameOf(NNameOf(v1)), j, WNameOf(NNameOf(v2)));
                badtag(genlocn,TagOf(v1),
                          "allocvars, var allocation failed");
              }
            }
        }
      }
      #endif
      /*}}}*/
    }
  if (num_var > MAXVAR) { memfree(sorted_var); }
  if (needs_quadalign)
    wssize = (wssize + 1) & (~1); /* round up to an even number */
  return (wssize);
}     
/*}}}*/
/*{{{  label allocation*/
/*{{{  PUBLIC int newlab ()*/
PUBLIC int newlab ( void )
{
  return (nextlabel++);
}
/*}}}*/
/*{{{  PUBLIC void initlabels()*/
PUBLIC void initlabels ( void )
{
  nextlabel = 0;
}
/*}}}*/
/*}}}*/
/*{{{  use count handling*/
#define HUGEVAL ((INT32)0x7fffffff)
/*{{{  PUBLIC void setusecount*/
PUBLIC void setusecount(treenode *const nptr, const INT32 n)
{
  /*DEBUG_MSG(("setusecount: setting usecount of %s to %ld\n",
             WNameOf(NNameOf(nptr)), n));*/
  SetNVUseCount(nptr, n);
}
/*}}}*/
/*{{{  PUBLIC void addusecount(nptr, n)*/
/* up the use count of a variable by n, where n has already been loop weighted
 * If the use count would overflow it is set to HUGEVAL
 */
PUBLIC void addusecount ( treenode *nptr , INT32 n )
{
  const INT32 oldcount = NVUseCountOf(nptr);
  if (n > (HUGEVAL - oldcount))
    SetNVUseCount(nptr, HUGEVAL);
  else
    {
    /*DEBUG_MSG(("addusecount: setting usecount of %s to %ld\n",
                 WNameOf(NNameOf(nptr)), oldcount + n));*/
      SetNVUseCount(nptr, oldcount + n);
    }
}
/*}}}*/
/*{{{  PUBLIC void upusecount(nptr, n)*/
/* up the use count of a variable n times.
 * If the use count would overflow it is set to HUGEVAL
 */
PUBLIC void upusecount (treenode *const nptr, const INT32 n)
{
  if (loop_weight == 0)
    ; /* skip */
  else if (n > (HUGEVAL / loop_weight))
    SetNVUseCount(nptr, HUGEVAL);
  else
    addusecount(nptr, loop_weight * n);
}
/*}}}*/
/*{{{  PUBLIC void uploop_weight(n)*/
/* up loop_weight n times
 * If the loop_weight would overflow it is set to HUGEVAL
 */
PUBLIC void uploop_weight ( INT32 n )
{
  if (n == 0 || loop_weight == 0)
    loop_weight = 0;
  else if (n > (HUGEVAL / loop_weight))
    loop_weight = HUGEVAL;
  else
    loop_weight *= n;
}
/*}}}*/
/*}}}*/
/*{{{  temporary slot allocation*/
/*{{{  comment*/
/*
 *  To allocate a temporary we create a new temporary node then look through
 *  the current variable table looking for a temporary of the same size
 *  which is unused.
 *  If an unused temporary is found we add the new temporary node  to the
 *    variable chain of the unused temporary. This gives the new temporary
 *    the same variable number as the unused temporary so they will share the
 *    same slot in workspace .
 *    The unused temporary is then resurrected and a pointer to the new
 *    temporary node returned.
 *  If an unused temporary is not found the new temporary node is added
 *    to the variable table.
 *
 * To free a temporary we simply kill the entry in the variable table given by
 * the tempories variable number.
 */
/*}}}*/

/*{{{  PUBLIC void alloctemp(treenode *tempptr)*/
PUBLIC void alloctemp ( treenode *tempptr )
{
  int i = var_base;
  /*INT32 size = wordsin(NTypeOf(tempptr));*/
  const INT32 size = numslots(tempptr);
  /* look for an unused temp of the same size */
#if 0
  while ((i < num_var) &&
         (TagOf(var_list[i]) != T_TEMP ||
          testbit(live_var, i) != 0 || size != numslots(var_list[i]) ))
    i++;
#else
  /* look for preeval temps too! */
  while ((i < num_var) &&
         ((TagOf(var_list[i]) != T_TEMP && TagOf(var_list[i]) != T_PREEVALTEMP) ||
          testbit(live_var, i) != 0 || size != numslots(var_list[i]) ))
    i++;
#endif
  if (i < num_var)
    {
      /* Add the new temp node to the unused tempories variable chain
       * and resurrect the old temporary
       */
      DEBUG_MSG(("alloctemp: Re-using temp%d\n", i));
      addnamechain(var_list[i], tempptr);
      SetNVOffset(tempptr, 0);
      resurrect_var(var_list[i]);
    }
  else
    {
      /* Allocate a new variable */
      DEBUG_MSG(("alloctemp: New temp%d\n", i));
      create_var(tempptr);
    }
}
/*}}}*/
/*{{{  PRIVATE treenode *newtypedtempnode(t, tptr, mode, type)*/
PRIVATE treenode *newtypedtempnode (const int t,    treenode *const tptr,
                                    const int mode, treenode *const type,
                                    const int current_lexlevel)
{
  treenode *const tempptr = newnamenode(t, genlocn, tempname_p, type, tptr,
                                  current_lexlevel, 0, mode);
  DEBUG_MSG(("newtypedtempnode\n"));
  SetNVOffset(tempptr, NO_SLOT);
  SetNVNext(tempptr, NULL);
  setusecount(tempptr, 1); /* make sure it always gets a slot allocated */
  return(tempptr);
}
/*}}}*/
/*{{{  PUBLIC treenode *newtempnode(t, tptr, mode)*/
PUBLIC treenode *newtempnode (const int t, treenode *const tptr,
                              const int mode, const int current_lexlevel)
{
#if 1
  /* We call gettype_main so that any type tree fabricated is created
     in real workspace, not temporary workspace */
  return (newtypedtempnode(t, tptr, mode, gettype_main(tptr), current_lexlevel));
#else
  /* tried using copytree here to fix bug INSdi03067 - unsuccessful */
  return (newtypedtempnode(t, tptr, mode, copytree(gettype(tptr), current_lexlevel), current_lexlevel));
#endif
}
/*}}}*/
/*{{{  PUBLIC treenode *gettemp(tptr, mode)*/
PUBLIC treenode *gettemp ( treenode *const tptr, const int mode )
{
  treenode *const tempptr = newtempnode(T_TEMP, tptr, mode, be_lexlevel);
  alloctemp(tempptr);
  return(tempptr);
}
/*}}}*/
/*{{{  PUBLIC treenode *gettypedtemp(tptr, mode, type)*/
PUBLIC treenode *gettypedtemp ( treenode *tptr , int mode , treenode *type )
{
  treenode *tempptr = newtypedtempnode(T_TEMP, tptr, mode, type, be_lexlevel);
  alloctemp(tempptr);
  return(tempptr);
}
/*}}}*/
/*{{{  PUBLIC treenode *getopdtemp(int opdmode, treenode *opd)*/
/****************************************************************************
 *
 *  getopdtemp takes an operand (opdmode, opd) and returns a suitable temp.
 *             node for it.
 *             This is a bit of a kludge as it pretends a pointer is an
 *             integer in workspace; ideally, we should have proper pointer
 *             types in the backend.
 *
 ****************************************************************************/
PUBLIC treenode *getopdtemp ( int opdmode , treenode *opd )
{
  return(gettemp(opd, (opdmode == P_PTR) ? NM_POINTER : NM_WORKSPACE));
}
/*}}}*/
/*{{{  PUBLIC void freetemp(temp)*/
PUBLIC void freetemp ( treenode *temp )
{
  kill_var(temp);
}
/*}}}*/
/*{{{  PUBLIC void freeiftemp(treenode *tptr)*/
/*{{{  comment*/
/*****************************************************************************
 *
 *  freeiftemp takes a treenode 'tptr', and if it is a temporary, or
 *             evaluated temporary, frees it.
 *
 *****************************************************************************/
/*}}}*/
PUBLIC void freeiftemp ( treenode *tptr )
{
  if (TagOf(tptr) == T_TEMP || TagOf(tptr) == T_PREEVALTEMP)
    freetemp(tptr);
}
/*}}}*/
/*}}}*/
/*{{{  PUBLIC void allocparams(nptr)*/
/*****************************************************************************
 *
 *  allocparams allocates workspace offsets to the formal parameters.
 *
 *****************************************************************************/
PUBLIC void allocparams ( treenode *nptr )
{
  treenode *fparams = NParamListOf(nptr);
  /*{{{  allocate workspace positions to the parameters*/
  {
    int hiddentempno = -1; /* numbering for the hidden parameters */
    /*int paramposn = 1;*/
    INT32 paramposn = NPMaxwspOf(nptr) + 1;
    int olddiagnostics = diagnostics;
    if (separatelycompiled(nptr)) diagnostics = FALSE;
    if (diagnostics)
      fprintf(outfile, "******* Parameter workspace allocation ******\n");
    while (!EndOfList(fparams))
      {
        treenode *thisfparam = ThisItem(fparams);
        switch(TagOf(thisfparam))
          {
            /*{{{  formal parameter*/
            case N_PARAM:
            case N_VALPARAM:
            case S_FNFORMALRESULT:
              if (basetype(gettype(thisfparam)) != S_TIMER)
                /*{{{  allocate the parameter*/
                {
                  SetNVOffset(thisfparam, paramposn);
                  if (diagnostics)
                    fprintf (outfile, "Parameter %s at workspace offset %ld, use count = %ld\n",
                             WNameOf(NNameOf(thisfparam)), NVOffsetOf(thisfparam),
                             NVUseCountOf(thisfparam));
                  allocate_overlays(NVNextOf(thisfparam), paramposn);
                }
                /*}}}*/
              else
                /* decrement paramposn to cancel out the increment below */
                paramposn--;
              break;
            /*}}}*/
            /*{{{  hidden array dimension*/
            case S_HIDDEN_PARAM:
              SetNVOffset(HExpOf(thisfparam), paramposn);
              /* added for more readability of assembly output: */
              SetNVVarNum(HExpOf(thisfparam), hiddentempno--); /* 14/3/91 */
              if (diagnostics)
                fprintf(outfile, "Hidden param at workspace offset %ld\n", paramposn);
              break;
            /*}}}*/
            /*{{{  static link*/
            case S_PARAM_STATICLINK:
              if (diagnostics)
                fprintf(outfile, "Static link at workspace offset %ld\n", paramposn);
              break;
            /*}}}*/
            /*{{{  vector space pointer*/
            case S_PARAM_VSP:
              {
                treenode *vptr;
                for (vptr = vars_mapped_onto_vsptr;
                     vptr != NULL;
                     vptr = NVAllocNextOf(vptr))
                  {
                    SetNVOffset(vptr, paramposn);
                    if (diagnostics)
                      printdiagnostic(vptr, TRUE, FALSE);
                    allocate_overlays(NVNextOf(vptr), paramposn);
                  }
                if (diagnostics)
                  fprintf(outfile, "Vectorspace pointer at workspace offset %ld\n", paramposn);
              }
              break;
            /*}}}*/
            /*{{{  function result pointer*/
            #if 0
            case S_FNFORMALRESULT:
              SetNVOffset(thisfparam, paramposn);
              if (diagnostics)
                fputs("Function result", outfile);
              break;
            #endif
            /*}}}*/
            default:
              badtag(genlocn, TagOf(thisfparam), "allocparams");
          }
        paramposn++;
        fparams = NextItem(fparams);
      }
    if (diagnostics)
      fprintf(outfile, "*********************************************\n");
    diagnostics = olddiagnostics;
  }
  /*}}}*/
}
/*}}}*/
/*{{{  PUBLIC void initalloc ()*/
PUBLIC void initalloc ( void )
{
  check_bitmap_relationships();

  staticlinkused = 0;
  initvsp(0);
  datasize = DS_MIN;
  fixedws_list = NULL;
  if (var_map == NULL) /* incase this is called more than once (in configurer) */
    var_map = alloc_var_map();
  if (live_var == NULL)
    live_var = alloc_live_var();
  if (var_list == NULL)
    var_list = alloc_var_list();
}
/*}}}*/
/*{{{  PUBLIC void freevarmap*/
PUBLIC void freevarmap(void)
{
  if (var_map  != NULL) { memfree(var_map);  var_map  = NULL; }
  if (live_var != NULL) { memfree(live_var); live_var = NULL; } /* bug 1075 24/9/91 */
  if (var_list != NULL) { memfree(var_list); var_list = NULL; } /* bug 1075 24/9/91 */
}
/*}}}*/
/*}}}*/
