/* $Id: code1.c,v 1.16 1998/10/05 11:53:56 dcw Exp $ */

/*
 *	instruction generation, code adjustment, code output
 *	Copyright (C) 1987 Inmos Limited
 *
 *	This program is free software; you can redistribute it and/or modify
 *	it under the terms of the GNU General Public License as published by
 *	the Free Software Foundation; either version 2 of the License, or
 *	(at your option) any later version.
 *
 *	This program is distributed in the hope that it will be useful,
 *	but WITHOUT ANY WARRANTY; without even the implied warranty of
 *	MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *	GNU General Public License for more details.
 *
 *	You should have received a copy of the GNU General Public License
 *	along with this program; if not, write to the Free Software
 *	Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 */

/*#define DEBUG*/

/*{{{  include files*/
#ifdef HAVE_CONFIG_H
#  include <config.h>
#endif
#include <stdio.h>
#include <string.h>
#include "includes.h"
#include "genhdr.h"
#include "instruct.h"
#include "instdef.h"
#include "generror.h"

#include "objwrt.h"
#include "gen1def.h"
#include "gen2def.h"
#include "gen4def.h"
#include "code1def.h"
#include "bind2def.h"
#include "bind3def.h"
#include "debugdef.h"
#include "profile.h"

#include "maplib.h" /* mapfile handling */ /* IMPORTED */
/*}}}*/

/*{{{  definitions*/
/* The actual code buffer size is CODE_SIZE * sizeof(INT32) */
#define CONFIGURING_CODE_SIZE 10000 /* approx  40K bytes */
#define COMPILING_CODE_SIZE   60000 /* approx 240K bytes */

PUBLIC  int req_code_size = 0; /* code size requested (in K) */
PRIVATE int max_code_size;     /* buffer size (in INT32s)    */


#define MAX_BYTES_IN_INSTRUCTION 8  /* 32 bit target wordsize maximum */
#define MAX_CONSTANT_SIZE        8  /* 8 bytes for a REAL64 etc */

/*{{{  code tags*/
#define C_NIL (-1)
#define C_SETLAB (-11)
#define C_USELAB (-12)
#define C_USELABDIFF (-13)
#define C_STARTFILE (-14)
#define C_ENDFILE (-15)
#define C_LOCATE (-16)
#define C_SCALE (-17)
#define C_TABLE (-18)
#define C_LIBPATCH (-19)
#define C_ADDRESSFIX (-20)
#define C_PROFADDRESSFIX (-21)
/*}}}*/

/*{{{  fields in a setlab record*/
#define SL_SIZE     3                                 /* Length of the record */
#define SL_ADDR     1                /* Last calculated address of the setlab */
#define SL_NEXT     2                 /* Pointer to next item on adjust chain */
/*}}}*/
/*{{{  fields in a uselab record*/
#define U_SIZE     6                                 /* Length of the record */
#define U_ADDR     1           /* Last calculated address of the label usage */
#define U_NEXT     2                 /* Pointer to next item on adjust chain */
#define U_DEF      3                          /* Pointer to label definition */
#define U_ADJ      4  /* Current estimated length of instruction using label */
#define U_OP       5                              /* Instruction using label */
/*}}}*/
/*{{{  fields in a uselabdiff record*/
#define UD_SIZE    7                                 /* Length of the record */
#define UD_ADDR    1       /* Last calculated address of the labeldiff usage */
#define UD_NEXT    2                 /* Pointer to next item on adjust chain */
#define UD_DEF1    3                 /* Pointer to definition of first label */
#define UD_ADJ     4  /* Current estimated length of instruction using label */
#define UD_OP      5                          /* Instruction using labeldiff */
#define UD_DEF2    6                /* Pointer to definition of second label */
/*}}}*/
/*{{{  fields in a label record*/
#define LB_SIZE    3                                 /* Length of the record */
#define LB_NAME    0                                           /* Label name */
#define LB_PRAD    1                  /* Address of the corresponding setlab */
#define LB_DEF     2                  /* Pointer to the corresponding setlab */
/*}}}*/
/*{{{  fields in a table record*/
#define T_SIZE     3                                 /* Length of the record */
#define T_SCALE    1            /* Pointer to field in related scale record  */
#define T_NEXT     2                 /* Pointer to next item on adjust chain */
/*}}}*/
/*{{{  fields in a scale record*/
#define SC_SIZE    4                                 /* Length of the record */
#define SC_VALUE   1              /* Last calculated size of the scale value */
#define SC_NEXT    2                 /* Pointer to next item on adjust chain */
#define SC_OP      3                         /* Instruction using case scale */
/*}}}*/
/*{{{  fields in a debug record*/
#define DB_SIZE    5                                 /* Length of the record */
#define DB_ADDR    1            /* Last calculated address of the debug item */
#define DB_NEXT    2                 /* Pointer to next item on adjust chain */
#define DB_LINE    3                     /* Source line number of debug item */
#define DB_DATA    4                       /* Label of associated code chunk */
/* bug 1158 means that multiple tables can have the same addresses;
   these are marked with address fixes, where the DB_DATA field is the label
   number of the section of code for that table
   CON 14/2/91
*/
/*}}}*/


/*{{{  segment definitions*/
typedef struct seginfo_struct
  {
    struct seginfo_struct *seg_next;
    treenode *seg_name;
    /*INT32 seg_entry;*/
  } seginfo_t;
/*}}}*/

#define IBUF_SIZE 256  /* amount copied at one time to code file */
/* This must be small enough to fit into the output file buffer */

/*}}}*/
/*{{{  global variables*/
/*{{{  public variables*/
/* All initialised to NULL */
PUBLIC void (*code_pseudo_link_fn)(treenode *);
PUBLIC void (*code_save_entrydata_fn)(const treenode *const nptr,
                              const INT32 offset, const INT32 code_size);
PUBLIC void (*code_link_map_fn)(void);
/*}}}*/

/* It is used for the T9000_gamma_badldnlp and the t450 nop workaround. */
PRIVATE int prev_instruction;
PRIVATE int dead;
PRIVATE treenode *libentrychain;
/*{{{  data for code buffer handling*/
PRIVATE INT32 *code = NULL;          /* Buffer to hold code during compaction */

PRIVATE int cstart;          /* Pointer to first free location in code buffer */
PRIVATE int cptr;                    /* Pointer to top of code in code buffer */
PRIVATE int lptr;          /* Pointer to bottom of label info. in code buffer */
PRIVATE int topl;       /* Pointer to top of chain of label defns. and usages */
PRIVATE int start; /* Pointer to start of current code segment in code buffer */
PRIVATE int coden;         /* Number of bytes of code in current code segment */
             /* Word used to pack bytes of code before placing in code buffer*/
PRIVATE BIT32 codew;
PRIVATE int partword;                         /* Used during packing of codew */
PRIVATE INT32 address;            /* Current estimate of code address within SC */
                     /* Pointer to label which marks entry point for section */
PRIVATE int section_label;
            /* Pointer to most recently declared scale record in code buffer */
PRIVATE int case_scale;
PRIVATE INT32 getword;              /* Word of four bytes read from code buffer */
                    /* Bytes packed into getword still to be written to ibuf */
PRIVATE int bytesleft;
PRIVATE int getptr;     /* Next word to be taken from code buffer by put_code */

PRIVATE BIT32 section_size; /* Current estimate of size of current code section */
PRIVATE BIT32 total_code_size;                             /* Total code size */
PRIVATE BIT32 bytes_output;
PRIVATE BYTE libstub[] = { I_PFIX, I_PFIX, I_PFIX, I_PFIX,
                           I_PFIX, I_PFIX, I_PFIX, I_PFIX };
  /* libstub is also used for entry point zero padding */
PRIVATE SOURCEPOSN oldlocn;

/*{{{  STOP on error handling*/
typedef enum {
  ERROR_STATUS_START,   /* At the start of a block    */
  ERROR_STATUS_CLEAN,   /* Error is known to be clear */
  ERROR_STATUS_DIRTY    /* Error may have been set    */
  } error_status_t;

PRIVATE error_status_t int_error_status;
PRIVATE error_status_t fpu_error_status;

#define LOCALISE_ERROR_HANDLING TRUE
/* TRUE if we want to wait until generating an FPU instruction before clearing
   the FPU error flag
   BUT we need a way to clear FPU error without clobbering integer stack.
   There isn't one, cos both FPTESTERR and FPUCLRERR hit 1 stack slot.
   AHAH! We now use the fact that FPU loads always use a CPU slot,
   so we can use that after the load has happened!  CO'N 12/2/90
   We can't use this system for normal errors, cos the 'TESTERR' instruction
   clobbers the stack
*/

#if LOCALISE_ERROR_HANDLING
#define ERROR_STATUS_FP_INIT ERROR_STATUS_START
#else
#define ERROR_STATUS_FP_INIT ERROR_STATUS_CLEAN
#endif
/*}}}*/

/*}}}*/
/*{{{  data for segment handling*/
PRIVATE seginfo_t *seginfo;

PRIVATE seginfo_t *mapinfo_list = NULL;
/*}}}*/
/*}}}*/

/*{{{  private routines*/
/*{{{  picture of the code buffer*/
/*****************************************************************************
 *
 *   <---- Code                                              Labels    ---->
 * ---------------------------------------------------------------------------
 * |  |   |    |                     |       |            | | | | | | | | |  |
 * |  |   |    | current code        |       | current    |   entry points   |
 * |  |   |    |     --->            |       | labels     |   for previously |
 * |  |   |    |                     |       |  <---      |   compacted      |
 * |  |   |    |                     |       |            |   routines       |
 * |  |   |    |                     |       |            | | | | | | | | |  |
 * |  |   |    |                     |       |            | | | | | | | | |  |
 * ---------------------------------------------------------------------------
 *  previously  ^                     ^       ^          ^
 *  compacted   cstart                cptr    lptr       section
 *  routines                                             label
 *
 *****************************************************************************/

/*{{{  format of previously compacted routines*/
/*****************************************************************************
 *
 * ---------------------------------------------------------------------------
 * | |                      | # |            | |
 * | |                      | # |            | |
 * | |                      | # |            | |
 * | |                      | # |            | |
 * | |                      | # |            | |
 * ---------------------------------------------------------------------------
 *  ^                        | ^              |
 *  |                        | |              |
 *  -------------------------  ----------------
 *  ^                          ^
 *  |                          |
 *  ---bytes in this routine ---
 *                           ^                ^
 *                           |                |
 *                      pointer to start of routine
 *****************************************************************************/
/*}}}*/

/*}}}*/
/*{{{  PRIVATE BOOL isbranchop(op)*/
/*****************************************************************************
 *
 *  isbranchop if the opcode 'op' represents a branch instruction.
 *
 *****************************************************************************/
PRIVATE BOOL isbranchop (const INT32 op )
{
  return (op == I_J)/* || (op == I_CJ)*/; /* Bug 1378 - 11/9/91 removed cj */
}
/*}}}*/
/*{{{  low-level code buffer handling*/
/*{{{  PRIVATE BYTE get_code()*/
/**************************************************************************
 *
 * get_code returns the next byte from the code buffer.
 *          The current position in the code buffer is represented by
 *          (getptr, getword, bytesleft).
 *
 **************************************************************************/
PRIVATE BYTE get_code ( void )
{
  BYTE r;
  if (bytesleft == 0)
    {
      getword = code[getptr++];
      bytesleft = 4;
    }
  r = (BYTE)(getword & 0xff);
  getword >>= 8;
  bytesleft--;
  return(r);
}
/*}}}*/
/*{{{  PRIVATE void add_code(c)*/
/***************************************************************************
 *
 * add_code adds a byte of code to the code buffer. Every time four bytes
 *          are assembled, they are written to the code buffer at cptr.
 *
 **************************************************************************/
PRIVATE void add_code (const BYTE c )
{
  codew |= ((BIT32)c << partword);
  coden++;
  address++;
  if (partword == 24)
    /*{{{  write codew to code buffer*/
    {
      code[cptr++] = codew;
      if (cptr > lptr)
        generr_i(GEN_CODE_BUFFER_FULL, cptr*(BIT32)sizeof(INT32));
      partword = 0;
      codew = 0;
    }
    /*}}}*/
  else
    {
      if (start == C_NIL)
        /*{{{  start new code section*/
        {
          start = cptr;
          cptr++;      /* Save word for code length */
        }
        /*}}}*/
      partword += 8;
    }
}
/*}}}*/
/*}}}*/
/*{{{  input*/
/*{{{  code input*/
/*{{{  support*/
/*{{{  PRIVATE int assemble_instruction(v)*/
/***************************************************************************
 *
 * assemble_instruction adds the prefix bytes required for operand to the
 *        buffer, and returns the length
 *
 **************************************************************************/
PRIVATE int assemble_instruction (BYTE buf[], const INT32 instruction, INT32 operand )
{
  int ptr;

  if (operand > 0)  /* most common */
    {
      ptr = 0;
      while (operand > 0)
        {
          buf[ptr++] = (BYTE)(I_PFIX | (operand & 0xf));
          operand >>= 4;
        }
    }
  else if (operand == 0)
    {
      buf[0] = 0;
      ptr = 1;
    }
  else if (operand >= -16)
    {
      buf[1] = I_NFIX | 0;
      buf[0] = (BYTE)(operand & 0xf);
      ptr = 2;
    }
  else
    {
      ptr = 0;
      operand = ~operand;
      while (operand > 15)
        {
          buf[ptr++] = (BYTE)(I_PFIX | ((~operand) & 0xf));
          operand >>= 4;
        }
      buf[ptr++] = (BYTE)(I_NFIX | operand);
    }

  buf[0] = (BYTE)(instruction | (int)(buf[0] & 0xf));

  return ptr;
}
/*}}}*/
/*{{{  PRIVATE void put_pfix (instruction, operand, size)*/
/*{{{  comment*/
/***************************************************************************
 *
 *  put_pfix adds the primary (instruction, operand) into the code buffer,
 *           prefixed if neccessary so that it occupies 'size' bytes.
 *
 **************************************************************************/
/*}}}*/
PRIVATE void put_pfix (const INT32 instruction, const INT32 operand, const INT32 size)
{
  int i;
  if (operand >= 0)
    for (i = 1; i < size; i++)
      add_code(I_PFIX | (BYTE)((operand >> (( size - i) * 4)) & 0xf));
  else
    {
      const int shifts = (int)size - 2;
      add_code(I_NFIX | (BYTE)((~(operand >> ((size - 1) * 4))) & 0xf));
      for (i = 0; i < shifts; i++)
        add_code(I_PFIX | (BYTE)((operand >> ((shifts - i) * 4)) & 0xf));
    }
  add_code((BYTE)(instruction | (operand & 0xf)));
}
/*}}}*/
/*}}}*/
/*{{{  PRIVATE BOOL is_multic_alu_op(const INT32 op)*/
PRIVATE BOOL is_multic_alu_op(const INT32 op)
{
  return(op == I_ADD     || op == I_BITCNT  || op == I_BITREVNBITS  || 
         op == I_CB      || op == I_CBU     ||
         op == I_CCNT1   || op == I_CFLERR  || op == I_CIR          || 
         op == I_CIRU    || op == I_CRCBYTE || 
         op == I_CRCWORD || op == I_CS      || op == I_CSNGL        || 
         op == I_CSU     || op == I_CSUB0   || 
         op == I_CWORD   || op == I_DIV     || op == I_FMUL         || 
         op == I_GT      || op == I_GTU     ||
         op == I_LADD    || op == I_LDIV    || op == I_LMUL         || 
         op == I_LSHL    || op == I_LSHR    ||
         op == I_LSUB    || op == I_MUL     || op == I_NORM         || 
         op == I_POSTNORMSN || op == I_PROD ||
         op == I_REM     || op == I_ROUNDSN || op == I_SATADD       || 
         op == I_SATMUL  || op == I_SATSUB  ||
         op == I_SLMUL   || op == I_SUB     || op == I_SULMUL       || 
         op == I_UNPACKSN|| op == I_XBWORD  ||
         op == I_XSWORD  || op == I_XWORD                                  );
}
/*}}}*/
/*{{{  PRIVATE BOOL is_alu_op(const INT32 op)*/
PRIVATE BOOL is_alu_op(const INT32 op)
{
  return(is_multic_alu_op(op) ||  
         op == I_AND     || op == I_DIFF  || op == I_LDIFF || 
         op == I_LSUM    || op == I_NOT   || op == I_OR    || 
         op == I_SHL     || op == I_SHR   || op == I_SUM   || 
         op == I_XDBLE   || op == I_XOR                          );

}
/*}}}*/
/*{{{  PRIVATE void wa_t450_nop_t9gamma_badldnlp(const int instruction)*/
PRIVATE void wa_t450_nop_t9gamma_badldnlp(const int instruction)
{
  if (T9000_gamma_badldnlp(&tx_global))
  {
    if ((instruction == I_LDNLP || instruction == I_STNL ) &&
        prev_instruction == I_LDNLP)
    {
      gensecondary(I_NOP);
      gencomment0("Workaround for bad ldnlp");
    }
  }
  else if (!inside_asm)
  {

    if (is_multic_alu_op(instruction)  &&
        is_alu_op(prev_instruction )       )
      {
        gensecondary(I_NOP);
        gencomment0("Workaround for alu, multicycle alu op pairs");
      }
  }
  if (t450a_workarounds && inside_asm)
    prev_instruction = I_NOP;
  else
    prev_instruction = instruction;
}
/*}}}*/
/*{{{  PRIVATE void code_primary(instruction, operand)*/
/***************************************************************************
 *
 * code_primary adds primary instruction, operand to the code buffer.
 *
 **************************************************************************/
PRIVATE void code_primary (const INT32 instruction, const INT32 operand )
{
  BYTE buf[MAX_BYTES_IN_INSTRUCTION];
  int ptr;
  if ((operand == 0) && ((instruction == I_LDNLP) ||
                              (instruction == I_AJW) ||
                              (instruction == I_ADC)))
    /* Null instruction */
    return;
  if (t450a_workarounds || T9000_gamma_badldnlp(&tx_global))
    wa_t450_nop_t9gamma_badldnlp((instruction == I_OPR)? 
                                 operand               :
                                 instruction | I_PRIMARY);
  ptr = assemble_instruction(buf, instruction, operand);

  while (--ptr >= 0)
    add_code(buf[ptr]);
}
/*}}}*/
/*{{{  PRIVATE void code_secondary(instruction)*/
/***************************************************************************
 *
 * code_secondary adds secondary instruction to the code buffer.
 *
 **************************************************************************/
PRIVATE void code_secondary (const int instruction )
{
  /*code_primary(I_OPR, instruction);*/
  code_primary(I_OPR,
     (instruction & I_NEGATIVE) ? (int)(-(instruction & (~I_NEGATIVE)))
                                : instruction);
}
/*}}}*/
/*}}}*/
/*{{{  labels and tables input*/
/*{{{  support*/
/*{{{  PRIVATE int mark_code(type, size)*/
/***************************************************************************
 *
 * mark_code inserts a record of type type and size size into the code
 *           buffer, and returns the index of the start of the record.
 *
 **************************************************************************/
PRIVATE int mark_code (const int type, const int size )
{
  int ptr;
  if (partword != 0)
    /*{{{  write partial word of code to code buffer*/
    {
      code[cptr++] = (INT32)codew;
      if (cptr > lptr)
        generr_i(GEN_CODE_BUFFER_FULL, cptr*(BIT32)sizeof(INT32));
      partword = 0;
      codew = 0;
    }
    /*}}}*/
  if (start != C_NIL)
    /*{{{  close up code segment*/
    {
      if (coden == 0)
        cptr = start;   /* Empty code segment, erase it */
      else
        code[start] = coden; /* Save length of code segment */
      start = C_NIL;
    }
    /*}}}*/
  coden = 0;
  ptr = cptr;
  if ((cptr + size) < lptr)
    /*{{{  create node, add it to adjust node chain*/
    {
      code[cptr] = type;
      /* Add node to adjust node chain */
      if (topl != C_NIL) code[topl + U_NEXT] = cptr;
      topl = cptr;
      /* Reserve space for the node */
      cptr += size;
    }
    /*}}}*/
  else
    generr_i(GEN_CODE_BUFFER_FULL, cptr*(BIT32)sizeof(INT32));
  return(ptr);
}
/*}}}*/
/*{{{  PRIVATE int addlabel(label)*/
/**************************************************************************
 *
 *  addlabel creates a new label record with name 'label', and returns
 *           the index of the new label.
 *
 **************************************************************************/
PRIVATE int addlabel (const int label )
{
  lptr -= LB_SIZE;
  if (lptr > cptr)
    {
      code[lptr + LB_NAME] = label;
      code[lptr + LB_DEF] = C_NIL;
      code[lptr + LB_PRAD] = 0;
    }
  else
  /*generr_i(GEN_LABEL_BUFFER_FULL, (max_code_size - lptr) / LB_SIZE);*/
    generr_i(GEN_CODE_BUFFER_FULL, cptr * (BIT32)sizeof(INT32));
  return(lptr);
}
/*}}}*/
/*{{{  PRIVATE int findlabel(label)*/
/**************************************************************************
 *
 * findlabel searches for label 'label' and if found returns the index
 *           of the label record. If not found, a new record is created
 *           for 'label' and its index is returned.
 *
 **************************************************************************/
PRIVATE int findlabel (const int label )
{
  int l = lptr;
  while ((l < max_code_size) && (code[l + LB_NAME] != label))
    l += LB_SIZE;
  if (l < max_code_size)    /* Label found */
    return(l);
  else
    {                  /* Create new label node */
      return(addlabel(label));
    }
}
/*}}}*/
/*}}}*/
/*{{{  PRIVATE void label_def(label, type)*/
/***************************************************************************
 *
 * label_def inserts a setlab record for label 'label' into the code buffer.
 *
 ***************************************************************************/
PRIVATE void label_def (const int label, const int type )
{
  const int lab = findlabel(label);
  if (code[lab + LB_DEF] != C_NIL)
    geninternal_i(GEN_MULTIPLY_DEFINED_LABEL, label);
  else
    {
      const int ptr = mark_code(type, SL_SIZE);
      code[ptr + SL_ADDR] = address;
      code[lab + LB_DEF] = ptr;
    }
}
/*}}}*/
/*{{{  PRIVATE void section_label_def(label, type)*/
/***************************************************************************
 *
 * section_label_def inserts a setlab record for label 'label'
 *                   into the code buffer, and sets it to
 *                   be the start of the current code section.
 *
 **************************************************************************/
PRIVATE void section_label_def (const int label, const int type )
{
  label_def(label, type);
  section_label = lptr;
}
/*}}}*/
/*{{{  PRIVATE void label_ref(instruction, label)*/
/***************************************************************************
 *
 * label_ref inserts a uselab record for label 'label' into the code buffer.
 *
 **************************************************************************/
PRIVATE void label_ref (const int instruction, const int label )
{
  const int l = findlabel(label);
  const int forward_reference = (code[l + LB_DEF] == C_NIL);
  const int ptr = mark_code(C_USELAB, U_SIZE);
  int adj;

  if (t450a_workarounds || T9000_gamma_badldnlp(&tx_global))
    wa_t450_nop_t9gamma_badldnlp(instruction | I_PRIMARY);  

  if (forward_reference)
    adj = 0; /* assume null jump */
  else
    adj = 2; /* backwards jumps require nfix and op */

  code[ptr + U_DEF] = (forward_reference) ? l : code[l + LB_DEF];
  code[ptr + U_ADDR] = address;
  code[ptr + U_OP] = instruction;
  code[ptr + U_ADJ] = adj;

  address += adj;
}
/*}}}*/
/*{{{  PRIVATE void label_diff(instruction, label1, label2)*/
/***************************************************************************
 *
 * label_diff inserts a uselabdiff record for label 'label'
 * into the code buffer.
 *
 **************************************************************************/
PRIVATE void label_diff (const int instruction, const int label1, const int label2 )
{
  const int ptr = mark_code(C_USELABDIFF, UD_SIZE);
  int lab;

  if (t450a_workarounds || T9000_gamma_badldnlp(&tx_global))
    wa_t450_nop_t9gamma_badldnlp(instruction | I_PRIMARY );  

  code[ptr + UD_ADDR] = address;
  code[ptr + UD_ADJ] = 0;
  /*{{{  set UD_DEF1*/
  lab = findlabel(label1);
  if (code[lab + LB_DEF] == C_NIL)
    code[ptr + UD_DEF1] = lab;
  else
    code[ptr + UD_DEF1] = code[lab + LB_DEF];
  /*}}}*/
  /*{{{  set UD_DEF2*/
  lab = findlabel(label2);
  if (code[lab + LB_DEF] == C_NIL)
    code[ptr + UD_DEF2] = lab;
  else
    code[ptr + UD_DEF2] = code[lab + LB_DEF];
  /*}}}*/
  code[ptr + UD_OP] = instruction;
}
/*}}}*/
/*{{{  PRIVATE void table_start()*/
/***************************************************************************
 *
 *  table_start inserts a table record marking the start of a table, into
 *              the code buffer.
 *
 **************************************************************************/
PRIVATE void table_start ( void )
{
  int ptr = mark_code(C_TABLE, T_SIZE);
  code[ptr + T_SCALE] = case_scale;
}
/*}}}*/
/*{{{  PRIVATE void table_end()*/
/***************************************************************************
 *
 *  table_end inserts a table record marking the end of a table, into
 *            the code buffer.
 *
 **************************************************************************/
PRIVATE void table_end ( void )
{
  mark_code(C_TABLE, T_SIZE);
}
/*}}}*/
/*{{{  PRIVATE void scale_ref(instruction)*/
/***************************************************************************
 *
 *  scale_ref inserts a scale record for the scale of the forthcoming table
 *            into the code buffer.
 *
 **************************************************************************/
PRIVATE void scale_ref (const int instruction )
{
  if (t450a_workarounds || T9000_gamma_badldnlp(&tx_global))
    wa_t450_nop_t9gamma_badldnlp(instruction | I_PRIMARY);  
  case_scale = mark_code(C_SCALE, SC_SIZE);
  code[case_scale + SC_VALUE] = 0;
  code[case_scale + SC_OP] = instruction;
  address++;
}
/*}}}*/
/*}}}*/
/*{{{  debug information input*/
/*{{{  PRIVATE void add_debug_info (type, line)*/
/*****************************************************************************
 *
 *  add_debug_info inserts a debug record into the code buffer,
 *                 'type' indicates the type of debug record, line gives
 *                 the corresponding source file line number.
 *
 *****************************************************************************/
PRIVATE void add_debug_info (const int type , treenode *const line, const int data )
{
  const int ptr = mark_code(type, DB_SIZE);
  code[ptr + DB_ADDR] = address;
  code[ptr + DB_LINE] = (INT32)line;
  code[ptr + DB_DATA] = data;
}
/*}}}*/
/*}}}*/
/*}}}*/
/*{{{  code adjustment*/
/*{{{  variables for adjust*/
PRIVATE INT32 extra;               /* Extra bytes of code added this iteration */
PRIVATE INT32 ajptr;               /* Pointer to current node on adjust chain */
/*}}}*/
/*{{{  routines for adjust*/
/*{{{  PRIVATE int getaddr(ldef)*/
/***************************************************************************
 *
 * getaddr returns the address (offset from beginning of current code
 *         section) of the label ldef.
 *
 **************************************************************************/
PRIVATE INT32 getaddr (const INT32 ldef )
{
  if (ldef >= section_label)
    /*{{{  previous routine*/
    return(section_size + extra + code[ldef + LB_PRAD]);
    /*}}}*/
  else if (ldef >= lptr)
    /*{{{  unresolved forward reference*/
    {
      /* resolve forward reference */
      INT32 l = code[ldef + LB_DEF];                  /* point l to the setlab */
      if (l == C_NIL)
        geninternal_i(GEN_ADJUST_ERROR, code[ldef + LB_NAME]);  /* no setlab */
      if (code[ajptr] == C_USELAB)
        code[ajptr + U_DEF] = l;           /* Point the uselab to the setlab */
      else
        /*{{{  point the uselabdiff to the setlab*/
        {
          if (code[ajptr + UD_DEF1] == ldef)
            code[ajptr + UD_DEF1] = l;
          else
            code[ajptr + UD_DEF2] = l;
        }
        /*}}}*/

      return(code[l + SL_ADDR] + extra);
    }
    /*}}}*/
  else if (ldef > ajptr)
    /*{{{  resolved forward reference*/
    return(code[ldef + SL_ADDR] + extra);
    /*}}}*/
  else
    /*{{{  backward reference*/
    return(code[ldef + SL_ADDR]);
    /*}}}*/
}
/*}}}*/
/*}}}*/
/*{{{  PRIVATE void adjust()*/
/***************************************************************************
 *
 * adjust adjusts the label offsets in the current code section to their
 *        optimal size.
 *
 **************************************************************************/
PRIVATE void adjust ( void )
{
  int passno = 0;
  /* Perform code compaction */
  extra = -1;
  section_size = address;
  code[topl + U_NEXT] = C_NIL; /* Mark end of list */
  while (extra != 0)
    /*{{{  iteration of code adjustment*/
    {
      passno++;
      if (diagnostics)
        fprintf(outfile, "\nCode adjustment, pass %d...", passno);
      ajptr = cstart;
      extra = 0;
      while (ajptr != C_NIL)
        {
          switch(code[ajptr])
            {
              /*{{{  C_SETLAB C_STARTFILE C_ENDFILE C_LOCATE C_LIBPATCH C_ADDRESSFIX C_PROFADDRESSFIX*/
              case C_SETLAB:
              case C_STARTFILE:
              case C_ENDFILE:
              case C_LOCATE:
              case C_LIBPATCH:
              case C_ADDRESSFIX:
              case C_PROFADDRESSFIX:
                code[ajptr + SL_ADDR] += extra;
                break;
              /*}}}*/
              /*{{{  C_USELAB*/
              case C_USELAB:
               {
                 INT32 addr, diff;
                 int newadj;

                 addr = getaddr(code[ajptr + U_DEF]);
                 code[ajptr + U_ADDR] += extra;
                 diff = addr - (code[ajptr + U_ADDR] + code[ajptr + U_ADJ]);
                 if (isbranchop(code[ajptr + U_OP]) && (diff == 0))
                   newadj = 0; /* we optimise jump 0 out */
                 else
                   newadj = ilength(diff);
                 if (newadj > code[ajptr + U_ADJ])
                   {
                     extra += newadj - code[ajptr + U_ADJ];
                     code[ajptr + U_ADJ] = newadj;
                   }
               }
               break;
              /*}}}*/
              /*{{{  C_USELABDIFF*/
              case C_USELABDIFF:
                {
                  INT32 addr1, addr2, diff;
                  int newadj;

                  addr1 = getaddr(code[ajptr + UD_DEF1]);
                  addr2 = getaddr(code[ajptr + UD_DEF2]);
                  diff = addr1 - addr2;
                  code[ajptr + UD_ADDR] += extra;

                  if (isbranchop(code[ajptr + UD_OP]) && (diff == 0))
                    newadj = 0; /* we optimise jump 0 out */
                  else
                    newadj = ilength(diff);

                  if (newadj > code[ajptr + UD_ADJ])
                    {
                      extra += newadj - code[ajptr + UD_ADJ];
                      code[ajptr + UD_ADJ] = newadj;
                    }
                }
                break;
              /*}}}*/
              /*{{{  C_TABLE*/
              case C_TABLE:
                {
                  INT32 case_start = ajptr;
                  INT32 case_extra = -1;
                  int max_size = 0;
                  BOOL first = TRUE;

                  while (case_extra != 0)
                    {
                      ajptr = code[case_start + U_NEXT];
                      case_extra = 0;
                      while (code[ajptr] != C_TABLE)
                        /*{{{  visit each uselab record in adjust chain*/
                        {
                          INT32 addr, diff;
                          int newadj;

                          addr = getaddr(code[ajptr + U_DEF]);
                          if (first)
                            code[ajptr + U_ADDR] += extra + case_extra;
                          else
                            code[ajptr + U_ADDR] += case_extra;
                          diff = addr - (code[ajptr + U_ADDR] + code[ajptr + U_ADJ]);
                          newadj = ilength(diff);
                          newadj = (int)max_INT32(newadj, max_size);
                          if (newadj > code[ajptr + U_ADJ])
                            {
                              case_extra += newadj - code[ajptr + U_ADJ];
                              code[ajptr + U_ADJ] = newadj;
                            }
                          max_size = (int)max_INT32(max_size, code[ajptr + U_ADJ]);
                          ajptr = code[ajptr + U_NEXT];
                        }
                        /*}}}*/
                      extra += case_extra;
                      first = FALSE;
                    }

                  code[code[case_start + T_SCALE] + SC_VALUE] = max_size;
                }
                break;
              /*}}}*/
              /*{{{  C_SCALE*/
              case C_SCALE:
                /* Scale is set when we go round the associated table */
                break;
              /*}}}*/
              default:
                generr_i(GEN_ADJUST_TAG_ERROR, code[ajptr]);
            }
          ajptr = code[ajptr + U_NEXT];
        }
      /* Move address up the number of bytes we have inserted */
      address += extra;
      section_size += extra;
    }
    /*}}}*/
  if (diagnostics)
    fprintf(outfile, "\nAdjustment complete.\n");
}
/*}}}*/
/*}}}*/
/*{{{  PRIVATE void compress_code_buffer()*/
/*{{{  comment*/
/***************************************************************************
 *
 * compress_code_buffer compresses the adjusted code section to the front
 *                      of the code buffer (cstart onwards), and moves
 *                      cstart to point to the first free word.
 *
 **************************************************************************/
/*}}}*/
PRIVATE void compress_code_buffer ( void )
{
  const int topc = mark_code(C_SETLAB, SL_SIZE);
  int cmptr = cstart;                    /* Index into the uncompressed code */
  cptr = cstart;                       /* Where we write the compressed code */
  start = C_NIL;
  /*{{{  update labels above section_label*/
  /* Update labels above section_label, ie. entry points of other code
     sections, so that their LB_PRAD field contains the offset from the
     beginning of the current code section */
  {
    int l = section_label + LB_SIZE;
    while (l < max_code_size)
      {
        code[l + LB_PRAD] += section_size;
        l += LB_SIZE;
      }
  }
  /*}}}*/
  /*{{{  set up codew*/
  codew = 0; coden = 0; partword = 0;
  /*}}}*/
  /*{{{  patch uselab and uselabdiff operands*/
  while (cmptr < topc)
    /*{{{  patch operands*/
    {
      switch((int)code[cmptr])
        {
          /*{{{  C_SETLAB*/
          case C_SETLAB:
            cmptr += SL_SIZE;
            break;
          /*}}}*/
          /*{{{  C_USELAB*/
          case C_USELAB:
            {
              INT32 addr = code[code[cmptr + U_DEF] + SL_ADDR],
                    diff = addr - (code[cmptr + U_ADDR] + code[cmptr + U_ADJ]);
              code[cmptr + U_DEF] = diff;
              cmptr += U_SIZE;
            }
            break;
          /*}}}*/
          /*{{{  C_USELABDIFF*/
          case C_USELABDIFF:
            {
              INT32 addr1 = code[code[cmptr + UD_DEF1] + SL_ADDR],
                    addr2 = code[code[cmptr + UD_DEF2] + SL_ADDR];
              code[cmptr + UD_DEF1] = addr1 - addr2;
              cmptr += UD_SIZE;
            }
            break;
          /*}}}*/
          /*{{{  C_TABLE*/
          case C_TABLE:
            cmptr += T_SIZE;
            break;
          /*}}}*/
          /*{{{  C_SCALE*/
          case C_SCALE:
            cmptr += SC_SIZE;
            break;
          /*}}}*/
          /*{{{  C_STARTFILE C_ENDFILE C_LOCATE C_LIBPATCH C_ADDRESSFIX C_PROFADDRESSFIX*/
          case C_STARTFILE:
          case C_ENDFILE:
          case C_LOCATE:
          case C_LIBPATCH:
          case C_ADDRESSFIX:
          case C_PROFADDRESSFIX:              
            cmptr += DB_SIZE;
            break;
          /*}}}*/
          default:   /* code chunk */
            {
              const int len = (int)code[cmptr++];
              int words = len / 4;
              if ((len % 4) != 0) /* allow for partword */
                words++;
              cmptr += words;
            }
            break;
        }
    }
    /*}}}*/
  /*}}}*/
  /*{{{  compress code buffer*/
  cmptr = cstart;
  while (cmptr < topc)
    switch((int)code[cmptr])
      {
        /*{{{  C_SETLAB*/
        case C_SETLAB:
          cmptr += SL_SIZE;
          break;
        /*}}}*/
        /*{{{  C_USELAB*/
        case C_USELAB:
          {
            const INT32 op   = code[cmptr + U_OP];
            const INT32 adj  = code[cmptr + U_ADJ];
            const INT32 diff = code[cmptr + U_DEF];
            /*{{{  a bit of peephole optimisation to remove j 0 and cj 0*/
            if (!isbranchop(op) || (diff != 0) || (adj != 0))
              put_pfix(op, diff, adj);
            /*}}}*/
            cmptr += U_SIZE;
          }
          break;
        /*}}}*/
        /*{{{  C_USELABDIFF*/
        case C_USELABDIFF:
          {
            INT32 diff = code[cmptr + UD_DEF1];
            code_primary(code[cmptr + UD_OP], diff);
            cmptr += UD_SIZE;
          }
          break;
        /*}}}*/
        /*{{{  C_SCALE*/
        case C_SCALE:
          code_primary(code[cmptr + SC_OP], code[cmptr + SC_VALUE]);
          cmptr += SC_SIZE;
          break;
        /*}}}*/
        /*{{{  C_TABLE*/
        case C_TABLE:
          cmptr += T_SIZE;
          break;
        /*}}}*/
        /*{{{  C_LOCATE*/
        case C_LOCATE:
          if (debugoutput)
            debug_gencodemark((treenode *)code[cmptr + DB_LINE],
             (bytes_output + section_size) - code[cmptr + DB_ADDR]);
            /* write_locate_item(rtltag(code[cmptr]), code[cmptr + DB_LINE],
                              code[cmptr + DB_ADDR]); */
          cmptr += DB_SIZE;
          break;
        /*}}}*/
        /*{{{  C_LIBPATCH*/
        case C_LIBPATCH:
          if (debugoutput)
            debug_genlibrpatch((treenode *)code[cmptr + DB_LINE],
             (bytes_output + section_size) - code[cmptr + DB_ADDR]);
          cmptr += DB_SIZE;
          break;
        /*}}}*/
        /*{{{  case C_PROFADDRESSFIX*/
        case C_PROFADDRESSFIX:
          profile_count_table_label = (bytes_output + section_size) - code[cmptr + DB_ADDR];
          cmptr += DB_SIZE;
          break;
        /*}}}*/
        /*{{{  C_ADDRESSFIX*/
        case C_ADDRESSFIX:
          if (debugoutput || sampling_profiling )
            {
              treenode *tptr = (treenode *)code[cmptr + DB_LINE];
              const int label = (int)code[cmptr + DB_DATA];
              BIT32 offset;
              DEBUG_MSG(("addressfix: label: %d, ", label));
              if (label == (-1)) /* normal addressfix */
                offset = (bytes_output + section_size) - code[cmptr + DB_ADDR];
              else /* bug 1158 - extra stuff */
                {
                #if 0
                  DEBUG_MSG(("findlabel: %d, ",                  findlabel(label)));
                  DEBUG_MSG(("code(DEF): %d, ",             code[findlabel(label)+LB_DEF]));
                  DEBUG_MSG(("code(PRAD): %d\n",            code[findlabel(label)+LB_PRAD]));
                  DEBUG_MSG(("code(SL_ADDR): ld, ",         code[findlabel(label)+SL_ADDR]));
                  DEBUG_MSG(("addr(SL_ADDR): %d, ", getaddr(code[findlabel(label)+SL_ADDR])));
                  DEBUG_MSG(("addr(PRAD): %d\n",    getaddr(code[findlabel(label)+LB_PRAD])));
                  DEBUG_MSG(("section_size: %d, bytes_output: %d\n", section_size, bytes_output));
                  DEBUG_MSG(("Correct: %d, ", (bytes_output + section_size) - code[findlabel(label)+LB_PRAD]));
                #endif
                  /* This is a horrible piece of black magic:
                     What we want is the address (offset from the end of code)
                     of a particular label.
                     After much experimentation, this expression gives the
                     correct result. Blame Martin Day for designing this
                     code cruncher in this way in the first place.
                     CON - 14/2/91
                  */
                  offset = (bytes_output + section_size) -
                           code[findlabel(label)+LB_PRAD];
                }
              DEBUG_MSG(("offset: %ld\n", offset));
              if (debugoutput)
                debug_genaddrfix(tptr, offset);
              if ((sampling_profiling || cgraph_profiling)  &&
                  (TagOf(tptr) == N_PROCDEF   ||
                   TagOf(tptr) == N_SFUNCDEF  ||
                   TagOf(tptr) == N_LFUNCDEF     )               )
                set_proftab_routine_address(profile_table, tptr, offset);
            }
          cmptr += DB_SIZE;
          break;
        /*}}}*/
        default:   /* code chunk */
          /*{{{  disassemble code chunk*/
          {
            const int len = (int)code[cmptr++];
            int i;
            int bytestogo = 0;
            BIT32 putword = 0; /* initialised to shut up gcc's optimiser */
            for (i = 0; i < len; i++)
              {
                if (bytestogo == 0)
                  {
                    putword = code[cmptr++];
                    bytestogo = 4;
                  }
                add_code((BYTE)(putword & 0xff));
                bytestogo--;
                putword >>= 8;
              }
          }
          break;
          /*}}}*/
      }
  /*}}}*/
  /*{{{  tidy up codew*/
  if (partword != 0)
    {
      code[cptr++] = codew;
      partword = 0;
      codew = 0;
    }
  /*}}}*/
  if (start != C_NIL)
    {
      code[cptr++] = start;       /* Append pointer to start of code section */
      code[start] = section_size;              /* Insert code section length */
    }
  cstart = cptr;
}
/*}}}*/
/*{{{  PRIVATE void diag_write_spaces*/
PRIVATE void diag_write_spaces(void)
{
  fputc('\t', outfile);
}
/*}}}*/
/*{{{  PRIVATE void diag_write_nl*/
PRIVATE void diag_write_nl(void)
{
  fputc('\n', outfile);
  diag_write_spaces();
}
/*}}}*/
/*{{{  PRIVATE void diag_write_comment*/
PRIVATE void diag_write_comment(void)
{
  diag_write_spaces();
  fputs("-- ", outfile);
}
/*}}}*/
/*{{{  PRIVATE void diag_write_nl_comment*/
PRIVATE void diag_write_nl_comment(void)
{
  fputc('\n', outfile);
  diag_write_comment();
}
/*}}}*/
/*{{{  PRIVATE void diag_write_nl_string*/
PRIVATE void diag_write_nl_string(const char *str)
{
  diag_write_nl();
  fputs(str, outfile);
}
/*}}}*/
/*{{{  PUBLIC void gencommentn(c, ...)*/
/***************************************************************************
 *
 *  gencommentn writes the printf format string 'c' (with parameters p1, p2,
 *             pn) as a comment to the debugging file.
 *
 **************************************************************************/
/* Write a comment */
PUBLIC void gencomment0 ( const char *const c )
{
  if (assembly_output && ((alloc_strategy & ALLOC_NOCOMMENTS) == 0))
    {
      diag_write_comment();
      fputs(c, outfile);
    }
}

/* Write a comment */
PUBLIC void gencomment1 (const char *const c, const BIT32 p1 )
{
  if (assembly_output && ((alloc_strategy & ALLOC_NOCOMMENTS) == 0))
    {
      diag_write_comment();
      fprintf (outfile, c, p1);
    }
}
PUBLIC void gencomments (const char *const c, const char * p1 )
{
  if (assembly_output && ((alloc_strategy & ALLOC_NOCOMMENTS) == 0))
    {
      diag_write_comment();
      fprintf (outfile, c, p1);
    }
}
/* Called after a function instance which leaves results on tstack */
PUBLIC void gensettsdepth (const int regresults)
{
  if (assembly_output && ((alloc_strategy & ALLOC_NOCOMMENTS) == 0))
  {
    fprintf (outfile, "\n\t-- stackdepth =  %d", regresults);
  }
}
/* Called after a function instance which leaves a result on fpstack */
PUBLIC void notefpresult (const char *type)
{
  if (assembly_output && ((alloc_strategy & ALLOC_NOCOMMENTS) == 0))
    fprintf (outfile, "\n\t-- fpstack =  %s", type);
}
#if 0  /* gencomment2 and gencomment3 are never used */
/* Write a comment */
PUBLIC void gencomment2 (const char *const c, const BIT32 p1, const BIT32 p2 )
{
  if (assembly_output && ((alloc_strategy & ALLOC_NOCOMMENTS) == 0))
    {
      diag_write_comment();
      fprintf (outfile, c, p1, p2);
    }
}

/* Write a comment */
PUBLIC void gencomment3 (const char *const c, const BIT32 p1, const BIT32 p2, const BIT32 p3 )
{
  if (assembly_output && ((alloc_strategy & ALLOC_NOCOMMENTS) == 0))
    {
      diag_write_comment();
      fprintf (outfile, c, p1, p2, p3);
    }
}
#endif
/*}}}*/
/*{{{  code disassembly*/
/*{{{  PRIVATE int look_at_instruction(length, operand)*/
/***************************************************************************
 *
 *  look_at_instruction disassembles the next instruction in the code
 *                      buffer (represented by getptr, bytesleft, getword)
 *                      into an instruction value and an operand value.
 *                      The instruction bytes are placed in buf.
 *                      The instruction len is returned.
 *
 **************************************************************************/
PRIVATE int look_at_instruction ( int *const instruction, INT32 *const operand,
        BYTE buf[], const int maxlen)
{
  int len = 0;
  *operand = 0;
  while (len < maxlen)
    /*{{{  read instruction byte*/
    {
      BYTE b = get_code();
      buf[len++] = b;
      *operand |= (int)(b & 0x0f);
      *instruction = b & 0xf0;
      if (*instruction == I_PFIX)
        *operand <<= 4;
      else if (*instruction == I_NFIX)
        *operand = (~(*operand)) << 4;
      else
        return len;
    };
    /*}}}*/
  return len;
}
/*}}}*/
/*{{{  PRIVATE int print_instruction(address)*/
/***************************************************************************
 *
 *  print_instruction disassembles the next instrution
 *                    and writes it to outfile.
 *                    Returns the length of the instruction disassembled.
 *
 **************************************************************************/
PRIVATE int print_instruction (const INT32 addr, const int bytestogo )
{
  int i_value, i_len;
  int i;
  INT32 operand;
  BYTE buf[MAX_BYTES_IN_INSTRUCTION];

  fprintf(outfile, "%8ld: ", addr);
  i_len = look_at_instruction (&i_value, &operand, buf,
                               (int)min_INT32(MAX_BYTES_IN_INSTRUCTION, bytestogo));

  for (i = 0; i < i_len; i++)
    fprintf (outfile, "%.2X ", (int)buf[i]);
  for (; i < MAX_BYTES_IN_INSTRUCTION; i++)
    fprintf (outfile, "   ");

  if (i_value == I_OPR)
    fprintf (outfile, "%-8s\n", secondaryname((int)operand));
  else
    fprintf (outfile, "%-8s%ld\n", primaryname(i_value), operand);

  return (i_len);
}
/*}}}*/
/*}}}*/
/*}}}*/
/*{{{  PUBLIC genstartconstblock(c)*/
PUBLIC void genstartconstblock (const char *c)
{
  gencomment0 (c);
}
/*}}}*/
/*{{{  PUBLIC genendconstblock*/
PUBLIC void genendconstblock ()
{
  gencomment0 ("End of constants");
}
/*}}}*/


/*{{{  PUBLIC void patch_jump(INT32 source_addr, INT32 target_addr)*/
#if 0
/* This is only called if configuring is TRUE */
PUBLIC void patch_jump(const INT32 source_addr, const INT32 target_addr)
{
  INT32 operand;
  int len, ptr;
  int codeptr = 0;
  BYTE buf[MAX_BYTES_IN_INSTRUCTION];

  DEBUG_MSG(("Patching a jump from %ld to %ld\n", source_addr, target_addr));

  operand = target_addr - source_addr;
  len = 1;
  while ((ilength(operand - len) > len) && (len < MAX_BYTES_IN_INSTRUCTION))
    len++;

  operand -= len;

  if (ilength(operand) < len)  /* allow for the smaller operand being */
    {                          /* a smaller length */
      /* Note - This can never happen for backward jumps */
      buf[0] = (I_PFIX | 0);
      ptr = 1;
    }
  else
    ptr = 0;
  /* len now holds the length of the patch to be inserted */
  if (assembly_output)
    fprintf(outfile, "\n-- Inserting %d byte patch at offset %ld for jump %ld to %ld",
            len, source_addr, operand, target_addr);

  len = assemble_instruction(&(buf[ptr]), I_J, operand);
  len += ptr;

  /*{{{  find the start of the correct code section*/
  {
    int section_start = cstart;
    INT32 code_addr = 0;
    while (section_start != 0)
      /*{{{  find a code section*/
      {
        INT32 code_length;
        section_start = (int)code[section_start - 1];
        code_length = code[section_start];
        if (code_addr == source_addr)
          {
            codeptr = section_start + 1;
            section_start = 0;  /* force a bomb out of loop */
            if (code_length < len)
              generr_i(GEN_CONF_PATCH_TOO_SMALL, len);
          }
        else
          code_addr += code_length;
      }
      /*}}}*/
    if (codeptr == 0)
      generr_i(GEN_CONF_CANT_FIND_PATCH, source_addr);
  }
  /*}}}*/
  /*{{{  patch the jump onto the code section*/
  {
    INT32 mask = 0xff;
    int shift = 0;
    while (--len >= 0)
      {
        code[codeptr] = (code[codeptr] & ~mask) | (int)(buf[len] << shift);
        shift += 8; mask <<= 8;
        if (shift == 32)
          {
            shift = 0;
            mask = 0xff;
            codeptr++;
          }
      }
  }
  /*}}}*/
}
#endif
/*}}}*/

/*{{{  PRIVATE void patch_lib_offsets ()*/
/***************************************************************************
 *
 *  patch_lib_offsets writes the offset of each library call stub into its
 *                    corresponding name node.
 *
 **************************************************************************/
PRIVATE void patch_lib_offsets (void)
{
  treenode *libcall;
  for(libcall = libentrychain; libcall != NULL; libcall = NLEntryNextOf(libcall))
    {
      const int labelpos = findlabel(NPLabelOf(libcall));
      SetNLEntryOffset(libcall, (int)code[labelpos + LB_PRAD]);
    }

  if (configuring)
    code_pseudo_link_fn(libentrychain);
}
/*}}}*/

/*{{{  PRIVATE void display_code_size*/
PRIVATE void display_code_size(const char *const prefix_str)
{
  if (!ims_asm_output)
    fprintf(outfile, "%sModule code size is %ld bytes\n", prefix_str, total_code_size);
}
/*}}}*/
/*{{{  PRIVATE void output_code_size()*/
/***************************************************************************
 *
 *  output_code_size writes the total code size to the object file
 *
 **************************************************************************/
PRIVATE void output_code_size ( void )
{
  /*INT32 nested_sc_size = sc_size();*/

  if (!configuring && information)
    {
      display_code_size("");
    #if 0
      if (nested_sc_size != 0)
        fprintf(outfile, ", nested SC code size is %ld bytes", nested_sc_size);
      fprintf(outfile, "\n");
    #endif
    }
  /*write_total_code(total_code_size, nested_sc_size);*/
}
/*}}}*/
/*{{{  PRIVATE void calculate_code_size*/
PRIVATE void calculate_code_size(void)
{
  int section_start = cstart;

  total_code_size = 0;
  while (section_start != 0)
    {
      section_start = (int)code[section_start - 1];
      total_code_size += code[section_start];
    }

  /*printf("calculate_code_size: size is %d\n", total_code_size);*/
}
/*}}}*/
/*{{{  PRIVATE void output_code()*/
/**************************************************************************
 *
 * output_code writes the entire code buffer to the object file
 *
 **************************************************************************/
PRIVATE void output_code ( void )
{
  int section_start = cstart;
  INT32 disassembly_address = 0;

  if (!disassemble) write_code_block_start((size_t)total_code_size);

  while (section_start != 0)
    /*{{{  output a code section*/
    {
      INT32 code_length;
      section_start = (int)code[section_start - 1];
      code_length = code[section_start];
      /*{{{  set up for get_code*/
      getptr = section_start + 1; bytesleft = 0;
      /*}}}*/
      if (!disassemble)
        /*{{{  write code section to object file*/
        {
          BIT32 l = code_length;
          DEBUG_MSG(("output_code: writing a section of %ld bytes\n", l));
          while (l > 0)
            {
              BYTE ibuf[IBUF_SIZE];
              size_t len = (size_t)min_INT32(IBUF_SIZE, l);
              int i;
              for (i = 0; i < len; i++)
                ibuf[i] = get_code();  /* this allows for endian-ness */
              write_code_block(len, ibuf);
              l -= len;
            }
        }
        /*}}}*/
      else
        /*{{{  disassemble code section to outfile*/
        {
          int i = 0;
          while (i < code_length)
            i += print_instruction(disassembly_address + i, (int)(code_length - i));
          disassembly_address += code_length;
        }
        /*}}}*/
    }
    /*}}}*/
}
/*}}}*/

/*{{{  PRIVATE BOOL name_is_visible*/
PRIVATE BOOL name_is_visible(const wordnode *const name, const seginfo_t *rest)
/* Returns TRUE if the name is not descoped by a later externally visible name */
{
  for (; rest != NULL; rest = rest->seg_next)
    if (NLexLevelOf(rest->seg_name) == 0 && 
        name == NNameOf(rest->seg_name)      ) 
      return FALSE;
  return TRUE;
}
/*}}}*/
/*{{{  PRIVATE INT32 offset_from_label*/
PRIVATE INT32 offset_from_label(const int label)
{
  if (label == NO_LABEL)
    return 0;
  return code[findlabel(label) + LB_PRAD];
}
/*}}}*/
/*{{{  PRIVATE void output_entry_points*/
/*{{{  comment*/
/***************************************************************************
 *
 *  output_entry_points writes all the declared entry points into the
 *                      object file. Must be called after output_code, as
 *                      the section labels must have been adjusted for
 *                      word alignment.
 *
 **************************************************************************/
/*}}}*/
PRIVATE void output_entry_points (void)
{
  seginfo_t *seg;

  /* first reverse the list back into the correct order */
  {
    seginfo_t *p = seginfo, *reverser = NULL, *temp;
    while (p != NULL)
      {
        temp        = p->seg_next;
        p->seg_next = reverser;
        reverser    = p;
        p           = temp;
      }
    seginfo = reverser;
  }

  for (seg = seginfo; seg != NULL; seg = seg->seg_next)
    {
      treenode *const n  = seg->seg_name;
      /*const INT32 offset = code[seg->seg_entry + LB_PRAD];*/
      /*const INT32 offset = code[findlabel(NPLabelOf(n)) + LB_PRAD];*/
      const INT32 offset = offset_from_label(NPLabelOf(n));
      /*assertion(section_label == findlabel(NPLabelOf(n)));*/
      if (NLexLevelOf(n) == 0                     &&
          name_is_visible(NNameOf(n), seg->seg_next)    )
      {
        if (configuring)
          code_save_entrydata_fn(n, offset, total_code_size);
        else if (information)
          fprintf(outfile, "%s %s requires %ld workspace and %ld vectorspace words\n",
                    (TagOf(n) == N_PROCDEF) ? "PROC" : "FUNCTION",
                    WNameOf(NNameOf(n)), NPDatasizeOf(n), NPVSUsageOf(n));
        if (disassemble)
        {
          diag_write_comment();
          fprintf(outfile, "Entry point %s, offset %ld\n",
                  WNameOf(NNameOf(n)), offset);
        }
        else
          write_entry_desc(n, offset);
      }
      else
      {
        if (disassemble)
        {
          diag_write_comment();
          fprintf(outfile, "LOcal Entry point %s, offset %ld\n",
                  WNameOf(NNameOf(n)), offset);
        }
        else
          write_local_entry(n, offset);
      }
    }
}
/*}}}*/
/*{{{  PRIVATE void output_mapfile_sectiontable*/
PRIVATE void output_mapfile_sectiontable(void)
{
  seginfo_t *this = mapinfo_list;
  while (this != NULL)
    {
      seginfo_t *const next = this->seg_next;
      treenode  *const nptr = this->seg_name;
      int label;
      const char *type;
      switch(TagOf(nptr))
        {
          /*{{{  data item*/
          case N_VALABBR: case N_VALRETYPE:
            {
              const treenode *const data = consttableof(NDeclOf(nptr));
              if (WLengthOf(CTValOf(data)) == 0)
                label = NO_LABEL;
              else
                label = CTLabelOf(data);
              type = (TagOf(data) == S_STRING) ? "string      " : "table       ";
              /*type = "data        ";*/
            }
            break;
          /*}}}*/
          /*{{{  code item*/
          default:
            label = NPLabelOf(nptr);
            type = "code        ";
            break;
          /*}}}*/
        }
      map_add_entry(mapfile_handle, NULL, WNameOf(NNameOf(nptr)), type,
                    offset_from_label(label), NULL);

      freevec(this, sizeof(seginfo_t));
      this = next;
    }
  map_set_size(mapfile_handle, NULL, total_code_size, 0);
}
/*}}}*/

/*{{{  PUBLIC void compress_code()*/
/***************************************************************************
 *
 * compress_code compresses the current code section down to the front
 *               of the code buffer, and prepares the code buffer for the
 *               input of a new code section.
 *
 **************************************************************************/
PUBLIC void compress_code ( void )
{
  adjust();
  compress_code_buffer();
  /*{{{  cut back labels*/
  /* The label at section_label is the entry point which we must keep.
     Save the entry point offset in this label.
     Delete all subsequent labels. */
  code[section_label + LB_DEF] = C_NIL;
  code[section_label + LB_PRAD] = 0;
  lptr = section_label;
  /*}}}*/
  /*{{{  cut back code*/
  bytes_output += section_size;
  cptr = cstart;
  start = C_NIL;
  topl = C_NIL;
  codew = 0;
  coden = 0;
  partword = 0;
  address = 0; /* New code section starts at offset 0 */
  /*}}}*/
}
/*}}}*/
/*{{{  instruction generation*/
/*{{{  PRIVATE void clean_fpu_flag*/
PRIVATE void clean_fpu_flag(const int instruction)
{
  if (inside_asm) return;

  if (has_fpu_core && (fpu_error_status == ERROR_STATUS_START) &&
      (errormode & ERRORMODE_TIMESLICECHECK))
    {
      /* This can be called with either FPTESTERR or FPUCLRERR,
         depending on whether we're doing localised error handling.
      */
      if (instruction & I_FPU_ENTRY_BIT) /* bug 1206 13/3/91 */
        genfpuentry(instruction);
      else
        gensecondary(instruction);
      fpu_error_status = ERROR_STATUS_CLEAN;
    }
}
/*}}}*/
/*{{{  PUBLIC void genstartblock()*/
/*****************************************************************************
*
*  genstartblock is used in STOP mode to remember to clear the error flag
*                after a process could have been timesliced by a jump
*                to the start of the block
*
******************************************************************************/
PUBLIC void genstartblock ( void )
{
  /* We cannot localise handling of the normal error flag, cos
     the TESTERR instruction clobbers the stack;
     For the FPU we can use FPUCLRERR which leaves all stacks alone */

  if (inside_asm) return;

  int_error_status = ERROR_STATUS_CLEAN;
  if (errormode & ERRORMODE_TIMESLICECHECK)
    gensecondary(I_TESTERR);

  fpu_error_status = ERROR_STATUS_START;
#if !LOCALISE_ERROR_HANDLING
  clean_fpu_flag(I_FPTESTERR);  /* FPTESTERR is faster than FPUCLRERR */
#endif
}
/*}}}*/
/*{{{  PUBLIC void coder_return_from_call*/
PUBLIC void coder_return_from_call(const BOOL clean_fpu)
{
  /* this is called after a 'call' instruction to reset any information
     about the error flags.
     Our current model assumes that the integer error flag will always
     be OK.
     If 'clean_fpu' is FALSE, we have to make sure that we clean the
     FPU error flag before we next use it.
  */
  if (!clean_fpu)
    fpu_error_status = ERROR_STATUS_START;
}
/*}}}*/
/*{{{  PRIVATE void dirty_error_flag*/
/*****************************************************************************
*
*  dirty_error_flag is used incase we've never got around to clearing
*                   the error flag in STOP mode.
*
******************************************************************************/
PRIVATE void dirty_error_flag (const BOOL fpu_not_int )
{
  /* We cannot localise handling of the normal error flag, cos
     the TESTERR instruction clobbers the stack;
     For the FPU we can use FPUCLRERR which leaves all stacks alone */

  if (inside_asm) return;

  if (fpu_not_int)
    {
#if 0  /* LOCALISE_ERROR_HANDLING */
      /* This doesn't work, cos FPUCLRERR DOES hit the CPU stack! */
      clean_fpu_flag(I_FPUCLRERR);
#endif
      fpu_error_status = ERROR_STATUS_DIRTY;
    }
  else
    int_error_status = ERROR_STATUS_DIRTY;
}
/*}}}*/
/*{{{  PUBLIC void mark_flag_clean*/
PUBLIC void mark_flag_clean(const BOOL fpu_not_int)
{
  if (fpu_not_int)
    fpu_error_status = ERROR_STATUS_CLEAN;
  else
    int_error_status = ERROR_STATUS_CLEAN;
}
/*}}}*/
/*{{{  PUBLIC void markdeadcode*/
PUBLIC void markdeadcode()
{
  dead = !inside_asm;
}
/*}}}*/
/*{{{  PUBLIC void genprimary (instruction, operand)*/
/*{{{  comment*/
/***************************************************************************
 *
 *  genprimary adds code byte representing primary (instruction, operand)
 *             to the code buffer
 *
 **************************************************************************/
/*}}}*/
PUBLIC void genprimary (const int instruction, const INT32 operand )
{
  if (!dead)
    {
      if (instruction == I_ADC) dirty_error_flag(FALSE);
      code_primary(instruction, operand);
      if (assembly_output)
        {
          if ((operand == 0) && ((instruction == I_LDNLP) ||
                                      (instruction == I_AJW) ||
                                      (instruction == I_ADC)))
            /* Null instruction */
            return;
          else
            {
              diag_write_nl();
              fprintf(outfile, "%-8s", primaryname(instruction));
              if ((alloc_strategy & ALLOC_NOOPERANDS) == 0)
                {
                  fprintf(outfile, "%ld", operand);
                  /* display large(ish) numbers in hex too */
                  if (operand > 4000 || operand < -4000)
                    gencomment1("#%lX", operand);
                }
            }
        }
    }
}
/*}}}*/
/*{{{  PUBLIC void genbranch (instruction, label)*/
/***************************************************************************
 *
 *  genbranch adds a record (representing a primary 'instruction' with
 *            operand the offset between the next instruction and the
 *            label 'label') to the code buffer.
 *
 **************************************************************************/
PUBLIC void genbranch (const int instruction, const int label )
{
  if (!dead)
    {
      if (instruction == I_ADC)
        /* Only happens inside ASM - eg ADC :label, or ADC routine.name */
        dirty_error_flag(FALSE);
      else
        checkerror(); /* INSdi02485 */ /* check before any transfer of control */
      label_ref(instruction, label);
      /*{{{  debugging*/
      if (assembly_output)
        {
          diag_write_nl();
          fprintf(outfile, "%-8sL%d", primaryname(instruction),
                  ((alloc_strategy & ALLOC_NOOPERANDS) == 0) ? label : 0);
        }
      /*}}}*/
      if (instruction == I_J) markdeadcode();
    }
}
/*}}}*/
/*{{{  PUBLIC void genjump*/
PUBLIC void genjump(const int label)
/* This is used to insert a jump. If code style flag is set,
   creates a non-timeslicable jump using ldc 0; cj
*/
{
  checkerror(); /* INSdi02485 */ /* check before any transfer of control */
  if ((code_style_flags & CODE_STYLE_CJ_NOT_J) != 0)
    {
      genprimary(I_LDC, 0);
      genbranch(I_CJ, label);
    }
  else
    genbranch(I_J, label);
}
/*}}}*/
/*{{{  PUBLIC void genlabeldiff (instruction, label1, label2)*/
/***************************************************************************
 *
 *  genlabeldiff adds a record (representing a primary 'instruction' with
 *               operand the offset between 'label1' and 'label2')
 *               to the code buffer.
 *
 **************************************************************************/
PUBLIC void genlabeldiff (const int instruction, const int label1, const int label2 )
{
  if (!dead)
    {
      if (instruction == I_ADC) dirty_error_flag(FALSE);
      label_diff(instruction, label1, label2);
      if (assembly_output)
        {
          const BOOL labels = ((alloc_strategy & ALLOC_NOOPERANDS) == 0);
          diag_write_nl();
          fprintf(outfile, "%-8sL%d - L%d",
                           primaryname(instruction),
                           labels ? label1 : 0, labels ? label2 : 0);
        }
    }
}
/*}}}*/
/*{{{  PUBLIC void genloadlabptr (thatlab, thislab,text)*/
PUBLIC void genloadlabptr (const int thatlab, const int thislab, const char * text)
{
  /* text defines what is at thatlab */
  genlabeldiff (I_LDC, thatlab, thislab);
  gensecondary (I_LDPI);
  if (text != NULL)
    gencomment0 (text);
  setlab (thislab);
}
/*}}}*/
/*{{{  PUBLIC void gensecondary (instruction)*/
/***************************************************************************
 *
 *  gensecondary adds the secondary 'instruction' to the code buffer.
 *
 **************************************************************************/
PUBLIC void gensecondary (int instruction )
{
  if (!dead)
    {
      /*{{{  check whether instruction may set error*/
      switch (instruction)
        {
          default:
            break;
          case I_CSUB0: case I_CCNT1: case I_SETERR:
          case I_ADD:   case I_SUB:   case I_MUL:   case I_DIV:  case I_REM:
          case I_CWORD: case I_CSNGL:
          case I_LADD:  case I_LSUB:  case I_LDIV:   case I_FMUL:
          case I_CFLERR:
          case I_FPCHKERR:
          case I_LDCNT: /* variableio 11/3/91 */
          case I_CB: case I_CBU: case I_CIR: case I_CIRU: case I_CS: case I_CSU: /* Added 5/9/91 */
          case I_CAUSEERROR: /* bug TS/1547 07/04/92 */
            dirty_error_flag(FALSE);
            break;

          case I_FPADD: case I_FPSUB: case I_FPMUL: case I_FPDIV: case I_FPREMFIRST:
          case I_FPLDNLADDSN:         case I_FPLDNLADDDB:
          case I_FPLDNLMULSN:         case I_FPLDNLMULDB:
          case I_FPGT:  case I_FPEQ:
          case I_FPINT: case I_FPRTOI32:

          case I_FPSQRT:     case I_FPREM: case I_FPRANGE:
          case I_FPMULBY2:   case I_FPDIVBY2:
          case I_FPEXPINC32: case I_FPEXPDEC32:
          case I_FPABS:
        /*case I_FPSETERR:*/ /* bug TS/1547 12/05/92 */
          case I_FPR32TOR64: case I_FPR64TOR32:
          case I_FPCHKI32:   case I_FPCHKI64:
          case I_FPADDDBSN:
            dirty_error_flag(TRUE);
            break;
        }
      /*}}}*/
      code_secondary(instruction);
      /*{{{  print out the instruction*/
      if (assembly_output)
        {
          diag_write_nl();
          if (instruction & I_STEP_BIT)
            {
               fputs("step ", outfile);
               instruction &= ~(I_STEP_BIT);
            }
          fprintf(outfile, "%-8s", secondaryname(instruction));
        }
      /*}}}*/
      /*{{{  check if we have been timesliced, also for loading FPU values*/
      switch (instruction)
        {
          default:
            break;
          case I_OUTWORD: case I_OUTBYTE: case I_OUT:
          case I_IN: case I_TIN:
          case I_ALTWT: case I_TALTWT:
          case I_VIN: case I_VOUT: /* variableio 3/11/91 */
            genstartblock();
            break;
          /* We rely on the fact that after an FPU load, there is always
             a spare register, so an FPUCLRERR is OK.
             Note that regsfor gives 1 for CONSTEXP, so the loads
             of zeroes are OK too. */
          case I_FPLDNLSN:    case I_FPLDNLDB:
          case I_FPLDNLSNI:   case I_FPLDNLDBI:
          case I_FPLDNLADDSN: case I_FPLDNLMULSN:
          case I_FPLDNLADDDB: case I_FPLDNLMULDB:
          case I_FPLDZEROSN:  case I_FPLDZERODB:
          case I_FPI32TOR32:  case I_FPB32TOR64:
          case I_FPI32TOR64:  /* bug TS/2006 15/12/92 */
            clean_fpu_flag(I_FPUCLRERR);
            break;
        #if 0
          /* NO - we can't blatantly clobber stopp, because we use it in
             PRIPAR and in RESCHEDULE, so we do this in the STOP code */
          case I_STOPP:  /* added for bug 1057 27/11/90 */
            markdeadcode();
            break;
        #endif
          case I_SETERR: /* added for bug 1057 27/11/90 */
            if (errormode & ERRORMODE_HALT) markdeadcode();
            break;
        }
      /*}}}*/
    }
}
/*}}}*/
/*{{{  PUBLIC void genboolinvert ()*/
/*****************************************************************************
 *
 *  genboolinvert generates code to invert a known boolean
 *
 *****************************************************************************/
PUBLIC void genboolinvert (void)
{
  genprimary(I_EQC, ZERO32);
}
/*}}}*/
/*{{{  PUBLIC void gencompare (relop,flags)*/
PUBLIC void gencompare (const int relop, const BOOL flags)
/* generates code to compare Breg, Areg
   if flags is FALSE result could be left in condition flags, else Areg */
{
  const char * opname;
  switch (relop)
  {
    case S_GR: opname = "gt"; break;
    case S_LE: opname = "le"; break;
    case S_NE: opname = "ne"; break;
    default:
    case S_EQ: opname = "eq"; break;
  }
  switch (relop)
  {
    case S_GR: gensecondary (I_GT); break;
    case S_LE: gensecondary (I_GT); genboolinvert (); break;
    case S_EQ: gensecondary (I_DIFF); genprimary (I_EQC, 0); break;
    case S_NE: gensecondary (I_DIFF); genprimary (I_EQC, 0);
               genboolinvert (); break;
  }
}
/*}}}*/
/*{{{  PUBLIC void checkerror_controlled*/
PUBLIC void checkerror_controlled(const int mode)
/* Checks the error flag, controlled by a supplied 'errormode' flag */
{
  if ((mode & ERRORMODE_NEED_FPCHKERR) &&
      (fpu_error_status == ERROR_STATUS_DIRTY))
    {
      /* This will automagically dirty the other flag */
      gensecondary(I_FPCHKERR);
      fpu_error_status = ERROR_STATUS_CLEAN;
    }
  if ((mode & ERRORMODE_NEED_STOPERR) &&
      (int_error_status == ERROR_STATUS_DIRTY))
    {
      gensecondary(I_STOPERR);
      int_error_status = ERROR_STATUS_CLEAN;
    }
}
/*}}}*/
/*{{{  PUBLIC void checkerror*/
PUBLIC void checkerror ( void )
/* Checks the error flag, controlled by the normal 'errormode' flag */
{
  if (!inside_asm && NEED_ERRORS)
    checkerror_controlled(errormode);
  else if (!inside_assertion)
    {
      if (fpu_error_status == ERROR_STATUS_DIRTY) /* INSdi02121 */
        fpu_error_status = ERROR_STATUS_CLEAN;
      if (int_error_status == ERROR_STATUS_DIRTY) /* INSdi02121 */
        int_error_status = ERROR_STATUS_CLEAN;
    }
}
/*}}}*/
/*{{{  PUBLIC void setlab (label)*/
/***************************************************************************
 *
 *  setlab adds a record representing the setting of label 'label' to the
 *         code buffer.
 *
 **************************************************************************/
PUBLIC void setlab (const int label )
{
  dead = FALSE;
  label_def(label, C_SETLAB);
  if (assembly_output && ((alloc_strategy & ALLOC_NOCOMMENTS) == 0))
    {
      fprintf(outfile, "\nL%d:", 
              ((alloc_strategy & ALLOC_NOOPERANDS) == 0) ? label : 0);
    }
}
/*}}}*/
/*{{{  PUBLIC void setsectionlab (label)*/
/***************************************************************************
 *
 *  setsectionlab adds a record representing the setting of label 'label'
 *                to the code buffer, and marks this label as indicating
 *                the start of the current code section (routine, table...)
 *
 **************************************************************************/
PUBLIC void setsectionlab (const int label )
{
  dead = FALSE;
  section_label_def(label, C_SETLAB);
  if (assembly_output && ((alloc_strategy & ALLOC_NOCOMMENTS) == 0))
    {
      fprintf(outfile, "\nL%d:", 
              ((alloc_strategy & ALLOC_NOOPERANDS) == 0) ? label : 0);
      gencomment0("section label");
    }
}
/*}}}*/
/*{{{  PUBLIC void commentexp (tptr)*/
/***************************************************************************
 *
 *  commentexp writes the expression 'tptr' as a comment to the debugging
 *             file.
 *
 **************************************************************************/
PUBLIC void commentexp ( treenode *const tptr )
{
  /*if (!source_output)*/ /* removed source_output 14/3/91 */
  if (!disassemble) /* added to prevent it on the zd option - CON 26/3/91 */
  {
    if (TagOf(tptr) != S_CONSTEXP)
      switch(nodetypeoftag(TagOf(tptr)))
        {
          case NAMENODE: /*case HIDDENPARAMNODE:*/
            break;
          default:
            diag_write_nl_comment();
            printexp (outfile, tptr);
            break;
        }
  }
}
/*}}}*/
/*{{{  PRIVATE seginfo_t *add_to_seginfo_list*/
PRIVATE seginfo_t *add_to_seginfo_list(seginfo_t *const list, treenode *const nptr)
{
  seginfo_t *const seg = (seginfo_t *)newvec(sizeof(seginfo_t));
  seg->seg_next  = list;
  seg->seg_name  = nptr;
  return seg;  
}
/*}}}*/
/*{{{  PUBLIC void add_entry_point(nptr)*/
#if 0
/*{{{  replaced by coder_add_entry*/
/***************************************************************************
 *
 *  add_entry_point adds the entry point of the current code section
 *                  (given by the current section_label)
 *                  to the list of entry points.
 *
 **************************************************************************/
PUBLIC void add_entry_point ( treenode *const nptr )
{
#if 0
  seginfo_t *const seg = (seginfo_t *)newvec(sizeof(seginfo_t));
  seg->seg_next  = seginfo;
  seg->seg_name  = nptr;
  /*seg->seg_entry = section_label;*/ /* Pointer (ie. subscript in code buffer)
                                     to current section_label, this will be
                                     changed to offset from start of module
                                     when code is output. */
  /*assertion(section_label == findlabel(NPLabelOf(nptr)));*/
  seginfo = seg;
#else
  seginfo = add_to_seginfo_list(seginfo, nptr);
#endif
}
/*}}}*/
#endif
/*}}}*/
/*{{{  PUBLIC void coder_add_mapfile_entry*/
PUBLIC void coder_add_mapfile_entry(treenode *const nptr)
{
  if (mapfile_handle != NULL)
    mapinfo_list = add_to_seginfo_list(mapinfo_list, nptr);
}
/*}}}*/
/*{{{  PUBLIC void coder_add_entry                     called from GEN1*/
PUBLIC void coder_add_entry(treenode *const nptr)
{
  if (diagnostics)
    gencomment0 ("coder add entry");
  if (NLexLevelOf(nptr) == 0)
  {
    /* only lexlevel 0 routines */
    if (ims_asm_output)
      write_asm_desc(outfile, nptr);
    seginfo = add_to_seginfo_list(seginfo, nptr);
  }
  coder_add_mapfile_entry(nptr);         /* all routines */
}
/*}}}*/
/*{{{  PUBLIC void add_const_block(l, p)*/
/***************************************************************************
 *
 *  add_const_block adds l bytes of code, pointed to by p, to the code
 *                 buffer.
 *
 **************************************************************************/
PUBLIC void add_const_block ( const int l , const BYTE *const p,
                          treenode *type_tptr, const char *type_string)
{
  int i;
  for (i = 0; i < l; i++)
    add_code(p[i]);
  if (assembly_output && ((alloc_strategy & ALLOC_NOCOMMENTS) == 0))
    /*{{{  print the block to outfile*/
    {
      for (i = 0; i < l; i++)
        {
          if ((i & 0xF) == 0)
            diag_write_nl_string(ims_asm_output ? "byte" : ".BYTE");
          /*fprintf (outfile, "#%-3x", p[i]);*/
          if (ims_asm_output)
            fprintf (outfile, "%s #%.2X", (i & 0xF) == 0 ? "" : ",", p[i]);
          else
            fprintf (outfile, " #%.2X", p[i]);
        }
    }
    /*}}}*/
}
/*}}}*/
/*{{{  PUBLIC void coder_jumptoentrypoint*/
PUBLIC void coder_jumptoentrypoint(const treenode *const nptr)
{
  if (ims_asm_output)
    {
      diag_write_nl();
      fprintf(outfile, "%-8s$%s", "j", WNameOf(NNameOf(nptr)));
      gencomment0("jump to entrypoint");
    }
}
/*}}}*/
/*{{{  PUBLIC void coder_mark_alignment*/
PUBLIC void coder_mark_alignment(void)
{
  if (ims_asm_output)
    {
      diag_write_nl();
      fputs("align", outfile);
    }
}
/*}}}*/
/*{{{  PUBLIC int get_padlen()*/
PUBLIC int get_padlen(const int len, const int alignment)
{
  /*return (alignment - (len % alignment)) % alignment;*/
  if (alignment > 1)
    return (alignment - (len & (alignment-1))) & (alignment-1);
  else
    return 0;
}
/*}}}*/
/*{{{  PRIVATE void subwordalign*/
PRIVATE void subwordalign(const int padlen, const BYTE *const padbytes)
{
  if (padlen > 0)
    {
      add_const_block(padlen, padbytes, NULL, NULL);
      gencomment1("add %d bytes for alignment", padlen);
    }
}
/*}}}*/
/*{{{  PRIVATE void genpadding*/
PRIVATE void genpadding(const int len)
{
  static BYTE padding[4] = { 0, 0, 0, 0 };
  if (len > 4) err_abort();
  subwordalign(len, padding);
}
/*}}}*/
/*{{{  PUBLIC int endofsection_padlen*/
PUBLIC int endofsection_padlen(const int alignment)
/* This returns the number of bytes which have to be added to the previous
   section to make it end on that alignment
*/
{
  return get_padlen((int)bytes_output, alignment);
}
/*}}}*/
/*{{{  PUBLIC void alignendofsection()*/
/***************************************************************************
 *
 *  alignendofsection ensures that the last byte of the
 *               current code section is word aligned.
 *               It must be called after the code has been put in the
 *               code buffer, but before it is compressed.
 *
 *               In effect, it ensures that the end of the previous section
 *               is on a word boundary.
 *
 **************************************************************************/
PUBLIC void alignendofsection (const int alignment )
{
  genpadding(endofsection_padlen(alignment));
}
/*}}}*/
/*{{{  PUBLIC void alignwholemodule()*/
PUBLIC void alignwholemodule(void)
/* This is used at the end of the object file to ensure that the whole
   object file is word aligned
*/
{
  if (!ims_asm_output)
    {
      const int padlen = endofsection_padlen(bytesperword);
      if (padlen > 0)
        {
          const int l = newlab();
          setsectionlab(l);
          /*alignendofsection(bytesperword);*/ /* bug 1066 11/12/90 */
          subwordalign(padlen, libstub);
          compress_code();
        }
    }
  coder_mark_alignment();
}
/*}}}*/

/*{{{  PUBLIC void endsection_align(wlen)              called from GEN1*/
PUBLIC void endsectionalign(const int wlen)
{
  /* first allow for round up already implicit in word-oriented output */
  /* as everything goes into buffer via add_data_lword or add_instr_lword */
  const int len = (wlen + 3) & (~3);
  const int alignment = 4;
  const int this_align = get_padlen(len, alignment);
  const int  end_align = endofsection_padlen(alignment);
  if (diagnostics)
    fprintf (outfile, "\n!endsection: wlen = %d, bytes_output = %d, this_align:%d, end_align:%d\n",
                          wlen, bytes_output, this_align, end_align);
  genpadding((end_align + this_align) & (alignment-1));
}
/*}}}*/

/*{{{  PUBLIC void genloopend (wsoffset, endlabel, looplabel) called from GEN1t*/
PUBLIC void genloopend (const int wsoffset, const int endlabel, const int looplabel)
{
  genprimary(I_LDLP, wsoffset + REPL_BASE);
  genlabeldiff(I_LDC, endlabel, looplabel);   /*      ldc endlabel - looplabel */
  gensecondary(I_LEND);                   /*          lend                 */
  setlab(endlabel);                       /* endlabel:                     */
}
/*}}}*/
/*{{{  PUBLIC void genaltend (altcount, altendlab)     called from GEN10t*/
PUBLIC void genaltend (const int altcount, const int altendlab)
{
  if (altcount != 0)
    gensecondary(I_ALTEND);
  setlab(altendlab);
}
/*}}}*/
/*{{{  PUBLIC void genloadconstant*/
/*****************************************************************************
 *
 *  genloadconstant loads the value of the constant c onto the expression stack
 *
 *****************************************************************************/
PUBLIC void genloadconstant (const INT32 c)
{
  if (kroc_flag) /* inappropriate for non-transputer target -- DCW */
    {
      /* I_MINT? */
      genprimary(I_LDC, c);
      return;
    }
  switch(constformat_of_const(targetintsize, c, 0))
    {
      case constformat_mint32_ldnlp:
        gensecondary(I_MINT);
        genprimary(I_LDNLP, (INT32)((BIT32)c - (BIT32)MOSTNEG_INT32) / 4 /*bytesperword*/);
        break;
      case constformat_mint32_adc:
        gensecondary(I_MINT);
        genprimary(I_ADC,   (INT32)((BIT32)c - (BIT32)MOSTNEG_INT32));
        break;
      case constformat_mint32_not:
        gensecondary(I_MINT);
        gensecondary(I_NOT);
        break;
      case constformat_ldinf_ldnlp:
        gensecondary(I_LDINF);
        genprimary(I_LDNLP, (INT32)((BIT32)c - (BIT32)INFINITY) / 4 /*bytesperword*/);
        break;
      case constformat_ldinf_adc:
        gensecondary(I_LDINF);
        genprimary(I_ADC,   (INT32)((BIT32)c - (BIT32)INFINITY));
        break;
      case constformat_mint16_ldnlp:
        gensecondary(I_MINT);
        genprimary(I_LDNLP, (INT32)((BIT32)c - (BIT32)MOSTNEG_INT16) / 2 /*bytesperword*/);
        break;
      case constformat_mint16_adc:
        gensecondary(I_MINT);
        genprimary(I_ADC,   (INT32)((BIT32)c - (BIT32)MOSTNEG_INT16));
        break;
      case constformat_ldc:
      case constformat_table:
        genprimary(I_LDC, c);
        break;
    }
}
/*}}}*/
/*{{{  PUBLIC geni64tor*/
PUBLIC void geni64tor (const int sourcemode, treenode *const source,
               const int desttype, const int roundmode, const int simple)
{
  loadopd(sourcemode, source, 0);              /*  ldp source   */
  if (!simple) gensecondary(I_DUP);            /*  dup          */
  gensecondary(I_FPB32TOR64);                  /*  fpb32tor64   */
  if (desttype == S_REAL32 && !has_directfp)
    genfpuentry(I_FPUNOROUND);                 /*  fpunoround   */
  if (simple)
    loadopd(sourcemode, source, 1);            /*  ldp source+1 */
  else
    genprimary(I_LDNLP, 1);                    /*  ldnlp  1     */
  gensecondary(I_FPI32TOR64);                  /*  fpi32or64    */
  if (has_directfp)     gensecondary(I_FPEXPINC32);/*  fpexpinc32   */
  else                  genfpuentry(I_FPUEXPINC32);
  if (desttype == S_REAL32 && !has_directfp)
    genfpuentry(I_FPUNOROUND);                 /*  fpunoround   */
  if (roundmode == S_TRUNC)
    {
      if (has_directfp) gensecondary(I_FPRZ);  /*  fprz         */
      else              genfpuentry(I_FPURZ);
    }
  if (desttype == S_REAL32 && has_directfp)
    gensecondary(I_FPADDDBSN);                 /*  fpadddbsn    */
  else
    gensecondary(I_FPADD);                     /*  fpadd        */
}
/*}}}*/
/*{{{  PUBLIC BOOL genreplicatorcheck (must_check, poslab, zerolab)*/
PUBLIC BOOL genreplicatorcheck (const BOOL must_check,
                                const int poslab, const int zerolab)
/* generates code to jump to poslab if Areg is positive, zerolab if it is
   zero, and (if must_check is true) to stop if it is negative.
   returns TRUE iff an explicit jump to bodylab has been generated
   (some implementations (eg 386) might do this if must_check was TRUE)
   (note that one 386 version had loop incrementing etc at head of loop) */
/* This simplified version assumes poslab = bodylab = next instruction */
{
  if (must_check)
    /*{{{  check the replicator (in Areg) is > 0*/
    {
      gensecondary(I_MINT);           /*          mint ( = mostpos PLUS 1)*/
      gensecondary(I_CSUB0);          /*          csub0                */
      gencomment0("error if negative");
      checkerror();
    }
    /*}}}*/
  genbranch (I_CJ, zerolab);           /*          cj    zerolab         */
  return FALSE;
}
/*}}}*/
/*{{{  PUBLIC void genloopheadtop (wsoffset)      called from GEN1t*/
PUBLIC void genloopheadtop (const int wsoffset)
{
  return;
}
/*}}}*/
/*{{{  PUBLIC void genloopheadbottom ()           called from GEN1t*/
PUBLIC void genloopheadbottom ()
{
  return;
}
/*}}}*/
/*{{{  PUBLIC void gen_init_chanarray*/
PUBLIC void gen_init_chanarray (const int pstride, const int cstride, const int elems)
/* generate a loop to initialise arrays of pointers and channels */
/* pointer to array has already been loaded into tstack */
{
  const BIT32 base = ZERO32;  /* Use bottom two words of workspace for */
  const BIT32 count = ONE32;  /* base and count */
  int looplab;
  genprimary(I_STL, base);              /*          stl  base      */
  gencomment0("PTR base");
  genprimary(I_LDC, elems);             /*          ldc  elements  */
  gencomment0("elements");
  genprimary(I_STL, count);             /*          stl  count     */
  gencomment0("INT count");
  looplab = newlab();
  setlab(looplab);                      /* looplab:                */
  genprimary(I_LDL, count);             /*          ldl   count     */
  gencomment0("INT count");
  genprimary(I_ADC, -1);                /*          adc   -1        */
  genprimary(I_STL, count);             /*          stl   count     */
  gencomment0("INT count");
  if (chanaspointer)
    {
      genprimary(I_LDL, base);          /*          ldl   base      */
      gencomment0("PTR base");
      genprimary(I_LDNLP, elems*pstride);   /*      ldnlp elements */
      gencomment0("elements");
      genprimary(I_LDL, base);          /*          ldl   base      */
      gencomment0("PTR base");
      genprimary(I_STNL, 0);            /*          stnl  0         */
      gencomment0("init channel pointer");
    }
  gennotprocess ();
  genprimary(I_LDL, base);              /*          ldl   base      */
  gencomment0("PTR base");
  genprimary(I_STNL, chanaspointer ? elems : 0); /*     stnl  e         */
  gencomment0("init channel word");
  genprimary(I_LDL, base);              /*          ldl   base      */
  gencomment0("PTR base");
  genprimary(I_LDNLP, 1);               /*          ldnlp 1        */
  genprimary(I_STL, base);              /*          stl   base      */
  gencomment0("PTR base");
  genprimary(I_LDL, count);             /*          ldl   count     */
  gencomment0("INT count");
  genprimary(I_EQC, ZERO32);            /*          eqc   0         */
  genbranch(I_CJ, looplab);             /*          cj    looplab   */
}
/*}}}*/

/*{{{  PUBLIC int genconstant(cptr)*/
/***************************************************************************
 *
 *  genconstant adds bytes representing the constant tptr to the code
 *              buffer (least significant byte first). Returns the
 *              number of WORDS written to the code.
 *
 **************************************************************************/
PUBLIC int genconstant ( treenode *tptr )
{
  BYTE v[MAX_CONSTANT_SIZE];
  BIT32 l;
  int i;
  const int w = (int)wordsin(gettype(tptr));
  /*{{{  put the code bytes representing tptr into v*/
  l = LoValOf(tptr);
  for (i = 0; i < 4; i++)
    {
      v[i] = (BYTE)(l & 0xff);
      l >>= 8;
    }
  l = HiValOf(tptr);
  for (i = 4; i < 8; i++)
    {
      v[i] = (BYTE)(l & 0xff);
      l >>= 8;
    }
  /*}}}*/
  add_const_block(w * bytesperword, v, NULL, NULL);
  return(w);
}
/*}}}*/
/*{{{  PRIVATE void genloadcasescale ()*/
/***************************************************************************
 *
 *  genloadcasescale adds a record representing a primary I_LDC
 *               to the code buffer, whose operand is the size
 *               in bytes of a forthcoming jump table.
 *
 **************************************************************************/
PRIVATE void genloadcasescale (void)
{
  if (!dead)
    {
      scale_ref(I_LDC);
      if (assembly_output) diag_write_nl_string("ldc     case.scale");
    }
}
/*}}}*/
/*{{{  PUBLIC void genstartjumptable ()*/
/***************************************************************************
 *
 *  genstartjumptable adds a record to the code buffer marking the start
 *                    of a jump table.
 *
 **************************************************************************/
PUBLIC void genstartjumptable ( void )
{
  if (!dead)
    {
      const int jlab = newlab();
      const int l    = newlab();
      genloadcasescale();                     /*           ldc   casescale       */
      gensecondary(I_PROD);                   /*           prod                  */
      /* now Areg will contain offset into table of jumps immediately following */
      genloadlabptr(jlab, l, "table base");
      gensecondary(I_BSUB);                   /*           bsub                  */
      gensecondary(I_GCALL);                  /*           gcall                 */
      setlab(jlab);                           /* jlab:                           */
      table_start();
      if (assembly_output) diag_write_nl_string("-- Jump table");
    }
}
/*}}}*/
/*{{{  PUBLIC void genjumptableentry (instruction, label)*/
/***************************************************************************
 *
 *  genjumptableentry adds a record representing a jump table entry
 *                    to the code buffer.
 *
 **************************************************************************/
PUBLIC void genjumptableentry (const int instruction, const int label )
{
  if (!dead)
    {
      genbranch(instruction, label);
      dead = FALSE;
    }
}
/*}}}*/
/*{{{  PUBLIC void genendjumptable ()*/
/***************************************************************************
 *
 *  genendjumptable adds a record to the code buffer marking the end
 *                  of a jump table.
 *
 **************************************************************************/
PUBLIC void genendjumptable ( void )
{
  if (!dead)
    {
      table_end();
      if (assembly_output) diag_write_nl_string("-- End of jump table");
    }
}
/*}}}*/
/*{{{  PUBLIC void genshiftimmediate (op,shiftcount)*/
PUBLIC void genshiftimmediate (const int op, const int shiftcount)
/* generate fast code for ((Areg) << shiftcount) == ((Areg) * 2^shiftcount)
   (where shiftcount is a constant)
*/
{
  if (op == I_SHR)
    /*{{{  rightshift*/
    {
      genloadconstant(shiftcount);
      gensecondary(I_SHR);
    }
    /*}}}*/
  else if (kroc_flag) /* inappropriate for non-transputer target -- DCW */
    /*{{{  leftshift*/
    {
      genloadconstant(shiftcount);
      gensecondary(I_SHL);
    }
    /*}}}*/
  else
    /*{{{  leftshift = mul by power of 2*/
    /* this code takes advantage of special transputer instructions */
    {
      const INT32 bcnts = lshiftcanusebcnt(shiftcount);

      if (bcnts > 0)
        /*{{{  we can use bcnt*/
        {
          int i;
          for (i = 0; i < bcnts; i++)
            gensecondary(I_BCNT);
        }
        /*}}}*/
      else if ((((code_style_flags & CODE_STYLE_SPACE_NOT_TIME) != 0) ||
                bytesperword == 2)
               && (shiftcount <= 3))
        {
          /* it is smaller to use prod, and code size mattters on 16-bit */
          genloadconstant((INT32)1 << shiftcount);
          gensecondary(I_PROD);
        }
      else if (has_dup && !T9000_instruction_timings &&
               (shiftcount == 1)) /* bug 1168 22/08/91 */
        {
          gensecondary(I_DUP);
          gensecondary(I_BSUB);
        }
      else if (has_wsubdb && (shiftcount == 3))
        {
          /* we've already checked that bytesperword != 2 when shiftcount == 3 */

          genloadconstant(0);
          gensecondary(I_WSUBDB); /* bug TS/1168 16/01/92 */
        }
      else
        {
          /* Bug 1168 - 14/3/91 - use a shift cos its quicker */
          /* This 'loadconstant' is a small +ve integer, so will not need
             a constant table
          */
          genloadconstant(shiftcount);
          gensecondary(I_SHL);
        }
    }
    /*}}}*/
}
/*}}}*/
/*{{{  PUBLIC void genfpuentry (int instruction)*/
/*****************************************************************************
 *
 *  genfpuentry generates an fpuentry instruction
 *
 *****************************************************************************/
PUBLIC void genfpuentry (const int instruction )
{
  if (!dead)
    {
      /*{{{  check whether instruction may set error*/
      switch (instruction)
        {
          default:
            break;
          case I_FPUSQRTFIRST:
          case I_FPUMULBY2:   case I_FPUDIVBY2:
          case I_FPUEXPINC32: case I_FPUEXPDEC32:
          case I_FPUABS:
          case I_FPUSETERR:
          case I_FPUR32TOR64: case I_FPUR64TOR32:
          case I_FPUCHKI32: case I_FPUCHKI64:
            dirty_error_flag(TRUE);
            break;
        }
      /*}}}*/
      /*code_primary(I_LDC, instruction & 0xff);*/
      code_primary(I_LDC, (int)(instruction & ~I_FPU_ENTRY_BIT));
      code_secondary(I_FPENTRY);
      if (assembly_output)
        {
          diag_write_nl();
          fprintf(outfile, "%-8s", secondaryname(instruction));
        }
    }
}
/*}}}*/
/*{{{  PUBLIC void declarelabel(label, offset)*/
#if 0 /* never used */
/***************************************************************************
 *
 *  declarelabel declares the label 'label' at offset 'offset' from the
 *               start of code.
 *
 **************************************************************************/
PUBLIC void declarelabel (const int label, const BIT32 offset )
{
  int lptr = addlabel(label);
  code[lptr + LB_PRAD] = offset;
}
#endif
/*}}}*/

/*{{{  PUBLIC noteprocentry ()                         called from GEN1t*/
PUBLIC void noteprocentry (void)
{
  /* just a placeholder in case need to save return link or whatever */
}
/*}}}*/
/*{{{  PUBLIC void genprocreturn (tag, regs, ws)       called from GEN1t*/
PUBLIC void genprocreturn (const int tag, const int regs, const int ws)
{
  gensettsdepth (regs);
  genprimary (I_AJW, ws);  /* maybe include adjustment implicit in RET */
  gensecondary (I_RET);
}
/*}}}*/
/*{{{  PUBLIC void gen_func_results (results)          called from GEN4t,11t*/
PUBLIC void gen_func_results (int results)
{
  gensettsdepth (results);
}
/*}}}*/
/*{{{  PUBLIC void genwidenshort ()            called from GEN4t*/
/*****************************************************************************
 *
 *  genwidenshort generates code to widen from INT16 to INT32
 *
 *****************************************************************************/
PUBLIC void genwidenshort (void)
{
  genprimary(I_LDC, INT16_CHECK_MASK);
  gensecondary (I_XWORD);
}
/*}}}*/
/*{{{  PUBLIC void gennotprocess ()            called from GEN1t*/
/*****************************************************************************
 *
 *  gennotprocess generates ldc,NOT_PROCESS for chan initialisation etc
 *
 *****************************************************************************/
PUBLIC void gennotprocess (void)
{
    gensecondary(I_MINT);
}
/*}}}*/
/*{{{  PUBLIC void genfppop ()            called from GEN8t*/
/*****************************************************************************
 *
 *  genfppop is called when a value can usefully be popped off top of FP stack
 *
 *****************************************************************************/
PUBLIC void genfppop(void)
{
}
/*}}}*/
/*}}}*/
/*{{{  run time locate information*/
/*{{{  PUBLIC void new_occam_line*/
/*****************************************************************************
 *
 *  combines coder_genlocate and mark_occam_line
 *
 *****************************************************************************/
PUBLIC void new_occam_line (treenode *laddress,
         BOOL write_linenumber, BOOL needs_locate, BOOL clear_tstack)
{
  if (debugoutput && needs_locate)
    add_debug_info(C_LOCATE, laddress, 0);
  /*{{{  write linenumber*/
  if (assembly_output && write_linenumber &&
      ((alloc_strategy & ALLOC_NOLINENUMS) == 0))
    {
      const SOURCEPOSN locn = sourcefileof(LocnOf(laddress));
      if (locn != oldlocn)
      {
        diag_write_nl_string("-- ");
        fprintf(outfile, "\"%s\": %d",
                       lookupfilename(FileNumOf(locn)), FileLineOf(locn));
        DEBUG_MSG((" addr: %lX", (BIT32)laddress));
        oldlocn = locn;
      }
    }
  /*}}}*/
}
/*}}}*/
#if 0
/*{{{  PUBLIC void coder_genlocate - takes treenode address as its param*/
/*****************************************************************************
 *
 *  coder_genlocate inserts a runtime locate item record into the code buffer
 *
 *****************************************************************************/
PUBLIC void coder_genlocate ( treenode *const addr, const BOOL write_linenumber )
{
  add_debug_info(C_LOCATE, addr, 0);
  /*{{{  debugging*/
  if (/*diagnostics*/ assembly_output && write_linenumber &&
      ((alloc_strategy & ALLOC_NOLINENUMS) == 0))
    {
      const SOURCEPOSN locn = sourcefileof(LocnOf(addr));
      /*diag_write_nl_string("-- Locate");*/
      diag_write_nl_string("-- ");
      fprintf(outfile, "\"%s\": %d",
                       lookupfilename(FileNumOf(locn)), FileLineOf(locn));
      DEBUG_MSG((" addr: %lX", (BIT32)addr));
    }
  /*}}}*/
  /*{{{  source output*/
  /* removed source_output 14/3/91 */
  /*if (source_output)
    output_source_code(get_from_index_table(addr), C_LOCATE);*/
  /*}}}*/
}
/*}}}*/
#endif
/*{{{  PUBLIC void coder_genaddrfix (address) - takes treenode address as its param*/
/*****************************************************************************
 *
 *  coder_genaddrfix inserts an addressfix item record into the code buffer
 *
 *****************************************************************************/
PUBLIC void coder_genaddrfix ( treenode *addr, const int label )
{
  add_debug_info(C_ADDRESSFIX, addr, label);
  /*{{{  debugging*/
  if (diagnostics)
    diag_write_nl_string("-- Addressfix");
  /*}}}*/
  /*{{{  source output*/
  /* removed source_output 14/3/91 */
  /*if (source_output)
    output_source_code(get_from_index_table(addr), C_ADDRESSFIX);*/
  /*}}}*/
}
/*}}}*/
/*{{{  PRIVATE void coder_genlibrpatch ()*/
/*****************************************************************************
 *
 *  coder_genlibrpatch inserts a libpatch item record into the code buffer
 *
 *****************************************************************************/
PRIVATE void coder_genlibrpatch ( void )
{
  add_debug_info(C_LIBPATCH, 0, 0);
  /*{{{  debugging*/
  if (diagnostics)
    diag_write_nl_string("-- Libpatch");
  /*}}}*/
}
/*}}}*/
/*{{{  PRIVATE void coder_genprofaddrfix ()*/
/*****************************************************************************
 *
 *  coder_genprofaddrfix inserts a profiler addressfix  item record 
 *  into the code buffer
 *
 *****************************************************************************/
PRIVATE void coder_genprofaddrfix ( void )
{
  add_debug_info(C_PROFADDRESSFIX, 0, 0);
  /*{{{  debugging*/
  if (diagnostics)
    diag_write_nl_string("-- Profile table Addressfix");
  /*}}}*/
}
/*}}}*/
/*}}}*/
/*{{{  PUBLIC void throw_the_result_away ()*/
PUBLIC void throw_the_result_away (void)
{
  /* ts_depth = 0; */
  gencomment0 ("throw the result away\n-- stackdepth = 0");
}
/*}}}*/
/*{{{  PUBLIC void throw_nested_result_away ()*/
PUBLIC void throw_nested_result_away (void)
{
  gencomment0 ("throw nested result away\n-- stackdepth -= 1");
}
/*}}}*/
/*{{{  PUBLIC void gencontrolsplit (label)*/
PUBLIC void gencontrolsplit (const int label)
{
  /* placeholder for possible need to note that tstack must match on two routes */
  genbranch (I_CJ, label);
}
/*}}}*/
/*{{{  PUBLIC void gencontroljoin (label)*/
PUBLIC void gencontroljoin (const int label)
{
  /* placeholder for possible need to note that tstack must match on two routes */
  setlab (label);
}
/*}}}*/
/*{{{  library handling*/
/*{{{  PUBLIC void add_to_libentries(nptr)*/
/*****************************************************************************
 *
 *  add_to_libentries adds the library entrypoint 'nptr' to the list of used
 *                    library entry points if it is not aready there.
 *
 *****************************************************************************/
PUBLIC void add_to_libentries ( treenode *nptr )
{
  treenode *l = libentrychain;
  while (l != NULL)
    {
      if (l == nptr)
        return;
      l = NLEntryNextOf(l);
    }
  SetNLEntryNext(nptr, libentrychain);
  libentrychain = nptr;                   /* Add to the entry point list */
}
/*}}}*/
/*{{{  PUBLIC void genlibstubs()*/
/*****************************************************************************
 *
 *  genlibstubs adds a stub for each library call to the code buffer
 *
 *****************************************************************************/
PUBLIC void genlibstubs ( void )
{
  const int stubsize = libpatchsize > sizeof(libstub) ? sizeof(libstub) : libpatchsize;
  treenode *libentry;
  for (libentry = libentrychain; libentry != NULL;
       libentry = NLEntryNextOf(libentry))
    {
      const char *name = WNameOf(NNameOf(libentry));
      const int l = newlab();
      SetNPLabel(libentry, l);          /* give the entry point a label */
      setsectionlab(l);
      gencomment1("Stub for library call %s", (BIT32)name);
      /*{{{  debugging*/
      if (debugoutput)
        coder_genlibrpatch();
      /*}}}*/
      /*{{{  ims_asm_output*/
      if (ims_asm_output)
        {
          /*fprintf(outfile, "\n$%s:", name);*/
          diag_write_nl();
          fprintf(outfile, "%-8s$%s", "extern", name);
          diag_write_nl();
          fprintf(outfile, "%-8s%d j codefix $%s 0", "patch", stubsize, name);
        }
      /*}}}*/
      add_const_block(stubsize, libstub, NULL, NULL);
      compress_code();  /* this must be done once for each library stub */
    }
}
/*}}}*/
/*{{{  PUBLIC  int genproftable(INT32 size)*/
/*****************************************************************************
 *
 *  genproftable creates and initializes the profile table in the code buffer
 *
 *****************************************************************************/
PUBLIC  int genproftable(INT32 proftsize)
{
  int proftabconstlab, i;
  BYTE *ptr;
  BYTE *profptr = (BYTE *)memalloc(sizeof(BYTE)*bytesperword*proftsize);
  for (i=0, ptr = profptr ; i < proftsize*bytesperword; i++)
    *(ptr + i) = 0;
  coder_genprofaddrfix();
  proftabconstlab = newlab();
  setsectionlab(proftabconstlab);
  if (assembly_output && ((alloc_strategy & ALLOC_NOCOMMENTS) == 0))
    gencomment0("Profiler table for the compilation unit ");
  add_const_block(proftsize*bytesperword, profptr, NULL, NULL);
  compress_code();  
  return proftabconstlab;
}
/*}}}*/
/*}}}*/

/*{{{  PUBLIC void write_object_file()*/
/*****************************************************************************
 *
 * write_object_file writes out the object file and closes it
 *
 *****************************************************************************/
PUBLIC void write_object_file ( void )
{
  calculate_code_size();

  if (assembly_output && ((alloc_strategy & ALLOC_NOCOMMENTS) == 0))
    display_code_size(diagnostics ? "" : "\n-- ");

  patch_lib_offsets();
  /* if (!disassemble && !assembly_output)  */
  if (tcoff_without_code || (!assembly_output)) /* MDP */
    {
      if (!tcoff_without_code)
        output_code();
      if (debugoutput) flush_debug_buffers();

      write_library_calls(TRUE, libentrychain, I_J);
      output_entry_points();
      output_code_size();
      output_mapfile_sectiontable();

      if (configuring)
        code_link_map_fn();  /* write the linker map into the file */
      if ((sampling_profiling || line_profiling || cgraph_profiling ) &&
          profile_table->count_table_size > 0                             )
        write_profile_section(profile_table);
      write_end_module ();
    }
  else if (disassemble)
    {
      if (!tcoff_without_code)
        output_code();
      output_entry_points();
    }
}
/*}}}*/
/*{{{  PUBLIC void close_object_file()*/
/*{{{  comment*/
/*****************************************************************************
*
*  close the object file - called by the harness
*
******************************************************************************/
/*}}}*/
PUBLIC void close_object_file (FILE *fptr, const char *const filename)
{
  if (fclose(fptr) == EOF)
  /*generr_s(GEN_CANNOT_CLOSE_OBJECT_FILE, filename);*/
    msg_out_s(SEV_FATAL, GEN, GEN_CANNOT_CLOSE_OBJECT_FILE, NOPOSN, filename);
}
/*}}}*/
/*{{{  PUBLIC FILE *open_object_file()*/
/*****************************************************************************
 *
 * open_object_file opens the object file for writing
 *
 *****************************************************************************/
PUBLIC FILE *open_object_file (const char *const filename)
{
  FILE *const fptr = fopen(filename, "wb");
  if (fptr == NULL)
  /*generr_s(GEN_CANNOT_OPEN_OBJECT_FILE, filename);*/
    msg_out_s(SEV_FATAL, GEN, GEN_CANNOT_OPEN_OBJECT_FILE, NOPOSN, filename);
  return fptr;
}
/*}}}*/

/*{{{  PUBLIC void initcode()*/
/***************************************************************************
 *
 * initcode initialises the coder.
 *
 **************************************************************************/
PUBLIC void initcode ( void )
{
  if (req_code_size != 0)
    {
      /* we multiply by 1126 (1024 + 10%) so that in effect we add 10%
         to the buffer size.
         (requested in K, so multiply by 1024 * 1.1)
         This allows for the other crap which is used in the code buffer */
      /* bug 1021 - 16/10/90 */
        max_code_size = req_code_size * (1126 / sizeof(INT32));
    }
  else
    max_code_size = configuring ? CONFIGURING_CODE_SIZE : COMPILING_CODE_SIZE;
  if (t450a_workarounds || T9000_gamma_badldnlp(&tx_global))
    prev_instruction = I_NOP;
  if (code == NULL) code = (INT32 *)memalloc(max_code_size * sizeof(INT32));
  cstart = 0;
  cptr = 0;
  lptr = max_code_size;
  start = C_NIL;
  topl  = C_NIL;
  codew = 0;
  coden = 0;
  partword = 0;
  address  = 0;
  case_scale = 0;
  seginfo = NULL;
  bytes_output = 0;
  /* debugnamesptr = 0; */
  libentrychain = NULL;
  dead = FALSE;
  int_error_status = ERROR_STATUS_CLEAN;
  fpu_error_status = ERROR_STATUS_FP_INIT;
}
/*}}}*/

/*{{{  PUBLIC void freecode*/
/* This is only called if configuring is TRUE */
PUBLIC void freecode(void)
{
  if (code != NULL)
    {
      memfree(code);
      code = NULL;
    }
}
/*}}}*/
