/*#define DEBUG*/
#define SHOWTSTACK

/*
 *	instruction generation, code adjustment, code output
 *	Copyright (C) 1987 Inmos Limited
 *
 *	This program is free software; you can redistribute it and/or modify
 *	it under the terms of the GNU General Public License as published by
 *	the Free Software Foundation; either version 2 of the License, or
 *	(at your option) any later version.
 *
 *	This program is distributed in the hope that it will be useful,
 *	but WITHOUT ANY WARRANTY; without even the implied warranty of
 *	MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *	GNU General Public License for more details.
 *
 *	You should have received a copy of the GNU General Public License
 *	along with this program; if not, write to the Free Software
 *	Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 */

/*
 *  This file is work-in-progress for a DEC alpha target version
 *  M.D.Poole@ukc.ac.uk (occam for all) work started 3rd March 1995
 */

/*{{{  include files*/
#ifdef HAVE_CONFIG_H
#  include <config.h>
#endif
#include <stdio.h>
#include <string.h>
#include "includes.h"
#include "genhdr.h"
#include "instruct.h"
#include "instdef.h"
#include "generror.h"

#include "objwrt.h"
#include "gen1def.h"
#include "gen2def.h"
#include "gen4def.h"
#include "code1def.h"
#include "bind2def.h"
#include "bind3def.h"
#include "debugdef.h"
#include "gendeca.h"
#include "disalpha.h"
#include "maplib.h" /* mapfile handling */ /* IMPORTED */
/*}}}*/

/*{{{  definitions*/
/* The actual code buffer size is CODE_SIZE * sizeof(INT32) */
#define CONFIGURING_CODE_SIZE 10000 /* approx  40K bytes */
#define COMPILING_CODE_SIZE   60000 /* approx 240K bytes */

PUBLIC  int req_code_size = 0; /* code size requested (in K) */
PRIVATE int max_code_size;     /* buffer size (in INT32s)    */


#define MAX_BYTES_IN_INSTRUCTION 8  /* 32 bit target wordsize maximum */
#define MAX_CONSTANT_SIZE        8  /* 8 bytes for a REAL64 etc */

/*{{{  code tags*/
#define C_NIL (-1)
#define C_SETLAB (-11)
#define C_USELAB (-12)
#define C_USELABDIFF (-13)
/* #define C_STARTFILE (-14) */
/* #define C_ENDFILE (-15)   */
#define C_LOCATE (-16)
#define C_SCALE (-17)
#define C_TABLE (-18)
#define C_LIBPATCH (-19)
#define C_ADDRESSFIX (-20)
/*}}}*/

/*{{{  fields in a setlab record*/
#define SL_SIZE     4                                 /* Length of the record */
#define SL_ADDR     1                /* Last calculated address of the setlab */
#define SL_NEXT     2                 /* Pointer to next item on adjust chain */
#define SL_TSTK     3                                 /* Actual tstack state  */
/*}}}*/

/*{{{  fields in a uselab record*/
#define U_SIZE     7                                 /* Length of the record */
#define U_ADDR     1           /* Last calculated address of the label usage */
#define U_NEXT     2                 /* Pointer to next item on adjust chain */
#define U_DEF      3                          /* Pointer to label definition */
#define U_ADJ      4  /* Current estimated length of instruction using label */
#define U_OP       5                              /* Instruction using label */
#define U_TSTK     6                                /* Expected tstack state */
#ifdef sun4           /* big-endian */
#define TSTK_UNDEF 0x7E7E7E00                       /* "~~~" */
#else
#define TSTK_UNDEF 0x007E7E7E                       /* "~~~" */
#endif
/*}}}*/
/*{{{  fields in a uselabdiff record*/
#define UD_SIZE    7                                 /* Length of the record */
#define UD_ADDR    1       /* Last calculated address of the labeldiff usage */
#define UD_NEXT    2                 /* Pointer to next item on adjust chain */
#define UD_DEF1    3                 /* Pointer to definition of first label */
#define UD_ADJ     4  /* Current estimated length of instruction using label */
#define UD_OP      5                          /* Instruction using labeldiff */
#define UD_DEF2    6                /* Pointer to definition of second label */
/*}}}*/
/*{{{  fields in a label record*/
#define LB_SIZE    3                                 /* Length of the record */
#define LB_NAME    0                                           /* Label name */
#define LB_PRAD    1                  /* Address of the corresponding setlab */
#define LB_DEF     2                  /* Pointer to the corresponding setlab */
/*}}}*/
/*{{{  fields in a table record*/
#define T_SIZE     3                                 /* Length of the record */
#define T_SCALE    1            /* Pointer to field in related scale record  */
#define T_NEXT     2                 /* Pointer to next item on adjust chain */
/*}}}*/
/*{{{  fields in a scale record*/
#define SC_SIZE    4                                 /* Length of the record */
#define SC_VALUE   1              /* Last calculated size of the scale value */
#define SC_NEXT    2                 /* Pointer to next item on adjust chain */
#define SC_OP      3                         /* Instruction using case scale */
/*}}}*/
/*{{{  fields in a debug record*/
#define DB_SIZE    5                                 /* Length of the record */
#define DB_ADDR    1            /* Last calculated address of the debug item */
#define DB_NEXT    2                 /* Pointer to next item on adjust chain */
#define DB_LINE    3                     /* Source line number of debug item */
#define DB_DATA    4                       /* Label of associated code chunk */
/* bug 1158 means that multiple tables can have the same addresses;
   these are marked with address fixes, where the DB_DATA field is the label
   number of the section of code for that table
   CON 14/2/91
*/
/*}}}*/


/*{{{  segment definitions*/
typedef struct seginfo_struct
  {
    struct seginfo_struct *seg_next;
    treenode *seg_name;
    /*INT32 seg_entry;*/
  } seginfo_t;
/*}}}*/

#define IBUF_SIZE 256  /* amount copied at one time to code file */
/* This must be small enough to fit into the output file buffer */

/*}}}*/
/*{{{  global variables*/
/*{{{  public variables*/
/* All initialised to NULL */
PUBLIC void (*code_pseudo_link_fn)(treenode *);
PUBLIC void (*code_save_entrydata_fn)(const treenode *const nptr,
                              const INT32 offset, const INT32 code_size);
PUBLIC void (*code_link_map_fn)(void);
/*}}}*/

PRIVATE int fpu_mode;
PRIVATE int dead;
PRIVATE treenode *libentrychain;
PRIVATE BOOL compressing;
PRIVATE SOURCEPOSN oldlocn;
PRIVATE BOOL showtstack;
PRIVATE BOOL assembly_optimise;
PRIVATE int special_inst;
PRIVATE int stzwlen;

/*{{{  data for code buffer handling*/
PRIVATE INT32 *code = NULL;          /* Buffer to hold code during compaction */

PRIVATE int cstart;          /* Pointer to first free location in code buffer */
PRIVATE int cptr;                    /* Pointer to top of code in code buffer */
PRIVATE int lptr;          /* Pointer to bottom of label info. in code buffer */
PRIVATE int topl;       /* Pointer to top of chain of label defns. and usages */
PRIVATE int start; /* Pointer to start of current code segment in code buffer */
PRIVATE int coden;         /* Number of bytes of code in current code segment */
PRIVATE INT32 address;            /* Current estimate of code address within SC */
                     /* Pointer to label which marks entry point for section */
PRIVATE int section_label;
            /* Pointer to most recently declared scale record in code buffer */
PRIVATE int case_scale;
PRIVATE INT32 getword;              /* Word of four bytes read from code buffer */
                    /* Bytes packed into getword still to be written to ibuf */
PRIVATE int bytesleft;
PRIVATE int getptr;     /* Next word to be taken from code buffer by put_code */

PRIVATE BIT32 section_size; /* Current estimate of size of current code section */
PRIVATE BIT32 total_code_size;                             /* Total code size */
PRIVATE BIT32 bytes_output;
PRIVATE BYTE padding[8] = {0x20, 0x20, 0x20, 0x20, 0x20, 0x20, 0x20, 0x20};

/*{{{  STOP on error handling*/
typedef enum {
  ERROR_STATUS_START,   /* At the start of a block    */
  ERROR_STATUS_CLEAN,   /* Error is known to be clear */
  ERROR_STATUS_DIRTY    /* Error may have been set    */
  } error_status_t;

PRIVATE error_status_t int_error_status;
PRIVATE error_status_t fpu_error_status;

#define LOCALISE_ERROR_HANDLING TRUE
/* TRUE if we want to wait until generating an FPU instruction before clearing
   the FPU error flag
   BUT we need a way to clear FPU error without clobbering integer stack.
   There isn't one, cos both FPTESTERR and FPUCLRERR hit 1 stack slot.
   AHAH! We now use the fact that FPU loads always use a CPU slot,
   so we can use that after the load has happened!  CO'N 12/2/90
   We can't use this system for normal errors, cos the 'TESTERR' instruction
   clobbers the stack
*/

#if LOCALISE_ERROR_HANDLING
#define ERROR_STATUS_FP_INIT ERROR_STATUS_START
#else
#define ERROR_STATUS_FP_INIT ERROR_STATUS_CLEAN
#endif
/*}}}*/

/*}}}*/
/*{{{  data for segment handling*/
PRIVATE seginfo_t *seginfo;

PRIVATE seginfo_t *mapinfo_list = NULL;
/*}}}*/
/*{{{  DEC alpha registers etc*/
PRIVATE RegA, RegB, RegC;
PRIVATE FRegA, FRegB, FRegC;
PRIVATE oldregA, oldregB, oldregC;
/*}}}*/
/*}}}*/

/*{{{  private routines*/
/*{{{  tstack diagnostics*/
/* a value between -3,3 is the change in depth of tstack */
/* a value >= 16 is 16 more than the new ts_depth to be set */
/*{{{  primaries*/
const int tsdiff_prim [16] = {16, 1, 0, 0,        /* 00 j, ldlp, pfix, ldnl */
                              1, 0, 0, 1,         /* 04 ldc, ldnlp, nfix, ldl */
                              0, 16, -1, 0,       /* 08 adc, call, cj, ajw */
                              0, -1, -2, 0};      /* 0C eqc, stl, stnl, opr */
/*}}}*/
/*{{{  secondaries*/
const int tsdiff_sec [0xB0] = {0, 0, -1, 16,      /* 00 rev, lb, bsub, endp */
                              -1, -1, 0, 16,      /* 04 diff, add, gcall, in */
                              -1, -1, -1, 16,     /* 08 prod, gt, wsub, out  */
                              -1, 16, 16, 16,     /* 0C sub, startp, outbyte, outword */
                              0, 0, 0, -1,        /* 10 seterr, ., resetch, csub0 */
                              0, 16, -2, -1,      /* 14 ., stopp, ladd, stlb */
                              -1, 1, -1, 0,       /* 18 sthf, norm, ldiv, ldpi */
                              -1, 1, 1, -1,       /* 1C stlf, xdble, ldpri, rem */
                              16, 16, 1, 0,       /* 20 ret, lend, ldtimer, . */
                              0, 0, 0, 0,         /* 24 ., ., ., . */
                              0, 1, 1, 16,        /* 28 ., testerr, testpranal, tin */
                              -1, 0, -2, -2,      /* 2C div, ., dist, disc */
                              -1, -1, 0, -1,      /* 30 diss, lmul, not, xor */
                              0, -1, -1, -1,      /* 34 bcnt, lshr, lshl, lsum */
                              -2, 0, -1, -2,      /* 38 lsub, runp, xword, sb */
                              0, -1, -1, 1,       /* 3C gajw, savel, saveh, wcnt */
                              -1, -1, 1, 0,       /* 40 shr, shl, mint, alt */
                              16, 0, -1, -1,      /* 44 altwt, altend, and, enbt */
                              -1, 0, -3, -1,      /* 48 enbc, enbs, move, or */
                              -1, -1, 0, -1,      /* 4C csngl, ccnt1, talt, ldiff */
                              -1, 16, -1, -1,     /* 50 sthb, taltwt, sum, mul */
                              -1, 0, -1, 0,       /* 54 sttimer, stoperr, cword, clrhalterr */
                              0, 1, 1, 16,        /* 58 sethalterr, testhalterr, dup, move2dinit */
                              16, 16, 16, 0,      /* 5C move2dall, move2dnonzero, move2dzero, . */
                              0, 0, 0, 19,        /* 60 ., ., ., unpacksn */
                              0, 0, 0, 0,         /* 64 ., ., ., . */
                              0, 0, 0, 0,         /* 68 ., ., ., . */
                              19, -2, 0, 0,       /* 6C postnormsn, roundsn, ., . */
                              1, 0, -1, 0,        /* 70 ., ldinf, fmul, cflerr */
                              -1, -1, -1, 0,      /* 74 crcword, crcbyte, bitcnt, bitrevword */
                              -1, -1, 0, 0,       /* 78 bitrevnbits, pop, ., . */
                              0, 0, 0, 0,         /* 7C ., ., ., . */
                              0, -1, -2, 0,       /* 80 ., wsubdb, fpldnldbi, fpchkerr */
                              -1, 0, -2, 0,       /* 84 fpstnldb, ., fpldnlsni, fpadd*/
                              -1, 0, -1, 0,       /* 88 fpstnlsn, fpsub, fpldnldb, fpmul */
                              0, 0, -1, 1,        /* 8C fpdiv, ., fpldnlsn, fpremfirst  */
                              1, 1, 1, 1,         /* 90 fpremstep, fpnan, fpordered, fpnotfinite */
                              1, 1, -1, 0,        /* 94 fpgt, fpeq, fpi32tor32, . */
                              -1, 0, 0, 0,        /* 98 fpi32tor64, ., fpb32tor64, . */
                              1, 0, -1, 0,        /* 9C fptesterr, fprtoi32, fpstnli32, fpldzerosn */
                              0, 0, 0, 0,         /* A0 fpldzerodb, fpint, ., fpdup */
                              0, 0, -1, 0,        /* A4 fprev, ., fpldnladddb, . */
                              -1, 0, -1, -1,      /* A8 fpldnlmuldb, ., fpldnladdsn, fpentry */
                              -1, 0, 0, 0};       /* AC fpldnlmulsn, ., ., .*/
/*}}}*/
PRIVATE char tstack [4];
PRIVATE int ts_depth, old_ts_depth;
PRIVATE int fs_depth, fs_width;
/*{{{  void check_tsdepth ()*/
PRIVATE void check_tsdepth ()
{
  if (ts_depth < 0)
  {
    fprintf (outfile, "\n%s stackdepth =  underflow !!!", COMMENT_INTRO);
    ts_depth = 0;
  }
  else if (ts_depth > 3)
  {
    fprintf (outfile, "\n%s stackdepth =  overflow !!!", COMMENT_INTRO);
    ts_depth = 3;
  }
}
/*}}}*/
/*{{{  void note_tstack ()*/
PRIVATE void note_tstack ()
{
  tstack [0] = (ts_depth >= 1) ? 'A' : '~';
  tstack [1] = (ts_depth >= 2) ? 'B' : '~';
  tstack [2] = (ts_depth >= 3) ? 'C' : '~';
  tstack [3] = '\0';
}
/*}}}*/
/*{{{  set_tregs*/
PRIVATE void set_tregs (const int depth)
{
  switch (depth)
  {
    case 0: RegA = REG_UNDEF; RegB = REG_UNDEF; RegC = REG_UNDEF; break;
    case 1: RegA = REG_0; RegB = REG_UNDEF; RegC = REG_UNDEF; break;
    case 2: RegA = REG_1; RegB = REG_0; RegC = REG_UNDEF; break;
    case 3: RegA = REG_2; RegB = REG_1; RegC = REG_0; break;
  }
}
/*}}}*/
/*{{{  void undefine_tstack ()*/
PRIVATE void undefine_tstack ()
{
  ts_depth = 0;
  set_tregs (ts_depth);
}
/*}}}*/
/*{{{  set_fregs*/
PRIVATE void set_fregs (const int depth)
{
  switch (depth)
  {
    case 0: FRegA = FREG_UNDEF; FRegB = FREG_UNDEF-1; FRegC = FREG_UNDEF-2; break;
    case 1: FRegA = FREG_0; FRegB = FREG_UNDEF; FRegC = FREG_UNDEF-1; break;
    case 2: FRegA = FREG_1; FRegB = FREG_0; FRegC = FREG_UNDEF; break;
    case 3: FRegA = FREG_2; FRegB = FREG_1; FRegC = FREG_0; break;
  }
  fs_depth = depth;
}
/*}}}*/
/*{{{  popfs ()*/
PRIVATE void popfs ()
{
  fs_depth--;
  FRegA--;
  FRegB--;
  FRegC--;
}
/*}}}*/
/*{{{  pushfs ()*/
PRIVATE void pushfs ()
{
  fs_depth++;
  FRegA++;
  FRegB++;
  FRegC++;
}
/*}}}*/
/*}}}*/
/*{{{  picture of the code buffer*/
/*****************************************************************************
 *
 *   <---- Code                                              Labels    ---->
 * ---------------------------------------------------------------------------
 * |  |   |    |                     |       |            | | | | | | | | |  |
 * |  |   |    | current code        |       | current    |   entry points   |
 * |  |   |    |     --->            |       | labels     |   for previously |
 * |  |   |    |                     |       |  <---      |   compacted      |
 * |  |   |    |                     |       |            |   routines       |
 * |  |   |    |                     |       |            | | | | | | | | |  |
 * |  |   |    |                     |       |            | | | | | | | | |  |
 * ---------------------------------------------------------------------------
 *  previously  ^                     ^       ^          ^
 *  compacted   cstart                cptr    lptr       section
 *  routines                                             label
 *
 *****************************************************************************/

/*{{{  format of previously compacted routines*/
/*****************************************************************************
 *
 * ---------------------------------------------------------------------------
 * | |                      | # |            | |
 * | |                      | # |            | |
 * | |                      | # |            | |
 * | |                      | # |            | |
 * | |                      | # |            | |
 * ---------------------------------------------------------------------------
 *  ^                        | ^              |
 *  |                        | |              |
 *  -------------------------  ----------------
 *  ^                          ^
 *  |                          |
 *  ---bytes in this routine ---
 *                           ^                ^
 *                           |                |
 *                      pointer to start of routine
 *****************************************************************************/
/*}}}*/

/*}}}*/
/*{{{  PRIVATE BOOL isbranchop(op)*/
/*****************************************************************************
 *
 *  isbranchop if the opcode 'op' represents a branch instruction.
 *
 *****************************************************************************/
PRIVATE BOOL isbranchop (const INT32 op )
{
  int trop = op & 0xFF;
  return (trop == I_J)/* || (trop == I_CJ)*/; /* Bug 1378 - 11/9/91 removed cj */
}
/*}}}*/
/*{{{  low-level code buffer handling*/
/*{{{  PRIVATE BYTE get_code()*/
/**************************************************************************
 *
 * get_code returns the next byte from the code buffer.
 * in the code buffer the bytes are stored in order of host endianness.
 * this function delivers bytes in order of target endianness.
 *          The current position in the code buffer is represented by
 *          (getptr, getword, bytesleft).
 *
 **************************************************************************/
PRIVATE BYTE get_code ( void )
{
  BYTE r;
  if (bytesleft == 0)
    {
      getword = code[getptr++];
      bytesleft = 4;
    }
  r = (BYTE)(getword & 0xff);
  getword >>= 8;
  bytesleft--;
  return(r);
}
/*}}}*/
/*}}}*/
/*{{{  PRIVATE void round_near ()*/
PRIVATE void round_near ()
{
  fpu_mode = FPU_N;
}
/*}}}*/
/*{{{  PRIVATE void gen_set_fround (mode)*/
PRIVATE void gen_set_fround (const int mode)
{
  if (fpu_mode != mode)
    fpu_mode = mode;
}
/*}}}*/
/*{{{  low level DEC alpha instruction packing*/
/*{{{  add_instr_lword*/
PRIVATE void add_instr_lword (const BIT32 lword)
{
  /*{{{  diagnostic symbolic printout of alpha instruction*/
  if (((diagnostics) || (disassemble)) && (!compressing))
  {
    int dumaddress = 0; /* not used */
    int asm_only = FALSE;
    int special;
    const char *ialpha_name, *operands;
    look_at_alpha_instruction (lword, &ialpha_name, &operands, dumaddress,
                       asm_only, &special);
    fprintf (outfile, "\n%8s %s", ialpha_name, operands);
  }
  /*}}}*/
  #if 0
  /*{{{  diagnostic printout of lword in alpha parts*/
  if (diagnostics)
  {
    BYTE opcode = lword >> 26;
    BYTE Ra = (lword >> 21) & 0x1F;
    BYTE Rb = (lword >> 16) & 0x1F;
    BYTE Rc = lword & 0x1F;
    BYTE lit = (lword >> 13) &0xFF;
    BYTE func = (lword >> 5) &0x3F;
    BIT16 disp = lword & 0xFFFF;
    fprintf (outfile, "\n!%8X (%2X %2X %2X %2X %4X)", lword, opcode, Ra, Rb, Rc, disp);
  }
  /*}}}*/
  #endif
  if (start == C_NIL)
    /*{{{  start new code section*/
    {
      start = cptr;
      cptr++;      /* Save word for code length */
    }
    /*}}}*/
  code[cptr++] = lword;
  coden+=4;
  address+=4;
  if (cptr > lptr)
    generr_i(GEN_CODE_BUFFER_FULL, cptr*(BIT32)sizeof(INT32));
}
/*}}}*/
/*{{{  add_data_lword*/
PRIVATE void add_data_lword (const BYTE * dstart, const int offset)
/* builds 4 bytes from start[offset] into a little-endian data lword */
{
  if (start == C_NIL)
    /*{{{  start new code section*/
    {
      start = cptr;
      cptr++;      /* Save word for code length */
    }
    /*}}}*/
  code[cptr++] = (((((dstart[offset+3] << 8) + dstart[offset+2]) << 8) +
                       dstart[offset+1]) << 8) + dstart[offset];
  coden+=4;
  address+=4;
  if (cptr > lptr)
    generr_i(GEN_CODE_BUFFER_FULL, cptr*(BIT32)sizeof(INT32));
}
/*}}}*/
/*{{{  alpha_instr_mem*/
PRIVATE void alpha_instr_mem (const int opcode, const int Ra,
                              const int Rb, const int memory_disp)
{
  const int tophalf = memory_disp & 0xFFFF0000;
  const int bottomhalf = memory_disp & 0xFFFF;
  assert ((tophalf == 0) || (tophalf == 0xFFFF0000));
  assert ((Ra >= 0) && (Rb >= 0));
  add_instr_lword ((((((opcode << 5) | Ra) << 5) | Rb) << 16) | bottomhalf);
}
/*}}}*/
/*{{{  alpha_instr_mem_func*/
PRIVATE void alpha_instr_mem_func (const int opcode, const int function,
                                   const int Ra, const int Rb)
{
  assert ((Ra >= 0) && (Rb >= 0));
  add_instr_lword ((((((opcode << 5) | Ra) << 5) | Rb) << 16) | function);
}
/*}}}*/
/*{{{  alpha_instr_branch*/
PRIVATE void alpha_instr_branch (const int opcode, const int Ra,
                                 const INT32 branch_disp)
{
  assert (Ra >= 0);
  add_instr_lword ((((opcode << 5) | Ra) << 21) | (branch_disp & 0x1FFFFF));
}
/*}}}*/
/*{{{  alpha_instr_operate*/
PRIVATE void alpha_instr_operate (const int opfunc, const int Ra,
                                  const int Rb, const int Rc)
{
  int opcode = opfunc >> 8;
  int func = opfunc & 0x7F;
  assert ((Ra >= 0) && (Rb >= 0) && (Rc >= 0));
  add_instr_lword ((((((((opcode << 5) | Ra) << 5) | Rb) << 11) | func) << 5) | Rc);
}
/*}}}*/
/*{{{  alpha_instr_operate_lit*/
PRIVATE void alpha_instr_operate_lit (const int opfunc, const int Ra,
                                       const int lit, const int Rc)
{
  int opcode = opfunc >> 8;
  int func = opfunc & 0x7F;
  assert ((Ra >= 0) && (Rc >= 0));
  add_instr_lword ((((((((((opcode << 5) | Ra) << 8) | (lit &0xFF)) << 1) | 1)
                                                     << 7) | func) << 5) | Rc);
}
/*}}}*/
/*{{{  alpha_instr_fpoperate*/
PRIVATE void alpha_instr_fpoperate (const int opfunc,
                   const int fa, const int fb, const int fc)
{
  int opcode = opfunc >> 12;
  int func = (opfunc & 0x7FF) ^ (fpu_mode << 6);
  assert ((fa >= 0) && (fb >= 0) && (fc >= 0));
  add_instr_lword ((((((((opcode << 5) | fa) << 5) | fb) << 11) | func) << 5) | fc);
  round_near();
}
/*}}}*/

/*{{{  PRIVATE void add_libstub ()*/
/***************************************************************************
 *
 *  add_libstub adds a zero word as placeholder for library proc call
 *
 **************************************************************************/
PRIVATE void add_libstub ()
{
  add_instr_lword (ZERO32);
}
/*}}}*/
/*}}}*/
/*{{{  input*/
/*{{{  code input*/
/*{{{  set_ts_prim*/
PRIVATE void set_ts_prim (const int instruction)
{
  int ts_diff = tsdiff_prim [instruction>>4];
  old_ts_depth = ts_depth;
  if (ts_diff >= 16)
    ts_depth = ts_diff - 16;
  else
    ts_depth += ts_diff;
  check_tsdepth ();
  set_tregs (ts_depth);
}
/*}}}*/
/*{{{  set_ts_sec*/
PRIVATE void set_ts_sec (const int instruction)
{
  int ts_diff;
  old_ts_depth = ts_depth;
  if (instruction < 0xB0)
  {
    ts_diff = tsdiff_sec [instruction];
    if (ts_diff >= 16)
      ts_depth = ts_diff - 16;
    else
      ts_depth += ts_diff;
    set_tregs (ts_depth);
  }
}
/*}}}*/
/*{{{  non-trivial instructions*/
/*{{{  PRIVATE void gen_ldc (reg, opd)*/
PRIVATE void gen_ldc (const int reg, const INT32 opd)
/* generates most economical load constant code */
/* MUST BE mirrored exactly in jlength */
{
  if (opd == 0)
    alpha_instr_operate (IAO_BIS, REGZERO, REGZERO, reg);
  else if (opd < (MOSTNEG_INT32 + 256))
    alpha_instr_operate_lit (IAO_ADDL, MIN, (opd - MOSTNEG_INT32), reg);
  else if ((opd < 0) && (opd > -256))
    alpha_instr_operate_lit (IAO_SUBL, REGZERO, -opd, reg);
  else
  /*{{{  build it in parts*/
  /* algorithm taken from page A-13 of Alpha Architecture Handbook */
  {
    int lowword = opd & 0xFFFF;
    int highword;
    int extra = 0;
    int tmp1;
    /*{{{  generate code to put opd into reg*/
    alpha_instr_mem (IA_LDA, reg, REGZERO, lowword);
    if ((lowword & 0x8000) != 0)
      lowword = lowword | 0xFFFF0000; /* sign extend */
    #if 0
    /*{{{  book algorithm wrongly coded -- book not clear!*/
    {
      int tmp2, exthigh;
      highword = opd >> 16;
      if ((highword & 0x8000) != 0)
        exthigh = highword | 0xFFFF0000; /* sign extend */
      else
        exthigh = highword;
      tmp1 = opd - lowword;
      tmp2 = tmp1 - (exthigh << 16);
      if (tmp2 != 0)
      {
        extra = 0x4000;
        highword = (tmp1 - 0x40000000) >> 16;
      }
    }
    /*}}}*/
    #else
    tmp1 = opd - lowword;   /* reduce by sign-extended low word */
    if (opd >= 0x7fff8000)
    {
      extra = 0x4000;
      tmp1 -= (extra << 16);
    }
    highword = tmp1 >> 16;
    #endif
    if (extra != 0)
      alpha_instr_mem (IA_LDAH, reg, reg, extra);
    if (highword != 0)
      alpha_instr_mem (IA_LDAH, reg, reg, highword);
    /*}}}*/
  }
  /*}}}*/
}
/*}}}*/
/*{{{  PRIVATE void gen_eqc (reg, opd)*/
PRIVATE void gen_eqc (const int reg, const INT32 opd)
/* replaces reg by 0 if it equals opd, else 1 */
{
  if ((opd < 256) && (opd >= 0))
    alpha_instr_operate_lit (IAO_CMPEQ, reg, opd, reg);
  else
  {
    /*{{{  generate code to put opd into REGSCRATCH*/
    gen_ldc (REGSCRATCH, opd);
    /*}}}*/
    alpha_instr_operate (IAO_CMPEQ, REGSCRATCH, reg, reg);
  }
}
/*}}}*/
/*{{{  PRIVATE void gen_adc (reg, opd)*/
PRIVATE void gen_adc (const int reg, const INT32 opd)
{
  if ((opd < 256) && (opd >= 0))
    alpha_instr_operate_lit (IAO_ADDLV, reg, opd, reg);
  else if ((opd > -256) && (opd < 0))
    alpha_instr_operate_lit (IAO_SUBLV, reg, -opd, reg);
  else
  {
    /*{{{  generate code to add opd into REGSCRATCH*/
    gen_ldc (REGSCRATCH, opd);
    alpha_instr_operate (IAO_ADDLV, REGSCRATCH, reg, reg);
    /*}}}*/
  }
}
/*}}}*/
/*{{{  PRIVATE void gen_load_var (reg, base, offset)*/
PRIVATE void gen_load_var (const int reg, const int base, const INT32 offset)
{
  /* generates one or more instructions according to size of offset */
  if ((offset >= 0x8000) || (offset < (-0x8000)))
  {
    if (reg == base)
    {
      assert (reg != REGSCRATCH);
      gen_ldc (REGSCRATCH, offset);
      alpha_instr_operate (IAO_ADDL, REGSCRATCH, base, REGSCRATCH);
      alpha_instr_mem (IA_LDL, reg, REGSCRATCH, 0);
    }
    else
    {
      gen_ldc (reg, offset);
      if (base != REGZERO)
        alpha_instr_operate (IAO_ADDL, reg, base, reg);
      alpha_instr_mem (IA_LDL, reg, reg, 0);
    }
  }
  else
    alpha_instr_mem (IA_LDL, reg, base, offset);
}
/*}}}*/
/*{{{  PRIVATE void gen_load_addr (reg, base, offset)*/
PRIVATE void gen_load_addr (const int reg, const int base, const INT32 offset)
{
  /* generates one or more instructions according to size of offset */
  if ((offset >= 0x8000) || (offset < (-0x8000)))
  {
    if (reg == base)
    {
      assert (reg != REGSCRATCH);
      gen_ldc (REGSCRATCH, offset);
      alpha_instr_operate (IAO_ADDL, REGSCRATCH, base, reg);
    }
    else
    {
      gen_ldc (reg, offset);
      if (base != REGZERO)
        alpha_instr_operate (IAO_ADDL, reg, base, reg);
    }
  }
  else
    alpha_instr_mem (IA_LDA, reg, base, offset);
}
/*}}}*/
/*{{{  PRIVATE void gen_store_var (reg, base, offset)*/
PRIVATE void gen_store_var (const int reg, const int base, const INT32 offset)
{
  /* generates one or more instructions according to size of offset */
  if ((offset >= 0x8000) || (offset < (-0x8000)))
  {
    assert (reg != REGSCRATCH);
    gen_ldc (REGSCRATCH, offset);
    if (base != REGZERO)
    {
      assert (base != REGSCRATCH);
      alpha_instr_operate (IAO_ADDL, REGSCRATCH, base, REGSCRATCH);
    }
    alpha_instr_mem (IA_STL, reg, REGSCRATCH, 0);
  }
  else
    alpha_instr_mem (IA_STL, reg, base, offset);
}
/*}}}*/
/*{{{  PRIVATE void gen_call_rel (offset, linkreg)*/
PRIVATE void gen_call_rel (const INT32 offset, const int linkreg)
{
  alpha_instr_branch (IA_BSR, linkreg, offset>>WSH);
}
/*}}}*/
/*{{{  PRIVATE gen_intrpt_ovr ()*/
PRIVATE void gen_intrpt_ovr ()
{
}
/*}}}*/
/*{{{  PRIVATE void gen_ldpi ()*/
PRIVATE void gen_ldpi ()
{
  gen_call_rel (0L, REGSCRATCH);
  alpha_instr_operate (IAO_ADDL, REGSCRATCH, RegA, RegA);
  alpha_instr_operate_lit (IAO_ADDL, RegA, 2 << WSH, RegA);
  /* offset of next instr from return addr from CALL above */
}
/*}}}*/
/*{{{  PRIVATE void gen_lend*/
PRIVATE void gen_lend()
{
  /*{{{  decrement count*/
  alpha_instr_mem (IA_LDL, REGSCRATCH, oldregB, 1 << WSH); /* count */
  alpha_instr_operate_lit (IAO_SUBL, REGSCRATCH, 1, REGSCRATCH);
  alpha_instr_mem (IA_STL, REGSCRATCH, oldregB, 1 << WSH); /* count - 1 */
  /*}}}*/
  /*{{{  jump over following instructions if count  = 0*/
  alpha_instr_branch (IA_BEQ, REGSCRATCH, 7); /* cond skip next instrs */
  /*}}}*/
  /*{{{  increment index*/
  alpha_instr_mem (IA_LDL, REGSCRATCH, oldregB, 0); /* index */
  alpha_instr_operate_lit (IAO_ADDL, REGSCRATCH, 1, REGSCRATCH);
  alpha_instr_mem (IA_STL, REGSCRATCH, oldregB, 0); /* index + 1 */
  /*}}}*/
  /*{{{  jump to top of loop*/
  gen_call_rel (0L, REGSCRATCH);
  alpha_instr_operate (IAO_SUBL, REGSCRATCH, oldregA, REGSCRATCH);
  alpha_instr_operate_lit (IAO_ADDL, REGSCRATCH, 3 << WSH, REGSCRATCH);
  alpha_instr_mem (IA_JMP, REGSCRATCH, REGSCRATCH, 0);
  /*}}}*/
}
/*}}}*/
/*{{{  PRIVATE void gen_stconst_regw (value, opd)*/
PRIVATE void gen_stconst_regw (const INT32 value, const int opd)
/* generates code to store a constant value at opd (assumed <128) bytes from EBP */
{
  assert ((opd > -128) && (opd < 128));
  gen_ldc (REGSCRATCH, value);
  alpha_instr_mem (IA_STL, REGSCRATCH, REGW, opd);
}
/*}}}*/
/*{{{  PRIVATE void gen_shiftop (op)*/
PRIVATE void gen_shiftop (const int op)
{
  if (op == IAO_SRL)
    alpha_instr_operate_lit (IAO_ZAPNOT, oldregB, 0xf, oldregB);
  /* note that transputer defines 32 place shift differently from alpha */
  alpha_instr_operate_lit (IAO_CMPEQ, oldregA, 32, REGSCRATCH);
  alpha_instr_operate (op, oldregB, oldregA, RegA);
  alpha_instr_operate (IAO_CMOVNE, REGSCRATCH, REGZERO, RegA);
}
/*}}}*/
/*{{{  PRIVATE void gen_lshiftop (op)*/
PRIVATE void gen_lshiftop (const int op)
{
  alpha_instr_operate_lit (IAO_ZAPNOT, oldregB, 0xf, oldregB);
  /* note that alpha defines 64 place shift differently from transputer */
  alpha_instr_operate_lit (IAO_CMPEQ, oldregA, 64, REGSCRATCH);
  alpha_instr_branch (IA_BEQ, REGSCRATCH, 2); /* cond skip next 2 instrs */
  alpha_instr_operate (IAO_BIS, REGZERO, REGZERO, REGSCRATCH);
  alpha_instr_branch (IA_BR, REGZERO, 3); /* skip next 3 instrs */
  alpha_instr_operate_lit (IAO_SLL, oldregC, 32, oldregC);
  alpha_instr_operate (IAO_BIS, oldregC, oldregB, oldregC);
  alpha_instr_operate (op, oldregC, oldregA, REGSCRATCH);
  alpha_instr_operate_lit (IAO_SRL, REGSCRATCH, 32, RegB);
  alpha_instr_operate_lit (IAO_ZAPNOT, REGSCRATCH, 0xf, RegA);
}
/*}}}*/
/*{{{  PRIVATE void gen_kcall (kentry, tregs_in, tregs_out)*/
PRIVATE void gen_kcall (const int kentry,
                        const int tregs_in, const int tregs_out)
/* uses jump table ending at address held in REGKBASE */
/* Can we assume that the kernel will not corrupt registers? */
/* If we have written it then we know what it does */
/* If it calls C we can assume DEC calling conventions obeyed */
/* Otherwise must preserve them somewhere */
/* CHECK CALLS IF this adds code here */
{
  BOOL needs_postshuffle = FALSE;
  /*{{{  pre-shuffle*/
  assert (old_ts_depth >= tregs_in);
  switch (tregs_in)
  {
    case 1:
    {
      assert (oldregA == REG_0);
      break;
    }
    case 2:
    {
      switch (tregs_out)
      {
        case 0:
          assert (oldregA == REG_1); assert (oldregB == REG_0); break;
        case 1:
          if (old_ts_depth == 3)
          {
            alpha_instr_operate (IAO_BIS, oldregC, oldregC, REGSCRATCH);
            alpha_instr_operate (IAO_BIS, oldregB, oldregB, REG_0);
            alpha_instr_operate (IAO_BIS, oldregA, oldregA, REG_1);
            alpha_instr_operate (IAO_BIS, REGSCRATCH, REGSCRATCH, REG_2);
            needs_postshuffle = TRUE;  break;
          }
        default:
          assert (old_ts_depth == 2); break;
      }
      break;
    }
    default:
      break;
  }
  /*}}}*/
  alpha_instr_operate_lit (IAO_SUBL, REGKBASE, 4*kentry, REGKLINK);
  alpha_instr_mem (IA_JMP, REGKLINK, REGKLINK, 0x4000);
  /*{{{  post-shuffle*/
  if (needs_postshuffle)
  {
    assert (tregs_out == 1); /* no other cases arise yet */
    alpha_instr_operate (IAO_BIS, REG_0, REG_0, RegA);
    alpha_instr_operate (IAO_BIS, REG_2, REG_2, RegB);
  }
  /*}}}*/
}
#define KCALLSIZE 2
/*}}}*/
/*{{{  PRIVATE void gen_ksubcall (kentry)*/
PRIVATE void gen_ksubcall (const int kentry)
/* special case for ENBC, ENBT, DISS, DIV and REM */
{
  /*{{{  pre-load into t10, t11*/
  alpha_instr_operate (IAO_BIS, oldregB, oldregB, t10);
  alpha_instr_operate (IAO_BIS, oldregA, oldregA, t11);
  /*}}}*/
  alpha_instr_operate_lit (IAO_SUBL, REGKBASE, 4*kentry, REGKLINK);
  alpha_instr_mem (IA_JMP, REGKLINK, REGKLINK, 0x4000);
  /*{{{  post-shuffle from t12*/
  alpha_instr_operate (IAO_BIS, t12, t12, RegA);
  if ((ts_depth > 1) && (oldregC != RegB))  /* probably never! */
    alpha_instr_operate (IAO_BIS, oldregC, oldregC, RegB);
  /*}}}*/
}
/*}}}*/
/*{{{  PRIVATE void gen_chk_sngl ()*/
PRIVATE void gen_chk_sngl ()
{
  /* generates code to check that either: */
  /* regB = -1 and regA < 0, or regB = 0 and regA >= 0 */
  /* First compare signs of regB, regA */
  /* then compare regB+1 with 2 (unsigned), */
  /* If the test fails the Kernel entry K_RANGERR is called */
  /* If successful oldregA is copied to RegA (same as oldregB) */

  alpha_instr_operate (IAO_XOR, oldregB, oldregA, REGSCRATCH);
  alpha_instr_branch (IA_BLT, REGSCRATCH, 4); /* cond skip next 4 instrs */

  alpha_instr_operate_lit (IAO_ADDL, oldregB, 1, oldregB);
  alpha_instr_operate_lit (IAO_CMPULE, oldregB, 2, REGSCRATCH);
  alpha_instr_operate (IAO_BIS, oldregA, oldregA, RegA);
  alpha_instr_branch (IA_BNE, REGSCRATCH, KCALLSIZE); /* cond skip next instrs */

  gen_kcall (K_RANGERR, 0, 0);
}
/*}}}*/
/*{{{  PRIVATE void gen_laddsub (op)*/
PRIVATE void gen_laddsub (const int op)
/* op is one of IAO_ADDQ, IAO_ADDLV, IAO_SUBQ or IAO_SUBLV */
{
  switch (op)
  {
    case IAO_ADDLV: break;
    case IAO_SUBLV: break;
    case IAO_ADDQ:
    case IAO_SUBQ:
      alpha_instr_operate_lit (IAO_ZAPNOT, oldregA, 0xf, oldregA);
      alpha_instr_operate_lit (IAO_ZAPNOT, oldregB, 0xf, oldregB);
      break;
  }
  alpha_instr_operate_lit (IAO_AND, oldregC, 1, REGSCRATCH);
  alpha_instr_operate (op, oldregB, oldregA, RegA);
  alpha_instr_operate (op, RegA, REGSCRATCH, RegA);
  switch (op)
  {
    case IAO_ADDLV: break;
    case IAO_SUBLV: break;
    case IAO_ADDQ:
    case IAO_SUBQ:
    {
      alpha_instr_operate_lit (IAO_SRA, RegA, 32, RegB);
      alpha_instr_operate_lit (IAO_AND, RegB, 1, RegB);
      break;
    }
  }
  alpha_instr_operate_lit (IAO_ZAPNOT, RegA, 0xf, RegA);
}
/*}}}*/
/*{{{  PRIVATE void gen_lmul*/
PRIVATE void gen_lmul()
{
  alpha_instr_operate_lit (IAO_ZAPNOT, oldregA, 0xf, oldregA);
  alpha_instr_operate_lit (IAO_ZAPNOT, oldregB, 0xf, oldregB);
  alpha_instr_operate_lit (IAO_ZAPNOT, oldregC, 0xf, oldregC);
  alpha_instr_operate (IAO_MULQ, oldregB, oldregA, REGSCRATCH);
  alpha_instr_operate (IAO_ADDQ, REGSCRATCH, oldregC, REGSCRATCH);
  alpha_instr_operate_lit (IAO_SRL, REGSCRATCH, 32, RegB);
  alpha_instr_operate_lit (IAO_ZAPNOT, REGSCRATCH, 0xf, RegA);
}
/*}}}*/
/*{{{  PRIVATE void gen_fmul*/
PRIVATE void gen_fmul()
{
  alpha_instr_operate (IAO_MULQ, oldregB, oldregA, REGSCRATCH);
  alpha_instr_operate_lit (IAO_SRA, REGSCRATCH, 31, RegA);
  alpha_instr_operate_lit (IAO_SLL, REGSCRATCH, 33, REGSCRATCH);
  alpha_instr_operate (IAO_CMPLT, REGSCRATCH, REGZERO, REGSCRATCH);
  alpha_instr_operate (IAO_ADDL, REGSCRATCH, RegA, RegA);
}
/*}}}*/
/*{{{  PRIVATE void gen_call (opd)*/
PRIVATE void gen_call (const INT32 opd)
{
  /* parameter stacking is done in genbranch */
  /* called code must save stacked Iptr */
  gen_call_rel (opd, REGRETURN);
  undefine_tstack ();
  /* expect a call of gen_func_results if it is a FUNCTION */
}
/*}}}*/
/*{{{  PRIVATE void gen_j (offset)*/
PRIVATE void gen_j (const INT32 offset)
{
  alpha_instr_branch (IA_BR, REGZERO, offset >> 2);
}
/*}}}*/
/*{{{  PRIVATE void gen_cj (offset)*/
PRIVATE void gen_cj (const int areg, const INT32 offset)
{
  alpha_instr_branch (IA_BEQ, areg, offset >> 2);
}
/*}}}*/
/*{{{  PRIVATE void gen_range_chk (lwb)*/
PRIVATE void gen_range_chk (const int lwb)
{
  /* generates code to check, using unsigned comparison */
  /* that oldregB lies between lwb and oldregA - (1 - lwb)    */
  /* If the test fails the Kernel entry K_RANGERR is called */
  /* The value in oldregB is left in regA  */
  if (lwb == 1)
  {
    alpha_instr_operate_lit (IAO_SUBL, oldregB, 1, oldregB);
  }
  else
    assert (lwb == 0);
  alpha_instr_operate (IAO_CMPULE, oldregA, oldregB, oldregA);
  alpha_instr_branch (IA_BEQ, oldregA, KCALLSIZE); /* cond skip next instr */
  gen_kcall (K_RANGERR, 0, 0);
  if (lwb == 1)
  {
    alpha_instr_operate_lit (IAO_ADDL, oldregB, 1, oldregB);
  }
}
/*}}}*/
/*{{{  PRIVATE void gen_checkword ()*/
PRIVATE void gen_checkword ()
/* only used when checkint INT16 */
{
  alpha_instr_operate (IAO_BIS, oldregB, oldregB, REGSCRATCH);
  alpha_instr_operate (IAO_ADDL, oldregA, oldregB, oldregB);
  alpha_instr_operate_lit (IAO_SLL, oldregA, 1, oldregA);
  gen_range_chk (0);
  alpha_instr_operate (IAO_BIS, REGSCRATCH, REGSCRATCH, oldregB);
}
/*}}}*/
/*{{{  PRIVATE void gen_testerr ()*/
PRIVATE void gen_testerr ()
{
  gen_ldc (RegA, 0);       /* temporary ! */
}
/*}}}*/
/*{{{  PRIVATE void gen_cflerr ()*/
PRIVATE void gen_cflerr ()
{
}
/*}}}*/
/*{{{  PRIVATE void gen_enbs ()*/
PRIVATE void gen_enbs ()
{
  alpha_instr_branch (IA_BEQ, oldregA, stzwlen);
  /* cond skip next instructions (load constant + store) */
  gen_stconst_regw (Z_READY, W_STATUS);
}
/*}}}*/
/*}}}*/
/*{{{  PRIVATE int jlength (const int instr, const INT32 opd)*/
PRIVATE int jlength (const int instr, const INT32 opd)
/* returns number of bytes in a "jump" instruction with this operand */
/* allow for the call to K_PAUSE for backward jumps */
/* and the ajw implicit in a call */
{
  if (instr == I_CALL) return 8; /* include ajw,-4 */
  if ((instr == I_J) && (opd < 0L)) return (4 + (4*KCALLSIZE));
  if (instr == I_LDC)
  /*{{{  copy of computation in gen_ldc*/
  {
    if (opd == 0)
      return 4;
    else if (opd < (MOSTNEG_INT32 + 256))
      return 4;
    else if ((opd < 0) && (opd > -256))
      return 4;
    else
    /*{{{  build it in parts*/
    /* algorithm taken from page A-13 of Alpha Architecture Handbook */
    {
      int lowword = opd & 0xFFFF;
      int highword;
      int extra = 0;
      int tmp1;
      int ibytes = 4;
      if ((lowword & 0x8000) != 0)
        lowword = lowword | 0xFFFF0000; /* sign extend */
      #if 0
      /*{{{  book algorithm wrongly coded -- book not clear!*/
      {
        int tmp2, exthigh;
        highword = opd >> 16;
        if ((highword & 0x8000) != 0)
          exthigh = highword | 0xFFFF0000; /* sign extend */
        else
          exthigh = highword;
        tmp1 = opd - lowword;
        tmp2 = tmp1 - (exthigh << 16);
        if (tmp2 != 0)
        {
          extra = 0x4000;
          highword = (tmp1 - 0x40000000) >> 16;
        }
      }
      /*}}}*/
      #else
      tmp1 = opd - lowword;   /* reduce by sign-extended low word */
      if (opd >= 0x7fff8000)
      {
        extra = 0x4000;
        tmp1 -= (extra << 16);
      }
      highword = tmp1 >> 16;
      #endif
      if (extra != 0)
        ibytes += 4;
      if (highword != 0)
        ibytes += 4;
      return ibytes;
    }
    /*}}}*/
  }
  /*}}}*/
  return 4; /* all other cases */
}
/*}}}*/
/*{{{  PRIVATE void code_primary(instruction, operand)*/
/***************************************************************************
 *
 * code_primary adds primary instruction, operand to the code buffer.
 *
 **************************************************************************/
PRIVATE void code_primary (const INT32 instruction, const INT32 operand )
{
  oldregA = RegA;
  oldregB = RegB;
  if ((operand == ZERO32) && ((instruction == I_LDNLP) ||
                              (instruction == I_AJW) ||
                              (instruction == I_ADC)))
    /* Null instruction */
    return;

    set_ts_prim (instruction);
    switch (instruction)
    {
      case I_LDL:  gen_load_var (RegA, REGW, operand << WSH); break;
      case I_LDLP: gen_load_addr (RegA, REGW, operand << WSH); break;
      case I_STL:  gen_store_var (oldregA, REGW, operand << WSH); break;
      case I_LDNL: gen_load_var (RegA, RegA, operand << WSH); break;
      case I_LDNLP:gen_load_addr (RegA, RegA, operand << WSH); break;
      case I_STNL: gen_store_var (oldregB, oldregA, operand << WSH); break;
      case I_EQC:  gen_eqc (RegA, operand); break;
      case I_LDC:  gen_ldc (RegA, operand); break;
      case I_ADC:  gen_adc (RegA, operand); break;
      case I_AJW:  gen_adc (REGW, operand << WSH); break;
      /* the next 3 should only occur via gen_jump_instr */
      case I_J:
      case I_CJ:
      case I_CALL: assert (FALSE);
    }
}
/*}}}*/
/*{{{  PRIVATE void code_secondary(instruction)*/
/***************************************************************************
 *
 * code_secondary adds secondary instruction to the code buffer.
 *
 **************************************************************************/
PRIVATE void code_secondary (const int instruction )
{
  oldregA = RegA;
  oldregB = RegB;
  oldregC = RegC;
  set_ts_sec (instruction);
  switch (instruction)
  {
    case I_FPSQRT:
    case I_FPORDERED:
    case I_XWORD:
    case I_UNPACKSN:
    case I_ROUNDSN:
    case I_POSTNORMSN:
    default:
    {
      genwarn_s(GEN_NOTALPHA, secondaryname (instruction));
      break;
    }
    /*{{{  integer operations*/
    case I_REV:
    {
        alpha_instr_operate (IAO_BIS, RegB, RegB, REGSCRATCH);
        alpha_instr_operate (IAO_BIS, RegA, RegA, RegB);
        alpha_instr_operate (IAO_BIS, REGSCRATCH, REGSCRATCH, RegA);
        break;
    }
    case I_POP:
    {
        if (oldregB != RegA) /* should be never! */
          alpha_instr_operate (IAO_BIS, oldregB, oldregB, RegA);
        break;
    }
    case I_RET:
    {
        gen_load_var (REGRETURN, REGW, 0);
        gen_adc (REGW, 4 << WSH);   /* may use REGSCRATCH */
        alpha_instr_mem (IA_JMP, REGZERO, REGRETURN, 0x8000);
        break;
    }
    case I_LDPI:        gen_ldpi (); break;
    case I_GAJW:
    {
        alpha_instr_operate (IAO_BIS, REGW, REGW, REGSCRATCH);
        alpha_instr_operate (IAO_BIS, RegA, RegA, REGW);
        alpha_instr_operate (IAO_BIS, REGSCRATCH, REGSCRATCH, RegA);
        break;
    }
    case I_GCALL:
    {
        alpha_instr_mem (IA_JMP, REGRETURN, RegA, 0); /* relies on compiler using it this way */
        break;
    }
    case I_MINT:
    {
        alpha_instr_operate (IAO_BIS, MIN, MIN, RegA);
        break;
    }
    case I_LEND:        gen_lend (); break;
    case I_CSUB0:       gen_range_chk (0); break;
    case I_CCNT1:       gen_range_chk (1); break;
    case I_SETERR:      gen_kcall (K_SETERR, 0, 0); break;
    case I_STOPERR:     gen_intrpt_ovr (); break;                   /* ?? */
    case I_TESTERR:     gen_testerr (); break;
    case I_BSUB:
        alpha_instr_operate (IAO_ADDL, oldregB, oldregA, RegA); break;
    case I_WSUB:
        alpha_instr_operate (IAO_S4ADDL, oldregB, oldregA, RegA); break;
    case I_WSUBDB:
        alpha_instr_operate (IAO_S8ADDL, oldregB, oldregA, RegA); break;
    case I_BCNT:
        alpha_instr_operate_lit (IAO_SLL, RegA, WSH, RegA); break;
    case I_WCNT:
    {
        alpha_instr_operate_lit (IAO_SRA, oldregA, WSH, RegA);
        alpha_instr_operate_lit (IAO_BIC, oldregB, WSH-1, RegB);
        break;
    }
    case I_LB:
    {
        alpha_instr_mem (IA_LDQ_U, REGSCRATCH, oldregA, 0);
        alpha_instr_operate (IAO_EXTBL, REGSCRATCH, oldregA, RegA);
        break;
    }
    case I_SB:
    {
        alpha_instr_mem (IA_LDA, REGSCRATCH, oldregA, 0);
        alpha_instr_mem (IA_LDQ_U, oldregA, REGSCRATCH, 0);
        alpha_instr_operate (IAO_INSBL, oldregB, REGSCRATCH, oldregB);
        alpha_instr_operate (IAO_MSKBL, oldregA, REGSCRATCH, oldregA);
        alpha_instr_operate (IAO_BIS, oldregA, oldregB, oldregA);
        alpha_instr_mem (IA_STQ_U, oldregA, REGSCRATCH, 0);
        break;
    }

    case I_AND:     alpha_instr_operate (IAO_AND, oldregB, oldregA, RegA); break;
    case I_OR:      alpha_instr_operate (IAO_BIS, oldregB, oldregA, RegA); break;
    case I_XOR:     alpha_instr_operate (IAO_XOR, oldregB, oldregA, RegA); break;
    case I_NOT:     alpha_instr_operate (IAO_EQV, REGZERO, RegA, RegA); break;
    case I_SHL:     gen_shiftop (IAO_SLL); break;
    case I_SHR:     gen_shiftop (IAO_SRL); break;
    case I_ADD:     alpha_instr_operate (IAO_ADDLV, oldregB, oldregA, RegA); break;
    case I_SUB:     alpha_instr_operate (IAO_SUBLV, oldregB, oldregA, RegA); break;
    case I_MUL:     alpha_instr_operate (IAO_MULLV, oldregB, oldregA, RegA); break;
    case I_GT:      alpha_instr_operate (IAO_CMPLT, oldregA, oldregB, RegA); break;
    case I_DIFF:    alpha_instr_operate (IAO_SUBL, oldregB, oldregA, RegA); break;
    case I_SUM:     alpha_instr_operate (IAO_ADDL, oldregB, oldregA, RegA); break;
    case I_PROD:    alpha_instr_operate (IAO_MULL, oldregB, oldregA, RegA); break;

    case I_XDBLE:    alpha_instr_operate_lit (IAO_ZAPNOT, oldregA, 0xf, RegA);
                     alpha_instr_operate_lit (IAO_SRA, oldregA, 32, RegB); break;
    case I_CWORD:    gen_checkword (); break; /* only called for INT16 */
    case I_CSNGL:    gen_chk_sngl (); break;
    case I_LADD:     gen_laddsub (IAO_ADDLV); break;
    case I_LSUB:     gen_laddsub (IAO_SUBLV); break;
    case I_LSUM:     gen_laddsub (IAO_ADDQ); break;
    case I_LDIFF:    gen_laddsub (IAO_SUBQ); break;
    case I_LMUL:     gen_lmul (); break;
    case I_LSHL:     gen_lshiftop (IAO_SLL); break;
    case I_LSHR:     gen_lshiftop (IAO_SRL); break;
    case I_NORM:
    {
      alpha_instr_operate_lit (IAO_ZAPNOT, oldregA, 0xf, oldregA);
      alpha_instr_operate_lit (IAO_SLL, oldregB, 32, oldregB);
      alpha_instr_operate (IAO_BIS, oldregA, oldregB, REGSCRATCH);
      alpha_instr_operate (IAO_BIS, REGZERO, REGZERO, RegC);
      alpha_instr_operate_lit (IAO_CMOVEQ, REGSCRATCH, 64, RegC);
      alpha_instr_branch (IA_BNE, RegC, 4); /* skip if was all zero */
      alpha_instr_branch (IA_BLT, REGSCRATCH, 3); /* skip if top bit set */
      alpha_instr_operate_lit (IAO_SLL, REGSCRATCH, 1, REGSCRATCH);
      alpha_instr_operate_lit (IAO_ADDL, RegC, 1, RegC);
      alpha_instr_branch (IA_BR, REGZERO, -4); /* loop */
      alpha_instr_operate_lit (IAO_SRL, REGSCRATCH, 32, RegB);
      alpha_instr_operate_lit (IAO_ZAPNOT, REGSCRATCH, 0xf, RegA);
      break;
    }
    /*}}}*/
    /*{{{  calls of gen_kcall for operations implemented by kernel subroutines*/
    case I_MOVE:    gen_kcall (K_MOVE, 3, 0); break;
    case I_DIV:     gen_ksubcall (K_DIV); break;
    case I_REM:     gen_ksubcall (K_REM); break;
    case I_LDIV:    gen_kcall (K_LDIV, 3, 2); break;
    case I_STARTP:  gen_kcall (K_STARTP, 2, 0); break;
    case I_ENDP:    gen_kcall (K_ENDP, 1, 0); break;
    case I_RUNP:    gen_kcall (K_RUNP, 1, 0); break;
    case I_STOPP:   gen_kcall (K_STOPP, 0, 0); break;
    case I_IN:      gen_kcall (K_IN, 3, 0); break;
    case I_OUT:     gen_kcall (K_OUT, 3, 0); break;
    case I_OUTWORD: gen_kcall (K_OUTWORD, 2, 0); break;
    case I_OUTBYTE: gen_kcall (K_OUTBYTE, 2, 0); break;
    case I_ALTWT:   gen_kcall (K_ALTWT, 0, 0); break;
    case I_DISS:    gen_ksubcall (K_DISS); break;
    case I_ENBC:    gen_ksubcall (K_ENBC); break;
    case I_DISC:    gen_kcall (K_DISC, 3, 1); break;
    case I_TIN:     gen_kcall (K_TIN, 1, 0); break;
    case I_TALTWT:  gen_kcall (K_TALTWT, 0, 0); break;
    case I_ENBT:    gen_ksubcall (K_ENBT); break;
    case I_DIST:    gen_kcall (K_DIST, 3, 1); break;
    /*{{{  floating support*/
    /* in case we decide to do these in kernel rather than by occam library fn
    case I_UNPACKSN: gen_kcall (K_UNPACKSN, 2, 3); break;
    case I_ROUNDSN:  gen_kcall (K_ROUNDSN, 3, 1); break;
    case I_POSTNORMSN: gen_kcall (K_POSTNORMSN, 3, 3); break;
    */
    /*}}}*/
    /*}}}*/
    /*{{{  oddments*/
    case I_ALT:      gen_stconst_regw (Z_ENABLING, W_STATUS); break;
    case I_ALTEND:   assert(FALSE); break; /* calls changed to use genaltend */
    case I_ENBS:     gen_enbs (); break;
    case I_LDTIMER:  gen_kcall (K_LDTIMER, 0 , 1); break;
    case I_TALT:     gen_stconst_regw (Z_ENABLING, W_STATUS);
                     gen_stconst_regw (Z_TIMENOTSET, W_TLINK); break;
    case I_LDINF:    gen_ldc (RegA, INFINITY); break;
    case I_CFLERR:   gen_cflerr (); break;
    case I_DUP:
        alpha_instr_operate_lit (IAO_ADDL, oldregA, 0, RegA); break;
    case I_FMUL:     gen_fmul (); break;
    /*}}}*/
    /*{{{  floating point instructions*/
    case I_FPRZ:        gen_set_fround (FPU_Z); break;
    case I_FPRM:        gen_set_fround (FPU_M); break;
    case I_FPRP:        gen_set_fround (FPU_P); break;
    case I_FPRN:        gen_set_fround (FPU_N); break;
    case I_FPCHKI32:    /* NEEDS MORE THOUGHT */ break;
    case I_FPCHKI64:    /* NEEDS MORE THOUGHT */ break;
    case I_FPCHKERR:
    {
      /* uses a quadword below Wptr as scratch memory for FPCR */
      /* OMIT THIS CHECK UNTIL WE UNDERSTAND THINGS BETTER! */
    #if 0
      alpha_instr_mem_func (IA_TRAPB, 0, 0, 0);
      alpha_instr_fpoperate (IAF_MF_FPCR, FREGSCRATCH, FREGSCRATCH, FREGSCRATCH);
      alpha_instr_mem_func (IA_TRAPB, 0, 0, 0);
      alpha_instr_mem (IA_STT, FREGSCRATCH, REGW, -8);
      alpha_instr_mem (IA_LDQ, REGSCRATCH, REGW, -8);
      alpha_instr_operate (IAO_CMPLT, REGSCRATCH, REGZERO, REGSCRATCH);
      alpha_instr_branch (IA_BEQ, REGSCRATCH, KCALLSIZE); /* cond skip next instr */
      gen_kcall (K_SETERR, 0, 0);
    #endif
      break;
    }
    case I_FPTESTERR: break;
    case I_FPREV:
    {
      alpha_instr_fpoperate (IAF_CPYS, FRegB, FRegB, FREGSCRATCH);
      alpha_instr_fpoperate (IAF_CPYS, FRegA, FRegA, FRegB);
      alpha_instr_fpoperate (IAF_CPYS, FREGSCRATCH, FREGSCRATCH, FRegA);
      break;
    }
    case I_FPDUP:
    {
      alpha_instr_fpoperate (IAF_CPYS, FRegA, FRegA, FRegA+1);
      pushfs(); break;
    }
    case I_FPLDNLDB:
    {
      pushfs();
      alpha_instr_mem (IA_LDT, FRegA, oldregA, 0);
      fs_width = 2; break;
    }
    case I_FPLDNLSN:
    {
      pushfs();
      alpha_instr_mem (IA_LDS, FRegA, oldregA, 0);
      fs_width = 1; break;
    }
    case I_FPLDNLADDDB:
    {
      alpha_instr_mem (IA_LDT, FRegA+1, oldregA, 0);
      alpha_instr_fpoperate (IAF_ADDT, FRegA, FRegA+1, FRegA);
      break;
    }
    case I_FPLDNLADDSN:
    {
      alpha_instr_mem (IA_LDS, FRegA+1, oldregA, 0);
      alpha_instr_fpoperate (IAF_ADDS, FRegA, FRegA+1, FRegA);
      break;
    }
    case I_FPLDNLMULDB:
    {
      alpha_instr_mem (IA_LDT, FRegA+1, oldregA, 0);
      alpha_instr_fpoperate (IAF_MULT, FRegA, FRegA+1, FRegA);
      break;
    }
    case I_FPLDNLMULSN:
    {
      alpha_instr_mem (IA_LDS, FRegA+1, oldregA, 0);
      alpha_instr_fpoperate (IAF_MULS, FRegA, FRegA+1, FRegA);
      break;
    }
    case I_FPLDNLDBI:
    {
      pushfs ();
      alpha_instr_operate (IAO_S8ADDL, oldregB, oldregA, oldregA);
      alpha_instr_mem (IA_LDT, FRegA, oldregA, 0); fs_width = 2; break;
    }
    case I_FPLDNLSNI:
    {
      pushfs ();
      alpha_instr_operate (IAO_S4ADDL, oldregB, oldregA, oldregA);
      alpha_instr_mem (IA_LDS, FRegA, oldregA, 0); fs_width = 1; break;
    }
    case I_FPSTNLDB:    alpha_instr_mem (IA_STT, FRegA, oldregA, 0); popfs(); break;
    case I_FPSTNLSN:    alpha_instr_mem (IA_STS, FRegA, oldregA, 0); popfs(); break;
    case I_FPSTNLI32:
    {
      alpha_instr_fpoperate (IAF_CVTTQ, FREGZERO, FRegA, FRegA);
      /* uses a quadword below Wptr as scratch memory */
      alpha_instr_mem (IA_STT, FRegA, REGW, -8); popfs();
      alpha_instr_mem (IA_LDL, REGSCRATCH, REGW, -8);
      alpha_instr_mem (IA_STL, REGSCRATCH, oldregA, 0); break;
    }
    case I_FPLDZEROSN:
    {
      pushfs ();
      alpha_instr_fpoperate (IAF_CPYS, FREGZERO, FREGZERO, FRegA);
      fs_width = 1; break;
    }
    case I_FPLDZERODB:
    {
      pushfs ();
      alpha_instr_fpoperate (IAF_CPYS, FREGZERO, FREGZERO, FRegA);
      fs_width = 2; break;
    }
    case I_FPADD:
    {
      alpha_instr_fpoperate (fs_width == 1 ? IAF_ADDS : IAF_ADDT, FRegB, FRegA, FRegB);
      popfs(); break;
    }
    case I_FPSUB:
    {
      alpha_instr_fpoperate (fs_width == 1 ? IAF_SUBS : IAF_SUBT, FRegB, FRegA, FRegB);
      popfs(); break;
    }
    case I_FPMUL:
    {
      alpha_instr_fpoperate (fs_width == 1 ? IAF_MULS : IAF_MULT, FRegB, FRegA, FRegB);
      popfs(); break;
    }
    case I_FPDIV:
    {
      alpha_instr_fpoperate (fs_width == 1 ? IAF_DIVS : IAF_DIVT, FRegB, FRegA, FRegB);
      popfs(); break;
    }
    case I_FPNAN:
    {
      alpha_instr_mem (IA_STS, FRegA, REGW, -8);
      alpha_instr_mem (IA_LDL, RegA, REGW, -8);
      alpha_instr_mem (IA_LDAH, REGSCRATCH, REGZERO, 0x7f80);
      alpha_instr_operate (IAO_AND, RegA, REGSCRATCH, RegA);
      alpha_instr_operate (IAO_CMPEQ, RegA, REGSCRATCH, RegA);
      alpha_instr_branch (IA_BEQ, RegA, 4); /* already false: cond skip next instrs */
      alpha_instr_mem (IA_LDL, REGSCRATCH, REGW, -8);
      alpha_instr_operate_lit (IAO_SLL, REGSCRATCH, 9, REGSCRATCH);
      alpha_instr_operate_lit (IAO_SRL, REGSCRATCH, 9, REGSCRATCH);
      alpha_instr_operate (IAO_CMOVEQ, REGSCRATCH, REGZERO, RegA); /* eliminate infinities */
      break;
    }
    case I_FPNOTFINITE:
    {
      alpha_instr_mem (IA_STS, FRegA, REGW, -8);
      alpha_instr_mem (IA_LDL, RegA, REGW, -8);
      alpha_instr_mem (IA_LDAH, REGSCRATCH, REGZERO, 0x7f80);
      alpha_instr_operate (IAO_AND, RegA, REGSCRATCH, RegA);
      alpha_instr_operate (IAO_CMPEQ, RegA, REGSCRATCH, RegA);
      break;
    }
    case I_FPGT:
    {
      alpha_instr_fpoperate (IAF_CMPTLE, FRegB, FRegA, FRegB);
      alpha_instr_mem (IA_STS, FRegB, REGW, -8);
      alpha_instr_mem (IA_LDL, RegA, REGW, -8);
      alpha_instr_operate_lit (IAO_CMPEQ, RegA, 0, RegA);
      popfs(); popfs(); break;
    }
    case I_FPEQ:
    {
      alpha_instr_fpoperate (IAF_CMPTEQ, FRegB, FRegA, FRegB);
      alpha_instr_mem (IA_STS, FRegB, REGW, -8);
      alpha_instr_mem (IA_LDL, RegA, REGW, -8);
      alpha_instr_operate_lit (IAO_CMPEQ, RegA, 0, RegA);
      alpha_instr_operate_lit (IAO_XOR, RegA, 1, RegA);
      popfs(); popfs(); break;
    }
    case I_FPI32TOR64:
    {
      pushfs ();
      alpha_instr_mem (IA_LDL, REGSCRATCH, oldregA, 0);
      alpha_instr_mem (IA_STQ, REGSCRATCH, REGW, -8);
      alpha_instr_mem (IA_LDT, FRegA, REGW, -8);
      alpha_instr_fpoperate (IAF_CVTQT, FREGZERO, FRegA, FRegA);
      fs_width = 2; break;
    }
    case I_FPI32TOR32:
    {
      pushfs ();
      alpha_instr_mem (IA_LDL, REGSCRATCH, oldregA, 0);
      alpha_instr_mem (IA_STQ, REGSCRATCH, REGW, -8);
      alpha_instr_mem (IA_LDT, FRegA, REGW, -8);
      alpha_instr_fpoperate (IAF_CVTQS, FREGZERO, FRegA, FRegA);
      fs_width = 1; break;
    }
    case I_FPINT:
    case I_FPRTOI32:
    {
      alpha_instr_fpoperate (IAF_CVTTQ, FREGZERO, FRegA, FRegA);
      alpha_instr_fpoperate (IAF_CVTQT, FREGZERO, FRegA, FRegA); break;
    }
    case I_FPR64TOR32:
    {
      alpha_instr_fpoperate (IAF_CVTTS, FREGZERO, FRegA, FRegA); break;
      fs_width = 1;
    }
    case I_FPREMSTEP: break; /* everything always done in first step */
    case I_FPREMFIRST:
    case I_FPREM:
      alpha_instr_fpoperate (IAF_DIVT, FRegB, FRegA, FREGSCRATCH);
      alpha_instr_fpoperate (IAF_CVTTQ, FREGZERO, FREGSCRATCH, FREGSCRATCH);
      alpha_instr_fpoperate (IAF_CVTQT, FREGZERO, FREGSCRATCH, FREGSCRATCH);
      alpha_instr_fpoperate (IAF_MULT, FREGSCRATCH, FRegA, FREGSCRATCH);
      alpha_instr_fpoperate (IAF_SUBT, FRegB, FREGSCRATCH, FRegB);
      alpha_instr_operate_lit (IAO_BIS, REGZERO, 1, RegA);
      popfs (); break;
    case I_FPEXPDEC32:
      alpha_instr_fpoperate (IAF_MULT, FREG_twom32, FRegA, FRegA); break;
    case I_FPEXPINC32:
      alpha_instr_fpoperate (IAF_MULT, FREG_two32, FRegA, FRegA); break;
    case I_FPMULBY2:
      alpha_instr_fpoperate (IAF_MULT, FREG_two, FRegA, FRegA); break;
    case I_FPDIVBY2:
      alpha_instr_fpoperate (IAF_MULT, FREG_half, FRegA, FRegA); break;
    case I_FPABS:
      alpha_instr_fpoperate (IAF_CPYS, FREGZERO, FRegA, FRegA); break;
    /*}}}*/
  }
}


/*}}}*/
/*}}}*/
/*{{{  labels and tables input*/
/*{{{  support*/
/*{{{  PRIVATE int mark_code(type, size)*/
/***************************************************************************
 *
 * mark_code inserts a record of type type and size size into the code
 *           buffer, and returns the index of the start of the record.
 *
 **************************************************************************/
PRIVATE int mark_code (const int type, const int size )
{
  int ptr;
  if (start != C_NIL)
    /*{{{  close up code segment*/
    {
      if (coden == 0)
        cptr = start;   /* Empty code segment, erase it */
      else
        code[start] = coden; /* Save length of code segment */
      start = C_NIL;
    }
    /*}}}*/
  coden = 0;
  ptr = cptr;
  if ((cptr + size) < lptr)
    /*{{{  create node, add it to adjust node chain*/
    {
      code[cptr] = type;
      /* Add node to adjust node chain */
      if (topl != C_NIL) code[topl + U_NEXT] = cptr;
      topl = cptr;
      /* Reserve space for the node */
      cptr += size;
    }
    /*}}}*/
  else
    generr_i(GEN_CODE_BUFFER_FULL, cptr*(BIT32)sizeof(INT32));
  return(ptr);
}
/*}}}*/
/*{{{  PRIVATE int addlabel(label)*/
/**************************************************************************
 *
 *  addlabel creates a new label record with name 'label', and returns
 *           the index of the new label.
 *
 **************************************************************************/
PRIVATE int addlabel (const int label )
{
  lptr -= LB_SIZE;
  if (lptr > cptr)
    {
      code[lptr + LB_NAME] = label;
      code[lptr + LB_DEF] = C_NIL;
      code[lptr + LB_PRAD] = 0;
    }
  else
  /*generr_i(GEN_LABEL_BUFFER_FULL, (max_code_size - lptr) / LB_SIZE);*/
    generr_i(GEN_CODE_BUFFER_FULL, cptr * (BIT32)sizeof(INT32));
  return(lptr);
}
/*}}}*/
/*{{{  PRIVATE int findlabel(label)*/
/**************************************************************************
 *
 * findlabel searches for label 'label' and if found returns the index
 *           of the label record. If not found, a new record is created
 *           for 'label' and its index is returned.
 *
 **************************************************************************/
PRIVATE int findlabel (const int label )
{
  int l = lptr;
  while ((l < max_code_size) && (code[l + LB_NAME] != label))
    l += LB_SIZE;
  if (l < max_code_size)    /* Label found */
    return(l);
  else
    {                  /* Create new label node */
      return(addlabel(label));
    }
}
/*}}}*/
/*}}}*/
/*{{{  PRIVATE void label_def(label, type)*/
/***************************************************************************
 *
 * label_def inserts a setlab record for label 'label' into the code buffer.
 *
 ***************************************************************************/
PRIVATE void label_def (const int label, const int type )
{
  const int lab = findlabel(label);
  if (code[lab + LB_DEF] != C_NIL)
    geninternal_i(GEN_MULTIPLY_DEFINED_LABEL, label);
  else
    {
      INT32 ptr;
      INT32 tstk;
      ptr = mark_code(type, SL_SIZE);
      note_tstack ();
      memcpy ((char *)&tstk, tstack, 4);  /* or should we use a union? */
      if (disassemble)
        fprintf (outfile, "\nL%d:", label);
      else if (diagnostics)
        fprintf (outfile, "\n%slabel_def %d", COMMENT_INTRO, label);
      code[ptr + SL_ADDR] = address;
      code[ptr + SL_TSTK] = tstk;
      code[lab + LB_DEF] = ptr;
    }
}
/*}}}*/
/*{{{  PRIVATE void section_label_def(label, type)*/
/***************************************************************************
 *
 * section_label_def inserts a setlab record for label 'label'
 *                   into the code buffer, and sets it to
 *                   be the start of the current code section.
 *
 **************************************************************************/
PRIVATE void section_label_def (const int label, const int type )
{
  label_def(label, type);
  section_label = lptr;
}
/*}}}*/
/*{{{  PRIVATE void label_ref(instruction, label, tstk)*/
/***************************************************************************
 *
 * label_ref inserts a uselab record for label 'label' into the code buffer.
 *
 **************************************************************************/
PRIVATE void label_ref (const int instruction, const int label,
                         const int tstk )
{
  const int l = findlabel(label);
  const int forward_reference = (code[l + LB_DEF] == C_NIL);
  const int ptr = mark_code(C_USELAB, U_SIZE);
  int adj;

  adj = 0; /* assume null jump */

  address += adj;
  if (diagnostics)
    {
      fprintf (outfile, "\n%s label_ref %d, I=%X", COMMENT_INTRO, label, instruction);
    }

  code[ptr + U_DEF] = (forward_reference) ? l : code[l + LB_DEF];
  code[ptr + U_ADDR] = address;
  code[ptr + U_OP] = instruction;
  code[ptr + U_ADJ] = adj;
  code[ptr + U_TSTK] = tstk;
}
/*}}}*/
/*{{{  PRIVATE void label_diff(instruction, label1, label2)*/
/***************************************************************************
 *
 * label_diff inserts a uselabdiff record for label 'label'
 * into the code buffer.
 *
 **************************************************************************/
PRIVATE void label_diff (const int instruction, const int label1, const int label2 )
{
  const int ptr = mark_code(C_USELABDIFF, UD_SIZE);
  int lab;

  code[ptr + UD_ADDR] = address;
  code[ptr + UD_ADJ] = 0;
  /*{{{  set UD_DEF1*/
  lab = findlabel(label1);
  if (code[lab + LB_DEF] == C_NIL)
    code[ptr + UD_DEF1] = lab;
  else
    code[ptr + UD_DEF1] = code[lab + LB_DEF];
  /*}}}*/
  /*{{{  set UD_DEF2*/
  lab = findlabel(label2);
  if (code[lab + LB_DEF] == C_NIL)
    code[ptr + UD_DEF2] = lab;
  else
    code[ptr + UD_DEF2] = code[lab + LB_DEF];
  /*}}}*/
  code[ptr + UD_OP] = instruction;
}
/*}}}*/
/*{{{  PRIVATE void table_start()*/
/***************************************************************************
 *
 *  table_start inserts a table record marking the start of a table, into
 *              the code buffer.
 *
 **************************************************************************/
PRIVATE void table_start ( void )
{
  int ptr = mark_code(C_TABLE, T_SIZE);
  code[ptr + T_SCALE] = case_scale;
  undefine_tstack ();
}
/*}}}*/
/*{{{  PRIVATE void table_end()*/
/***************************************************************************
 *
 *  table_end inserts a table record marking the end of a table, into
 *            the code buffer.
 *
 **************************************************************************/
PRIVATE void table_end ( void )
{
  mark_code(C_TABLE, T_SIZE);
}
/*}}}*/
/*{{{  PRIVATE void scale_ref(instruction)*/
/***************************************************************************
 *
 *  scale_ref inserts a scale record for the scale of the forthcoming table
 *            into the code buffer.
 *
 **************************************************************************/
PRIVATE void scale_ref (const int instruction )
{
  case_scale = mark_code(C_SCALE, SC_SIZE);
  code[case_scale + SC_VALUE] = ZERO32;
  code[case_scale + SC_OP] = instruction;
  address += jlength(I_LDC, 1);
}
/*}}}*/
/*}}}*/
/*{{{  debug information input*/
/*{{{  PRIVATE void add_debug_info (type, line)*/
/*****************************************************************************
 *
 *  add_debug_info inserts a debug record into the code buffer,
 *                 'type' indicates the type of debug record, line gives
 *                 the corresponding source file line number.
 *
 *****************************************************************************/
PRIVATE void add_debug_info (const int type , treenode *const line, const int data )
{
  const int ptr = mark_code(type, DB_SIZE);
  code[ptr + DB_ADDR] = address;
  code[ptr + DB_LINE] = (INT32)((PTRINT)line & 0xffffffff);
  code[ptr + DB_DATA] = data;
}
/*}}}*/
/*}}}*/
/*}}}*/
/*{{{  code adjustment*/
/*{{{  variables for adjust*/
PRIVATE INT32 extra;               /* Extra bytes of code added this iteration */
PRIVATE INT32 ajptr;               /* Pointer to current node on adjust chain */
/*}}}*/
/*{{{  routines for adjust*/
/*{{{  PRIVATE int getaddr(ldef)*/
/***************************************************************************
 *
 * getaddr returns the address (offset from beginning of current code
 *         section) of the label ldef.
 *
 **************************************************************************/
PRIVATE INT32 getaddr (const INT32 ldef )
{
  if (ldef >= section_label)
    /*{{{  previous routine*/
    return(section_size + extra + code[ldef + LB_PRAD]);
    /*}}}*/
  else if (ldef >= lptr)
    /*{{{  unresolved forward reference*/
    {
      /* resolve forward reference */
      INT32 l = code[ldef + LB_DEF];                  /* point l to the setlab */
      if (l == C_NIL)
        geninternal_i(GEN_ADJUST_ERROR, code[ldef + LB_NAME]);  /* no setlab */
      if (code[ajptr] == C_USELAB)
        /*{{{  Point the uselab to the setlab*/
        {
          code[ajptr + U_DEF] = l;
          if ((code[l + SL_TSTK] == code[ajptr + U_TSTK]) ||
              (code[l + SL_TSTK] == TSTK_UNDEF) ||
              (code[ajptr + U_TSTK] == TSTK_UNDEF))
            code[l + SL_TSTK] = TSTK_UNDEF;
          else
          {
            fprintf (outfile, "\n%s Warning tstk at def (%d) is %s, should be %s ",
                              COMMENT_INTRO,
                              (int)l,
                              (char *)(&code[l + SL_TSTK]),
                              (char *)(&code[ajptr + U_TSTK]));
            code[l + SL_TSTK] = code[ajptr + U_TSTK];
          }
        }
        /*}}}*/
      else
        /*{{{  point the uselabdiff to the setlab*/
        {
          if (code[ajptr + UD_DEF1] == ldef)
            code[ajptr + UD_DEF1] = l;
          else
            code[ajptr + UD_DEF2] = l;
        }
        /*}}}*/

      return(code[l + SL_ADDR] + extra);
    }
    /*}}}*/
  else if (ldef > ajptr)
    /*{{{  resolved forward reference*/
    return(code[ldef + SL_ADDR] + extra);
    /*}}}*/
  else
    /*{{{  backward reference*/
    return(code[ldef + SL_ADDR]);
    /*}}}*/
}
/*}}}*/
/*}}}*/
/*{{{  PRIVATE void adjust()*/
/***************************************************************************
 *
 * adjust adjusts the label offsets in the current code section to their
 *        optimal size.
 *
 **************************************************************************/
PRIVATE void adjust ( void )
{
  int passno = 0;
  /* Perform code compaction */
  extra = -1;
  section_size = address;
  code[topl + U_NEXT] = C_NIL; /* Mark end of list */
  while (extra != 0)
    /*{{{  iteration of code adjustment*/
    {
      passno++;
      if (diagnostics)
        fprintf(outfile, "\n!Code adjustment, pass %d...", passno);
      ajptr = cstart;
      extra = 0;
      while (ajptr != C_NIL)
        {
          switch(code[ajptr])
            {
              /*{{{  C_SETLAB C_LOCATE C_LIBPATCH C_ADDRESSFIX*/
              case C_SETLAB:
              case C_LOCATE:
              case C_LIBPATCH:
              case C_ADDRESSFIX:
                code[ajptr + SL_ADDR] += extra;
                break;
              /*}}}*/
              /*{{{  C_USELAB*/
              case C_USELAB:
               {
                 INT32 addr, diff;
                 int newadj;
                 INT32 instr;

                 addr = getaddr(code[ajptr + U_DEF]);
                 code[ajptr + U_ADDR] += extra;
                 diff = addr - (code[ajptr + U_ADDR] + code[ajptr + U_ADJ]);
                 instr = code[ajptr + U_OP];
                 if (isbranchop(instr) && (diff == ZERO32))
                   newadj = 0; /* we optimise jump 0 out */
                 else
                   newadj = jlength(instr, diff);
                 if (newadj > code[ajptr + U_ADJ])
                   {
                     if (diagnostics)
                       fprintf(outfile, "\n!USELAB:increasing extra by %d - %ld ",
                                       newadj, (long)code[ajptr + U_ADJ]);
                     extra += newadj - code[ajptr + U_ADJ];
                     code[ajptr + U_ADJ] = newadj;
                   }
               }
               break;
              /*}}}*/
              /*{{{  C_USELABDIFF*/
              case C_USELABDIFF:
                {
                  INT32 addr1, addr2, diff;
                  int newadj;
                  INT32 instr;

                  addr1 = getaddr(code[ajptr + UD_DEF1]);
                  addr2 = getaddr(code[ajptr + UD_DEF2]);
                  diff = addr1 - addr2;
                  code[ajptr + UD_ADDR] += extra;

                  instr = code[ajptr + UD_OP] & 0xff;
                  if (isbranchop(instr) && (diff == ZERO32))
                    newadj = 0; /* we optimise jump 0 out */
                  else
                    newadj = jlength(instr, diff);

                  if (newadj > code[ajptr + UD_ADJ])
                    {
                     if (diagnostics)
                       fprintf(outfile, "\n!USELABDIFF:increasing extra by %d - %ld ",
                                       newadj, (long)code[ajptr + U_ADJ]);
                      extra += newadj - code[ajptr + UD_ADJ];
                      code[ajptr + UD_ADJ] = newadj;
                    }
                }
                break;
              /*}}}*/
              /*{{{  C_TABLE*/
              case C_TABLE:
                {
                  INT32 case_start = ajptr;
                  INT32 case_extra = -1;
                  int max_size = 0;
                  BOOL first = TRUE;

                  while (case_extra != 0)
                    {
                      ajptr = code[case_start + U_NEXT];
                      case_extra = 0;
                      while (code[ajptr] != C_TABLE)
                        /*{{{  visit each uselab record in adjust chain*/
                        {
                          INT32 addr, diff;
                          int newadj;
                          INT32 instr;

                          addr = getaddr(code[ajptr + U_DEF]);
                          if (first)
                            code[ajptr + U_ADDR] += extra + case_extra;
                          else
                            code[ajptr + U_ADDR] += case_extra;
                          diff = addr - (code[ajptr + U_ADDR] + code[ajptr + U_ADJ]);
                          instr = code[ajptr + UD_OP] & 0xff;
                          newadj = jlength(instr, diff);
                          newadj = (int)max_INT32(newadj, max_size);
                          if (newadj > code[ajptr + U_ADJ])
                            {
                              case_extra += newadj - code[ajptr + U_ADJ];
                              code[ajptr + U_ADJ] = newadj;
                            }
                          max_size = (int)max_INT32(max_size, code[ajptr + U_ADJ]);
                          ajptr = code[ajptr + U_NEXT];
                        }
                        /*}}}*/
                      extra += case_extra;
                      first = FALSE;
                    }

                  code[code[case_start + T_SCALE] + SC_VALUE] = max_size;
                }
                break;
              /*}}}*/
              /*{{{  C_SCALE*/
              case C_SCALE:
                /* Scale is set when we go round the associated table */
                break;
              /*}}}*/
              default:
                generr_i(GEN_ADJUST_TAG_ERROR, code[ajptr]);
            }
          ajptr = code[ajptr + U_NEXT];
        }
      /* Move address up the number of bytes we have inserted */
      if (diagnostics)
        fprintf(outfile, " extra = %ld, address = %ld, section_size = %ld",
                           (long)extra, (long)address, (long)section_size);
      address += extra;
      section_size += extra;
    }
    /*}}}*/
  /* assert (passno <= 2); check that multiple passes never needed on alpha */
  if (diagnostics)
    fprintf(outfile, "\n!Adjustment complete.\n");
}
/*}}}*/
/*}}}*/
/*{{{  PRIVATE void gen_jump_instr (instruction, opd, size)*/
PRIVATE void gen_jump_instr (const int instruction, const INT32 opd,
                             const int size_wanted)
{
  INT32 operand = opd;
  int instr = instruction & 0xFF;
  int min_size = jlength (instruction, operand);
  int num_nops = size_wanted - min_size;
  assert (num_nops % 4 == 0);
  /*{{{  generate num_nops NOPs*/
  while (num_nops > 0)
  {
    alpha_instr_operate (IAO_BIS, REGZERO, REGZERO, REGZERO);
    num_nops -= 4;
  }
  /*}}}*/
  /*{{{  generate minimum number of instruction bytes*/
  switch (instr)
  {
    case I_J:    if (operand < 0L)
                 {
                   ts_depth = 0; old_ts_depth = 0;
                   gen_kcall (K_PAUSE, 0, 0);
                 }
                 gen_j (operand); break;
    case I_CALL: gen_adc (REGW, (-4) << WSH);
                 gen_call (operand); break;
    case I_CJ:   {
                   int areg = instruction >> 8;
                   gen_cj (areg, operand);
                   break;
                 }
    default:     fprintf (outfile, "\n!gen_jump bad instr (#%X)", instruction);
                 break;
  }
  /*}}}*/
}
/*}}}*/
/*{{{  PRIVATE void compress_code_buffer()*/
/*{{{  comment*/
/***************************************************************************
 *
 * compress_code_buffer compresses the adjusted code section to the front
 *                      of the code buffer (cstart onwards), and moves
 *                      cstart to point to the first free word.
 *
 **************************************************************************/
/*}}}*/
PRIVATE void compress_code_buffer ( void )
{
  const int topc = mark_code(C_SETLAB, SL_SIZE);
  int cmptr = cstart;                    /* Index into the uncompressed code */
  cptr = cstart;                       /* Where we write the compressed code */
  start = C_NIL;
  /*{{{  update labels above section_label*/
  /* Update labels above section_label, ie. entry points of other code
     sections, so that their LB_PRAD field contains the offset from the
     beginning of the current code section */
  {
    int l = section_label + LB_SIZE;
    while (l < max_code_size)
      {
        code[l + LB_PRAD] += section_size;
        l += LB_SIZE;
      }
  }
  /*}}}*/
  coden = 0;
  /*{{{  patch uselab and uselabdiff operands*/
  while (cmptr < topc)
    /*{{{  patch operands*/
    {
      switch((int)code[cmptr])
        {
          /*{{{  C_SETLAB*/
          case C_SETLAB:
            cmptr += SL_SIZE;
            break;
          /*}}}*/
          /*{{{  C_USELAB*/
          case C_USELAB:
            {
              INT32 addr = code[code[cmptr + U_DEF] + SL_ADDR],
                    diff = addr - (code[cmptr + U_ADDR] + code[cmptr + U_ADJ]);
              code[cmptr + U_DEF] = diff;
              cmptr += U_SIZE;
            }
            break;
          /*}}}*/
          /*{{{  C_USELABDIFF*/
          case C_USELABDIFF:
            {
              INT32 addr1 = code[code[cmptr + UD_DEF1] + SL_ADDR],
                    addr2 = code[code[cmptr + UD_DEF2] + SL_ADDR];
              code[cmptr + UD_DEF1] = addr1 - addr2;
              cmptr += UD_SIZE;
            }
            break;
          /*}}}*/
          /*{{{  C_TABLE*/
          case C_TABLE:
            cmptr += T_SIZE;
            break;
          /*}}}*/
          /*{{{  C_SCALE*/
          case C_SCALE:
            cmptr += SC_SIZE;
            break;
          /*}}}*/
          /*{{{  C_LOCATE C_LIBPATCH C_ADDRESSFIX*/
          case C_LOCATE:
          case C_LIBPATCH:
          case C_ADDRESSFIX:
            cmptr += DB_SIZE;
            break;
          /*}}}*/
          default:   /* code chunk */
            {
              const int len = (int)code[cmptr++];
              int words = len / 4;
              cmptr += words;
            }
            break;
        }
    }
    /*}}}*/
  /*}}}*/
  /*{{{  compress code buffer*/
  cmptr = cstart;
  while (cmptr < topc)
    switch((int)code[cmptr])
      {
        /*{{{  C_SETLAB*/
        case C_SETLAB:
          cmptr += SL_SIZE;
          break;
        /*}}}*/
        /*{{{  C_USELAB*/
        case C_USELAB:
          {
            const INT32 op   = code[cmptr + U_OP];
            const INT32 adj  = code[cmptr + U_ADJ];
            const INT32 diff = code[cmptr + U_DEF];
            /*{{{  a bit of peephole optimisation to remove j 0 and cj 0*/
            if (!isbranchop(op) || (diff != ZERO32) || (adj != ZERO32))
              /* put_pfix(op, diff, adj); */
              gen_jump_instr(op, diff, adj);
            /*}}}*/
            cmptr += U_SIZE;
          }
          break;
        /*}}}*/
        /*{{{  C_USELABDIFF*/
        case C_USELABDIFF:
          {
            /* NB avoid calling code_primary as it affects tstack */
            INT32 diff = code[cmptr + UD_DEF1];
            int instr = (int)code[cmptr + UD_OP] & 0xff;
            int areg = (int)code[cmptr + UD_OP] >> 8;
            assert (instr == I_LDC);
            gen_ldc (areg, diff);
            cmptr += UD_SIZE;
          }
          break;
        /*}}}*/
        /*{{{  C_SCALE*/
        case C_SCALE:
          {
            /* NB avoid calling code_primary as it affects tstack */
            INT32 value = code[cmptr + SC_VALUE];
            int instr = (int)code[cmptr + SC_OP] & 0xff;
            int areg = (int)code[cmptr + SC_OP] >> 8;
            assert (instr == I_LDC);
            gen_ldc (areg, value);
            cmptr += SC_SIZE;
          }
          break;
        /*}}}*/
        /*{{{  C_TABLE*/
        case C_TABLE:
          cmptr += T_SIZE;
          break;
        /*}}}*/
        /*{{{  C_LOCATE*/
        case C_LOCATE:
          if (debugoutput)
            debug_gencodemark((treenode *)((PTRINT) code[cmptr + DB_LINE]),
             (bytes_output + section_size) - code[cmptr + DB_ADDR]);
            /* write_locate_item(rtltag(code[cmptr]), code[cmptr + DB_LINE],
                              code[cmptr + DB_ADDR]); */
          cmptr += DB_SIZE;
          break;
        /*}}}*/
        /*{{{  C_LIBPATCH*/
        case C_LIBPATCH:
          if (debugoutput)
            debug_genlibrpatch((treenode *)((PTRINT)code[cmptr + DB_LINE]),
             (bytes_output + section_size) - code[cmptr + DB_ADDR]);
          cmptr += DB_SIZE;
          break;
        /*}}}*/
        /*{{{  C_ADDRESSFIX*/
        case C_ADDRESSFIX:
          if (debugoutput)
            {
              const int label = (int)code[cmptr + DB_DATA];
              BIT32 offset;
              DEBUG_MSG(("addressfix: label: %d, ", label));
              if (label == (-1)) /* normal addressfix */
                offset = (bytes_output + section_size) - code[cmptr + DB_ADDR];
              else /* bug 1158 - extra stuff */
                {
                #if 0
                  DEBUG_MSG(("findlabel: %d, ",                  findlabel(label)));
                  DEBUG_MSG(("code(DEF): %d, ",             code[findlabel(label)+LB_DEF]));
                  DEBUG_MSG(("code(PRAD): %d\n",            code[findlabel(label)+LB_PRAD]));
                  DEBUG_MSG(("code(SL_ADDR): ld, ",         code[findlabel(label)+SL_ADDR]));
                  DEBUG_MSG(("addr(SL_ADDR): %d, ", getaddr(code[findlabel(label)+SL_ADDR])));
                  DEBUG_MSG(("addr(PRAD): %d\n",    getaddr(code[findlabel(label)+LB_PRAD])));
                  DEBUG_MSG(("section_size: %d, bytes_output: %d\n", section_size, bytes_output));
                  DEBUG_MSG(("Correct: %d, ", (bytes_output + section_size) - code[findlabel(label)+LB_PRAD]));
                #endif
                  /* This is a horrible piece of black magic:
                     What we want is the address (offset from the end of code)
                     of a particular label.
                     After much experimentation, this expression gives the
                     correct result. Blame Martin Day for designing this
                     code cruncher in this way in the first place.
                     CON - 14/2/91
                  */
                  offset = (bytes_output + section_size) -
                           code[findlabel(label)+LB_PRAD];
                }
              DEBUG_MSG(("offset: %ld\n", offset));
              debug_genaddrfix((treenode *)((PTRINT)code[cmptr + DB_LINE]), (long)offset);
            }
          cmptr += DB_SIZE;
          break;
        /*}}}*/
        default:   /* code chunk */
          /*{{{  copy code chunk*/
          {
            int i;
            const int len = (int)code[cmptr++];
            const int nlwords = len >> 2;
            for (i = 0; i < nlwords; i++)
              add_instr_lword (code[cmptr++]);
          }
          break;
          /*}}}*/
      }
  /*}}}*/
  if (start != C_NIL)
    {
      code[cptr++] = start;       /* Append pointer to start of code section */
      code[start] = section_size;              /* Insert code section length */
    }
  cstart = cptr;
}
/*}}}*/
/*{{{  PRIVATE void diag_write_spaces*/
PRIVATE void diag_write_spaces(void)
{
  fputc('\t', outfile);
}
/*}}}*/
/*{{{  PRIVATE void diag_write_nl*/
PRIVATE void diag_write_nl(void)
{
  fputc('\n', outfile);
  diag_write_spaces();
}
/*}}}*/
/*{{{  PRIVATE void diag_write_comment*/
PRIVATE void diag_write_comment(void)
{
  diag_write_spaces();
  fputs(COMMENT_INTRO, outfile);
}
/*}}}*/
/*{{{  PRIVATE void diag_write_nl_comment*/
PRIVATE void diag_write_nl_comment(void)
{
  fputc('\n', outfile);
  fputs(COMMENT_INTRO, outfile);
}
/*}}}*/
/*{{{  PRIVATE void diag_write_nl_string*/
PRIVATE void diag_write_nl_string(const char *str)
{
  diag_write_nl();
  fputs(str, outfile);
}
/*}}}*/
/*{{{  PUBLIC void gencommentn(c, ...)*/
/***************************************************************************
 *
 *  gencommentn writes the printf format string 'c' (with parameters p1, p2,
 *             pn) as a comment to the debugging file.
 *
 **************************************************************************/
/* Write a comment */
PUBLIC void gencomment0 ( const char *const c )
{
  if (disassemble && ((alloc_strategy & ALLOC_NOCOMMENTS) == 0))
    {
      diag_write_comment();
      fputs(c, outfile);
    }
}

/* Write a comment */
PUBLIC void gencomment1 (const char *const c, const BIT32 p1 )
{
  if (disassemble && ((alloc_strategy & ALLOC_NOCOMMENTS) == 0))
    {
      diag_write_comment();
      fprintf (outfile, c, p1);
    }
}

/* Write a comment */
PUBLIC void gencomments (const char *const c, const char * p1 )
{
  if (disassemble && ((alloc_strategy & ALLOC_NOCOMMENTS) == 0))
    {
      diag_write_comment();
      fprintf (outfile, c, p1);
    }
}
/*}}}*/
/*{{{  PRIVATE topcrease*/
PRIVATE void topcrease ()
{
  fputs("\n!{{{  ", outfile);
}
/*}}}*/
/*{{{  PRIVATE bottomcrease*/
PRIVATE void bottomcrease ()
{
  fputs("\n!}}}", outfile);
}
/*}}}*/
/*{{{  PUBLIC markdeadcode*/
PUBLIC void markdeadcode()
{
  dead = !inside_asm;
}
/*}}}*/
/*{{{  PUBLIC genstartconstblock(c)*/
PUBLIC void genstartconstblock (const char *c)
{
  if (assembly_optimise)
  {
    topcrease ();
    fprintf(outfile, "Constant Block %s", c);
  }
  else
    gencomment0 (c);
}
/*}}}*/
/*{{{  PUBLIC genendconstblock*/
PUBLIC void genendconstblock ()
{
  if (assembly_optimise)
    bottomcrease ();
  else
    gencomment0 ("End of constants");
}
/*}}}*/
/*}}}*/

/*{{{  PRIVATE void patch_lib_offsets ()*/
/***************************************************************************
 *
 *  patch_lib_offsets writes the offset of each library call stub into its
 *                    corresponding name node.
 *
 **************************************************************************/
PRIVATE void patch_lib_offsets (void)
{
  treenode *libcall;
  for(libcall = libentrychain; libcall != NULL; libcall = NLEntryNextOf(libcall))
    {
      const int labelpos = findlabel(NPLabelOf(libcall));
      SetNLEntryOffset(libcall, (int)code[labelpos + LB_PRAD]);
    }

  if (configuring)
    code_pseudo_link_fn(libentrychain);
}
/*}}}*/

/*{{{  PRIVATE void display_code_size*/
PRIVATE void display_code_size(const char *const prefix_str)
{
  if (!ims_asm_output)
    fprintf(outfile, "%s!Module code size is %ld bytes\n", prefix_str, (long)total_code_size);
}
/*}}}*/
/*{{{  PRIVATE void output_code_size()*/
/***************************************************************************
 *
 *  output_code_size writes the total code size to the object file
 *
 **************************************************************************/
PRIVATE void output_code_size ( void )
{
  /*INT32 nested_sc_size = sc_size();*/

  if (!configuring && information)
    {
      display_code_size("");
    #if 0
      if (nested_sc_size != ZERO32)
        fprintf(outfile, ", nested SC code size is %ld bytes", (long)nested_sc_size);
      fprintf(outfile, "\n");
    #endif
    }
  /*write_total_code(total_code_size, nested_sc_size);*/
}
/*}}}*/
/*{{{  PRIVATE void calculate_code_size*/
PRIVATE void calculate_code_size(void)
{
  int section_start = cstart;

  total_code_size = 0;
  while (section_start != 0)
    {
      section_start = (int)code[section_start - 1];
      total_code_size += code[section_start];
    }

  /*printf("calculate_code_size: size is %d\n", total_code_size);*/
}
/*}}}*/
/*{{{  PRIVATE BIT32 get_lword ()*/
PRIVATE BIT32 get_lword ()
/* builds a 32-bit instruction from 4 code bytes (least sig first) */
{
  int i;
  int shift = 0;
  BIT32 wordval = 0;
  for (i=0; i<4; i++)
  {
    BYTE b = get_code();
    wordval += ((BIT32)b << shift);
    shift += 8;
  }
  return (wordval);
}
/*}}}*/
/*{{{  PRIVATE void output_code()*/
/**************************************************************************
 *
 * output_code writes the entire code buffer to the object file
 * NOT called if tcoff_without_code is TRUE
 *
 **************************************************************************/
PRIVATE void output_code ( void )
{
  int section_start = cstart;
  INT32 disassembly_address = 0;

  if ((!assembly_output) && (!disassemble))
    write_code_block_start((size_t)total_code_size);

  while (section_start != 0)
    /*{{{  output a code section*/
    {
      INT32 code_length;
      section_start = (int)code[section_start - 1];
      code_length = code[section_start];
      /*{{{  set up for get_code*/
      getptr = section_start + 1; bytesleft = 0;
      code_ptr = (BYTE *)&code[getptr];
      /*}}}*/
      if (assembly_output)
        /*{{{  disassemble code section to outfile*/
        {
          int i = 0;
          fprintf (outfile, "!section at %lX:\n", (long)disassembly_address);
          while (i < code_length)
          {
            BIT32 iword = get_lword ();
            BOOL is_data_word = (special_inst == 1) &&
                ((iword >> 16) != ((IA_STL << 10) | (REGRETURN << 5) | REGW));
            /*{{{  special cases after a RET instruction (and relative jump)*/
            if (is_data_word)
            {
              print_const_word (outfile, iword, 0, disassembly_address + i,
                                 !disassemble);
            }
            /*}}}*/
            else
            {
              special_inst = 0;
              print_instruction_word (outfile, iword, 0, disassembly_address + i,
                                    &special_inst, !disassemble);
            }
            i += 4;
          }
          disassembly_address += code_length;
        }
        /*}}}*/
      else
        if (!disassemble)
        /*{{{  write code section to object file*/
        {
          BIT32 l = code_length;
          DEBUG_MSG(("output_code: writing a section of %ld bytes\n", (long)l));
          while (l > 0)
            {
              BYTE ibuf[IBUF_SIZE];
              size_t len = (size_t)min_INT32(IBUF_SIZE, l);
              int i;
              for (i = 0; i < len; i++)
                ibuf[i] = get_code();  /* this allows for endian-ness */
              write_code_block(len, ibuf);
              l -= len;
            }
        }
        /*}}}*/
    }
    /*}}}*/
}
/*}}}*/

/*{{{  PRIVATE BOOL name_is_visible*/
PRIVATE BOOL name_is_visible(const wordnode *const name, const seginfo_t *rest)
/* Returns TRUE if the name is not descoped by a later externally visible name */
{
  for (; rest != NULL; rest = rest->seg_next)
    if (name == NNameOf(rest->seg_name)) return FALSE;
  return TRUE;
}
/*}}}*/
/*{{{  PRIVATE INT32 offset_from_label*/
PRIVATE INT32 offset_from_label(const int label)
{
  if (label == NO_LABEL)
    return 0;
  return code[findlabel(label) + LB_PRAD];
}
/*}}}*/
/*{{{  PRIVATE void output_entry_points*/
/*{{{  comment*/
/***************************************************************************
 *
 *  output_entry_points writes all the declared entry points into the
 *                      object file. Must be called after output_code, as
 *                      the section labels must have been adjusted for
 *                      word alignment.
 *
 **************************************************************************/
/*}}}*/
PRIVATE void output_entry_points (void)
{
  seginfo_t *seg;

  /* first reverse the list back into the correct order */
  {
    seginfo_t *p = seginfo, *reverser = NULL, *temp;
    while (p != NULL)
      {
        temp        = p->seg_next;
        p->seg_next = reverser;
        reverser    = p;
        p           = temp;
      }
    seginfo = reverser;
  }

  for (seg = seginfo; seg != NULL; seg = seg->seg_next)
    {
      treenode *const n  = seg->seg_name;
      /*const INT32 offset = code[seg->seg_entry + LB_PRAD];*/
      /*const INT32 offset = code[findlabel(NPLabelOf(n)) + LB_PRAD];*/
      const INT32 offset = offset_from_label(NPLabelOf(n));
      /*assertion(section_label == findlabel(NPLabelOf(n)));*/
      if (name_is_visible(NNameOf(n), seg->seg_next))
        {
          if (configuring)
            code_save_entrydata_fn(n, offset, total_code_size);
          else if (information)
            fprintf(outfile, "%s %s requires %ld workspace and %ld vectorspace words\n",
                    (TagOf(n) == N_PROCDEF) ? "PROC" : "FUNCTION",
                    WNameOf(NNameOf(n)), (long)NPDatasizeOf(n), (long)NPVSUsageOf(n));

          if (disassemble)
            {
              diag_write_nl_comment();
              fprintf(outfile, "Entry point %s, offset %ld\n",
                      WNameOf(NNameOf(n)), (long)offset);
            }
          else if (!assembly_output)
            write_entry_desc(n, offset);
        }
    }
}
/*}}}*/
/*{{{  PRIVATE void output_mapfile_sectiontable*/
PRIVATE void output_mapfile_sectiontable(void)
{
  seginfo_t *this = mapinfo_list;
  while (this != NULL)
    {
      seginfo_t *const next = this->seg_next;
      treenode  *const nptr = this->seg_name;
      int label;
      const char *type;
      switch(TagOf(nptr))
        {
          /*{{{  data item*/
          case N_VALABBR: case N_VALRETYPE:
            {
              const treenode *const data = consttableof(NDeclOf(nptr));
              if (WLengthOf(CTValOf(data)) == 0)
                label = NO_LABEL;
              else
                label = CTLabelOf(data);
              type = (TagOf(data) == S_STRING) ? "string      " : "table       ";
              /*type = "data        ";*/
            }
            break;
          /*}}}*/
          /*{{{  code item*/
          default:
            label = NPLabelOf(nptr);
            type = "code        ";
            break;
          /*}}}*/
        }
      map_add_entry(mapfile_handle, NULL, WNameOf(NNameOf(nptr)), type,
                    offset_from_label(label), NULL);

      freevec(this, sizeof(seginfo_t));
      this = next;
    }
  map_set_size(mapfile_handle, NULL, total_code_size, 0);
}
/*}}}*/

/*{{{  PUBLIC void compress_code()*/
/***************************************************************************
 *
 * compress_code compresses the current code section down to the front
 *               of the code buffer, and prepares the code buffer for the
 *               input of a new code section.
 *
 **************************************************************************/
PUBLIC void compress_code ( void )
{
  compressing = TRUE;
  adjust();
  compress_code_buffer();
  /*{{{  cut back labels*/
  /* The label at section_label is the entry point which we must keep.
     Save the entry point offset in this label.
     Delete all subsequent labels. */
  code[section_label + LB_DEF] = C_NIL;
  code[section_label + LB_PRAD] = 0;
  lptr = section_label;
  /*}}}*/
  /*{{{  cut back code*/
  bytes_output += section_size;
  cptr = cstart;
  start = C_NIL;
  topl = C_NIL;
  coden = 0;
  address = 0; /* New code section starts at offset 0 */
  /*}}}*/
  compressing = FALSE;
}
/*}}}*/
/*{{{  instruction generation*/
/*{{{  PRIVATE void clean_fpu_flag*/
PRIVATE void clean_fpu_flag(const int instruction)
{
  if (inside_asm) return;

  if (has_fpu_core && (fpu_error_status == ERROR_STATUS_START) &&
      (errormode & ERRORMODE_TIMESLICECHECK))
    {
      /* This can be called with either FPTESTERR or FPUCLRERR,
         depending on whether we're doing localised error handling.
      */
      if (instruction & I_FPU_ENTRY_BIT) /* bug 1206 13/3/91 */
        genfpuentry(instruction);
      else
        gensecondary(instruction);
      fpu_error_status = ERROR_STATUS_CLEAN;
    }
}
/*}}}*/
/*{{{  PRIVATE void dirty_error_flag*/
/*****************************************************************************
*
*  dirty_error_flag is used incase we've never got around to clearing
*                   the error flag in STOP mode.
*
******************************************************************************/
PRIVATE void dirty_error_flag (const BOOL fpu_not_int )
{
  /* We cannot localise handling of the normal error flag, cos
     the TESTERR instruction clobbers the stack;
     For the FPU we can use FPUCLRERR which leaves all stacks alone */

  if (inside_asm) return;

  if (fpu_not_int)
    {
#if 0  /* LOCALISE_ERROR_HANDLING */
      /* This doesn't work, cos FPUCLRERR DOES hit the CPU stack! */
      clean_fpu_flag(I_FPUCLRERR);
#endif
      fpu_error_status = ERROR_STATUS_DIRTY;
    }
  else
    int_error_status = ERROR_STATUS_DIRTY;
}
/*}}}*/
/*{{{  PRIVATE void gensettsdepth (regresults)*/
PRIVATE void gensettsdepth (const int regresults)
/* called after a function instance which leaves results on tstack */
{
  if (assembly_output && ((alloc_strategy & ALLOC_NOCOMMENTS) == 0))
  {
    fprintf (outfile, "\n%s stackdepth =  %d", COMMENT_INTRO, regresults);
  }
  ts_depth += regresults;
  set_tregs (ts_depth);
}
/*}}}*/
/*{{{  PRIVATE void gen_tcall_parms ()*/
PRIVATE void gen_tcall_parms ()
{
  if (oldregC != REG_UNDEF)
  {
    alpha_instr_mem (IA_STL, oldregC, REGW, (-1)<<WSH);
  }
  if (oldregB != REG_UNDEF)
  {
    alpha_instr_mem (IA_STL, oldregB, REGW, (-2)<<WSH);
  }
  if (oldregA != REG_UNDEF)
  {
    alpha_instr_mem (IA_STL, oldregA, REGW, (-3)<<WSH);
  }
}
/*}}}*/
/*{{{  PUBLIC void genprimary (instruction, operand)*/
/*{{{  comment*/
/***************************************************************************
 *
 *  genprimary adds code byte representing primary (instruction, operand)
 *             to the code buffer
 *
 **************************************************************************/
/*}}}*/
PUBLIC void genprimary (const int instruction, const INT32 operand )
{
  if (!dead)
    {
      note_tstack ();
      if (instruction == I_ADC) dirty_error_flag(FALSE);
      code_primary(instruction, operand);
      if (disassemble)
        {
          if ((operand == ZERO32) && ((instruction == I_LDNLP) ||
                                      (instruction == I_AJW) ||
                                      (instruction == I_ADC)))
            /* Null instruction */
            return;
          else
            {
              diag_write_nl();
              #ifdef SHOWTSTACK
              fprintf(outfile, "!%3s %-8s", tstack, primaryname(instruction));
              #else
              fprintf(outfile, "!%-8s", primaryname(instruction));
              #endif
              if ((alloc_strategy & ALLOC_NOOPERANDS) == 0)
                {
                  fprintf(outfile, "%ld", (long)operand);
                  /* display large(ish) numbers in hex too */
                  if (operand > 4000 || operand < -4000)
                    gencomment1("#%lX", operand);
                }
            }
        }
    }
}
/*}}}*/
/*{{{  PUBLIC void genbranch (instruction, label)*/
/***************************************************************************
 *
 *  genbranch adds a record (representing a primary 'instruction' with
 *            operand the offset between the next instruction and the
 *            label 'label') to the code buffer.
 *
 **************************************************************************/
PUBLIC void genbranch (const int instruction, const int label )
{
  if (!dead)
    {
      int real_instr = instruction & 0x3fff;  /* I_KEEPT - 1 */
      if (real_instr == I_ADC)
        /* Only happens inside ASM - eg ADC :label, or ADC routine.name */
        dirty_error_flag(FALSE);
      else
        checkerror(); /* INSdi02485 */ /* check before any transfer of control */
      note_tstack ();
      oldregA = RegA;
      oldregB = RegB;
      oldregC = RegC;
      set_ts_prim (real_instr);
      {
        int tstk;
        memcpy ((char *)&tstk, tstack, 4);
        /*{{{  if CJ remember which register to test*/
        if (real_instr == I_CJ)
          real_instr += (oldregA << 8);
        /*}}}*/
        /*{{{  if call put params in workspace*/
        else if (real_instr == I_CALL)
          gen_tcall_parms ();
        /*}}}*/
        label_ref(real_instr, label, tstk);
      }
      /*{{{  debugging*/
      if (disassemble)
        {
          diag_write_nl();
          #ifdef SHOWTSTACK
          fprintf(outfile, "!%3s %-8sL%d", tstack, primaryname(instruction),
                  ((alloc_strategy & ALLOC_NOOPERANDS) == 0) ? label : 0);
          #else
          fprintf(outfile, "!%-8sL%d", primaryname(instruction),
                  ((alloc_strategy & ALLOC_NOOPERANDS) == 0) ? label : 0);
          #endif
        }
      /*}}}*/
      if (instruction == I_J)
      {
        markdeadcode();
      }
    }
}
/*}}}*/
/*{{{  PUBLIC void genjump*/
PUBLIC void genjump(const int label)
/* This is used to insert a jump. If code style flag is set,
   creates a non-timeslicable jump using ldc 0; cj
*/
{
  checkerror(); /* INSdi02485 */ /* check before any transfer of control */
  if ((code_style_flags & CODE_STYLE_CJ_NOT_J) != 0)
    {
      genprimary(I_LDC, 0);
      genbranch(I_CJ, label);
    }
  else
    genbranch(I_J, label);
}
/*}}}*/
/*{{{  PUBLIC void genlabeldiff (instruction, label1, label2)*/
/***************************************************************************
 *
 *  genlabeldiff adds a record (representing a primary 'instruction' with
 *               operand the offset between 'label1' and 'label2')
 *               to the code buffer.
 *
 **************************************************************************/
PUBLIC void genlabeldiff (const int instruction, const int label1, const int label2 )
{
  if (!dead)
    {
      if (instruction == I_ADC) dirty_error_flag(FALSE);
      note_tstack ();
      set_ts_prim (instruction);
      label_diff(instruction + (RegA << 8), label1, label2);
      if (disassemble || diagnostics)
        {
          const BOOL labels = ((alloc_strategy & ALLOC_NOOPERANDS) == 0);
          diag_write_nl();
          #ifdef SHOWTSTACK
          fprintf(outfile, "!%3s %-8sL%d - L%d", tstack,
                           primaryname(instruction),
                           labels ? label1 : 0, labels ? label2 : 0);
          #else
          fprintf(outfile, "!%-8sL%d - L%d",
                           primaryname(instruction),
                           labels ? label1 : 0, labels ? label2 : 0);
          #endif
        }
    }
}
/*}}}*/
/*{{{  PUBLIC void gensecondary (instruction)*/
/***************************************************************************
 *
 *  gensecondary adds the secondary 'instruction' to the code buffer.
 *
 **************************************************************************/
PUBLIC void gensecondary (int instruction )
{
  if (!dead)
    {
      /*{{{  check whether instruction may set error*/
      switch (instruction)
        {
          default:
            break;
          case I_CSUB0: case I_CCNT1: case I_SETERR:
          case I_ADD:   case I_SUB:   case I_MUL:   case I_DIV:  case I_REM:
          case I_CWORD: case I_CSNGL:
          case I_LADD:  case I_LSUB:  case I_LDIV:   case I_FMUL:
          case I_CFLERR:
          case I_FPCHKERR:
          case I_LDCNT: /* variableio 11/3/91 */
          case I_CB: case I_CBU: case I_CIR: case I_CIRU: case I_CS: case I_CSU: /* Added 5/9/91 */
          case I_CAUSEERROR: /* bug TS/1547 07/04/92 */
            dirty_error_flag(FALSE);
            break;

          case I_FPADD: case I_FPSUB: case I_FPMUL: case I_FPDIV: case I_FPREMFIRST:
          case I_FPLDNLADDSN:         case I_FPLDNLADDDB:
          case I_FPLDNLMULSN:         case I_FPLDNLMULDB:
          case I_FPGT:  case I_FPEQ:
          case I_FPINT: case I_FPRTOI32:

          case I_FPSQRT:     case I_FPREM: case I_FPRANGE:
          case I_FPMULBY2:   case I_FPDIVBY2:
          case I_FPEXPINC32: case I_FPEXPDEC32:
          case I_FPABS:
        /*case I_FPSETERR:*/ /* bug TS/1547 12/05/92 */
          case I_FPR32TOR64: case I_FPR64TOR32:
          case I_FPCHKI32:   case I_FPCHKI64:
          case I_FPADDDBSN:
            dirty_error_flag(TRUE);
            break;
        }
      /*}}}*/
      note_tstack ();
      code_secondary(instruction);
      /*{{{  print out the instruction*/
      if (disassemble)
        {
          diag_write_nl();
          if (instruction & I_STEP_BIT)
            {
               fputs("step ", outfile);
               instruction &= ~(I_STEP_BIT);
            }
          #ifdef SHOWTSTACK
          fprintf(outfile, "!%3s %-8s", tstack, secondaryname(instruction));
          #else
          fprintf(outfile, "!%-8s", secondaryname(instruction));
          #endif
        }
      /*}}}*/
      /*{{{  check if we have been timesliced, also for loading FPU values*/
      switch (instruction)
        {
          default:
            break;
          case I_OUTWORD: case I_OUTBYTE: case I_OUT:
          case I_IN: case I_TIN:
          case I_ALTWT: case I_TALTWT:
          case I_VIN: case I_VOUT: /* variableio 3/11/91 */
            genstartblock();
            break;
          /* We rely on the fact that after an FPU load, there is always
             a spare register, so an FPUCLRERR is OK.
             Note that regsfor gives 1 for CONSTEXP, so the loads
             of zeroes are OK too. */
          case I_FPLDNLSN:    case I_FPLDNLDB:
          case I_FPLDNLSNI:   case I_FPLDNLDBI:
          case I_FPLDNLADDSN: case I_FPLDNLMULSN:
          case I_FPLDNLADDDB: case I_FPLDNLMULDB:
          case I_FPLDZEROSN:  case I_FPLDZERODB:
          case I_FPI32TOR32:  case I_FPB32TOR64:
          case I_FPI32TOR64:  /* bug TS/2006 15/12/92 */
            clean_fpu_flag(I_FPUCLRERR);
            break;
        #if 0
          /* NO - we can't blatantly clobber stopp, because we use it in
             PRIPAR and in RESCHEDULE, so we do this in the STOP code */
          case I_STOPP:  /* added for bug 1057 27/11/90 */
            markdeadcode();
            break;
        #endif
          case I_SETERR: /* added for bug 1057 27/11/90 */
            if (errormode & ERRORMODE_HALT) markdeadcode();
            break;
        }
      /*}}}*/
    }
}
/*}}}*/
/*{{{  PUBLIC void setlab (label)*/
/***************************************************************************
 *
 *  setlab adds a record representing the setting of label 'label' to the
 *         code buffer.
 *
 **************************************************************************/
PUBLIC void setlab (const int label )
{
  dead = FALSE;
  label_def(label, C_SETLAB);
}
/*}}}*/
/*{{{  PUBLIC void setsectionlab (label)*/
/***************************************************************************
 *
 *  setsectionlab adds a record representing the setting of label 'label'
 *                to the code buffer, and marks this label as indicating
 *                the start of the current code section (routine, table...)
 *
 **************************************************************************/
PUBLIC void setsectionlab (const int label )
{
  dead = FALSE;
  section_label_def(label, C_SETLAB);
}
/*}}}*/
/*{{{  PUBLIC void commentexp (tptr)*/
/***************************************************************************
 *
 *  commentexp writes the expression 'tptr' as a comment to the debugging
 *             file.
 *
 **************************************************************************/
PUBLIC void commentexp ( treenode *const tptr )
{
  /*if (!source_output)*/ /* removed source_output 14/3/91 */
  /* if (!disassemble) */ /* added to prevent it on the zd option - CON 26/3/91 */
  if (diagnostics)
  {
    if (TagOf(tptr) != S_CONSTEXP)
      switch(nodetypeoftag(TagOf(tptr)))
        {
          case NAMENODE: /*case HIDDENPARAMNODE:*/
            break;
          default:
            diag_write_nl_comment();
            printexp (outfile, tptr);
            break;
        }
  }
}
/*}}}*/
/*{{{  PUBLIC void coder_jumptoentrypoint*/
PUBLIC void coder_jumptoentrypoint(const treenode *const nptr)
{
  if (ims_asm_output)
    {
      diag_write_nl();
      fprintf(outfile, "%-8s$%s", "j", WNameOf(NNameOf(nptr)));
      gencomment0("jump to entrypoint");
    }
}
/*}}}*/
/*{{{  PUBLIC void genfpuentry (int instruction)*/
/*****************************************************************************
 *
 *  genfpuentry generates an fpuentry instruction
 *
 *****************************************************************************/
PUBLIC void genfpuentry (const int instruction )
{
  if (!dead)
    {
      /*{{{  check whether instruction may set error*/
      switch (instruction)
        {
          default:
            break;
          case I_FPUSQRTFIRST:
          case I_FPUMULBY2:   case I_FPUDIVBY2:
          case I_FPUEXPINC32: case I_FPUEXPDEC32:
          case I_FPUABS:
          case I_FPUSETERR:
          case I_FPUR32TOR64: case I_FPUR64TOR32:
          case I_FPUCHKI32: case I_FPUCHKI64:
            dirty_error_flag(TRUE);
            break;
        }
      /*}}}*/
      /*code_primary(I_LDC, instruction & 0xff);*/
      {
        int fpinst = 0;
        switch (instruction)
        {
          case I_FPUSQRTFIRST: fpinst = I_FPSQRT; break;
          case I_FPURZ: fpinst = I_FPRZ; break;
          case I_FPURM: fpinst = I_FPRM; break;
          case I_FPURP: fpinst = I_FPRP; break;
          case I_FPURN: fpinst = I_FPRN; break;
          case I_FPUEXPDEC32: fpinst = I_FPEXPDEC32; break;
          case I_FPUEXPINC32: fpinst = I_FPEXPINC32; break;
          case I_FPUABS: fpinst = I_FPABS; break;
          case I_FPUCHKI64: fpinst = I_FPCHKI64; break;
          case I_FPUCHKI32: fpinst = I_FPCHKI32; break;
          case I_FPUDIVBY2: fpinst = I_FPDIVBY2; break;
          case I_FPUMULBY2: fpinst = I_FPMULBY2; break;
          case I_FPUR64TOR32: fpinst = I_FPR64TOR32; break;
          case I_FPUNOROUND: /* fpinst = I_FPNOROUND; break; */
          case I_FPUSETERR:
          case I_FPUCLRERR:
          case I_FPUSQRTSTEP:
          case I_FPUSQRTLAST:
          case I_FPUR32TOR64:
            fpinst = 0; break;
          default:
            printf ("Trying to generate FPU instr %d\n", instruction);
            assert (FALSE); break;
        }
        if (fpinst != 0)
          code_secondary(fpinst);
      }
      #if 0
      code_primary(I_LDC, instruction & ~I_FPU_ENTRY_BIT);
      code_secondary(I_FPENTRY);
      #endif
      if (disassemble)
        {
          diag_write_nl();
          #ifdef SHOWTSTACK
          fprintf(outfile, "!%3s %-8s", tstack, secondaryname(instruction));
          #else
          fprintf(outfile, "!%-8s", secondaryname(instruction));
          #endif
        }
    }
}
/*}}}*/
/*{{{  PUBLIC void add_const_block(l, p)*/
/***************************************************************************
 *
 *  add_const_block adds l bytes of constants, pointed to by p, to the code
 *                 buffer.
 *  the bytes are assumed to have been cumulated assuming host endianness
 *  and, if necessary are rotated here to the endianness of the target
 *
 **************************************************************************/
PUBLIC void add_const_block ( const int l , const BYTE * const p,
                          treenode *type_tptr, const char *type_string)
{
  /* may need to be rewritten if and when we need qword alignment */
  int oddbytes = l & 3;
  int i;
  for (i = 0; i < l-oddbytes; i+=4)
    add_data_lword (p,i);
  if (oddbytes > 0)
  {
    BYTE lastword [4];
    for (i = 0; i < 4; i++)
      lastword [i] = (i < oddbytes) ? p [l - oddbytes + i] : '\0';
    add_data_lword (&lastword[0], 0);
  }
  if (disassemble && ((alloc_strategy & ALLOC_NOCOMMENTS) == 0))
    /*{{{  print the block to outfile*/
    {
      for (i = 0; i < l; i++)
        {
          if ((i & 0xF) == 0)
            diag_write_nl_string("byte");
          fprintf (outfile, "%s #%.2X", (i & 0xF) == 0 ? "" : ",", p[i]);
        }
      if (oddbytes > 0)
        for (i = oddbytes; i < 4; i++)
          fprintf (outfile, ", #%.2X", 0);
    }
    /*}}}*/
}
/*}}}*/
/*}}}*/
/*{{{  code alignment*/
#define QUADWORD 8
/*{{{  PUBLIC void coder_mark_alignment*/
PUBLIC void coder_mark_alignment(void)
{
  if (ims_asm_output)
    {
      diag_write_nl();
      fputs("align", outfile);
    }
}
/*}}}*/
/*{{{  PUBLIC int get_padlen()*/
PUBLIC int get_padlen(const int len, const int alignment)
{
  if (alignment > 1)
    return (alignment - (len & (alignment-1))) & (alignment-1);
  else
    return 0;
}
/*}}}*/
/*{{{  PRIVATE void subwordalign*/
PRIVATE void subwordalign(const int padlen, const BYTE *const padbytes)
{
  if (padlen > 0)
    {
      add_const_block(padlen, padbytes, NULL, "padding");
      gencomment1("add %d bytes for alignment", padlen);
    }
}
/*}}}*/
/*{{{  PRIVATE void genpadding*/
PRIVATE void genpadding(const int len)
{
  if (len >= QUADWORD) err_abort();
  subwordalign(len, padding);
}
/*}}}*/
/*{{{  PUBLIC int endofsection_padlen*/
PUBLIC int endofsection_padlen(const int alignment)
/* This returns the number of bytes which have to be added to the previous
   section to make it end on that alignment
*/
{
  return get_padlen((int)bytes_output, alignment);
}
/*}}}*/
/*{{{  PUBLIC void alignendofsection()*/
/***************************************************************************
 *
 *  alignendofsection called after outputting a constant section
 *               to ensure that its start will be quad word aligned
 *
 **************************************************************************/
PUBLIC void alignendofsection (const int alignment )
{
  if (diagnostics)
    fprintf (outfile, "\n!alignendofsection, bytes_output = %d, coden = %d",
                            bytes_output, coden);
  genpadding(get_padlen(bytes_output+coden, QUADWORD));
}
/*}}}*/
/*{{{  PUBLIC void endsection_align(wlen)              called from GEN1*/
PUBLIC void endsectionalign(const int wlen)
{
  /* first allow for round up already implicit in word-oriented output */
  /* as everything goes into buffer via add_data_lword or add_instr_lword */
  const int len = (wlen + 3) & (~3);
  const int alignment = QUADWORD;
  const int this_align = get_padlen(len, alignment);
  const int  end_align = endofsection_padlen(alignment);
  if (diagnostics)
    fprintf (outfile, "\n!endsection: wlen = %d, bytes_output = %d, this_align:%d, end_align:%d\n",
                          wlen, bytes_output, this_align, end_align);
  genpadding((end_align + this_align) & (alignment-1));
}
/*}}}*/
/*{{{  PUBLIC void alignwholemodule()*/
PUBLIC void alignwholemodule(void)
/* This is used at the end of the object file to ensure that the whole
   object file is quad word aligned
*/
{
  if (!ims_asm_output)
    {
      const int padlen = endofsection_padlen(QUADWORD);
      if (padlen > 0)
        {
          const int l = newlab();
          if (diagnostics)
            fprintf (outfile, "\n!alignwholemodule");
          setsectionlab(l);
          alpha_instr_operate (IAO_BIS, REGZERO, REGZERO, REGZERO); /* NOP */
          /* subwordalign(padlen, padding);  */
          compress_code();
        }
    }
  coder_mark_alignment();
}
/*}}}*/
/*}}}*/
/*{{{  routines called to pass information to coder*/
/*{{{  PUBLIC void genstartblock()*/
/*****************************************************************************
*
*  genstartblock is used in STOP mode to remember to clear the error flag
*                after a process could have been timesliced by a jump
*                to the start of the block
*
******************************************************************************/
PUBLIC void genstartblock ( void )
{
  /* We cannot localise handling of the normal error flag, cos
     the TESTERR instruction clobbers the stack;
     For the FPU we can use FPUCLRERR which leaves all stacks alone */

  if (inside_asm) return;

  int_error_status = ERROR_STATUS_CLEAN;
  if (errormode & ERRORMODE_TIMESLICECHECK)
    gensecondary(I_TESTERR);

  fpu_error_status = ERROR_STATUS_START;
#if !LOCALISE_ERROR_HANDLING
  clean_fpu_flag(I_FPTESTERR);  /* FPTESTERR is faster than FPUCLRERR */
#endif
}
/*}}}*/
/*{{{  PUBLIC void mark_flag_clean*/
PUBLIC void mark_flag_clean(const BOOL fpu_not_int)
{
  if (fpu_not_int)
    fpu_error_status = ERROR_STATUS_CLEAN;
  else
    int_error_status = ERROR_STATUS_CLEAN;
}
/*}}}*/
/*{{{  PRIVATE seginfo_t *add_to_seginfo_list*/
PRIVATE seginfo_t *add_to_seginfo_list(seginfo_t *const list, treenode *const nptr)
{
  seginfo_t *const seg = (seginfo_t *)newvec(sizeof(seginfo_t));
  seg->seg_next  = list;
  seg->seg_name  = nptr;
  return seg;
}
/*}}}*/
/*{{{  PUBLIC void coder_add_mapfile_entry*/
PUBLIC void coder_add_mapfile_entry(treenode *const nptr)
{
}
/*}}}*/
/*{{{  PUBLIC void coder_add_entry                     called from GEN1*/
PUBLIC void coder_add_entry(treenode *const nptr)
{
  if (diagnostics)
    gencomment0 ("coder add entry");
  if (NLexLevelOf(nptr) == 0)
  {
    /* only lexlevel 0 routines */
    if (ims_asm_output)
      write_asm_desc(outfile, nptr);
    seginfo = add_to_seginfo_list(seginfo, nptr);
  }
  coder_add_mapfile_entry(nptr);         /* all routines */
}
/*}}}*/
/*{{{  PRIVATE void reset_fregs()*/
PRIVATE void reset_fregs()
{
  set_fregs (0);
  fpu_mode = FPU_N;

}
/*}}}*/
/*{{{  PUBLIC noteprocentry ()                         called from GEN1t*/
PUBLIC void noteprocentry (void)
{
  /*{{{  remark-comment*/
  if (assembly_optimise)
    {
      diag_write_nl_comment ();
      fprintf(outfile, "... noteprocentry");
    }
  /*}}}*/
  alpha_instr_mem (IA_STL, REGRETURN, REGW, 0); /* save return link */
  reset_fregs ();
}
/*}}}*/
/*{{{  PUBLIC void notefpresult (type)*/
PUBLIC void notefpresult (const char *type)
/* called after a function instance which leaves results on fpstack */
{
  set_fregs (1);
  if (disassemble && ((alloc_strategy & ALLOC_NOCOMMENTS) == 0))
    fprintf (outfile, "\n%s fpstack =  %s", COMMENT_INTRO, type);
}
/*}}}*/
/*{{{  PUBLIC void coder_return_from_call*/
PUBLIC void coder_return_from_call(const BOOL clean_fpu)
{
  /* this is called after a 'call' instruction to reset any information
     about the error flags.
     Our current model assumes that the integer error flag will always
     be OK.
     If 'clean_fpu' is FALSE, we have to make sure that we clean the
     FPU error flag before we next use it.
  */
  if (!clean_fpu)
    fpu_error_status = ERROR_STATUS_START;
}
/*}}}*/
/*}}}*/
/*{{{  instruction sequence generation*/
/*{{{  PUBLIC void gencontrolsplit (label)*/
PUBLIC void gencontrolsplit (const int label)
{
  /* placeholder for possible need to note that tstack must match on two routes */
  genbranch (I_CJ, label);
}
/*}}}*/
/*{{{  PUBLIC void gencontroljoin (label)*/
PUBLIC void gencontroljoin (const int label)
{
  /* placeholder for possible need to note that tstack must match on two routes */
  setlab (label);
}
/*}}}*/
/*{{{  PRIVATE void gensignext (frombits)*/
PRIVATE void gensignext (const int frombits)
{
  alpha_instr_operate_lit (IAO_SLL, RegA, 64 - frombits, RegA);
  alpha_instr_operate_lit (IAO_SRA, RegA, 64 - frombits, RegA);
}
/*}}}*/
/*{{{  PUBLIC void throw_the_result_away ()*/
PUBLIC void throw_the_result_away (void)
{
  if (ts_depth > 0)
  {
    undefine_tstack ();
    gencomment0 ("throw the result away\n-- stackdepth = 0");
  }
}
/*}}}*/
/*{{{  PUBLIC void throw_nested_result_away ()*/
PUBLIC void throw_nested_result_away (void)
{
  /*{{{  pre-comment*/
  if (assembly_optimise)
    {
      topcrease();
      fprintf(outfile, "throw nested result away");
    }
  /*}}}*/
  note_tstack ();
  /*{{{  print out the (pseudo!) instruction*/
  if (disassemble)
    {
      diag_write_nl ();
      #ifdef SHOWTSTACK
      fprintf(outfile, "!%3s %-8s", tstack, "REVPOP");
      #else
      fprintf(outfile, "!%-8s", "REVPOP");
      #endif
    }
  /*}}}*/
  oldregA = RegA;
  set_ts_sec (I_POP);          /* behaves like POP */
  alpha_instr_operate (IAO_BIS, oldregA, oldregA, RegA);
  /*{{{  post-comment*/
  if (assembly_optimise)
    bottomcrease();
  /*}}}*/
}
/*}}}*/
/*{{{  PUBLIC void checkerror_controlled*/
PUBLIC void checkerror_controlled(const int mode)
/* Checks the error flag, controlled by a supplied 'errormode' flag */
{
  if ((mode & ERRORMODE_NEED_FPCHKERR) &&
      (fpu_error_status == ERROR_STATUS_DIRTY))
    {
      /* This will automagically dirty the other flag */
      gensecondary(I_FPCHKERR);
      fpu_error_status = ERROR_STATUS_CLEAN;
    }
  if ((mode & ERRORMODE_NEED_STOPERR) &&
      (int_error_status == ERROR_STATUS_DIRTY))
    {
      gensecondary(I_STOPERR);
      int_error_status = ERROR_STATUS_CLEAN;
    }
}
/*}}}*/
/*{{{  PUBLIC void checkerror*/
PUBLIC void checkerror ( void )
/* Checks the error flag, controlled by the normal 'errormode' flag */
{
  if (!inside_asm && NEED_ERRORS)
    checkerror_controlled(errormode);
  else if (!inside_assertion)
    {
      if (fpu_error_status == ERROR_STATUS_DIRTY) /* INSdi02121 */
        fpu_error_status = ERROR_STATUS_CLEAN;
      if (int_error_status == ERROR_STATUS_DIRTY) /* INSdi02121 */
        int_error_status = ERROR_STATUS_CLEAN;
    }
}
/*}}}*/
/*{{{  PUBLIC int genconstant(cptr)                    called from GEN1t*/
/***************************************************************************
 *
 *  genconstant adds bytes representing the constant tptr to the code
 *              buffer (least significant byte first). Returns the
 *              number of WORDS written to the code.
 *
 **************************************************************************/
PUBLIC int genconstant ( treenode *tptr )
{
  BYTE v[MAX_CONSTANT_SIZE];
  BIT32 l;
  int i;
  const treenode *ty = gettype(tptr);
  const int w = (int)wordsin(ty);
  /*{{{  put the code bytes representing tptr into v*/
  l = LoValOf(tptr);
  for (i = 0; i < 4; i++)
    {
      v[i] = (BYTE)(l & 0xff);
      l >>= 8;
    }
  l = HiValOf(tptr);
  for (i = 4; i < 8; i++)
    {
      v[i] = (BYTE)(l & 0xff);
      l >>= 8;
    }
  /*}}}*/
  add_const_block(w * bytesperword, v, NULL, tagstring(TagOf(ty))); /* MDP */
  return(w);
}
/*}}}*/
/*{{{  PUBLIC void gencondbranch (jumpcondop, label)   called from GEN7t*/
/***************************************************************************
 *
 *  gencondbranch adds a record (representing a conditional branch with
 *            operand the offset between the next instruction and the
 *            label 'label') to the code buffer.
 *  equivalent to genbranch (I_CJ, label) in the absence of support for
 *            more than one conditional branch test
 *  in the general case jumpcondop may be S_EQ, S_NE, S_GR or S_LE
 *
 **************************************************************************/
PUBLIC void gencondbranch (const int jumpcondop, const int label )
{
  assert (jumpcondop == S_EQ);
  genbranch (I_CJ, label);
}
/*}}}*/
/*{{{  PUBLIC void genloadlabptr (thatlab, thislab,text)     called from GEN1t,11t*/
PUBLIC void genloadlabptr (const int thatlab, const int thislab, const char * text)
{
  /* text defines what is at thatlab */
  genlabeldiff (I_LDC, thatlab, thislab);
  gensecondary (I_LDPI);
  if (text != NULL)
    gencomment0 (text);
  setlab (thislab);
}
/*}}}*/
/*{{{  PUBLIC void genloopheadtop (wsoffset)      called from GEN1t*/
PUBLIC void genloopheadtop (const int wsoffset)
{
  /*{{{  pre-comment*/
  if (assembly_optimise)
    {
      topcrease ();
      fprintf(outfile, "genloophead (%d)", wsoffset);
    }
  /*}}}*/
}
/*}}}*/
/*{{{  PUBLIC void genloopheadbottom ()           called from GEN1t*/
PUBLIC void genloopheadbottom ()
{
  /*{{{  post-comment*/
  if (assembly_optimise)
    {
      bottomcrease ();
    }
  /*}}}*/
}
/*}}}*/
/*{{{  PRIVATE void gen_inc_local (offset, increment)*/
PRIVATE void gen_inc_local (const int offset, const int increment)
{
  gen_load_var (REGSCRATCH, REGW, offset << WSH);
  if (increment > 0)
    alpha_instr_operate_lit (IAO_ADDL, REGSCRATCH, increment, REGSCRATCH);
  else
    alpha_instr_operate_lit (IAO_SUBL, REGSCRATCH, -increment, REGSCRATCH);
  gen_store_var (REGSCRATCH, REGW, offset << WSH);
}
/*}}}*/
/*{{{  PUBLIC void genloopend (wsoffset, endlabel, looplabel) called from GEN1t*/
PUBLIC void genloopend (const int wsoffset, const int endlabel, const int looplabel)
{
  /*{{{  pre-comment*/
  if (assembly_optimise)
    {
      topcrease();
      fprintf(outfile, "genloopend (%d, %d, %d)", wsoffset, endlabel, looplabel);
    }
  /*}}}*/
  gen_inc_local (wsoffset + REPL_COUNT, -1); /* leaves decremented count in REGSCRATCH */
  label_ref(I_CJ+(REGSCRATCH<<8), endlabel, TSTK_UNDEF);
  gen_inc_local (wsoffset + REPL_BASE, 1);
  label_ref(I_J, looplabel, TSTK_UNDEF);
  setlab(endlabel);                       /* endlabel:                     */
  /*{{{  post-comment*/
  if (assembly_optimise)
    bottomcrease();
  /*}}}*/
}
/*}}}*/
/*{{{  PUBLIC void genaltend (altcount, endlabel)       called from GEN10t*/
PUBLIC void genaltend (const int altcount, const int altendlab)
{
  note_tstack ();
  set_ts_sec (I_ALTEND);
  if (altcount != 0)
  {
    alpha_instr_branch (IA_BSR, REGKLINK, 0);
    setlab(altendlab);
    alpha_instr_mem (IA_LDL, REGSCRATCH, REGW, 0);
    alpha_instr_operate (IAO_ADDL, REGSCRATCH, REGKLINK, REGKLINK);
    alpha_instr_mem (IA_JMP, REGKLINK, REGKLINK, 0);
  }
  else
    setlab(altendlab);
  /*{{{  print out the instruction*/
  if (disassemble)
    {
      diag_write_nl();
      #ifdef SHOWTSTACK
      fprintf(outfile, "!%3s %-8s", tstack, secondaryname(I_ALTEND));
      #else
      fprintf(outfile, "!%-8s", secondaryname(I_ALTEND));
      #endif
    }
  /*}}}*/
}
/*}}}*/
/*{{{  PUBLIC void genprocreturn (tag, regs, ws)       called from GEN1t*/
PUBLIC void genprocreturn (const int tag, const int regs, const int ws)
{
  set_ts_sec (I_RET);
  dead = FALSE;     /* must generate even if preceding code is dead */
                    /* this is to facilitate sensible disassembly only! */
  /*{{{  PROC*/
  if (tag == S_PROCDEF)
  {
    /*{{{  pre-comment*/
    if (assembly_optimise)
      {
        topcrease();
        fprintf(outfile, "genprocreturn (%d, %d)", regs, ws);
      }
    /*}}}*/
    gen_load_var (REGRETURN, REGW, ws << WSH);
    gen_adc (REGW, (ws + 4)<< WSH);  /* may use REGSCRATCH */
    alpha_instr_mem (IA_JMP, REGZERO, REGRETURN, 0x8000);
    /*{{{  post-comment*/
    if (assembly_optimise)
      bottomcrease ();
    /*}}}*/
  }
  /*}}}*/
  /*{{{  FUNCTION*/
  else
  {
    /*{{{  pre-comment*/
    if (assembly_optimise)
      {
        topcrease ();
        fprintf(outfile, "genfuncreturn (%d, %d)", regs, ws);
      }
    /*}}}*/
    if (old_ts_depth != regs)
      fprintf(outfile, "ts_depth = %d", old_ts_depth);

    /* register results transferred in appropriate registers */
    gen_load_var (REGRETURN, REGW, ws << WSH);
    gen_adc (REGW, (ws + 4)<< WSH);  /* may use REGSCRATCH */
    alpha_instr_mem (IA_JMP, REGZERO, REGRETURN, 0x8000);
    /*{{{  post-comment*/
    if (assembly_optimise)
      bottomcrease ();
    /*}}}*/
  }
  /*}}}*/
  new_occam_line (NULL, FALSE, TRUE, TRUE);
}
/*}}}*/
/*{{{  PUBLIC void gen_func_results (results)          called from GEN4t,11t*/
PUBLIC void gen_func_results (int results)
{
  /*{{{  remark-comment*/
  if (assembly_optimise)
    {
      diag_write_nl_comment();
      fprintf(outfile, "... genfuncresults (%d)", results);
    }
  /*}}}*/
  gensettsdepth (results);
  /* assume they are already in simulated tstack registers */
}
/*}}}*/
/*{{{  PRIVATE void genloadcasescale ()*/
/***************************************************************************
 *
 *  genloadcasescale adds a record representing a primary 'instruction' I_LDC
 *               to the code buffer, whose operand is the size
 *               in bytes of a forthcoming jump table.
 *
 **************************************************************************/
PRIVATE void genloadcasescale ()
{
  /*{{{  remark-comment*/
  if (assembly_optimise)
    {
      diag_write_nl_comment();
      fprintf(outfile, "... genloadcasescale");
    }
  /*}}}*/
  if (!dead)
    {
      note_tstack ();
      set_ts_prim (I_LDC);
      scale_ref(I_LDC + (RegA<<8));
      if (disassemble)
      {
        diag_write_nl ();
        if (showtstack)
          fprintf(outfile, "!%3s ", tstack);
        fprintf(outfile, "!ldc     case.scale");
      }
    }
}
/*}}}*/
/*{{{  PUBLIC void genstartjumptable ()                called from GEN9*/
/***************************************************************************
 *
 *  genstartjumptable adds a record to the code buffer marking the start
 *                    of a jump table.
 *
 **************************************************************************/
PUBLIC void genstartjumptable ( void )
{
  /* assumes regA will contain offset into table of jumps immediately following */
  /*{{{  pre-comment*/
  if (assembly_optimise)
    {
      topcrease ();
      fprintf(outfile, "genstartjumptable");
    }
  /*}}}*/
  if (!dead)
    {
      const int jlab = newlab();
      const int l    = newlab();
      genloadcasescale();                     /*           ldc   casescale       */
      gensecondary(I_PROD);                   /*           prod                  */
      /* now Areg will contain offset into table of jumps immediately following */
      genloadlabptr(jlab, l, "table base");
      gensecondary(I_BSUB);                   /*           bsub                  */
      gensecondary(I_GCALL);                  /*           gcall                 */
      setlab(jlab);                           /* jlab:                           */
      table_start();
    }
  /*{{{  post-comment*/
  if (assembly_optimise)
    bottomcrease ();
  /*}}}*/
}
/*}}}*/
/*{{{  PUBLIC void genjumptableentry (instruction, label)     from GEN9*/
/***************************************************************************
 *
 *  genjumptableentry adds a record representing a jump table entry
 *                    to the code buffer.
 *
 **************************************************************************/
PUBLIC void genjumptableentry (const int instruction, const int label )
{
  if (!dead)
    {
      genbranch(instruction, label);
      dead = FALSE;
    }
}
/*}}}*/
/*{{{  PUBLIC void genendjumptable ()                  called from GEN9*/
/***************************************************************************
 *
 *  genendjumptable adds a record to the code buffer marking the end
 *                  of a jump table.
 *
 **************************************************************************/
PUBLIC void genendjumptable ( void )
{
  /*{{{  remark-comment*/
  if (assembly_optimise)
    {
      diag_write_nl_comment ();
      fprintf(outfile, "... genendjumptable");
    }
  /*}}}*/
  if (!dead)
    {
      table_end();
    }
}
/*}}}*/
/*{{{  PUBLIC void genboolinvert ()             called from GEN5t, 7t*/
/*****************************************************************************
 *
 *  genboolinvert generates code to invert a known boolean
 *
 *****************************************************************************/
PUBLIC void genboolinvert (void)
{
  /*{{{  pre-comment*/
  if (assembly_optimise)
    {
      topcrease ();
      fprintf(outfile, "genboolinvert");
    }
  /*}}}*/
  alpha_instr_operate_lit (IAO_XOR, RegA, 1, RegA);
  /*{{{  post-comment*/
  if (assembly_optimise)
    bottomcrease ();
  /*}}}*/
}
/*}}}*/
/*{{{  PUBLIC void genloadconstant*/
/*****************************************************************************
 *
 *  genloadconstant loads the value of the constant c onto the expression stack
 *
 *****************************************************************************/
PUBLIC void genloadconstant (const INT32 c)
{
  /*{{{  pre-comment*/
  if (assembly_optimise)
    {
      topcrease ();
      fprintf(outfile, "genloadconstant (%ld)", (long)c);
    }
  /*}}}*/
  genprimary(I_LDC, c);
  /*{{{  post-comment*/
  if (assembly_optimise)
    bottomcrease ();
  /*}}}*/
}
/*}}}*/
/*{{{  PUBLIC void genshiftimmediate (op,shiftcount)*/
PUBLIC void genshiftimmediate (const int op, const int shiftcount)
/* generate fast code for ((RegA) << shiftcount) == ((RegA) * 2^shiftcount)
   (where shiftcount is a non-negative constant)
*/
{
  char lr = (op == I_SHR) ? 'r' : 'l';
  /*{{{  pre-comment*/
  if (assembly_optimise)
    {
      topcrease ();
      fprintf(outfile, "gen%cshiftimmediate (%d)", lr, shiftcount);
    }
  /*}}}*/
  if (shiftcount == 0); /* nothing needed */
  else if (shiftcount == 32) /* turn anything into zero */
    alpha_instr_operate (IAO_BIS, REGZERO, REGZERO, RegA);
  else if (op == I_SHR)
    /*{{{  rightshift*/
    {
      alpha_instr_operate_lit (IAO_ZAPNOT, RegA, 0xf, RegA);
      alpha_instr_operate_lit (IAO_SRL, RegA, shiftcount, RegA);
    }
    /*}}}*/
  else
    /*{{{  leftshift = mul by power of 2*/
    {
      alpha_instr_operate_lit (IAO_SLL, RegA, shiftcount, RegA);
    }
    /*}}}*/
  /*{{{  post-comment*/
  if (assembly_optimise)
    bottomcrease ();
  /*}}}*/
}
/*}}}*/
/*{{{  PUBLIC void gennotprocess ()            called from GEN1t*/
/*****************************************************************************
 *
 *  gennotprocess generates ldc,NOT_PROCESS for chan initialisation etc
 *
 *****************************************************************************/
PUBLIC void gennotprocess (void)
{
  /*{{{  remark-comment*/
  if (assembly_optimise)
    {
      diag_write_nl_comment();
      fprintf(outfile, "... gennotprocess");
    }
  /*}}}*/
  genloadconstant(NOT_PROCESS);
}
/*}}}*/
/*{{{  PUBLIC void genfppop ()            called from GEN8t*/
/*****************************************************************************
 *
 *  genfppop is called when a value can usefully be popped off top of FP stack
 *
 *****************************************************************************/
PUBLIC void genfppop(void)
{
}
/*}}}*/
/*{{{  PUBLIC void genwidenshort ()            called from GEN4t*/
/*****************************************************************************
 *
 *  genwidenshort generates code to widen from INT16 to INT32
 *
 *****************************************************************************/
PUBLIC void genwidenshort (void)
{
  gensignext (16);
}
/*}}}*/
/*{{{  PUBLIC void gencompare (relop,flags)    called from GEN4t*/
PUBLIC void gencompare (const int relop, const BOOL flags)
/* generates code to compare regB, regA
   if flags is FALSE result could be left in condition flags, else regA */
{
  const char * opname;
  switch (relop)
  {
    case S_GR: opname = "gt"; break;
    case S_LE: opname = "le"; break;
    case S_NE: opname = "ne"; break;
    default:
    case S_EQ: opname = "eq"; break;
  }
  /*{{{  pre-comment*/
  if (assembly_optimise)
    {
      topcrease ();
      fprintf(outfile, "gencompare (%s,%s)", opname, flags ? "condflags" : "regA");
    }
  /*}}}*/
  oldregA = RegA;
  oldregB = RegB;
  switch (relop)
  {
    case S_GR: gensecondary (I_GT); break;
    case S_LE:
               {
                 set_ts_sec (I_GT);
                 alpha_instr_operate (IAO_CMPLT, oldregB, oldregA, RegA);
                 break;
               }
    case S_EQ:
               {
                 set_ts_sec (I_GT);
                 alpha_instr_operate (IAO_CMPEQ, oldregB, oldregA, RegA);
                 break;
               }
    case S_NE:
               {
                 set_ts_sec (I_GT);
                 alpha_instr_operate (IAO_CMPEQ, oldregB, oldregA, RegA);
                 genboolinvert (); break;
               }
  }
  /*{{{  post-comment*/
  if (assembly_optimise)
    bottomcrease ();
  /*}}}*/
}
/*}}}*/
/*{{{  PUBLIC geni64tor                        called from GEN12t*/
PUBLIC void geni64tor (const int sourcemode, treenode *const source,
               const int desttype, const int roundmode, const int simple)
{
  loadopd(sourcemode, source, 0);              /*  ldp source   */
  if (has_i64tor)
  {
    gensecondary(I_FPLDNLDB);                    /*  load int64 to FregA */
    if (desttype == S_REAL32)
    {
      gencomment0 ("geni64tor32");
      alpha_instr_fpoperate (IAF_CVTQT, FREGZERO, FRegA, FRegA);
      fs_width = 1;
    }
    else
    {
      gencomment0 ("geni64tor64");
      alpha_instr_fpoperate (IAF_CVTQT, FREGZERO, FRegA, FRegA);
      fs_width = 2;
    }
  }
  else
  {
    if (!simple) gensecondary(I_DUP);            /*  dup          */
    gensecondary(I_FPB32TOR64);                  /*  fpb32tor64   */
    if (desttype == S_REAL32 && !has_directfp)
      genfpuentry(I_FPUNOROUND);                 /*  fpunoround   */
    if (simple)
      loadopd(sourcemode, source, 1);            /*  ldp source+1 */
    else
      genprimary(I_LDNLP, 1);                    /*  ldnlp  1     */
    gensecondary(I_FPI32TOR64);                  /*  fpi32or64    */
    if (has_directfp)     gensecondary(I_FPEXPINC32);/*  fpexpinc32   */
    else                  genfpuentry(I_FPUEXPINC32);
    if (desttype == S_REAL32 && !has_directfp)
      genfpuentry(I_FPUNOROUND);                 /*  fpunoround   */
    if (roundmode == S_TRUNC)
      {
        if (has_directfp) gensecondary(I_FPRZ);  /*  fprz         */
        else              genfpuentry(I_FPURZ);
      }
    if (desttype == S_REAL32 && has_directfp)
      gensecondary(I_FPADDDBSN);                 /*  fpadddbsn    */
    else
      gensecondary(I_FPADD);                     /*  fpadd        */
  }
}
/*}}}*/
/*{{{  PUBLIC void gentstop (staddress)*/
PUBLIC void gentstop ( treenode *const staddress )
{
  new_occam_line (staddress, TRUE, TRUE, TRUE);
  /* if (debugoutput)
    coder_genlocate(staddress, TRUE); */
  if (errormode & ERRORMODE_STOP_IS_SETERR)
    {
      if (H_series)
        {
          genloadconstant(ERR_INTERR);
          gensecondary(I_CAUSEERROR);
        }
      else
        gensecondary(I_SETERR);
    }
  if (errormode & ERRORMODE_STOP_IS_STOPP)
    gensecondary (I_STOPP);
  markdeadcode();
}
/*}}}*/
/*{{{  PUBLIC void gen_init_chanarray*/
PUBLIC void gen_init_chanarray (const int pstride, const int cstride, const int elems)
/* generate a loop to initialise arrays of pointers and channels */
/* pointer to array has already been loaded into tstack */
{
  const BIT32 base = ZERO32;  /* Use bottom two words of workspace for */
  const BIT32 count = ONE32;  /* base and count */
  int looplab;
  /*{{{  pre-comment*/
  if (assembly_optimise)
    {
      topcrease ();
      fprintf(outfile, "geninitchanarray (%d, %d, %d)", pstride, cstride, elems);
    }
  /*}}}*/
  genprimary(I_STL, base);              /*          stl  base      */
  gencomment0("base");
  genprimary(I_LDC, elems);             /*          ldc  elements  */
  gencomment0("elements");
  genprimary(I_STL, count);             /*          stl  count     */
  gencomment0("count");
  looplab = newlab();
  setlab(looplab);                      /* looplab:                */
  genprimary(I_LDL, count);             /*          ldl   count     */
  gencomment0("count");
  genprimary(I_ADC, -1);                /*          adc   -1        */
  genprimary(I_STL, count);             /*          stl   count     */
  gencomment0("count");
  if (chanaspointer)
    {
      genprimary(I_LDL, base);          /*          ldl   base      */
      gencomment0("base");
      genprimary(I_LDNLP, elems*pstride);   /*      ldnlp elements */
      gencomment0("elements");
      genprimary(I_LDL, base);          /*          ldl   base      */
      gencomment0("base");
      genprimary(I_STNL, 0);            /*          stnl  0         */
      gencomment0("init channel pointer");
    }
  gennotprocess();                      /*          mint           */
  genprimary(I_LDL, base);              /*          ldl   base      */
  gencomment0("base");
  genprimary(I_STNL, chanaspointer ? elems : 0); /*     stnl  e         */
  gencomment0("init channel word");
  genprimary(I_LDL, base);              /*          ldl   base      */
  gencomment0("base");
  genprimary(I_LDNLP, 1);               /*          ldnlp 1        */
  genprimary(I_STL, base);              /*          stl   base      */
  gencomment0("base");
  genprimary(I_LDL, count);             /*          ldl   count     */
  gencomment0("count");
  genprimary(I_EQC, ZERO32);            /*          eqc   0         */
  genbranch(I_CJ, looplab);             /*          cj    looplab   */
  /*{{{  post-comment*/
  if (assembly_optimise)
    bottomcrease ();
  /*}}}*/
}
/*}}}*/
/*{{{  PUBLIC BOOL genreplicatorcheck (must_check, poslab, zerolab)*/
PUBLIC BOOL genreplicatorcheck (const BOOL must_check,
                                const int poslab, const int zerolab)
/* generates code to jump to poslab if regA is positive, zerolab if it is
   zero, and (if must_check is true) to stop if it is negative.
   returns TRUE iff a jump to bodylab has been generated
   (some implementations might do this if must_check was TRUE) */
{
  /*{{{  pre-comment*/
  if (assembly_optimise)
    {
      topcrease ();
      fprintf(outfile, "genreplicator%s (%d, %d)", must_check ? "check": "skip", poslab, zerolab);
    }
  /*}}}*/
  if (must_check)
    /*{{{  check the replicator (in regA) is > 0*/
    {
      gensecondary(I_MINT);           /*          mint ( = mostpos PLUS 1)*/
      gensecondary(I_CSUB0);          /*          csub0                */
      gencomment0("error if negative");
      checkerror();
    }
    /*}}}*/
  genbranch (I_CJ, zerolab);           /*          cj    zerolab         */
  /*{{{  post-comment*/
  if (assembly_optimise)
    bottomcrease ();
  /*}}}*/
  return FALSE;
}
/*}}}*/
/*{{{  PUBLIC void control_split ()                         called from GEN5t,7t*/
PUBLIC void control_split (void)
{
  /*{{{  remark-comment*/
  if (assembly_optimise)
    {
      diag_write_nl_comment();
      fprintf(outfile, "... controlsplit");
    }
  /*}}}*/
  /* just a placeholder in case need to normalise tstack on control split */
}
/*}}}*/
/*{{{  PUBLIC void control_join ()                          called from GEN5t,7t*/
PUBLIC void control_join (void)
{
  /*{{{  remark-comment*/
  if (assembly_optimise)
    {
      diag_write_nl_comment();
      fprintf(outfile, "... controljoin");
    }
  /*}}}*/
  /* just a placeholder in case need to normalise tstack on control join */
}
/*}}}*/
/*{{{  PUBLIC void gengcall (const BOOL savereturn)*/
void gengcall (const BOOL savedreturn)
{
  if (savedreturn)
  /* in this case called function is assumed to save REGRETURN
     so we must get it from where it already is! */
     alpha_instr_mem (IA_LDL, REGRETURN, REGW, 0);
  gensecondary (I_GCALL);
}
/*}}}*/
/*}}}*/
/*{{{  run time locate information*/
/*{{{  PUBLIC void new_occam_line       from GEN1t,7t,8t,9t,10t*/
/*****************************************************************************
 *
 *  combines coder_genlocate and mark_occam_line
 *
 *****************************************************************************/
PUBLIC void new_occam_line (treenode *laddress,
         BOOL write_linenumber, BOOL needs_locate, BOOL clear_tstack)
{
  if (debugoutput && needs_locate)
    add_debug_info(C_LOCATE, laddress, 0);
  /*{{{  write linenumber*/
  if (disassemble && write_linenumber &&
      ((alloc_strategy & ALLOC_NOLINENUMS) == 0))
    {
      const SOURCEPOSN locn = sourcefileof(LocnOf(laddress));
      if (locn != oldlocn)
      {
        diag_write_nl_comment();
        fprintf(outfile, "\"%s\": %d",
                       lookupfilename(FileNumOf(locn)), FileLineOf(locn));
        DEBUG_MSG((" addr: %lX", (BIT32)laddress));
        oldlocn = locn;
      }
    }
  /*}}}*/
  /*{{{  reset tstack*/
  if (clear_tstack && ts_depth > 0)
  {
    if (diagnostics)
      gencomment1 (" stack depth = %d", ts_depth);
    undefine_tstack ();
  }
  /*}}}*/
}
/*}}}*/
/*{{{  PUBLIC void coder_genaddrfix (addr) - takes treenode address as its param*/
/*****************************************************************************
 *
 *  coder_genaddrfix inserts an addressfix item record into the code buffer
 *
 *****************************************************************************/
PUBLIC void coder_genaddrfix ( treenode *addr, const int label )
{
  add_debug_info(C_ADDRESSFIX, addr, label);
  /*{{{  debugging*/
  if (diagnostics)
    diag_write_nl_string("-- Addressfix");
  /*}}}*/
  /*{{{  source output*/
  /* removed source_output 14/3/91 */
  /*if (source_output)
    output_source_code(get_from_index_table(addr), C_ADDRESSFIX);*/
  /*}}}*/
}
/*}}}*/
/*{{{  PRIVATE void coder_genlibrpatch ()*/
/*****************************************************************************
 *
 *  coder_genlibrpatch inserts a libpatch item record into the code buffer
 *
 *****************************************************************************/
PRIVATE void coder_genlibrpatch ( void )
{
  add_debug_info(C_LIBPATCH, 0, 0);
  /*{{{  debugging*/
  if (diagnostics)
    diag_write_nl_string("-- Libpatch");
  /*}}}*/
}
/*}}}*/
/*}}}*/
/*{{{  library handling*/
/*{{{  PUBLIC void add_to_libentries(nptr)*/
/*****************************************************************************
 *
 *  add_to_libentries adds the library entrypoint 'nptr' to the list of used
 *                    library entry points if it is not aready there.
 *
 *****************************************************************************/
PUBLIC void add_to_libentries ( treenode *nptr )
{
  treenode *l = libentrychain;
  while (l != NULL)
    {
      if (l == nptr)
        return;
      l = NLEntryNextOf(l);
    }
  SetNLEntryNext(nptr, libentrychain);
  libentrychain = nptr;                   /* Add to the entry point list */
}
/*}}}*/
/*{{{  PUBLIC void genlibstubs()*/
/*****************************************************************************
 *
 *  genlibstubs adds a stub for each library call to the code buffer
 *
 *****************************************************************************/
PUBLIC void genlibstubs ( void )
{
  const int stubsize = libpatchsize;
  treenode *libentry;
  for (libentry = libentrychain; libentry != NULL;
       libentry = NLEntryNextOf(libentry))
    {
      const char *name = WNameOf(NNameOf(libentry));
      const int l = newlab();
      SetNPLabel(libentry, l);          /* give the entry point a label */
      setsectionlab(l);
      gencomments("Stub for library call %s", name);
      /*{{{  debugging*/
      if (debugoutput)
        coder_genlibrpatch();
      /*}}}*/
      /*{{{  ims_asm_output*/
      if (ims_asm_output)
        {
          /*fprintf(outfile, "\n$%s:", name);*/
          diag_write_nl();
          fprintf(outfile, "%-8s$%s", "extern", name);
          diag_write_nl();
          fprintf(outfile, "%-8s%d j codefix $%s 0", "patch", stubsize, name);
        }
      /*}}}*/
      add_libstub();
      compress_code();  /* this must be done once for each library stub */
    }
}
/*}}}*/
/*}}}*/

/*{{{  PUBLIC void close_object_file()*/
/*{{{  comment*/
/*****************************************************************************
*
*  close the object file - called by the harness
*
******************************************************************************/
/*}}}*/
PUBLIC void close_object_file (FILE *fptr, const char *const filename)
{
  if (fclose(fptr) == EOF)
  /*generr_s(GEN_CANNOT_CLOSE_OBJECT_FILE, filename);*/
    msg_out_s(SEV_FATAL, GEN, GEN_CANNOT_CLOSE_OBJECT_FILE, NOPOSN, filename);
}
/*}}}*/
/*{{{  PUBLIC FILE *open_object_file()*/
/*****************************************************************************
 *
 * open_object_file opens the object file for writing
 *
 *****************************************************************************/
PUBLIC FILE *open_object_file (const char *const filename)
{
  FILE *const fptr = fopen(filename, "wb");
  if (fptr == NULL)
  /*generr_s(GEN_CANNOT_OPEN_OBJECT_FILE, filename);*/
    msg_out_s(SEV_FATAL, GEN, GEN_CANNOT_OPEN_OBJECT_FILE, NOPOSN, filename);
  return fptr;
}
/*}}}*/
/*{{{  PUBLIC void write_object_file()*/
/*****************************************************************************
 *
 * write_object_file writes out the object file and closes it
 *
 *****************************************************************************/
PUBLIC void write_object_file ( void )
{
  calculate_code_size();

  if (assembly_output && ((alloc_strategy & ALLOC_NOCOMMENTS) == 0))
    display_code_size(diagnostics ? "" : "\n");

  patch_lib_offsets();
  if (tcoff_without_code || (!assembly_output))
    {
      if (!tcoff_without_code) output_code();

      if (!disassemble)
      {
        if (debugoutput) flush_debug_buffers();

        write_library_calls(TRUE, libentrychain, I_J);
        output_entry_points();
        output_code_size();
        output_mapfile_sectiontable();

        if (configuring)
          code_link_map_fn();  /* write the linker map into the file */

        write_end_module ();
      }
    }
  else if (assembly_output)
    {
      output_code(); /* disassemble code buffer contents to outfile */
      output_entry_points();
    }
}
/*}}}*/

/*{{{  PRIVATE void initgenalpha()*/
PRIVATE void initgenalpha()
{
  reset_fregs ();
}
/*}}}*/
/*{{{  PUBLIC void initcode()*/
/***************************************************************************
 *
 * initcode initialises the coder.
 *
 **************************************************************************/
PUBLIC void initcode ( void )
/***************************************************************************
 *
 * initcode initialises the coder.
 *
 **************************************************************************/
{
  if (req_code_size != 0)
  {
    /* we multiply by 1126 (1024 + 10%) so that in effect we add 10%
       to the buffer size.
       (requested in K, so multiply by 1024 * 1.1)
       This allows for the other crap which is used in the code buffer */
    /* bug 1021 - 16/10/90 */
      max_code_size = req_code_size * (1126 / sizeof(INT32));
  }
  else
    max_code_size = configuring ? CONFIGURING_CODE_SIZE : COMPILING_CODE_SIZE;

  if (code == NULL) code = (INT32 *)memalloc(max_code_size * sizeof(INT32));
  cstart = 0;
  cptr = 0;
  lptr = max_code_size;
  start = C_NIL;
  topl  = C_NIL;
  coden = 0;
  address  = 0;
  case_scale = 0;
  seginfo = NULL;
  bytes_output = ZERO32;
  /* debugnamesptr = 0; */
  libentrychain = NULL;
  dead = FALSE;
  int_error_status = ERROR_STATUS_CLEAN;
  fpu_error_status = ERROR_STATUS_FP_INIT;
  initgenalpha();
  undefine_tstack ();
  special_inst = 0;
  compressing = FALSE;
  oldlocn = NOPOSN;
  assembly_optimise = diagnostics;
  stzwlen = 1 + (jlength(I_LDC, Z_READY)>>WSH);
  #ifdef SHOWTSTACK
  showtstack = TRUE;
  #else
  showtstack = FALSE;
  #endif
}

/*}}}*/
/*{{{  PUBLIC void freecode*/
/* This is only called if configuring is TRUE */
PUBLIC void freecode(void)
{
  if (code != NULL)
    {
      memfree(code);
      code = NULL;
    }
}
/*}}}*/
/*{{{  PUBLIC  int genproftable(INT32 size)              dummy*/
/*****************************************************************************
 *
 *  genproftable creates and initializes the profile table in the code buffer
 *
 *****************************************************************************/
PUBLIC  int genproftable(INT32 proftsize)
{
  /* assert (!(sampling_profiling || line_profiling || cgraph_profiling )); */
  return -1;
}
/*}}}*/

