/* $Id: gen10.c,v 1.5 1997/06/06 14:22:51 mdp2 Exp $ */

/*
 *	code generator - channel input and output generation
 *	Copyright (C) 1987 Inmos Limited
 *
 *	This program is free software; you can redistribute it and/or modify
 *	it under the terms of the GNU General Public License as published by
 *	the Free Software Foundation; either version 2 of the License, or
 *	(at your option) any later version.
 *
 *	This program is distributed in the hope that it will be useful,
 *	but WITHOUT ANY WARRANTY; without even the implied warranty of
 *	MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *	GNU General Public License for more details.
 *
 *	You should have received a copy of the GNU General Public License
 *	along with this program; if not, write to the Free Software
 *	Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 */

/*#define DEBUG */
/*
 *  MDP changes:
 *               in taltcheck: report when result of ENBC, etc is thrown away
 *               in taltdisables: report that result of DISC, etc is thrown away
 *               in tlengthcheck: call throw_the_result_away
 */

/*{{{  include files*/
#ifdef HAVE_CONFIG_H
#  include <config.h>
#endif
#include <stdio.h>
#include "includes.h"
#include "instruct.h"
#include "generror.h"
#include "genhdr.h"
#include "trandef.h"
#include "bind1def.h"
#include "bind2def.h"
#include "bind3def.h"
#include "gen1def.h"
#include "gen2def.h"
#include "gen4def.h"
#include "gen7def.h"
#include "gen9def.h"
#include "gen10def.h"
#include "gen11def.h"
#include "gen12def.h"
#include "code1def.h"
/*}}}*/

/*{{{  private variables*/
/*#define MAX_NESTED_ALTS 10*/ /* There is now no hard limit */

PRIVATE treenode **replaltstack;
PRIVATE int replaltstacklen;
PRIVATE INT32 replaltvaluesbase; /* WS offset of first saved replicator */

/* The following variables are only live while generating code for
   the enabling and disabling sequences, and since these can't be
   nested, there is no need to save the when recursing.
*/
PRIVATE int nalternatives;
PRIVATE int alternativecount;
PRIVATE BOOL inreplalt;
PRIVATE BOOL altcheckneeded;
PRIVATE BOOL altcheckdone;

PRIVATE int alt_repl_scope_level;

/* replaltvector is used when mapping enables, and when mapping bodies */
PRIVATE treenode *replaltvector; /* Used in mapaltenable */
PRIVATE BOOL premapped_alt; /* TRUE if mapping top level of an ALT */

PRIVATE treenode *intnodeptr = NULL; /* used for processing counted arrays */

#define virtualinstring      "VIRTUAL.IN%"
#define virtualoutstring     "VIRTUAL.OUT%"
#define virtualoutbytestring "VIRTUAL.OUT.BYTE%"
#define virtualoutwordstring "VIRTUAL.OUT.WORD%"
#define virtualvinstring     "VIRTUAL.VIN%"  /* variableio 11/3/91 */
#define virtualvoutstring    "VIRTUAL.VOUT%" /* variableio 11/3/91 */
/*}}}*/

/*{{{  PRIVATE treenode *iocallname(ioinst)*/
PRIVATE treenode *iocallname (const int ioinst)
{
  const char *str = NULL; /* initialised to shut up gcc's optimiser */
  switch(ioinst)
    {
      case I_IN      : str = virtualinstring;      break;
      case I_OUT     : str = virtualoutstring;     break;
      case I_OUTWORD : str = virtualoutwordstring; break;
      case I_OUTBYTE : str = virtualoutbytestring; break;
      case I_VIN     : str = virtualvinstring;     break; /* variableio 11/3/91 */
      case I_VOUT    : str = virtualvoutstring;    break; /* variableio 11/3/91 */
    }
  DEBUG_MSG(("In iocallname: str = %s\n", str));
  return vlibentry(processlibname(str, 1, vlibsuffix));
}
/*}}}*/
/*{{{  PRIVATE mapioop(ioinst, iobycall)*/
PRIVATE void mapioop (const int ioinst, const BOOL usecall)
{
  DEBUG_MSG(("In mapioop: ioinst = %d, usecall = %d\n", ioinst, usecall));
  if (usecall)
    {
      treenode *nptr = iocallname(ioinst);
      INT32 ws, vs;
      getprocwsandvs(nptr, &ws, &vs);
      datasize = max_INT32(datasize, (ws + REG_PARAMS + INS_EXTRA));
      /* vs will always be zero */
      add_to_libentries(nptr);
    }
  else
    {
      datasize = max_INT32(datasize, DS_IO);
      if (ioinst == I_OUTWORD || ioinst == I_OUTBYTE)
        reservelowworkspace(1);
    }
}
/*}}}*/

/*{{{  PRIVATE treenode *fixporttype(portuse)*/
/*****************************************************************************
 *
 *  fixporttype converts a port namenode (used in the tree 'portuse')
 *              from being a 'PORT OF t' to being a 't'.
 *              This unpleasant trick allows us to use the general
 *              assign code on port input and port output.
 *              The original type tree is returned so that it can be
 *              put back at the earliest possible opportunity (see
 *              routine 'restoreporttype').
 *
 *****************************************************************************/
PRIVATE treenode *fixporttype (treenode *const portuse, treenode ***const typeaddr)
{
  treenode *portname = nameof(portuse);
  treenode **porttype = NTypeAddr(portname);
  treenode *oldtype;
  treenode **oldtypeaddr;
#if 0
  printf("fixporttype: ");
  printtree(stdout, 0, *porttype);
  printf("\n");
#endif
  while (TagOf(*porttype) == S_ARRAY) porttype = ARTypeAddr(*porttype);
  oldtype = *porttype;
  oldtypeaddr = porttype; /* INSdi03067 */
  *porttype = ProtocolOf(*porttype);
  SetNVDevice(portname, TRUE);
  *typeaddr = oldtypeaddr; /* INSdi03067 */
  return oldtype;
}
/*}}}*/
/*{{{  PRIVATE void restoreporttype(portuse, porttype)*/
/*****************************************************************************
 *
 *  restoreporttype sets the type field of the port namenode
 *                  (used in the tree 'portuse') to 'porttype'.
 *
 *****************************************************************************/
PRIVATE void restoreporttype ( treenode *const portuse , treenode *const porttype,
             treenode **const oldtypeaddr)
{
  treenode *portname = nameof(portuse);
  treenode **type = NTypeAddr(portname);

#if 0
  printf("restoreporttype: portuse: ");
  printtree(stdout, 0, portuse);
  printf("\n, portname: ");
  printtree(stdout, 0, portname);
  printf("\n, porttype: ");
  printtree(stdout, 0, porttype);
  printf("\n, oldtypeaddr: ");
  printtree(stdout, 0, *oldtypeaddr);
  printf("\n, type: ");
  printtree(stdout, 0, *type);
  printf("\n");
#endif

  *oldtypeaddr = porttype; /* restore the original copy */ /* INSdi03067 */

  while (TagOf(*type) == S_ARRAY)
    type = ARTypeAddr(*type);

  if (TagOf(*type) != S_PORT) /* INSdi03067 */
    /* a temp has been inserted, which doesn't share a common type tree */
    {
      *type = newtypenode(S_PORT, LocnOf(porttype), NULL, *type); /* INSdi03067 */
    }
}
/*}}}*/
/*{{{  PRIVATE int numberofalternatives(tptr)*/
/*{{{  comment*/
/*****************************************************************************
 *
 *  numberofalternatives count the number of alternatives in an ALT
 *                       construct: a replicated alternative counts as 1.
 *
 *****************************************************************************/
/*}}}*/
PRIVATE int numberofalternatives ( treenode *tptr )
{
  while (isspecification(tptr))
    tptr = DBodyOf(tptr);
  switch(TagOf(tptr))
    {
      case S_ALTERNATIVE:
        return 1;
      case S_ALT:
      case S_PRIALT:
        {
          int count = 0;
          for (tptr = CBodyOf(tptr); !EndOfList(tptr); tptr = NextItem(tptr))
            count += numberofalternatives(ThisItem(tptr));
          return count;
        }
      case S_REPLALT:
      case S_PRIREPLALT:
        return numberofalternatives(ReplCBodyOf(tptr));
      default :
        badtag(genlocn, TagOf(tptr), "numberofalternatives");
    }
  return (0); /* Not reached */
}
/*}}}*/
/*{{{  PRIVATE int countnestedalts(tptr)*/
/*{{{  comment*/
/*****************************************************************************
 *
 *  countnestedalts counts the max number of nested levels in an ALT
 *
 *****************************************************************************/
/*}}}*/
#define NESTED_ALTS  0  /* count all levels of the ALT */
#define NESTED_REPLS 1  /* just count the replicated levels */
PRIVATE int countnestedalts ( treenode *tptr, int flag)
{
  DEBUG_MSG(("countnestedalts: %s\n", flag == NESTED_ALTS ? "nested ALTs" : "nested repls"));
  while (isspecification(tptr))
    tptr = DBodyOf(tptr);
  switch(TagOf(tptr))
    {
      case S_ALTERNATIVE:
        if (flag == NESTED_ALTS)
          return 1;
        else
          return 0;
      case S_ALT:
      case S_PRIALT:
        {
          int count = 0;
          for (tptr = CBodyOf(tptr); !EndOfList(tptr); tptr = NextItem(tptr))
            count = (int)max_INT32(count, countnestedalts(ThisItem(tptr), flag));
          if (flag == NESTED_ALTS)
            return (count + 1);
          else  /* (flag == NESTED_REPLS) */
            return (count);
        }
      case S_REPLALT:
      case S_PRIREPLALT:
        return (countnestedalts(ReplCBodyOf(tptr), flag) + 1);
      default :
        badtag(genlocn, TagOf(tptr), "countnestedalts");
    }
  return (0); /* Not reached */
}
/*}}}*/
/*{{{  PRIVATE BOOL hastruealtguard(tptr)*/
/*****************************************************************************
 *
 *  hastruealtguard returns TRUE if the ALT construct tptr has at least
 *                  one TRUE Boolean guard
 *
 *****************************************************************************/
PRIVATE BOOL hastruealtguard ( treenode *tptr )
{
  while (tptr != NULL)
    {
      tptr = skipspecifications(tptr);
      switch(TagOf(tptr))
        {
          /*{{{  S_ALT S_PRIALT*/
          case S_ALT: case S_PRIALT:
            for (tptr = CBodyOf(tptr); !EndOfList(tptr); tptr = NextItem(tptr))
              if (hastruealtguard(ThisItem(tptr)))
                return TRUE;
            break;
          /*}}}*/
          /*{{{  S_ALTERNATIVE*/
          case S_ALTERNATIVE:
            return istrueguard(AltGuardOf(tptr));
          /*}}}*/
          /*{{{  S_REPLALT S_PRIREPLALT*/
          case S_REPLALT: case S_PRIREPLALT:
            tptr = ReplCBodyOf(tptr);
            break;
          /*}}}*/
        }
    }
  return FALSE;
}
/*}}}*/
/*{{{  PRIVATEPARAM int do_alt_check_enabling_functions*/
PRIVATEPARAM int do_alt_check_enabling_functions(treenode **tptr, void *voidptr)
{
  treenode *const nptr = *tptr;
  treenode *const fptr = voidptr; /* namenode of fn */
#if 0
  printf("do_alt_check_enabling_functions: Looking at %s, \"%s\", scope: %d\n",
         itagstring(TagOf(nptr)),
         WNameOf(NNameOf(nptr)), NScopeOf(nptr));
#endif
  if ((TagOf(nptr) == N_REPL) && (NScopeOf(nptr) >= alt_repl_scope_level))
    msg_out_ss(SEV_FATAL, GEN, GEN_REPL_FN_IN_ALT_GUARD, genlocn,
               WNameOf(NNameOf(fptr)), WNameOf(NNameOf(nptr)));
  return CONTINUE_WALK;
}
/*}}}*/
/*{{{  PUBLIC void alt_check_enabling_functions*/
PUBLIC void alt_check_enabling_functions(treenode *const tptr)
{
  if (alt_repl_scope_level != (-1)) /* inside a repl ALT */
    {
    #if 0
      printf("Found FINSTANCE: %s, scope: %d\n",
             WNameOf(NNameOf(INameOf(tptr))), NScopeOf(INameOf(tptr)));
    #endif
      if (NScopeOf(INameOf(tptr)) > alt_repl_scope_level)
        walk_free_vars(INameOf(tptr), do_alt_check_enabling_functions,
                       INameOf(tptr) /*LocnAddr(tptr)*/);
    }
}
/*}}}*/
/*{{{  PRIVATEPARAM void mapaltenables(tptr)*/
/*****************************************************************************
 *
 *  mapaltenables allocates workspace for enabling each guard in
 *                the ALT tree 'tptr'.
 *
 *****************************************************************************/
PRIVATEPARAM void mapaltenables ( treenode *tptr )
{
  const BOOL saved_premapped_alt = premapped_alt;
  premapped_alt = FALSE;
  DEBUG_MSG(("mapaltenables: replaltstacklen is %d\n", replaltstacklen));
  switch(TagOf(tptr))
    {
      default:               /* Must be a leading specification */
        /* modified so that this calls mappreprocess now - bug 776 5/10/90 */
        mapdeclandbody(tptr, mapaltenables, TRUE, TRUE);
        break;
      /*{{{  S_ALT S_PRIALT*/
      case S_ALT: case S_PRIALT:
        SetCTemp(tptr, replaltvector);
        if (!saved_premapped_alt) mappreprocess(tptr);
        mapconstruction(CBodyOf(tptr), mapaltenables);
        break;
      /*}}}*/
      /*{{{  S_REPLALT S_PRIREPLALT*/
      case S_REPLALT: case S_PRIREPLALT:
        {
          const INT32 oldweight = loop_weight;
          const int   oldscope  = alt_repl_scope_level;
          uploop_weight((INT32)2); /* Enable + disables */
          if (alt_repl_scope_level == (-1))
            alt_repl_scope_level = NScopeOf(ReplCNameOf(tptr));
          SetReplCTemp(tptr, replaltvector);
          replaltstacklen++;
          DEBUG_MSG(("mapaltenables: found a repl, replaltstacklen is now %d\n", replaltstacklen));
          if (!saved_premapped_alt) mappreprocess(tptr);
          maprepl(tptr, mapaltenables);
          replaltstacklen--;
          /*addusecount(replaltvector, NVUseCountOf(ReplCNameOf(tptr)));*/
          loop_weight = oldweight;
          alt_repl_scope_level = oldscope;
        }
        break;
      /*}}}*/
      /*{{{  S_ALTERNATIVE*/
      case S_ALTERNATIVE:
        {
          treenode **const guard = AltGuardAddr(tptr);
          /*{{{  map for enable/disable*/
          mappreprocess(tptr);
          {
          /*printf("mapenables: alternative: repl scope level is %d\n", alt_repl_scope_level);*/
            if (AltChanExpOf(tptr) == NULL)
              mapbool(guard);
            else
              {
                treenode **chanexp = AltChanExpAddr(tptr); /* added 2/11/90 bug 779 */
                treenode **enabler;
                int mode           = P_EXP;

                if (AltTimeExpOf(tptr) == NULL)
                  {
                    /*{{{  mask off virtual i/o bit if required*/
                    /*const BOOL usecall = iobycall && maybevirtualchan(*channel);*/
                    /*printf("mapaltenables: realign_virtual_channels? %d, virtual? %d\n",
                           realign_virtual_channels, maybevirtualchan(*chanexp));*/
                    if (realign_virtual_channels && maybevirtualchan(*chanexp))
                      {
                        if (chanaspointer)
                          {
                            /* bug INSdi02103 */
                            *chanexp = newdopnode(S_BITAND, genlocn, *chanexp,
                                       newconstant((int)(-bytesperword)), S_INT);
                          }
                        else
                          {
                            /* Cannot do this, cos I'm too lazy to implement it, and we
                               think that in practice the zw (!chanaspointer) option will
                               always be used when iobycall is FALSE */
                            /* This call to memo_err makes sure that we only issue the error msg
                               once per compilation unit */
                            if (memo_err(GEN, GEN_ENABLE_CHAN_ALIGNMENT_PROBLEM, NOPOSN))
                              msg_out(SEV_WARN, GEN, GEN_ENABLE_CHAN_ALIGNMENT_PROBLEM, LocnOf(tptr));
                          }
                      }
                    /*}}}*/
                    enabler = chanexp;
                    mode = chanaspointer ? P_EXP : P_PTR;
                  }
                else /* timer input */
                  {
                    mapexpopd(P_EXP, chanexp); /* Bug 288 22/5/90 */
                    enabler = AltTimeExpAddr(tptr);
                  }

                if (T9000_alpha_badalt(&tx_global))
                  /*{{{  T9000_alpha_badalt*/
                  {
                    mapexpopd(P_EXP, guard);
                    mapexpopd(mode, enabler);
                    if (!issimplelocal(*enabler, be_lexlevel))
                      {
                        /* need to re-load channel for 'cj' trick */
                        *enabler = gettemp(*enabler, NM_WORKSPACE);
                        upusecount(*enabler, 2);
                        freetemp(*enabler);
                      }
                  }
                  /*}}}*/
                else
                if (!istrueguard(*guard) && cancauseerror(*enabler))
                  {
                    mapexpopd(P_EXP, guard);
                    mapexpopd(mode, enabler);
                  }
                else
                  mapload2regs(mode, enabler, P_EXP, guard);
              }
          }
          /*}}}*/
          /*SetAltTemp(tptr, replaltvector);*/
          addusecount(replaltvector, replaltstacklen);
        }
        break;
      /*}}}*/
    }
  premapped_alt = saved_premapped_alt;
}
/*}}}*/
/*{{{  PRIVATEPARAM void mapaltbodies*/
/*****************************************************************************
 *
 *  mapaltbodies allocates workspace for the body of each guarded process in
 *                the ALT tree 'tptr'.
 *
 *****************************************************************************/
PRIVATEPARAM void mapaltbodies ( treenode *tptr )
{
  treenode *const save_vsp_nptr = enclosing_vsp_nptr;
  int *const scope = save_scope();
  tptr = resurrect_specs(tptr);
  switch(TagOf(tptr))
    {
      /*{{{  S_ALT S_PRIALT*/
      case S_ALT: case S_PRIALT:
        mapconstruction(CBodyOf(tptr), mapaltbodies);
        break;
      /*}}}*/
      /*{{{  S_REPLALT S_PRIREPLALT*/
      case S_REPLALT: case S_PRIREPLALT:
        {
          /* Don't include the enabling and disabling loops in the usage
             calculation for the replaltvector */
          INT32 enable_count = NVUseCountOf(ReplCNameOf(tptr));
          resurrect_var(ReplCNameOf(tptr));
          mapaltbodies(ReplCBodyOf(tptr));
          addusecount(ReplCTempOf(tptr), NVUseCountOf(ReplCNameOf(tptr)) - enable_count);
        }
        break;
      /*}}}*/
      /*{{{  S_ALTERNATIVE*/
      case S_ALTERNATIVE:
      #if 0
        /* Have to be careful, because alternatives
           will not yet have been through mappreprocess. */
        {
          treenode *input = AltInputOf(tptr);
          /*{{{  map the deferred specification and the input*/
          {
            /* mapdeclandbody will descope the specs before the input
             * so we must resurrect them before we process the body
             */
            /* modified so that this calls mappreprocess now - bug 776 5/10/90 */
            mapdeclandbody(input, mapinput, TRUE, TRUE);
            input = resurrect_specs(input);
          }
          /*}}}*/
          /*{{{  map the body*/
          {
            switch(inputtypeof(input))
              {
                /*{{{  case input*/
                case INP_CASE_INPUT:
                  break;   /* There is no body */
                /*}}}*/
                /*{{{  everything else*/
                default:
                  mapprocess(AltBodyOf(tptr));
                  break;
                /*}}}*/
              }
          }
          /*}}}*/
        }
      #else
        mapdeclandbody(AltBodyOf(tptr), mapprocess, TRUE, TRUE);
      #endif
        break;
      /*}}}*/
    }
  restore_scope(scope);
  enclosing_vsp_nptr = save_vsp_nptr;
}
/*}}}*/
/*{{{  PRIVATE void taltcheck()*/
PRIVATE void taltcheck (const BOOL inskippingalt, const BOOL known_true_guard)
{
  if (altcheckneeded)
    /*{{{  do a check*/
    {
      if (known_true_guard ||
          (!inreplalt && alternativecount == 1))
      {
        genprimary(I_STL, 0);
      }
      else
        {
          genprimary(I_LDL, 0);
          gensecondary(I_OR);
          if (!inreplalt && !inskippingalt && alternativecount == nalternatives)
            {
              genprimary(I_LDC, 1);
              gensecondary(I_CCNT1);
              altcheckdone = TRUE;
            }
          else
          {
            genprimary(I_STL, 0);
          }
        }
    }
    /*}}}*/
    else
      throw_the_result_away();
}
/*}}}*/
/*{{{  PRIVATEPARAM void taltenables*/
/*****************************************************************************
 *
 *  taltenables enables all the guards in the ALT or replicated ALT tree tptr,
 *              including those of any nested alternatives.
 *
 *****************************************************************************/
PRIVATEPARAM void taltenables ( treenode *tptr, void *const voidptr, const INT32 dummy)
{
  treenode *aptr = tspecs(tptr);
  DEBUG_MSG(("taltenables\n"));
  switch(TagOf(aptr))
    {
      /*{{{  S_ALT S_PRIALT*/
      case S_ALT: case S_PRIALT:
        DEBUG_MSG(("taltenables: found a nested ALT or PRIALT\n"));
        for (aptr = CBodyOf(aptr); !EndOfList(aptr); aptr = NextItem(aptr))
          taltenables(ThisItem(aptr), voidptr, dummy);
        break;
      /*}}}*/
      /*{{{  S_ALTERNATIVE*/
      case S_ALTERNATIVE:
        {
          treenode *const guard = AltGuardOf(aptr);
          BOOL inskippingalt = FALSE;

          DEBUG_MSG(("taltenables: found a branch of the ALT\n"));
          alternativecount++;

          new_occam_line (aptr, TRUE, TRUE, FALSE);
          /* if (debugoutput)
            coder_genlocate(aptr, TRUE); */
          /* Added 27/10/92 for a comment in the zb display */

          /* bug INSdi02047 05/05/93 */
          temp_clear_all_evaluated(); /* bug TS/1900 22/10/92 */

          tpreexp(guard);

          if (AltChanExpOf(aptr) == NULL)
            /*{{{  skip guard*/
            {
              tbool(guard, MANY_REGS);
              gensecondary(I_ENBS);
              taltcheck(inskippingalt, FALSE);
            }
            /*}}}*/
          else
            {
              int skiplab = NO_LABEL; /* bug TS/1827 14/08/92 */
              treenode *enabler;
              int mode = P_EXP;
              if (AltTimeExpOf(aptr) == NULL)
                /*{{{  set up enabling channel*/
                {
                  treenode *inputchannel = AltChanExpOf(aptr);
                  /* Don't try and preevaluate it here */
                  if (TagOf(inputchannel) == T_TEMP)
                    inputchannel = NDeclOf(inputchannel);

                  enabler = inputchannel;
                  mode = chanaspointer ? P_EXP : P_PTR;
                }
                /*}}}*/
              else
                /*{{{  timer alt*/
                {
                  treenode *time = /*RHSOf(input)*/AltTimeExpOf(aptr);
                  tpreexp(/*LHSOf(input)*/AltChanExpOf(aptr)); /* Bug 288 22/5/90 */

                  /* Don't try and preevaluate it here */
                  if (T9000_alpha_badalt(&tx_global) && (TagOf(time) == T_TEMP))
                    time = NDeclOf(time);

                  enabler = time;
                }
                /*}}}*/

              tpreexp(enabler);
              if (T9000_alpha_badalt(&tx_global))
                /*{{{  T9000_alpha_badalt*/
                {
                  /* strictly we only need to do this when disabling,
                     but we do it here too so that the mapper doesn't
                     get confused */
                  if (!istrueguard(guard))
                    {
                      skiplab = newlab(); /* bug TS/1827 14/08/92 */
                      inskippingalt = TRUE;
                      tguard(guard, TRUE, skiplab);
                    }
                  /* This isn't necessary cos we only evaluate this once */
                  /* If we need to re-instate it, we need to remove the
                     call to NDeclOf(...) when setting up inputchannel */
                  /*mode = simplify(mode, enabler);*/
                  texpopd(mode, enabler, MANY_REGS);
                  loadconstant(1);
                }
                /*}}}*/
              else
              if (!istrueguard(guard) && cancauseerror(enabler))
                {
                  skiplab = newlab(); /* bug TS/1827 14/08/92 */
                  inskippingalt = TRUE;
                  tguard(guard, TRUE, skiplab);
                  texpopd(mode, enabler, MANY_REGS);
                  loadconstant(1);
                }
              else
                tload2regs(mode, enabler, P_EXP, guard, FALSE, TRUE);
              /*{{{  T9000_gamma_carryin*/
              if ((AltTimeExpOf(aptr) != NULL) && T9000_gamma_carryin(&tx_global))
                {
                  genprimary(I_LDC, 0);
                  gensecondary(I_SUB);
                  gencomment0("Set carry flag for ALU Adder CarryIn bug");
                }
              /*}}}*/
              gensecondary((AltTimeExpOf(aptr) == NULL) ? I_ENBC : I_ENBT);
              taltcheck(inskippingalt,
                        skiplab != NO_LABEL); /* INSdi02264 22/06/93 */
              if (skiplab != NO_LABEL) /* bug TS/1827 14/08/92 */
                setlab(skiplab);
            }
          temp_unevaluate_all(); /* bug TS/1900 22/10/92 */
        }
        break;
      /*}}}*/
      /*{{{  S_REPLALT S_PRIREPLALT*/
      case S_REPLALT: case S_PRIREPLALT:
        {
          const BOOL savedinreplalt = inreplalt;
          DEBUG_MSG(("taltenables: found a nested REPLALT or PRIREPLALT\n"));
          tpreexp(ReplCStartExpOf(aptr));
          tpreexp(ReplCLengthExpOf(aptr));
          inreplalt = TRUE;
          trepl(aptr, taltenables, voidptr, dummy, FALSE);
          inreplalt = savedinreplalt;
        }
        break;
      /*}}}*/
    }
}
/*}}}*/
/*{{{  PRIVATEPARAM void taltdisables*/
/*****************************************************************************
 *
 *  taltdisables disables all the guards in the ALT or replicated ALT tree
 *               tptr, including those of any nested alternatives.
 *
 *****************************************************************************/
PRIVATEPARAM void taltdisables ( treenode *tptr , void *const voidptr, const INT32 altendlab )
{
  treenode *aptr = skipspecifications(tptr);
  DEBUG_MSG(("taltdisables, replaltstacklen is %d\n", replaltstacklen));    
  /*{{{  translate leading specifications for the second time*/
  /* These specifications can be translated in REDUCED error mode as
     we have already translated them once, and if anything was going to
     go wrong, it would have done so already.
     Don't do 'tpreexp' a second time. */
  #if 0  /* We've now guaranteed (in transformalt) that all leading specs are necessary*/
  while (isspecification(tptr))
    {
      if (usedinaltguards(DNameOf(tptr), aptr))
        tspecification(tptr);
      tptr = DBodyOf(tptr);
    }
  #endif
  tptr = tspecs(tptr);
  /*}}}*/
  switch(TagOf(aptr))
    {
      /*{{{  S_ALT S_PRIALT*/
      case S_ALT: case S_PRIALT:
        for (aptr = CBodyOf(aptr); !EndOfList(aptr); aptr = NextItem(aptr))
          taltdisables(ThisItem(aptr), voidptr, altendlab);
        break;
      /*}}}*/
      /*{{{  S_ALTERNATIVE*/
      case S_ALTERNATIVE:
        {
          treenode *const guard = AltGuardOf(aptr);
          const int startlab = newlab();
          int skiplab = NO_LABEL; /* bug TS/1827 14/08/92 */

          SetAltLabel(aptr, startlab);

          new_occam_line (aptr, TRUE, TRUE, FALSE);
          /* if (debugoutput)
            coder_genlocate(aptr, TRUE); */
          /* Added 27/10/92 for a comment in the zb display */

          if (AltChanExpOf(aptr) == NULL)
            /*{{{  skip guard*/
            {
              tbool(guard, MANY_REGS);
              genlabeldiff(I_LDC, startlab, (int)altendlab);
              gensecondary(I_DISS);
            }
            /*}}}*/
          else
            {
              treenode *disabler;
              int mode = P_EXP;
              if (AltTimeExpOf(aptr) == NULL)
                /*{{{  get channel address*/
                {
                  /* bug 779 2/11/90 - modified to use AltChanExpOf */
                  treenode *inputchannel = /*LHSOf(input)*/ AltChanExpOf(aptr);
                  /* Don't try and preevaluate it here */
                  if (!T9000_alpha_badalt(&tx_global) &&
                      TagOf(inputchannel) == T_TEMP)
                    inputchannel = NDeclOf(inputchannel);

                  disabler = inputchannel;
                  mode = chanaspointer ? P_EXP : P_PTR;
                }
                /*}}}*/
              else
                disabler = AltTimeExpOf(aptr);

              if (T9000_alpha_badalt(&tx_global))
                /*{{{  T9000_alpha_badalt*/
                {
                  if (!istrueguard(guard))
                    {
                      skiplab = newlab(); /* bug TS/1827 14/08/92 */
                      tguard(guard, TRUE, skiplab);
                    }
                  mode = simplify(mode, disabler);

                  if (!testflag)
                    {
                      /*gensecondary(I_LDPRODID);*/ /* Not on alpha! */
                      genprimary(I_LDC, 0x10E);
                      gencomment0("Silicon bug fix;");
                      gensecondary(I_LDCONF);
                      gencomment0("Slooow instruction to help prefetch get ahead");
                    }

                  texpopd(mode, disabler, MANY_REGS);
                  texpopd(mode, disabler, MANY_REGS);
                  gencomment0("Silicon bug fix; guard := chan/time");
                }
                /*}}}*/
              else
              if (!istrueguard(guard) && cancauseerror(disabler))
                /*{{{  guard ; cj; chan/time ; ldc 1*/
                {
                  skiplab = newlab(); /* bug TS/1827 14/08/92 */
                  tguard(guard, TRUE, skiplab);
                  texpopd(mode, disabler, MANY_REGS);
                  loadconstant(1);
                }
                /*}}}*/
              else
                tload2regs(mode, disabler, P_EXP, guard, FALSE, TRUE);
              /*{{{  T9000_gamma_carryin*/
              if ((AltTimeExpOf(aptr) != NULL) && T9000_gamma_carryin(&tx_global))
                {
                  genprimary(I_LDC, 0);
                  gensecondary(I_SUB);
                  gencomment0("Set carry flag for ALU Adder CarryIn bug");
                }
              /*}}}*/
              genlabeldiff(I_LDC, startlab, (int)altendlab);
              /*{{{  T9000_alpha_badalt*/
              if (T9000_alpha_badalt(&tx_global) && !testflag)
                {
                  gensecondary(I_NOP);
                  gencomment0("try to ensure the following group together");
                  /*genprimary(I_LDC, 1);*/
                  texpopd(mode, disabler, MANY_REGS);
                  gencomment0("Silicon bug fix - push into D reg");
                  genprimary(I_CJ, 0);
                  gencomment0("Pop it off again to sort out the ALU");
                }
              /*}}}*/
              gensecondary((AltTimeExpOf(aptr) == NULL) ? I_DISC : I_DIST);
            }

          /*{{{  select current value of any enclosing replicators*/
          {
            if (replaltstacklen > 0)
              {
                int i;
                if (skiplab == NO_LABEL) skiplab = newlab(); /* bug TS/1827 14/08/92 */
                genbranch(I_CJ, skiplab);                   /*          cj    skiplab   */
                for (i = 0; i < replaltstacklen; i++)
                  {
                    treenode *replname = replaltstack[i];
                    loadname(replname, REPL_BASE);          /*          ldl   i        */
                    /*storeinname(replname, REPL_SELECTED);*//*         stl   i + 2    */
                    genprimary(I_STL, replaltvaluesbase + i); /*        stl   saved_repl */
                    gencomment1("INT saved repl %d", i);
                  }
              }
              else
                  throw_the_result_away();
          }
          /*}}}*/
          if (skiplab != NO_LABEL) /* bug TS/1827 14/08/92 */
            setlab(skiplab);                          /* skiplab:                */
        }
        break;
      /*}}}*/
      /*{{{  S_REPLALT S_PRIREPLALT*/
      case S_REPLALT: case S_PRIREPLALT:
        replaltstack[replaltstacklen] = ReplCNameOf(aptr);
        replaltstacklen++;
        DEBUG_MSG(("taltdisables: found a repl, replaltstacklen is now %d\n", replaltstacklen));
        trepl(aptr, taltdisables, voidptr, altendlab, FALSE);
        replaltstacklen--;
        break;
      /*}}}*/
    }
}
/*}}}*/
/*{{{  PRIVATE void taltbodies(tptr, joinlab)*/
/*****************************************************************************
 *
 *  taltbodies generates code for all the ALT bodies contained in the ALT or
 *             replicated ALT tree, tptr, including the bodies of nested ALTs.
 *
 *****************************************************************************/
PRIVATE void taltbodies ( treenode *tptr , int joinlab,
                          treenode **altbodystack, int altbodystackptr )
{
  DEBUG_MSG(("taltbodies: altbodystackptr is %d\n", altbodystackptr));
  altbodystack[altbodystackptr] = tptr;
  tptr = skipspecifications(tptr);
  switch(TagOf(tptr))
    {
      /*{{{  S_ALT S_PRIALT*/
      case S_ALT: case S_PRIALT:
        for (tptr = CBodyOf(tptr); !EndOfList(tptr); tptr = NextItem(tptr))
          taltbodies(ThisItem(tptr), joinlab, altbodystack, altbodystackptr+1);
        break;
      /*}}}*/
      /*{{{  S_ALTERNATIVE*/
      case S_ALTERNATIVE:
        {
          /*treenode *input = AltInputOf(tptr);*/
          treenode *body = AltBodyOf(tptr);

          setlab(AltLabelOf(tptr));

          /*{{{  translate specifications leading up to this alternative*/
          {
            /* These specifications can be translated in REDUCED error mode as
               we have already translated them at least twice, and if anything
               was going to go wrong, it would have done so already. */
            int i;
            for (i = 0; i < (altbodystackptr+1); i++)
              {
                treenode *t = altbodystack[i];
                t = tspecs(t);
              }
          }
          /*}}}*/
        #if 0 /* bug TS/1626 21/04/92 */
          /*{{{  translate specifications moved in front of the input*/
          input = tspecs(input);
          /*}}}*/
          switch(inputtypeof(input))
            {
              /*{{{  input, port input, timer input, tagged input*/
              case INP_INPUT:
              case INP_PORT_INPUT:
              case INP_TIMER_INPUT:
              case INP_TAGGED_INPUT:
                tinput(input);
                tprocess(body);
                break;
              /*}}}*/
              /*{{{  case input*/
              case INP_CASE_INPUT:
                tinput(input);
                break;  /* There is no body */
              /*}}}*/
              /*{{{  skip, delayed input*/
              case INP_SKIP:
              case INP_DELAYED_INPUT:
                tprocess(body);
                break;
              /*}}}*/
            }
        #else
          tprocess(body);
        #endif
          genjump(joinlab);
        }
        break;
      /*}}}*/
      /*{{{  S_REPLALT S_PRIREPLALT*/
      case S_REPLALT: case S_PRIREPLALT:
        /* Change the position of the replicator from replicated_i to saved i */
        {
          treenode *replname = ReplCNameOf(tptr);
        /*INT32 newoffset = NVOffsetOf(replname) + (REPL_SELECTED - REPL_BASE);*/
        /*INT32 newoffset = replaltvaluesbase + replaltstacklen;*/
          INT32 newoffset = replaltvaluesbase + replaltstacklen - nameoffsetof(NLexLevelOf(replname));
          DEBUG_MSG(("taltbodies: in replalt, replaltstacklen is %d\n", replaltstacklen));
          replaltstacklen++;
          SetNVOffset(replname, newoffset);
          if (!isconst(ReplCLengthExpOf(tptr)) || (LoValOf(ReplCLengthExpOf(tptr)) != 0))
            taltbodies(ReplCBodyOf(tptr), joinlab, altbodystack, altbodystackptr+1);
          replaltstacklen--;
        }
        break;
      /*}}}*/
    }
}
/*}}}*/

/*{{{  PUBLIC void tioop*/
PUBLIC void tioop (const int ioinst, const BOOL usecall)
{
  DEBUG_MSG(("In tioop: ioinst = %d, usecall = %d\n", ioinst, usecall));
  if (usecall)
    {
      treenode *const nptr = iocallname(ioinst);
      genbranch (I_CALL, NPLabelOf(nptr));
      if (assembly_output)
        gencomments("Call %s",WNameOf(NNameOf(nptr)));
      coder_return_from_call(FALSE); /* bug TS/2005 15/12/92 */
    }
  else
    {
      gensecondary(ioinst);
      /*{{{  T9000_alpha_badrunp*/
      if (T9000_alpha_badrunp(&tx_global))
        {
          gensecondary(I_NOP);
          gencomment0("Fix for scheduler problem");
          gensecondary(I_NOP);
          gencomment0("Fix for scheduler problem");
        }
      /*}}}*/
    }
}
/*}}}*/
/*{{{  PRIVATE void mapoutputitem*/
/*****************************************************************************
 *
 *  mapoutputitem allocates workspace required for outputting
 *                (outputitemmode, outputitem) on channel
 *                (channelmode, channel)
 *
 *****************************************************************************/
PRIVATE void mapoutputitem (const int channelmode,    treenode **const channel,
                                  int outputitemmode, treenode **const outputitem,
                            treenode *const comm)
{
  const BOOL usecall = iobycall && maybevirtualchan(*channel);
  DEBUG_MSG(("mapoutputitem: usecall is %d\n", usecall));
  if (TagOf(*outputitem) == S_COLON2)
    /*{{{  map counted output*/
    {
      treenode **countexp = LeftOpAddr(*outputitem);
      treenode **arrayexp = RightOpAddr(*outputitem);
      treenode *dimexp = dimexpof(*arrayexp, 0);
      int countmode = P_EXP;
      const int constantcount = isconst(*countexp);
      const int type = ntypeof(*countexp);
      const SOURCEPOSN locn = LocnOf(*outputitem);
      if (has_variableio)
        /*{{{  use the special variable length io instruction*/
        {
          /* we modify the left hand side of the tree node to hold
             the actual number of bytes to be output.
          */
          treenode *arraytype = ARTypeOf(gettype(*arrayexp));

          /*{{{  Firstly convert to an INT if necessary:*/
          if (isquadlength(type))
            *countexp = makeconversion(type, S_INT, *countexp, S_EXACT, locn);
          else if (!fitsinword(type) || isshortint(type))
            *countexp = newmopnode(S_EXACT, locn, *countexp, S_INT);
          /*}}}*/

          /*{{{  Now insert possible range check:*/
          if (RANGECHECKING && (!constantcount || !isconst(dimexp)))
            {
              treenode *maxlen = isconst(dimexp) ?
                        newconstant(LoValOf(dimexp) + 1)
                      : newdopnode(S_ADD, locn, dimexp, newconstant(1), S_INT);
              *countexp = newdopnode(S_CSUB0, locn, *countexp, maxlen, S_INT);
            }
          /*}}}*/

          /*{{{  Now multiply by the size of each element of the rhs:*/
          if (bytesin(arraytype) != 1)
            /* scaletreeof also works if the size is not a constant */
            *countexp = newdopnode(S_TIMES, locn, *countexp,
                                   scaletreeof(arraytype, 1, be_lexlevel), S_INT);
          /*}}}*/

          *countexp = foldexp(*countexp);
          mapload3regs(P_PTR, arrayexp, channelmode, channel, P_EXP, countexp);
          mapioop(I_VOUT, usecall);
        }
        /*}}}*/
      else
        {
          /*{{{  preevaluate count to a temporary if necessary*/
          if (!constantcount)
            {
              if (!issimplelocal(*countexp, be_lexlevel)) /* || !fitsinword(type))*/ /* bug 1132 29/1/91 */
                {
                  *countexp = gettemp(*countexp, NM_WORKSPACE);
                  upusecount (*countexp, 1);
                  mapsimpleassign(type, P_TEMP, countexp, P_EXP, NDeclAddr(*countexp));
                  countmode = P_TEMP;
                }
              if (isshortint(type))
                upusecount(*countexp, 2); /* sign extension */ /* bug 706 13/9/90 */
            }
          /*}}}*/
          /*{{{  output the count*/
          mapoutputitem(channelmode, channel, countmode, countexp, NULL);
          /*}}}*/
          /*{{{  output the array from 0 for count*/
          if /* count is variable or count is constant but not zero */
            (!constantcount || (LoValOf(*countexp) != 0) ||
                               (HiValOf(*countexp) != 0))
              /*{{{  output the array from 0 for count*/
              {
                treenode *checkedcount;
                int checkcount = FALSE;
                /*{{{  map the conversion (check) if a quad word*/
                /* Added for bug 1130 29/1/91 */
                if (!constantcount && isquadlength(type))
                  {
                    const int old = switch_to_temp_workspace();
                    treenode *conversion = makeconversion (type, S_INT, *countexp, S_EXACT, locn);
                    switch_to_prev_workspace(old);
                    mapexp(&conversion);
                  }
                /*}}}*/
                /*{{{  check the count if it or the dimension are not constant*/
                if (RANGECHECKING && (!constantcount || !isconst(dimexp)))
                  {
                    const int old = switch_to_temp_workspace();
                    checkedcount = newdopnode(S_CCNT1, NOPOSN, *countexp, dimexp, S_INT);
                    switch_to_prev_workspace(old);
                    checkcount = TRUE;
                  }
                else
                  checkedcount = *countexp;
                /*}}}*/
                if (constantcount)
                  {
                     if (checkcount) mapexp(&checkedcount);
                     mapload2regs(P_PTR, arrayexp, channelmode, channel);
                  }
                else
                  /*{{{  ldptr array; ldptr channel; ld length*/
                  {
                    treenode *lengthexp;
                    const int old = switch_to_temp_workspace();
                    /* 10 is just a dummy value as we don't know the real size,
                       and it doesn't matter what it is anyway, just as long
                       as the lengthexp takes two registers to load */
                    lengthexp = newdopnode(S_TIMES, NOPOSN, checkedcount, newconstant(10), S_INT);
                    switch_to_prev_workspace(old);
                    mapload3regs(P_PTR, arrayexp, channelmode, channel, P_EXP, &lengthexp);
                  }
                  /*}}}*/
                mapioop(I_OUT, usecall);
              }
              /*}}}*/
          /*}}}*/
          /*{{{  free the count temporary if neccessary*/
          if (countmode == P_TEMP)
            freetemp(*countexp);
          /*}}}*/
        }
    }
    /*}}}*/
  else
    {
      const int t = ntypeof(*outputitem);
      BOOL freeoutputitem = FALSE;
      BOOL useout;
      /*{{{  This bit added by CO'N for eg. out ! x + x, REAL32 on T8*/
      if (has_fpu_core && isreal(t) &&
          !isaddressableopd(outputitemmode, *outputitem))
        {
          /*printf("out ! x + x\n");*/ /* INSdi02071 */
          if (isconst(*outputitem))
            placeintable(*outputitem);
          else
            {
            #if 0 /* This isn't quite right, or is it? */
              treenode *type = gettype(*outputitem);
              *outputitem = gettemp(*outputitem, NM_WORKSPACE);
              upusecount(*outputitem, 1);
              mapsimpleassign(TagOf(type), P_TEMP, outputitem, outputitemmode,
                              NDeclAddr(*outputitem));
              outputitemmode = P_TEMP;
              freeoutputitem = TRUE;
            #endif
              mapfpexp(outputitem);
              *outputitem = gettemp (*outputitem, NM_WORKSPACE);
              outputitemmode = P_TEMP;
              freeoutputitem = TRUE;
            }
        }
      /*}}}*/

      /* Bug 1203 - 12/3/91 - don't use outbyte and/or outword if the
         output item is already addressable.
         However, if it is a scalar constant, we haven't yet decided whether to
         put it into the constant table, so isaddressable will not yet
         return TRUE.
         If we would not normally use a constant table, we still
         do an outbyte or outword.
      */
      /* Bug TS/1504 - 07/04/92:
         If we're optimising for space, DO use the shorter forms if possible.
      */
      useout = ((code_style_flags & CODE_STYLE_SPACE_NOT_TIME) == 0) &&
               (isaddressableopd(outputitemmode, *outputitem) ||
                ((TagOf(*outputitem) == S_CONSTEXP) &&
                 shouldbeinconstanttable(*outputitem)) );

      if (!useout && /* bug 1203 12/3/91 */
          istargetintsize(t))
        /*{{{  outword*/
        {
          mapload2regs(channelmode, channel, outputitemmode, outputitem);
          mapioop(I_OUTWORD, usecall);
        }
        /*}}}*/
      else if (!useout && /* bug 1203 12/3/91 */
               istargetbytesize(t))
        /*{{{  outbyte*/
        {
          mapload2regs(channelmode, channel, outputitemmode, outputitem);
          mapioop(I_OUTBYTE, usecall);
        }
        /*}}}*/
      else
        /*{{{  out*/
        {
          treenode *type = gettype(*outputitem);
          const INT32 s = bytesin(type);
          /*{{{  if outputitem is not addressable, evaluate it into a temporary*/
          if (!isaddressableopd(outputitemmode, *outputitem))
            {
              if (isconst(*outputitem))
                placeintable(*outputitem);
              else
                {
                  *outputitem = gettemp(*outputitem, NM_WORKSPACE);
                  upusecount(*outputitem, 1);
                  mapsimpleassign(TagOf(type), P_TEMP, outputitem, outputitemmode,
                                  NDeclAddr(*outputitem));
                  outputitemmode = P_TEMP;
                  freeoutputitem = TRUE;
                }
            }
          /*}}}*/

          outputitemmode = ptrmodeof(outputitemmode);
          if (s == (-1))
            /*{{{  length is an expression*/
            {
              int loadseq;
              treenode *lengthexp;

              /*int old = switch_to_temp_workspace();*/
              lengthexp = scaletreeof(type, 1, be_lexlevel);
              /*switch_to_prev_workspace(old);*/

              loadseq = mapload3regs(outputitemmode, outputitem, channelmode, channel,
                                     P_EXP, &lengthexp);
              /* Ideally we should save this loadseq value on the tree somewhere
                 so that we don't have to recalculate it in the code generator */
              {
                /* bug TS/1626 24/04/92 - save the length expression */
                treenode *rhs = RHSOf(comm);
                NewNextItem(addtofront(lengthexp, NULL), rhs);
              }
            }
            /*}}}*/
          else
            mapload2regs(outputitemmode, outputitem, channelmode, channel);
          mapioop(I_OUT, usecall);
        }
        /*}}}*/
      if (freeoutputitem)
        freetemp(*outputitem);
    }
}
/*}}}*/
/*{{{  PUBLIC void mapinputitem*/
/*****************************************************************************
 *
 *  mapinputitem allocates workspace required for inputting inputitem
 *                on channel (channelmode, channel)
 *
 *****************************************************************************/
PUBLIC void mapinputitem (const int channelmode,   treenode **const channel,
                          const int inputitemmode, treenode **const inputitem,
                          treenode *const comm)
{
  const BOOL usecall = iobycall && maybevirtualchan(*channel);
  DEBUG_MSG(("mapinputitem, usecall %d\n", usecall));
  if (TagOf(*inputitem) == S_COLON2)
    {
      const SOURCEPOSN locn = LocnOf(*inputitem);
      if (has_variableio) /* variableio 14/3/91 */
        /*{{{  Use vin instruction*/
        {
          /*
          Simplistic version to start with;
          we do VIN with maxsize = size of the type of the rhs.
          We then do LDCNT and store to a temporary.
          We then divide this by the size of one element of the array, to get
          the count. This is converted to the correct type, and assigned
          to the count expression.

          The important bit to note is that we MUST NOT timeslice between
          the VIN and the LDCNT. In the 'usecall' case, the VIN actually
          returns the count.
          So the philosophy is to always execute the LDCNT immediately after
          the VIN.

          Modify the tree to look like:

                   S_COLON2
                   /     \
              S_LIST     rhs
              /    \
           maxsize S_LIST   -   S_LIST   -   S_LIST   -   NULL
                     |            |            |
                   temp         expression    countexp

          where expression is Type-convert(temp / (SIZE array[0]))

          Technically this should be an unsigned divide, because we can
          communicate more than half the address space, so the number of
          bytes transferred is actually an unsigned number.
          However, with this compiler it is impossible to declare an array
          of such a size; hence any such large number is impossible.
          Therefore a normal division is ok.
          Note that the optimisation which inserts a shift will automatically
          perform an unsigned divide.

          Optimisation: if the count exp is simple enough, and the divide
          is simple enough, don't generate the temporary, but simply
          do the divide and store directly in the code generator.
          We mark this by inserting a CONSTEXP node where the expression should
          be, with the size of the divide to be performed,
          and temp is left NULL.
          */

          treenode **countexp   = LeftOpAddr(*inputitem);
          treenode **arrayexp   = RightOpAddr(*inputitem);
          const int counttype   = ntypeof(*countexp);
          treenode *maxsize     = scaletreeof(gettype(*arrayexp), 1, be_lexlevel);
          treenode *temp        = NULL;
          treenode *expression  = NULL;
          treenode *arraytype   = ARTypeOf(gettype(*arrayexp));
          const INT32 b = bytesin(arraytype);
          const int exponent = whichpowerof2(b);

          mapload3regs(P_PTR, arrayexp, channelmode, channel, P_EXP, &maxsize);
          mapioop(I_VIN, usecall);

          /* decide whether we can optimise the storage so that we don't need
             to store into a temporary
          */
          if ((b != (-1)) /* size of rhs is constant */
           && (fitsinword(counttype) || isdoublelength(counttype))
           /*&& issimplelocal(*countexp))*/
           && (regsforstore(*countexp) < MAXREGS))
            /*{{{  Mark the simple case*/
            {
              expression = newconstant(b);
              if (exponent <= 1) mapexp(&expression);
              /* and leave temp = NULL */
              mapstoreinopd(P_EXP, countexp);
            }
            /*}}}*/
          else
            /*{{{  Build expression to divide down the byte count*/
            {
              temp = gettypedtemp(NULL, NM_WORKSPACE, newleafnode(S_INT, locn));
              if (b == 1)
                expression = temp;
              else if (exponent > 0)
                expression = newdopnode(S_RSHIFT, locn, temp, newconstant(exponent), S_INT);
              else /* either b == -1, or b is not a small power of 2 */
                expression = newdopnode(S_DIV, locn, temp, scaletreeof(arraytype, 1, be_lexlevel), S_INT);

              if (isquadlength(counttype))
                expression = makeconversion(S_INT, counttype, expression, S_EXACT, locn);
              else if (!istargetintsize(counttype))
                expression = newmopnode(S_EXACT, locn, expression, counttype);

              upusecount(temp, 1);
              if (TagOf(temp) != T_TEMP)
                badtag(locn, TagOf(temp), "mapinputitem");
              SetTag(temp, T_PREEVALTEMP);

              mapsimpleassign(counttype, P_EXP, countexp, P_EXP, &expression);
              freetemp(temp);
            }
            /*}}}*/

          *countexp = addtofront(maxsize,
                      addtofront(temp,
                      addtofront(expression,
                      addtofront(*countexp,  NULL))));
        }
        /*}}}*/
      else
        /*{{{  do counted input*/
        {
          treenode **countexp = LeftOpAddr(*inputitem),
                   **arrayexp = RightOpAddr(*inputitem);
          int countmode = P_EXP;
          const int counttype = ntypeof(*countexp);
          treenode *checkedcount;
          /*{{{  declare a temporary to hold count, if necessary*/
          {
            if (!issimplelocal(*countexp, be_lexlevel))
              /*{{{  input to a temporary; move to the destination*/
              {
                *countexp = gettemp(*countexp, NM_WORKSPACE);
                upusecount (*countexp, 1);
                countmode = P_TEMP;
                /*{{{  input to a temporary*/
                mapinputitem(channelmode, channel, P_TEMPPTR, countexp, NULL);
                /*}}}*/
                /*{{{  move to the destination*/
                mapsimpleassign(counttype, P_EXP, NDeclAddr(*countexp), P_TEMP, countexp);
                /*}}}*/
              }
              /*}}}*/
            else
              /*{{{  input to the destination*/
              {
                countmode = P_EXP;
                mapinputitem(channelmode, channel, P_PTR, countexp, comm);
              }
              /*}}}*/
            if (isshortint(counttype))
              upusecount(*countexp, 2); /* sign extension */
          }
          /*}}}*/
          /*{{{  convert a long int to int if neccessary*/
          /* It will be in a temp if it isn't a local simple BYTE, or short INT */
          /* The only conversions that need checking are long ints to int */
          /* No mapping required */ /* WRONG!!!!!! */
          /*{{{  map the conversion (check) if a quad word*/
          /* Added for bug 1130 29/1/91 */
          if (isquadlength(counttype))
            {
              const int old = switch_to_temp_workspace();
              treenode *conversion = makeconversion(counttype, S_INT, *countexp, S_EXACT, locn);
              switch_to_prev_workspace(old);
              mapexp(&conversion);
            }
          /*}}}*/
          /*}}}*/
          /*{{{  if count = 0 goto skiplab*/
          /* No mapping required: we've already mapped the count */
          /*}}}*/
          /*{{{  add rangecheck to count expression if necessary*/
          if (RANGECHECKING)
            {
              const int old = switch_to_temp_workspace();
              checkedcount = newdopnode(S_CCNT1, NOPOSN, *countexp, dimexpof(*arrayexp, 0), S_INT);
              switch_to_prev_workspace(old);
            }
          else
            checkedcount = *countexp;
          /*}}}*/
          /*{{{  input array from 0 for count*/
          {
            treenode *lengthexp;
            /*{{{  lengthexp = checkedcount * scale*/
            {
              const int old = switch_to_temp_workspace();
              /* 10 is just a dummy value as we don't know the real size,
                 and it doesn't matter what it is anyway, just as long
                 as the lengthexp takes two registers to load */
              lengthexp = newdopnode(S_TIMES, NOPOSN, checkedcount, newconstant(10), S_INT);
              switch_to_prev_workspace(old);
            }
            /*}}}*/
            mapload3regs(P_PTR, arrayexp, channelmode, channel, P_EXP, &lengthexp);
            mapioop(I_IN, usecall);
          }
          /*}}}*/
          /*{{{  free the temporary holding count, if necessary*/
          if (countmode == P_TEMP)
            freetemp(*countexp);
          /*}}}*/

        }
        /*}}}*/
    }
  else
    {
      treenode *type = gettype(*inputitem);
      if (bytesin(type) == (-1))
        /*{{{  length is an expression*/
        {
          int loadseq;
          treenode *lengthexp;

          /*const int old = switch_to_temp_workspace();*/
          lengthexp = scaletreeof(type, 1, be_lexlevel);
          /*switch_to_prev_workspace(old);*/

          loadseq = mapload3regs(inputitemmode, inputitem, channelmode, channel,
                                 P_EXP, &lengthexp);
          /* Ideally we should save this loadseq value on the tree somewhere
             so that we don't have to recalculate it in the code generator */
          {
            /* bug TS/1626 24/04/92 - save the length expression */
            treenode *rhs = RHSOf(comm);
            NewNextItem(addtofront(lengthexp, NULL), rhs);
          }
        }
        /*}}}*/
      else
        mapload2regs(inputitemmode, inputitem, channelmode, channel);
      mapioop(I_IN, usecall);
    }
}
/*}}}*/
/*{{{  PUBLIC void mapoutput(tptr)*/
/*****************************************************************************
 *
 *  mapoutput allocates workspace and temporaries for an OUTPUT node
 *
 *****************************************************************************/
PUBLIC void mapoutput ( treenode *tptr )
{
  treenode *channel      = LHSOf(tptr);
  treenode *outputlist   = RHSOf(tptr);
  const int noutputitems = listitems(outputlist);
  treenode *channeltype  = gettype(channel);
  DEBUG_MSG(("mapoutput\n"));
  if (TagOf(channeltype) != S_CHAN)
    /*{{{  handle a PORT output*/
    {
      treenode **oldtypeaddr; /* INSdi03067 */
      treenode *const porttype = fixporttype(channel, &oldtypeaddr);
      const int type = ntypeof(channel);
      treenode **outputitem = ThisItemAddr(outputlist);

      /* If we're using device access instructions for ports, we want to make
         sure that we don't try anything fancy; eg FPU arith.
         or conversions from real->int which try to store directly from
         an FPU instruction.
         Thus we insert a temp. This may be less than optimal, but at least
         it is correct.
         We have to do this for 'long' arithmetic too, because any lib calls
         will otherwise pass the actual result (the port) as a pointer,
         and the body of the function doesn't know it is supposed to be
         doing device access to it.
         This is not necessary for short arith. because this is returned
         on the integer stack.
         Conor O'Neill: 08/07/92.
      */

      if (tx_global.hasdevaccess
        && !T9000_alpha_nodevaccess(&tx_global)
        && !isaddressable(*outputitem) && !isconst(*outputitem)
       && (   isreal(type)
           || isdoublelength(type)
           || isquadlength(type)
        /* || (isshortint(type) && !use_shortintops)*/ /* not necessary */
           || (has_fpu_core && isconversion(*outputitem) && isreal(ntypeof(OpOf(*outputitem))))
          )
         )
        {
          *outputitem = gettemp(*outputitem, NM_WORKSPACE);
          upusecount(*outputitem, 1);
          mapsimpleassign(type, P_TEMP, outputitem, P_EXP, NDeclAddr(*outputitem));
          mapsimpleassign(type, P_EXP,  LHSAddr(tptr), P_TEMP, outputitem);
          freetemp(*outputitem);
        }
      else
        mapsimpleassign(type, P_EXP, LHSAddr(tptr), P_EXP, outputitem);
      restoreporttype(LHSOf(tptr), porttype, oldtypeaddr);
    }
    /*}}}*/
  else
    /*{{{  handle a channel output*/
    {
      int channelmode = chanaspointer ? P_EXP : P_PTR;
      int channeltempmode = tempmodeof(channelmode);

      const BOOL counted_array = TagOf(ThisItem(outputlist)) == S_COLON2;
      BOOL create_channel_temp = !issimplelocal(channel, be_lexlevel);
      if (ActionTypeOf(tptr) == NULL) /* bug TS/1626 24/4/92 */
        create_channel_temp &=
              ((noutputitems > 1) || (counted_array && !has_variableio));
      else
        create_channel_temp &= counted_array && !has_variableio;

      if (create_channel_temp)
        /*{{{  preevaluate pointer to channel into temporary before doing outputs*/
        {
          DEBUG_MSG(("mapoutput: preevaluating pointer to chan\n"));
          mapexpopd(channelmode, LHSAddr(tptr));
          channel = getopdtemp(channelmode, channel);
          SetLHS(tptr, channel);
          upusecount (channel, 1);
          channelmode = tempmodeof(channelmode);
        }
        /*}}}*/
      if (ActionTypeOf(tptr) != NULL) /* bug TS/1626 24/04/92 */
        mapoutputitem(channelmode, LHSAddr(tptr), P_EXP, ThisItemAddr(outputlist), tptr);
      else
        while (!EndOfList(outputlist))
          /*{{{  map this outputitem, move to next*/
          {
            mapoutputitem(channelmode, LHSAddr(tptr), P_EXP, ThisItemAddr(outputlist), tptr);
            outputlist = NextItem(outputlist);
          }
          /*}}}*/
      if (channelmode == channeltempmode)
        freetemp(channel);
    }
    /*}}}*/
}
/*}}}*/
/*{{{  PUBLIC void mapinput(tptr)*/
/*****************************************************************************
 *
 *  mapinput allocates workspace and temporaries for an INPUT node
 *
 *****************************************************************************/
PUBLIC void mapinput ( treenode *tptr )
{
  if (TagOf(tptr) != S_SKIP)
    {
      treenode *channel = LHSOf(tptr);
      treenode **channeladdr = LHSAddr(tptr);
      treenode *inputlist = skipspecifications(RHSOf(tptr));
      int channelmode = chanaspointer ? P_EXP: P_PTR;
      int channeltempmode = tempmodeof(channelmode);

      switch(inputtypeof(tptr))
        {
          /*{{{  channel input*/
          case INP_INPUT:
            {
              const BOOL counted_array = TagOf(ThisItem(inputlist)) == S_COLON2;
              BOOL create_channel_temp = !issimplelocal(channel, be_lexlevel);
              if (ActionTypeOf(tptr) == NULL) /* bug TS/1626 24/4/92 */
                create_channel_temp &=
                      ((listitems(inputlist) > 1) || (counted_array && !has_variableio));
              else
                create_channel_temp &= counted_array && !has_variableio;

              datasize = max_INT32(datasize, DS_IO);

              if (create_channel_temp)
                /*{{{  preevaluate pointer to channel into temporary before doing outputs*/
                {
                  mapexpopd(channelmode, channeladdr);
                  channel = getopdtemp(channelmode, channel);
                  SetLHS(tptr, channel);
                  upusecount(channel, 1);
                  channelmode = tempmodeof(channelmode);
                }
                /*}}}*/
              if (ActionTypeOf(tptr) != NULL) /* bug TS/1626 24/04/92 */
                mapinputitem(channelmode, LHSAddr(tptr), P_PTR, ThisItemAddr(inputlist), tptr);
              else
                while (!EndOfList(inputlist))
                  /*{{{  map this inputitem, move to next*/
                  {
                    mapinputitem(channelmode, LHSAddr(tptr), P_PTR, ThisItemAddr(inputlist), tptr);
                    inputlist = NextItem(inputlist);
                  }
                  /*}}}*/
              if (channelmode == channeltempmode)
                freetemp(channel);
            }
            break;
          /*}}}*/
          /*{{{  read timer*/
          case INP_TIMER_INPUT:
            mapexp(channeladdr); /* Just to mark the timer as used, also for subscript checks */
            mapaddr(ThisItemAddr(inputlist));
            break;
          /*}}}*/
          /*{{{  delayed input*/
          case INP_DELAYED_INPUT:
            datasize = max_INT32(datasize, DS_WAIT);
            mapexp(channeladdr);
            inputlist = RHSOf(tptr);
            mapexp(&inputlist);
            /*{{{  T9000_alpha_badrunp*/
            if (T9000_alpha_badrunp(&tx_global))
              {
                treenode *reserved;
                const BOOL need_temp = !issimple(inputlist);
                if (need_temp)
                  {
                    inputlist = gettemp(inputlist, NM_WORKSPACE);
                    upusecount(inputlist, 1);
                  }
                reserved = alloc_fixedws(0, 1);
                kill_var(reserved);
                if (need_temp)
                  {
                    freetemp(inputlist);
                    SetRHS(tptr, inputlist);
                  }
              }
            /*}}}*/
            break;
          /*}}}*/
          /*{{{  port input*/
          case INP_PORT_INPUT:
            {
              treenode **oldtypeaddr; /* INSdi03067 */
              treenode *const porttype = fixporttype(channel, &oldtypeaddr);
              mapsimpleassign(ntypeof(channel),
                              P_EXP, ThisItemAddr(inputlist),
                              P_EXP, channeladdr);
              restoreporttype(*channeladdr, porttype, oldtypeaddr);
            }
            break;
          /*}}}*/
          /*{{{  case input*/
          #if 0 /* this is now transformed into something simpler */
          /*{{{  comment     What the tree looks like after mapping*/
          /*
                                S_CASE_INPUT
                                 /      \
          (holds channel ptr) T_TEMP   T_TEMP (holds inputted tag)
                                          \
                                        S_LIST
                                        /    \
                                   S_VARIANT  S_LIST
                                   /             .
                                S_LIST            .
                                /    \
              (const val S_CONSTEXP    .
               of tag)      /           .
                      N_TAGDEF

          N.B. The temporary inserted in front of the channel is optional, and only
               inserted if the channel is not a simple local.
               There is ALWAYS a temporary inserted in front of the variant list, to
               hold the tag.
          */
          /*}}}*/
          case INP_CASE_INPUT:
            datasize = max_INT32(datasize, DS_IO);
            {
              if (!issimplelocal(channel, be_lexlevel))
                /*{{{  preevaluate pointer to channel into temporary before doing inputs*/
                {
                  mapexpopd(channelmode, channeladdr);
                  SetLHS(tptr, getopdtemp(channelmode, channel));
                  channelmode = tempmodeof(channelmode);
                  channel = LHSOf(tptr);
                  upusecount(channel, 1);
                }
                /*}}}*/
              /*{{{  load tag into tag temporary*/
              SetRHS(tptr, gettypedtemp(RHSOf(tptr), NM_WORKSPACE, newleafnode(S_BYTE, NOPOSN)));
              upusecount(RHSOf(tptr), 1);
              mapexpopd(channelmode, LHSAddr(tptr));
              mapioop(I_IN, iobycall && maybevirtualchan(channel));
              freetemp(RHSOf(tptr));
              /*}}}*/
              /*{{{  map the body*/
              {
                int savedglobchannelmode = globchannelmode;
                treenode **savedglobchannelptr = globchannelptr;

                /* Set up global variables for mapvariant */
                globchannelmode = channelmode;
                globchannelptr = LHSAddr(tptr);
                mapconstruction(inputlist, mapprocess);

                globchannelptr = savedglobchannelptr;
                globchannelmode = savedglobchannelmode;
              }
              /*}}}*/
              if (channelmode == channeltempmode)
                freetemp(channel);
            }
            break;
          #endif
          /*}}}*/
          /*{{{  tagged input*/
          #if 0 /* this is now transformed into something simpler */
          /*{{{  comment     What the tree looks like after mapping*/
          /*
                                S_TAGGED_INPUT
                                 /      \
          (holds channel ptr) T_TEMP   T_TEMP (holds inputted tag)
                                          \
                                        S_LIST
                                        /    \
                      (const val S_CONSTEXP  other input items
                       of tag)      /           .
                               N_TAGDEF          .

          N.B. The temporary inserted in front of the channel is optional, and only
               inserted if the channel is not a simple local.
               There is ALWAYS a temporary inserted in front of the input item list, to
               hold the tag.
          */
          /*}}}*/
          case INP_TAGGED_INPUT:
            datasize = max_INT32(datasize, DS_IO);
            {
              if (!issimplelocal(channel, be_lexlevel))
                /*{{{  preevaluate pointer to channel into temporary before doing inputs*/
                {
                  mapexpopd(channelmode, channeladdr);
                  SetLHS(tptr, getopdtemp(channelmode, channel));
                  channelmode = tempmodeof(channelmode);
                  channel = LHSOf(tptr);
                  upusecount(channel, 1);
                }
                /*}}}*/
              /*{{{  load tag into tag temporary*/
              SetRHS(tptr, gettypedtemp(RHSOf(tptr), NM_WORKSPACE, newleafnode(S_BYTE, NOPOSN)));
              mapexpopd(channelmode, LHSAddr(tptr));
              mapioop(I_IN, iobycall && maybevirtualchan(channel));
              upusecount(RHSOf(tptr),1);
              freetemp(RHSOf(tptr));
              /*}}}*/

              inputlist = NextItem(inputlist); /* Move past the tag */
              while (!EndOfList(inputlist))
                /*{{{  map this inputitem, move to next*/
                {
                  mapinputitem(channelmode, LHSAddr(tptr), P_PTR, ThisItemAddr(inputlist), tptr);
                  inputlist = NextItem(inputlist);
                }
                /*}}}*/
              if (channelmode == channeltempmode)
                freetemp(channel);
            }
            break;
          #endif
          /*}}}*/
          default:
            badtag(LocnOf(tptr), TagOf(tptr), "mapinput");
        }
    }
}
/*}}}*/
/*{{{  PUBLIC void set_alt_ds*/
PUBLIC void set_alt_ds(treenode *const tptr)
{
  alt_ds = timerguardinalt(tptr) ? DS_WAIT : DS_IO;
}
/*}}}*/
/*{{{  PUBLIC void mapalt(tptr)*/
PUBLIC void mapalt ( treenode *tptr )
{
  /* The following mapping does the enables and disables */
  /* Reserve wsp0 from the enables and disables.
      During enabling it may be used as a temporary to trap degenerate
     ALTs, during disabling it is used by the disabling instructions.
   */
  const BOOL saved_premapped_alt  = premapped_alt;
  const int saved_replaltstacklen = replaltstacklen;
  treenode *const saved_replaltvector = replaltvector;
  treenode *reserved, *replvector, *type;
  const int repls = countnestedalts(tptr, NESTED_REPLS);
  type = newtypenode(S_ARRAY, LocnOf(tptr), newconstant(repls),
                      newleafnode(S_INT, LocnOf(tptr)));
  SetARDim(type, repls);
  DEBUG_MSG(("mapalt: nested repls is %d\n", repls));
  replvector = gettypedtemp(tptr, NM_WORKSPACE, type);
  if (nodetypeoftag(TagOf(tptr)) == CNODE)
    SetCTemp(tptr, replvector);
  else
    SetReplCTemp(tptr, replvector);
  replaltvector = replvector;
  reserved = alloc_fixedws(0, 1);
  insidealtguard = TRUE;
  alt_repl_scope_level = -1; /* bug TS/1893 14/01/93 */
  set_alt_ds(tptr);

  datasize = max_INT32(datasize, alt_ds); /* bug TS/1774 30/07/92 */

  replaltstacklen = 0; /* used while mapping for usage count of replaltvector */
  premapped_alt = TRUE; /* indicate that mappreprocess has already been called */
  mapaltenables(tptr);
  kill_var(reserved);
  insidealtguard = FALSE;
  /* Map ALT bodies above ALT enables */
  mapaltbodies(tptr);
  freetemp(replvector);

  replaltvector   = saved_replaltvector;
  replaltstacklen = saved_replaltstacklen;
  premapped_alt   = saved_premapped_alt;
}
/*}}}*/

/*{{{  PRIVATE void tlengthcheck*/
PRIVATE void tlengthcheck(treenode *const lengthexp, treenode *const prottype)
{
  if ((TagOf(prottype) == S_ANY) && /* bug 412 29/8/90 */
      RANGECHECKING) /* bug 1139 4/2/91 */
    {
      /* Check for zero length communications on a CHAN OF ANY */
      texp(lengthexp, MANY_REGS);
      genprimary(I_EQC, 0);
      genprimary(I_LDC, 1);
      gensecondary(I_CSUB0);
      gencomment0("check for zero length comm");
      throw_the_result_away ();
    }
}
/*}}}*/
/*{{{  PRIVATE void tsignextend*/
PRIVATE void tsignextend(treenode *const nptr, const BOOL already_zeroed)
{
  if (RANGECHECKING || !already_zeroed)
    {
      loadname(nptr, 0);
      if (use_shortintops) /* variableio 12/3/91 */
        /* When half word support is inserted properly, we can junk all this
           stuff, and simply insert type conversions into the tree which
           will then generate the most efficient code anyway.
           Until then, this is a quick bodge which improves iserver protocol...
           CON - 12/3/91
        */
        gensecondary(I_XSWORD);
      else
        {
          if (!already_zeroed)
            {
              loadconstant(typemask(S_INT16));
              gensecondary(I_AND);
            }
          if (RANGECHECKING)
            {
              /* if not range checking, we assume that the count is never negative! */
              if (T9000_alpha_noxsword(&tx_global)) /* OK if masked first */
                gensecondary(I_XSWORD);
              else
                {
                  genwidenshort ();
                }
            }
        }
      storeinname(nptr, 0);
    }
}
/*}}}*/
/*{{{  PRIVATE void toutputitem*/
/*{{{  comment*/
/*****************************************************************************
 *
 *  toutputitem generates code to output (outputitemmode, outputitem) on
 *              channel (channelmode, channel)
 *
 *****************************************************************************/
/*}}}*/
PRIVATE void toutputitem (const int channelmode,    treenode *const channel,
                                int outputitemmode, treenode *const outputitem,
                          treenode *const prottype, treenode *const comm)
{
  const BOOL usecall = iobycall && maybevirtualchan(channel);
  DEBUG_MSG(("toutputitem: usecall is %d\n", usecall));
  if (TagOf(outputitem) == S_COLON2)
    {
      if (has_variableio) /* variableio 11/3/91 */
        /*{{{  use vout instruction*/
        {
          treenode *countexp = LeftOpOf(outputitem);
          treenode *arrayexp = RightOpOf(outputitem);
          BOOL preeval_e2, preeval_e3;
          int loadseq;

          tpreexp(arrayexp); /* This wasn't done earlier for S_COLON2 */

          loadseq = giveorder(P_PTR, arrayexp, channelmode, channel, P_EXP, countexp,
                              &preeval_e2, &preeval_e3);
          tload3regs(P_PTR, arrayexp, channelmode, channel, P_EXP, countexp, loadseq, TRUE);
          checkerror();
          tioop(I_VOUT, usecall);
        }
        /*}}}*/
      else
        /*{{{  do counted output*/
        {
          treenode *countexp = LeftOpOf(outputitem),
                   *arrayexp = RightOpOf(outputitem);
          int countmode = P_EXP;
          const int constantcount = isconst(countexp);
          int counttype = ntypeof(countexp);
          treenode *origtype = NTypeOf(countexp); /* bug 864 29/1/91 */
          int skiplab = (-1); /* initialised to shut up gcc's optimiser */
          treenode *dimexp = dimexpof(arrayexp, 0);
          BOOL zeroed_temp = FALSE;
          treenode *leftprottype, *rightprottype;
          /*{{{  set up leftprottype and rightprottype*/
          if (TagOf(prottype) == S_ANY)
            {
              leftprottype  = prottype;
              rightprottype = prottype;
            }
          else
            {
              leftprottype  = LeftOpOf(prottype);
              rightprottype = RightOpOf(prottype);
            }
          /*}}}*/
          if (preeval(countmode, countexp))
            /*{{{  preevaluate count to a temporary*/
            {
              if (!use_shortintops
                  && isshortint(counttype)  /* bug 706 13/9/90 */
                     /* bug 1294: don't zero var if it will be optimised to a word move */
                     /* (Same tests as in moveopd) */ /* bug 1294 11/6/91 */
                  && (isaddressableopd(P_EXP, NDeclOf(countexp))
                    &&  needtemptoload(P_EXP, NDeclOf(countexp))))
                { zero_local_var(countexp); zeroed_temp = TRUE; }
              tsimpleassign(counttype, P_TEMP, countexp, P_EXP, NDeclOf(countexp), MANY_REGS);
              /*SetTag(countexp, T_PREEVALTEMP);*/
              temp_mark_as_evaluated(countexp); /* bug TS/1900 22/10/92 */
              countmode = P_TEMP;
            }
            /*}}}*/
          /*{{{  output the count*/
          toutputitem(channelmode, channel, countmode, countexp, leftprottype, NULL);
          /*}}}*/
          /*{{{  skip if count is 0, check count against size of array*/
          if (!constantcount)
            {
              /* It will be in a temp if it isn't a local simple BYTE, or short INT */
              /* The only conversions that need checking are long ints to int */

              if (isshortint(counttype)) /* bug 706 13/9/90 */
                /* must be a simplelocal, or would have been mapped to a temporary */
                tsignextend(countexp, zeroed_temp);
              else if (!fitsinword(counttype))
                /*{{{  convert long int temporary to int temporary*/
                {
                  if (RANGECHECKING)
                    {
                      treenode *checktree;
                      const int old = switch_to_temp_workspace();
                      if (isquadlength(counttype)) /* bug 1130 29/1/91 */
                        checktree = makeconversion(counttype, S_INT, countexp, S_EXACT, LocnOf(outputitem));
                      else
                        checktree = newmopnode(S_EXACT, NOPOSN, countexp, S_INT);
                      switch_to_prev_workspace(old);
                      texp(checktree, MANY_REGS);
                      checkerror(); /* bug 1202 11/3/91 */
                    }
                  if (intnodeptr == NULL)
                    {
                      const int old = switch_to_real_workspace();
                      intnodeptr = newleafnode(S_INT, NOPOSN);
                      switch_to_prev_workspace(old);
                    }
                  SetNType(countexp, intnodeptr); /* bug 864 - see below */
                }
                /*}}}*/
              texpopd(countmode, countexp, MANY_REGS);
              skiplab = newlab();
              genbranch(I_CJ, skiplab);
            }
          /*}}}*/
          /*{{{  preprocess the arrayexp*/
          /* The array expression may contain segments whose base and length
             cannot be calculated until the count has been inputted.  */
          DEBUG_MSG(("\ntoutputitem: COLON2: calling preexp on arrayexp"));
          tpreexp(arrayexp);
          /*}}}*/
          /*{{{  output the array from 0 for count*/
          if /* count is variable or count is constant but not zero */
            (!constantcount || (LoValOf(countexp) != 0))
              /*{{{  output the array from 0 for count*/
              {
                treenode *arraytype = ARTypeOf(gettype(arrayexp)); /* bug 1205 13/3/91 */
                const INT32 b = bytesin(arraytype);
                treenode *checkedcount;
                int checkcount = FALSE;
                /*{{{  add range checking to countexp if neccessary*/
                if (RANGECHECKING && (!constantcount || !isconst(dimexp)))
                  {
                    const int old = switch_to_temp_workspace();
                    DEBUG_MSG(("\ntoutputitem: COLON2: inserting range check"));
                    checkedcount = newdopnode(S_CCNT1, NOPOSN, countexp, dimexp, S_INT);
                    switch_to_prev_workspace(old);
                    checkcount = TRUE;
                  }
                else
                  {
                    DEBUG_MSG(("\ntoutputitem: COLON2: no range check"));
                    checkedcount = countexp;
                  }
                /*}}}*/
                if (constantcount)
                  /*{{{  ldptr array; ldptr channel; ldc length*/
                  {
                    DEBUG_MSG(("\ntoutputitem: COLON2: calling texp on checkedcount"));
                    if (checkcount)
                      texp(checkedcount, MANY_REGS);
                    DEBUG_MSG(("\ntoutputitem: COLON2: calling tload2regs"));
                    tload2regs(P_PTR, arrayexp, channelmode, channel, FALSE, TRUE);
                    loadconstant(b * LoValOf(countexp));
                  }
                  /*}}}*/
                else
                  /*{{{  ldptr array; ldptr channel; ld length*/
                  {
                    treenode *lengthexp;
                    BOOL preeval_e2, preeval_e3;
                    int loadseq;
                    /*{{{  lengthexp = checkedcount * scale*/
                    if (b != 1)
                      {
                        int old = switch_to_temp_workspace();
                      #if 0 /* bug 1205 13/3/91 */
                        lengthexp = newdopnode(S_TIMES, NOPOSN, checkedcount,
                                               newconstant(b), S_INT);
                      #else
                        lengthexp = newdopnode(S_TIMES, NOPOSN, checkedcount,
                                               scaletreeof(arraytype, 1, be_lexlevel), S_INT);
                      #endif
                        switch_to_prev_workspace(old);
                      }
                    else
                      lengthexp = checkedcount;
                    /*}}}*/
                    DEBUG_MSG(("\ntoutputitem: COLON2: calling tload3regs"));
                    loadseq = giveorder(P_PTR, arrayexp, channelmode, channel, P_EXP, lengthexp,
                                          &preeval_e2, &preeval_e3);
                    tload3regs(P_PTR, arrayexp, channelmode, channel, P_EXP, lengthexp, loadseq, TRUE);
                  }
                  /*}}}*/
                checkerror(); /* bug 1202 11/3/91 */
                tioop(I_OUT, usecall);
              }
              /*}}}*/
          /*}}}*/
          SetNType(countexp, origtype); /* bug 864 */
          if (!constantcount)
            setlab(skiplab);
        }
        /*}}}*/
    }
  else
    {
      /* bug INSdi02071 - just do this here? */
      outputitemmode = simplify(outputitemmode, outputitem);
      {
      /* Bug 1203 - 12/3/91 - don't use outbyte and/or outword if the
         output item is already addressable.
         Unlike in the mapping phase, if the value is a scalar constant
         and has been put into a constant table, then isaddressable
         will return TRUE, and everything else falls out correctly.
      */
      /* Bug TS/1504 - 07/04/92:
         If we're optimising for space, DO use the shorter forms if possible.
      */
      /*const int addressable = isaddressableopd(outputitemmode, outputitem);*/
      /*const BOOL useout = isaddressableopd(outputitemmode, outputitem);*/
      const BOOL useout = ((code_style_flags & CODE_STYLE_SPACE_NOT_TIME) == 0) &&
                          isaddressableopd(outputitemmode, outputitem);
      const int t = ntypeof(outputitem);
      if (!useout && /* bug 1203 12/3/91 */
          istargetintsize(t))
        /*{{{  ldptr channel; ld outputitem; outword*/
        {
          tload2regs(channelmode, channel, outputitemmode, outputitem, FALSE, TRUE);
          checkerror(); /* bug 1202 11/3/91 */
          tioop(I_OUTWORD, usecall);
        }
        /*}}}*/
      else if (!useout && /* bug 1203 12/3/91 */
               istargetbytesize(t))
        /*{{{  ldptr channel; ld outputitem; outbyte*/
        {
          tload2regs(channelmode, channel, outputitemmode, outputitem, FALSE, TRUE);
          checkerror(); /* bug 1202 11/3/91 */
          tioop(I_OUTBYTE, usecall);
        }
        /*}}}*/
      else
        /*{{{  ldptr outputitem; ldptr channel; ld length; out*/
        {
          treenode *const type = gettype(outputitem);

          const INT32 s = bytesin(type);
          /*{{{  preevaluate the outputitem if necessary*/
          #if 0 /* bug 743 makes this evaluate twice 21/9/90 */
          if (preeval(outputitemmode, outputitem))
            {
              int tmode = tempmodeof(outputitemmode);
              tsimpleassign(TagOf(type), tmode, outputitem, outputitemmode,
                            NDeclOf(outputitem), MANY_REGS);
              outputitemmode = tmode;
            }
          #else
          outputitemmode = simplify(outputitemmode, outputitem);
          #endif
          /*}}}*/
          outputitemmode = ptrmodeof(outputitemmode);
          if (s == (-1))
            /*{{{  length is an expression*/
            {
              /*{{{  check hidden dimensions*/
              if ((TagOf(prottype) != S_ANY) && RANGECHECKING)
                tcheckdimensions(type, prottype);
              /*}}}*/
              /*{{{  load operands for an 'out' instruction*/
              {
                BOOL preeval_e2, preeval_e3;
                int loadseq;
                treenode *lengthexp;
                int lengthmode = P_EXP;
              #if 0
                /* bug TS/1626 24/04/92 */
                int old = switch_to_temp_workspace();
                lengthexp = scaletreeof((TagOf(prottype) != S_ANY) ? prottype : type,
                                        1, be_lexlevel);
                switch_to_prev_workspace(old);
              #else
                lengthexp = ThisItem(NextItem(RHSOf(comm)));
                lengthmode = simplify(lengthmode, lengthexp);
              #endif

                /* We only do this call so that we have a loadseq parameter for
                   tload3regs, ideally we should make the binder save loadseq
                   in the tree. */
                loadseq = giveorder(outputitemmode, outputitem,
                                    channelmode, channel,
                                    lengthmode, lengthexp, &preeval_e2, &preeval_e3);
                tlengthcheck(lengthexp, prottype);
                tload3regs(outputitemmode, outputitem, channelmode, channel,
                           lengthmode, lengthexp, loadseq, TRUE);
              }
              /*}}}*/
            }
            /*}}}*/
          else
            {
              tload2regs(outputitemmode, outputitem, channelmode, channel, FALSE, TRUE);
              genprimary(I_LDC, s);
            }
          checkerror(); /* bug 1202 11/3/91 */
          tioop(I_OUT, usecall);
        }
        /*}}}*/
      }
    }
}
/*}}}*/
/*{{{  PUBLIC void tinputitem*/
/*{{{  comment*/
/*****************************************************************************
 *
 *  tinputitem generates code to input 'inputitem' on channel
 *              (channelmode, channel)
 *
 *****************************************************************************/
/*}}}*/
PUBLIC void tinputitem (const int channelmode,   treenode *const channel,
                        const int inputitemmode, treenode *const inputitem,
                        treenode *const prottype, treenode *const comm)
{
  const BOOL usecall = iobycall && maybevirtualchan(channel);
  DEBUG_MSG(("tinputitem: usecall is %d\n", usecall));
  if ((TagOf(inputitem) == S_COLON2) && has_variableio) /* variableio 14/3/91 */
    /*{{{  Use the VIN instruction*/
    {
      treenode *arrayexp   = RightOpOf(inputitem);
      treenode *lhs        = LeftOpOf(inputitem);
      treenode *maxsize    = ThisItem(lhs); /* Max size of input in bytes */
      treenode *temp       = ThisItem(NextItem(lhs)); /* Temp to store the byte count */
      treenode *expression = ThisItem(NextItem(NextItem(lhs)));
      treenode *count      = ThisItem(NextItem(NextItem(NextItem(lhs))));
      const int counttype  = ntypeof(count);
      BOOL preeval_e2, preeval_e3;
      int loadseq;

      tpreexp(arrayexp); /* This wasn't done earlier for S_COLON2 */

      loadseq = giveorder(P_PTR, arrayexp, channelmode, channel, P_EXP, maxsize,
                          &preeval_e2, &preeval_e3);
      tload3regs(P_PTR, arrayexp, channelmode, channel, P_EXP, maxsize, loadseq, TRUE);
      checkerror();
      tioop(I_VIN, usecall);
      if (!usecall) gensecondary(I_LDCNT);
      checkerror();
      if (temp == NULL) /* optimised case */
        {
          /*{{{  Divide by the size of each array element*/
          {
            const INT32 b = LoValOf(expression);
            const int exponent = whichpowerof2(b);
            if (b == 1)
              ; /* skip */
            else if (exponent > 0)
              {
                 genprimary(I_LDC, exponent);
                 gensecondary(I_SHR);
              }
            else
              {
                texp(expression, MAXREGS-1); /* can only take 1 register cos constant */
                if (t450a_workarounds)
                  t450_divrem_workaround(I_DIV);
                else
                  gensecondary (I_DIV); 
              }
          }
          /*}}}*/
          /*{{{  Type convert and store*/
          if (isdoublelength(counttype))
            {
              gensecondary(I_XDBLE);
              storeinopd(P_EXP, count, 0, MAXREGS - 2);
              storeinopd(P_EXP, count, 1, MAXREGS - 1);
            }
          else
            {
              ttypeconversion(S_INT, counttype);
              storeinopd(P_EXP, count, 0, MAXREGS - 1);
            }
          /*}}}*/
        }
      else /* general case */
        {
          storeinname(temp, 0);
          tsimpleassign(counttype, P_EXP, count, P_EXP, expression, MANY_REGS);
        }
    }
    /*}}}*/
  else if (TagOf(inputitem) == S_COLON2)
    /*{{{  do counted input*/
    {
      treenode *countexp = LeftOpOf(inputitem),
               *arrayexp = RightOpOf(inputitem);
      treenode *leftprottype, *rightprottype;
      treenode *origexp  = countexp;          /* added for bug 864 */
      treenode *origtype = NTypeOf(countexp); /* ditto */
      int countmode = P_EXP;
      const int counttype = ntypeof(countexp);
      const int skiplab = newlab();

      /*{{{  set up leftprottype and rightprottype*/
      if (TagOf(prottype) == S_ANY)
        {
          leftprottype  = prottype;
          rightprottype = prottype;
        }
      else
        {
          leftprottype  = LeftOpOf(prottype);
          rightprottype = RightOpOf(prottype);
        }
      /*}}}*/
      /*{{{  input the count*/
      if (preeval(countmode, countexp))
        {
          countmode = P_TEMP;
          /* we initialise INT16 temporaries here, unlike normal, so that
             the sign extension goes better (bug 706 11/9/90) */
          if (/*istargetbytesize(counttype)*/ isshorttype(counttype))
            zero_local_var(countexp);
          /*{{{  input to the temporary*/
          tinputitem(channelmode, channel, P_TEMPPTR, countexp, leftprottype, NULL);
          /*}}}*/
          if (isshortint(counttype)) /* bug 706 11/9/90 */
            tsignextend(countexp, TRUE);
          /*{{{  move count to its real destination*/
          tsimpleassign(counttype, P_EXP, NDeclOf(countexp), P_TEMP, countexp, MANY_REGS);
          /*SetTag(countexp, T_PREEVALTEMP);*/
          temp_mark_as_evaluated(countexp); /* bug TS/1900 22/10/92 */

          /*}}}*/
        }
      else
        {
          countmode = P_EXP;
          /*{{{  input the count*/
          tinputitem(channelmode, channel, P_PTR, countexp, leftprottype, NULL);
          /*}}}*/
          if (isshortint(counttype)) /* bug 706 11/9/90 */
            /* must be a simplelocal, or would have been mapped to a temporary */
            tsignextend(countexp, FALSE);
        }
      /*}}}*/
      /*{{{  convert a long int to int if neccessary*/
      /* It will be in a temp if it isn't a local simple BYTE, or short INT */
      /* The only conversions that need checking are long ints to int */
      if (!fitsinword(counttype))
        /*{{{  convert long int temporary to int temporary*/
        {
          if (RANGECHECKING)
            {
              treenode *checktree;
              int old = switch_to_temp_workspace();
              if (isquadlength(counttype)) /* bug 1130 29/1/91 */
                checktree = makeconversion(counttype, S_INT, countexp, S_EXACT, LocnOf(inputitem));
              else
                checktree = newmopnode(S_EXACT, NOPOSN, countexp, S_INT);
              switch_to_prev_workspace(old);
              texp(checktree, MANY_REGS);
              checkerror(); /* bug 1202 11/3/91 */
            }
          if (intnodeptr == NULL)
            {
              const int old = switch_to_real_workspace();
              intnodeptr = newleafnode(S_INT, NOPOSN);
              switch_to_prev_workspace(old);
            }
          SetNType(countexp, intnodeptr); /* bug 864 - see below */
        }
        /*}}}*/
      /*}}}*/
      /*{{{  if count = 0 goto skiplab*/
      texpopd(countmode, countexp, MANY_REGS);
      genbranch(I_CJ, skiplab);
      /*}}}*/
      /*{{{  preprocess the arrayexp*/
      /* The arrray expression may contain segments whose base and length
         cannot be calculated until the count has been inputted.  */
      tpreexp(arrayexp);
      /*}}}*/
      /*{{{  add rangecheck to count expression if neccessary*/
      if (RANGECHECKING)
        {
          const int old = switch_to_temp_workspace();
          countexp = newdopnode(S_CCNT1, NOPOSN, countexp, dimexpof(arrayexp, 0), S_INT);
          switch_to_prev_workspace(old);
        }
      /*}}}*/
      /*{{{  input array from 0 for count*/
      {
        int loadseq;
        BOOL preeval_e2, preeval_e3;
        treenode *lengthexp;
        /*{{{  lengthexp = countexp * scale*/
        {
          treenode *arraytype = ARTypeOf(gettype(arrayexp));
          const INT32 b = bytesin(arraytype);
          if (b != 1)
            {
              const int old = switch_to_temp_workspace();
            #if 0 /* bug 1205 13/3/91 */
              lengthexp = newdopnode(S_TIMES, NOPOSN, countexp,
                                     newconstant(b), S_INT);
            #else
              lengthexp = newdopnode(S_TIMES, NOPOSN, countexp,
                                     scaletreeof(arraytype, 1, be_lexlevel), S_INT);
            #endif
              switch_to_prev_workspace(old);
            }
          else
            lengthexp = countexp;
        }
        /*}}}*/
        /* We only do this call so that we have a loadseq parameter for
           tload3regs, ideally we should make the binder save loadseq
           in the tree. */
        loadseq = giveorder(P_PTR, arrayexp, channelmode, channel,
                            P_EXP, lengthexp, &preeval_e2, &preeval_e3);
        tload3regs(P_PTR, arrayexp, channelmode, channel, P_EXP, lengthexp, loadseq, TRUE);
        checkerror(); /* bug 1202 11/3/91 */
        tioop(I_IN, usecall);
      }
      /*}}}*/
      /*{{{  clear up the mess*/
      /* bug 864 - if countexp was a long, we temporarily modified it
         to pretend that it was an INT. This reclaims the type tree. 28/1/91.
      */
      SetNType(origexp, origtype);
      /*}}}*/
      setlab(skiplab);
    }
    /*}}}*/
  else
    /*{{{  input an element*/
    {
      treenode *type = gettype(inputitem);
      const INT32 s = bytesin(type);
    #if 0 /*def DEBUG*/
      DEBUG_MSG(("tinputitem: bytesin is %ld, type is", s));
      printtree(0, type);
      DEBUG_MSG(("\n"));
    #endif
      if (s == (-1))
        /*{{{  length is an expression*/
        {
          /*{{{  check hidden dimensions*/
          if ((TagOf(prottype) != S_ANY) && RANGECHECKING)
            tcheckdimensions(type, prottype);
          /*}}}*/
          /*{{{  load operands for an 'in' instruction*/
          {
            BOOL preeval_e2, preeval_e3;
            int loadseq;
            treenode *lengthexp;
            int lengthmode = P_EXP;

          #if 0
            /* bug TS/1626 24/04/92 */
            int old = switch_to_temp_workspace();
            lengthexp = scaletreeof((TagOf(prottype) != S_ANY) ? prottype : type,
                                    1, be_lexlevel);
            switch_to_prev_workspace(old);
          #else
            lengthexp = ThisItem(NextItem(RHSOf(comm)));
            lengthmode = simplify(lengthmode, lengthexp);
          #endif

            /* We only do this call so that we have a loadseq parameter for
               tload3regs, ideally we should make the binder save loadseq
               in the tree. */
            loadseq = giveorder(inputitemmode, inputitem, channelmode, channel,
                                lengthmode, lengthexp, &preeval_e2, &preeval_e3);
            tlengthcheck(lengthexp, prottype);
            tload3regs(inputitemmode, inputitem, channelmode, channel,
                       lengthmode, lengthexp, loadseq, TRUE);
          }
          /*}}}*/
        }
        /*}}}*/
      else
        {
          tload2regs(inputitemmode, inputitem, channelmode, channel, FALSE, TRUE);
          genprimary(I_LDC, s);
        }
      checkerror(); /* bug 1202 11/3/91 */
      tioop(I_IN, usecall);
    }
    /*}}}*/
}
/*}}}*/
/*{{{  PUBLIC void toutput(tptr)*/
/*****************************************************************************
 *
 *  toutput generates code for an OUTPUT node
 *
 *****************************************************************************/
PUBLIC void toutput ( treenode *tptr )
{
  treenode *channel = LHSOf(tptr);
  treenode *outputlist = RHSOf(tptr);
  treenode *channeltype = gettype(channel);
  DEBUG_MSG(("toutput\n"));
  if (TagOf(channeltype) != S_CHAN)
    /*{{{  handle PORT output*/
    {
      treenode **oldtypeaddr; /* INSdi03067 */
      treenode *porttype;
      tpreexp(channel);
      tpreexp(ThisItem(outputlist));
      porttype = fixporttype(channel, &oldtypeaddr);
      tsimpleassign(ntypeof(channel), P_EXP, channel,
                              P_EXP, ThisItem(outputlist), MANY_REGS);
      restoreporttype(channel, porttype, oldtypeaddr);
    }
    /*}}}*/
  else if (ActionTypeOf(tptr) != NULL) /* bug TS/1626 23/04/92 */
    /*{{{  all the work was done in trans*/
    {
      int channelmode = chanaspointer ? P_EXP : P_PTR;
      tpreexp(channel);
      channelmode = simplify(channelmode, channel);
      tpreexp(ThisItem(outputlist));
      toutputitem(channelmode, channel, P_EXP, ThisItem(outputlist),
                  ActionTypeOf(tptr), tptr);
    }
    /*}}}*/
  else
    /*{{{  handle CHAN output*/
    {
      int channelmode = chanaspointer ? P_EXP : P_PTR;
      treenode *protocol = ProtocolOf(channeltype);
      BOOL outputtag = FALSE; /* True when we are generating code to output a tag */

      /*{{{  find the protocol type*/
      if (TagOf(protocol) == N_SPROTDEF)
        protocol = NTypeOf(protocol);
      else if (TagOf(protocol) == N_TPROTDEF)
        /* Pick out the type tree of the tag node */
        {
          protocol = NTypeOf(CExpOf(ThisItem(outputlist)));
          outputtag = TRUE;
        }
      /*}}}*/

      tpreexp(channel);
      channelmode = simplify(channelmode, channel);
      while (!EndOfList(outputlist))
        /*{{{  output item of list, move to next item*/
        {
          treenode *prottype = ((protocol != NULL) && (TagOf(protocol) == S_LIST))
                                 ? ThisItem(protocol) : protocol;
          tpreexp(ThisItem(outputlist));
          toutputitem(channelmode, channel, P_EXP, ThisItem(outputlist), prottype, tptr);
          if ((protocol != NULL) && (TagOf(protocol) == S_LIST) && !outputtag)
            protocol = NextItem(protocol);
          outputtag = FALSE;
          outputlist = NextItem(outputlist);
        }
        /*}}}*/
    }
    /*}}}*/
}
/*}}}*/
/*{{{  PUBLIC void tinput(tptr)*/
/*****************************************************************************
 *
 *  tinput generates code for an INPUT node
 *
 *****************************************************************************/
PUBLIC void tinput ( treenode *tptr )
{
  treenode *channel = LHSOf(tptr),
           *inputlist = RHSOf(tptr);

  DEBUG_MSG(("tinput\n"));
  switch (inputtypeof(tptr))
    {
      /*{{{  channel input*/
      case INP_INPUT:
        if (ActionTypeOf(tptr) != NULL) /* bug TS/1626 23/04/92 */
          {
            int channelmode = chanaspointer ? P_EXP : P_PTR;
            tpreexp(channel);
            channelmode = simplify(channelmode, channel);
            tpreexp(ThisItem(inputlist));
            tinputitem(channelmode, channel, P_PTR, ThisItem(inputlist),
                       ActionTypeOf(tptr), tptr);
          }
        else
        {
          int channelmode = chanaspointer ? P_EXP : P_PTR;
          treenode *protocol = ProtocolOf(gettype(channel));
          if (TagOf(protocol) == N_SPROTDEF)
            protocol = NTypeOf(protocol);
          tpreexp(channel);
          channelmode = simplify(channelmode, channel);
          while (!EndOfList(inputlist))
            /*{{{  input item of list, move to next item*/
            {
              treenode *prottype = (TagOf(protocol) == S_LIST) ? ThisItem(protocol)
                                                               : protocol;
              tpreexp(ThisItem(inputlist));
              tinputitem(channelmode, channel, P_PTR, ThisItem(inputlist), prottype, tptr);
              if (TagOf(protocol) == S_LIST)
                protocol = NextItem(protocol);
              inputlist = NextItem(inputlist);
            }
            /*}}}*/
        }
        break;
      /*}}}*/
      /*{{{  timer input*/
      case INP_TIMER_INPUT:
        {
          treenode *dest = ThisItem(inputlist);
          tpreexp(dest);
          tpreexp(channel);         /* Bug 288 22/5/90 */
          /*texp(channel, MANY_REGS);*/ /* Bug 288 22/5/90 */
          if (regsfor(dest) < MAXREGS)
            /*{{{  ldtimer; st dest*/
            {
              gensecondary(I_LDTIMER);
              storeinelement(dest, 0, MAXREGS - 1);
            }
            /*}}}*/
          else
            /*{{{  make pointer to dest, generate source, store through pointer*/
            {
              BIT32 offset = loadelementpointeroffset(dest, MANY_REGS);
              gensecondary(I_LDTIMER);
              gensecondary(I_REV);
              genprimary(I_STNL, offset);
            }
            /*}}}*/
        }
        break;
      /*}}}*/
      /*{{{  delayed input*/
      case INP_DELAYED_INPUT:
        tpreexp(inputlist);
        tpreexp(channel);         /* Bug 288 22/5/90 */
        /*texp(channel, MANY_REGS);*/ /* Bug 288 22/5/90 */
        /*{{{  T9000_alpha_badrunp*/
        if (T9000_alpha_badrunp(&tx_global))
          {
            int mode = P_EXP;
            mode = simplify(mode, inputlist);
            gensecondary(I_TALT);
            gencomment0("Can't use tin because of Scheduler bug");
            texp(inputlist, MANY_REGS);
            genprimary(I_LDC, 1);
            /*{{{  T9000_gamma_carryin*/
            if (T9000_gamma_carryin(&tx_global))
              {
                genprimary(I_LDC, 0);
                gensecondary(I_SUB);
                gencomment0("Set carry flag for ALU Adder CarryIn bug");
              }
            /*}}}*/
            gensecondary(I_ENBT);
            /*{{{  T9000_gamma_carryin*/
            if (T9000_gamma_carryin(&tx_global))
              {
                genprimary(I_LDC, 0);
                gensecondary(I_SUB);
                gencomment0("Set carry flag for ALU Adder CarryIn bug");
              }
            /*}}}*/
            gensecondary(I_TALTWT);
          }
        /*}}}*/
        else
          {
            texp(inputlist, MANY_REGS);
            gensecondary(I_TIN);
          }
        break;
      /*}}}*/
      /*{{{  port input*/
      case INP_PORT_INPUT:
        tpreexp(channel);
        {
          treenode **oldtypeaddr; /* INSdi03067 */
          treenode *porttype = fixporttype(channel, &oldtypeaddr);
          tpreexp(ThisItem(inputlist));
          tsimpleassign(ntypeof(channel), P_EXP, ThisItem(inputlist),
                                  P_EXP, channel, MANY_REGS);
          restoreporttype(channel, porttype, oldtypeaddr);
        }
        break;
      /*}}}*/
      /*{{{  case input*/
      #if 0 /* This is now transformed into something simpler */
      case INP_CASE_INPUT:
        tcaseinput(tptr);
        break;
      #endif
      /*}}}*/
      /*{{{  tagged input*/
      #if 0 /* This is now transformed into something simpler */
      case INP_TAGGED_INPUT:
        {
          int channelmode = chanaspointer ? P_EXP : P_PTR;
          treenode *tagtemp = inputlist;
          treenode *protocol;
          inputlist = NDeclOf(tagtemp);

          /* Pick out the type tree of the tag node */
          /* Tag node is the first item of the input list, but the mapper will
             have put a temporary on top of the input list */
          protocol = NTypeOf(CExpOf(ThisItem(inputlist)));

          tpreexp(channel);
          /*{{{  preevaluate pointer to channel, if neccessary*/
          channelmode = simplify(channelmode, channel);
          /*}}}*/
          /*{{{  input the tag*/
          zero_local_var(tagtemp); /* will always be a BYTE */
          tload2regs(P_TEMPPTR, tagtemp, channelmode, channel, FALSE, TRUE);
                                                          /*           ldlp  temp      */
                                                          /*           ldptr channel   */
          loadconstant(1);                            /*           ldc   1         */
          tioop(I_IN, iobycall && maybevirtualchan(channel)); /*       in              */
          /*SetTag(tagtemp, T_PREEVALTEMP);*/
          temp_mark_as_evaluated(tagtemp); /* bug TS/1900 22/10/92 */

          /*}}}*/
          /*{{{  compare inputted tag with required tag*/
          {
            treenode *snode;
            int continuelab = newlab();
            int old = switch_to_temp_workspace();
            snode = newdopnode(S_EQ, NOPOSN, tagtemp, ThisItem(inputlist), S_BYTE);
            switch_to_prev_workspace(old);
            tguard(snode, FALSE, continuelab);
            /*{{{  source output*/
            /* removed source_output 14/3/91 */
            /*if (source_output)
              so_stop();*/
            /*}}}*/
            tstop(tptr);
            setlab(continuelab);
          }
          /*}}}*/
          /*{{{  input the rest of the items on the list*/
          inputlist = NextItem(inputlist);
          while (!EndOfList(inputlist))
            {
              tpreexp(ThisItem(inputlist));
              tinputitem(channelmode, channel, P_PTR, ThisItem(inputlist),
                         ThisItem(protocol), tptr);
              protocol  = NextItem(protocol);
              inputlist = NextItem(inputlist);
            }
          /*}}}*/
        }
        break;
      #endif
      /*}}}*/
      default:
        badtag(LocnOf(tptr), TagOf(tptr), "tinput");
    }
}
/*}}}*/
/*{{{  PUBLIC void talt(tptr)*/
/*{{{  comment*/
/*****************************************************************************
 *
 *  talt generates code for a complete ALT construct
 *
 *****************************************************************************/
/*}}}*/
PUBLIC void talt ( treenode *tptr )
{
  const int joinlab   = newlab();
  const int altendlab = newlab();
  const int timedalt = timerguardinalt(tptr);
  const INT32 savedreplaltvaluesbase   = replaltvaluesbase;
  const int   savedreplaltstacklen     = replaltstacklen;
  treenode **const savedreplaltstack   = replaltstack;
  treenode *const  savedrepls = (nodetypeoftag(TagOf(tptr)) == CNODE) ?
                                CTempOf(tptr) : ReplCTempOf(tptr);
  /* This will be the number of nested levels of replicators: */
  const int repllevels = (int)allocsize(savedrepls); /* bug 1254 8/5/91 */

  DEBUG_MSG(("talt: mallocing size %d replaltstack\n", repllevels));
  /* add 1 to malloc-ed size to prevent problems with zero byte malloc */
  replaltstack = (treenode **)memalloc(sizeof(treenode *) * repllevels + 1);
  replaltstacklen = 0;
  /*replaltvaluesbase = NVOffsetOf(savedrepls);*/
  replaltvaluesbase = NVOffsetOf(savedrepls) + nameoffsetof(NLexLevelOf(savedrepls));

  nalternatives = numberofalternatives(tptr);
  if (nalternatives != 0)
    /*{{{  generate the alternatives*/
    {
      /* the following are only used while enabling, so can't be nested */
      alternativecount = 0;
      inreplalt        = FALSE;
      altcheckneeded   = (NEED_ERRORS
                       && ((errormode & ERRORMODE_STOP) == 0) /* INSdi02265 */
                       && !hastruealtguard(tptr));
      altcheckdone     = FALSE;

      gensecondary(timedalt ? I_TALT : I_ALT);
      insidealtguard = TRUE;
      alt_ds = timedalt ? DS_WAIT : DS_IO;
      /*{{{  check if we need to initialise degenerate alt trap*/
      if (altcheckneeded)
        {
          BOOL inittemp = TRUE;
          treenode *alt = tptr;
          while (alt != NULL && TagOf(alt) != S_ALTERNATIVE)
            /*{{{  walk down to the first alternative*/
            {
              alt = skipspecifications(alt);
              switch(TagOf(alt))
                {
                  case S_ALT: case S_PRIALT:
                    alt = CBodyOf(alt);
                    break;
                  case S_LIST:
                    alt = ThisItem(alt);
                    break;
                  case S_REPLALT: case S_PRIREPLALT:
                    alt = NULL;
                    break;
                  case S_ALTERNATIVE:
                    break;
                }
            }
            /*}}}*/

          if (alt != NULL && TagOf(alt) == S_ALTERNATIVE)
            /*{{{  if (don't need to init. temp) inittemp = FALSE*/
            {
              treenode *const guard = AltGuardOf(alt);
              if (AltChanExpOf(alt) == NULL)
                inittemp = FALSE;
              else if (AltTimeExpOf(alt) == NULL)
                {
                  if (istrueguard(guard) || !cancauseerror(AltChanExpOf(alt)))
                    inittemp = FALSE;
                }
              else
                {
                  if (istrueguard(guard) || !cancauseerror(AltTimeExpOf(alt)))
                    inittemp = FALSE;
                }
            }
            /*}}}*/

          if (inittemp)
            {
            /*genprimary(I_LDC, 1);*/
              genprimary(I_LDC, 0); /* INSdi02266 22/06/93 */
              genprimary(I_STL, 0);
              gencomment0("INT");
            }
        }
      /*}}}*/
      taltenables(tptr, NULL, 0);
      /*{{{  check for degenerate ALT (if not done already)*/
      if (altcheckneeded && !altcheckdone)
        {
          genprimary(I_LDL, 0);
          genprimary(I_LDC, 1);
          gensecondary(I_CCNT1);
        }
      /*}}}*/
      if (alternativecount == 0) /* all replicators must have had zero counts */
        /*{{{  stop*/
        {
          /* removed source_output 14/3/91 */
          /*if (source_output)
            so_stop();*/
          tstop (tptr);
        }
        /*}}}*/
      else
        {
          /*{{{  T9000_gamma_carryin*/
          if (timedalt && T9000_gamma_carryin(&tx_global))
            {
              genprimary(I_LDC, 0);
              gensecondary(I_SUB);
              gencomment0("Set carry flag for ALU Adder CarryIn bug");
            }
          /*}}}*/
          gensecondary(timedalt ? I_TALTWT : I_ALTWT);
        }
      disable_csub0 = TRUE;
      taltdisables(tptr, NULL, altendlab);
      disable_csub0 = FALSE;
      insidealtguard = FALSE;
      /* if (alternativecount != 0)
        gensecondary(I_ALTEND);
      setlab(altendlab);
      */
      genaltend (alternativecount, altendlab);
      /*{{{  generate all the bodies*/
      {
        treenode **altbodystack;
        int nestedalts = countnestedalts(tptr, NESTED_ALTS);
        DEBUG_MSG(("talt: mallocing size %d altbodystack\n", nestedalts));
        altbodystack = (treenode **)memalloc(sizeof(treenode *) * nestedalts);
        taltbodies(tptr, joinlab, altbodystack, 0); /* Generate all the bodies */
        DEBUG_MSG(("talt: freeing size %d altbodystack\n", nestedalts));
        memfree(altbodystack);
      }
      /*}}}*/
      setlab(joinlab);
      genstartblock();
    }
    /*}}}*/
  else
    {
      tstop (tptr);
    }
  DEBUG_MSG(("talt: freeing size %d replaltstack\n", repllevels));
  memfree(replaltstack);

  replaltvaluesbase = savedreplaltvaluesbase;
  replaltstacklen   = savedreplaltstacklen;
  replaltstack      = savedreplaltstack;
}
/*}}}*/
