/* $Id: gen11.c,v 1.8 1997/10/02 14:14:28 mdp2 Exp $ */

/*
 *	code generator - operand loading and storing
 *	Copyright (C) 1987 Inmos Limited
 *
 *	This program is free software; you can redistribute it and/or modify
 *	it under the terms of the GNU General Public License as published by
 *	the Free Software Foundation; either version 2 of the License, or
 *	(at your option) any later version.
 *
 *	This program is distributed in the hope that it will be useful,
 *	but WITHOUT ANY WARRANTY; without even the implied warranty of
 *	MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *	GNU General Public License for more details.
 *
 *	You should have received a copy of the GNU General Public License
 *	along with this program; if not, write to the Free Software
 *	Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 */

/*#define DEBUG*/

/*{{{  include files*/
#ifdef HAVE_CONFIG_H
#  include <config.h>
#endif
#include <stdlib.h>
#include <stdio.h>
#include <string.h>
#include "includes.h"
#include "generror.h"
#include "instruct.h"
#include "genhdr.h"
#include "trandef.h"
#include "bind1def.h"
#include "bind2def.h"
#include "bind3def.h"
#include "gen1def.h"
#include "gen2def.h"
#include "gen4def.h"
#include "gen5def.h"
#include "gen8def.h"
#include "gen11def.h"
#include "gen12def.h"
#include "code1def.h"
/*}}}*/

/*{{{  constants*/
/* These are done as bit patterns so that multiple cases can be tested
   efficiently - CON 24/7/91
*/
#define MOVEDIRN_LOAD    1
#define MOVEDIRN_STORE   2
#define MOVEDIRN_LOADPTR 4
#define MOVEDIRN_LOADEXT 8 /* load and sign extend */
/*}}}*/

/*{{{  local variables*/
PRIVATE INT32 wspadjusts[MAX_LEX_LEVELS];
PRIVATE BIT32 staticlinkoffsets[MAX_LEX_LEVELS];
/*}}}*/

/*{{{  forward declarations*/
#if 0 /* no constructors are passed to the backend any more */
PRIVATE void tconstructorassign (treenode *dest, treenode *source);
#endif
/*}}}*/

/*{{{  scope adjustment handling*/
/*{{{  PUBLIC void adjustworkspace (adjustment)*/
/*****************************************************************************
 *
 *  adjustworkspace updates the code generator's view of where the workspace
 *                  pointer is.  All workspace accesses will be 'adjustment'
 *                  slots higher than they were.
 *
 *****************************************************************************/
PUBLIC void adjustworkspace ( INT32 adjustment )
{
  wspadjusts[be_lexlevel] += adjustment;
#ifdef DEBUG
  if (adjustment != 0)
    DEBUG_MSG(("adjustworkspace: adjusting be_lexlevel %d by %ld to %ld\n",
               be_lexlevel, adjustment, wspadjusts[be_lexlevel]));
#endif
  asmvalues[ASMNAME_WSSIZE] += adjustment;
  asmvalues[ASMNAME_STATIC] += adjustment;
  asmvalues[ASMNAME_VSPTR ] += adjustment;
}
/*}}}*/
/*{{{  PUBLIC void setadjust (lexlevel, v)*/
/*****************************************************************************
 *
 *  setadjust  initialises the code generator's view of where the workspace
 *             pointer is for lexical level 'lexlevel'.
 *
 *****************************************************************************/
PUBLIC void setadjust ( int lexlevel , INT32 v )
{
  DEBUG_MSG(("setadjust: setting adjust for lexlevel %d to %ld\n", lexlevel, v));
  wspadjusts[lexlevel] = v;
}
/*}}}*/
/*}}}*/
/*{{{  static link and vector space pointer routines*/
/*{{{  PUBLIC void setsloffset(lexlevel, v)*/
/*****************************************************************************
 *
 *  setsloffset sets the static link offset for 'lexlevel' to v
 *
 *****************************************************************************/
PUBLIC void setsloffset ( int lexlevel , INT32 v )
{
  DEBUG_MSG(("setsloffset: setting sloffset for lexlevel %d to %ld\n", lexlevel, v));
  staticlinkoffsets[lexlevel] = v;
}
/*}}}*/
/*{{{  PRIVATE INT32 sloffsetof(level)*/
/*****************************************************************************
 *
 *  sloffsetof returns the workspace offset of the static link at lex level
 *             'level' relative to the static link pointer of lex level
 *             'level - 1' (or if level is the current lexical level, relative
 *             to the current value of wptr).
 *
 *****************************************************************************/
PRIVATE INT32 sloffsetof ( int level )
{
  if (level == be_lexlevel)
    return((staticlinkoffsets[level] & OFFSET_BITS) + wspadjusts[level]);
  else
    {
      INT32 sloffset = staticlinkoffsets[level] & OFFSET_BITS;
      if (staticlinkoffsets[level + 1] & REPL_FLAG)
        sloffset += wspadjusts[level];
      return(sloffset);
    }
}
/*}}}*/
/*{{{  PUBLIC INT32 nameoffsetof(level)*/
/*{{{  comment*/
/*****************************************************************************
 *
 *  nameoffsetof returns the adjustment to the workspace offset of a
 *               variable at lexical level 'level'.  If 'level' is the current
 *               lexical level, this is relative to the current wptr,
 *               otherwise it is relative to the static link at offset
 *               'level - 1'.
 *
 *****************************************************************************/
/*}}}*/
PUBLIC INT32 nameoffsetof ( int level )
{
  if (level == be_lexlevel)
    return(wspadjusts[level]);
  else if (staticlinkoffsets[level + 1] & REPL_FLAG)
    return(wspadjusts[level]);
  else
    return(0);
}
/*}}}*/
/*{{{  PRIVATE BOOL loadlex (l)*/
/*{{{  comment*/
/*****************************************************************************
 *
 *  loadlex takes a lexical level, l, and returns FALSE if it is the current
 *          level, otherwise it loads a pointer to the workspace at level l
 *          and returns TRUE.
 *
 *****************************************************************************/
/*}}}*/
PRIVATE BOOL loadlex ( int l )
{
  /* Generate static link code to fetch workspace at level l */
  if (l != be_lexlevel)
    {
      int level = be_lexlevel - 1;
      INT32 sloffset = sloffsetof(be_lexlevel);
      genprimary(I_LDL, sloffset);
      gencomment0("PTR staticlink");
      for (; level > l; level--)
        /*{{{  load static link for level*/
        {
          /* If it is a repl PAR static link the static link contains the
             workspace pointer of the routine setting up the repl
             AT THE TIME OF SETTING UP THE REPL */
          genprimary (I_LDNL, sloffsetof(level));
          gencomment1("PTR staticlink %d", level - l);
        }
        /*}}}*/
      return(TRUE);
    }
  else
    return(FALSE);
}
/*}}}*/
/*{{{  PUBLIC void loadstaticlink (newlevel)*/
PUBLIC void loadstaticlink ( int newlevel )
{
  if (newlevel == (be_lexlevel + 1))
    genprimary(I_LDLP, wspadjusts[be_lexlevel]);
    /* This should be the current workspace offset */
  else
    {
      int nonlocal = loadlex(newlevel);
      genprimary(nonlocal ? I_LDNL : I_LDL, sloffsetof(newlevel));
      /* if required static link isn't normalised, normalise it */
      if (staticlinkoffsets[newlevel] & REPL_FLAG)
        genprimary(I_LDNLP, wspadjusts[newlevel - 1]);
    }
  gencomment0("PTR staticlink");
}
/*}}}*/
/*{{{  PUBLIC void loadreplreturnlink ()*/
/*****************************************************************************
 *
 *  loadreplreturnlink loads the static link required when ending a replicated
 *                     parallel process
 *
 *****************************************************************************/
PUBLIC void loadreplreturnlink ( void )
{
  genprimary(I_LDL, sloffsetof(be_lexlevel));
  gencomment0("PTR repljoin");
}
/*}}}*/
/*{{{  PUBLIC void loadnewvsp(offset)*/
PUBLIC void loadnewvsp (const BIT32 offset)
{
  genprimary(I_LDL, vspoffset + wspadjusts[be_lexlevel]);
  gencomment0("PTR vsp");
  genprimary(I_LDNLP, offset);
}
/*}}}*/
/*}}}*/
/*{{{  temp restoring*/
/* bug TS/1900 22/10/92 */
PRIVATE treenode *evaluated_temp_list;
/*{{{  PUBLIC void temp_clear_all_evaluated*/
PUBLIC void temp_clear_all_evaluated(void)
{
  evaluated_temp_list = NULL;
}
/*}}}*/
/*{{{  PUBLIC void temp_mark_as_evaluated*/
PUBLIC void temp_mark_as_evaluated(treenode *const nptr)
{
  SetTag(nptr, T_PREEVALTEMP);
  SetNVNextTemp(nptr, evaluated_temp_list);
  evaluated_temp_list = nptr;
}
/*}}}*/
/*{{{  PUBLIC void temp_unevaluate_all*/
PUBLIC void temp_unevaluate_all(void)
{
  treenode *tptr;
  for (tptr = evaluated_temp_list; tptr != NULL; tptr = NVNextTempOf(tptr))
    SetTag(tptr, T_TEMP);
  evaluated_temp_list = NULL;
}
/*}}}*/

/*}}}*/
/*{{{  PUBLIC int simplify (opdmode, opd)*/
/*****************************************************************************
 *
 *  simplify evaluates an expression into a temporary, if required
 *
 *****************************************************************************/
PUBLIC int simplify ( int opdmode , treenode *opd )
{
  if (preeval(opdmode, opd))
    {
      const int newmode = tempmodeof(opdmode);
      /* Mark opd as initialised so that in the course of initialising it
         we don't start to recursively initialise it. */
      /*SetTag(opd, T_PREEVALTEMP);*/
      temp_mark_as_evaluated(opd); /* bug TS/1900 22/10/92 */

      if (NModeOf(opd) == NM_POINTER)
        /*{{{  treat temporary as an abbreviation*/
        {
          if (opdmode != P_PTR)
            geninternal_is(GEN_BAD_OPD, opdmode, "simplify");
          loadopd(opdmode, NDeclOf(opd), 0);
          storeinname(opd, 0);
        }
        /*}}}*/
      else
        /*{{{  treat temporary as a declaration*/
        tsimpleassign(ntypeof(opd), newmode, opd,
                            opdmode, NDeclOf(opd), MANY_REGS);
        /*}}}*/
      opdmode = newmode;
    }
  return opdmode;
}
/*}}}*/
/*{{{  PUBLIC void loadconstant*/
/*****************************************************************************
 *
 *  loadconstant loads the value of the constant c onto the expression stack
 *
 *****************************************************************************/
PUBLIC void loadconstant (const INT32 c)
{
#if 0
/* moved to code1 so that target-dependent optimisations can be done away from here */
  switch(constformat_of_const(targetintsize, c, 0))
    {
      case constformat_mint32_ldnlp:
        gensecondary(I_MINT);
        genprimary(I_LDNLP, (INT32)((BIT32)c - (BIT32)MOSTNEG_INT32) / 4 /*bytesperword*/);
        break;
      case constformat_mint32_adc:
        gensecondary(I_MINT);
        genprimary(I_ADC,   (INT32)((BIT32)c - (BIT32)MOSTNEG_INT32));
        break;
      case constformat_mint32_not:
        gensecondary(I_MINT);
        gensecondary(I_NOT);
        break;
      case constformat_ldinf_ldnlp:
        gensecondary(I_LDINF);
        genprimary(I_LDNLP, (INT32)((BIT32)c - (BIT32)INFINITY) / 4 /*bytesperword*/);
        break;
      case constformat_ldinf_adc:
        gensecondary(I_LDINF);
        genprimary(I_ADC,   (INT32)((BIT32)c - (BIT32)INFINITY));
        break;
      case constformat_mint16_ldnlp:
        gensecondary(I_MINT);
        genprimary(I_LDNLP, (INT32)((BIT32)c - (BIT32)MOSTNEG_INT16) / 2 /*bytesperword*/);
        break;
      case constformat_mint16_adc:
        gensecondary(I_MINT);
        genprimary(I_ADC,   (INT32)((BIT32)c - (BIT32)MOSTNEG_INT16));
        break;
      case constformat_ldc:
      case constformat_table:
        genprimary(I_LDC, c);
        break;
    }
#else
  genloadconstant (c);
#endif
}
/*}}}*/
/*{{{  PRIVATE void tbyteoffset*/
PRIVATE void tbyteoffset(const INT32 offset)
{
  if (offset == 0)
    /* skip */ ;
  else if ((bytesperword == 2) && ((offset & 0x1) == 0))
    genprimary(I_LDNLP, offset / 2 /* bytesperword */);
  else if ((bytesperword == 4) && ((offset & 0x3) == 0))
    genprimary(I_LDNLP, offset / 4 /* bytesperword */);
  else if (use_bsub_not_adc) /* T9000 optimisation 18/7/91 */
    {
      genprimary(I_LDC, offset);
      gensecondary(I_BSUB);
    }
  else
    genprimary(I_ADC, offset);
}
/*}}}*/
/*{{{  load/store/loadptr       names*/
/*{{{  PRIVATE void movepointer*/
PRIVATE void movepointer(const int dirn, const int inonlocal, const INT32 word,
                         const int type, BOOL *const need_signextend,
                         const BOOL devaccess)
/*
  This assumes a pointer has already been loaded, and does a load/store off it
  It requires 'word' to be a word offset, if we're using word sized data.
*/
{
  if (dirn == MOVEDIRN_STORE) checkerror ();
  if (istargetbytesize(type))
    {
      tbyteoffset(word); /* In practice this is never non-zero */
      if (dirn != MOVEDIRN_LOADPTR) /* bug 738 5/11/90 */
        /*{{{  load/store a byte*/
        {
          if (devaccess)
            /*{{{  device instructions*/
            {
              /*{{{  T9000_gamma_baddevaccess*/
              if (T9000_gamma_baddevaccess(&tx_global))
                {
                  gensecondary(I_INTDIS);
                  gencomment0("bug fix for device access");
                }
              /*}}}*/
              gensecondary((dirn == MOVEDIRN_STORE) ? I_DEVSB : I_DEVLB);
              /*{{{  T9000_gamma_baddevaccess*/
              if (T9000_gamma_baddevaccess(&tx_global))
                {
                  gensecondary(I_INTENB);
                  gencomment0("bug fix for device access");
                }
              /*}}}*/
            }
            /*}}}*/
          else
            gensecondary((dirn == MOVEDIRN_STORE) ? I_SB : I_LB);
        }
        /*}}}*/
    }
  else if (use_shortintops && isshortint(type)) /* T9000 shorts 17/7/91 */
    {
      tbyteoffset(word * 2); /* In practice this is never non-zero */
      if (dirn != MOVEDIRN_LOADPTR)
        /*{{{  load/store a sixteen*/
        {
          if (devaccess)
            /*{{{  device access*/
            {
              /*{{{  T9000_gamma_baddevaccess*/
              if (T9000_gamma_baddevaccess(&tx_global))
                {
                  gensecondary(I_INTDIS);
                  gencomment0("bug fix for device access");
                }
              /*}}}*/
              gensecondary((dirn == MOVEDIRN_STORE) ? I_DEVSS : I_DEVLS);
              /*{{{  T9000_gamma_baddevaccess*/
              if (T9000_gamma_baddevaccess(&tx_global))
                {
                  gensecondary(I_INTENB);
                  gencomment0("bug fix for device access");
                }
              /*}}}*/
              if (dirn == MOVEDIRN_LOADEXT) gensecondary(I_XSWORD);
            }
            /*}}}*/
          else
            gensecondary((dirn == MOVEDIRN_STORE)   ? I_SS :
                         (dirn == MOVEDIRN_LOADEXT) ? I_LSX :
                                                      I_LS);
        }
        /*}}}*/
      if (need_signextend != NULL)
        *need_signextend = FALSE;
    }
  else
    /*{{{  load/store a word*/
    {
      if (devaccess)
        {
          if (dirn == MOVEDIRN_LOADPTR)
            genprimary(inonlocal, word);
          else
            {
              genprimary(I_LDNLP, word);
              /*{{{  T9000_gamma_baddevaccess*/
              if (T9000_gamma_baddevaccess(&tx_global))
                {
                  gensecondary(I_INTDIS);
                  gencomment0("bug fix for device access");
                }
              /*}}}*/
              gensecondary((dirn == MOVEDIRN_STORE) ? I_DEVSW : I_DEVLW);
              /*{{{  T9000_gamma_baddevaccess*/
              if (T9000_gamma_baddevaccess(&tx_global))
                {
                  gensecondary(I_INTENB);
                  gencomment0("bug fix for device access");
                }
              /*}}}*/
            }
        }
      else
        genprimary(inonlocal, word);
    }
    /*}}}*/
}
/*}}}*/
/*{{{  PRIVATE void movename (nptr, w, ilocal, inonlocal)*/
/*{{{  comment*/
/*****************************************************************************
 *
 *  movename loads or stores a name or a pointer to a name.
 *           nptr represents the name.
 *           For a long type, w is the word which we wish to load/store.
 *           ilocal is the load/store instruction used for a local object,
 *           inonlocal is the load/store instruction used for a non-local
 *           object.
 *
 *****************************************************************************/
/*}}}*/
PRIVATE void movename (treenode *const nptr, const INT32 w,
                       const int ilocal, const int inonlocal)
{
#if 0 /* now S_FNFORMALRESULT is a namenode */
  if (TagOf(nptr) == S_FNFORMALRESULT)
    /*{{{  move formal result (ie same as a pointer)*/
    {
      const int level     = HLexLevelOf(nptr);
      const BOOL nonlocal = loadlex(level);
      const INT32 wsposn  = HOffsetOf(nptr) + w + nameoffsetof(level);

      genprimary((nonlocal ? inonlocal : ilocal), wsposn);
    }
    /*}}}*/
  else
#endif
    {
      if (NVUseCountOf(nptr) == 0) /* This should never happen */
        geninternal_s(GEN_BAD_USECOUNT, WNameOf(NNameOf(nptr)));

      if (isplaced(nptr))
        /*{{{  move to/from a placed variable*/
        {
          loadconstant(NVOffsetOf(nptr));
          /*genprimary(inonlocal, w);*/ /* removed 8/11/91 for bug TS/1467 */
        }
        /*}}}*/
      else
        /*{{{  move to/from a normal variable*/
        {
          const int level     = NLexLevelOf(nptr);
          const BOOL nonlocal = loadlex(level);
          const INT32 wsposn = NVOffsetOf(nptr) + w + nameoffsetof(level);
          genprimary((nonlocal ? inonlocal : ilocal), wsposn);
        }
        /*}}}*/
    }
  /*{{{  debugging*/
  if (assembly_output && ((alloc_strategy & ALLOC_NOCOMMENTS) == 0))
  {
    /*{{{  comment on the type MDP*/
    {
      treenode *type = follow_user_type(NTypeOf(nptr));
      if (TagOf(type) == S_ARRAY)
        type = follow_user_type(ARTypeOf(type));
      if (ispointer(nptr))
        gencomment0("PTR");
      gencomment0(tagstring(TagOf(type))); /* MDP */
    }
    /*}}}*/
    switch(TagOf(nptr))
      {
        case T_TEMP: case T_PREEVALTEMP:
          gencomment1("$temp%d", NVVarNumOf(nptr));
          break;
      #if 0 /* now S_FNFORMALRESULT is a namenode */
        case S_FNFORMALRESULT:
          gencomment0("formalresult");
          break;
      #endif
        default:
          gencomment0(WNameOf(NNameOf(nptr)));
          break;
      }
    if (w != 0)
      fprintf(outfile, " + %d", w);
  }
  /*}}}*/
}
/*}}}*/
/*{{{  PUBLIC void loadname (nptr, w)*/
PUBLIC void loadname(treenode *const nptr, const INT32 w)
{
  if (chanaspointer && issimplechan(nptr))
    /* Very special case */
    movename(nptr, w, I_LDLP, I_LDNLP);
  else
    {
      if (ispointer(nptr))
        {
          movename(nptr, 0, I_LDL, I_LDNL);
          genprimary(I_LDNL, w); /* should only be called on word-sized objects */
        }
      else
        movename(nptr, w, I_LDL, I_LDNL);
    }
}
/*}}}*/
/*{{{  PUBLIC void storeinname (nptr, w)*/
PUBLIC void storeinname(treenode *const nptr, const INT32 w)
{
  checkerror();
  movename(nptr, w, I_STL, I_STNL);
}
/*}}}*/
/*{{{  PUBLIC void zero_local_var*/
PUBLIC void zero_local_var(treenode *const nptr)
{
  loadconstant(0);
  storeinname(nptr, 0);
  gencomment0("INIT");
}
/*}}}*/
/*{{{  PRIVATE int return_subscript*/
PRIVATE int return_subscript(treenode *const tptr)
{
  treenode *const base_type = basetype_tree(gettype(tptr));
#ifdef OCCAM2_5
  const int b = (TagOf(base_type) == S_RECORD) ? bytesperword : (int)bytesin(base_type);
#else
  const int b = (int)bytesin(base_type);
#endif
  if      (can_use_wsubdb(b, bytesperword, tptr)) return I_WSUBDB;
  else if (b >= bytesperword)                     return I_WSUB;
  else if (use_shortintops && b == 2)             return I_SSUB; /* T9000 shorts 17/7/91 */
  else                                            return I_BSUB;
}
/*}}}*/
/*{{{  PRIVATE void add_scaled_offset*/
PRIVATE void add_scaled_offset(treenode *const tptr, const INT32 offset)
{
  const int isubscript = return_subscript(tptr);
  if (isubscript == I_BSUB)
    tbyteoffset(offset);
  else if (isubscript == I_SSUB) /* T9000 shorts 17/7/91 */
    tbyteoffset((BIT32)offset * 2);
  else
    genprimary (I_LDNLP, offset);
}
/*}}}*/
/*{{{  PUBLIC void loadnamepointer*/
PUBLIC void loadnamepointer(treenode *nptr, const INT32 w)
{
  treenode *name = NULL;
  if (TagOf(nptr) == N_VALABBR || TagOf(nptr) == N_VALRETYPE)
    {
      treenode *cp = DValOf(NDeclOf(nptr));
      name = nptr;
      if ((TagOf(cp) == S_STRING) ||
          (TagOf(cp) == S_CONSTCONSTRUCTOR))
        nptr = cp;
    }

  if ((TagOf(nptr) == S_STRING) || (TagOf(nptr) == S_CONSTCONSTRUCTOR))
    /*{{{  load a pointer to a constant table*/
    {
      int label = NO_LABEL;
      if (WLengthOf(CTValOf(nptr)) == 0)
        {
          /* optimisation for null strings */
          genprimary(I_LDLP, 0);
        }
      else
        {
          const int constlab = CTLabelOf(nptr);
          label = newlab();
          genloadlabptr(constlab, label,
           (name == NULL) ? NULL : WNameOf(NNameOf(name)));
          label = NO_LABEL;  /* suppress setlab below */
        }
      if (label != NO_LABEL)
        setlab(label);
      add_scaled_offset(nptr, w); /* Added for bug 1071 13/12/90 */
    }
    /*}}}*/
  else if (ispointer(nptr))
    {
    #if 0 /* INSdi02168 has demonstrated that this fails if w != 0 */
      movename(nptr, w, I_LDL, I_LDNL);
    #else
      movename(nptr, 0, I_LDL, I_LDNL);
      genprimary(I_LDNLP, w);
    #endif
    /*if (w != 0) printf("loadnamepointer: pointer; genlocn: %8X, w=%ld\n", genlocn, w);*/
    }
  else
    movename(nptr, w, I_LDLP, I_LDNLP);
}
/*}}}*/
/*{{{  load/store/loadptr       arrays*/
/*{{{  PRIVATE INT32 loadarraypointer*/
/*{{{  comment*/
/*****************************************************************************
 *
 *  loadarraypointer loads a pointer to the array element 'tptr' into Areg,
 *                   using at most regs registers.
 *                   A (word) offset from the loaded pointer to the element
 *                   is returned.
 *
 *****************************************************************************/
/*}}}*/
/* A transformed subscript tree */
PRIVATE INT32 loadarraypointer(treenode *const tptr, const INT32 word, const int regs)
{
  const BOOL segmentitem = (TagOf(tptr) == S_SEGMENTITEM);
  const int isubscript = return_subscript(tptr);
  treenode *nptr = nameof(tptr);
  treenode *subscriptexp = segmentitem ? SSubscriptExpOf(tptr) : ASExpOf(tptr);
  INT32 offset = (segmentitem ? SOffsetOf(tptr) : ASOffsetOf(tptr)) + word;

  if (directload(P_EXP, tptr, be_lexlevel))
    /*{{{  load local pointer*/
    {
      loadnamepointer (nptr, offset);
      offset = 0;
    }
    /*}}}*/
  else
    /*{{{  load a pointer*/
    {
      if (subscriptexp != NULL)
        /*{{{  perform subscript calculation*/
        {
          texp(subscriptexp, regs);
          loadnamepointer (nptr, 0);
          gensecondary (isubscript);
        }
        /*}}}*/
      else
        loadnamepointer (nptr, 0);
    }
    /*}}}*/
  if (isubscript == I_BSUB)
    /*{{{  add in the offset now*/
    {
      tbyteoffset(offset);
      offset = 0;
    }
    /*}}}*/
  return (offset);
}
/*}}}*/
/*{{{  PRIVATE void movearrayitem*/
PRIVATE void movearrayitem(treenode *const tptr, const INT32 word, const int regs,
                           const int dirn, const int inonlocal)
{
  treenode *nptr = nameof(tptr);
  int type = basetype(gettype(nptr));
#ifdef OCCAM2_5
  if (type == S_RECORD) type = basetype(gettype(tptr));
#endif

/*printf("movearrayitem: type is %s\n", itagstring(type));*/

  if ((dirn != MOVEDIRN_LOADPTR) &&
      !use_shortintops && /* T9000 shorts 17/7/91 */
      (type == S_INT16) && (targetintsize != S_INT16))
    badtag(genlocn, type, "movearrayitem");
  else
    {
      if (((dirn & (MOVEDIRN_LOADEXT | MOVEDIRN_LOAD)) != 0) &&
          directload(P_EXP, tptr, be_lexlevel))
        loadname (nptr, ASOffsetOf(tptr) + word);

      else if ((dirn == MOVEDIRN_STORE) &&
          directstore(P_EXP, tptr, be_lexlevel))
        storeinname(nptr, ASOffsetOf(tptr) + word);

      else
        /*{{{  ldptr item; load/store*/
        {
          INT32 offset = loadarraypointer(tptr, word, regs);
          if ((dirn == MOVEDIRN_LOADPTR) ||
              (bytesinscalar(type) < bytesperword))
            {
              add_scaled_offset(tptr, offset);
              offset = 0;
            }

          if (dirn != MOVEDIRN_LOADPTR)
            movepointer(dirn, inonlocal, offset, type, NULL, is_devaccess(nptr));
        }
        /*}}}*/
    }
}
/*}}}*/
/*}}}*/
/*{{{  load/store/loadptr       elements*/
/*{{{  PRIVATE void moveelement (tptr, word, dirn, regs)*/
/* Load or store the element represented by tptr.
If dirn is MOVEDIRN_LOAD load the element into Areg.
If dirn is MOVEDIRN_LOADEXT load the element into Areg and sign-extend.
if dirn is MOVEDIRN_STORE store Areg into the element.
If dirn is MOVEDIRN_LOADPTR load a pointer to the element into Areg.
If the element is long, load/store the word'th word of it. */
  /* modified 5/11/90 by CO'N for bug 738, to work with word != 0 when
     dirn == MOVEDIRN_LOADPTR. (IE until now, it has always assumed that the offset
     is zero when loading a pointer.
     The idea is that it simply loads a pointer to the same data which it
     would have loaded if dirn was MOVEDIRN_LOAD
  */
PRIVATE void moveelement(treenode *const tptr, const INT32 word, const int dirn, const int regs)
{
  int ilocal, inonlocal;
  const int tag = TagOf(tptr);
  BOOL need_signextend = (dirn == MOVEDIRN_LOADEXT) &&
       use_shortintops && isshortint(ntypeof(tptr)); /* bug 1345 23/7/91 */

  /*{{{  set up ilocal and inonlocal*/
  switch(dirn)
    {
      case MOVEDIRN_LOAD: case MOVEDIRN_LOADEXT:
        if (chanaspointer && issimplechan(tptr))
          /* Very special case */
          {
            ilocal = I_LDLP;
            inonlocal = I_LDNLP;
          }
        else
          {
            ilocal = I_LDL;
            inonlocal = I_LDNL;
          }
        break;
      case MOVEDIRN_STORE:
        ilocal = I_STL;
        inonlocal = I_STNL;
        break;
      case MOVEDIRN_LOADPTR:
        ilocal = I_LDLP;
        inonlocal = I_LDNLP;
        break;
      default:
        ilocal = I_LDL; inonlocal = I_LDNL; /* to shut up warning messages */
        geninternal_i(GEN_BAD_MOVEELEMENT_DIRN, dirn);
    }
  /*}}}*/

  switch(tag)
    /*{{{  cases*/
    {
      /*{{{  decl repl param var.abbrev retype preevaltemp*/
      case N_DECL: case N_REPL: case N_PARAM: case N_VALPARAM:
      case N_ABBR: case N_RETYPE: case T_PREEVALTEMP:
      case T_TEMP: /* code commoned-ed up 5/11/90 */
      case S_FNFORMALRESULT: /* code commoned-up by CON - 15/11/93 */
        {
          const int type = ntypeof(tptr);
          /* T_TEMP code commoned-ed up 5/11/90 */
          if ((tag == T_TEMP) &&
              ((dirn & (MOVEDIRN_LOAD | MOVEDIRN_LOADEXT)) != 0) &&
               /*!use_shortintops &&*/ /* T9000 shorts 17/7/91 */
               isshortint(type))
            simplify(P_EXP, tptr); /* Move the value to a temporary */
          if (ispointer(tptr))
            {
              movename(tptr, 0, I_LDL, I_LDNL);
              movepointer(dirn, inonlocal, word, type,
                          &need_signextend, is_devaccess(tptr));
            }
          else
            {
              if (dirn == MOVEDIRN_STORE) checkerror ();
              movename(tptr, word, ilocal, inonlocal);
            }
        }
        break;
      /*}}}*/
      /*{{{  N_VALABBR N_VALRETYPE*/
      case N_VALABBR:
      case N_VALRETYPE:
        /* Could be a constant table. If so, we can only load a pointer */
        {
          treenode *cp = DValOf(NDeclOf(tptr));
          const int t = TagOf(cp);
          if (t == S_STRING || t == S_CONSTEXP  || t == S_CONSTCONSTRUCTOR)
            /*{{{  its a constant table, load a pointer to it*/
            {
              moveelement(cp, word, dirn, regs);
              need_signextend = FALSE;
            }
            /*}}}*/
          else if (ispointer(tptr)) /* if VAL abbr, must be greater than word sized */
            {
              movename(tptr, 0, I_LDL, I_LDNL);
              /* if (dirn != MOVEDIRN_LOADPTR) */ /* test removed for bug 738 5/11/90 */
                genprimary(inonlocal, word);
            }
          else
            movename(tptr, word, ilocal, inonlocal);
          break;
        }
      /*}}}*/
      /*{{{  S_ARRAYITEM S_SEGMENTITEM*/
      case S_ARRAYITEM:
      case S_SEGMENTITEM:
      case S_RECORDSUB: case S_RECORDITEM:
        /* tnestedsegments(tptr, regs); */
        /*{{{  see if we have to preevaluate a constructor*/
        #if 0 /* no constructors are passed to the backend any more */
        {
        #if 0
          treenode *const nptr = nameof((tag == S_ARRAYITEM || tag == S_RECORDITEM) ? ASBaseOf(tptr) : SNameOf(tptr));
        #else
          treenode *const nptr = nameof(tptr);
        #endif
          if (preeval(P_EXP, nptr))
            {
              /*SetTag(nptr, T_PREEVALTEMP);*/
              temp_mark_as_evaluated(nptr); /* bug TS/1900 22/10/92 */
              tconstructorassign(nptr, NDeclOf(nptr));
            }
        }
        #endif
        /*}}}*/

        movearrayitem(tptr, word, regs, dirn, inonlocal);
        need_signextend = FALSE;

        break;
      /*}}}*/
      /*{{{  S_STRING S_CONSTCONSTRUCTOR N_LABELDEF, and PROCs*/
      case S_STRING:
      case S_CONSTCONSTRUCTOR:
      case N_LABELDEF:
      case N_PROCDEF:                   case N_LIBPROCDEF: /* bug TS/1240 14/04/92 */
      case N_LFUNCDEF: case N_SFUNCDEF: case N_LIBFUNCDEF:
        {
          const BOOL consttable = (tag == S_STRING) || (tag == S_CONSTCONSTRUCTOR);
          if (consttable && (WLengthOf(CTValOf(tptr)) == 0))
            {
              /* optimisation for null strings */
              genprimary(I_LDLP, 0);
            }
          else
            {
              const int constlab = (tag == N_LABELDEF)  ? (int)NVOffsetOf(tptr)
                                   : consttable         ? CTLabelOf(tptr)
                                                        : NPLabelOf(tptr);
              const int l = newlab();
              genloadlabptr(constlab, l, NULL);
            }
          /* I suppose that here we should add an offset if 'word' is
             non-zero - bug 738 5/11/90.  */
          /* type(S_STRING) and type(S_CONSTCONSTRUCTOR) etc are S_ARRAY,
             so we always load a word offset:
             (I would guess that this never happens, since it seems wrong
              for S_STRING etc, but at least it is consistent with
              what is done for names, etc. CO'N 5/11/90) */
          if ((dirn & (MOVEDIRN_LOAD | MOVEDIRN_LOADPTR | MOVEDIRN_LOADEXT)) != 0)
            genprimary(inonlocal, word);
          else
            geninternal_i(GEN_BAD_CONSTANT_LOAD, dirn);
        }
        break;
      /*}}}*/
      /*{{{  S_FNFORMALRESULT*/
      #if 0
      case S_FNFORMALRESULT:
        /* A function result pointer is effectively an abbreviation to the
           result destination */
        /*genprimary(I_LDL, HOffsetOf(tptr));*/
        movename(tptr, 0, I_LDL, I_LDNL);
        /*gencomment0("formalresult");*/
        movepointer(dirn, inonlocal, word, TagOf(HExpOf(tptr)),
                    &need_signextend, FALSE);
        break;
      #endif
      /*}}}*/
      /*{{{  S_CONSTEXP*/
      case S_CONSTEXP:
        if (isinconstanttable(tptr))
          texp_constanttable(tptr, inonlocal, (int)word);
        else if ((dirn & (MOVEDIRN_LOAD | MOVEDIRN_LOADEXT)) != 0)
          loadconstant(wordof(tptr, (int)word));
        else
          geninternal_i(GEN_BAD_MOVEELEMENT_DIRN, dirn);
        need_signextend = FALSE;
        break;
      /*}}}*/
      default:
        badtag(genlocn, tag, "moveelement");
    }
    /*}}}*/

  if (need_signextend)
    gensecondary(I_XSWORD); /* bug 1345 23/7/91 */
}
/*}}}*/
/*{{{  PUBLIC void loadelement(tptr, word, regs)*/
PUBLIC void loadelement (treenode *const tptr, const INT32 word,
                         const int regs, const BOOL signextend_result)
{
  moveelement(tptr, word,
              signextend_result ? MOVEDIRN_LOADEXT : MOVEDIRN_LOAD,
              regs);
}
/*}}}*/
/*{{{  PUBLIC void storeinelement(tptr, word, regs)*/
PUBLIC void storeinelement ( treenode *tptr , INT32 word , int regs )
{
  moveelement(tptr, word, MOVEDIRN_STORE, regs);
}
/*}}}*/
/*{{{  PUBLIC void loadelementpointer(tptr, regs)*/
PUBLIC void loadelementpointer(treenode *const tptr, const INT32 word, const int regs )
{
  /* modified 5/11/90 to take a 'word' parameter too */
  moveelement(tptr, word, MOVEDIRN_LOADPTR, regs);
}
/*}}}*/
/*{{{  PUBLIC BIT32 loadelementpointeroffset(tptr, regs)*/
PUBLIC INT32 loadelementpointeroffset(treenode *const tptr, const int regs)
{
  INT32 offset = 0;
  if (TagOf(tptr) == S_ARRAYITEM || TagOf(tptr) == S_RECORDITEM)
    offset = loadarraypointer(tptr, 0, regs); /* better code */
  else
    moveelement(tptr, 0, MOVEDIRN_LOADPTR, regs);
  return(offset);
}
/*}}}*/
/*}}}*/
/*{{{  load/store/loadptr       opds*/
/*{{{  PUBLIC void loadopd(opdmode, opd, word)*/
PUBLIC void loadopd ( int opdmode , treenode *opd , INT32 word )
{
  /* This is normally used for double length etc, so we can ignore
     the sign-extension parameter to loadelement by setting it TRUE
  */
  switch (opdmode)
    {
      case P_EXP:     loadelement(opd, word, MANY_REGS, TRUE);  break;
      case P_TEMP:    loadname(opd, word);                      break;
      case P_PTR:     loadelementpointer(opd, word, MANY_REGS); break;
      case P_TEMPPTR: loadnamepointer(opd, word);               break;
    }
}
/*}}}*/
/*{{{  PUBLIC void storeinopd (opdmode, opd, word, regs)*/
/*****************************************************************************
 *
 *  storeinopd pulls the top of stack and places it in (opdmode, opd)
 *
 *****************************************************************************/
PUBLIC void storeinopd ( int opdmode , treenode *opd , INT32 word , int regs )
{
  switch (opdmode)
    {
      case P_TEMPPTR:
      case P_TEMP:
      case P_EXP:      storeinelement(opd, word, regs);
                       break;
      default:
        geninternal_is(GEN_BAD_OPD, opdmode, "storeinopd");
    }
}
/*}}}*/
/*{{{  PUBLIC void loadopdpointer (opdmode, opd)          NEVER USED*/
#if 0 /* loadopdpointer is never used */
PUBLIC void loadopdpointer ( int opdmode , treenode *opd )
{
  switch(opdmode)
    {
      case P_TEMPPTR:
      case P_EXP:
      case P_TEMP:    loadelementpointer(opd, 0, MANY_REGS); break;
      default:
        geninternal_is(GEN_BAD_OPD, opdmode, "loadopdpointer");
    }
}
#endif /* loadopdpointer is never used */
/*}}}*/
/*{{{  BIT32 loadopdpointeroffset (opdmode, opd)*/
PRIVATE INT32 loadopdpointeroffset ( int opdmode , treenode *opd )
{
  INT32 offset = 0;
  switch(opdmode)
    {
      case P_TEMPPTR:
      case P_EXP:
      case P_TEMP:    offset = loadelementpointeroffset (opd, MANY_REGS);
                      break;
      default:
        geninternal_is(GEN_BAD_OPD, opdmode, "loadopdpointeroffset");
    }
  return(offset);
}
/*}}}*/
/*}}}*/
/*}}}*/

/*{{{  assigninfo_t*/
typedef struct
  {
    treenode *adest;
    treenode *asource;
    int atype;
    int aregs;
    int astoreregs;
    BOOL aevaluated;
  } assigninfo_t;
/*}}}*/

/*{{{  PRIVATE void setregdests (nregresults, regdests, destlist, instancedfn)*/
/*{{{  comment*/
/*****************************************************************************
 *
 *  setregdests takes the left- and right- hand sides of a call to a
 *              multivalued function and builds up a table, 'regdests' of
 *              the destination elements of the register results of the
 *              function. On exit, '*nregresults' contains the number of
 *              register results returned by the function.
 *
 *****************************************************************************/
/*}}}*/
PRIVATE void setregdests ( int *nregresults , treenode **regdests [MAXREGS ], treenode *destlist , treenode *instancedfn )
{
  int i;
  int regsused = 0;
  treenode *ftypelist = FnTypeListOf(NTypeOf(instancedfn));
  while (!EndOfList(ftypelist) && (regsused < MAXREGS))
    {
      if ( result_return_style(ThisItem(ftypelist)) == return_style_alu)
        regdests[regsused++] = ThisItemAddr(destlist);
      destlist  = NextItem(destlist);
      ftypelist = NextItem(ftypelist);
    }

  /* The list appears in the wrong order, so reverse it */
  for (i = 0; i < regsused / 2; i++)
    {
      /* Swap elements i and j (where j is the i'th from top) */
      const int j = regsused - 1 - i;
      treenode **temp = regdests[i];
      regdests[i] = regdests[j];
      regdests[j] = temp;
    }

  *nregresults = regsused;
}
/*}}}*/
/*{{{  PRIVATE int regsforassignments (assignment, nassignments)*/
/*****************************************************************************
 *
 *  regsforassignments returns the number of registers required to perform
 *                     all unevaluated assignments in the table 'assignment'.
 *
 *****************************************************************************/
PRIVATE int regsforassignments ( assigninfo_t assignment [], const int nassignments )
{
  int i;
  int maxr = -1;
  for (i = 0; i < nassignments; i++)
    if (!assignment[i].aevaluated)
      maxr = (int)max_INT32(maxr, assignment[i].aregs);
  return(maxr);
}
/*}}}*/
/*{{{  PRIVATE int ndependson (i, assignment, dependson, nassignments)*/
/*****************************************************************************
 *
 *  ndependson returns the number of assignments which assignment i
 *             depends on.
 *
 *****************************************************************************/
PRIVATE int ndependson ( int i , assigninfo_t assignment [], char dependson [], int nassignments )
{
  int d = 0;
  int j;
  for (j = 0; j < nassignments; j++)
    if (!assignment[j].aevaluated && dependson[(i * nassignments) + j])
      d = d + 1;
  return(d);
}
/*}}}*/
/*{{{  PRIVATE int ndependants (i, assignment, dependson, nassignments)*/
/*****************************************************************************
 *
 *  ndependants returns the number of assignments which depend upon
 *              assignment i.
 *
 *****************************************************************************/
PRIVATE int ndependants ( int i , assigninfo_t assignment [], char dependson [], int nassignments )
{
  int dependants = 0;
  int j;
  for (j = 0; j < nassignments; j++)
    if (!assignment[j].aevaluated && dependson[(j * nassignments) + i])
      dependants++;
  return(dependants);
}
/*}}}*/
/*{{{  PRIVATE int forced_assignment (assignment, dependson, nassignments)*/
/*{{{  comment*/
/*****************************************************************************
 *
 *  forced_assignment finds the best assignment to perform from the
 *                    assignment table 'assignment'.
 *                    The best assignment is the one which is depended upon
 *                    by the most other assignments (this breaks the greatest
 *                    number of cyclic dependencies), or in the case of a
 *                    draw, the assignment which uses most registers.
 *
 *****************************************************************************/
/*}}}*/
PRIVATE int forced_assignment ( assigninfo_t assignment [], char dependson [], int nassignments )
{
  int best_assignment = -1, maxdependants = -1, maxregisters = -1, i;

  for (i = 0; i < nassignments; i++)
    if (!assignment[i].aevaluated)
      {
        const int d = ndependants(i, assignment, dependson, nassignments);
        if
          /*{{{  best assignment so far*/
          ((d > maxdependants) ||
            ((d == maxdependants) && (assignment[i].aregs > maxregisters)))
          /*}}}*/
            /*{{{  choose this assignment*/
            {
              best_assignment = i;
              maxdependants = d;
              maxregisters = assignment[i].aregs;
            }
            /*}}}*/
      }

  return(best_assignment);
}
/*}}}*/
/*{{{  PUBLIC void mapmoveopd (destmode, destptr, sourcemode, sourceptr)*/
/*****************************************************************************
 *
 *  mapmoveopd maps out the moving of (sourcemode, sourceptr) to
 *             (destmode, destptr)  : both operands must be addressable
 *             N.B. At the moment, this routine assumes that source must
 *             be moved to dest using a move instruction, as other cases
 *             are pulled out by mapsimpleassign.
 *
 *****************************************************************************/
PUBLIC void mapmoveopd ( int destmode , treenode **destptr , int sourcemode , treenode **sourceptr )
{
  /*{{{  default to a move instruction (with some last minute optimisations)*/
  {
    treenode *desttype = gettype_main(*destptr);  /* we now want to keep this type tree */
    INT32 b;

    /*{{{  turn expressions into pointers*/
    sourcemode = ptrmodeof(sourcemode);
    destmode   = ptrmodeof(destmode);
    /*}}}*/

    b = bytesin(desttype);
    if (b == -1) b = bytesin(gettype(*sourceptr));
    if ((b == 1) || (b == bytesperword &&
                     check_aligned(*sourceptr, (int)b) &&
                     check_aligned(*destptr,   (int)b)))
      mapload2regs(destmode, destptr, sourcemode, sourceptr);
    else if (b == (-1))
      /*{{{  length is an expression*/
      {
        treenode *lengthexp;
        treenode **lengthexpp = &lengthexp;  /* just in case */

        /*switch_to_temp_workspace();*/
        lengthexp = scaletreeof(desttype, 1, be_lexlevel);
        /*switch_to_real_workspace();*/

        /* Now we save the length of the block move - CO'N 18/5/90 */
        if (TagOf(*destptr) == S_ARRAYITEM)
          {
            DEBUG_MSG(("mapmoveopd: S_ARRAYITEM\n"));
            SetASLength(*destptr, lengthexp);
            lengthexpp = ASLengthAddr(*destptr);
          }
        else if (TagOf(*destptr) == S_SEGMENTITEM)
          {
            DEBUG_MSG(("mapmoveopd: S_SEGMENTITEM\n"));
            SetSLength(*destptr, lengthexp);
            lengthexpp = SLengthAddr(*destptr);
          }
      #ifdef DEBUG
        else
          DEBUG_MSG(("mapmoveopd: neither S_ARRAYITEM nor S_SEGMENTITEM\n"));
      #endif

        mapload3regs(sourcemode, sourceptr, destmode, destptr,
                   P_EXP, lengthexpp);
        /* And, as usual, throw away the load sequence */
      }
      /*}}}*/
    else if (b == 0)
      return; /* bug TS/1782 26/08/92 */
    else
      /*{{{  length is a constant*/
      mapload2regs(sourcemode, sourceptr, destmode, destptr);
      /*}}}*/
  }
  /*}}}*/
}
/*}}}*/
/*{{{  PRIVATE treenode *makeconstructordesttree (dest)*/
#if 0 /* no constructors are passed to the backend any more */
/*****************************************************************************
 *
 *  makeconstructordesttree builds an ARRAYITEM tree for the first element
 *                          of the constructor destination tree, 'dest'.
 *
 *****************************************************************************/
PRIVATE treenode *makeconstructordesttree ( treenode *dest )
{
  treenode *dexp = dummyexp_p; /* added for bug 679 31/8/90 */
  BIT32 doffset  = 0;
  treenode *desttree;

#if 0
  if (TagOf(dest) == T_TEMP)  /* temporary bodge for bug TS/1263 11/12/91 */
    dest = NDeclOf(dest);
#endif

  if (TagOf(dest) == S_ARRAYITEM)
    {
      dexp = ASExpOf(dest);
      doffset = ASOffsetOf(dest);
      dest = newarraysubnode(S_ARRAYSUB, NOPOSN, ASBaseOf(dest), dummyexp_p);
    }
  desttree = newarraysubnode(S_ARRAYITEM, NOPOSN, dest, dexp);
  SetASOffset(desttree, doffset);

#ifdef DEBUG
  fputs("*desttree from makeconstructordesttree is :", outfile);
  printtree(outfile, 0, desttree);
  fputc('\n', outfile);
#endif

  return(desttree);
}
#endif
/*}}}*/
/*{{{  PUBLIC void mapconstructorassign(destmode, destptr, sourcemode, sourceptr)*/
#if 0 /* no constructors are passed to the backend any more */
/*{{{  comment*/
/*****************************************************************************
 *
 *  mapconstructorassign maps the assignment of the constructor, source, to
 *                       dest
 *
 *****************************************************************************/
/*}}}*/
PUBLIC void mapconstructorassign ( int destmode , treenode **destptr , int sourcemode , treenode **sourceptr )
{
  if (sourcemode != P_EXP)
    geninternal_is(GEN_BAD_OPD, sourcemode, "mapconstructorassign");

#if 0 /* No VALOFs can return arrays - bug TS/1442 4/11/91 */
  if (isspecification(*sourceptr) || TagOf(*sourceptr) == S_VALOF) /* bug TS/1442 28/10/91 */
    {
      mapvalof(*sourceptr, destptr);
      return;
    }
#endif

  if (TagOf(*sourceptr) == S_CONSTRUCTOR)
    sourceptr = LitExpAddr(*sourceptr);
  else
    badtag(genlocn, TagOf(*sourceptr), "mapconstructorassign");

#ifdef DEBUG
  fputs("*destptr before mapping is :", outfile);
  printtree(outfile, 0, *destptr);
  fputc('\n', outfile);
#endif

  {
    int old = switch_to_temp_workspace();
    treenode *desttree = makeconstructordesttree(*destptr); /* bug 337 */
    switch_to_prev_workspace(old);

    for(; !EndOfList(*sourceptr); sourceptr = NextItemAddr(*sourceptr))
      {
        mapsimpleassign(ntypeof(ThisItem(*sourceptr)), destmode, &desttree,
                        P_EXP, ThisItemAddr(*sourceptr));
      }

#ifdef DEBUG
    fputs("desttree after mapping is :", outfile);
    printtree(outfile, 0, desttree);
    fputc('\n', outfile);
#endif
  }
#if 0
  {
    for (; !EndOfList(*sourceptr); sourceptr = NextItemAddr(*sourceptr))
        mapsimpleassign(ntypeof(ThisItem(*sourceptr)), destmode, destptr,
                        P_EXP, ThisItemAddr(*sourceptr));
  }
#endif
#ifdef DEBUG
  fputs("*destptr after mapping is :", outfile);
  printtree(outfile, 0, *destptr);
  fputc('\n', outfile);
#endif
}
#endif
/*}}}*/
/*{{{  PUBLIC void mapsimpleassign(type, destmode, dest, sourcemode, source)  ***/
/*{{{  comment*/
/*****************************************************************************
 *
 *  mapsimpleassign maps out the assignment of source to dest, where
 *                  source and dest are single items of type 'type'.
 *
 *****************************************************************************/
/*}}}*/
PUBLIC void mapsimpleassign (const int type, int destmode,   treenode **dest,
                                             int sourcemode, treenode **source)
{
  DEBUG_MSG(("mapsimpleassign: type: %s, destmode: %s, sourcemode: %s\n",
             itagstring(type), opdmode_string(destmode), opdmode_string(sourcemode)));
  if (has_fpu_core && isreal(type)
      && !isaddressableopd(sourcemode, *source) /* bug 353 11/9/90 */
      && !isconstopd(sourcemode, *source)) /* bug 708 12/9/90 */
    mapfpassign(destmode, dest, sourcemode, source);
  else if (istargetintsize(type) || istargetbytesize(type)
           || (use_shortintops && isshortint(type))) /* T9000 shorts 17/7/91 */
    /*{{{  single-length assign*/
    {
      if (has_fpu_core && isconversion(*source) && isreal(ntypeof(OpOf(*source)))
          && !isshortint(type)) /* otherwise breaks on i16 := INT16 ROUND r */
        /*{{{  real to int conversion as source*/
        {
          BOOL desttempused = FALSE;

          if (T9000_alpha_nofpconv(&tx_global)) /* Sections 4.2.7, 4.2.8, 4.2.11, 4.2.12 of SW-0341-2 */
            {
              treenode *new = makeconversion(ntypeof(OpOf(*source)), type, OpOf(*source), TagOf(*source), LocnOf(*source));
              freenode(source);
              *source = new;
              mapsimpleassign(type, destmode, dest, sourcemode, source);
              return;
            }

          if (!issimpleopd(destmode, *dest)) /* added for bug 739 24/9/90 */
            {
              *dest = gettemp(*dest, NM_POINTER);
              destmode = tempmodeof(destmode);
              mapexp(NDeclAddr(*dest));
              upusecount(*dest, 1);
              desttempused = TRUE;
            }
          mapfpexp(OpAddr(*source));
          destmode = ptrmodeof(destmode);
          mapexpopd(destmode, dest);
          if (desttempused)
            freetemp(*dest);
        }
        /*}}}*/
      else if (directstore(destmode, *dest, be_lexlevel))
        /*{{{  load, then store*/
        {
          mapexpopd(sourcemode, source);
          mapstoreinopd(destmode, dest);
        }
        /*}}}*/
      else
        /*{{{  load pointer to dest and load source, store through pointer*/
        {
          destmode = ptrmodeof(destmode);
          mapload2regs(sourcemode, source, destmode, dest);
        }
        /*}}}*/
    }
    /*}}}*/
  else if (isdoublelength(type))
    mapmovelopd(destmode, dest, sourcemode, source);
  else
    /*{{{  special cases for assign*/
    switch (type)
      {
        /* other cases  still to come */
        /*{{{  INT64 REAL64 on 16-bit machine*/
        case S_INT64:
        case S_REAL64:
          mapmoveqopd(destmode, dest, sourcemode, source);
          break;
        /*}}}*/
        /*{{{  INT16 on 32-bit machine*/
        case S_INT16:
          if (isaddressableopd(sourcemode, *source) || isconstopd(sourcemode, *source))
            {
              const BOOL needtempdest   = needtemptoload(destmode, *dest);
              const BOOL notconst       = !isconstopd(sourcemode, *source);
              const BOOL needtempsource = needtemptoload(sourcemode, *source);
              DEBUG_MSG(("mapsimpleassign: INT16: needtempdest=%d, notconst=%d, needtempsource=%d\n",
                         needtempdest, notconst, needtempsource));
              if (needtempdest || (notconst && needtempsource))
                /*{{{  ldptr source; ldptr dest; ldc 2; move*/
                {
                  DEBUG_MSG(("mapsimpleassign: INT16: first case, mapping for move\n"));
                  destmode = ptrmodeof(destmode);
                  sourcemode = ptrmodeof(sourcemode);
                /*mapload2regs(destmode, dest, sourcemode, source);*/
                  mapload2regs(sourcemode, source, destmode, dest); /* bug 1375 22/8/91 */
                }
                /*}}}*/
              else
                /*{{{  ld source; st dest*/
                {
                  DEBUG_MSG(("mapsimpleassign: INT16: first case, turn into INT\n"));
                  mapexpopd(sourcemode, source);
                  mapstoreinopd(destmode, dest);
                }
                /*}}}*/
            }
          else
            {
              DEBUG_MSG(("mapsimpleassign: INT16: second case\n"));
              if (!issimpleopd(destmode, *dest) || ispointer(*dest))
                /*{{{  temp := source; dest := temp*/
                {
                  *source = gettemp(*source, NM_WORKSPACE);
                  mapsimpleassign(type, P_TEMP, source, sourcemode,NDeclAddr(*source));
                  sourcemode = P_TEMP;
                  mapsimpleassign(type, destmode, dest, sourcemode, source);
                  freetemp(*source);
                }
                /*}}}*/
              else
                /*{{{  ld source; st dest*/
                {
                  mapexpopd(sourcemode, source);
                  mapstoreinopd(destmode, dest);
                }
                /*}}}*/
            }
          break;
        /*}}}*/
        /*{{{  ARRAY*/
        case S_ARRAY: case S_RECORD:
        #ifdef OCCAM2_5
          if ((type == S_RECORD) && (bytesin(gettype(*dest)) <= bytesperword))
            /*{{{  assign directly*/
            mapsimpleassign(S_INT, destmode, dest, sourcemode, source);
            /*}}}*/
          else
        #endif
            /*{{{  do a fake 'quad length' move*/
            mapmoveqopd(destmode, dest, sourcemode, source);
            /*}}}*/
        #if 0 /* constructors don't get this far any more */
          if (isaddressableopd(sourcemode, *source))
            mapmoveopd(destmode, dest, sourcemode, source);
          else
            /*{{{  rhs must contain a constructor*/
            {
            #if 0
              fputs("mapsimpleassign: source is", outfile);
              printtree(0, *source);
              fputs("\ndest is", outfile);
              printtree(0, *dest);
              fputs("\n", outfile);
            #endif

              /* Special case channel constuctors **ACS*** */
              if (basetype(NTypeOf(nameof(*dest))) != S_CHAN &&
                  usedin(*dest, *source, be_lexlevel))
                /*{{{  preevaluate source to a temporary*/
                {
                  /*printf("mapsimpleassign: usedin is TRUE, so creating temp\n");*/

                  /* Make a temporary to evaluate the source into */
                  *source = gettemp(*source, NM_WORKSPACE);

                  /* Now we load each element of the constructor and store it in the temp */
                  mapconstructorassign(P_TEMP, source, P_EXP, NDeclAddr(*source));

                  /* Move the temporary to the destination */
                  mapmoveopd(destmode, dest, sourcemode, source);

                  freetemp(*source);
                }
                /*}}}*/
              else
                /*{{{  move each element of the source to the dest*/
                {
                  BOOL desttempused = FALSE;
                  /*{{{  preevaluate a pointer to the destination if neccessary*/
                  /* I've turned this off for the moment, till I have time to
                     fix it!  IT WAS TOTALLY WRONG AS IT STANDS   CO'N 20/2/90 */
                  /*if (!issimplelocalopd(destmode, *dest))*/ /* testing for bug 679 31/8/90 */
                  if (!issimpleopd(destmode, *dest)) /* testing for bug 679 31/8/90 */
                    {
                      DEBUG_MSG(("mapsimpleassign: turning constructor dest into a pointer\n"));
                      /**dest = gettemp(*dest, NM_WORKSPACE);*/
                      *dest = gettemp(*dest, NM_POINTER);
                      destmode = P_TEMP;
                      /**dest = gettemp(newmopnode(S_ADDRESSOF, LocnOf(*dest), *dest, 0), NM_WORKSPACE);
                      destmode = P_TEMPPTR;*/
                      mapexp(NDeclAddr(*dest));
                      upusecount(*dest, 1);
                      desttempused = TRUE;
                    }
                  /*}}}*/
                  /* Now we load each element of the constructor and store it in the dest */
                  mapconstructorassign(destmode, dest, sourcemode, source);
                  /*{{{  free up the destination pointer if neccessary*/
                  if (desttempused)
                    freetemp(*dest);
                  /*}}}*/
                }
                /*}}}*/
            }
            /*}}}*/
        #endif
          break;
        /*}}}*/
        default:
          /*geninternal_is(GEN_ERROR_IN_ROUTINE, 4, "mapsimpleassign");*/
          badtag(genlocn, type, "mapsimpleassign");
      }
    /*}}}*/
}
/*}}}*/
/*{{{  PRIVATE void mapmultiassign (mdestlist, msourcelist)  ***/
/* N.B. !!! Doesn't work for reals with inline fp */
/*{{{  comment*/
/*****************************************************************************
 *
 *  mapmultiassign takes a list of destinations, 'mdestlist' and a list of
 *                 sources, 'msourcelist', reorders them and
 *                 allocates and inserts temporaries required.  Note that
 *                 'special' temporary nodes are inserted for sources which
 *                 are held in registers before storing.
 *                 This routine only works for lists of expression sources,
 *                 multiple assignments which have multi-valued functions or
 *                 valofs on the right-hand side are handled separately.
 *                 Note in all the multiple assignment code that we can have
 *                 a multiple assignment nested inside another multiple
 *                 assignment (in a valof), so all the routines have to be
 *                 reentrant.
 *
 *                 Modified to use memalloc rather than fixing MAX_ASSIGNMENTS
 *
 *****************************************************************************/
/*}}}*/
PRIVATE void mapmultiassign ( treenode *mdestlist , treenode *msourcelist )
{
  treenode *sourcelist = msourcelist,
           *destlist   = mdestlist;
  int nassignments     = listitems(destlist);
  assigninfo_t *assignment = (assigninfo_t *)memalloc( (sizeof(assigninfo_t))
                                                      * nassignments);
  char *dependson = (char *)memalloc((sizeof(char)) * nassignments * nassignments);
  int  *evalorder = (int  *)memalloc((sizeof(int))  * nassignments);

  int i, j, ndone = 0;
  int regtemp[MAXREGS];
  int nregtemps = 0;
  DEBUG_MSG(("mapmultiassign:\n"));
  /*{{{  initialise 'assignment'*/
  for (i = 0; i < nassignments; i++)
    {
      assigninfo_t *ass = &(assignment[i]);
      treenode *dest   = ThisItem(destlist);
      treenode *source = ThisItem(sourcelist);
      int type = ntypeof(dest);
      ass->atype = type;
      ass->adest = dest;
      ass->asource = source;
      ass->aevaluated = FALSE;
      /*{{{  set ass->aregs and ass->astoreregs*/
      if (istargetbytesize(type) || istargetintsize(type) ||
          (use_shortintops && isshortint(type))) /* T9000 shorts 18/7/91 */
        {
          const int sr = regsfor(source);
          const int dr = regsforstore(dest);
          ass->astoreregs = dr;
          ass->aregs = ((sr == dr) ? sr + 1 : (int)max_INT32(sr, dr));
        }
      else
        {
          ass->astoreregs = MAXREGS;
          ass->aregs = MAXREGS;
        }
      DEBUG_MSG(("mapmultiassign: assignment %d, type %s, storeregs:%d, aregs:%d\n",
                  i, itagstring(ass->atype), ass->astoreregs, ass->aregs));
      /* There is a hole here : a scalar INT16 source (on T4) to scalar INT16
         dest only takes one register */
      /*}}}*/
      destlist = NextItem(destlist);
      sourcelist = NextItem(sourcelist);
    }
  /*}}}*/
  /*{{{  initialise 'dependson'*/
  for (i = 0; i < nassignments; i++)
    for (j = 0; j < nassignments; j++)
      {
        dependson[(nassignments * i) + j] = (i == j) ? FALSE :
                  usedin(assignment[i].adest, assignment[j].asource, be_lexlevel);
        DEBUG_MSG(("mapmultiassign: %d dependson %d ? %d\n", i, j, dependson[(nassignments * i) + j]));
      }
  /*}}}*/
  while (ndone < nassignments)
    /*{{{  move the sources to destinations, temporaries, or registers*/
    {
      /*{{{  do the assignments which aren't dependant upon any others*/
      BOOL done_an_assignment = TRUE;
      while (done_an_assignment)
        {
          done_an_assignment = FALSE;
          for (i = 0; i < nassignments; i++)
            if (!assignment[i].aevaluated &&
                (ndependson(i, assignment, dependson, nassignments) == 0))
              {
                assignment[i].aevaluated = TRUE;
                done_an_assignment = TRUE;
                evalorder[ndone++] = i;
                DEBUG_MSG(("mapmultiassign: doing asignment %d\n", i));
                mapsimpleassign(assignment[i].atype,
                                P_EXP, &(assignment[i].adest),
                                P_EXP, &(assignment[i].asource));
                /*{{{  resurrect newly formed temps*/
                /* INSdi03198 12/11/93 */
                /* If the 'mapsimpleassign' generated a temporary, then we must
                   resurrect it so that we don't clobber it.
                   This means that it is OK for 'tmultiassign' to delay the
                   re-assignment until the end of the multiassign.
                */
                if (TagOf(assignment[i].asource) == T_TEMP)
                  {
                    DEBUG_MSG(("mapmultiassign: resurrecting temp for assignment %d\n", i));
                    resurrect_var(assignment[i].asource);
                  }
                /*}}}*/
              }
        }
      /*}}}*/
      if (ndone < nassignments)
        /*{{{  break a cycle of dependency*/
        {
          int freeregs = MAXREGS - nregtemps;
          int type;
          i = forced_assignment(assignment, dependson, nassignments);
          assignment[i].aevaluated = TRUE;
          evalorder[ndone++] = i;
          DEBUG_MSG(("mapmultiassign: breaking cycle with assignment %d ", i));
          type = assignment[i].atype;
          if
            (fitsinregister(type) &&
             !(has_fpu_core && isreal(type)) && /* added for bug 1002 4/10/90 */
             (regsforassignments(assignment, nassignments) <= freeregs - 1) &&
             (assignment[i].astoreregs <= freeregs - 1))
              /*{{{  we can keep the source in a register*/
              {
                DEBUG_MSG(("in register\n"));
                mapexp(&(assignment[i].asource));
                assignment[i].asource = newtempnode(T_REGTEMP, assignment[i].asource,
                                                    NM_WORKSPACE, be_lexlevel);
                regtemp[nregtemps++] = i;
              }
              /*}}}*/
            else
              /*{{{  assign source i to a temporary*/
              {
                /* bug 1002 4/10/90 - cannot simply create scalars for all scalar
                   variables - we mustn't if it is a long, or if on FPU */
                if (fitsinregister(type) && !(has_fpu_core && isreal(type)))
                  /*{{{  create a scalar temporary and generate into that*/
                  {
                    DEBUG_MSG(("scalar temporary\n"));
                    mapexp(&(assignment[i].asource));
                    assignment[i].asource = gettemp(assignment[i].asource,
                                                    NM_WORKSPACE);
                    upusecount(assignment[i].asource,1);
                  }
                  /*}}}*/
                else
                  /*{{{  create a (vector/long/real) temporary and assign into that*/
                  {
                    DEBUG_MSG(("(vector/long/real) temporary\n"));
                    /* This branch creates a temporary _before_ evaluating the rhs, which
                       is okay for vectors, but suboptimal for scalar expressions */
                    assignment[i].asource = gettemp(assignment[i].asource, NM_WORKSPACE);
                    /* it would seem that we could just change NM_WORKSPACE to NM_DEFAULT,
                    but if we do, the temp is never initialised to point at vectorspace,
                    and the vectorspace is never allocated */
                    /*assignment[i].asource = gettemp(assignment[i].asource, NM_DEFAULT);*/ /* bug 235 17/9/90 */
                    mapsimpleassign(type, P_TEMP, &(assignment[i].asource),
                                          P_EXP, NDeclAddr(assignment[i].asource));
                  }
                  /*}}}*/
              }
              /*}}}*/
        }
        /*}}}*/
    }
    /*}}}*/
  /*{{{  store any values held in registers*/
  for (i = nregtemps - 1; i >= 0; i--)
    {
      DEBUG_MSG(("mapmultiassign: saving register temp %d\n", regtemp[i]));
      mapaddr(&(assignment[regtemp[i]].adest));
    }
  /*}}}*/
  /*{{{  move temporaries to destinations*/
  for (i = 0; i < nassignments; i++)
    if (TagOf(assignment[i].asource) == T_TEMP)
      /*{{{  move temporary i to destination*/
      {
        DEBUG_MSG(("mapmultiassign: moving temp for multiassignment %d back\n", i));
        mapsimpleassign(assignment[i].atype, P_EXP, &(assignment[i].adest),
                              P_TEMP, &(assignment[i].asource));
        freetemp(assignment[i].asource);
      }
      /*}}}*/
  /*}}}*/
  /*{{{  remake the multi-assign node with the new order*/
  destlist = mdestlist;
  sourcelist = msourcelist;
  for (i = 0; i < nassignments; i++)
    {
      assigninfo_t *ass = &(assignment[evalorder[i]]);
      NewItem(ass->adest, destlist);
      NewItem(ass->asource, sourcelist);
      destlist   = NextItem(destlist);
      sourcelist = NextItem(sourcelist);
    }
  /*}}}*/
  memfree(assignment);
  memfree(dependson);
  memfree(evalorder);
}
/*}}}*/
/*{{{  PUBLIC void mapassign (destptr, sourceptr)*/
/*****************************************************************************
 *
 *  mapassign maps the assignment of 'sourceptr' to 'destptr'.
 *            The source and destinations may be single items or lists.
 *
 *****************************************************************************/
PUBLIC void mapassign ( treenode **destptr , treenode **sourceptr )
{
  treenode *dest = *destptr,
           *source = *sourceptr;
  if (TagOf(dest) == S_LIST)
    /*{{{  deal with multiple assignments*/
    {
      if (listitems(source) > 1)
        mapmultiassign(dest, source);
      else
        /*{{{  multi-valued function or valof*/
        {
          treenode *sourceitem = ThisItem(source);
          const int tag = TagOf(sourceitem);
          if (isspecification(sourceitem) || (tag == S_VALOF))
            mapvalof(sourceitem, destptr);
          else if (tag == S_FINSTANCE)
            {
              if ((TagOf(INameOf(sourceitem)) != N_PREDEFFUNCTION) ||
                  !mappredef(sourceitem, dest))
                {  
                  treenode *instancedfn = INameOf(sourceitem);
                  /*{{{  map multi-valued function instance*/
                  {
                    treenode *paramlist = IParamListOf(sourceitem);
                    paramlist = augmentparams(paramlist,
                                              FnParamsOf(NTypeOf(instancedfn)), dest);
                    SetIParamList(sourceitem, paramlist);
                    /* Insert temporaries in front of any results assigned through result
                       pointers which are aliased in the function */
                    /*{{{  insert temporaries in front of aliased results*/
                    {
                      treenode *p;
                      for(p = paramlist; !EndOfList(p); p = NextItem(p))
                        {
                          treenode *thisparam = ThisItem(p);
                          if (TagOf(thisparam) == S_FNACTUALRESULT)
                            {
                              treenode *r = HExpOf(thisparam);
                            #if 0 /* bug TS/1563 26/05/92 */
                              if (usedin(r, paramlist, be_lexlevel) ||
                                  isafreevarof(basedecl(r), instancedfn))
                            #else
                              if (temp_required_for_fn_dest(r, sourceitem))
                            #endif
                                SetHExp(thisparam, gettemp(r, NM_WORKSPACE));
                            }
                        }
                    }
                    /*}}}*/
                    mapinstance(sourceitem);
                    /*{{{  store the register results*/
                    {
                      int nregresults;
                      treenode **regdests[MAXREGS];
                      setregdests(&nregresults, regdests, dest, instancedfn);
                      if (nregresults > 0)
                        mapstoreregs(regdests, nregresults);
                    }
                    /*}}}*/
                    /*{{{  move temporary results to their real locations, free temps*/
                    {
                      treenode *p;
                      for (p = paramlist; !EndOfList(p); p = NextItem(p))
                        {
                          treenode *const thisparam = ThisItem(p);
                          if (TagOf(thisparam) == S_FNACTUALRESULT)
                            {
                              treenode *const pnode = HExpOf(thisparam);
                              if (TagOf(pnode) == T_TEMP)
                                {
                                  treenode *const r = NDeclOf(pnode);
                                  mapsimpleassign(ntypeof(r), P_EXP, NDeclAddr(pnode),
                                                             P_TEMP, HExpAddr(thisparam));
                                  freetemp(pnode);
                                }
                            }
                        }
                    }
                    /*}}}*/
                  }
                  /*}}}*/
                }
            }
          else
            geninternal_is(GEN_ERROR_IN_ROUTINE, 5, "mapassign");
        }
        /*}}}*/
    }
    /*}}}*/
  else
    mapsimpleassign (ntypeof(dest), P_EXP, destptr, P_EXP, sourceptr);
}
/*}}}*/

/*{{{  PUBLIC void moveopd (destmode, dest, sourcemode, source)*/
/*****************************************************************************
 *
 * moveopd copies (sourcemode, source) to (destmode, dest)
 *         source and dest are assumed to be addressable.
 *
 *****************************************************************************/
PUBLIC void moveopd ( int destmode , treenode *dest , int sourcemode , treenode *source )
{
  const BOOL dest_devaccess   = is_devaccess(dest);
  const BOOL source_devaccess = is_devaccess(source);
  int type = ntypeof(dest);
  DEBUG_MSG(("moveopd\n"));

  /*{{{  see if we can pretend its an integer assign*/
  if (!use_shortintops && /* T9000 shorts 17/7/91 */
      isshortint(type))
    {
      /* If the source isn't addressable, we must be able to store directly
         to dest, otherwise the source would have to be preevaluated to a
         temporary.
         Otherwise, we need a temp to load test on source and dest
         decides whether we can pretend it's an integer assign */
      if (!isaddressableopd(sourcemode, source) ||
          (!needtemptoload(destmode, dest) &&
           !needtemptoload(sourcemode, source)))
        {
          DEBUG_MSG(("moveopd: converted INT16 move into INT move\n"));
          type = targetintsize;
        }
      #if 0 /* bug 328 11/9/90 */
      else
        {
          if (destmode == P_TEMP)
            zero_local_var(dest);
        }
      #endif
    }
  else if (isshorttype(type) &&
      issimpleopd(destmode, dest) && issimpleopd(sourcemode, source))
    {
      DEBUG_MSG(("moveopd: converted short move into INT move\n"));
      type = targetintsize;
    }
  /*}}}*/

  if (istargetintsize(type))
    /*{{{  optimise single length move*/
    {
      if (preeval(destmode, dest))
        /*{{{  ldptr dest; stl temp; source; ldl temp; stnl offset*/
        {
          const BIT32 offset = loadelementpointeroffset(NDeclOf(dest), MANY_REGS);
          storeinname(dest, 0);
          texpopd(sourcemode, source, MANY_REGS);
          loadname(dest, 0);
          movepointer(MOVEDIRN_STORE, I_STNL, offset, targetintsize, NULL, dest_devaccess);
        }
        /*}}}*/
      else if (regsforaddropd(destmode, dest) < MAXREGS)
        /*{{{  source; st dest*/
        {
          texpopd(sourcemode, source, MANY_REGS);
          storeinopd(destmode, dest, 0, MAXREGS - 1);
        }
        /*}}}*/
      else /* We know destmode isn't TEMP or TEMPPTR due to regsfor result */
        /*{{{  ldptr dest; source; rev; stnl offset*/
        {
          BIT32 offset = loadelementpointeroffset(dest, MAXREGS);
          texpopd(sourcemode, source, MANY_REGS);
          gensecondary(I_REV);
          movepointer(MOVEDIRN_STORE, I_STNL, offset, targetintsize, NULL, dest_devaccess);
        }
        /*}}}*/
    }
    /*}}}*/
  else if (isdoublelength(type))
    /*{{{  optimise double length move*/
    movelopd (destmode, dest, sourcemode, source);
    /*}}}*/
  else
    /*{{{  default to a move instruction (with some last minute optimisations)*/
    {
      treenode *desttype = gettype(dest);
      treenode *sourcetype = gettype(source);
      INT32 b;

      sourcemode = ptrmodeof(sourcemode);
      destmode   = ptrmodeof(destmode);

      b = bytesin(desttype);
      if (b == (-1)) b = bytesin(sourcetype);

      if
        /*{{{  can be optimised to a load and a store*/
        ( !source_devaccess && !dest_devaccess &&
        (
          (b == 1)
        ||((b == bytesperword) &&
           check_aligned(source, (int)b) && check_aligned(dest, (int)b))
        ||((b < bytesperword) && (b == 2) && use_shortintops && /* T9000 shorts 17/7/91 */
           check_aligned(source, (int)b) && check_aligned(dest, (int)b))
        ))
          {
            const int size_tag = (b == 1) ? S_BYTE :
                                 (b == 2) ? S_INT16 :
                                            targetintsize;

            tload2regs(destmode, dest, sourcemode, source, FALSE, FALSE);

            /* no need to sign extend, cos we're just about to store it */
            movepointer(MOVEDIRN_LOAD,  I_LDNL, 0, size_tag, NULL, source_devaccess);
            gensecondary(I_REV);
            movepointer(MOVEDIRN_STORE, I_STNL, 0, size_tag, NULL, dest_devaccess);
          }

        /*}}}*/
      else
        /*{{{  use a move instruction*/
        {
          if (b == (-1))
            /*{{{  length is an expression*/
            {
              BOOL preeval_e2, preeval_e3;
              int loadseq;
              treenode *lengthexp;

              if (TagOf(dest) == S_ARRAYITEM)
                {
                  DEBUG_MSG(("moveopd: dest is an ARRAYITEM; picking up length\n"));
                  lengthexp = ASLengthOf(dest);
                }
              else if (TagOf(dest) == S_SEGMENTITEM)
                {
                  DEBUG_MSG(("moveopd: dest is a SEGMENTITEM; picking up length\n"));
                  lengthexp = SLengthOf(dest);
                }
              else
                {
                  const int old = switch_to_temp_workspace();
                  DEBUG_MSG(("moveopd: neither ARRAYITEM nor SEGMENTITEM\n"));
                  lengthexp = scaletreeof(desttype, 1, be_lexlevel);
                  switch_to_prev_workspace(old);
                }

              /* We only do this call so that we have a loadseq parameter for
                 tload3regs, ideally we should make the binder save loadseq
                 in the tree. */
              loadseq = giveorder(sourcemode, source, destmode, dest,
                                  P_EXP, lengthexp, &preeval_e2, &preeval_e3);
              DEBUG_MSG(("moveopd: calling tload3regs; loadseq is %d\n", loadseq));
              tload3regs(sourcemode, source, destmode, dest,
                         P_EXP, lengthexp, loadseq, FALSE);
            }
            /*}}}*/
          else if (b == 0)
            return; /* bug TS/1782 26/08/92 */
          else
            /*{{{  length is a constant*/
            {
              tload2regs(sourcemode, source, destmode, dest, FALSE, FALSE);
              loadconstant(b);
            }
            /*}}}*/
          checkerror ();

          if (source_devaccess || dest_devaccess)
            /*{{{  device move*/
            {
              /*{{{  T9000_gamma_baddevaccess*/
              if (T9000_gamma_baddevaccess(&tx_global))
                {
                  /*{{{  issue warning*/
                  /* This warning is issued because turning off interrupts
                     during a devmove could have severe effects on
                     interrupt latency.
                     However, these devmove instructions are only generated
                     for oddities such as PORTs which are bigger than a word,
                     and will hence be very rare, so there doesn't seem
                     to be much point in worrying about it.

                     As a compromise, we just warn the user about the problem.
                     If this matters, they can either buy a chip which works,
                     (my preferred solution)
                     or code around it by changing the shape of the PORT.

                     When this bug work-around is removed, remember to remove
                     the warning message too.

                     CON - 7/3/94.
                  */
                  genwarn(GEN_DEVMOVE_LATENCY_PROBLEM);
                  /*}}}*/
                  gensecondary(I_INTDIS);
                  gencomment0("bug fix for device access");
                }
              /*}}}*/
              gensecondary(I_DEVMOVE);
              /*{{{  T9000_gamma_baddevaccess*/
              if (T9000_gamma_baddevaccess(&tx_global))
                {
                  gensecondary(I_INTENB);
                  gencomment0("bug fix for device access");
                }
              /*}}}*/
            }
            /*}}}*/
          else
            gensecondary(I_MOVE);
        }
        /*}}}*/
    }
    /*}}}*/
}
/*}}}*/

/*{{{  PRIVATE void tconstructorassign(dest, source)*/
#if 0 /* no constructors are passed to the backend any more */
/*****************************************************************************
 *
 *  tconstructorassign generates code to assign source to dest, where source
 *                     is a constructor.
 *
 *****************************************************************************/
PRIVATE void tconstructorassign ( treenode *dest , treenode *source )
{
  treenode *desttree;
  treenode *typetree;
  int old;
  INT32 scale;
#if 0
  fprintf(outfile, "\ntconstructorassign: source is :");
  printtree(0, source);
  fprintf(outfile, "\ntconstructorassign: dest is :");
  printtree(0, dest);
  fprintf(outfile, "\n");
#endif

#if 0 /* No VALOFs can return arrays - bug TS/1442 4/11/91 */
  if (isspecification(source) || TagOf(source) == S_VALOF)
    {
      tvalof(source, dest); /* bug TS/1442 28/10/91 */
      return;
    }
#endif

  /*{{{  find the constructor expression list*/
  if (TagOf(source) == S_CONSTRUCTOR)
    source = LitExpOf(source);
  else
    badtag(genlocn, TagOf(source), "tconstructorassign");
  /*}}}*/
  typetree = gettype(ThisItem(source));
  /*{{{  set up scale*/
  {
    const int bytes = bytesinscalar(basetype(typetree));
    scale = (bytes >= bytesperword)         ? wordsin(typetree) :
            (use_shortintops && bytes == 2) ? bytesin(typetree) / 2 : /* T9000 shorts 18/7/91 */
                                              bytesin(typetree);
  }
  /*}}}*/
  old = switch_to_temp_workspace();
  desttree = makeconstructordesttree(dest);
  switch_to_prev_workspace(old);
  /*{{{  assign each element of the constructor*/
  while (!EndOfList(source))
    {
      tsimpleassign(TagOf(typetree), P_EXP, desttree, P_EXP, ThisItem(source),
             MANY_REGS);
      SetASOffset(desttree, ASOffsetOf(desttree) + scale);
      source = NextItem(source);
    }
  /*}}}*/
}
#endif
/*}}}*/
/*{{{  PUBLIC void tsimpleassign (type, destmode, dest, sourcemode, source, regs)*/
/*{{{  comment*/
/*****************************************************************************
 *
 *  tsimpleassign generates code to evaluate (sourcemode, source) and store the
 *         result in (destmode, dest).
 *         'type' is the type of the destination, 'regs' is the maximum
 *         number of registers available for the assignment.
 *
 *
 *****************************************************************************/
/*}}}*/

/* The largest number of registers we may use */
PUBLIC void tsimpleassign ( int type , int destmode , treenode *dest , int sourcemode , treenode *source , int regs )
{
  const BOOL dest_devaccess = is_devaccess(dest);
  int r = (regs == MANY_REGS) ? MAXREGS : regs;
  /*{{{  pick out cases which can be turned into an INT assign*/
  /*if (istargetbytesize(type) && issimpleopd(destmode, dest))*/
  if ((istargetbytesize(type) ||
       (use_shortintops && isshortint(type))) /* T9000 shorts 17/7/91 */
      && issimpleopd(destmode, dest))
    /* Byte sized types are ok, because we can load bytes directly.
       Other short types (the only one currently is INT16 on a 32-bit
       machine) can only be treated as an INT  assign if both source
       and destination are simple. */
    type = targetintsize;
  else if ((type == S_CHAN) || (type == S_PORT))
    type = targetintsize;
  /*}}}*/

  if (has_fpu_core && isreal(type)
      && !isaddressableopd(sourcemode, source) /* re-inserted 11/9/90 bug 353 */
      && !isconstopd(sourcemode, source))      /* bug 708 12/9/90 */
    tfpassign(type, destmode, dest, sourcemode, source, regs);
  else if (istargetintsize(type))
    /*{{{  optimise single length assign*/
    if (has_fpu_core && isconversion(source) && isreal(ntypeof(OpOf(source))))
      /*{{{  special case a conversion from real on an fp processor*/
      {
        destmode = ptrmodeof(destmode);
        destmode = simplify(destmode, dest); /* added 24/9/90 for bug 739 */
        tfpexp(OpOf(source), regs, MANY_REGS);    /*      fp.source    */
        if (TagOf(source) == S_TRUNC)
          {
            if (has_directfp) gensecondary(I_FPRZ);/*     (fprz)       */
            else              genfpuentry(I_FPURZ);
          }
        gensecondary(CONVERSIONCHECKING ? I_FPRTOI32 /*   fprtoi32     */
                                        : I_FPINT);  /*   or fpint     */
        loadopd(destmode, dest, 0);               /*      ldp   dest   */
        checkerror ();
        gensecondary(I_FPSTNLI32);                /*      fpstnli32    */
      }
      /*}}}*/
    else if (preeval(destmode, dest))
      /*{{{  ldptr dest; stl temp; source; ldl temp; stnl*/
      {
        const BIT32 offset = loadopdpointeroffset(destmode, NDeclOf(dest));
        storeinname(dest, 0);
        texpopd_main(sourcemode, source, MANY_REGS, FALSE);
        loadnamepointer(dest, 0);
        movepointer(MOVEDIRN_STORE, I_STNL, offset, type, NULL, dest_devaccess);
      }
      /*}}}*/
    else if (directstore(destmode, dest, be_lexlevel))
      /*{{{  simplest case: generate source, store in dest*/
      {
        /* Only sign extend if we are initialising a temporary */
        const BOOL signextend_result = (TagOf(dest) == T_PREEVALTEMP);
        texpopd_main(sourcemode, source, regs, signextend_result);
        storeinopd(destmode, dest, 0, r - 1);
      }
      /*}}}*/
    else if (preeval(sourcemode, source))
      /*{{{  source; stl temp; ldptr dest; ldl temp; rev; stnl*/
      {
        BIT32 offset;
        texpopd_main(sourcemode, NDeclOf(source), MANY_REGS, FALSE);
        storeinname(source, 0);
        sourcemode = P_TEMP;
        offset = loadopdpointeroffset(destmode, dest);
        loadname(source, 0);
        gensecondary(I_REV);
        movepointer(MOVEDIRN_STORE, I_STNL, offset, type, NULL, dest_devaccess);
      }
      /*}}}*/
    else if (regsforopd(destmode, dest) < r)
      /*{{{  source; ldptr dest; stnl*/
      {
        BIT32 offset;
        texpopd_main(sourcemode, source, regs, FALSE);
        offset = loadelementpointeroffset(dest, r - 1);
        movepointer(MOVEDIRN_STORE, I_STNL, offset, type, NULL, dest_devaccess);
      }
      /*}}}*/
    else
      /*{{{  ldptr dest; source; rev; stnl*/
      {
        BIT32 offset = loadelementpointeroffset(dest, regs);
        texpopd_main(sourcemode, source, r - 1, FALSE);
        gensecondary(I_REV);
        movepointer(MOVEDIRN_STORE, I_STNL, offset, type, NULL, dest_devaccess);
      }
      /*}}}*/
    /*}}}*/
  else if (isdoublelength(type))
    movelopd(destmode, dest, sourcemode, source);
  else
    /*{{{  special cases for assign*/
    switch (type)
      {
        /*{{{  BYTE BOOL*/
        /* dest MUST be via a pointer */
        case S_BYTE:
        case S_BOOL:
          destmode = ptrmodeof(destmode);
          tload2regs(sourcemode, source, destmode, dest, FALSE, FALSE);
          movepointer(MOVEDIRN_STORE, I_STNL, 0, type, NULL, dest_devaccess);
          break;
        /*}}}*/
        /*{{{  INT16 on 32-bit machine*/
        case S_INT16: /* we must be on a 32-bit machine */
          if (use_shortintops) /* T9000 shorts 17/7/91 */
            { /* equivalent code to BYTE etc */
              destmode = ptrmodeof(destmode);
              tload2regs(sourcemode, source, destmode, dest, FALSE, FALSE);
              movepointer(MOVEDIRN_STORE, I_STNL, 0, type, NULL, dest_devaccess);
            }
          else
            {
              if (preeval(sourcemode, source))
                {
                  if (NModeOf(source) == NM_POINTER)
                    sourcemode = ptrmodeof(sourcemode); /* bug 1340 - 18/7/91 */
                  /*fprintf(outfile, "preeval stuff; genlocn is %lX, sourcemode is now %d\n", genlocn, sourcemode);*/
                  sourcemode = simplify(sourcemode, source);
                }
              moveopd (destmode, dest, sourcemode, source);
            }
          break;
        /*}}}*/
        /*{{{  INT64 REAL64 on 16-bit machine*/
        case S_INT64:
        case S_REAL64:
          moveqopd(destmode, dest, sourcemode, source);
          break;
        /*}}}*/
        /*{{{  ARRAY*/
        case S_ARRAY: case S_RECORD:
        #ifdef OCCAM2_5
          if ((type == S_RECORD) && (bytesin(gettype(dest)) <= bytesperword))
            /*{{{  assign directly*/
            tsimpleassign(S_INT, destmode, dest, sourcemode, source, regs);
            /*}}}*/
          else
        #endif
            /*{{{  do a fake 'quad length' move*/
            moveqopd(destmode, dest, sourcemode, source);
            /*}}}*/
        #if 0 /* constructors don't get this far any more */
          if (isaddressableopd(sourcemode, source))
            moveopd(destmode, dest, sourcemode, source);
          else
            /*{{{  right-hand side must contain a constructor*/
            {
              if (TagOf(dest) == T_TEMP)
                destmode = P_PTR;
              destmode = simplify(destmode, dest);

              if (preeval(sourcemode, source))
                {
                  /*SetTag(source, T_PREEVALTEMP);*/
                  temp_mark_as_evaluated(source); /* bug TS/1900 22/10/92 */
                  tconstructorassign(source, NDeclOf(source));
                  moveopd(destmode, dest, P_TEMP, source);
                }
              else
                {
                  if (TagOf(dest) == T_TEMP)
                    /*SetTag(dest, T_PREEVALTEMP);*/
                    temp_mark_as_evaluated(dest); /* bug TS/1900 22/10/92 */
                  tconstructorassign(dest, source);
                }
            }
            /*}}}*/
        #endif
          break;
        /*}}}*/
        default:
          geninternal_is(GEN_ERROR_IN_ROUTINE, 3, "tsimpleassign");
      }
    /*}}}*/
}
/*}}}*/
/*{{{  PRIVATE void tmultiassign (mdestlist, msourcelist)*/
/*{{{  comment*/
/*****************************************************************************
 *
 *  tmultiassign takes a multiple - assignment tree, tptr, which has been
 *               reordered by the mapper, and generates code for it.
 *
 *****************************************************************************/
/*}}}*/
PRIVATE void tmultiassign ( treenode *mdestlist , treenode *msourcelist )
{
  /*{{{  range check each assignment if neccessary*/
  if (RANGECHECKING)
    {
      treenode *destlist = mdestlist,
               *sourcelist = msourcelist;
      while (!EndOfList(destlist))
        {
          treenode *thisdest = ThisItem(destlist);
          if (ntypeof(thisdest) == S_ARRAY)
            tcheckdimensions(gettype(thisdest), gettype(ThisItem(sourcelist)));
          destlist = NextItem(destlist);
          sourcelist = NextItem(sourcelist);
        }
    }
  /*}}}*/
  /*{{{  do the assignment*/
  {
    treenode *destlist = mdestlist;
    treenode *sourcelist = msourcelist;
    treenode *regdest[MAXREGS];
    int nregtemps = 0, i;
    while (!EndOfList(destlist))
      /*{{{  move source to register, temporary, or directly to dest*/
      {
        /*{{{  do this assignment*/
        {
          treenode *dest   = ThisItem(destlist);
          treenode *source = ThisItem(sourcelist);
          if (TagOf(source) == T_TEMP)
            /*{{{  move to temporary*/
            tsimpleassign(ntypeof(dest), P_TEMP, source,
                   P_EXP, NDeclOf(source), MAXREGS - nregtemps);
            /*}}}*/
          else if (TagOf(source) == T_REGTEMP)
            /*{{{  load to register*/
            {
              texpopd(P_EXP, NDeclOf(source), MANY_REGS);
              regdest[nregtemps++] = dest;
            }
            /*}}}*/
          else
            /*{{{  move directly to dest*/
            tsimpleassign(ntypeof(dest), P_EXP, dest, P_EXP, source, MAXREGS - nregtemps);
            /*}}}*/
        }
        /*}}}*/
        destlist = NextItem(destlist);
        sourcelist = NextItem(sourcelist);
      }
      /*}}}*/
    /*{{{  store register temporaries in their destinations*/
    for (i = nregtemps - 1; i >= 0; i--)
      storeinopd(P_EXP, regdest[i], 0, (MAXREGS - 1) - i);
    /*}}}*/
    /*{{{  move temporaries to their destinations*/
    destlist = mdestlist;
    sourcelist = msourcelist;
    while (!EndOfList(destlist))
      {
        treenode *dest = ThisItem(destlist),
                 *source = ThisItem(sourcelist);
        if (TagOf(source) == T_TEMP)
          tsimpleassign(ntypeof(dest), P_EXP, dest, P_TEMP, source, MANY_REGS);
        destlist = NextItem(destlist);
        sourcelist = NextItem(sourcelist);
      }
    /*}}}*/
  }
  /*}}}*/
}
/*}}}*/
/*{{{  PUBLIC void tassign (dest, source)*/
/*****************************************************************************
 *
 *  tassign generates code to assign source to dest.
 *          source and dest may be lists, in which case it is assumed that
 *          they have been reordered and temporaries inserted as necessary.
 *
 *****************************************************************************/
PUBLIC void tassign ( treenode *dest , treenode *source )
{
  if (TagOf(dest) == S_LIST)
    /*{{{  we have a multiple assignment*/
    {
      if (listitems(source) > 1)
        tmultiassign(dest, source);
      else
        /*{{{  multi-valued function or valof*/
        /* N.B. Here I assume that functions and valofs always have scalar results
                - if they have vector results then hidden dimension size checking
                code must be generated. */
        {
          treenode *sourceitem = ThisItem(source);
          const int tag = TagOf(sourceitem);
          if (isspecification(sourceitem) || (tag == S_VALOF))
            tvalof(sourceitem, dest);
          else if (tag == S_FINSTANCE)
            {
              if (TagOf(INameOf(sourceitem)) == N_PREDEFFUNCTION)
                tpredef(sourceitem, dest);
              else
                {
                  tinstance(sourceitem);
                  /* store the register results */
                  /*{{{  store the register results*/
                  {
                    treenode **regdests[MAXREGS];
                    int nregresults;
                    setregdests(&nregresults, regdests, dest, INameOf(sourceitem));
                    if (nregresults > 0)
                    {
                      /* If assembly_output && 
                         ((alloc_strategy & ALLOC_NOCOMMENTS) == 0 */
                      gen_func_results (nregresults);
                      tstoreregs(regdests, nregresults);
                    }
                  }
                  /*}}}*/
                  /*{{{  move any results put into temporaries to their real destinations*/
                  {
                    treenode *paramlist;
                    for (paramlist = IParamListOf(sourceitem);
                         !EndOfList(paramlist); paramlist = NextItem(paramlist))
                      {
                        treenode *thisparam = ThisItem(paramlist);
                        if (TagOf(thisparam) == S_FNACTUALRESULT)
                          {
                            treenode *p = HExpOf(thisparam);
                            if (TagOf(p) == T_TEMP)
                              {
                                treenode *r = NDeclOf(p);
                                tsimpleassign(ntypeof(r), P_EXP, r, P_TEMP, p, MANY_REGS);
                              }
                          }
                      }
                  }
                  /*}}}*/
                }
            }
          else
            geninternal_is(GEN_ERROR_IN_ROUTINE, 5, "tassign");
        }
        /*}}}*/
    }
    /*}}}*/
  else
    {
      const int type = ntypeof(dest);
      if ((type == S_ARRAY) && RANGECHECKING)
        tcheckdimensions(gettype(dest), gettype(source));
      tsimpleassign (type, P_EXP, dest, P_EXP, source, MANY_REGS);
    }
}
/*}}}*/
