/* $Id: chk4.c,v 1.3 1997/03/25 14:52:40 djb1 Exp $ */

/*
 *	Occam two checker - driver for scoping and semantic analysis
 *	Copyright (C) 1987 Inmos Limited
 *
 *	This program is free software; you can redistribute it and/or modify
 *	it under the terms of the GNU General Public License as published by
 *	the Free Software Foundation; either version 2 of the License, or
 *	(at your option) any later version.
 *
 *	This program is distributed in the hope that it will be useful,
 *	but WITHOUT ANY WARRANTY; without even the implied warranty of
 *	MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *	GNU General Public License for more details.
 *
 *	You should have received a copy of the GNU General Public License
 *	along with this program; if not, write to the Free Software
 *	Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 */

/* #define DEBUG  */

/*{{{  include files*/
#ifdef HAVE_CONFIG_H
#  include <config.h>
#endif
#include <stdio.h>
#include <string.h>
#include "feinc.h"
#include "lexdef.h" /* init_predefs_lexer */
#include "chkerror.h"
#include "syndef.h" /* rpredef */
#include "chk1def.h"
#include "chk2def.h"
#include "chk4def.h"
#include "chkdef.h"
#include "objlib.h" /* Used for checking validity of library entries */

#include "predefhd.h" /* setting up HOSTEDGE for configurer */
/*}}}  */

/*{{{  definitions*/
/*#define NAME_STACK_SIZE 2000*/
/*#define UNDEC_NAME_STACK_SIZE 50*/

#define DEFAULT_NAMESTACK_SIZE_MAIN  256 /* this is extended when necessary */
#define DEFAULT_NAMESTACK_SIZE_ATTR   32 /* this is extended when necessary */
#define DEFAULT_NAMESTACK_SIZE_UNDEC   1 /* this is extended when necessary */

/*{{{  struct namestackentry*/
typedef struct namestackentry
  {
    wordnode *wordptr;
    treenode *namedesc;
    /*struct namestackentry *n_next;*/
  } namestackentry_t;

typedef int scopestack_t;

typedef struct
  {
    namestackentry_t *base;
    namestackentry_t *top;
    scopestack_t      count;
    int               max;
  } namestack_t;

/*}}}  */
/*}}}  */

/*{{{  PRIVATE variables*/
PRIVATE namestack_t namestack_main;
PRIVATE namestack_t namestack_undeclared;

PRIVATE BOOL insidepripar, nestedpripar;
/*PRIVATE BOOL insideinline;*/ /* no longer necessary to distinguish */
PRIVATE BOOL insideseparatelycompiledsc = FALSE;
PRIVATE BOOL nestedtimer; /* There is a TIMER inside this routine */
PRIVATE BOOL nestedplace; /* There is a PLACE AT inside this routine */
PRIVATE BOOL nestedport;  /* There is a PORT inside this routine */
PRIVATE scopestack_t label_scope; /* used for testing for duplicate label names */

#if 1 /*def CONFIG*/
PRIVATE int configcheckstate;
PRIVATE int configcheckscope;
PRIVATE int configcount;
PRIVATE int networkcount;
PRIVATE int mappingcount;

PRIVATE namestack_t namestack_attributes;
#endif

PRIVATE int check_lexlevel;
/*}}}  */

/*{{{  forward declarations*/
PUBLIC void scopeandcheck(treenode **tptr);
#if 1 /*def CONFIG*/
PRIVATEPARAM int configcheck(treenode *tptr, void *);
#endif
/*}}}  */

/*{{{  name stack handling*/
#if 0
/*{{{  namestack abstraction - linked list*/
/*{{{  PRIVATE void newnamestackentry()*/
PRIVATE void newnamestackentry(namestack_t *const namestack,
        treenode *defn, wordnode *name)
{
  namestackentry_t *const old = namestack->top;
  namestackentry_t *const ptr = (namestackentry_t *)newvec(sizeof(namestackentry_t));

  namestack->count++;

  ptr->wordptr  = name;
  ptr->namedesc = defn;
  ptr->n_next   = old;

  DEBUG_MSG(("newnamestackentry: %s, returning %x\n",
             (name == NULL) ? "(NULL)" : WNameOf(name), ptr));

  namestack->top = ptr;
}
/*}}}  */
/*{{{  PRIVATE void delete_namestackentry*/
PRIVATE void delete_namestackentry(namestack_t *const namestack)
{
  namestackentry_t *const temp = namestack->top;
  namestack->top = namestack->top->n_next;
  freevec(temp, sizeof(namestackentry_t));

  namestack->count--;
}
/*}}}  */

/*{{{  PRIVATE void init_namestack*/
PRIVATE void init_namestack(namestack_t *const namestack, const size_t size)
{
  USE_VAR(size); /* prevent unused variable warning */
  namestack->top   = NULL;
  namestack->count = 0;
  newnamestackentry(namestack, NULL, NULL);

  namestack->base  = namestack->top;
  namestack->count = 0;
}
/*}}}  */
/*{{{  PRIVATE void walk_namestack*/
PRIVATE void walk_namestack(namestack_t *const namestack,
        void (*f)(wordnode *,treenode *,void *), void *const voidptr)
{
  namestackentry_t *nmptr;
  for (nmptr = namestack->top;
       nmptr != namestack->base;
       nmptr = nmptr->n_next)
    f(nmptr->wordptr, nmptr->namedesc, voidptr);
}
/*}}}  */
/*{{{  PRIVATE treenode *searchnamestack (n)*/
PRIVATE treenode *searchnamestack (register namestack_t *const namestack,
                                   register wordnode *const n )
{
  register namestackentry_t *stack = namestack->top;
  /* In a large example program, the number of iterations in this routine
     were estimated as follows:
     example 1: Average number of iterations per call: 100,
                Average number of items on the name stack: 814.
     example 2: Average number of iterations per call: 150,
                Average number of items on the name stack: 900.

     Thus, this loop is very speed critical. In example 2, this one
     routine cost 4.1% of the execution time of the compiler.
  */

  /* we MUST ensure that it's on the stack somewhere */
  namestack->base->wordptr = n; /* its namedesc is already NULL */

#if 1
  /*{{{  normal copy*/
  /* NOTE - this loop is very speed critical. It is called thousands
     of times! */
  while (stack->wordptr != n)
    stack = stack->n_next;
  return stack->namedesc;
  /*}}}  */
#else
  /*{{{  diagnostic copy*/
  {
    static int calls = 0;
    static int total_len = 0;
    static int total_searches = 0;
    int len = 0;
    int searches = 0;
    treenode *ans;
    while (stack->wordptr != n)
      {
        stack = stack->n_next;
        len++;
      }
    ans = stack->namedesc;
    searches = len;
    while (stack != NULL)
      {
        stack = stack->n_next;
        len++;
      }

    calls ++;
    total_len += len;
    total_searches += searches;
    printf("srch: call:%d, len:%d, s:%d, tlen:%d, t_s:%d, a_len:%d, a_s:%d\n",
           calls, len, searches, total_len, total_searches, total_len/calls, total_searches/calls);
    return ans;
  }
  /*}}}  */
#endif
}
/*}}}  */
/*}}}  */
#else
/*{{{  namestack abstraction - realloced array*/
/*{{{  PRIVATE void newnamestackentry()*/
PRIVATE void newnamestackentry(namestack_t *const namestack,
        treenode *defn, wordnode *name)
{
  if (namestack->count == namestack->max - 1)
    /*{{{  full stack - realloc double the size*/
    {
      const int new_max = namestack->max * 2;
      namestackentry_t *const new = memalloc(new_max * sizeof(namestackentry_t));
      memcpy(new, namestack->base,    namestack->max * sizeof(namestackentry_t));
      namestack->top = new + (namestack->top - namestack->base);
      memfree(namestack->base);
      namestack->base = new;
      namestack->max  = new_max;
      /*printf("doubling namestack to %d\n", new_max);*/
    }
    /*}}}  */

  namestack->top++;
  namestack->top->wordptr  = name;
  namestack->top->namedesc = defn;
  namestack->count++;
}
/*}}}  */
/*{{{  PRIVATE void delete_namestackentry*/
PRIVATE void delete_namestackentry(namestack_t *const namestack)
{
  namestack->top--;
  namestack->count--;
}
/*}}}  */

/*{{{  PRIVATE void init_namestack*/
PRIVATE void init_namestack(namestack_t *const namestack, const size_t size)
{
  namestack->max   = size;
  namestack->base  = (size == 0) ? NULL : memalloc(size * sizeof(namestackentry_t));
  namestack->top   = namestack->base;
  namestack->count = 0;

  if (size != 0)
    namestack->base->namedesc = NULL;
}
/*}}}  */
/*{{{  PRIVATE void walk_namestack*/
PRIVATE void walk_namestack(namestack_t *const namestack,
        void (*f)(wordnode *,treenode *,void *), void *const voidptr)
{
  int i;
  for (i = namestack->count; i > 0; i--)
    {
      namestackentry_t *const nmptr = &namestack->base[i];
      f(nmptr->wordptr, nmptr->namedesc, voidptr);
    }
}
/*}}}  */
/*{{{  PRIVATE treenode *searchnamestack (n)*/
PRIVATE treenode *searchnamestack (register namestack_t *const namestack,
                                   register wordnode *const n )
{
  register namestackentry_t *t = namestack->top;

  /* we MUST ensure that it's on the stack somewhere */
  namestack->base->wordptr = n; /* its namedesc is already NULL */

  /* NOTE - this loop is very speed critical. It is called thousands
     of times! */
  while (t->wordptr != n)
    t--;
  return t->namedesc;
}
/*}}}  */
/*}}}  */
#endif

/*{{{  PRIVATE treenode *findname (n)*/
#if 1
/* above was commented set to 0 so that macro invoked instead */
PUBLIC treenode *findname ( wordnode *n ) /* changed to public */
{
  /* Look up the name pointed to by n on the name stack.
     Return NULL if not found */
  return searchnamestack(&namestack_main, n);
}
#else
#define findname(n) searchnamestack(&namestack_main, (n))
#endif
/*}}}  */
/*{{{  PRIVATEPARAM void addname (nptr)*/
PRIVATEPARAM void addname ( treenode *const nptr, void *const voidptr )
{
  USE_VAR(voidptr); /* remove unused variable warning */

  /*{{{  (optionally) warn about descoping names*/
  {
    const BOOL param = ((TagOf(nptr) == N_PARAM) ||
                        (TagOf(nptr) == N_VALPARAM));
    if ((( param && current_fe_data->fe_warning_descoped_p
                 && !insideseparatelycompiledsc )
      || (!param && current_fe_data->fe_warning_descoped_n ) )
     && (findname(NNameOf(nptr)) != NULL))
      msg_out_s(SEV_WARN, CHK, CHK_NAME_IS_DESCOPED, LocnOf(nptr),
                WNameOf(NNameOf(nptr)));
  }
  /*}}}  */

  SetNScope(nptr, namestack_main.count);
  newnamestackentry(&namestack_main, nptr, NNameOf(nptr));

  DEBUG_MSG(("addname: %s, ptr is %x, namep is %d\n",
             WNameOf(namestack_main.top->wordptr), namestack_main.top, namestack_main.count));
}
/*}}}  */
/*{{{  PRIVATE treenode *addundeclaredname (name)*/
/* Pointer to node containing text of name */
PRIVATE treenode *addundeclaredname ( wordnode *const name )
{
  /*treenode *const nptr = declname(N_DECL, chklocn, name, undeclaredp, NULL);*/
  treenode *const nptr =
    newnamenode(N_DECL, chklocn, name, undeclaredp, NULL, 0, 0, NM_DEFAULT);

  newnamestackentry(&namestack_undeclared, nptr, name);

  return nptr;
}
/*}}}  */
/*{{{  PRIVATE treenode *findundeclaredname (n)*/
#if 0
PRIVATE treenode *findundeclaredname ( wordnode *n )
{
  /* Look up the name pointed to by n on the name stack.
     Return NULL if not found */
  return searchnamestack(&namestack_undeclared, n);
}
#else
#define findundeclaredname(n) searchnamestack(&namestack_undeclared, (n))
#endif
/*}}}  */
/*{{{  PUBLIC treenode *lookupname (n)*/ /* was PRIVATE */
PUBLIC treenode *lookupname ( wordnode *n, treenode *tptr )
{
  /* Look up the name pointed to by n on the name stack.
     If found, return a pointer to the name record,
     otherwise make new entry. */
  /* tptr is the namenode returned from findname previously */
  if (tptr == NULL)
    /*{{{  check to see if this undeclared variable has already been seen*/
    {
      const SOURCEPOSN locn = err_sourcefileof(chklocn);
      tptr = findundeclaredname (n);
      if (tptr == NULL)
        /*{{{  Never been seen before, create an undeclared name node*/
        {
          tptr = addundeclaredname(n);
          chkerr_s(CHK_NAME_NOT_DECLARED, locn, WNameOf(n));
        }
        /*}}}  */
      {
        treenode *old = NDeclOf(tptr);
        /* This test added 16/10/90 so we don't repeat previous lines */
        if ((old == NULL) || (LocnOf(old) != locn))
          { /* add to list of undeclared instances */
            treenode *dptr = newlistnode(S_LIST, locn, NULL, old);
            SetNDecl(tptr, dptr);
          }
      }
    }
    /*}}}  */
  return (tptr);
}
/*}}}  */
/*{{{  PRIVATE void printlinenums (file, listptr)*/
PRIVATE void printlinenums ( FILE *file , treenode *listptr )
{
  int current_file = (-1);
  while (listptr != NULL)
    {
      if (current_file != (-1))
        fputc(',', file);
      if (FileNumOf(LocnOf(listptr)) != current_file)
        {
          current_file = FileNumOf(LocnOf(listptr));
          fprintf(file, " \"%s\" line%s ",
                  fe_lookupfilename(current_fe_handle, current_file),
          ((NextItem(listptr) != NULL) &&
           (FileNumOf(LocnOf(NextItem(listptr))) == current_file)) ? "s" : "");
        }
      fprintf(file, "%d", FileLineOf(LocnOf(listptr)));
      listptr = NextItem(listptr);
    }
}
/*}}}  */
/*{{{  PRIVATE void sub_printundeclarednames*/
PRIVATE void sub_printundeclarednames(wordnode *const name, treenode *const nptr,
        void *const voidptr)
{
  if (TagOf(NTypeOf(nptr)) == S_UNDECLARED)
    {
      FILE *const file = voidptr;
      reverselist(NDeclAddr(nptr));
      fprintf (file, "%s undeclared on", WNameOf(name));
      printlinenums (file, NDeclOf(nptr));
      fputc('\n', file);
    }
}
/*}}}  */
/*{{{  PUBLIC void printundeclarednames*/
PUBLIC void printundeclarednames ( FILE *const file )
{
  walk_namestack(&namestack_undeclared, sub_printundeclarednames, file);
}
/*}}}  */
/*{{{  PUBLIC int inscope (nptr)                   NEVER USED*/
#if 0 /* inscope is never used */
/* Return TRUE if the symbol nptr is in scope, FALSE otherwise */
PUBLIC int inscope ( treenode *nptr )
{
  return (nptr == findname(NNameOf(nptr)));
}
#endif /* inscope is never used */

/*}}}  */

/*{{{  NODE attribute handling (Configurer/NDL reader)*/
#if 1 /*def CONFIG*/
PRIVATE const int controlport_attrs[] = {PD_ATTR_CONTROL, PD_ATTR_DATA};
PRIVATE const int route_attrs      [] = {PD_ATTR_OUTWARD, PD_ATTR_RETURN};
/*{{{  PRIVATE BOOL in_attribute_list*/
PRIVATE BOOL in_attribute_list(const int attr, const int attrs[], const int n)
{
  int i;
  for (i = 0; i < n; i++)
    if (attr == attrs[i]) return TRUE;
  return FALSE;
}
/*}}}  */

/*{{{  PRIVATE treenode *find_attribute (n)*/
#if 0
PRIVATE treenode *find_attribute ( wordnode *n )
{
  /* Look up the name pointed to by n on the name stack.
     Return NULL if not found */

  return searchnamestack(&namestack_attributes, n);
}
#else
#define find_attribute(n) searchnamestack(&namestack_attributes, (n))
#endif
/*}}}  */
/*{{{  PRIVATE void add_attribute*/
PRIVATE void add_attribute(treenode *nptr)
{
  newnamestackentry(&namestack_attributes, nptr, NNameOf(nptr));
}
/*}}}  */
/*{{{  PRIVATE init_attributes*/
PRIVATE void init_attributes(void)
{
  /* prepare the attributes for S_NODE */
  init_namestack(&namestack_attributes,
                 ((current_fe_data->fe_lang & FE_LANG_OCCAM) != 0) ? 0 :
                 DEFAULT_NAMESTACK_SIZE_ATTR);
}
/*}}}  */

#endif
/*}}}  */

/*{{{  PRIVATE int current_scope()*/
#if 0 /* make it a macro instead */
PRIVATE BIT32 current_scope(void)
{
  return namestack_main.count;
}
#else
#define current_scope() namestack_main.count
#endif
/*}}}  */
/*{{{  PRIVATE scopestack_t markscopenames()*/
#if 0 /* make it a macro instead */
PRIVATE scopestack_t markscopenames(void)
{
  DEBUG_MSG(("markscopenames: namestack is %x (%s), namep is %d\n", namestack_main.top, WNameOf(namestack_main.top->wordptr), namestack_main.count));
  return namestack_main.count;
}
#else
#define markscopenames() namestack_main.count
#endif
/*}}}  */
/*{{{  PRIVATE void chk_descopename*/
PRIVATE void chk_descopename(treenode *const nptr)
{
  if (!NUsedOf(nptr) &&
      ((NTypeOf(nptr) == NULL) || TagOf(NTypeOf(nptr)) != S_UNDECLARED))
    {
      const char *const namestring = WNameOf(NNameOf(nptr));
      const SOURCEPOSN locn = LocnOf(nptr);
      switch (TagOf(nptr))
        /*{{{  cases*/
        {
          default:          break;
          case N_DECL:
          case N_ABBR:
          case N_RETYPE:    if (current_fe_data->fe_warning_unused_v)
                              msg_out_s(SEV_WARN, CHK, CHK_NAME_NOT_USED, locn, namestring);
                            break;
          case N_PARAM:
          case N_VALPARAM:  if (current_fe_data->fe_warning_unused_p)
                              msg_out_s(SEV_WARN, CHK, CHK_PARAM_NOT_USED, locn, namestring);
                            break;
          case N_VALABBR:
          case N_VALRETYPE: if ((current_fe_data->fe_warning_unused_v) && !isconst(nptr))
                              msg_out_s(SEV_WARN, CHK, CHK_NAME_NOT_USED, locn, namestring);
                            break;
          case N_INLINEPROCDEF: /* bug TS/2045 18/01/93 */
          case N_INLINEFUNCDEF:
          case N_PROCDEF:
          case N_SFUNCDEF:
          case N_LFUNCDEF:  if ((NLexLevelOf(nptr) != 0) && (current_fe_data->fe_warning_unused_r))
                              msg_out_s(SEV_WARN, CHK, CHK_ROUTINE_NOT_USED, locn,namestring);
                            break;
        }
        /*}}}  */
    }
}
/*}}}  */
/*{{{  PRIVATE void descopenames (scopestack_t n)*/
PRIVATE void descopenames ( const scopestack_t n )
{
  DEBUG_MSG(("descopenames: marker is %d\n", n));
  if (namestack_main.count < n) err_abort();
  while (namestack_main.count > n)
    {
      DEBUG_MSG(("descopenames: removing %x (%s) (%d)\n",
                 namestack_main.top, WNameOf(namestack_main.top->wordptr), namestack_main.count));
      chk_descopename(namestack_main.top->namedesc);
      delete_namestackentry(&namestack_main);
    }
}
/*}}}  */

/*}}}  */

/*{{{  PRIVATEPARAM int decllab(treenode *tptr)*/
/*{{{  comment*/
/*****************************************************************************
 *
 *  decllab is the routine called by the tree walker on each node
 *          when we are declaring GUY code labels.
 *          Don't go into routine bodies, but do go into expressions.
 *
 *****************************************************************************/
/*}}}  */
PRIVATEPARAM int decllab ( treenode *tptr, void *voidptr )
{
  /* We do successive specifications in a loop to avoid excessive recursion */
  if (isspecification(tptr))
    {
      /* S_LABELDEF will only ever appear as a specification on its own */
      if (TagOf(tptr) == S_LABELDEF)
        {
          treenode *const nptr = DNameOf(tptr);
          const treenode *const prev = findname(NNameOf(nptr));
          if (prev != NULL && NScopeOf(prev) >= label_scope) /* bug 1348 21/8/91 */
            chkerr_s(CHK_DUPLICATE_LABEL, LocnOf(tptr), WNameOf(NNameOf(nptr)));
          else
            addname(nptr, NULL);
          return STOP_WALK; /* body of S_LABELDEF is always NULL */
        }
      tptr = skipspecifications(tptr); /* skip nested PROCs etc */
      prewalktree(tptr, decllab, voidptr);
      return STOP_WALK;
    }
  else
    return CONTINUE_WALK;
}
/*}}}  */
/*{{{  PRIVATE void decllabels(treenode *tptr)*/
/*{{{  comment*/
/*****************************************************************************
 *
 *  decllabels takes a parse tree and brings into scope all the labels in
 *             that tree. It does not look inside nested routines.
 *
 *****************************************************************************/
/*}}}  */
PRIVATE void decllabels ( treenode *tptr )
{
  const scopestack_t saved_label_scope = label_scope;
  label_scope = current_scope();
  prewalktree(tptr, decllab, NULL);
  label_scope = saved_label_scope;
}
/*}}}  */
/*{{{  PRIVATE void scopeandchecksc (tptr)*/
/*{{{  comment*/
/*****************************************************************************
 *
 *  scopeandchecksc applies scoping and type checking to separately
 *                  compiled routine body tptr.
 *
 *****************************************************************************/
/*}}}  */
PRIVATE void scopeandchecksc ( treenode **tptr )
{
  while (*tptr != NULL)
    {
      treenode *t = *tptr;
      if (TagOf(t) != S_NAME) chklocn = LocnOf(t);
      switch(TagOf(t))
        {
          default:
            return;
          /*{{{  SEQ PRIPAR          break*/
          case S_SEQ:
            tptr = CBodyAddr(t);
            break;
          case S_PRIPAR:
            nestedpripar = TRUE;
            tptr = CBodyAddr(t);
            break;
          /*}}}  */
          /*{{{  OUTPUT, INPUT       break*/
          case S_OUTPUT:
          case S_INPUT:
            tptr = LHSAddr(t);
            break;
          /*}}}  */
          /*{{{  NAME                return*/
          case S_NAME :
            {
              wordnode *n = (wordnode *)t;
              *tptr = lookupname(n, findname(n));
            }
            return;
          /*}}}  */
          /*{{{  LIST                break*/
          case S_LIST :
            scopeandchecksc(ThisItemAddr(t));
            tptr = NextItemAddr(t);
            break;
          /*}}}  */
        }
    }
}
/*}}}  */

/*{{{  PRIVATE treenode *cset(tptr)*/
/*****************************************************************************
 *
 *  cset performs semantic checking on a set node
 *
 *****************************************************************************/
PRIVATE treenode *cset ( treenode *volatile tptr )
/* 'volatile' is inserted just to prevent an erroneous gcc warning:
   warning: argument `tptr' may be clobbered by `longjmp'
*/
{
  jmp_buf savedenv;
  DEBUG_MSG(("cset\n"));
  memcpy((char *)savedenv, (char *)env, sizeof(env));

  if (setjmp (env) != 0)
    tptr = NULL;
  else
    {
      int set_type; /* type of the SET expression */
      chklocn = LocnOf(tptr);

      /*{{{  check that it is a NODE/ROUTE*/
      {
        treenode *t = typecheck (STDevOf(tptr), unknownnodeptr);
        if (!sametype(TagOf(t), S_NODE) && !sametype(TagOf(t), S_ROUTE))
          chkreport_ss(CHK_INVTYPE_CONFIG, chklocn, "SET", "NODE");
        set_type = TagOf(t);
      }
      /*}}}  */
      /*{{{  check the attribute settings*/
      {
      #if 1
        /* we're basically doing a checkasslist here, but we can't call
           typecheck_main on lhs cos fields on their own are disallowed */
        int varno = 1;
        treenode *lhs = STAttrNameOf(tptr);
        treenode *rhs = STAttrExpOf(tptr);

        while (!EndOfList(lhs) && !EndOfList(rhs))
          /*{{{  check one attribute*/
          {
            treenode *type;
            treenode *attr;
            /*{{{  get type of lhs*/
            {
              chk_permit_lonely_fields = TRUE;
            #if 0 /* This only allows simple names here */
              {
                wordnode *n = (wordnode *)ThisItem(lhs);
                NewItem(lookupname(n, find_attribute(n)), lhs);
                attr = ThisItem(lhs);
                if (TagOf(attr) != N_FIELD)
                  chkreport_s(CHK_INVTYPE_ATTR, chklocn, WNameOf(NNameOf(attr)));
                type = NTypeOf(attr);
              }
            #else
              {
                wordnode *n;
                treenode **t = ThisItemAddr(lhs);

                /*{{{  loop through subscripts*/
                while (TagOf(*t) == S_ARRAYSUB)
                  t = ASBaseAddr(*t);
                /*}}}  */
                /*{{{  scope the attribute etc*/
                if (TagOf(*t) != S_NAME)
                  chkreport(CHK_ILLEGAL_CONFIG_CONSTRUCT, chklocn);
                n = (wordnode *)(*t);
                *t = lookupname(n, find_attribute(n));

                attr = *t;
                if (TagOf(attr) != N_FIELD)
                  chkreport_s(CHK_INVTYPE_ATTR, chklocn, WNameOf(NNameOf(attr)));

                scopeandcheck(ThisItemAddr(lhs));
                chklocn = LocnOf(tptr); /* somehow this gets hit by scopeandcheck */

                /*}}}  */

              }
              type = typecheck(ThisItem(lhs), unknownnodeptr);
            #endif
              chk_permit_lonely_fields = FALSE;
            }
            /*}}}  */
            /*{{{  check rhs is same type as lhs*/
            {
              if (!typesequivalent(type, typecheck(ThisItem(rhs), type), FALSE))
                {
                  char buf[20]; /* enough for a decimal number */
                  sprintf(buf, "%d", varno);
                  chkreport_s(CHK_INVTYPE_MULTI_ASS, chklocn, buf);
                }
              else if (set_type != S_UNDECLARED)
                {
                  /* Some attributes are only permitted for ROUTEs,
                     and vice-versa. Hence route_attr must be TRUE iff
                     the type is S_ROUTE */
                  const BOOL route_attr =
                    in_attribute_list(NModeOf(attr), route_attrs,
                         sizeof(route_attrs)/sizeof(route_attrs[0]));
                  if (route_attr != sametype(set_type, S_ROUTE))
                    chkerr_ss(CHK_FIELD_NOT_VALID, LocnOf(tptr),
                              WNameOf(NNameOf(attr)), tagstring(set_type));
                }

              /*{{{  you may not SET a link, etc*/
              switch(TagOf(basetype_tree(type)))
                {
                  case S_INT: case S_BYTE: case S_BOOL:
                  case S_NODE: case S_ARC:
                    break;
                  default:
                    chkreport_s(CHK_FIELD_NOT_SET, chklocn, WNameOf(NNameOf(attr)));
                    break;
                }
              /*}}}  */

            }
            /*}}}  */

            lhs = NextItem(lhs);
            rhs = NextItem(rhs);
            varno++;
          }
          /*}}}  */

        if (!EndOfList(lhs))
          chkreport (CHK_TOO_MANY_VARS, chklocn);
        else if (!EndOfList(rhs))
          chkreport (CHK_TOO_FEW_VARS, chklocn);
      #else
        checkasslist(STAttrNameOf(tptr), STAttrExpOf(tptr));
      #endif
      }
      /*}}}  */
      /*{{{  fold the expressions*/
      SetSTDev(tptr, foldexp(STDevOf(tptr)));
      if (STAttrExpOf(tptr) != NULL)
        SetSTAttrExp(tptr, foldexplist(STAttrExpOf(tptr)));
      /*}}}  */
    }

  memcpy((char *)env, (char *)savedenv, sizeof(env));
  DEBUG_MSG(("cset finished\n"));
  chk_permit_lonely_fields = FALSE; /* incase we aborted via an error */
  return (tptr);
}
/*}}}  */
/*{{{  PRIVATE treenode *cconnect(tptr)*/
/*****************************************************************************
 *
 *  cconnect performs semantic checking on a connect node
 *
 *****************************************************************************/
PRIVATE treenode *cconnect ( treenode *volatile tptr )
/* 'volatile' is inserted just to prevent an erroneous gcc warning:
   warning: argument `tptr' may be clobbered by `longjmp'
*/
{
  jmp_buf savedenv;
  DEBUG_MSG(("cconnect\n"));
  memcpy((char *)savedenv, (char *)env, sizeof(env));
  if (setjmp (env) != 0)
    tptr = NULL;
  else
  {
    treenode *t;
    chklocn = LocnOf(tptr);

    /*{{{  check the two EDGEs*/
    t = typecheck (ConnectFromEdgeOf(tptr), edgenodeptr);
    if (!sametype(TagOf(t), S_EDGE))
      chkreport_ss(CHK_INVTYPE_CONFIG, chklocn, "CONNECT", "EDGE");
    SetConnectFromEdge(tptr, foldexp(ConnectFromEdgeOf(tptr)));

    t = typecheck (ConnectToEdgeOf(tptr), edgenodeptr);
    if (!sametype(TagOf(t), S_EDGE))
      chkreport_ss(CHK_INVTYPE_CONFIG, chklocn, "CONNECT", "EDGE");
    SetConnectToEdge(tptr,   foldexp(ConnectToEdgeOf(tptr)));
    /*}}}  */

    if (ConnectArcOf(tptr) != NULL)
      /*{{{  check the ARC*/
      {
        t = typecheck (ConnectArcOf(tptr), arcnodeptr);
        if (!sametype(TagOf(t), S_ARC))
          chkreport_ss(CHK_INVTYPE_CONFIG, chklocn, "WITH", "ARC");
        SetConnectArc(tptr, foldexp(ConnectArcOf(tptr)));
      }
      /*}}}  */
  }
  memcpy((char *)env, (char *)savedenv, sizeof(env));
  DEBUG_MSG(("cconnect finished\n"));

  return (tptr);
}
/*}}}  */
/*{{{  PRIVATE treenode *cmap(tptr)*/
/*****************************************************************************
 *
 *  cmap performs semantic checking on a map node
 *
 *****************************************************************************/
PRIVATE treenode *cmap ( treenode *volatile tptr )
/* 'volatile' is inserted just to prevent an erroneous gcc warning:
   warning: argument `tptr' may be clobbered by `longjmp'
*/
{
  jmp_buf savedenv;
  DEBUG_MSG(("cmap\n"));
  memcpy((char *)savedenv, (char *)env, sizeof(env));
  if (setjmp (env) != 0)
    tptr = NULL;
  else
  {
    treenode *t, *maplist;
    treenode *sourcetype = NULL; /* initialised to shut up gcc's optimiser */
    chklocn = LocnOf(tptr);

    if( (current_fe_data->fe_lang & FE_LANG_NDL) != 0 )
      /*{{{  Checks for NDL MAP statement*/
      {
        /*{{{  check the destination of the mapping*/
        t = typecheck(MapDestOf(tptr), nodenodeptr);
        if (!sametype(TagOf(t), S_EDGE))
          chkreport(CHK_INV_RHS_NDL_MAPPING, chklocn);
        SetMapDest(tptr, foldexp(MapDestOf(tptr)));
        /*}}}  */
        /*{{{  check the source*/
        maplist = MapSourceOf(tptr);
        if( !EndOfList(NextItem(maplist)) )
          chkreport(CHK_NDL_MAPPING_TOO_MANY_SOURCES, chklocn);
        else
          {
            t = typecheck(ThisItem(maplist), edgenodeptr);
            if (!sametype(TagOf(t), TagOf(edgenodeptr)))
              chkreport(CHK_INV_LHS_NDL_MAPPING, chklocn);
            NewItem(foldexp(ThisItem(maplist)), maplist);
          }
        /*}}}  */
      }
      /*}}}  */
    else
      /*{{{  Checks for configurer MAP statement*/
      {
        /*{{{  check the destination of the mapping*/
        t = typecheck(MapDestOf(tptr), nodenodeptr);
        if (sametype(TagOf(t), S_NODE))
          sourcetype = nodenodeptr;
        else if (sametype(TagOf(t), S_ARC))
          sourcetype = channodeptr;
        else
          chkreport(CHK_INV_MAPPING_RHS, chklocn);
        SetMapDest(tptr, foldexp(MapDestOf(tptr)));
        /*}}}  */
        /*{{{  check the sources*/
        for (maplist = MapSourceOf(tptr); !EndOfList(maplist); maplist = NextItem(maplist))
          {
            t = typecheck(ThisItem(maplist), sourcetype);
            if (!sametype(TagOf(t), TagOf(sourcetype)))
              chkreport_s(CHK_INV_MAPPING_LHS, chklocn, tagstring(TagOf(sourcetype)));
            NewItem(foldexp(ThisItem(maplist)), maplist);
          }
        /*}}}  */
        /*{{{  check the priority*/
        if (MapPriOf(tptr) != NULL)
          {
            if (TagOf(sourcetype) != S_NODE)
              chkreport(CHK_INV_MAPPING_NOPRI, chklocn);
            t = typecheck(MapPriOf(tptr), intnodeptr);
            if (!sametype(TagOf(t), S_INT))
              chkreport(CHK_INV_MAPPING_PRI, chklocn);
            SetMapPri(tptr, foldexp(MapPriOf(tptr)));
          }
        /*}}}  */
      }
      /*}}}  */
  }
  memcpy((char *)env, (char *)savedenv, sizeof(env));
  DEBUG_MSG(("cmap finished\n"));

  return (tptr);
}
/*}}}  */

/*{{{  PRIVATE void lookup_special_namespace*/
PRIVATE void lookup_special_namespace(treenode **tptr, treenode *const type_tree)
{
  wordnode *const name = (wordnode *)*tptr;
  treenode *nptr;
  switch (TagOf(type_tree))
    {
      /*{{{  NODE attributes*/
      case S_NODE: case S_CONTROLPORT:
        nptr = lookupname(name, find_attribute(name));
        if (TagOf(nptr) == N_FIELD)
          /*{{{  check validity of that attribute*/
          {
            /* Some attributes are only permitted for CONTROLPORT,
               and vice-versa. Hence controlport_attr must be TRUE iff
               type is S_CONTROLPORT */
            const BOOL controlport_attr =
              in_attribute_list(NModeOf(nptr), controlport_attrs,
                   sizeof(controlport_attrs)/sizeof(controlport_attrs[0]));
            if (controlport_attr != (TagOf(type_tree) == S_CONTROLPORT))
              chkerr_ss(CHK_FIELD_NOT_VALID, LocnOf(*tptr),
                        WNameOf(NNameOf(nptr)), tagstring(TagOf(type_tree)));
          }
          /*}}}  */
        *tptr = nptr;
        break;
      /*}}}  */
      /*{{{  fields of RECORDs*/
      #ifdef OCCAM2_5
      case S_RECORD:
        /*{{{  look for a field name*/
        {
          treenode *t;
          for (nptr = NULL, t = ARTypeOf(type_tree);
               nptr == NULL && t != NULL;
               t = DBodyOf(t))
            if (NNameOf(DNameOf(t)) == name)
              nptr = DNameOf(t);
        }
        /*}}}  */

        if (nptr == NULL)
          scopeandcheck(tptr);
        else
          *tptr = nptr;
        break;
      #endif
      /*}}}  */
      /*{{{  fields of tagged PROTOCOLs*/
      #ifdef OCCAM2_5
      case N_TPROTDEF:
        /*{{{  look for a field of a tagged protocol*/
        {
          treenode *t;
          for (nptr = NULL, t = NTypeOf(type_tree);
               nptr == NULL && !EndOfList(t);
               t = NextItem(t))
            if (NNameOf(ThisItem(t)) == name)
              nptr = ThisItem(t);
        }
        /*}}}  */
        if (nptr != NULL)
          *tptr = nptr;
        break;
      #endif
      /*}}}  */
      default:
        break;
    }
}
/*}}}  */

/*{{{  PRIVATE treenode *csubscript*/
/*****************************************************************************
 *
 *  csubscript turns array subscripting into record subscripting if necessary
 *
 *****************************************************************************/
PRIVATE treenode *csubscript ( treenode *const tptr )
{
#ifdef OCCAM2_5
  if (TRUE)
#else
  if ((current_fe_data->fe_lang & FE_LANG_OCCAM) == 0) /* Allow records */
#endif
    {
      treenode *const type_tree = typecheck(ASBaseOf(tptr), unknownnodeptr);
      treenode *const sub_tree  = follow_user_type(type_tree);
      /*{{{  debugging*/
      #ifdef DEBUG
      DEBUG_MSG(("csubscript: tree of Base is"));
      printtree(stdout, 0, ASBaseOf(tptr));
      DEBUG_MSG(("\ncsubscript: type tree of Base is"));
      printtree(stdout, 0, type_tree);
      DEBUG_MSG(("\n"));
      #endif
      /*}}}  */
      switch (TagOf(sub_tree))
        {
          /*{{{  some sort of record type*/
          case S_NODE: case S_CONTROLPORT:
          #ifdef OCCAM2_5
          case S_RECORD:
          #endif
            if (TagOf(ASIndexOf(tptr)) == S_NAME)
              lookup_special_namespace(ASIndexAddr(tptr), sub_tree);
            else
              scopeandcheck(ASIndexAddr(tptr));

            if (TagOf(ASIndexOf(tptr)) != N_FIELD)
              /*{{{  error*/
              {
                chkerr_s(CHK_RSUB_NOT_FIELD, LocnOf(tptr),
                         (nodetypeoftag(TagOf(type_tree)) == NAMENODE)
                         ? WNameOf(NNameOf(type_tree)) : tagstring(TagOf(type_tree)));
              }
              /*}}}  */

            SetTag(tptr, S_RECORDSUB);
            break;
          /*}}}  */
          default:
            scopeandcheck(ASIndexAddr(tptr));
            break;
        }
    }
  else
    scopeandcheck(ASIndexAddr(tptr));
#ifdef DEBUG
  DEBUG_MSG(("csubscript: tree after processing is"));
  printtree(stdout, 0, tptr);
  DEBUG_MSG(("\n"));
#endif
  return (tptr);
}
/*}}}  */
/*{{{  PRIVATE treenode *cpragma*/
PRIVATE treenode *cpragma(treenode *volatile tptr,
        treenode *last_decl, treenode *current_params)
/* 'volatile' is inserted just to prevent an erroneous gcc warning:
   warning: argument `tptr' may be clobbered by `longjmp'
*/
{
  jmp_buf savedenv;
  memcpy((char *)savedenv, (char *)env, sizeof(env));
  if (setjmp(env) != 0)
    tptr = NULL;
  else
    {
      switch((pragma_name_tag_t)NModeOf(DNameOf(tptr)))
        {
          /*{{{  NOUSAGECHECK/NOTALIASED*/
          case pragma_name_shared:
          case pragma_name_aliased:
          case pragma_name_assumeconst:
            {
              treenode *l;
              for (l = DValOf(tptr); !EndOfList(l); l = NextItem(l))
                {
                  treenode *const pname = ThisItem(l);
                  switch(TagOf(pname))
                    {
                      case N_DECL:
                      case N_PARAM:  case N_VALPARAM:
                      case N_ABBR:   case N_VALABBR:
                      case N_RETYPE: case N_VALRETYPE:
                        check_adjacent_name(LocnOf(tptr), pname, last_decl, current_params, SEV_ERR);
                        break;
                      default:
                        chkreport_ss(CHK_BAD_PRAGMA_SHARED_NAME, chklocn,
                                     WNameOf(NNameOf(DNameOf(tptr))), WNameOf(NNameOf(pname)));
                        break;
                    }
                  switch((pragma_name_tag_t)NModeOf(DNameOf(tptr)))
                    {
                      case pragma_name_shared:
                        SetNVShared(pname, TRUE);
                        break;
                      case pragma_name_aliased:
                        SetNVAliased(pname, TRUE);
                        SetNVShared(pname, TRUE); /* set shared for notaliased too! */
                        break;
                      case pragma_name_assumeconst:
                        SetNVAssumeConst(pname, TRUE);
                        break;
                      default:
                        break;
                    }
                }
            }
            break;
          /*}}}  */
          /*{{{  NESTEDTIMER/NESTEDPLACE/NESTEDPORT/BADLYBEHAVED*/
          case pragma_name_nestedtimer:
          case pragma_name_nestedplace:
          case pragma_name_nestedport:
          case pragma_name_badlybehaved:
            {
              treenode *const pname = DValOf(tptr);
              if (nametypeoftag(TagOf(pname)) != NAMENODE_ROUTINE)
                chkreport_ss(CHK_BAD_PRAGMA_SHARED_NAME, chklocn,
                             WNameOf(NNameOf(DNameOf(tptr))), WNameOf(NNameOf(pname)));
              check_adjacent_name(LocnOf(tptr), pname, last_decl, NULL, SEV_ERR);
              switch (NModeOf(DNameOf(tptr)))
                {
                  case pragma_name_nestedtimer: SetNPNestedTimer(pname, TRUE); break;
                  case pragma_name_nestedplace: SetNPNestedPlace(pname, TRUE); break;
                  case pragma_name_nestedport:  SetNPNestedPort (pname, TRUE); break;
                  case pragma_name_badlybehaved:
                    SetNPNestedTimer(pname, TRUE);
                    SetNPNestedPlace(pname, TRUE);
                    SetNPNestedPort (pname, TRUE);
                    break;
                  default:                                                     break;
                }
            }
            break;
          /*}}}  */
          /*{{{  everything else*/
          case pragma_name_linkage:
          case pragma_name_translate:
          case pragma_name_external:
          case pragma_name_comment:
          case pragma_name_hardware:
            break;
          /*}}}  */
        }
    }
  memcpy((char *)env, (char *)savedenv, sizeof(env));
  return tptr;
}
/*}}}  */
/*{{{  PRIVATE void scope_predefs*/
PRIVATE void scope_predefs(void)
{
  BOOL more_predefs = TRUE;
  DEBUG_MSG(("scope_predefs\n"));
  init_predefs_lexer();
  while (more_predefs)
    {
      treenode *const nptr = rpredef();
      if (nptr == NULL)
        more_predefs = FALSE;
      else
        {
          DEBUG_MSG(("scope_predefs: %s\n", WNameOf(NNameOf(nptr))));
        #if 0  /* use the normal list, rather than have individual entries */
          switch(NModeOf(nptr))
            {
              case PD_HOSTEDGE:
                fe_set_predefname(current_fe_handle, nptr, fe_predefname_HOST_EDGE);
                break;
              case PD_DEFAULTNODE:
                fe_set_predefname(current_fe_handle, nptr, fe_predefname_DEFAULT_NODE);
                break;
              default:
                break;
            }
        #endif

          if (TagOf(nptr) == N_FIELD)
            add_attribute(nptr); /* not globally scoped */
          else
            {
              addname(nptr, NULL); /* Declare the predefined name to the scoper */

              /* Add to the list of visible predefines */ /* bug TS/1625 16/03/92 */
              fe_set_predefname(current_fe_handle,
                    addtofront(nptr, fe_get_predefname(current_fe_handle, fe_predefname_LIST)),
                                fe_predefname_LIST);
            }
        }
    }
}
/*}}}  */
/*{{{  PRIVATE void scopeprocorfunc*/
PRIVATE treenode *current_proc;
PRIVATE treenode *current_params;
PRIVATE void scopeprocorfunc(treenode *const tptr)
{
  treenode *const name = DNameOf(tptr);
  treenode *params = NParamListOf(name);
  const scopestack_t namestackmarker = markscopenames();
  const BOOL oldinsidepripar = insidepripar;
  const BOOL oldnestedpripar = nestedpripar;
  const BOOL oldnestedtimer  = nestedtimer;
  const BOOL oldnestedplace  = nestedplace;
  const BOOL oldnestedport   = nestedport;
  /*int oldinsideinline = insideinline;*/
  const int old_lexlevel   = check_lexlevel;
  treenode *const old_proc = current_proc;
  insidepripar = FALSE;
  nestedpripar = FALSE;
  nestedtimer  = FALSE;
  nestedplace  = FALSE;
  nestedport   = FALSE;
  /*insideinline = inline(name);*/
  insideseparatelycompiledsc = separatelycompiled(name);
  current_proc = name;
  check_lexlevel++;

  /*{{{  scope and check result types*/
  #ifdef OCCAM2_5
  if ((NTypeOf(name) != NULL) &&
      (TagOf(NTypeOf(name)) == S_FNTYPE)) /* its a function */
    {
      treenode *fn_type;
      scopeandcheck(FnTypeListAddr(NTypeOf(name)));
      for (fn_type = FnTypeListOf(NTypeOf(name)); !EndOfList(fn_type);
           fn_type = NextItem(fn_type))
        ctype(ThisItem(fn_type));
    }
  #endif
  /*}}}  */

  /*{{{  scope parameters*/
  {
    treenode *p;
    for (p = params; !EndOfList(p); p = NextItem(p))
      scopeandcheck(NTypeAddr(ThisItem(p)));
  }
  /*}}}  */
  params = (treenode *)cparmlist(params, TagOf(tptr), LocnOf(tptr), insideseparatelycompiledsc);
  walklist(addname, params, NULL);            /* Add parameters  */
  SetNParamList(name, params);
  current_params = params;

  if (*current_fe_data->fe_guyinserts != 0)
    decllabels(DValOf(tptr));   /* Bring guy labels into scope */
  if (separatelycompiled(name))
    /*{{{  mark parameters as used, scopeandcheck the body*/
    {
      for (; !EndOfList(params); params = NextItem(params))
        SetNUsed(ThisItem(params), TRUE);
      scopeandchecksc(DValAddr(tptr));
    }
    /*}}}  */
  else
    scopeandcheck(DValAddr(tptr));               /* Check procedure/function */
  cdeclaration(tptr);
  /*{{{  Nested properties such as PRI PARs*/
  if (nestedpripar) SetNPNestedPriPar(name, TRUE);
  if (nestedtimer)  SetNPNestedTimer (name, TRUE);
  if (nestedplace)  SetNPNestedPlace (name, TRUE);
  if (nestedport)   SetNPNestedPort  (name, TRUE);
  #if 0
  if (nestedtimer)  printf("scopeprocorfunc: %s has a nestedtimer\n", WNameOf(NNameOf(name)));
  if (nestedplace)  printf("scopeprocorfunc: %s has a nestedplace\n", WNameOf(NNameOf(name)));
  if (nestedport)   printf("scopeprocorfunc: %s has a nestedport\n",  WNameOf(NNameOf(name)));
  #endif
  /*}}}  */
  current_proc   = old_proc;
  current_params = NULL;
  check_lexlevel = old_lexlevel;
  insidepripar   = oldinsidepripar;
  nestedpripar   = oldnestedpripar;
  /*insideinline = oldinsideinline;*/
  nestedtimer    = oldnestedtimer;
  nestedplace    = oldnestedplace;
  nestedport     = oldnestedport;
  insideseparatelycompiledsc = FALSE;
  descopenames(namestackmarker);                                 /* Descope names   */
}
/*}}}  */
/*{{{  PRIVATE void permit_hwname*/
PRIVATE void permit_hwname(treenode *const tptr,
        const int tag, const int fe_tag)
{
  if ((current_fe_data->fe_lang & FE_LANG_CONFIG3) != 0)
    {
      treenode *const bname = nameof(tptr); /* skip thru subscripts */
      wordnode *const wname = (wordnode *)bname;
      treenode *const tname = findname(wname);

      if ((tname == NULL) &&
          (findundeclaredname(wname) == NULL)) /* not in scope */
        {
          treenode *const nptr = addundeclaredname(wname);
          treenode *type = newleafnode(tag, NOPOSN);
          treenode *bptr;
          for (bptr = tptr; TagOf(bptr) == S_ARRAYSUB; bptr = ASBaseOf(bptr))
            type = newtypenode(S_ARRAY, NOPOSN, NULL, type);
          SetNType(nptr, type);
        #ifdef DEBUG
          DEBUG_MSG(("permit_hwname: creating %s, tptr is", WNameOf(wname)));
          printtree(stdout, 0, tptr);
          DEBUG_MSG(("\npermit_hwname: type is"));
          printtree(stdout, 0, type);
          DEBUG_MSG(("\n"));
        #endif

          fe_set_predefname(current_fe_handle,
                            addtofront(nptr, fe_get_predefname(current_fe_handle, fe_tag)),
                            fe_tag);
        }
    }
}
/*}}}  */
/*{{{  PRIVATE void scope_offsetof*/
#ifdef OCCAM2_5
PRIVATE void scope_offsetof(treenode *const tptr)
{
  treenode *type;
  scopeandcheck(LeftOpAddr(tptr));
  type = follow_user_type(LeftOpOf(tptr));
  if (TagOf(type) == S_RECORD)
    {
      lookup_special_namespace(RightOpAddr(tptr), type);
      if (TagOf(RightOpOf(tptr)) != N_FIELD)
        chkerr_ss(CHK_OFFSETOF_NOT_FIELD, LocnOf(tptr),
                  WNameOf(NNameOf(RightOpOf(tptr))),
                  WNameOf(NNameOf(LeftOpOf(tptr))));
    }
  else
    chkerr_s(CHK_OFFSETOF_NOT_RECORD, LocnOf(tptr),
             WNameOf(NNameOf(LeftOpOf(tptr))));
}
#endif
/*}}}  */
/*{{{  PRIVATE void scope_communication*/
#ifdef OCCAM2_5
PRIVATE void scope_communication(treenode *const tptr)
/* This routine scopes the tag values for tagged protocols.
   It leaves anything it doesn't recognise alone, so that
   'scopeandcheck' can be run afterwards.
*/
{
  treenode *const channel = LHSOf(tptr);
  treenode *const type_tree = typecheck(channel, unknownnodeptr);
  if ((TagOf(type_tree) == S_CHAN) &&
      (TagOf(ProtocolOf(type_tree)) == N_TPROTDEF))
    {
      treenode *const rhs = RHSOf(tptr);
      switch(TagOf(tptr))
        {
          /*{{{  output/ tagged input*/
          case S_OUTPUT: case S_TAGGED_INPUT:
            if (!EndOfList(rhs) && (TagOf(ThisItem(rhs)) == S_NAME))
              lookup_special_namespace(ThisItemAddr(rhs), ProtocolOf(type_tree));
            break;
          /*}}}  */
          /*{{{  case input*/
          case S_CASE_INPUT:
            {
              treenode *variantlist;
              for (variantlist = rhs; !EndOfList(variantlist); variantlist = NextItem(variantlist))
                {
                  treenode *const variant = skipspecifications(ThisItem(variantlist));
                  if (variant != NULL)
                    {
                      treenode *const this_variant = VRTaggedListOf(variant);
                      if (!EndOfList(this_variant) && (TagOf(ThisItem(this_variant)) == S_NAME))
                        lookup_special_namespace(ThisItemAddr(this_variant), ProtocolOf(type_tree));
                    }
                }
            }
            break;
          /*}}}  */
        }
    }
}
#endif
/*}}}  */

/*{{{  PRIVATE void check_nested_properties_place*/
PRIVATE void check_nested_properties_place(treenode *tptr)
/* bug TS/1983 17/12/92 */
{
  treenode *type = NTypeOf(DNameOf(tptr));
  while (TagOf(type) == S_ARRAY)
    type = ARTypeOf(type);

  if ((TagOf(type) == S_PORT) || (TagOf(type) == S_CHAN))
    nestedport = TRUE;

  nestedplace = TRUE;
}
/*}}}  */
/*{{{  PRIVATE void check_nested_properties_decl*/
PRIVATE void check_nested_properties_decl(treenode *tptr)
/* bug TS/1983 17/12/92 */
{
  treenode *name = DNameOf(tptr);
  if (TagOf(name) == S_LIST) name = ThisItem(name);
  if (basetype(NTypeOf(name)) == S_TIMER)
    nestedtimer = TRUE;
}
/*}}}  */

/*{{{  PRIVATE void scopeandcheck*/
/*{{{  comment*/
/*****************************************************************************
 *
 *  scopeandcheck applies scoping and type checking to parse tree
 *
 *****************************************************************************/
/*}}}  */
PUBLIC void scopeandcheck ( treenode **tptr )
{
  static BOOL guy_not_asm = FALSE;  /* YES - static but local to this function! */
  if (tptr != NULL)
    while (*tptr != NULL)
      {
        treenode *const t = *tptr;
        const int tag = TagOf(t);
        const int nodetype = nodetypeoftag(tag);
        if ((nodetype != WORDNODE) && (nodetype != NAMENODE))
          chklocn = LocnOf(t);
        DEBUG_MSG(("scopeandcheck: %s at %d\n", itagstring(tag), (int)chklocn));
        switch (nodetype)
          {
            default:
              return;
            /*{{{  specification*/
            case DECLNODE:
              {
                const scopestack_t namestackmarker = markscopenames();
                treenode **root = tptr;
                treenode *tt = t;
                treenode *last_decl = NULL;
                for (; (*tptr != NULL) && isspecification(*tptr); tptr = DBodyAddr(tt))
                  {
                    treenode *name;
                    tt = *tptr;
                    name = DNameOf(tt);
                    chklocn = LocnOf(tt);
                    switch (TagOf(tt))
                    {
                      /*{{{  ABBR, VALABBR, RETYPE, VALRETYPE, TPROTDEF, SPROTDEF*/
                      case S_ABBR: case S_VALABBR: case S_RETYPE: case S_VALRETYPE:
                      case S_SPROTDEF: CASE_CONFIG_SPEC
                      #ifdef OCCAM2_5
                      case S_TYPEDECL:
                      #endif
                        {
                          scopeandcheck(NTypeAddr(name));
                          scopeandcheck(DValAddr(tt));
                          if ((treenode *)cdeclaration(tt) != (treenode *)NULL)
                            {
                              addname(name, NULL);
                              last_decl = tt; /* bug TS/1910 26/01/93 */
                            }
                        }
                        break;
                      case S_TPROTDEF:
                        {
                          treenode *n;
                          for (n = NTypeOf(name); !EndOfList(n); n = NextItem(n))
                            {
                            #ifdef OCCAM2_5
                              if (current_fe_data->fe_visible_tags)
                                addname(ThisItem(n), NULL);
                            #else
                              addname(ThisItem(n), NULL);
                            #endif
                              scopeandcheck(NTypeAddr(ThisItem(n)));
                            }
                          if ((treenode *)cdeclaration(tt) != (treenode *)NULL)      /* Scope the protocol name */
                            {
                              addname(name, NULL);
                              last_decl = tt; /* bug TS/1910 26/01/93 */
                            }
                        }
                        break;
                      /*}}}  */
                      /*{{{  PROCDEF FUNCDEF*/
                      case S_PROCDEF: case S_SFUNCDEF: case S_LFUNCDEF:
                        scopeprocorfunc(tt);
                        addname(DNameOf(tt), NULL);
                        last_decl = tt; /* bug TS/1910 26/01/93 */
                        break;
                      /*}}}  */
                      /*{{{  DECL*/
                      case S_DECL:
                        {
                          #ifdef DECL_EQ
                          if (DValOf(tt) == NULL)
                            synerror(TRUE, CHK_MISSING_INIT, LocnOf(tt), 0);
                          else
                          #endif
                          if (TagOf(name) == S_LIST)
                            /* They all share the same type tree so do this once */
                            {
                              treenode **typeaddr = NTypeAddr(ThisItem(name));
                              scopeandcheck(typeaddr);
                              /* if the type is a typedecl, this hasn't scoped any trailing
                                 declarations; so we do it directly:
                              */
                              for (; !EndOfList(name); name = NextItem(name))
                                SetNType(ThisItem(name), *typeaddr);
                            }
                          else
                            scopeandcheck(NTypeAddr(name));

                          if ((treenode *)cdeclaration(tt) != (treenode *)NULL) /* Check type and name */
                            {
                              name = DNameOf(tt);
                              if (TagOf(name) == S_LIST)
                                walklist(addname, name, NULL);
                              else
                                addname(name, NULL);
                              check_nested_properties_decl(tt);
                              last_decl = tt; /* bug TS/1910 26/01/93 */
                            }
                        }
                        break;
                      /*}}}  */
                      /*{{{  PLACE WSPLACE VSPLACE*/
                      case S_PLACE: case S_WSPLACE: case S_VSPLACE:
                      case S_PLACEON:
                        if (TagOf(tt) == S_PLACEON)
                          permit_hwname(PlaceExpOf(tt), S_ARC, fe_predefname_HARDWARE_ARC);
                        scopeandcheck(DNameAddr(tt));
                        scopeandcheck(PlaceExpAddr(tt));
                        *tptr = (treenode *)callocation(tt, last_decl);
                        if (*tptr != NULL && TagOf(tt) == S_PLACE) /* PLACE ... AT */
                          check_nested_properties_place(*tptr);
                        break;
                      /*}}}  */
                      /*{{{  COMMENT RECURSIVE*/
                      /***********************  Start comment out *****************************/
                      /*|*{{{  RECURSIVE*|*/
                      /*case S_RECURSIVE :*/
                        /*{*/
                          /*treenode *item = DValOf(tt);*/
                          /*while (!EndOfList(item))*/
                            /*{*/
                              /*treenode *name = DNameOf(ThisItem(item));*/
                              /*treenode *rest = NextItem(item);*/
                              /*wordnode *id = NNameOf(name);*/
                              /*while (!EndOfList(rest))*/
                              /*{*/
                                /*if (id == NNameOf(DNameOf(ThisItem(rest))))*/
                                  /*synerror (SYN_MULTIPLE_RECURSIVE, flocn, (BIT32)WNameOf(id));*/
                                /*rest = NextItem(rest);*/
                              /*}*/
                              /*addname(name, NULL);*/
                              /*item = NextItem(item);*/
                            /*}*/
                          /*item = DValOf(tt);*/
                          /*while (!EndOfList(item))*/
                            /*{*/
                              /*scopeprocorfunc(ThisItem(item));*/
                              /*item = NextItem(item);*/
                            /*}*/
                        /*}*/
                        /*break;*/
                      /*|*}}}*|*/
                       /***********************   End comment out  *****************************/
                      /*}}}  */
                      /*{{{  labeldef*/
                      case S_LABELDEF:
                        /*if (insideinline)
                          chkrecoverreport(CHK_LABEL_INSIDE_INLINE, chklocn, 0);*/
                        break;
                      /*}}}  */
                      /*{{{  PRAGMA*/
                      case S_PRAGMA: /* bug 829 19/9/91 */
                        switch((pragma_name_tag_t)NModeOf(name))
                          {
                            case pragma_name_shared:
                            case pragma_name_aliased:
                            case pragma_name_nestedtimer:
                            case pragma_name_nestedplace:
                            case pragma_name_nestedport:
                            case pragma_name_badlybehaved:
                            case pragma_name_assumeconst:
                              scopeandcheck(DValAddr(tt));
                              break;
                            case pragma_name_linkage:
                            case pragma_name_translate:
                            case pragma_name_external:
                            case pragma_name_comment:
                            case pragma_name_hardware:
                              break;
                          }
                        cpragma(tt, last_decl, current_params);
                        break;
                      /*}}}  */
                      default:
                        badtag(chklocn, TagOf(tt), "scopeandcheck");
                    }
                  }
                current_params = NULL; /* bug TS/1910 26/01/93 */
                if ((*tptr != NULL) && (TagOf(*tptr) == S_VALOF))
                  /*{{{  do checking*/
                  {
                    /* Scope and check body and result list here so we can include the
                     * preceeding specs in check of valof
                     */
                    scopestack_t valofnamestackmarker = markscopenames();
                    treenode *const ttt = *tptr;
                    scopeandcheck(VLBodyAddr(ttt));
                    descopenames(valofnamestackmarker);            /* Descope names within Valof */
                    scopeandcheck(VLResultListAddr(ttt));
                    descopenames(namestackmarker);                 /* Descope names before Valof */
                    *root = (treenode *)cvalof(*root, current_scope());
                  }
                  /*}}}  */
                else
                  {
                    scopeandcheck(tptr);
                    descopenames(namestackmarker);
                  }
                return;
              }
            /*}}}  */
            /*{{{  process*/
            /*{{{  Constructor*/
            case CNODE:
              current_params = NULL; /* bug TS/1910 26/01/93 */
              if (tag == S_PRIPAR)
                /*{{{  PRI PAR - return*/
                {
                  const BOOL savedinsidepripar = insidepripar;
                  if (insidepripar)
                    chkerr(CHK_NESTED_PRI_PAR, chklocn);
                  insidepripar = TRUE;
                  nestedpripar = TRUE;
                  scopeandcheck(CBodyAddr(t));
                  if (listitems(CBodyOf(t)) != 2)
                    chkerr(CHK_INV_PRI_PAR, LocnOf(t));
                  insidepripar = savedinsidepripar;
                  return;
                }
                /*}}}  */
              else if (((current_fe_data->fe_lang & FE_LANG_OCCAM) != 0) &&
                       (tag == S_PLACEDPAR)) /* bug TS/1623 09/04/92 */
                chkerr_s(CHK_NOT_IMPLEMENTED, chklocn, tagstring(S_PLACEDPAR));
              guy_not_asm = (tag == S_GUY);
              if (guy_not_asm && current_fe_data->fe_warning_guy)
                {
                  /* we pass 'NOPOSN' to the memo function so that it doesn't just
                     restrict the 'memo' to this one line of source code */
                  if (fe_memo_err(current_fe_handle, CHK, CHK_WARN_GUY, NOPOSN))
                    msg_out(SEV_WARN, CHK, CHK_WARN_GUY, chklocn);
                }
              tptr = CBodyAddr(t);
              break;
            /*}}}  */
            /*{{{  REPLSEQ REPLPAR REPLIF REPLALT PRIREPLALT   return*/
            case REPLCNODE:
              {
                /*int savedinsidepripar = insidepripar;*/
                const BOOL replpar = parrepl(tag);
                const scopestack_t namestackmarker = markscopenames();
                current_params = NULL; /* bug TS/1910 26/01/93 */
                scopeandcheck(ReplCStartExpAddr(t));
                scopeandcheck(ReplCLengthExpAddr(t));
                if ((treenode *)crepl(t) != (treenode *)NULL)
                  {
                    const int old_lexlevel = check_lexlevel;
                    /*{{{  Check repl is constant (Compiler and configurer only)*/
                    if (((current_fe_data->fe_lang & FE_LANG_NDL) == 0) &&
                        replpar && !isconst(ReplCLengthExpOf(t)))
                      chkerr(CHK_REPL_NOT_CONST, chklocn);
                    /*}}}  */
                    /*{{{  REPLPRIPAR / PLACEDREPLPAR is not implemented*/
                    if ((tag == S_PRIREPLPAR) ||
                        (((current_fe_data->fe_lang & FE_LANG_OCCAM) != 0) &&
                         (tag == S_PLACEDREPLPAR))) /* bug TS/1623 09/04/92 */
                      {
                        chkerr_s(CHK_NOT_IMPLEMENTED, chklocn, tagstring(tag));
                        #if 0  /* no point in leaving this here! */
                        if (insidepripar)
                          chkerr(CHK_NESTED_PRI_PAR, chklocn);
                        insidepripar = TRUE;
                        nestedpripar = TRUE;
                        #endif
                      }
                    /*}}}  */
                    check_lexlevel += replpar;
                    addname(ReplCNameOf(t), NULL);
                    scopeandcheck(ReplCBodyAddr(t));
                    check_lexlevel = old_lexlevel;
                    #if 0
                    if (tag == S_PRIREPLPAR)
                      insidepripar = savedinsidepripar;
                    #endif
                  }
                descopenames(namestackmarker);
              }
              return;
            /*}}}  */
            /*{{{  WHILE CHOICE                                return*/
            case CONDNODE:
              scopeandcheck(CondGuardAddr(t));
              scopeandcheck(CondBodyAddr(t));
              if (tag != S_SELECTION && tag != S_VARIANT)
                *tptr = (treenode *)ccond(t);
              return;
            /*}}}  */
            /*{{{  ALTERNATIVE                                 return*/
            case ALTNODE:
              scopeandcheck(AltGuardAddr(t));
              scopeandcheck(AltInputAddr(t));
              scopeandcheck(AltBodyAddr(t));
              *tptr = (treenode *)calt(t);
              return;
            /*}}}  */
            /*{{{  PINSTANCE FINSTANCE                         return*/
            case INSTANCENODE:
              {
                treenode **params = IParamListAddr(t);
                treenode *name;
                scopeandcheck(INameAddr(t));
                if (*params != NULL)
                  scopeandcheck(params);

                name = INameOf(t);
                /*{{{  if it is actually a type conversion => RETURN*/
                if ((TagOf(name) == N_TYPEDECL) &&
                    (listitems(IParamListOf(t)) == 1))
                  {
                    const SOURCEPOSN locn = LocnOf(t);
                    treenode *const explist = IParamListOf(t);
                    freenode(tptr);
                    *tptr = newmopnode(S_EXACT, locn,
                            newlistnode(S_LIST, locn, name, explist), S_NAME);
                    return;
                  }
                /*}}}  */

                if (insidepripar && TagOf(name) != N_DECL && NPNestedPriParOf(name))
                  /* N.B. tag is N_DECL if the lookupname failed */
                  chkerr_s(CHK_NESTED_PRI_PROC, chklocn, WNameOf(NNameOf(name)));

                /*{{{  check for 'nested' attributes*/
                /* bug TS/1982 17/12/92 */
                if (NPNestedPriParOf(name)) nestedpripar = TRUE;

                /* bug TS/1983 17/12/92 */
                if (NPNestedTimerOf(name))  nestedtimer  = TRUE;
                if (NPNestedPlaceOf(name))  nestedplace  = TRUE;
                if (NPNestedPortOf(name))   nestedport   = TRUE;

                #if 0
                printf("instance: \"%s\", nestedtimer: %d, nestedplace: %d, nestedport: %d\n",
                       WNameOf(NNameOf(name)),
                       NPNestedTimerOf(name), NPNestedPlaceOf(name), NPNestedPortOf(name));
                #endif
                /*}}}  */

                if (((current_fe_data->fe_lang & FE_LANG_OCCAM) != 0) &&
                    separatelycompiled(name))
                  checklibproctype(name, LocnOf(t), FALSE);

                if (tag == S_PINSTANCE)
                  *tptr = (treenode *)cinstance(t);
                return ;
              }
            /*}}}  */
            /*{{{  CASE_INPUT, DELAYED_INPUT, OUTPUT, INPUT, TAGGED_INPUT, ASS, CASE  return*/
            case ACTIONNODE:
              scopeandcheck(LHSAddr(t));
              /*{{{  check for special scope rules*/
              #ifdef OCCAM2_5
              switch(TagOf(t))
                {
                  case S_OUTPUT: case S_TAGGED_INPUT: case S_CASE_INPUT:
                    scope_communication(t);
                    break;
                  default:
                    break;
                }
              #endif
              /*}}}  */
              scopeandcheck(RHSAddr(t));
              *tptr = (tag == S_CASE) ? (treenode *)ccase(t) : (treenode *)caction(t);
              return;
            /*}}}  */
            /*}}}  */
            /*{{{  expression*/
            /*{{{  monadics            return*/
            case MOPNODE:
              tptr = OpAddr(t);
              break;
            /*}}}  */
            /*{{{  literals*/
            case LITNODE:
              if (tag == S_CONSTRUCTOR)
                scopeandcheck(LitExpAddr(t));
              tptr = LitTypeAddr(t);
              break;
            /*}}}  */
            /*{{{  consttablenode*/
            #ifdef OCCAM2_5
                        case CONSTTABLENODE:
                          if (tag == S_STRING && ConstTableTypeOf(t) != NULL)
                            tptr = ConstTableTypeAddr(t);
                          else
                            return;
                          break;
            #endif
            /*}}}  */
            /*}}}  */
            /*{{{  dyadics             break*/
            case DOPNODE:
              switch(tag)
                {
                  case S_GUYCODE: case S_GUYSTEP:
                    scopeandcheck(RightOpAddr(t));
                    /* guy_not_asm was set up by enclosing S_GUY or S_ASM */
                    *tptr = (treenode *)cguy_or_asm(t, guy_not_asm, current_proc);
                    return;
                #ifdef OCCAM2_5
                  case S_OFFSETOF:
                    scope_offsetof(t);
                    return;
                #endif
                  default:
                    scopeandcheck(LeftOpAddr(t));
                    tptr = RightOpAddr(t);
                    break;
                }
              break;
            /*}}}  */
            /*{{{  VALOF                       return*/
            case VALOFNODE:
              {
                const scopestack_t namestackmarker = markscopenames();
                scopeandcheck(VLBodyAddr(t));
                descopenames(namestackmarker);      /* Descope names within Valof */
                scopeandcheck(VLResultListAddr(t));
                *tptr = (treenode *)cvalof(t, current_scope());
              }
              return;
            /*}}}  */
            /*{{{  arrayconstructor*/
            #ifdef ARRAYCONSTRUCTOR
            case S_ARRAYCONSTRUCTOR :  unknown
              {
                scopestack_t namestackmarker = markscopenames();
                scopeandcheck(ACStartExpAddr(t));
                scopeandcheck(ACLengthExpAddr(t));
                addname(ACNameOf(t), NULL);
                scopeandcheck(ACValExpAddr(t));
                descopenames(namestackmarker);
              }
              return;
            #endif
            /*}}}  */
            /*{{{  conditional expression*/
            #ifdef CONDEXP
            case CONDEXPNODE:
              {
                scopeandcheck(CondExpGuardAddr(t));
                scopeandcheck(CondExpTrueAddr(t));
                tptr = CondExpFalseAddr(t);
              }
              break;
            #endif

            /*}}}  */

            /*{{{  element*/
            case ARRAYSUBNODE:
              scopeandcheck(ASBaseAddr(t));
              *tptr = csubscript(t);  /* convert to S_RECORDSUB if necessary */
              return;
            case SEGMENTNODE:
              scopeandcheck(SNameAddr(t));
              scopeandcheck(SStartExpAddr(t));
              scopeandcheck(SLengthExpAddr(t));
              tptr = SSubscriptExpAddr(t);
              break;
            /*}}}  */
            /*{{{  type*/
            case TYPENODE :
              switch (tag)
                {
                  /*{{{  record*/
                  case S_RECORD:
                    {
                      treenode *field;
                      for (field = ARTypeOf(t); field != NULL; field = DBodyOf(field))
                        {
                          treenode *this_field = DNameOf(field);
                          if (TagOf(this_field) == S_LIST) this_field = ThisItem(this_field);
                          scopeandcheck(NTypeAddr(this_field));
                        }
                    }
                    return;
                  /*}}}  */
                  /*{{{  array*/
                  case S_ARRAY:
                    scopeandcheck(ARDimLengthAddr(t));
                    tptr = ARTypeAddr(t);
                    break;
                  /*}}}  */
                  /*{{{  channel / port*/
                  case S_CHAN: case S_PORT:
                    tptr = ProtocolAddr(t);
                    break;
                  /*}}}  */
                  default:
                    badtag(chklocn, tag, "scopeandcheck");
                }
              break;
            /*}}}  */
            /*{{{  PROCESSOR*/
            case PROCESSORNODE:
              {
                const int old_lexlevel = check_lexlevel;
                permit_hwname(ProcessorExpOf(t), S_NODE, fe_predefname_HARDWARE_NODE);
                scopeandcheck(ProcessorExpAddr(t));
                check_lexlevel++;
                scopeandcheck(ProcessorBodyAddr(t));
                check_lexlevel = old_lexlevel;
                SetProcessorScope(t, current_scope());
                *tptr = (treenode *)cprocessor(t);
              }
              return;
            /*}}}  */
            /*{{{  NAME*/
            case WORDNODE :
              if (tag != S_ASMNAME)
                {
                  wordnode *const n = (wordnode *)t;
                  treenode *const nptr = lookupname(n, findname(n));
                  SetNUsed(nptr, TRUE);
                  *tptr = nptr;

                  /*printf("scopeandcheck: name: %s, lexlevel %d, current level %d\n",
                         WNameOf(NNameOf(*tptr)), NLexLevelOf(*tptr), check_lexlevel);*/

                  if ((TagOf(nptr) == N_LABELDEF) && (NLexLevelOf(nptr) != check_lexlevel))
                    chkerr(CHK_NONLOCAL_LABEL, chklocn); /* bug TS/1524 13/12/91 */
                }
              return;
            /*}}}  */
            /*{{{  list*/
            case LISTNODE :
              {
                treenode *tt;
                for (tt = t; !EndOfList(tt); tt = NextItem(tt))
                  scopeandcheck(ThisItemAddr(tt));
              }
              return;
            /*}}}  */
            /*{{{  structconstructor*/
            #if 0
            case S_STRUCTCONSTRUCTOR : unknown
              {
                if (MOpTypeOf(t) == S_NODE)
                  {
                    t = OpOf(t);              /* This will be a list */
                    if (!EndOfList(t))        /* Skip device type    */
                      t = NextItem(t);
                    while (!EndOfList(t))
                      {
                        scopeandcheck(ThisItemAddr(t));
                        t = NextItem(t);
                      }
                    *tptr = cnetnode(*tptr);
                  }
              }
              return;
            #endif
            /*}}}  */
            /*{{{  config*/
            case CONFIGNODE:
              if (tag == S_SET)
                /*{{{  allow for hardware names*/
                permit_hwname(STDevOf(t), S_NODE, fe_predefname_HARDWARE_NODE);
                /*}}}  */
              scopeandcheck(STDevAddr(t)); /* also MapSourceAddr */
              if (tag == S_MAP)
                /*{{{  allow for hardware names*/
                {
                  treenode *const name = nameof(ThisItem(MapSourceOf(t)));
                  if (basetype(NTypeOf(name)) == S_CHAN)
                    permit_hwname(MapDestOf(t), S_ARC, fe_predefname_HARDWARE_ARC);
                  else
                    permit_hwname(MapDestOf(t), S_NODE, fe_predefname_HARDWARE_NODE);
                }
                /*}}}  */
              if (tag != S_SET)
                scopeandcheck(STAttrNameAddr(t));
              scopeandcheck(STAttrExpAddr(t));
              switch(tag)
                {
                  case S_CONNECT: *tptr = cconnect(t); break;
                  case S_MAP:     *tptr = cmap(t);     break;
                  case S_SET:     *tptr = cset(t);     break;
                }
              return;
            /*}}}  */
          }
      }
}
/*}}}  */


/*{{{  configcheck*/
#if 1 /*def CONFIG*/
/* Checking states:
 S_END  outermost VALs
 S_NETWORK   inside a network description
 S_MAPPING   inside a MAPPING construct
 S_CONFIG    inside a CONFIG construct
 S_PROCESSOR inside a PROCESSOR construct
*/
/*{{{  PRIVATE void invalidconfigconstruct()*/
PRIVATE void invalidconfigconstruct(treenode *tptr)
{
  if (configcheckstate == S_END)
    chkerr(CHK_ILLEGAL_CONFIG_CONSTRUCT, LocnOf(tptr));
  else
    chkerr_s(CHK_ILLEGAL_CONSTRUCT, LocnOf(tptr), tagstring(configcheckstate));
}
/*}}}  */
/*{{{  PRIVATE void changestate()*/
PRIVATE void changestate(treenode *tptr, treenode *new_tptr, int req_state, int new_state)
{
  if (configcheckstate == req_state)
    {
      DEBUG_MSG(("changestate: from %s to %s\n", tagstring(req_state), tagstring(new_state)));
      configcheckstate = new_state;
      prewalkproctree(new_tptr, configcheck, NULL);
      configcheckstate = req_state;
      DEBUG_MSG(("changestate: back %s to %s\n", tagstring(new_state), tagstring(req_state)));
    }
  else
    invalidconfigconstruct(tptr);
}
/*}}}  */
/*{{{  PRIVATEPARAM int configcheck()*/
PRIVATEPARAM int configcheck(treenode *tptr, void *voidptr)
{
  /* rscunit has ensured that if we're genuinely at the top level, all
     we will see are S_VALABBR, S_VALRETYPE, S_TPROTDEF, S_SPROTDEF,
     S_PROCDEF, S_LFUNCDEF, S_SFUNCDEF, S_NETWORK, S_CONFIG, S_MAPPING
  */
  DEBUG_MSG(("configcheck: %s\n", itagstring(TagOf(tptr))));
  /*{{{  Some node types are always valid*/
  switch (nodetypeoftag(TagOf(tptr)))
    {
      case CONSTEXPNODE: case CONSTTABLENODE: case LITNODE:
      case MOPNODE: case DOPNODE:
      case SEGMENTNODE:
      case NAMENODE:
        return CONTINUE_WALK;
      default:
        break;
    }
  /*}}}  */
  switch (TagOf(tptr))
    {
      /*{{{  default: PROCESSOR only*/
      default:
        if (configcheckstate != S_PROCESSOR)
          invalidconfigconstruct(tptr);
        break;
      /*}}}  */
      /*{{{  S_CONFIG*/
      case S_CONFIG:
        configcount++;
        if (configcount > 1)
          chkerr_s(CHK_DUPLICATE_CONSTRUCT, LocnOf(tptr), tagstring(S_CONFIG));
        changestate(tptr, DValOf(tptr), S_END, S_CONFIG);
        break; /* continue after the CONFIG */
      /*}}}  */
      /*{{{  S_NETWORK*/
      case S_NETWORK:
        networkcount++;
        if (networkcount > 1)
          chkerr_s(CHK_DUPLICATE_CONSTRUCT, LocnOf(tptr), tagstring(S_NETWORK));
        changestate(tptr, DValOf(tptr), S_END, S_NETWORK);
        break; /* continue after the NETWORK */
      /*}}}  */
      /*{{{  S_MAPPING*/
      case S_MAPPING:
        mappingcount++;
        if (mappingcount > 1)
          chkerr_s(CHK_DUPLICATE_CONSTRUCT, LocnOf(tptr), tagstring(S_NETWORK));
        changestate(tptr, DValOf(tptr), S_END, S_MAPPING);
        break; /* continue after the MAPPING */
      /*}}}  */
      /*{{{  S_PROCESSOR*/
      case S_PROCESSOR:
        configcheckscope = ProcessorScopeOf(tptr);
        changestate(tptr, ProcessorBodyOf(tptr), S_CONFIG, S_PROCESSOR);
        return STOP_WALK;
      /*}}}  */
      /*{{{  NETWORK or MAPPING only*/
      case S_DO: case S_REPLDO: case S_SET:
      case S_MAP:
        if (configcheckstate != S_NETWORK && configcheckstate != S_MAPPING)
          invalidconfigconstruct(tptr);
        break;
      /*}}}  */
      /*{{{  CONFIG or PROCESSOR only*/
      case S_PAR: case S_REPLPAR:
        if (configcheckstate != S_CONFIG && configcheckstate != S_PROCESSOR)
          invalidconfigconstruct(tptr);
        break;
      /*}}}  */
      /*{{{  toplevel or CONFIG only*/
      case S_PLACEON:
      #ifdef OCCAM2_5
      case S_TYPEDECL:
      #endif
        if (configcheckstate != S_CONFIG && configcheckstate != S_END)
          invalidconfigconstruct(tptr);
        break;
      /*}}}  */
      /*{{{  nowhere except config*/
      case S_PLACEDPAR: case S_PLACEDREPLPAR:
        if (configcheckstate != S_CONFIG)
          invalidconfigconstruct(tptr);
        break;
      /*}}}  */
      /*{{{  nowhere except network*/
      case S_CONNECT: case S_RECORDSUB:
        if (configcheckstate != S_NETWORK)
          invalidconfigconstruct(tptr);
        break;
      /*}}}  */
      /*{{{  PINSTANCE*/
      case S_PINSTANCE: /* bug TS/1465 12/11/91 */
        if (((current_fe_data->fe_lang & FE_LANG_NDL) == 0)
          &&(configcheckstate != S_PROCESSOR))
          invalidconfigconstruct(tptr);
        break;
      /*}}}  */
      /*{{{  PROC/FUNCTION*/
      case S_PROCDEF: case S_LFUNCDEF: case S_SFUNCDEF:
        DEBUG_MSG(("configcheck: PROC/FUNCTION %s (%s)\n",
                   WNameOf(NNameOf(DNameOf(tptr))), itagstring(TagOf(DNameOf(tptr)))));
        if ((current_fe_data->fe_lang & FE_LANG_NDL) != 0) /* bug TS/1465 12/11/91 */
          {
            changestate(tptr, DValOf(tptr), configcheckstate, S_NETWORK);
          }
        else if (separatelycompiled(DNameOf(tptr)))
          {
            if (configcheckstate == S_NETWORK || configcheckstate == S_MAPPING)
              invalidconfigconstruct(tptr);
          }
        else /* locally defined */
          {
            if (configcheckstate != S_PROCESSOR)
              invalidconfigconstruct(tptr);
            prewalkproctree(DValOf(tptr), configcheck, voidptr);
          }
        break;
      /*}}}  */
      /*{{{  declarations*/
      case S_DECL: case S_ABBR: /* case S_RETYPE: */
        {
          treenode *name = DNameOf(tptr);
          int type;
          if (TagOf(name) == S_LIST) name = ThisItem(name);
          type = basetype(NTypeOf(name));
          if (((configcheckstate == S_END || configcheckstate == S_NETWORK) &&
               network_datatype(type)                                     ) ||
              ((configcheckstate == S_CONFIG || configcheckstate == S_END ) &&
               (type == S_CHAN)                                           ) ||
              ((configcheckstate == S_PROCESSOR) && !network_datatype(type))    )
            ;
          else
            invalidconfigconstruct(tptr);
        }
        break;
      /*}}}  */
      /*{{{  PLACEMENTs*/
      case S_PLACE: case S_WSPLACE: case S_VSPLACE:
        if (configcheckstate != S_PROCESSOR)
          invalidconfigconstruct(tptr);
        else if (NScopeOf(DNameOf(tptr)) < configcheckscope)
          chkerr_s(CHK_ILLEGAL_PLACE, LocnOf(tptr), WNameOf(NNameOf(DNameOf(tptr))));
        break;
      /*}}}  */
      /*{{{  Always valid*/
      case S_VALABBR: case S_VALRETYPE:
      case S_TPROTDEF: case S_SPROTDEF:
      case S_IF: /*case S_REPLIF:*/ case S_CHOICE:
      case S_SKIP: case S_STOP: case S_LIST: case S_END:
      case S_ARRAYSUB:
      case S_PRAGMA: /* bug 829 20/9/91 */
        break;
      /*}}}  */
    }
  return CONTINUE_WALK;
}
/*}}}  */
#endif
/*}}}  */

/*{{{  PUBLIC void scopeandcheck_main()*/
PUBLIC void scopeandcheck_main(treenode **const tptr, const BOOL allowpredefs, const BOOL toplevel)
{
  jmp_buf savedenv;
  memcpy((char *)savedenv, (char *)env, sizeof(env));

  if (setjmp(env) == 0)
    {
      if (allowpredefs)
        scope_predefs();
      check_lexlevel = 0;
      scopeandcheck(tptr);
      /*{{{  check for high level semantics*/
      if (((current_fe_data->fe_lang & (FE_LANG_NDL | FE_LANG_CONFIG2 | FE_LANG_CONFIG3)) != 0) &&
          (fe_get_errorcount(current_fe_handle) == 0) && toplevel)
        {
          configcheckstate = S_END; /*(compilemode == COMP_PROGRAM ? S_END : S_PROCESSOR);*/
          configcount = 0; networkcount = 0; mappingcount = 0;
          prewalkproctree(*tptr, configcheck, NULL);
          if (configcount == 0 &&
              ((current_fe_data->fe_lang & FE_LANG_NDL) == 0)) /* bug TS/1465 12/11/91 */
            chkerr_s(CHK_ZERO_CONSTRUCT, NOPOSN, tagstring(S_CONFIG));
          if (networkcount == 0 &&
              ((current_fe_data->fe_lang & FE_LANG_CONFIG3) == 0))
            chkerr_s(CHK_ZERO_CONSTRUCT, NOPOSN, tagstring(S_NETWORK));
        }
      /*}}}  */
    }
  memcpy((char *)env, (char *)savedenv, sizeof(env));

  /* This added to prevent later 'out of memory' errors reporting
     a silly line number - CON 22/1/91 */
  flocn   = NOPOSN;
  chklocn = NOPOSN;
}
/*}}}  */

/*{{{  PUBLIC void scopeinit ()*/
PUBLIC void scopeinit ( void )
{
  /*{{{  initialise real namestack*/
  init_namestack(&namestack_main, DEFAULT_NAMESTACK_SIZE_MAIN);
  /*}}}  */
  /*{{{  initialise undeclared namestack*/
  init_namestack(&namestack_undeclared, DEFAULT_NAMESTACK_SIZE_UNDEC);
  /*}}}  */
  insidepripar = FALSE;
  /*insideinline = FALSE;*/
#if 1 /*def CONFIG*/
  init_attributes();
#endif
}
/*}}}  */
