/* $Id: err2.c,v 1.4 1997/04/30 11:14:55 jm40 Exp $ */

/*
 *	Error reporting strings
 *	Copyright (C) 1992 Inmos Limited
 *
 *	This program is free software; you can redistribute it and/or modify
 *	it under the terms of the GNU General Public License as published by
 *	the Free Software Foundation; either version 2 of the License, or
 *	(at your option) any later version.
 *
 *	This program is distributed in the hope that it will be useful,
 *	but WITHOUT ANY WARRANTY; without even the implied warranty of
 *	MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *	GNU General Public License for more details.
 *
 *	You should have received a copy of the GNU General Public License
 *	along with this program; if not, write to the Free Software
 *	Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 */

/*#define DEBUG*/

/*{{{  include files*/
#ifdef HAVE_CONFIG_H
#  include <config.h>
#endif
#include <stdio.h>
#include "midinc.h"

#include "err2.h"

#include "lexerror.h"
#include "synerror.h"
#include "chkerror.h"
#include "useerror.h"
#include "vtierror.h"

/*}}}*/

/*{{{  PUBLIC const char *anymessagestring*/
PUBLIC const char *anymessagestring(const int n)
{
  switch(n)
    {
      case ANY_BAD_TAG:
        return "Unknown tag %s in %s";
      case ANY_OBJFILE_WRITE_ERROR:
        return "Cannot write to object file: error code %d";
      case ANY_OUT_OF_SPACE:
      /*return("Run out of symbol space");*/
      #if 0 /*def CONFIG*/
        return("Not enough memory to configure this program");
      #else
        return("Not enough memory to compile this program");
      #endif
      case ANY_FATAL_ABORT:
        return("Internal inconsistency found in routine \"%s\"");
      case ANY_FILEINFO:
        return("At line %d of file \"%s\"...");
      default:
        return NULL;
    }
}
/*}}}*/
/*{{{  PUBLIC const char *synmessagestring*/
PUBLIC const char *synmessagestring ( const int n)
{
  switch (n)
    {
      /*{{{  LEX errors*/
      case LEX_E_QUOTE:
        return ("Expected single quote, found \"%c\"");
      case LEX_E_DOUBLEQUOTE:
        return ("Expected double quotes, found \"%c\"");
      case LEX_E_HEXDIGIT:
        return ("Expected hex digit, found \"%c\"");
      case LEX_ILLEGAL_CHAR:
        return ("Illegal character (ASCII %d)");
      /*case LEX_FORMAT_ERROR:
        return ("Format error");*/
      case LEX_NUMBER_ERROR:
        /*return("In number");*/
        return("Expected digit, found \"%c\"");
      case LEX_SIGN_ERROR:
        return("Exponent must be signed");
      case LEX_EOF:
        return ("Unexpected end of file");
      case LEX_ILLEGAL_STRING_CHAR:
        return("Illegal character \"%c\" following \"*\" in string");
      case LEX_STRING_ERROR:
        return("Illegal character (ASCII %d) in string");
      case LEX_CHAR_ERROR:
        return("Illegal character constant");
      case LEX_BAD_STRING_LENGTH:
        return("String too long to insert \"*l\" value");
      case LEX_BAD_STRING_CONTINUATION:
        return("Illegal string continuation");
      case LEX_BAD_INDENT:
        return("Incorrect indentation of string continuation");
      case LEX_BAD_STRING_LENGTH_POSN:
        return("'*l' must be at start of string");
      case LEX_READ_ERROR:
        return("Error code %d when reading file");
      /*case LEX_TOO_MANY_FILES:
        return("Files nested too deeply: maximum depth is %d");*/
      case LEX_FILE_OPEN_ERROR:
        return("Cannot open file \"%s\"");
      case LEX_FILE_LENGTH_ERROR:
        return("Maximum filename length is %d characters");
      /*
      case LEX_BAD_SCDESC_FORMAT:
        return("Bad descriptor file tag %d");
      case LEX_BAD_SC_COMPATIBILITY:
        return("Object code not compatible with current version (%s)");
      case LEX_BAD_SC_PROCTYPE:
        return("Object code compiled for incompatible target (%s)");
      case LEX_BAD_SC_ERRORMODE:
        return("Object code compiled in incompatible error mode (%s)");
      case LEX_TOO_MANY_NESTED_SCS:
        return("Cannot open \"%s\" - too many nested SCs");
      */
      case LEX_MISSING_ENTRYPOINT:
        return("Missing entrypoint \"%s\"");
      #if 0 /* now done in ANY_MODULE */
      case LEX_OBJFILE_WRITE_ERROR:
        /*return("%d writing to object file");*/
        return("Cannot write to object file: error code %d");
      #endif
      /* No longer relevant cos there's now no hard limit
      case LEX_TOO_MANY_ENTRYPOINTS:
        return("Too many entry points in file \"%s\"");
      */
      case LEX_BAD_DESC_FORMAT:
        return("Bad object file format");
      case LEX_LIBRARY_NOT_SC:
        return("\"%s\" is a library, not an SC");
      case LEX_HASHWORD_ERROR:
        return("Badly formed compiler directive");
      /*case LEX_TCOFF_IOERR :
        return ("TCOFF IO ERROR: 64 bit number!");*/
      /*case LEX_TCOFF_BAD_LCHAR :
        return ("Characters 'l%c' unknown in record string");*/
      /*case LEX_TCOFF_BAD_CHAR :
        return ("Character %c unknown in record string");*/
      case LEX_TCOFF_BAD_DESC :
        return ("Peculiar descriptor : \"%s\"");
      /*
      case LEX_TCOFF_BAD_PROCESSOR :
        return ("Unknown processor type in object file");
      case LEX_TCOFF_BAD_MODE :
        return ("Unknown error mode in object file");
      */
      case LEX_TCOFF_UNMATCHED_ENDMODULE :
        return ("Unmatched END_MODULE");
      /*case LEX_TCOFF_DESC_OVERFLOW :
        return ("Descriptor overflow");*/
      /*#ifdef CONFIG*/
      case LEX_TCOFF_NOT_LINKED:
        return ("\"%s\" is not a linked object file");
      /*#else*/
      case LEX_TCOFF_WRONG_FORMAT:
        return ("\"%s\" is not a valid object file");
      /*#endif*/
      case LEX_NAME_TRUNCATED:
        return("Name too long - maximum length is %d characters");
      case LEX_NUMBER_TRUNCATED:
        return("Number too long - maximum length is %d characters");
      /*case LEX_LINE_TRUNCATED:
        return("Line too long - maximum length is %d characters");*/
      
      /*}}}*/
      /*{{{  SYN errors*/
      case SYN_TOO_MANY_NAMES:
        return("One name only allowed in specification or retype");
      case SYN_PROTOCOL_NOT_DECLARED:
        return ("Protocol is not declared");
      /*case SYN_ILLEGAL_CONVERSION:
        return ("Illegal conversion");*/
      case SYN_INVALID_PROCESS:
        return ("Invalid process");
      case SYN_BAD_INDENT:
        return ("Incorrect indentation");
      /*case SYN_ILLEGAL_TAG:
        return ("Illegal treenode tag");*/
      case SYN_ILLEGAL_DECL_TYPE:
        return ("Illegal type %s for declaration");
      case SYN_LIST_IN_IO:
        return ("Only one element allowed on left-hand side of input or output");
      case SYN_SC_ERROR:
        return ("Item not allowed at outermost level of compilation unit");
      case SYN_STRING_TOO_LONG:
        return ("String is too long for buffer (max %d bytes)");
      case SYN_FILE_OPEN_ERROR:
        return("Cannot open file \"%s\"");
      case SYN_MISSING_FILENAME:
        return("Missing filename");
      case SYN_BAD_GUYLINE:
        return("Badly formed code insert line");
      case SYN_GUY_NOT_ENABLED:
        return("Code insertion is not enabled");
      case SYN_GUY_IS_OBSOLETE:
        return "Implementation restriction: GUY is no longer supported: use ASM";
      case SYN_OPTION_IN_BAD_POS:
        return("Option in illegal position");
      case SYN_BAD_OPTION:
        return("Unrecognised option \"%c\" in option string");
      #if 0
      case SYN_E_HCOMMENT:
        return("Expected string after #COMMENT");
      case SYN_E_OPTION:
        return("Expected string after #OPTION");
      #endif
      case SYN_E_STRING_AFTER:
        return "Expected string after %s";
      /*case SYN_PROG_ERROR:
        return ("Item not allowed in PROGRAM unit");*/
      case SYN_BAD_DESCRIPTOR:
        return ("Descriptor has incorrect format");
      case SYN_SC_EMPTY:
        return ("No PROC or FUNCTION declared in SC module");
      case SYN_BAD_PRAGMA_NAME :
        return ("Unknown #PRAGMA name: %s; ignored");
      /*case SYN_BAD_PRAGMA_POSN:
        return ("#PRAGMA not allowed in this position");*/
      case SYN_BAD_PRAGMA_DIRECTIVE :
        return ("Badly formed #PRAGMA %s directive");
      case SYN_STDLIB_HAS_VS:
        return ("Standard library routine %s requires vectorspace");
      /*case SYN_DUPLICATE_IN_LIB :
        return ("Routine %s imported by multiple #USEs");*/
      case SYN_NOTHING_IN_LIB :
        return ("No compatible entrypoints found in \"%s\"");
      #if 0
      case SYN_TRANSLATE_NUL :
      /*return ("TRANSLATE ignored: String for %s contains NUL character");*/
        return ("TRANSLATE ignored: String for %s contains NUL character");
      #endif
      case SYN_TRANSLATE_ILLEGAL_CHAR :
        return ("TRANSLATE ignored: String for %s contains illegal character: %s");
      case SYN_TRANSLATE_DUPLICATE_IN :
        return ("TRANSLATE ignored: Name %s has already been used");
      case SYN_TRANSLATE_DUPLICATE_EX :
        return ("TRANSLATE ignored: String \"%s\" has already been used");
      case SYN_TRANSLATE_SEQUENCE :
        return ("TRANSLATE ignored: Module containing \"%s\" has already been loaded");
      case SYN_SC_IS_OBSOLETE :
        return ("#SC \"%s\" is obsolete; use #USE instead");
      case SYN_DIFFERENT_DESCRIPTORS:
        return ("Differing definitions of %s in library; check formal parameters and channel usage");
      case SYN_PLACE_NAME:
        return ("You may only PLACE a name");
      case SYN_SKIPPING_DEFN:
        return ("Skipping over %s definition");
      case SYN_PREDEF_ERROR:
        return ("Predefine error"); /* internal error */
      #ifdef USER_DEFINED_OPERATORS
      case SYN_UDO_NOT_SUPPORTED:
        return ("a binary operator here");
      case SYN_UDO_MULTI_RET_PARAMS:
        return ("Operator overloading function cannot have multiple return parameters");
      case SYN_UDO_NO_RET_PARAM:
        return ("Operator overloading function must have one return parameter");
      case SYN_UDO_TWO_PARAMS:
        return ("two parameters for binary operator function");
      case SYN_UDO_ONE_PARAM:
        return ("one parameter for unary operator function");
      case SYN_UDO_ONE_OR_TWO_PARAMS:
        return ("either one or two parameters for operator function");
      #endif
      #ifdef INITIAL_DECL
      case SYN_INITDECL_ADIM_MISSING:
        return ("Open arrays not supported in INIITAL declarations");
      #endif
      #ifdef RESOURCES
      case SYN_INITDECL_RESOURCE:
        return ("RESOURCE type not allowed in INITIAL declarations");
      #endif
      /*}}}*/
      /*{{{  expected "%s" found "%s" messages*/
      /*case SYN_E_TOKEN:  abort();*/ /* this is processed externally */
      case SYN_E_EXPR_OR_SPEC:    return "expression or specification";
      case SYN_E_GUARD_OR_SPEC:   return "guard or specification";
      /*case SYN_E_ELEMENT_OR_SPEC: return "element or specification";*/
      case SYN_E_SPEC:            return "specification";
      case SYN_E_PTYPE:           return "primitive type";
      case SYN_E_SPECIFIER:       return "specifier";
      case SYN_E_ELEMENT:         return "element";
      case SYN_E_OPERAND:         return "operand";
      /*case SYN_E_INTTYPE:         return "integer type";*/
      case SYN_E_EXPR:            return "expression";
      case SYN_E_IS_OR_RETYPES:   return "IS or RETYPES";
      case SYN_E_ACTION:          return "\"?\", \"!\" or \":=\"";
      case SYN_E_PAR_OR_ALT:      return "PAR or ALT";
      case SYN_E_RBOX_OR_FROM:    return "\"]\" or FROM";
      case SYN_E_AT_OR_IN:        return "AT or IN";
      case SYN_E_NAME:            return "name";
      case SYN_E_NEWLINE:         return "new line";
      case SYN_E_COLON_IS_RETYPES:return "\":\", IS or RETYPES";
      case SYN_E_COLON_IS:        return "\":\" or IS";
      case SYN_E_WS_OR_VS:        return "WORKSPACE or VECSPACE";
      /*case SYN_E_PROCESSOR_TYPE:  return "processor type";*/
      #if 0
      case SYN_E_LINK_DEL:        return "INTERVAL or DELETING";
      #endif
      /*}}}*/
      default:
        return NULL;
    }
}
/*}}}*/
/*{{{  PUBLIC const char *chkmessagestring*/
PUBLIC const char *chkmessagestring ( const int n )
{
  switch (n)
    {
      /*{{{  cases*/
      /*case CHK_UNKNOWN_TYPE:
        return ("Unknown type %s");*/
      case CHK_TYPES_DIFF:
        return ("Operands have different types for operator %s");
      case CHK_INVTYPE:
        return ("Type of operand invalid for operator %s");
      case CHK_CONSTRUCTOR_TYPE_MISMATCH:
        return ("Table elements are of different types");
      case CHK_INVCOUNT:
        return ("Shift count must be of type INT");
      case CHK_INV_EXACT:
        return ("Type of operand invalid for exact type conversion %s");
      case CHK_INV_ROUND_TRUNC:
        return ("Type of operand invalid for \'ROUND\' or \'TRUNC\'");
      case CHK_ASS_UNBALANCED:
        return ("Number of variables does not match number of expressions");
      case CHK_INV_SUBSCRIPT:
        return ("%s subscripted too many times");
      case CHK_INVTYPE_MULTI_ASS:
        return ("Variable type does not match assigned expression %s");
      /*case CHK_INVTYPE_SINGLE_ASS:
        return ("Variable type does not match assigned expression%s");*/ /* expects NUL string */
      case CHK_ABBR_TYPE_MISMATCH:
        return ("Expression type does not match abbreviation specifier");
      /*case CHK_UNKNOWN_NODE:
        return ("Unknown node tag %d encountered");*/
      case CHK_INVTYPE_FRESULT:
       return ("Expression %s does not match type of corresponding function result");
      case CHK_TOOMANYEXPS:
        return ("Function result list contains too many expressions");
      case CHK_TOOFEWEXPS:
        return ("Function result list contains too few expressions");
      case CHK_RETYPE_TYPE_MISMATCH:
        return ("Size of RETYPE does not match the expression size");
      case CHK_INV_FUNC_LIST:
        return ("Multi-valued function may not be used as expression operand");
      case CHK_INV_VALOF_LIST:
        return ("Multi-valued valof may not be used as expression operand");
      case CHK_ADIM_NOT_CONST:
        return ("Array dimension is not a constant expression");
      case CHK_ADIM_NOT_INT:
        return ("Array dimension is not of type INT");
      case CHK_INV_PROT:
        return ("Invalid channel protocol");
      /*case CHK_INVTYPETAG:
        return ("Invalid type tag %d");*/
      case CHK_TOO_FEW_VARS:
        return ("There are more expressions than variables in multiple assignment");
      case CHK_TOO_MANY_VARS:
        return ("There are more variables than expressions in multiple assignment");
      /*case CHK_INV_SIZE_OPERAND:
        return ("\'SIZE\' operand has invalid type");*/
      case CHK_CFOLD_OFLOW:
        return ("Overflow when evaluating constant expression");
      /*case CHK_NAMENOTCONST:
        return ("Constant expression contains non-constant name");*/
      case CHK_NAMENOTARRAY:
        /*return ("Subscripted operand is not an array");*/
        return ("Item may not be subscripted");
      /*case CHK_SUBSCRIPT_NOTCONST:
        return ("Constant expression contains non-constant subscript");*/
      case CHK_SUBSCRIPT_OUT_OF_RANGE:
        return ("Subscript out of range");
      case CHK_INV_DECL_TAG:
        return ("Invalid tag %s in decl node");
      case CHK_ADIM_NEGATIVE:
        return ("Array dimension must be greater than zero");
      case CHK_ADIM_TOOBIG:
        return ("Array dimension is too large");
      /*case CHK_INV_CONSTRUCT:
        return ("Invalid construct tag %s");*/
      /*case CHK_INV_LIST:
        return ("Invalid list tag %d");*/
      case CHK_INV_SEGMENT_OPERAND:
        return ("Segmented element is not an array");
      /*case CHK_SLENGTH_NOT_CONST:
        return ("Segment length not constant");*/
      case CHK_ADIM_MISSING:
        return ("Array type must have dimension specified");
      case CHK_INVTYPE_SPEC:
        return ("Illegal specifier");
      /*case CHK_INV_EXP:
        return ("Invalid expression %d");*/
      case CHK_INVTYPE_CONSTEXP:
        return ("Constant expression has invalid type %d");
      case CHK_EXP_NOT_CONST:
        return ("Expression is not constant");
      case CHK_TOO_MANY_ELSES:
        return ("More than one ELSE in CASE process");
      case CHK_MULTIPLE_CASE:
        return ("Option is used more than once in this CASE");
      case CHK_INVTYPE_SELECTOR:
        return ("Invalid selector type");
      case CHK_SELN_NOTCONST:
        return ("Option must be a constant expression");
      case CHK_INVTYPE_SELN:
        return ("Type mismatch for CASE option");
      /*case CHK_UNRESOLVED_INTTYPE:
        return ("Integer number value must have a type specifier");*/
      #ifndef OCCAM2_5
      case CHK_UNRESOLVED_REALTYPE:
        return ("Real number value must have a type specifier");
      #endif
      case CHK_TYPE_MISMATCH:
        return ("Type mismatch");
      case CHK_TYPE_MISMATCH_EXPECTED:
        return "Type mismatch: expected %s";
      case CHK_ASUB_NOT_INT:
        return ("Array index expression must be of type INT");
      case CHK_FN_VAR_PARAM:
        return ("Formal parameter to function must be VAL");
      case CHK_PTYPE_MISMATCH:
        return ("I/O list item %d does not match protocol");
      case CHK_TOO_FEW_PITEMS:
        return ("Not enough items in i/o list");
      case CHK_TOO_MANY_PITEMS:
        return ("Too many items in i/o list");
      case CHK_BAD_PTAG:
        return ("Illegal tag in i/o list");
      case CHK_BAD_ANY_PROT:
        return ("Cannot send zero length segment on CHAN OF ANY");
      case CHK_INVTYPE_STARTEXP:
        return ("Replicator start expression not of type INT");
      case CHK_INVTYPE_LENGTHEXP:
        return ("Replicator count expression not of type INT");
      case CHK_REPL_NOT_CONST:
        return ("Implementation restriction: PAR replicator must have a constant count value");
      case CHK_REPL_NEGATIVE:
        return ("Negative replicator counts are invalid");
      case CHK_REPL_OVERFLOW:
        return "Replicator variable overflows";
      case CHK_INVTYPE_GUARD:
        return ("Guard must be of type BOOL");
      #if 0
      case CHK_INV_VAL:
      /*return ("Invalid type specification for VAL parameter or abbreviation");*/
        if ((current_fe_data->fe_lang & FE_LANG_OCCAM) != 0)
          return ("Cannot use VAL on a CHAN, PORT, or TIMER");
        else if ((current_fe_data->fe_lang & FE_LANG_NDL) != 0)
          return ("Cannot use VAL on a hardware item");
        else
          return ("Cannot use VAL on a CHAN, PORT, TIMER, or hardware item");
      #else
      case CHK_INV_VAL_OCCAM:
        return ("Cannot use VAL on a CHAN, PORT, or TIMER");
      #endif
      case CHK_NAME_NOT_PROC:
        return ("%s is not a procedure");
      case CHK_INVVARIABLE_PARAM:
        return ("Parameter %d is not a variable");
      case CHK_INVTYPE_PARAM:
        return ("Type mismatch for parameter %d in call of %s");
      case CHK_TOO_FEW_ACTUALS:
        return ("Not enough actual parameters in call of %s");
      case CHK_TOO_MANY_ACTUALS:
        return ("Too many actual parameters in call of %s");
      case CHK_INV_CINDEX:
        return ("Invalid subscript into table");
      case CHK_NOT_CHANNEL:
        return ("Left-hand side must be of type CHAN or PORT");
      case CHK_INV_TIMER_INPUT:
        return ("Invalid timer input");
      case CHK_TIME_TYPE_MISMATCH:
        return ("Time expression must be of type INT");
      case CHK_NOT_CHANNEL_OR_TIMER:
        return ("Left-hand side must be of type CHAN, PORT, or TIMER");
      case CHK_NOT_TIMER:
        return ("Left-hand side must be of type TIMER");
      case CHK_INV_PORT_PROT:
        return ("Invalid port protocol");
      case CHK_INDISTINCT_TAG:
        return ("Tag name %s conflicts with another tag name");
      case CHK_NOT_A_TAG:
        return ("I/O list does not begin with a tag");
      case CHK_LOST_VARIANT:
        return ("Cannot find variant");
      case CHK_NAME_NOT_FUNCTION:
        return ("%s is not a function");
      case CHK_LOST_VALOF:
        return ("Lost valof, found %d");
      case CHK_VALOF_SCOPE:
        return ("Side-effect: %s is not assignable within VALOF");
      case CHK_BAD_VALOF:
        return ("%s is not allowed within a VALOF");
      case CHK_VALOF_CALL_SCOPE:
       return ("Side-effect: instanced PROC assigns to free variable %s");
      case CHK_PLACE_IN_VALOF:
        return ("Possible side-effect: PLACED variable %s");
      case CHK_PLACE_IN_VALOF_CALL:
        return ("Possible side-effect: instanced PROC has PLACED variable %s");
      case CHK_SUBSCRIPT_RANGE:
        return ("Subscript value %d is out of range");
      case CHK_SEG_START_RANGE:
        return ("Segment start value %d is out of range");
      case CHK_SEG_LENGTH_RANGE:
        return ("Segment length %d is longer than array");
      case CHK_SEG_RANGE:
        return ("Segment end value %d is out of range");
      /*case CHK_CONSTANT_SEGMENT_ERROR:
        return ("Error in constant segment folding, tag = %d,");*/
      case CHK_BAD_PLACE_NAME:
        return ("%s may not be placed");
      case CHK_INVTYPE_PLACEMENT:
        return("Placement address must be of type INT");
      case CHK_TOO_MANY_MULTIVALUES:
        return("Only one expression list allowed on right-hand side of assignment");
      case CHK_NAME_NOT_DECLARED:
        return("%s is not declared");
      /*case CHK_TAG_CONFLICT:
        return ("Tag name %s conflicts with protocol name");*/
      #ifndef OCCAM2_5
      case CHK_BAD_FUNCTION_TYPE:
        return ("Function types must be scalar");
      case CHK_BAD_VALOF_TYPE:
        return ("VALOF RESULT types must be scalar");
      #endif
      case CHK_BAD_INPUT_PROTOCOL:
        return ("Simple input is not allowed on a channel with a tagged protocol");
      case CHK_BAD_CASE_INPUT_PROTOCOL:
      case CHK_BAD_TAGGED_INPUT_PROTOCOL:
        return("Case input is not allowed on a channel without a tagged protocol");
      case CHK_ALT_INPUT_PORT:
        return("PORT input is not allowed in an alternative");
      case CHK_BAD_WSPLACE:
        return("Only vector declarations may be placed");
      case CHK_NAME_MULTIPLY_PLACED:
        return("%s has already been placed");
      case CHK_MULTIPLE_VARIANT:
        return ("Tag %s is used more than once in case input");
      case CHK_TOO_MANY_TAGS:
        return ("Too many tags in protocol definition - %d allowed");
      case CHK_BAD_DEST:
        return ("Cannot write to %s");
      /*case CHK_BAD_COUNT:
        return ("counted array protocol not allowed on anarchic channel");*/
      case CHK_BAD_COLON2:
        return "Right hand part of counted array communication is not an array";
      case CHK_COUNT_OUT_OF_RANGE:
        return ("Count is out of range for I/O list item %d");
      /*case CHK_ENTRYPOINT_NOT_FOUND:
        return("Cannot find entry point %s in entry point list");*/
      /*case CHK_NAME_STACK_OVERFLOW:
        return("Implementation limit: maximum of %d names in scope");*/
      case CHK_UNIMPLEMENTED_ASMCODE:
        return("Implementation restriction: Instruction %s not available");
      case CHK_BAD_GUY_OPERAND:
        return("Instruction operand must be a scalar constant expression or a name");
      case CHK_DISABLED_GUYCODE:
        return("Instruction %s is not available in current code insertion mode");
      case CHK_INVALID_GUYCODE:
        return("Instruction %s is not available on target processor");
      case CHK_BAD_GUYCODE:
        return("Instruction %s does not exist");
      case CHK_BAD_STEPCODE:
        return("STEP instruction must be a secondary");
      case CHK_ASM_BAD_STORE:
        return("Operand to a store must be an element");
      case CHK_ASM_TOO_BIG:
        return ("Operand does not fit into a word");
      case CHK_MISSING_OPERAND:
        /*return("Instruction %s takes an operand");*/
        return("Instruction %s has too few operands");
      case CHK_EXTRANEOUS_OPERAND:
        /*return("Instruction %s does not take an operand");*/
        return("Instruction %s has too many operands");
      case CHK_BAD_VALOF_CALL:
        return ("Instanced PROC contains %s, not allowed within a VALOF");
      case CHK_INVTYPE_PROCEXP:
        return("Processor expression must be of type %s");
      #if 0  /* #ifndef CONFIG */
      case CHK_INVPROCTYPE :
        return("Unrecognised processor type `%s'\n");
      #endif
      case CHK_NESTED_PRI_PAR:
        return("PRI PAR is nested inside another PRI PAR");
      case CHK_NESTED_PRI_PROC:
        return("%s contains a nested PRI PAR");
      case CHK_INV_PRI_PAR:
        return("Implementation restriction: A PRI PAR must have two component processes");
      case CHK_NOT_IMPLEMENTED:
        return("Implementation restriction: %s not implemented");
      case CHK_UNKNOWN_CONSTRUCTOR_SIZE:
        return("Implementation restriction: Table size must be known");
      case CHK_ASS_NAME_USED_MULTIPLY:
        return ("%s appears more than once on left-hand side of assignment");
      #ifdef CONDEXP
      case CHK_CONDEXP_NOT_BOOL :
        return("Condition in expression must be of type BOOL");
      #endif
      case CHK_INV_ABBR :
        return ("Right hand side of abbreviation must be assignable");
      case CHK_INV_RETYPE :
        /*return ("Retyping an item of type CHAN, PORT, or TIMER is not allowed");*/
        /*return ("Retyping an item of type TIMER is not allowed");*/
        return ("You may only RETYPE data items");
      case CHK_NAME_NOT_USED:
        return ("%s is not used");
      case CHK_PARAM_NOT_USED:
        return ("Parameter %s is not used");
      case CHK_ROUTINE_NOT_USED:
        return ("Routine %s is not used");
      case CHK_BAD_VALRETYPE:
        return ("RETYPE cannot be VAL for CHAN, PORT or TIMER");
      case CHK_BAD_CHANRETYPE:
        /*return ("CHAN, PORT or TIMER may only be RETYPEd to CHAN etc");*/
        return ("RETYPE of CHAN or PORT must be VAL");
      case CHK_RETYPE_NOTYPE:
        return ("RETYPE must have a type specifier");
      case CHK_BAD_DEST_CONSTRUCTOR:
        return ("Cannot write to table");
      case CHK_BAD_DEST_STRING:
        return ("Cannot write to string");
      case CHK_INV_SHIFT:
        return ("Shift count is out of range");
      case CHK_INV_MULTIASSIGN:
        return ("Implementation restriction: Arrays may not be unknown size");
      /*case CHK_LABEL_INSIDE_INLINE:
        return ("Implementation restriction: No labels inside INLINE routines");*/
      case CHK_NAME_IS_DESCOPED:
        return ("Name %s descopes a previous declaration");
      /*case CHK_ARRAY_SIZE_OVERFLOW:
        return ("Implementation restriction: Array size is too large");*/
      case CHK_RSUB_NOT_FIELD:
      #if 0 /*def CONFIG*/
        return ("Expected a NODE attribute, found a subscript expression");
      /*return ("Subscript expression is not a NODE attribute");*/
      #else
        return ("Expected a field name of %s, found a subscript expression");
      /*return ("Subscript expression is not a field name of %s");*/
      #endif
      case CHK_LONELY_FIELD:
      /*return ("Field selector %s may not appear on its own");*/
        return ("Attribute %s may not appear on its own");
      case CHK_LIB_WRONG_TARGET:
        return ("%s %s is not compiled for a compatible processor type or mode");
      case CHK_LIB_WRONG_IO:
        return ("%s %s is not compiled with library i/o");
      case CHK_LIB_WRONG_IO_CONFIG:
        return ("%s %s is not compiled with library i/o; this is required unless both interactive debugging and virtual routing are disabled");
      case CHK_INV_ASSERT:
        return ("Parameter to ASSERT is FALSE");
      case CHK_DUPLICATE_LABEL:
        return "Label `%s' multiply defined in this routine";
      case CHK_BAD_PRAGMA_SHARED_NAME:
      /*return "%s may not be shared";*/
        return "May not apply #PRAGMA %s to `%s'";
      case CHK_NONLOCAL_LABEL:
      /*return "Access to labels at different lexical level is not permitted";*/
        return "Access to non-local labels is not permitted";
      case CHK_CHAN_OF_ANY_USED:
      /*return "CHAN OF ANY is obsolete - use PROTOCOL name IS ANY for greater security";*/
        return "CHAN OF ANY is obsolete: use PROTOCOL name IS ANY";
      case CHK_CHAN_OF_ANY_PARAM_PASSING:
        return "Obsolete channel type conversion: use channel RETYPE";
      case CHK_OUTERLEVEL_CONSTANT:
        return "Implementation restriction: Expression for outermost level VAL must be constant";
      case CHK_COLON2_INPUT:
      /*  return ("Cannot use length `%s' in array part of counted array input");*/
        return ("Using length `%s' in array part of counted array input is obsolete");
      case CHK_BAD_ASM_INLINE_OP:
      /*return ("Cannot access formal parameter `%s' inside INLINE routine");*/
        return ("Cannot use formal parameter `%s' as operand to primary instruction inside INLINE routine");
      case CHK_WARN_GUY:
        return "GUY construct is no longer supported. You should use the ASM construct";
      case CHK_MISSING_TAG_IN_CASE_INPUT:
        return "Tag %s is not handled in CASE input";
      case CHK_NON_EXHAUSTIVE_CASE_INPUT:
        return "CASE input does not handle all tags";
      case CHK_NOT_ADJACENT_NAME:
        return "PRAGMA or PLACEment must immediately follow declaration of %s";
      /*case CHK_SEGMENT_CONSTRUCTOR:
        return "Implementation restriction: Cannot take a segment of a constructor";*/
      case CHK_UNPLACED_PORT:
        return "PORT %s must be PLACED";
      case CHK_INV_TYPE_DECORATION:
        return "Invalid type for literal";
      case CHK_TAG_IN_EXPRESSION:
        return "Illegal use of PROTOCOL tag \"%s\" in an expression";
      case CHK_RETYPINGREAL64:
        return "Retyping from/to REAL64 when words are swapped";
      /*}}}*/
      /*{{{  occam 2.5 errors*/
      #ifdef OCCAM2_5
      case CHK_BAD_USERTYPE:
        return "Invalid type for type declaration";
      case CHK_ARRAY_CONVERSION:
        return "Cannot convert to an array type";
      case CHK_ILLEGAL_FIELD_DECL:
        return "Illegal field item declared inside RECORD";
      case CHK_DUPLICATE_FIELD_DECL:
        return "Duplicate field name `%s'";
      case CHK_UNALIGNED_PACKED_FIELD:
        return "Implementation restriction: Field `%s' must be aligned to a multiple of %s bytes for a PACKED RECORD";
      case CHK_UNALIGNED_PACKED_RECORD:
        return "Implementation restriction: PACKED RECORD `%s' must be an integral number of words long";
      case CHK_BAD_CONSTRUCTOR_TYPE:
        return "Invalid type decoration for table";
      case CHK_REC_LIT_TYPE_MISMATCH:
        return "Incorrect type for field `%s' in record literal";
      case CHK_REC_LIT_TOO_SMALL:
        return "Too few items in record literal";
      case CHK_REC_LIT_TOO_LARGE:
        return "Too many items in record literal";
      case CHK_OFFSETOF_NOT_RECORD:
        return "%s is not a RECORD type name";
      case CHK_OFFSETOF_NOT_FIELD:
        return "%s is not a field of %s";
      case CHK_BAD_RESHAPES:
        return "Can only RESHAPE from an array to an array of similar base type";
      case CHK_BAD_TYPENAME:
        return "Name %s is not a TYPE";
      case  CHK_NO_RECF_ZERO:
        return ("The number of RECORD fields must be greater than zero");
      case CHK_BAD_STRING_TYPE:
        return "Invalid type decoration for string";
      #endif
      /*}}}*/
      #if 1 /*def CONFIG*/
      /*{{{  Configurer checker errors*/
      case CHK_ILLEGAL_CONFIG_CONSTRUCT:
        return ("Illegal item in configuration code");
      case CHK_ILLEGAL_CONSTRUCT:
        return ("Illegal item inside %s construct");
      case CHK_DUPLICATE_CONSTRUCT:
        return ("Multiple %s constructs not permitted");
      case CHK_ZERO_CONSTRUCT:
        return ("No %s construct");
      
      #if 0
      case CHK_INVTYPE_EDGE:
        return("CONNECT expression must be of type %s");
      case CHK_INVTYPE_ARC:
        return("WITH expression must be of type %s");
      case CHK_INVTYPE_DEVICE:
        return("SET expression must be of type %s");
      #else
      case CHK_INVTYPE_CONFIG:
        return("%s expression must be of type %s");
      #endif
      
      case CHK_INVTYPE_ATTR:
        return("Unknown attribute name \"%s\"");
      case CHK_INV_MAPPING_RHS:
        return("Right hand side of mapping must be of type NODE or ARC");
      case CHK_INV_MAPPING_LHS:
        return("Left hand side of mapping must be of type %s");
      case CHK_INV_MAPPING_PRI:
        return("Priority expression of mapping must be of type INT");
      case CHK_INV_MAPPING_NOPRI:
        return("No priority expression permitted when mapping CHANs");
      case CHK_INV_PLACEON_RHS:
        return("Right hand side of mapping must be of type ARC");
      case CHK_ILLEGAL_PLACE:
        return("Cannot PLACE %s which was declared outside a PROCESSOR construct");
      case CHK_CALLING_CONVENTION:
        return("FUNCTION %s returns a REAL result but is compiled for wrong calling convention");
      case CHK_FIELD_NOT_VALID:
        return "Attribute %s is not available for a %s";
      case CHK_FIELD_NOT_SET:
        return "Attribute %s may not be SET";
      case CHK_INV_VAL_NDL:
          return ("Cannot use VAL on a hardware item");
      case CHK_INV_VAL_CONFIG:
          return ("Cannot use VAL on a CHAN, PORT, TIMER, or hardware item");
      case CHK_INV_RHS_NDL_MAPPING:
          return ("Right hand side of mapping must be of type EDGE");
      case CHK_INV_LHS_NDL_MAPPING:
          return ("Left hand side of mapping must be of type EDGE");
      case CHK_NDL_MAPPING_TOO_MANY_SOURCES:
          return ("Map statement must have only one source");
      #ifdef USER_DEFINED_OPERATORS
      case CHK_UDO_NOT_SUPPORTED:
          return ("Overloading on this operator is not currently supported");
      #endif
      #if 0
      case CHK_INV_CONFIG:
        return ("Invalid configuration");
      case CHK_BAD_NETDECL :
        return("Invalid declaration in network description");
      case CHK_BAD_ATTRIBUTE :
        return("Invalid %s attribute");
      case CHK_MISSING_ATTRIBUTE :
        return("Missing %s attribute");
      case CHK_TOOMANY_ATTRIBUTES :
        return("Too many attributes for node");
      case CHK_MISSING_INIT:
        return ("Declaration must have initialisation");
      #endif
      
      
      /*}}}*/
      #endif
      default:
        return (NULL);
    }
}
/*}}}*/
/*{{{  PUBLIC const char *usemessagestring*/
PUBLIC const char *usemessagestring ( const int n)
{
  switch (n)
    {
      /*{{{  cases*/
      /*case USE_BAD_TAG:
        sprintf(errorstring, "Illegal tag %d in routine %s",
                (long int)p1, (char *)p2);
        return(errorstring);*/
      /*case USE_GLOBAL_FREE_VARS:
        return("There are global free variables");*/
      case USE_VAR_VAR_ABBREVIATED:
        return ("`%s' cannot be used (it has been abbreviated)");
      case USE_VAR_VAL_ABBREVIATED:
        return ("`%s' cannot be altered (it has been VAL abbreviated)");
      case USE_BAD_CHAN_PARAM:
        return ("Channel parameter `%s' cannot be used for both input and output");
      case USE_BAD_FREE_CHAN:
        return ("Free channel `%s' cannot be used for both input and output");
      case USE_WRITTEN_IN_PAR:
        return ("Variable `%s' is assigned to in parallel");
      case USE_READ_AND_WRITTEN_IN_PAR:
        return ("Variable `%s' is read and assigned to in parallel");
      case USE_OUTPUT_IN_PAR:
      /*return ("Channel `%s' is output on in parallel");*/
        return ("Parallel outputs on channel `%s'");
      case USE_INPUT_IN_PAR:
      /*return ("Channel `%s' is input on in parallel");*/
        return ("Parallel inputs on channel `%s'");
      /* I've removed the fixed limit - CO'N 17/2/90
      case USE_TOO_MANY_OVERLAP_CHECKS:
        return("Too many overlap checks required: maximum is %d");
      */
      /*case USE_VAR_TABLE_SUBSCRIPTS :
        return "All subscripts in channel tables must be constant";*/
      case USE_VAR_FREEVAR:
      case USE_VAL_FREEVAR:
        return ("Aliasing error, `%s' is a free variable");
      /*case USE_FUTURE_COLON2_INPUT:*/
      /*return ("Cannot use length `%s' in array part of counted array input");*/
      /*return ("Future implementation restriction - Cannot use length `%s' in array part of counted array input");*/
      case USE_COLON2_INPUT:
        return ("Cannot use length `%s' in array part of counted array input");
      case USE_NO_ALIASCHECK:
        return "Alias and usage checking is disabled";
      case USE_NO_USAGECHECK:
        return "Usage checking is disabled";
      case USE_TIMER_IN_PRI_PAR:
        return "Implementation restriction: TIMER `%s' should not be read in both high and low priority branches of a PRI PAR";
      /*}}}*/
      default:
        return (NULL);
    }
}
/*}}}*/
/*{{{  PUBLIC const char *vtimessagestring*/
PUBLIC const char *vtimessagestring ( const int n)
{
  switch (n)
    {
      /*{{{  cases*/
      /*case VTI_STACK_OVERFLOW:
        return("Workspace stack overflow");*/
      /*case VTI_STACK_UNDERFLOW:
        return("Workspace stack underflow");*/
      /*case VTI_BAD_TAG:
        return("Illegal tag %d in routine %s");*/
      /*case VTI_NAMETRANS_OVERFLOW :
        return("Name translation stack overflow");*/
      case VTI_NAMETRANS_UNDERFLOW :
        return("Name translation stack underflow");
      /*case VTI_TOO_MANY_NAMETRANS :
        return("Too many names for name translation");*/
      case VTI_ARRAY_SIZE_OVERFLOW:
        return ("Implementation restriction: Array size is too large");
      case VTI_EXPAND_NOT_CONSTANT:
        return "Actual parameter corresponding to formal parameter %s must be a constant";
      /*}}}*/
      default:
        return (NULL);
    }
}
/*}}}*/
