/* $Id: lex1.c,v 1.5 1997/09/16 15:58:52 djb1 Exp $ */


/*
 *	occam two lexical analyser
 *	Copyright (C) 1987 Inmos Limited
 *
 *	This program is free software; you can redistribute it and/or modify
 *	it under the terms of the GNU General Public License as published by
 *	the Free Software Foundation; either version 2 of the License, or
 *	(at your option) any later version.
 *
 *	This program is distributed in the hope that it will be useful,
 *	but WITHOUT ANY WARRANTY; without even the implied warranty of
 *	MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *	GNU General Public License for more details.
 *
 *	You should have received a copy of the GNU General Public License
 *	along with this program; if not, write to the Free Software
 *	Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 */

/*#define DEBUG*/
/*#define DEBUG_NEXTSYMB*/

/*{{{  include files */
#ifdef HAVE_CONFIG_H
#  include <config.h>
#endif
#include <stdio.h>
#include <string.h>

#include "popen_re.h" /* IMPORTED */

#include "feinc.h"
#include "lexerror.h"
#include "lexdef.h"
#include "chkdef.h" /* current_fe_handle etc */
#include "objrdr.h"
#include "predefhd.h"
/*}}}*/

/*{{{  definitions */
/* String termination character,
   mustn't clash with any valid occam character */
#define CLOSE_QUOTE 257

#define MINLEGALCH 32
#define MAXLEGALCH 126

/*{{{  filestruct_t */
/* This is used to keep track of the nesting of #INCLUDE files */

typedef struct filestruct_s
  {
    struct filestruct_s *next;
    FILE *fileptr;
    char fmode;
    SOURCEPOSN ffileposn;
    int fbaseindent;
    int fstackposition;
    char fname[MAX_FILENAME_LENGTH];
  } filestruct_t;
/*}}}*/
/*}}}*/

/*{{{  global variables */
PUBLIC FILE *infile;

PUBLIC int symb;
PUBLIC int symbindent, lineindent;

PUBLIC wordnode *lexword;
PUBLIC char literalv[MAXSTRING_SIZE + 1];
PUBLIC int literalp;
PUBLIC int pdnumber;  /* **SHC 5-Apr-1988 */

PRIVATE char current_filename[MAX_FILENAME_LENGTH];
/*PUBLIC int current_file;*/
PUBLIC SOURCEPOSN flocn = NOPOSN;

PUBLIC int lexmode;

PUBLIC int currentfilenum;
PUBLIC int allow_asmnames;

#ifdef USER_DEFINED_OPERATORS
extern BOOL user_defined_operators;
#endif
/*}}}*/
/*{{{  local variables */
PRIVATE int ch;
PRIVATE int stringcontinuation;
PRIVATE int insertlength;
PRIVATE int linep;
PRIVATE int currentindent;
PRIVATE int baseindent;
PRIVATE int string_indent; /* bug 1251 25/9/91 */ /* indentation of line containing the start of the string */

PRIVATE int totallines;

/* TRUE when inputline cannot read any more from file */
PRIVATE BOOL endoffile;

/* TRUE when the lexer has sent an S_END symbol: any call to nextsymb
   after this is an unexpected end of file. */
PRIVATE BOOL sentend;

/*{{{  line buffer handling */
PRIVATE int linecount;  /* lines in this file */

PRIVATE int         linebuffer_len;  /* length of current line */
PRIVATE const char *linebuffer_ptr;  /* ptr to current line */

PRIVATE BOOL eof;                           /* TRUE when last symb was END     */
PRIVATE BOOL newlineflag;                   /* TRUE when last symb was NEWLINE */
PRIVATE BOOL desc_eof;        /* TRUE when we hit the end of a descriptor file */
PRIVATE filestruct_t *filestack = NULL;
PRIVATE int filestackptr = 0;
/*}}}*/

/*}}}*/

/*{{{  forward declarations */
PRIVATE void bufinit (void);
/*}}}*/

/*{{{  error handler */
/*{{{  PUBLIC void nextline () */
/* Skip to start of next line */
PUBLIC void nextline ( void )
{
  while (symb != S_NEWLINE)
    nextsymb ();
  while (symb == S_NEWLINE)
    nextsymb ();
}
/*}}}*/
/*{{{  PUBLIC void skiplines (indent) */
PUBLIC void skiplines (const int indent )
{
  /*int i = 0;*/

  if (symbindent > indent)  /* bug 1018 2/11/90 */
    nextline(); /* force at least a single new line, even if this line starts ok */

  while (lineindent > indent)
    {
      nextline ();
      /*++i;*/
    }
#if 0
  if (i != 0)
    fprintf (errfile, "Skipped %d line%s\n", i, (i=1) ? "" : "s");
#endif
  while (symb == S_COMMENT)
    nextline ();
}
/*}}}*/
/*}}}*/
/*{{{  predefine handling */
/*{{{  predef_lang_t */
typedef enum
  {
    A  = FE_LANG_ALL,                       /* all languages */
    O  = FE_LANG_OCCAM,                     /* compiler only */
    C  = FE_LANG_CONFIG2 | FE_LANG_CONFIG3, /* configurer only */
    CC = FE_LANG_OCCAM   | C,               /* compiler and configurer */
    N  = FE_LANG_NDL,                       /* NDL reader only */
    CN = FE_LANG_CONFIG2 | N,               /* configurer H/W and NDL reader */
    C2 = FE_LANG_CONFIG2                    /* Version 2 configurer only */
  } predef_lang_t;
/*}}}*/

/*{{{  predefined routines defns. */
PRIVATE const struct predeflinestruct
  /*{{{  structure format */
  {
    predef_lang_t langs;
    int pnumber;
    const char *pline;
  } predeflines[] =
  /*}}}*/
  /* These don't HAVE to be in the correct order, but it is clearer */
  {
    /*{{{  longarithmetic */
    { CC, PD_LONGADD, "FI(VI,VI,VI):LONGADD"},
    { CC, PD_LONGSUM, "FI,I(VI,VI,VI):LONGSUM"},
    { CC, PD_LONGSUB, "FI(VI,VI,VI):LONGSUB"},
    { CC, PD_LONGDIFF,"FI,I(VI,VI,VI):LONGDIFF"},
    { CC, PD_LONGPROD,"FI,I(VI,VI,VI):LONGPROD"},
    { CC, PD_LONGDIV, "FI,I(VI,VI,VI):LONGDIV"},
    /*}}}*/
    /*{{{  shift */
    { CC, PD_SHIFTRIGHT,"FI,I(VI,VI,VI):SHIFTRIGHT"},
    { CC, PD_SHIFTLEFT, "FI,I(VI,VI,VI):SHIFTLEFT"},
    /*}}}*/
    /*{{{  normalise,fracmul */
    { CC, PD_NORMALISE,"FI,I,I(VI,VI):NORMALISE"},
    { CC, PD_FRACMUL,  "FI(VI,VI):FRACMUL"},
    /*}}}*/
    /*{{{  shift,arithmeticshift,rotate */
    { CC, PD_ASHIFTRIGHT,"FI(VI,VI):ASHIFTRIGHT"},
    { CC, PD_ASHIFTLEFT, "FI(VI,VI):ASHIFTLEFT"},
    { CC, PD_ROTATERIGHT,"FI(VI,VI):ROTATERIGHT"},
    { CC, PD_ROTATELEFT, "FI(VI,VI):ROTATELEFT"},
    /*}}}*/
    /*{{{  causeerror */
    { CC, PD_CAUSEERROR,"P():CAUSEERROR"},
    /*}}}*/
    /*{{{  kernelrun,loadchannel,loadchannelvector,loadbytevector */
    { CC, PD_KERNELRUN,              "P(V[]Y,VI,[]I,VI):KERNEL.RUN"},
    { CC, PD_LOADINPUTCHANNEL,       "P(I,C):LOAD.INPUT.CHANNEL"},
    { CC, PD_LOADOUTPUTCHANNEL,      "P(I,C):LOAD.OUTPUT.CHANNEL"},
    { CC, PD_LOADINPUTCHANNELVECTOR, "P(I,[]C):LOAD.INPUT.CHANNEL.VECTOR"},
    { CC, PD_LOADOUTPUTCHANNELVECTOR,"P(I,[]C):LOAD.OUTPUT.CHANNEL.VECTOR"},
    { CC, PD_LOADBYTEVECTOR,         "P(I,V[]Y):LOAD.BYTE.VECTOR"},
    /*}}}*/
    /*{{{  unpacksnroundsn */
    { CC, PD_UNPACKSN,"FI,I,I(VI):UNPACKSN"},
    { CC, PD_ROUNDSN, "FI(VI,VI,VI):ROUNDSN"},
    /*}}}*/
    /*{{{  draw2d,clip2d,move2d */
    { CC, PD_DRAW2D,"P(V[][]Y,VI,VI,[][]Y,VI,VI,VI,VI):DRAW2D"},
    { CC, PD_CLIP2D,"P(V[][]Y,VI,VI,[][]Y,VI,VI,VI,VI):CLIP2D"},
    { CC, PD_MOVE2D,"P(V[][]Y,VI,VI,[][]Y,VI,VI,VI,VI):MOVE2D"},
    /*}}}*/
    /*{{{  crcbyteandword */
    { CC, PD_CRCWORD,"FI(VI,VI,VI):CRCWORD"},
    { CC, PD_CRCBYTE,"FI(VI,VI,VI):CRCBYTE"},
    /*}}}*/
    /*{{{  bitops */
    { CC, PD_BITCOUNT,   "FI(VI,VI):BITCOUNT"},
    { CC, PD_BITREVWORD, "FI(VI):BITREVWORD"},
    { CC, PD_BITREVNBITS,"FI(VI,VI):BITREVNBITS"},
    /*}}}*/
    /*{{{  floatingpoint */
    { CC, PD_ABS,       "FR3(VR3):ABS"},
    { CC, PD_ISNAN,     "FB(VR3):ISNAN"},
    { CC, PD_NOTFINITE, "FB(VR3):NOTFINITE"},
    { CC, PD_ORDERED,   "FB(VR3,VR3):ORDERED"},
    { CC, PD_MINUSX,    "FR3(VR3):MINUSX"},
    { CC, PD_MULBY2,    "FR3(VR3):MULBY2"},
    { CC, PD_DIVBY2,    "FR3(VR3):DIVBY2"},
    { CC, PD_SQRT,      "FR3(VR3):SQRT"},
    { CC, PD_FPINT,     "FR3(VR3):FPINT"},
    { CC, PD_DABS,      "FR6(VR6):DABS"},
    { CC, PD_DISNAN,    "FB(VR6):DISNAN"},
    { CC, PD_DNOTFINITE,"FB(VR6):DNOTFINITE"},
    { CC, PD_DORDERED,  "FB(VR6,VR6):DORDERED"},
    { CC, PD_DMINUSX,   "FR6(VR6):DMINUSX"},
    { CC, PD_DMULBY2,   "FR6(VR6):DMULBY2"},
    { CC, PD_DDIVBY2,   "FR6(VR6):DDIVBY2"},
    { CC, PD_DSQRT,     "FR6(VR6):DSQRT"},
    { CC, PD_DFPINT,    "FR6(VR6):DFPINT"},
    { CC, PD_SCALEB,    "FR3(VR3,VI):SCALEB"},
    { CC, PD_DSCALEB,   "FR6(VR6,VI):DSCALEB"},
    { CC, PD_COPYSIGN,  "FR3(VR3,VR3):COPYSIGN"},
    { CC, PD_DCOPYSIGN, "FR6(VR6,VR6):DCOPYSIGN"},
    { CC, PD_NEXTAFTER, "FR3(VR3,VR3):NEXTAFTER"},
    { CC, PD_DNEXTAFTER,"FR6(VR6,VR6):DNEXTAFTER"},
    { CC, PD_LOGB,      "FR3(VR3):LOGB"},
    { CC, PD_DLOGB,     "FR6(VR6):DLOGB"},
    { CC, PD_FLOATINGUNPACK, "FI,R3(VR3):FLOATING.UNPACK"},
    { CC, PD_DFLOATINGUNPACK,"FI,R6(VR6):DFLOATING.UNPACK"},
    { CC, PD_ARGUMENTREDUCE, "FB,I32,R3(VR3,VR3,VR3):ARGUMENT.REDUCE"},
    { CC, PD_DARGUMENTREDUCE,"FB,I32,R6(VR6,VR6,VR6):DARGUMENT.REDUCE"},
    /*}}}*/
    /*{{{  IEEEarithmetic */
    { CC, PD_REAL32OP,    "FR3(VR3,VI,VR3):REAL32OP"},
    { CC, PD_REAL64OP,    "FR6(VR6,VI,VR6):REAL64OP"},
    { CC, PD_IEEE32OP,    "FB,R3(VR3,VI,VI,VR3):IEEE32OP"},
    { CC, PD_IEEE64OP,    "FB,R6(VR6,VI,VI,VR6):IEEE64OP"},
    { CC, PD_REAL32REM,   "FR3(VR3,VR3):REAL32REM"},
    { CC, PD_REAL64REM,   "FR6(VR6,VR6):REAL64REM"},
    { CC, PD_REAL32EQ,    "FB(VR3,VR3):REAL32EQ"},
    { CC, PD_REAL64EQ,    "FB(VR6,VR6):REAL64EQ"},
    { CC, PD_REAL32GT,    "FB(VR3,VR3):REAL32GT"},
    { CC, PD_REAL64GT,    "FB(VR6,VR6):REAL64GT"},
    { CC, PD_IEEECOMPARE, "FI(VR3,VR3):IEEECOMPARE"},
    { CC, PD_DIEEECOMPARE,"FI(VR6,VR6):DIEEECOMPARE"},
    { CC, PD_IEEE32REM,   "FB,R3(VR3,VR3):IEEE32REM"},
    { CC, PD_IEEE64REM,   "FB,R6(VR6,VR6):IEEE64REM"},
    /*}}}*/
    /*{{{  reschedule/assert */
    { CC, PD_RESCHEDULE,"P():RESCHEDULE"},
    { CC, PD_ASSERT,    "P(VB):ASSERT"},
    { CC, PD_WSSIZEOF,  "FI(VI):WSSIZEOF"},
    { CC, PD_VSSIZEOF,  "FI(VI):VSSIZEOF"},
    /*}}}*/
    /*{{{  Used by the compiler to update profile counts */
    { CC, PD_UPDATE_PROFCOUNT, "P(VI):UPDATE.PROFCOUNT"},
    /*}}}*/  
    /*{{{  configurer s/w attributes */
    { C,  PD_IMS_GET_PDATA, "FI():IMS.GET.PDATA"},
    /*}}}*/
    /*{{{  configurer attributes */
    { CN, PD_ATTR_LINK,          ". [4]E :link"},
    { CN, PD_ATTR_TYPE,          ". []Y :type"},
    { CN, PD_ATTR_MEMSIZE,       ". I :memsize"},
    { CN, PD_ATTR_ROMSIZE,       ". I :romsize"},
    { CN, PD_ATTR_ROOT,          ". B :root"},
    { CN, PD_ATTR_MEMSTART,      ". I :memstart"},
    { CN, PD_ATTR_NUM_LINKS,     ". I :numlinks"},
    { C,  PD_ATTR_ORDER_CODE,    ". I :order.code"},
    { C,  PD_ATTR_ORDER_VS,      ". I :order.vs"},
    { C,  PD_ATTR_LOCATION_CODE, ". I :location.code"},
    { C,  PD_ATTR_LOCATION_WS,   ". I :location.ws"},
    { C,  PD_ATTR_LOCATION_VS,   ". I :location.vs"},
    { C,  PD_ATTR_RESERVED,      ". I :reserved"},
    { C2, PD_ATTR_TOLERANCE,     ". I :tolerance"},
    { C2, PD_ATTR_LINKQUOTA,     ". I :linkquota"},
    { C2, PD_ATTR_ROUTECOST,     ". I :routecost"},
    { C,  PD_ATTR_ORDER_WS,      ". I :order.ws"},
    { C,  PD_ATTR_NODEBUG,       ". B :nodebug"},
    { C,  PD_ATTR_NOPROFILE,     ". B :noprofile"},
    
    { C2, PD_HOSTEDGE,           "E :HOST"},
    { N,  PD_DEFAULTNODE,        "N :DEFAULT"},
    /*}}}*/
    /*{{{  NDL reader attributes */
    { N,  PD_ATTR_LINK_SPEED_MULTIPLY  , ". I :link.speed.multiply"},
    { N,  PD_ATTR_LINK_SPEED_DIVIDE    , ". []I :link.speed.divide"},
    { N,  PD_ATTR_CONTROL              , ". E :control"},
    { N,  PD_ATTR_CONTROL_UP           , ". E :control.up"},
    { N,  PD_ATTR_CONTROL_DOWN         , ". E :control.down"},
    { N,  PD_ATTR_CONTROL_SPEED_DIVIDE , ". [2]I :control.speed.divide"},
    { N,  PD_ATTR_MEMCONFIG            , ". []Y :memconfig"},
    { N,  PD_ATTR_PREAMBLE             , ". []Y :preamble"},
    { N,  PD_ATTR_BOOTSTRAP            , ". []Y :bootstrap"},
    { N,  PD_ATTR_MEMORY               , ". [][3]I :memory"},
    { N,  PD_ATTR_CACHESIZE            , ". I :cachesize"},
    /*{ N,PD_ATTR_LINK_TXMODE          , ". [4]B :link.Txmode"},*/
    { N,  PD_ATTR_LINK_BYTE_MODE       , ". [4]B :link.byte.mode"},
    { N,  PD_ATTR_EVENT_OUT            , ". [4]B :event.out"},
    { N,  PD_ATTR_LOCAL_ROM            , ". B :local.rom"},
    { N,  PD_ATTR_SYSTEM_ROM           , ". B :system.rom"},
    { N,  PD_ATTR_PMI_CONFIG_INROM     , ". B :pmi.config.inrom"},
    { N,  PD_ATTR_LINKSET_INROM        , ". B :linkset.inrom"},
    { N,  PD_ATTR_BOOTSTRAP_INROM      , ". B :bootstrap.inrom"},
    { N,  PD_ATTR_CACHE_CONFIG_INROM   , ". B :cache.config.inrom"},
    /*{ N,PD_ATTR_INTERVAL             , ". [][2]I :interval"},*/
    { N,  PD_ATTR_LINK_SELECT          , ". []I :link.select"},
    { N,  PD_ATTR_DISCARD              , ". []B :discard"},
    { N,  PD_ATTR_DELETE               , ". []B :delete"},
    { N,  PD_ATTR_RANDOM               , ". []B :random"},
    /*{ N,PD_ATTR_RANDOMSEED           , ". I :randomseed"},*/
    { N,  PD_ATTR_RANDOM_INTERVAL      , ". [2]I :random.interval"},
    { N,  PD_ATTR_NODE                 , ". N :node"},
    { N,  PD_ATTR_LINKS                , ". [][2]I :links"},
    { N,  PD_ATTR_START_CODE           , ". Y :start.code"},
    { N,  PD_ATTR_IDENTIFY_CODE        , ". Y :identify.code"},
    { N,  PD_ATTR_IDENTIFY_RESPONSE    , ". Y :identify.response"},
    { N,  PD_ATTR_EXPECTED_IDENTITY    , ". I :expected.identity"},
    { N,  PD_ATTR_ERROR_RESPONSE       , ". Y :error.response"},
    { N,  PD_ATTR_ERROR_HANDSHAKE      , ". Y :error.handshake"},
    { N,  PD_ATTR_INITIALISATION       , ". []I :initialisation"},
    { N,  PD_ATTR_DATA                 , ". E :data"},
    { N,  PD_ATTR_OUTWARD              , ". []A :outward"},
    { N,  PD_ATTR_RETURN               , ". []A :return"},
    { N,  PD_ATTR_LINK_GROUPS          , ". [][2]I :link.groups"},
    { N,  PD_ATTR_INTERVAL_SEPARATOR   , ". []I :interval.separator"},
    { N,  PD_ATTR_ROM_FILE             , ". []Y :rom.file"},
    { N,  PD_ATTR_REBOOT_FROM_LINK     , ". B :reboot.from.link"},
    { N,  PD_ATTR_BOOT_FROM_ROM        , ". B :boot.from.rom"},
    { N,  PD_ATTR_MODE                 , ". I :mode"},
    { N,  PD_ATTR_DS_LINK              , ". [4]E :ds.link"},
    
    /*}}}*/
    { A, 0, ""}
};
/*}}}*/

/*{{{  predecode */
PRIVATE char *addto ( char *l , const char *s )
{
  while (*s != '\0')
    *(l++) = *(s++);
  return l;
}

/* Decode the predefine string into line */
PRIVATE void predecode ( char *l , const char *s )
{
  while (*s != '\0' && *s != ':')
    /*{{{  decode */
    {
      const char *expand;
      switch (*s)
        {
          case 'B' : expand = "BOOL";        break;
          case 'C' : expand = "CHAN OF ANY"; break;
          case 'F' : expand = "FUNCTION ";   break;
          case 'I' : expand = "INT";         break;
          case 'P' : expand = "PROC";        break;
          case 'R' : expand = (*++s == '3' ? "REAL32" : "REAL64"); break;
          case 'V' : expand = "VAL ";        break;
          case 'Y' : expand = "BYTE";        break;
    #if 1 /*def CONFIG*/
          case 'A' : expand = "ARC";         break;
          case 'E' : expand = "EDGE";        break;
          case 'N' : expand = "NODE";        break;
    #endif
          default  : expand = NULL;          break;
        }
      if (expand != NULL)
        l = addto(l, expand);
      else
        *(l++) = *s;
      s++;
    }
    /*}}}*/
  if (*s == ':') s++;
  l = addto(l, s);
  l = addto(l, ":\n");
  *l = '\0';
}
/*}}}*/
/*{{{  PRIVATE char *readpredefline (line) */
/*{{{  comment */
/*****************************************************************************
 *
 *  readpredefline copies a predefined name specification into line,
 *                 and returns a pointer to the first character of line.
 *
 *****************************************************************************/
/*}}}*/
PRIVATE char *readpredefline ( char *const line, int *const line_len)
{
  if (linecount < ((sizeof(predeflines) / sizeof(struct predeflinestruct))-1))
    {
      if ((predeflines[linecount].langs & current_fe_data->fe_lang) != 0)
        predecode(line, predeflines[linecount].pline);
      else
        strcpy(line, "\n");
      pdnumber = predeflines[linecount].pnumber; /* **SHC 5-Apr-1988 */
      *line_len = strlen(line);
      return(line);
    }
  return(NULL);
}
/*}}}*/
/*{{{  PUBLIC void init_predefs_lexer */
PUBLIC void init_predefs_lexer(void)
{
  lexmode = LEX_PREDEFS;
  strcpy(current_filename, "");
  desc_eof = FALSE;
  bufinit ();
  nextsymb();
  if (symb == S_NEWLINE)
    nextline();
}
/*}}}*/
/*{{{  PUBLIC const char *get_predef_string */
PUBLIC const char *get_predef_string(const int pdno)
{
  int i;
  for (i = 0; i <  ((sizeof(predeflines) / sizeof(struct predeflinestruct))-1); i++)
    if (predeflines[i].pnumber == pdno)
      {
        const char *const s = strrchr(predeflines[i].pline, ':');
        if (s == NULL) return NULL;
        return s+1; /* skip past the colon char */
      }
  return NULL;
}
/*}}}*/
/*}}}*/

/*{{{  input buffer */
/* This version allows lines of arbitrary length */

/*This _doesn't_ impose a hard limit on the length of lines;
  they are read in in buffer chunks of this size.
*/
#define LINEMAX             255

PRIVATE char *linebuffer = NULL;
PRIVATE size_t linesize  = LINEMAX;

/*{{{  PRIVATE char *readsourceline */
PRIVATE char *readsourceline(int *const line_len)
{
  char *ptr;
  int spaces = 0;
  BOOL expandtabs = TRUE; /* expand only until first non-whitespace character */
  char buf[LINEMAX];
  const char *tab_ptr = NULL;
  register const char *src_ptr = buf;
  register char *local_linebuffer = linebuffer;
  register int dest_col = 0;

  ptr = fgets(buf, LINEMAX, infile);
  if (ptr == NULL) return NULL;

  while (TRUE)
    {
      register int c = *src_ptr++;
      switch(c)
        {
          case '\n': /* hit the end of a line */
          eol_label:
            local_linebuffer[dest_col  ] = '\n';
            local_linebuffer[dest_col+1] = '\0';
            *line_len = dest_col + 1;
            return local_linebuffer;
          case '\0': /* hit the end of a buffer before end of line */
            DEBUG_MSG(("readsourceline: reading more from file\n"));
            ptr = fgets(buf, LINEMAX, infile);
            /*if (ptr == NULL) return NULL;*/
            if (ptr == NULL) goto eol_label; /* INSdi03520 allow for final line with no lf */
            src_ptr = buf;
            break;
          case '\t':
            DEBUG_MSG(("readsourceline: found a tab\n"));
            if (expandtabs)
              {
                src_ptr--; /* point back to the tab character */
                if (spaces != 0)
                  {
                    c = ' '; spaces--;
                    goto addchar_label;
                  }
                if (src_ptr == tab_ptr) /* at the end of the last tab */
                  src_ptr++; /* finished 'processing' the tab, now move on */
                else
                  {
                    tab_ptr = src_ptr; /* remember this position */
                    spaces  = TABSIZE - (dest_col % TABSIZE);
                  }
                break; /* break out of the whole switch */
                /* else we have finished all the spaces for this tab */
              }
            DEBUG_MSG(("readsourceline: dropping through from tab\n"));
            /* FALL THROUGH */
          default:
            expandtabs = FALSE; /* don't expand tabs once a char has been seen */
            /* FALL THROUGH */
          case ' ':
          addchar_label:
            local_linebuffer[dest_col++] = c;
            if (dest_col >= linesize)
              {
                char *const new = memalloc(linesize * 2 + 2); /* 2 extra for "\n\0" */
                DEBUG_MSG(("readsourceline: extending linebuffer to %ld\n", linesize*2));
                memcpy(new, linebuffer, linesize);
                memfree(linebuffer);
                linebuffer = new;
                local_linebuffer = linebuffer;
                linesize *= 2;
              }
            break;
        }
    }
}
/*}}}*/

/*{{{  PRIVATE void readline () */
/*{{{  comment */
/*****************************************************************************
 *
 *  readline updates lineptr to point to a structure containing
 *           the next line of source, reading
 *           a line from the input file if neccessary.
 *
 *****************************************************************************/
/*}}}*/
PRIVATE void readline ( void )
{
  do
    {
      DEBUG_MSG(("readline: calling inputline\n"));
      /*{{{  input a line */
      /*****************************************************************************
       *
       *  inputline reads a text line from the input file into a structure element
       *            of array buf.
       *
       *****************************************************************************/
      {
        const char *ptr;
        /*{{{  read in line */
        switch (lexmode)
          {
            default: /*case LEX_SOURCE:*/
              ptr = readsourceline(&linebuffer_len);
              break;
            case LEX_SC:
            case LEX_LIB:
            case LEX_STDLIB:
              ptr = readdescriptorline(linebuffer, lexmode == LEX_STDLIB,
                                       &linebuffer_len);
              break;
          #if 1 /*ndef CONFIG*/ /* no #PRAGMA EXTERNAL when configuring */
            case LEX_EXTERNAL:
              ptr = readexternalline(&linebuffer_len);
              break;
          #endif
            case LEX_PREDEFS:
              ptr = readpredefline(linebuffer, &linebuffer_len);
              break;
          }
        /*}}}*/
        if (ptr == NULL)
          /*{{{  end of file */
          {
          #if 0  /* This changed 23/4/90 by CO'N */
            if ((lexmode == LEX_PREDEFS) || (lexmode == LEX_EXTERNAL) || feof(infile))
              endoffile = TRUE;
            else
              lexfatal_i(LEX_READ_ERROR, flocn, ferror(infile));
          #else
            if ((lexmode == LEX_SOURCE) && (!feof(infile)))
              lexfatal_i(LEX_READ_ERROR, flocn, ferror(infile));
            else
              endoffile = TRUE;
          #endif
            linebuffer_len = 0;
            DEBUG_MSG(("inputline: end of file\n"));
          }
          /*}}}*/
        else
          /*{{{  calculate the length of the line */
          {
            switch(lexmode)
              {
                case LEX_STDLIB: case LEX_PREDEFS:
                  break;
                default:
                  fe_sourcehash(current_fe_handle, ptr, linebuffer_len);
                  break;
              }
          }
          /*}}}*/
        linebuffer_ptr = ptr;
      }
      /*}}}*/
      eof |= endoffile;
      linecount++;
      /*if (lexmode == LEX_SOURCE) totallines++;*/
      totallines += (lexmode == LEX_SOURCE);
    }
  while (!eof && (linebuffer_len == 0));

  linep = 0;
}
/*}}}*/
/*{{{  PRIVATE int rch () */
/* Read a character from the input file */
#if 0
PRIVATE int rch ( void )
{
  if (linep >= linebuffer_len)
    readline();

  currentindent = linep + baseindent;
  return (eof ? EOF : linebuffer_ptr[linep++]);
}
#else
#define rch() ( ((linep >= linebuffer_len) ? readline() : (void)0) , \
                currentindent = linep + baseindent, \
                (eof ? EOF : linebuffer_ptr[linep++]) )
#endif
/*}}}*/
/*{{{  PRIVATE int rsch() */
/* Read a character from the input file, give error message if hit EOF */
PRIVATE int rsch ( void )
{
  const int c = rch();
  if (c == EOF)
    lexfatal(LEX_EOF, flocn);
  return (c);
}
/*}}}*/
/*{{{  PRIVATE void bufinit () */
PRIVATE void bufinit ( void )
{
  linecount   = 0;
  sentend     = FALSE;
  eof         = FALSE;
  endoffile   = FALSE;
  readline ();
  symbindent  = 0;
  lineindent  = 0;
  newlineflag = TRUE;
  stringcontinuation = FALSE;
  ch = rch();
}
/*}}}*/

/*{{{  PRIVATE void suspend_file() */
/*****************************************************************************
 *
 *  suspend_file saves the current state of infile in the filestack.
 *
 *****************************************************************************/
PRIVATE void suspend_file (const char *const name, const int mode, FILE *const new_fptr )
{
  if (infile != NULL)
    {
      filestruct_t *fptr   = memalloc(sizeof(filestruct_t));
      fptr->fileptr        = infile;
      strcpy(fptr->fname, current_filename);
      fptr->fmode          = lexmode;
      SetFileNum (fptr->ffileposn, currentfilenum);
      SetFileLine(fptr->ffileposn, linecount);
      fptr->fbaseindent    = baseindent;
      fptr->fstackposition = filestackptr;
      fptr->next           = filestack;
      filestack            = fptr;
      filestackptr++;
      DEBUG_MSG(("suspend_file: new filestackptr is %d, name is %s, saved baseindent is %d\n",
                 filestackptr, name, baseindent));
    }
  if (mode != LEX_EXTERNAL)
    {
      infile = new_fptr;
      strcpy (current_filename, name);
    }
  currentfilenum = fe_addfilename(current_fe_handle,
                                  (mode == LEX_EXTERNAL) ?
                                  NULL : lookupword(name, strlen(name)),
                                  (mode == LEX_SOURCE),
                                  currentfilenum, FileLineOf(flocn));
  SetFileLine(flocn, 0);
  SetFileNum(flocn, currentfilenum);
}
/*}}}*/
/*{{{  PRIVATE void patchup() */
PRIVATE void patchup()
{
  switch(lexmode)
    {
      case LEX_SC: /* obsolete */
      case LEX_LIB: case LEX_STDLIB: case LEX_EXTERNAL:
        patchdescriptors(current_filename);
        break;
      default:
        break;
    }
}
/*}}}*/
/*{{{  PRIVATE void resume_file() */
/*{{{  comment */
/*****************************************************************************
 *
 *  resume_file resumes the top entry of the file stack
 *
 *****************************************************************************/
/*}}}*/
PRIVATE void resume_file ( void )
{
  patchup();

  /*{{{  pop the details off the stack */
  {
    filestruct_t *fptr = filestack;
  
    if (lexmode != LEX_EXTERNAL)
      fclose(infile);
  
    strcpy(current_filename, fptr->fname);
    infile     = fptr->fileptr;
  
    /* We resume a line after where we suspended */
    lexmode    = fptr->fmode;
    baseindent = fptr->fbaseindent;
    flocn      = fptr->ffileposn;
    DEBUG_MSG(("resume_file: setting baseindent to %d\n", baseindent));
    {
      filestruct_t *fnext = filestack->next;
      memfree(filestack); /* freeup the top of the stack */
      filestack = fnext;
      filestackptr--;
    }
  
  }
  /*}}}*/

  bufinit();
  linecount      = FileLineOf(flocn) + 1;
  currentfilenum = FileNumOf(flocn);
}
/*}}}*/
/*{{{  PRIVATE void close_file() */
/*{{{  comment */
/*****************************************************************************
 *
 *  close_file closes the currently open file
 *
 *****************************************************************************/
/*}}}*/
PRIVATE void close_file ( void )
{
  DEBUG_MSG(("close_file\n"));
  patchup();
  if (lexmode != LEX_PREDEFS)
    fclose(infile);
  infile = NULL;
  sentend = FALSE;
  eof = FALSE;
  endoffile = FALSE;
}
/*}}}*/
/*{{{  PUBLIC BOOL open_file(name, mode) */
/*{{{  comment */
/*****************************************************************************
 *
 *  open_file opens the file of name 'name', in mode 'mode', returns TRUE
 *            if opened ok, otherwise returns FALSE.
 *
 *****************************************************************************/
/*}}}*/
PUBLIC BOOL open_file (const char *const name, const int mode, const int fileindent )
{
  FILE *fptr = NULL;
  BOOL external_ok = FALSE;
  DEBUG_MSG(("open_file: filename: %s, fileindent: %d\n", name, fileindent));
  if ((strlen(name) >= MAX_FILENAME_LENGTH) &&
      (mode != LEX_EXTERNAL)) /* bug 331 29/8/90 */
    lexerr_i(LEX_FILE_LENGTH_ERROR, flocn, MAX_FILENAME_LENGTH);
  else
    {
      switch(mode)
        {
          case LEX_SOURCE:
            fptr = popen_relative(name, current_fe_data->fe_pathname,
                                  NULL, "r", NULL,
                                  memalloc, memfree, NULL);
            break;
          case LEX_SC:
          case LEX_LIB:
          case LEX_STDLIB:   fptr = open_descfile(name, mode);
                             break;
        #if 1 /*ndef CONFIG*/ /* No #PRAGMA EXTERNAL when configuring */
          case LEX_EXTERNAL: external_ok = init_external(name);
                             break;
        #endif
        }
    }
  if ((fptr != NULL) || external_ok)
    /*{{{  set up input buffers */
    {
      suspend_file(name, mode, fptr);
      baseindent = fileindent;
      lexmode = mode;
      desc_eof = FALSE;
      bufinit ();
      return(TRUE);
    }
    /*}}}*/
  else
    return(FALSE);
}
/*}}}*/
/*}}}*/

/*{{{  routines */
/*{{{  PRIVATE int readtag (name) */
PRIVATE int readtag ( char *const name )
{
  register int i = 0;
  register int temp_ch = ch;
  if (lexmode != LEX_STDLIB)
    while (isalnum(temp_ch) || (temp_ch == '.'))
      {
        if (i < (MAXNAMELENGTH - 1)) /* added for bug 1020 18/10/90 */
          name[i++] = temp_ch;
        temp_ch = rch ();
      }
  else /* lexmode == LEX_STDLIB */
    while (isalnum(temp_ch) || (temp_ch == '.') ||
           (temp_ch == '@') || (temp_ch == '%') || (temp_ch == '_') ||
           (temp_ch == '|') || (temp_ch == '$'))
      {
        if (i < (MAXNAMELENGTH - 1)) /* added for bug 1020 18/10/90 */
          name[i++] = temp_ch;
        temp_ch = rch ();
      }
  ch = temp_ch;
  name[i] = '\0';
  if (i >= (MAXNAMELENGTH - 1)) /* added for bug 1020 18/10/90 */
    lexwarn_i(LEX_NAME_TRUNCATED, flocn, MAXNAMELENGTH-1);
  return i;
}
/*}}}*/
/*{{{  PRIVATE void readdecimal () */
/* Read decimal digits from the input file and place them in
   literalv from position literalp onwards.
   Update literalp. */
PRIVATE void readdecimal ( void )
{
  int temp_ch = ch;
  if (isdigit(temp_ch))
    {
      while (isdigit(temp_ch))
        {
          if (literalp < (MAXSTRING_SIZE - 5)) /* bug 1020 18/10/90 */
            literalv[literalp++] = temp_ch;
          temp_ch = rch ();
        }
      ch = temp_ch;
      /* we subtract 5 here to allow for sign characters, exponents, etc */
      if (literalp >= (MAXSTRING_SIZE - 5))
        lexwarn_i(LEX_NUMBER_TRUNCATED, flocn, MAXSTRING_SIZE - 5);
    }
  else
    {
      lexerr_i(LEX_NUMBER_ERROR, flocn, temp_ch);
      literalv[literalp++] = '0';
    }
}
/*}}}*/
/*{{{  PRIVATE void rexponent () */
/* Read an exponent from the input file and place it in literalv
   from literalp onwards */
PRIVATE void rexponent ( void )
{
  literalv[literalp++] = 'E';
  ch = rch ();
  if ((ch == '+') || (ch == '-'))
    {
      literalv[literalp++] = ch;
      ch = rch ();
    }
  else
    {
      lexerr(LEX_SIGN_ERROR, flocn);
      /*lexerror (LEX_NUMBER_ERROR, flocn, 0);*/
      literalv[literalp++] = '+';
    }
  readdecimal ();
}
/*}}}*/
/*{{{  PRIVATE int hexval () */
/* Return the hexadecimal value of the ASCII hex digit in ch */
PRIVATE int hexval ( void )
{
  if (isdigit(ch))
    return (ch - '0');
  if ((ch >= 'A') && (ch <= 'F'))
    return (ch - 'A' + 10);
  lexerr_i(LEX_E_HEXDIGIT, flocn, ch);
  return (0);
}
/*}}}*/
/*{{{  PRIVATE int rescapech () */
PRIVATE int rescapech (const BOOL inside_string)
/* This routine modified for bug 1399 25/9/91 - and merged string and character
   processing */
{
  switch(ch)
    {
      /*{{{  '*' - special character */
      case '*':
        ch = rsch();
        /*{{{  skip over carriage-returns */
        while (ch == '\r' && inside_string)
          ch = rsch(); /* bug TS/2013 05/01/93 */
        /*}}}*/
        if (inside_string)
          switch (ch)
            {
              /*{{{  '\n' */
              case '\n':
                {
                  ch = rsch();
                #if 0
                  while (ch == ' ') ch = rsch();
                  SetFileLine(flocn, linecount);
                #else /* allow blank lines and comments - bug 1103 10/1/91 */
                  {
                    BOOL repeating;
                    do
                      {
                        repeating = FALSE;
                        while ((ch == ' ') || (ch == '\n')) ch = rsch(); /* bug 1103 10/1/91 */
                        SetFileLine(flocn, linecount);
                        if (ch == '-') /* check for comments */
                          {
                            /*{{{  check indent of comment  */
                            symbindent = currentindent;
                            /*lineindent = symbindent;*/
                            if (symbindent < string_indent)
                              lexerr(LEX_BAD_INDENT, flocn);
                            lineindent = string_indent; /* bug 1251 25/9/91 */
                            
                            /*}}}*/
                            ch = rsch();
                            if (ch != '-')
                              lexerr(LEX_BAD_STRING_CONTINUATION, flocn);
                            while (ch != '\n') ch = rsch();
                            repeating = TRUE;
                          }
                      }
                    while (repeating);
                  }
                #endif
                  if (ch != '*')
                    lexerr(LEX_BAD_STRING_CONTINUATION, flocn);
                  /*{{{  check indent of string continuation */
                  symbindent = currentindent;
                  /*lineindent = symbindent;*/
                  if (symbindent < string_indent)
                    lexerr(LEX_BAD_INDENT, flocn);
                  lineindent = string_indent; /* bug 1251 25/9/91 */
                  
                  /*}}}*/
                  ch = rsch();
                  return(rescapech(inside_string));
                }
              /*}}}*/
              /*{{{  l L */
              case 'l': case 'L':
                if (!stringcontinuation && (literalp == 0))
                  {
                    insertlength = TRUE;
                    return (' ');
                  }
                else
                  {
                    lexerr_i(LEX_BAD_STRING_LENGTH_POSN, flocn, ch);
                    return ('?');
                  }
              /*}}}*/
              default: break;
            }
        switch(ch)
          {
            case 'c': case 'C':        return ('\r');
            case 'n': case 'N':        return ('\n');
            case 't': case 'T':        return ('\t');
            case 's': case 'S':        return (' ');
            case '\'':                 return ('\'');
            case '"':                  return ('"');
            case '*':                  return ('*');
            case '#':
              {
                int h;
                ch = rsch();
                h = hexval();
                ch = rsch();
                return ((h << 4) | hexval());
              }
            default:
              lexerr_i(LEX_ILLEGAL_STRING_CHAR, flocn, ch);
              return (ch);
          }
        break;
      /*}}}*/
      /*{{{  ''' or '"' */
      case '\'':
        break;
      case '"':
        if (inside_string)
          return CLOSE_QUOTE;
        break;
      /*}}}*/
      /*{{{  default */
      default:
        if (ch >= MINLEGALCH && ch <= MAXLEGALCH)
          return ch;
        break;
      /*}}}*/
    }

  if (inside_string)
    {
      lexerr_i(LEX_STRING_ERROR, flocn, ch);
      return CLOSE_QUOTE;
    }
  else
    {
      lexerr(LEX_CHAR_ERROR, flocn);
      return (ch);
    }
}
/*}}}*/
/*{{{  PRIVATE int readnumber () */
/* Read a number from the input file.
The number may be an integer or a real.
A string representing the number is placed in literalv.
Return S_UREALLIT or S_UINTLIT */
PRIVATE int readnumber ( void )
{
  literalp = 0;
  readdecimal ();
  if (ch == '.')
    /*{{{  read rest of decimal number */
    {
      literalv[literalp++] = '.';
      ch = rch ();
      readdecimal ();
      if (ch == 'E')
        rexponent ();
      literalv[literalp] = '\0';
      return (S_UREALLIT);
    }
    /*}}}*/
  else
    /*{{{  return integer */
    {
      literalv[literalp] = '\0';
      return (S_UINTLIT);
    }
    /*}}}*/
}
/*}}}*/
/*{{{  PRIVATE int readhex () */
PRIVATE int readhex ( void )
{
  int temp_ch = ch;
  while (isdigit(temp_ch) || ((temp_ch >= 'A') && (temp_ch <= 'F')))
    {
      if (literalp < (MAXSTRING_SIZE - 1)) /* bug 1020 18/10/90 */
        literalv[literalp++] = temp_ch;
      temp_ch = rch ();
    }
  ch = temp_ch;
  if (literalp == 1)
    {
      lexerr_i(LEX_E_HEXDIGIT, flocn, ch);
      /*lexerror(LEX_NUMBER_ERROR, flocn, 0);*/
      return (S_ILLEGALSYMB);
    }
  else if (literalp >= (MAXSTRING_SIZE - 1))
    lexwarn_i(LEX_NUMBER_TRUNCATED, flocn, MAXSTRING_SIZE);
  literalv[literalp] = '\0';
  return (S_UINTLIT);
}
/*}}}*/
/*{{{  PRIVATE int rhashword (matchstring, s) */
/* Read characters from the input stream and ensure that they match successive
characters of the string 'matchstring'.  If the whole of matchstring is
matched return 's', otherwise return S_ILLEGALSYMB. */
PRIVATE int rhashword (const char *matchstring , const int s )
{
  int temp_ch = rch();
  while (*matchstring != '\0' && temp_ch == *matchstring)
    {
      temp_ch = rch();
      matchstring++;
    }
  ch = temp_ch;
  if (*matchstring == '\0')
    return s;
  else
    {
      lexerr(LEX_HASHWORD_ERROR, flocn);
      return S_ILLEGALSYMB;
    }
}
/*}}}*/
/*{{{  PRIVATE int readbyteliteral () */
/* Read a byte literal string from the input file and place it in literalv.
Return S_UBYTELIT */
PRIVATE int readbyteliteral ( void )
{
  ch = rescapech(FALSE);
  literalv[0] = '\'';
  literalv[1] = ch  ;
  literalv[2] = '\'';
  literalv[3] = '\0';
  literalp = 3;
  ch = rsch ();
  if (ch == '\'')
    ch = rch ();
  else
    {
      symbindent = currentindent;
      lexerr_i(LEX_E_QUOTE, flocn, ch);
      return S_ILLEGALSYMB;
    }
  return (S_UBYTELIT);
}
/*}}}*/
/*{{{  PRIVATE int readstringliteral () */
PRIVATE int readstringliteral ( void )
{
  insertlength = FALSE;
  literalp = 0;
  if (!stringcontinuation)  /* ch has already been escaped */
    {
      string_indent = lineindent; /* bug 1251 25/9/91 */ /* moved up one line 27/11/91 for bug TS/1493 */
      ch = rescapech(TRUE);
      DEBUG_MSG(("readstringliteral: setting string_indent to %d ", string_indent));
    }
  stringcontinuation = FALSE;

  while ((ch != CLOSE_QUOTE) && (literalp < MAXSTRING_SIZE))
    /*{{{  read a character into literalp */
    {
      literalv[literalp++] = ch;
      ch = rsch();
      ch = rescapech(TRUE);
    }
    /*}}}*/
  literalv[literalp] = '\0';
  stringcontinuation = (ch != CLOSE_QUOTE);
  if (insertlength)
    /*{{{  insert the length if the string is short enough */
    {
      if (!stringcontinuation)
        literalv[0] = literalp - 1;
      else
        {
          lexerr(LEX_BAD_STRING_LENGTH, flocn);
          literalv[0] = '?';
        }
    }
    /*}}}*/

  if (stringcontinuation)
    return (S_STRINGCONT);
  else
    /*{{{  read another character, return S_STRING */
    {
      ch = rch ();
      return(S_STRING);
    }
    /*}}}*/
}
/*}}}*/

/*}}}*/

/*{{{  PUBLIC void nextsymb () */
#ifdef DEBUG_NEXTSYMB
  #define nextsymb local_nextsymb
#endif

/*{{{  comment */
/*****************************************************************************
 *
 *  nextsymb sets symb and symbindent to value and indentation of next symbol
 *           from source file.  If we start a new line, lineindent is set
 *           to the indentation of the first symbol on the line.
 *
 *****************************************************************************/
/*}}}*/
PUBLIC void nextsymb ( void )
{
  int temp_ch;

  while (TRUE)
    /*{{{  find the next symbol */
    {
      if (sentend)
        lexfatal(LEX_EOF, flocn);
    
      if (newlineflag)
        /*{{{  skip leading spaces and set lineindent */
        {
          newlineflag = FALSE;
          temp_ch = ch;
          while (temp_ch == ' ')  /* skip leading spaces */
            temp_ch = rch();
          ch = temp_ch;
          lineindent = currentindent;
          DEBUG_MSG(("nextsymb: setting lineindent to: %d\n", lineindent));
        }
        /*}}}*/
    
      symbindent = currentindent;
      SetFileLine(flocn, linecount);
    
      if (stringcontinuation)
        /*{{{  read the continuing string */
        {
          symb = readstringliteral();
          return;
        }
        /*}}}*/
      else
        /*{{{  switch on ch */
        switch (ch)
          {
            /*{{{  newline                              return */
            case '\n':
              ch = rch ();
              newlineflag = TRUE;
              symb = S_NEWLINE;
              return;
            /*}}}*/
            /*{{{  space / tab                  break */
            case ' ':
              while ((temp_ch = rch()) == ' ')
                ;  /* SKIP */
              ch = temp_ch;
              break;
            case '\t':
            case '\r': /* bug TS/2013 04/01/93 */
              ch = rch();
              break;
            /*}}}*/
            /*{{{  0 1 2 3 4 5 6 7 8 9                  return */
            case '0': case '1': case '2': case '3': case '4':
            case '5': case '6': case '7': case '8': case '9':
              symb = readnumber();
              return;
            /*}}}*/
            /*{{{  a - z  A - Z                         return */
            /*{{{  case a - z A - Z */
            case 'a': case 'b': case 'c': case 'd': case 'e':
            case 'f': case 'g': case 'h': case 'i': case 'j':
            case 'k': case 'l': case 'm': case 'n': case 'o':
            case 'p': case 'q': case 'r': case 's': case 't':
            case 'u': case 'v': case 'w': case 'x': case 'y':
            case 'z':
            case 'A': case 'B': case 'C': case 'D': case 'E':
            case 'F': case 'G': case 'H': case 'I': case 'J':
            case 'K': case 'L': case 'M': case 'N': case 'O':
            case 'P': case 'Q': case 'R': case 'S': case 'T':
            case 'U': case 'V': case 'W': case 'X': case 'Y':
            case 'Z':
            /*}}}*/
              {
                char name [MAXNAMELENGTH];
                const int len = readtag(name);
                lexword = lookupword (name, len);
                symb = WTagOf(lexword);
                return;
              }
            /*}}}*/
            /*{{{  dot                                  return */
            case '.':  /* All special ASM names begin with dot */
              if (allow_asmnames)
                {  /* simply read a name */
                  char name [MAXNAMELENGTH];
                  const int len = readtag(name);
                  lexword = lookupword (name, len);
                  symb = WTagOf(lexword);
                }
              else
                {
                  ch = rch ();
                  symb = S_DOT;
                }
              return;
            /*}}}*/
            /*{{{  ( ) [ ] + , = ! ? ~ & ;              return */
            case '(': ch = rch (); symb = S_LPAREN;    return;
            case ')': ch = rch (); symb = S_RPAREN;    return;
            case ']': ch = rch (); symb = S_RBOX;      return;
            case '+': 
              ch = rch (); 
              /*{{{  user defined operator additions (for ++)*/
              #ifdef USER_DEFINED_OPERATORS
              if ((ch == '+') && user_defined_operators)
              {
                ch = rch ();
                symb = S_DPLUS;
                return;
              }
              #endif
              /*}}}*/
              symb = S_ADD;       
              return;
            case ',': ch = rch (); symb = S_COMMA;     return;
            case '=': 
              ch = rch (); 
              /*{{{  user defined operator additions (for ==)*/
              #ifdef USER_DEFINED_OPERATORS
              if ((ch == '=') && user_defined_operators)
              {
                ch = rch ();
                symb = S_DEQUALS;
                return;
              }
              #endif
              /*}}}*/
              symb = S_EQ;        
              return;
            case '!': 
              ch = rch (); 
              /*{{{  user defined opeartor additions (for !!)*/
              #ifdef USER_DEFINED_OPERATORS
              if ((ch == '!') && user_defined_operators)
              {  
                ch = rch ();
                symb = S_DEXCLAIMATION;
                return;
              }
              #endif
              /*}}}*/
              symb = S_OUTPUT;    
              return;
            /*{{{  case '%' and '^' (user defined operators only)*/
            #ifdef USER_DEFINED_OPERATORS
            case '^': ch = rch (); symb = S_HAT;       return; /* Added by Jim */
            case '%':
              if (user_defined_operators)
              {
                ch = rch ();
                if (ch == '%')
                {
                  ch = rch ();
                  symb = S_DPERCENT;
                  return;
                } 
                else
                {
                  if (ch == '>')
                  {
                    ch = rch ();
                    symb = S_PERCENTRIGHT;
                    return;
                  }
                } 
                symb = S_PERCENT;
                return;
              }
            #endif
            /*}}}*/
            case '?':  /* added by Jim */
              ch = rch (); 
              /*{{{  user defined operator additions (for ??)*/
              #ifdef USER_DEFINED_OPERATORS
              if ((ch == '?') && user_defined_operators)
                {
                  ch = rch ();
                  symb = S_DQUESTIONM;
                }
              else
              #endif
              /*}}}*/
              symb = S_INPUT;     
              return;
            /*{{{  case '@' for @@,  @>,  @*/
            #ifdef USER_DEFINED_OPERATORS
            case '@':
              if (user_defined_operators)
              {
                ch = rch ();
                if (ch == '@')
                {
                  ch = rch ();
                  symb = S_DAMPERSAT;
                  return;
                }
                else 
                {
                  if (ch == '>')
                  {
                    ch = rch ();
                    symb = S_AMPERSATRIGHT;
                    return;
                  }
                }
                    symb = S_AMPERSAT;
                    return;
                  }
            #endif
            /*}}}*/
            case '~': ch = rch (); symb = S_BITNOT;    return;
            case '&': 
              ch = rch (); 
              /*{{{  user defined operator additions for &, &&, &> etc*/
              #ifdef USER_DEFINED_OPERATORS
              if ((ch == '&') && user_defined_operators)
              {
                ch = rch ();
                symb = S_DAMPERSAND;
                return;
              }
              else
              {
                if ((ch == '>') && user_defined_operators)
                {
                  ch = rch ();
                  symb = S_AMPERSANDRIGHT;
                  return;
                }
              }
              #endif
              /*}}}*/
              symb = S_AMPERSAND; 
              return;
            case ';': ch = rch (); symb = S_SEMICOLON; return;
            /*}}}*/
            /*{{{  *                                    return */
            case '*':
              #if 0  /* we've checked for this outside the switch */
              if (stringcontinuation)
                {
                  ch = rsch ();
                  symb = readstringliteral();
                  return;
                }
              else
              #endif
                {
                  ch = rch ();
                  symb = S_MULT;
                  return;
                }
            /*}}}*/
            /*{{{  [                                    return */
            case '[':
              ch = rch ();
              if (ch == ']')
                {
                  ch = rch ();
                  symb = S_BOX;
                  return;
                }
              /*{{{  user defined operator extensions for [>*/
              #ifdef USER_DEFINED_OPERATORS
              if ((ch == '>') && (user_defined_operators)) /* Added by Jim for rotates 17/1/97 */
                {
                  ch = rch ();
                  symb = S_ROTATERIGHT;
                  return;
                }
              #endif
              /*}}}*/
              symb = S_LBOX;
              return;
            /*}}}*/
            /*{{{  :                                    return */
            case ':':
              ch = rch ();
              if (ch == '=')
                {
                  ch = rch ();
                  symb = S_ASS;
                  return;
                };
              if (ch == ':')
                {
                  ch = rch ();
                  symb = S_COLON2;
                  return;
                };
              symb = S_COLON;
              return;
            /*}}}*/
            /*{{{  $                                    return */
            case '$':
              ch = rch();
              #ifdef USER_DEFINED_OPERATORS
              if ((ch == '$') && (user_defined_operators)) /* Added by Jim */
                {
                  ch = rch ();
                  symb = S_DDOLLAR;
                  return;
                }
              #endif
              literalv[0] = '#';
              literalp = 1;
              symb = readhex();
              return;
            /*}}}*/
            /*{{{  #                                    return */
            case '#':
              {
                /* note that #INCLUDE is always allowed! */
                const BOOL directives_allowed =
                  (current_fe_data->fe_lang & FE_LANG_NDL) == 0;
            
                literalv[0] = '#';
                literalp = 1;
                switch(ch = rch())
                  {
                    /*{{{  0-9, A, B, D, E, F must be a hex number     return */
                    case '0': case '1': case '2': case '3': case '4':
                    case '5': case '6': case '7': case '8': case '9':
                    case 'A': case 'B':           case 'D': case 'E': case 'F':
                      symb = readhex();
                      return;
                    /*}}}*/
                    /*{{{  C  hex number or   #COMMENT                 return */
                    case 'C':
                      if ((ch = rch ()) == 'O' && directives_allowed)
                        symb = rhashword("MMENT", S_HCOMMENT);
                      else
                        {
                          literalv[literalp++] = 'C';
                          symb = readhex();
                        }
                      return;
                    /*}}}*/
                    /*{{{  E  hex number or   #EXTERNAL                return */
                    #if 0 /* #EXTERNAL replaced by #PRAGMA */
                    case 'E':
                      if ((ch = rch()) == 'X' && directives_allowed)
                        symb = rhashword("TERNAL", S_EXTERNAL);
                      else
                        {
                          literalv[literalp++] = 'E';
                          symb = readhex();
                        }
                      return;
                    #endif
                    /*}}}*/
                    /*{{{  I                  #INCLUDE or #IMPORT      return */
                    case 'I':
                      ch = rch();
                      if (ch == 'N') /* #INCLUDE is always allowed */
                        symb = rhashword("CLUDE", S_INCLUDE);
                      else if (ch == 'M' && directives_allowed)
                        symb = rhashword("PORT", S_IMPORT);
                      else
                        {
                          lexerr(LEX_HASHWORD_ERROR, flocn);
                          symb = S_ILLEGALSYMB;
                        }
                      return;
                    /*}}}*/
                    /*{{{  #OPTION, #PRAGMA, #SC, #USE                 return */
                    case 'N': case 'O': case 'P': case 'S': case 'U':
                      if (directives_allowed)
                        switch(ch)
                          {
                            case 'N': symb = rhashword("ETWORK", S_HNETWORK); break;
                            case 'O': symb = rhashword("PTION",  S_OPTION);   break;
                            case 'P': symb = rhashword("RAGMA",  S_PRAGMA);   break;
                            case 'S': symb = rhashword("C",      S_SC);       break;
                            case 'U': symb = rhashword("SE",     S_USE);      break;
                          }
                      else
                        {
                          lexerr(LEX_HASHWORD_ERROR, flocn);
                          symb = S_ILLEGALSYMB;
                        }
                      return;
                    /*}}}*/
                    /*{{{  a, b, c, d, e, f   badly formed hex number  return */
                    case 'a': case 'b': case 'c': case 'd': case 'e': case 'f':
                      lexerr_i(LEX_E_HEXDIGIT, flocn, ch);
                      /*lexerror(LEX_NUMBER_ERROR, flocn, 0);*/
                      symb = S_ILLEGALSYMB;
                      return;
                    /*}}}*/
                    /*{{{  default            error                    return */
                    default:
                      lexerr(LEX_HASHWORD_ERROR, flocn);
                      symb = S_ILLEGALSYMB;
                      return;
                    /*}}}*/
                  }
              }
              break; /* just to make sure */
            /*}}}*/
            /*{{{  -                            break / return */
            case '-':
              ch = rsch ();
              if (ch == '-')  /* Read past comment */
                {
                  literalp = 0;
                  temp_ch = rsch ();
                  while (temp_ch != '\n')
                    {
                      /* don't copy into literalv, so it can't overflow:
                         bug 1020 18/10/90 */
                      /* literalv[literalp++] = ch; */
                      temp_ch = rsch ();
                    }
                  ch = '\n';
                  if (!current_fe_data->fe_ignore_comments) /* bug 853 18/1/91 */
                    { symb = S_COMMENT; return; }
                  /* otherwise drop round loop */
                }
            #ifdef CONDEXP
              else if (ch == '>')
                { ch = rsch(); symb = S_RIGHTARROW; return; }
            #endif
              else
                { symb = S_SUBTRACT; return; }
              break;
            /*}}}*/
            /*{{{  /                                    return */
            case '/':
              if ((ch = rch()) == '\\')
                {
                  ch = rch();
                  symb = S_BITAND;
                  return;
                };
              symb = S_DIV;
              return;
            /*}}}*/
            /*{{{  \                                    return */
            case '\\':
              if ((ch = rch()) == '/')
                {
                  ch = rch ();
                  symb = S_BITOR;
                  return;
                };
              symb = S_REM;
              return;
            /*}}}*/
            /*{{{  <                                    return */
            case '<':
              switch (ch = rch())
                {
                  case '<': ch = rch (); symb = S_LSHIFT; return;
                  case '=': ch = rch (); symb = S_LE; return;
                  case '>': ch = rch (); symb = S_NE; return;
                  /*{{{  user defined operator extensions (for &>, [>, %>, @>)*/
                  #ifdef USER_DEFINED_OPERATORS
                  case ']': if (user_defined_operators) { ch = rch (); symb = S_ROTATELEFT; return; }
                  case '@': if (user_defined_operators) { ch = rch (); symb = S_LEFTAMPERSAT; return; }
                  case '%': if (user_defined_operators) { ch = rch (); symb = S_LEFTPERCENT; return; }
                  case '&': if (user_defined_operators) { ch = rch (); symb = S_LEFTAMPERSAND; return; }
                  #endif
                  /*}}}*/
                  default:  symb = S_LS; return;
                };
            /*}}}*/
            /*{{{  >                                    return */
            case '>':
              switch (ch = rch())
                {
                  case '<': ch = rch (); symb = S_XOR; return;
                  case '=': ch = rch (); symb = S_GE; return;
                  case '>': ch = rch (); symb = S_RSHIFT; return;
                  default:  symb = S_GR; return;
                };
            /*}}}*/
            /*{{{  '                                    return */
            case '\'':
              ch = rsch ();
              symb = readbyteliteral();
              return;
            /*}}}*/
            /*{{{  "                                    return */
            case '"':
              ch = rsch ();
              symb = readstringliteral();
              return;
            /*}}}*/
            /*{{{  EOF                          break / return */
            case EOF:
              if (filestackptr > 0)
                resume_file();
              else
                {
                  close_file();
                  symb = S_END;
                  sentend = TRUE;
                  return;
                }
              break;
            /*}}}*/
            /*{{{  default                      break */
            default:
              lexerr_i(LEX_ILLEGAL_CHAR, flocn, ch);
              symb = S_ILLEGALSYMB;
              ch = rch ();
              DEBUG_MSG(("nextsymb: looping for next char... "));
              break;
            /*}}}*/
          }
        /*}}}*/
    }
    /*}}}*/
}

#ifdef DEBUG_NEXTSYMB
#undef nextsymb
PUBLIC void nextsymb ( void )
{
  local_nextsymb();
  printlex(current_fe_data->fe_outfile);
}
#endif

/*}}}*/

/*{{{  PUBLIC void printlex () */
PUBLIC void printlex (FILE *fptr)
{
#ifdef DEBUG_NEXTSYMB
  fprintf(fptr, "%d:", symbindent);
#endif
  switch(symb)
    {
      default:
        fprintf(fptr, "%s ", tagstring(symb));
        break;
      /*{{{  S_NEWLINE; */
      case S_NEWLINE:
        fprintf(fptr, "NEWLINE\n");
        break;
      /*}}}*/
      /*{{{  S_END */
      case S_END:
        fprintf(fptr, "END\n");
        break;
      /*}}}*/
      /*{{{  S_UBYTELIT */
      case S_UBYTELIT:
        fprintf(fptr,"BYTELIT:");
        printbyte(fptr, literalv[1]);
        break;
      /*}}}*/
      /*{{{  S_INTLIT */
      #if 0
      case S_INTLIT:
        fprintf (fptr, "INTLIT:%.*s ", literalp, literalv);
        break;
      #endif
      /*}}}*/
      /*{{{  S_UINTLIT */
      case S_UINTLIT:
        fprintf (fptr, "UINTLIT:%.*s ", literalp, literalv);
        break;
      /*}}}*/
      /*{{{  S_NAME */
      case S_NAME:
        fprintf (fptr, "NAME:%s ", WNameOf(lexword));
        break;
      /*}}}*/
      /*{{{  S_STRING */
      case S_STRING:
        fprintf (fptr, "STRING:");
        printstring(fptr, literalv, literalp);
        break;
      /*}}}*/
      /*{{{  S_STRINGCONT */
      case S_STRINGCONT:
        fprintf(fptr, "STRINGCONT:");
        printstring(fptr, literalv, literalp);
        break;
      /*}}}*/
      /*{{{  S_UREALLIT */
      case S_UREALLIT:
        fprintf (fptr, "UREALLIT:%.*s ", literalp, literalv);
        break;
      /*}}}*/
      /*{{{  S_COMMENT */
      case S_COMMENT:
        fprintf(fptr, "-- %.*s", literalp, literalv);
        break;
      /*}}}*/
    }
}
/*}}}*/

/*{{{  PUBLIC void lexinit () */
PUBLIC void lexinit ( void )
{
  flocn = NOPOSN;
  set_vtilocn(&flocn);
  filestackptr = 0;
  baseindent = 0;
  currentfilenum = (-1);
  SetFileLine(flocn, 0);
  SetFileNum(flocn, currentfilenum);
  allow_asmnames = FALSE;
  totallines = 0;
  if (linebuffer == NULL)
    linebuffer = memalloc(linesize + 2); /* 2 extra for "\n\0" */

  init_lex_error(&symb, &symbindent, &baseindent, &lexword, &literalp, literalv);

  /* predefines are now read in the checker */
}
/*}}}*/
/*{{{  PUBLIC void lexfinish */
PUBLIC void lexfinish(FILE *fptr)
{
  USE_VAR(fptr); /* stop unused variable warning */
}
/*}}}*/
/*{{{  PUBLIC int lex_totallines */
PUBLIC int lex_totallines(void)
{
  return totallines;
}
/*}}}*/
