/* $Id: syn2.c,v 1.13 1998/09/03 11:41:28 djb1 Exp $ */

/*
 *	Occam two syntax analyser 2
 *	Copyright (C) 1987 Inmos Limited
 *
 *	This program is free software; you can redistribute it and/or modify
 *	it under the terms of the GNU General Public License as published by
 *	the Free Software Foundation; either version 2 of the License, or
 *	(at your option) any later version.
 *
 *	This program is distributed in the hope that it will be useful,
 *	but WITHOUT ANY WARRANTY; without even the implied warranty of
 *	MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *	GNU General Public License for more details.
 *
 *	You should have received a copy of the GNU General Public License
 *	along with this program; if not, write to the Free Software
 *	Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 */

/*#define DEBUG*/

/*{{{  include files*/
#ifdef HAVE_CONFIG_H
#  include <config.h>
#endif
#include <stdio.h>
#include "feinc.h"
#include "lexerror.h"
#include "synerror.h"
#include "lexdef.h"
#include "syn1def.h"
#include "syn2def.h"
#include "syn3def.h"
#include "syndef.h"
#include "chkdef.h" /* current_fe_handle etc */
#include "objrdr.h" /* needed for addtoentrypointlist (correct declaration of library routines) */
#include "objlib.h" /* only needed cos objtrans.h requires it */
#include "objtrans.h" /* #PRAGMA TRANSLATE */
#include "predefhd.h" /* #PRAGMAs */
/*}}}*/

/*{{{  #defines*/
#define STRING_RECORD_SIZE 1000
/*}}}*/

/*{{{  forward declarations*/
PRIVATE treenode *rspecorexpr (BOOL *const specflag , const int pindent,
        const BOOL exprlist_permitted, const BOOL fn_permitted);
#ifdef OCCAM2_5
PRIVATE treenode *rspecifier_constr ( void );
PRIVATE treenode *rtypeoroperand(void);
PRIVATE treenode *rspecorexprlist(BOOL *specflag);
#endif
#ifdef USER_DEFINED_OPERATORS
extern BOOL user_defined_operators;
#endif

/*}}}*/

/*{{{  support*/
/*{{{  PUBLIC BOOL mustbespec(s)*/
PUBLIC BOOL mustbespec(const int s)
{
  switch (s)
    {
      case S_VAL: case S_PROC: case S_PROTOCOL: case S_CHAN:
      case S_PORT: case S_TIMER: case S_BOX: case S_PLACE:
      #ifdef RESOURCES
      case S_RESOURCE:
      #endif
      case S_INLINE:
      CASE_CONFIG_SPEC
      case S_PRAGMA: /* bug 829 19/9/91 */
      case S_HCOMMENT: /* bug 829 20/9/91 */
      case S_HNETWORK:
    #ifdef OCCAM2_5
      case S_DATA:
    #endif
        return TRUE;
      default :
        return FALSE;
   }
}
/*}}}*/
/*{{{  PRIVATE BOOL mustbeexp(s)*/
PRIVATE BOOL mustbeexp (const int s)
{
  switch (s)
    {
     case S_MOSTPOS: case S_MOSTNEG: case S_SUBTRACT: case S_MINUS:
     case S_BITNOT: case S_NOT: case S_SIZE: case S_LPAREN:
     case S_TRUE: case S_FALSE:
     case S_UBYTELIT: case S_UINTLIT: case S_UREALLIT:
     case S_STRING: case S_STRINGCONT:
     case S_VALOF:
   #ifdef OCCAM2_5
     case S_BYTESIN: case S_OFFSETOF:
   #endif
     #include "casemops.h"
     #include "casedops.h"
     /* may need dops too, Jim 20/2/97 */
        return TRUE;
      default :
        return FALSE;
   }
}
/*}}}*/
/*}}}*/
/*{{{  lists*/
/*{{{  PUBLIC treenode *rproclist (readitem, indent)*/
/*****************************************************************************
 *
 *  rproclist reads a list of items. 'readitem' is a function called to read
 *            each item.  Each item of the list must have the same indentation
 *            as the first item: a smaller indentation terminates the list,
 *            a greater indentation is an error.
 *
 *****************************************************************************/
PUBLIC treenode *rproclist (treenode *(*readitem)(void), const int indent)
{
  treenode *root = NULL;
  treenode **pptr = &root;
  DEBUG_MSG(("rproclist... "));
  while (symbindent >= indent)
    /*{{{  create a list node and read another item*/
    {
      DEBUG_MSG(("rproclist: looking for next process, symb:symbindent is %s:%d...\n",
                 tagstring(symb), symbindent));
      switch (symb)
        {
          case S_COMMENT:
            nextline();
            break;
          case S_INCLUDE: case S_SC: case S_USE: case S_IMPORT: /* added for bug 1208 18/10/91 */
            (void)rfile(); /* just loop even if there is an error */
            break;
          default:
            if (symbindent > indent)
              /*{{{  error so skip to next process if any*/
              {
                synerr(SYN_BAD_INDENT, flocn);
                skiplines (indent);
              }
              /*}}}*/
            else
              /*{{{  parse a process item, and append to list*/
              {
                const SOURCEPOSN locn = flocn;
                treenode *const item = (*readitem)();
                if (item != NULL)
                  /*{{{  make new node*/
                  {
                    *pptr = newlistnode (S_LIST, locn, item, NULL);
                    pptr = NextItemAddr(*pptr);
                  }
                  /*}}}*/
              }
              /*}}}*/
            break;
        }
    }
    /*}}}*/
  DEBUG_MSG(("rproclist: returning NOW... "));
  return (root);
}
/*}}}*/
/*{{{  PUBLIC treenode *rlist (p, sep)*/
/*****************************************************************************
 *
 *  rlist reads a list of items.  'p' is a function which reads each item,
 *        and 'sep' is the token which separates each item.
 *        ie. parse   p { sep p }
 *
 *****************************************************************************/

/* may return NULL if error found */
PUBLIC treenode *rlist ( treenode *(*p)(void), const int sep )
{
  treenode *tptr, *a, *item;
  SOURCEPOSN locn = flocn;
  DEBUG_MSG(("rlist... "));
  /*{{{  parse item*/
  if ((item = (*p)()) == NULL)
    return NULL;
  /*}}}*/
  tptr = newlistnode(S_LIST, locn, item, NULL);
  a = tptr;
  while (symb == sep)
    /*{{{  do another item*/
    {
      nextsymb ();
      /*{{{  ignore line break*/
      if (checklinebreak ())
        return NULL;
      /*}}}*/
      /*{{{  parse item*/
      if ((item = (*p)()) == NULL)
        return NULL;
      /*}}}*/
      NewNextItem(newlistnode (S_LIST, locn, item, NULL), a);
      a = NextItem(a);
      locn = flocn;
    }
    /*}}}*/
  return (tptr);
}
/*}}}*/
/*}}}*/




/* expression parsers may return NULL if an error is found , */
/* but will not skip to the end of a line,                   */
/* except when a VALOF is parsed.                            */
/*{{{  expressions*/
/*{{{  PRIVATE wordnode *rnameorstring (typelist, RETURNED overload, num_of_params)*/
/*{{{  COMMENT*/
 /*This is a modified version of the rname - it checks for a function being*/
 /*declared surrounded by quotes - eg "+", and then generates a valid string*/
 /*from that symbol - eg "+" -> udo.ADD. The caller then adds the type names*/
 /*(as strings) to this to construct the overloading functions name.*/
 /*overload becomes set if it is an overloaded function, and num_of_params*/
 /*is set to the number of parameters expected by that operator - if it can*/
 /*be either one or two (in the case of "-") then is becomes -1.*/
/*}}}*/
#ifdef USER_DEFINED_OPERATORS
/*{{{  forward declaration*/
/* forward delcaration */
PRIVATE char *param_type_string ( treenode *typetree, size_t *length );
/*}}}*/

PRIVATE wordnode *rnameorstring (treenode* typelist, BOOL* overload, int* num_of_params )
{
  if (!user_defined_operators)
    return (rname());
  DEBUG_MSG(("rnameorstring... "));
  if ((symb == S_NAME) || (symb == S_ASMNAME))
    {
      wordnode *const name = lexword;
      nextsymb ();
      *overload = FALSE;
      return (name);
    }
  else
  {
  if (TagOf(typelist)==S_LIST)
  {
    synerr_e(SYN_UDO_MULTI_RET_PARAMS, flocn, symb);
    return NULL;
  }
    /* this next section of code converts a "+" or whatever as a function name */
  /* into a wordnode with a pseudo name - ie "+" -> udo.ADD. */
  if (symb == S_STRING) /* deal with overloaded operators */
  {                     /* added by Jim Moores 18/10/96 */
      /*{{{  allocate memory to string*/
      char *name = (char *)malloc(256); /* should change to a constant */
      /*}}}*/
      int namelen = 0;
      wordnode *name_word = NULL;
      wordnode *wordtocompare = lookupword(literalv, literalp);
      /*{{{  zero string memory*/
      int index=0;
      for(index=0; index<256; name[index++]=0);
      /*}}}*/
      /*{{{  is it a binary operator?  Make string for it*/
      if (wordtocompare==lookupword("+", 1))
        name = strcpy(name,"udo.ADD.\0");
      if (wordtocompare==lookupword("/", 1))
        name = strcpy(name,"udo.DIV.\0");
      if (wordtocompare==lookupword("*", 1))
        name = strcpy(name,"udo.MUL.\0");
      if (wordtocompare==lookupword("\\", 1))
        name = strcpy(name,"udo.REM.\0");
      if (wordtocompare==lookupword("PLUS",4))
        name = strcpy(name,"udo.PLU.\0");
      if (wordtocompare==lookupword("TIMES",5))
        name = strcpy(name,"udo.TIM.\0");
      if (wordtocompare==lookupword("AFTER",5))
        name = strcpy(name,"udo.AFT.\0");
      if (wordtocompare==lookupword("AND",3))
        name = strcpy(name,"udo.AND.\0");
      if (wordtocompare==lookupword("OR",2))
        name = strcpy(name,"udo.ORo.\0");
      if (wordtocompare==lookupword("><",2))
        name = strcpy(name,"udo.XOR.\0");
      if (wordtocompare==lookupword("/\\",2))
        name = strcpy(name,"udo.BAN.\0");
      if (wordtocompare==lookupword("\\/",2))
        name = strcpy(name,"udo.BOR.\0");
      if (wordtocompare==lookupword(">>",2))
        name = strcpy(name,"udo.RSH.\0");
      if (wordtocompare==lookupword("<<",2))
        name = strcpy(name,"udo.LSH.\0");
      if (wordtocompare==lookupword("=",1))
        name = strcpy(name,"udo.EQA.\0");
      if (wordtocompare==lookupword("<>",2))
        name = strcpy(name,"udo.NEQ.\0");
      if (wordtocompare==lookupword("<",1))
        name = strcpy(name,"udo.LTH.\0");
      if (wordtocompare==lookupword(">",1))
        name = strcpy(name,"udo.GRT.\0");
      if (wordtocompare==lookupword("<=",2))
        name = strcpy(name,"udo.LTE.\0");
      if (wordtocompare==lookupword(">=",2))
        name = strcpy(name,"udo.GTE.\0");
      /*}}}*/
      if (name[0] != '\0') /* match found */
        *num_of_params = 2;
      else
      {
        /*{{{  is it a unary operator?  Make string for it*/
        if (wordtocompare==lookupword("~",1))
          name = strcpy(name,"udo.BNT.\0");
        if (wordtocompare==lookupword("NOT",3))
          name = strcpy(name,"udo.NOT.\0");
        /*}}}*/
        if (name[0] != '\0') /* match found */
          *num_of_params = 1;
        else
        {
          /*{{{  is it an operator that can be either?  Make string for it*/
          if (wordtocompare==lookupword("-",1)) /* Special case, "-" can be monadic or diadic */
            name = strcpy(name,"udo.SUB.\0");
          if (wordtocompare==lookupword("MINUS",5))
            name = strcpy(name,"udo.MIN.\0");
          if (wordtocompare==lookupword("??",2))
            name = strcpy(name,"udo.QMQ.\0");
          if (wordtocompare==lookupword("@@",2))
            name = strcpy(name,"udo.ATA.\0");
          if (wordtocompare==lookupword("$$",2))
            name = strcpy(name,"udo.DOD.\0");
          if (wordtocompare==lookupword("%",1))
            name = strcpy(name,"udo.PER.\0");
          if (wordtocompare==lookupword("%%",2))
            name = strcpy(name,"udo.PPR.\0");
          if (wordtocompare==lookupword("&&",2))
            name = strcpy(name,"udo.ANA.\0");
          if (wordtocompare==lookupword("<%",2))
            name = strcpy(name,"udo.LPR.\0");
          if (wordtocompare==lookupword("%>",2))
            name = strcpy(name,"udo.PRR.\0");
          if (wordtocompare==lookupword("<&",2))
            name = strcpy(name,"udo.LAN.\0");
          if (wordtocompare==lookupword("&>",2))
            name = strcpy(name,"udo.ANR.\0");
          if (wordtocompare==lookupword("[>",2))
            name = strcpy(name,"udo.RTR.\0");
          if (wordtocompare==lookupword("<]",2))
            name = strcpy(name,"udo.RTL.\0");
          if (wordtocompare==lookupword("++",2))
            name = strcpy(name,"udo.PLP.\0");
          if (wordtocompare==lookupword("^",1))
            name = strcpy(name,"udo.HAT.\0");
          if (wordtocompare==lookupword("!!",2))
            name = strcpy(name,"udo.EXE.\0");
          if (wordtocompare==lookupword("==",2))
            name = strcpy(name,"udo.EQE.\0");
          if (wordtocompare==lookupword("<@",2))
            name = strcpy(name,"udo.LAT.\0");
          if (wordtocompare==lookupword("@>",2))
            name = strcpy(name,"udo.ATR.\0");
          if (wordtocompare==lookupword("@",1))
            name = strcpy(name,"udo.AMA.\0");
          /*}}}*/
          if (name[0] != '\0') /* match found */
            *num_of_params = -1;
          else
          {
            /*{{{  otherwise its an error*/
            synerr_e(SYN_UDO_NOT_SUPPORTED,flocn,symb);
            return NULL;
            /*}}}*/
          }
        }
      }
      /*{{{  make wordnode and return it (with overload flag set)*/
      namelen = 8;
      name_word = lookupword(name,namelen);
      nextsymb ();
      *overload = TRUE;
      return (name_word);
      /*}}}*/
    }
  else
    {
      synerr_e (SYN_E_NAME, flocn, symb);
      return NULL;
    }
  }
}
#endif
/*}}}*/
/*{{{  PUBLIC wordnode *rname ()*/
PUBLIC wordnode *rname ( void )
{
  DEBUG_MSG(("rname... "));
  if ((symb == S_NAME) || (symb == S_ASMNAME))
    {
      wordnode *const name = lexword;
      nextsymb ();
      return (name);
    }
  else
    {
      synerr_e (SYN_E_NAME, flocn, symb);
      return NULL;
    }
}
/*}}}*/
/*{{{  PUBLIC treenode *rsegment(node)*/
PUBLIC treenode *rsegment ( treenode *const node )
{
  /*{{{  parsing*/
  /*
    the '[' and element have already been parsed,
    node is the element we are taking the slice of :

      'FROM' exp 'FOR' exp ']'

    occam2.5: permits the dropping of 'FROM' or 'FOR' expression,
              but NOT both
  */
  /*}}}*/
  treenode *start  = NULL;
  treenode *length = NULL;
  const SOURCEPOSN locn = flocn;
#ifdef OCCAM2_5
  if (symb != S_FOR)
    {
      if (checkfor(S_FROM) ||
          checklinebreak() ||
          ((start = rexp()) == NULL))
        return NULL;
    }
  if (symb != S_RBOX)
    {
      if (checkfor(S_FOR)  ||
          checklinebreak() ||
          ((length = rexp()) == NULL))
        return NULL;
    }
  if (checkfor(S_RBOX))
    return NULL;
#else
  if (checkfor(S_FROM)  ||
      checklinebreak () ||
      ((start = rexp ()) == NULL) ||
      checkfor (S_FOR)  ||
      checklinebreak () ||
      ((length = rexp ()) == NULL) ||
      checkfor (S_RBOX) )
    return NULL;
#endif
  return newsegmentnode (S_SEGMENT, locn, node, start, length);
}
/*}}}*/
/*{{{  PRIVATE treenode *rendofconstructor*/
PRIVATE treenode *rendofconstructor(treenode *const node, const SOURCEPOSN locn)
{
  treenode *type = NULL;
#ifdef OCCAM2_5
  if (symb == S_LPAREN)
    /*{{{  type decoration*/
    {
      nextsymb();
      type = (treenode *)rname();
      if (checkfor(S_RPAREN)) return NULL;
    }
    /*}}}*/
#endif
  return newlitnode(S_CONSTRUCTOR, locn, node, type);
}
/*}}}*/
/*{{{  PRIVATE treenode *rconstructor(node)*/
PRIVATE treenode *rconstructor ( treenode *node )
{
  /*{{{  parsing*/
  /*
    The opening box & first element have been parsed,
    node is is the first element.
    we do :
             {',' exp} ']'
  */
  /*}}}*/
  if (symb == S_COMMA)
    {                                                 /* read list of items */
      treenode *list;
      nextsymb ();
      if (checklinebreak ())
        return NULL;
      if ((list = rlist (rexp, S_COMMA)) == NULL)
        return NULL;
      node = newlistnode (S_LIST, flocn, node, list);
    }
  else
    node = newlistnode(S_LIST, flocn, node, NULL);     /* one item list      */
  if (checkfor (S_RBOX))
    return NULL;
  return rendofconstructor(node, flocn);
}
/*}}}*/
/*{{{  PRIVATE treenode *rarrayconstructor()*/
#ifdef ARRAYCONSTRUCTOR
PRIVATE treenode *rarrayconstructor ( void )
{
  /*{{{  parsing*/
  /*
    the '[' has been parsed,

      'ARRAY' name 'FROM' exp 'FOR' exp 'OF' exp ']'

  */
  /*}}}*/
  treenode *nptr, *rcptr, *start, *length, *val;

  if (checkfor(S_ARRAYCONSTRUCTOR) ||
      rrepl(&nptr, &start, &length) ||
      checkfor (S_OF) ||
      checklinebreak () ||
      ((val = rexp ()) == NULL) ||
      checkfor (S_RBOX))
    return NULL;

  rcptr = newreplcnode (S_ARRAYCONSTRUCTOR, flocn, nptr, start, length, val);
  SetNDecl(nptr, rcptr);
  return rcptr;
}
#endif

/*}}}*/
/*{{{  PUBLIC treenode *rsubscript (a)*/
/* The element built up so far */
/* may return NULL */
/* Parse
'[' expression ']' { '[' expression ']' }
*/
PUBLIC treenode *rsubscript ( treenode *a )
{
  SOURCEPOSN locn = flocn;
  treenode *exp;
  DEBUG_MSG(("rsubscript... "));
  if (symb != S_LBOX)
    {
      synetoken (S_LBOX);
      return NULL;
    }
  while (symb == S_LBOX)
    {
      nextsymb();
      if ((exp = rexp ()) == NULL)
        return NULL;
      if (checkfor (S_RBOX))
        return NULL;
      a = newarraysubnode (S_ARRAYSUB, locn, a, exp);
      /* if (symb == S_LBOX) a = rsubscript (a); */
    }
  return (a);
}
/*}}}*/
/*{{{  PUBLIC treenode *relement ()*/
/* may return NULL if error found */
PUBLIC treenode *relement ( void )
{
  SOURCEPOSN locn = flocn;
  treenode *a;

  DEBUG_MSG(("relement... "));
  if (symb == S_NAME)
    a = (treenode *)rname(); /* name or subscripted element */
  else if (symb == S_LBOX)
    /*{{{  segment*/
    {
      treenode *name;
      nextsymb ();
      if ((name = relement ()) == NULL)
        return NULL;
      a = rsegment (name);
    }
    /*}}}*/
  else
    {
      synerr_e (SYN_E_ELEMENT, locn, symb);
      return NULL;
    }

  if (a != NULL && symb == S_LBOX)
    a = rsubscript (a);

  return (a);
}
/*}}}*/
/*{{{  PUBLIC treenode *rinstance (tag, locn, name)*/
PUBLIC treenode *rinstance ( int tag , SOURCEPOSN locn , treenode *name )
{
  /*{{{  parsing*/
  /*
    Assume the name has already been read
    '(' exp {',' exp } ')'
   */
  /*}}}*/
  treenode *params;
  if (checkfor (S_LPAREN))
    return NULL;
  if (symb != S_RPAREN)
    {
      /* bug TS/1487 - remove all these bits from here! 26/11/91 */
      /*const int saved_indent = linebreakindent;*/ /* bug TS/1457 11/11/91 */
      /*linebreakindent = (-1);*/ /* bug TS/1251 25/9/91 */ /* removed 26/11/91 */
      params = rlist(rexp, S_COMMA);
      /*linebreakindent = saved_indent;*/ /* added 11/11/91 */ /* removed 26/11/91 */
      if (params == NULL)
        return NULL;
    }
  else
    params = NULL;
  if (checkfor (S_RPAREN))
    return NULL;
  return newinstancenode (tag, locn, name, params);
}
/*}}}*/
/*{{{  PRIVATE treenode *rrestofoperand*/
PRIVATE treenode *rrestofoperand(treenode *tptr, const SOURCEPOSN locn)
{
  if (symb == S_LPAREN)
    tptr = rinstance(S_FINSTANCE, locn, tptr);
#ifndef OCCAM2_5
  else if (symb == S_LBOX)
      tptr = rsubscript(tptr);
#endif
#ifdef OCCAM2_5
  if (symb == S_LBOX)
  /*  We expect either name[exp]... or type conversion of
      the form name table[exp]. We have parsed already name and
      the result is held in tptr  */
  {
    treenode *exp;
    nextsymb ();
    if ((exp = rexp ()) == NULL) return NULL;
    if (symb == S_RBOX && TagOf(exp) != S_CONSTRUCTOR )
    {
      nextsymb ();
      tptr = newarraysubnode (S_ARRAYSUB, locn, tptr, exp);
      if (symb == S_LBOX)
        tptr = rsubscript (tptr);
      return tptr;
    }
    /* type conversion of the form: name table[exp] bug INSdi03578.
       We have parsed already name and the result is held in tptr */
    else if (symb == S_RBOX && TagOf(exp) == S_CONSTRUCTOR )
    {
    /* A constructor with one  element which is a constructor. */
      nextsymb ();
      exp = rendofconstructor(newlistnode(S_LIST, flocn, exp, NULL), flocn);
    }
    else
      exp = rconstructor(exp);
    if (exp == NULL)
      return NULL;
    else if (symb == S_LBOX)
      exp = rsubscript (exp);
    if (tptr != NULL)
    {
      tptr = newlistnode(S_LIST, locn, tptr,
                         newlistnode(S_LIST, locn, exp, NULL));
    }
    tptr = newmopnode(S_EXACT, locn, tptr, S_NAME);
  }
#endif
  return tptr;
}
/*}}}*/
/*{{{  PRIVATE treenode *rvalof ()*/
/* Terminates with symb at start of new line */
/*{{{  what we are parsing*/
/*  Parse
valof = 'VALOF'
process
'RESULT' expression.list

including the newline following the expression list,
or
valof = specification
valof
*/
/*}}}*/
PRIVATE treenode *rvalof ( void )
{
  int indent = symbindent;              /* The indent of the 'VALOF' keyword */
  int savedlinebreakindent;
  SOURCEPOSN locn = flocn;
  treenode *p, *n, *e;

  DEBUG_MSG(("rvalof... "));
  /*{{{  VALOF*/
  if (checkfor (S_VALOF))
    goto error;
  if (checknlindent (indent + 2))
    goto error2;
  /*}}}*/
  /*{{{  PROCESS*/
  savedlinebreakindent = linebreakindent;
  linebreakindent = (-1);
  if (((p = rprocess()) == NULL) || (checkindent (indent + 2)))
    goto error2;
  ignorecomments(indent + 2);
  /*}}}*/
  /*{{{  RESULT list*/
  if (checkfor (S_RESULT) || ((e = rlist (rexp, S_COMMA)) == NULL))
    goto error;
  /*}}}*/
  n = newvalofnode (S_VALOF, locn, p, e);
  checknewline ();
  linebreakindent = savedlinebreakindent;
  return (n);

error:
  nextline ();
error2:
  skiplines (indent);
  return NULL;
}
/*}}}*/
/*{{{  PRIVATE treenode *rcondexp(node)*/
#ifdef CONDEXP
PRIVATE treenode *rcondexp ( treenode *node )
{
  /*{{{  parsing*/
  /*
      the boolean expression has been parsed already :
          '->' exp ',' exp
        | '->' exp ',' conditional.expression
   */
  /*}}}*/
  treenode *true, *false;
  if (checkfor (S_RIGHTARROW) ||
      checklinebreak () ||
      ((true = rexp ()) == NULL) ||
      checkfor (S_COMMA) ||
      checklinebreak () ||
      ((false = rexp ()) == NULL))
    return NULL;
  if (symb == S_RIGHTARROW)
    false = rcondexp(false);
  return newcondexpnode (S_CONDEXP, flocn, node, true, false);
}
#endif

/*}}}*/
/*{{{  PRIVATE treenode *rstring*/
PRIVATE treenode *rstring()
/* This routine reads a string, allowing for long strings.
   It does NOT advance the symbol to the next symbol after the string
*/
{
  treenode *node;
  SOURCEPOSN locn = flocn;
  if (symb == S_STRING)
    {
      literalv[literalp] = '\0';
      node = newconsttablenode(S_STRING, locn, lookupword(literalv, literalp), NULL);
    }
  else /* (symb == S_STRINGCONT) */
    {
      BOOL more = TRUE;
      int len = literalp;
      int stringsize = STRING_RECORD_SIZE;
      char *string = memalloc(stringsize);
      memcpy(string, literalv, literalp);
      nextsymb();
      while (more)
        {
          if ((len + literalp) > stringsize)
            /*{{{  make string a bit bigger*/
            {
              char *newstring;
              stringsize += STRING_RECORD_SIZE;
              newstring = memalloc(stringsize);
              memcpy(newstring, string, len);
              memfree(string);
              string = newstring;
            }
            /*}}}*/
          memcpy(string + len, literalv, literalp);
          len += literalp;
          locn = flocn;
          if (symb == S_STRINGCONT)
            nextsymb();
          else /* symb == S_STRING */
            more = FALSE;
        }
      node = newconsttablenode(S_STRING, locn, lookupword(string, len), NULL);
      memfree(string);
    }
  return node;
}
/*}}}*/
/*{{{  PRIVATE treenode *roperand ()*/
/* On error, leave symb unchanged, except when VALOF found */
/*{{{  what we are parsing*/
/* Parse
'(' expression ')'
| '(' valof ')'
| '(' conditional.expression ')'
| literal
| element
| '[' expression { ',' expression } ']'
| name '(' expression { ',' expression } ')'
*/
/*}}}*/
PRIVATE treenode *roperand ( void )
{
  const SOURCEPOSN locn = flocn;
  treenode *node = NULL;
  DEBUG_MSG(("roperand... "));
  switch (symb)
    {
      /*{{{  case S_LPAREN could be expression, specification or valof*/
      case S_LPAREN:
        {
          BOOL specflag = FALSE;
          treenode **a = &node;
          int pindent = symbindent;
          int indent;
          nextsymb ();
          indent = symbindent; /* The indent of first symbol after left paren */
          /*{{{  read spec or expression into *a*/
          {
            const BOOL fn_permitted = ((current_fe_data->fe_lang & FE_LANG_NDL) == 0);
            *a = rspecorexpr (&specflag, pindent, FALSE, fn_permitted);
            if (*a == NULL)
              if (!specflag)
                return NULL;  /* previous syntax error */
          }
          /*}}}*/
          if (specflag)
            /*{{{  read further specifications, a valof, and check it*/
            {
              while (symb != S_VALOF && symbindent == indent)
                {
                  if (*a != NULL) a = DBodyAddr(*a);  /* move a along list */
                  while (symb == S_COMMENT)           /* skip comments    */
                    if (checknlindent (indent))
                      return NULL;
                  *a = rspecification ();             /* parse expression  */
                }
              if (symb != S_VALOF)                    /* Check for valof   */
                {
                  synetoken(S_VALOF);
                  return NULL;
                }
              if (checkindent (indent))
                return NULL;
              if (*a != NULL)
                a = DBodyAddr(*a);                       /* move a along list   */
              if ((*a = rvalof ()) == NULL)
                return NULL;                             /* parse VALOF         */
              ignorecomments(pindent);                   /* throw away comments */
            }
            /*}}}*/
          #ifdef CONDEXP
          if (symb == S_RIGHTARROW)
            *a = rcondexp(*a);
          #endif
          if (checkfor (S_RPAREN)) return NULL;
        }
        break;
      /*}}}*/
      /*{{{  case S_TRUE FALSE*/
      case S_TRUE: case S_FALSE:
        node = newleafnode(symb, locn);
        nextsymb ();
        break;
      /*}}}*/
      /*{{{  S_UINTLIT S_UREALLIT S_UBYTELIT*/
      case S_UINTLIT: case S_UREALLIT: case S_UBYTELIT:
        {
          const int s = symb;
          treenode *ss = NULL;
          wordnode *w = lookupword(literalv, literalp);
          nextsymb ();
          if (symb == S_LPAREN)
            /*{{{  we have a type specifier*/
            { /* '(' simple type ')' */
              nextsymb ();
            #if 0
              if (((s == S_UINTLIT || s == S_UBYTELIT) &&
                     (symb == S_INT   || symb == S_BYTE || symb == S_INT64 ||
                        symb == S_INT32 || symb == S_INT16)) ||
                  (s == S_UREALLIT && (symb == S_REAL32 || symb == S_REAL64)))
                /*s = littag(symb);*/
                ss = newleafnode(symb, locn);
              else
                {
                  synerr(SYN_ILLEGAL_CONVERSION, locn);
                  return NULL;
                }
              nextsymb ();
            #else
              ss = rspecifier();
              if (ss == NULL) return NULL;
            #endif
              if (checkfor (S_RPAREN))
                return NULL;
            }
            /*}}}*/
          node = newlitnode (s, locn, (treenode *)w, ss);
        }
        break;
      /*}}}*/
      /*{{{  case S_STRING, S_STRINGCONT*/
      case S_STRING: case S_STRINGCONT:
        node = rstring();
        nextsymb();
      #ifdef OCCAM2_5
        if (symb == S_LPAREN)
        /*{{{  type decoration*/
        {
          treenode *type = NULL;
          nextsymb();
          type = (treenode *)rname();
          if (checkfor(S_RPAREN)) return NULL;
          SetConstTableType(node, type);
        }
        /*}}}*/
      #endif
        if (symb == S_LBOX) node = rsubscript(node);
        break;
      /*}}}*/
      /*{{{  case S_NAME*/
      case S_NAME: case S_ASMNAME:
        node = rrestofoperand((treenode *)rname(), locn);
        break;
      /*}}}*/
      /*{{{  case S_LBOX*/
      case S_LBOX:
        nextsymb ();
      #ifdef ARRAYCONSTRUCTOR
        if (symb == S_ARRAYCONSTRUCTOR)
          node = rarrayconstructor();
        else
      #endif
          {
            if ((node = rexp ()) == NULL)
              return NULL;
          #ifdef OCCAM2_5
            if ((symb == S_FROM) || (symb == S_FOR))
          #else
            if (symb == S_FROM)
          #endif
              node = rsegment(node);
            else
              node = rconstructor(node);
          }
        if (node != NULL && symb == S_LBOX)
          node = rsubscript (node);
        break;
      /*}}}*/
      /*{{{  BYTESIN and OFFSETOF*/
      #ifdef OCCAM2_5
      case S_BYTESIN:
        {
          treenode *op;
          nextsymb ();
          if (checklinebreak ()
           || checkfor(S_LPAREN)
           || checklinebreak()
           || ((op = rtypeoroperand()) == NULL)
           || checkfor(S_RPAREN))
            return NULL;
          node = newmopnode (S_BYTESIN, locn, op, 0);
        }
        break;
      case S_OFFSETOF:
        {
          wordnode *name;
          wordnode *field;
          nextsymb ();
          if (checklinebreak ()
           || checkfor(S_LPAREN)
           || checklinebreak())
            return NULL;
          name = rname();
          if (checkfor(S_COMMA)
           || checklinebreak()) return NULL;
          field = rname();
          if (checkfor(S_RPAREN)) return NULL;
          node = newdopnode(S_OFFSETOF, locn,
                 (treenode *)name, (treenode *)field, 0);
        }
        break;
      #endif
      /*}}}*/
      default:
        synerr_e (SYN_E_OPERAND, flocn, symb);
        break;
    }
#ifdef OCCAM2_5
  if (node != NULL && symb == S_LBOX)
    node = rsubscript(node);
#endif
  return node;
}
/*}}}*/
/*{{{  PRIVATE treenode *rconversion*/
PRIVATE treenode *rconversion(const int type_tag, treenode *const type, const SOURCEPOSN locn)
{
  treenode *op;
  treenode *conversion;
  int tag = S_EXACT;

  if ((symb == S_ROUND) || (symb == S_TRUNC))
    {
      tag = symb;
      nextsymb ();
      /*{{{  ignore line break*/
      if (checklinebreak ())
        return NULL;
      /*}}}*/
    }
  if ((op = roperand ()) == NULL)
    return NULL;
  if (type != NULL)
    {
      op = newlistnode(S_LIST, locn, type,
           newlistnode(S_LIST, locn, op, NULL));
    }
  conversion = newmopnode(tag, locn, op, type_tag);
  return (conversion);
}
/*}}}*/
/*{{{  PRIVATE treenode *rrestofexp (a, locn)*/
/* May return NULL if error found */
/* Parse
{ dyadic.op operand }
*/
PRIVATE treenode *rrestofexp ( treenode *a , SOURCEPOSN locn )
{
  BOOL assoc = TRUE;
  int assoc_symb = S_END; /* anything except S_AND or S_OR */ /* bug 1361 19/9/91 */
  /*{{{  build up expression*/
  while (assoc)
    {
      treenode *operand;
      const int operator = symb;
      switch (symb)
        {
          /*{{{  S_AND S_OR*/
          case S_AND:
          case S_OR:
            if (assoc_symb == S_END) /* first time through */ /* bug 1361 19/9/91 */
              assoc_symb = symb;
            else if (symb != assoc_symb) /* mixtures of AND and OR */ /* bug 1361 19/9/91 */
              return (a); /* do the same as the default case */
            assoc = TRUE;
            break;
          /*}}}*/
          /*{{{  ADD SUBTRACT MULT DIV REM BITAND BITOR XOR LSHIFT RSHIFT PLUS...*/
          case S_ADD: case S_SUBTRACT: case S_MULT: case S_DIV: case S_REM:
          case S_BITAND: case S_BITOR: case S_XOR:
          case S_LSHIFT: case S_RSHIFT:
          case S_PLUS: case S_MINUS: case S_TIMES:
          case S_EQ: case S_NE: case S_LS: case S_LE: case S_GR: case S_GE: case S_AFTER:
          #include "casedops.h" /* added by Jim 2/2/97 */
            assoc = FALSE;
            break;
          /*}}}*/
          default:
            return (a);
        };
      nextsymb ();
      if (checklinebreak ())
        return NULL;
      if ((operand = roperand ()) == NULL)
        return NULL;
      a = newdopnode (operator, locn, a, operand, 0);
    };
  /*}}}*/
  return (a);
}
/*}}}*/
/*{{{  PRIVATE treenode *rtypeoroperand*/
#ifdef OCCAM2_5
PRIVATE treenode *rtypeoroperand(void)
{
  const SOURCEPOSN locn = flocn;

  if (istypetag(symb) || symb == S_BOX)
    return rspecifier();

  if (symb == S_LBOX)
    /*{{{  S_LBOX    we don't know which*/
    {
      treenode *spec;
      treenode *e;
      nextsymb ();
      if ((e = rexp ()) == NULL) return NULL;
      if (symb == S_RBOX)
        {
          nextsymb();
          if ((spec = rspecifier_constr ()) == NULL)
          /* One element constructor */
          {
            e = rendofconstructor(newlistnode(S_LIST, flocn, e, NULL), flocn);
            return e;
          }
          else
            return (newtypenode (S_ARRAY, locn, e, spec));
        }
      if ((symb == S_FROM) || (symb == S_FOR)) /* Note we know we're in OCCAM2_5 */
        e = rsegment(e);
      else
        e = rconstructor(e);
      if (e != NULL && symb == S_LBOX)
        e = rsubscript (e);
      return e;
    }
    /*}}}*/

  return roperand();
}
#endif
/*}}}*/
/*{{{  PUBLIC  treenode *rexp ()*/
/* Expressions may be broken after an expression operator - the indentation
of the first symbol on the continuation line must be at least as great
as that of the first line.
*/
PUBLIC treenode *rexp ( void )
{
  const SOURCEPOSN locn = flocn;
  DEBUG_MSG(("rexp... "));

  switch (symb)
    {
      /*{{{  case S_MOSTPOS S_MOSTNEG*/
      case S_MOSTPOS:
      case S_MOSTNEG:
        {
          const int s = symb;
        #if 0
          /*{{{  old version*/
          nextsymb ();
          switch (symb)
            {
              case S_BYTE: case S_INT: case S_INT16: case S_INT32: case S_INT64:
                {
                  const int t = symb;
                  nextsymb ();
                  return (newmopnode (s, locn, newleafnode(t, locn), 0));
                }
            #ifdef OCCAM2_5
              case S_NAME:
                return newmopnode(s, locn, (treenode *)rname(), 0);
            #endif
              default:
                synerr_e (SYN_E_INTTYPE, locn, symb);
                return NULL;
            };
          /*}}}*/
        #else
          treenode *op;
          nextsymb();
          op = rspecifier();
          return (op == NULL) ? NULL : newmopnode(s, locn, op, 0);
        #endif
        }
      /*}}}*/
      /*{{{  case S_BOOL S_BYTE S_INT S_INTn S_REALn*/
      case S_BOOL: case S_BYTE:
      case S_INT: case S_INT16: case S_INT32: case S_INT64:
      case S_REAL32: case S_REAL64:
        {
          const int s = symb;
          nextsymb();
          return rconversion(s, NULL, locn);
        }
      /*}}}*/
      /*{{{  LBOX TRUE FALSE BYTELIT UINTLIT UREALLIT STRING STRINGCONT LPAREN*/
      case S_LBOX: case S_TRUE: case S_FALSE: case S_ASMNAME:
      case S_UBYTELIT: case S_UINTLIT: case S_UREALLIT:
      case S_STRING: case S_STRINGCONT: case S_LPAREN:
        {
          treenode *op = roperand ();
          return (op == NULL ? NULL : rrestofexp (op, locn));
        }
      /*}}}*/
      /*{{{  case S_NAME*/
      case S_NAME:
      #ifdef OCCAM2_5
        {
          treenode *op = (treenode *)rname();
          switch(symb)
            {
              case S_ROUND: case S_TRUNC:  /* obviously conversions */
              /*{{{  or beginning of an operand*/
              /* Anything which might start an operand is permitted here,
                 because we're looking at
                  expression = conversion
                             = name operand
                     or
                  expression = operand
                             = element | name ( expression )
              */

              case S_NAME:
              /*case S_LPAREN:*/ /* ambiguous, decode as fn instance, then decide
                                    later */
              /*}}}*/
      #if 0
              case S_LBOX:
      #endif
              case S_TRUE: case S_FALSE:
              case S_UINTLIT: case S_UREALLIT: case S_UBYTELIT:
              case S_STRING: case S_STRINGCONT:
                return rconversion(S_NAME, op, locn);
              default:
                /* do the same as 'roperand', except that we have already
                   read the first name */
                op = rrestofoperand(op, locn);
                return (op == NULL ? NULL : rrestofexp (op, locn));
            }
        }
      #else
        {
          treenode *op = roperand ();
          return (op == NULL ? NULL : rrestofexp (op, locn));
        }
      #endif

      /*}}}*/
      /*{{{  case S_SUBTRACT S_MINUS S_BITNOT S_NOT S_SIZE*/
      case S_SUBTRACT: case S_MINUS: case S_BITNOT: case S_NOT:
      #include "casedops.h"
        {
          treenode *op;
          int s = symb;
          if (s == S_SUBTRACT)
            s = S_NEG;
          else if (s == S_MINUS)
            s = S_UMINUS;
          /* added by Jim */
          #ifdef USER_DEFINED_OPERATORS
          switch (s)
          {
            case S_DQUESTIONM:
              s = S_M_DQUESTIONM;
              break;
            case S_DAMPERSAT:
              s = S_M_DAMPERSAT;
              break;
            case S_DDOLLAR:
              s = S_M_DDOLLAR;
              break;
            case S_PERCENT:
              s = S_M_PERCENT;
              break;
            case S_DPERCENT:
              s = S_M_DPERCENT;
              break;
            case S_DAMPERSAND:
              s = S_M_DAMPERSAND;
              break;
            case S_LEFTPERCENT:
              s = S_M_LEFTPERCENT;
              break;
            case S_PERCENTRIGHT:
              s = S_M_PERCENTRIGHT;
              break;
            case S_AMPERSANDRIGHT:
              s = S_M_AMPERSANDRIGHT;
              break;
            case S_LEFTAMPERSAND:
              s = S_M_LEFTAMPERSAND;
              break;
            case S_ROTATELEFT:
              s = S_M_ROTATELEFT;
              break;
            case S_ROTATERIGHT:
              s = S_M_ROTATERIGHT;
              break;
            case S_DPLUS:
              s = S_M_DPLUS;
              break;
            case S_HAT:
              s = S_M_HAT;
              break;
            case S_DEXCLAIMATION:
              s = S_M_DEXCLAIMATION;
              break;
            case S_DEQUALS:
              s = S_M_DEQUALS;
              break;
            case S_LEFTAMPERSAT:
              s = S_M_LEFTAMPERSAT;
              break;
            case S_AMPERSATRIGHT:
              s = S_M_AMPERSATRIGHT;
              break;
            case S_AMPERSAT:
              s = S_M_AMPERSAT;
              break;
          }
          #endif
          nextsymb ();
          if (checklinebreak ())
            return NULL;
          op = roperand ();
          return (op == NULL ? NULL : newmopnode (s, locn, op, 0));
        }
      case S_SIZE:
        {
          treenode *op;
          nextsymb ();
          if (checklinebreak ())
            return NULL;
        #ifdef OCCAM2_5
          op = rtypeoroperand();
        #else
          op = roperand ();
        #endif
          return (op == NULL ? NULL : newmopnode (S_SIZE, locn, op, 0));
        }
      /*}}}*/
      /*{{{  BYTESIN and OFFSETOF*/
      #ifdef OCCAM2_5
      case S_BYTESIN: case S_OFFSETOF:
        {
          treenode *op = roperand ();
          return (op == NULL ? NULL : rrestofexp (op, locn));
        }
      #endif
      /*}}}*/
      default:
        synerr_e (SYN_E_EXPR, locn, symb);
        return NULL;
    }
}
/*}}}*/

/*{{{  PRIVATE treenode *rtypeetc*/
#ifdef OCCAM2_5
PRIVATE treenode *rtypeetc(BOOL *const specflag, int *const error_code,
        const BOOL exprlist_permitted, const BOOL fn_permitted)
/*{{{  what we are parsing*/
/* This reads a line which has a type decl at the start
   (ie INT, BYTE, etc, or, more problematically, a name)
   and decides whether it is
     type TRUNC/ROUND operand
     type operand
     type FUNCTION etc
     type, type FUNCTION etc
     type name :
     type name, name :
     type name IS/RETYPES ...
     name IS ...
     name, exp, exp, ... <NL>   -- (CASE selection)

if exprlist_permitted is TRUE, we must allow for a list of expressions
  (eg selection list in a CASE)
if fn_permitted is TRUE, we can look for comma separated lists of names
   as fn definitions. We need this falg to be able to distinguish
   between names in a selection list, and named types in function definitions.

*/
/*}}}*/
{
  SOURCEPOSN locn = flocn;
  const int indent = symbindent;
  const int ttag = symb;
  const BOOL line_starts_with_name = ttag == S_NAME;
  treenode *type;

  *error_code = 0;

  /*{{{  read in user name/type, then move to next symbol*/
  if (line_starts_with_name)
    type = (treenode *)rname();
  else
    { nextsymb(); type = NULL; }
  /*}}}*/

  if ((symb == S_TRUNC) || (symb == S_ROUND))
    /*{{{  conversion  so expression*/
    {
      treenode *const conversion = rconversion(ttag, type, locn);
      if (conversion == NULL)
        { /*goto error;*/ *error_code = 1; return NULL; }
      return conversion;
    }
    /*}}}*/

  else if (line_starts_with_name && (symb == S_IS))
    /*{{{  it was an abbreviation*/
    {
      *specflag = TRUE;
      return (rrestofspec (NULL, type, locn, indent, TRUE));
    }
    /*}}}*/

  else if ((fn_permitted &&
           ((symb == S_COMMA) || (symb == S_FUNCTION) || (symb == S_INLINE)))
         || (exprlist_permitted && (symb == S_COMMA)))
    /*{{{  it must be a function definition, or a CASE selection list*/
    {
      BOOL local_specflag = FALSE;
      treenode *a = line_starts_with_name ? type : newleafnode(ttag, locn);

      if (symb == S_COMMA)
        /*{{{  read in the rest of the specifiers/expressions*/
        {
          treenode *list;
          nextsymb ();
          if (checklinebreak ())
            { /*goto error2;*/ *error_code = 2; return NULL; }
          list = (fn_permitted && exprlist_permitted)
                               ? rspecorexprlist(&local_specflag)
                : fn_permitted ? rlist(rspecifier, S_COMMA)
                               : rlist(rexp, S_COMMA);
          if (list == NULL)
            { /*goto error;*/  *error_code = 1; return NULL; }
          a = newlistnode (S_LIST, locn, a, list);
        }
        /*}}}*/
      if (fn_permitted && ((symb == S_FUNCTION) || (symb == S_INLINE)))
        {
          *specflag = TRUE;
          return (rfunctiondef (a, locn, indent));
        }
      else
        /*{{{  must be an expression list*/
        {
          return a; /* expression list */
        }
        /*}}}*/
    }
    /*}}}*/

  else if ((symb == S_NAME))
    /*{{{  could still be conversion or specification*/
    {
      treenode *list = NULL;
      wordnode *name = rname ();
      if (symb == S_COMMA)
        {
        #if 0 /* we can never accept a list of expressions here ! */
          if (explist)
            /*{{{  still don't know*/
            {
              nextsymb ();
              if (checklinebreak ())
                { /*goto error2;*/ *error_code = 2; return NULL; }
              if ((list = rlist (rexp, S_COMMA)) == NULL)
                { /*goto error;*/  *error_code = 1; return NULL; }
              /* we have type in ttag, first name in name, expression list in list */
            }
            /*}}}*/
          else
        #endif
            /*{{{  it must be a declaration*/
            {
              treenode *name_list;
              treenode *nptr;
              *specflag = TRUE;
              nextsymb ();
              if (checklinebreak ())
                { /*goto error2;*/ *error_code = 2; return NULL; }
              if ((name_list = rlist ((treenode *(*)())rname, S_COMMA)) == NULL)
                { /*goto error;*/  *error_code = 1; return NULL; }
              nptr = newlistnode (S_LIST, locn, (treenode *)name, name_list);
              return rrestofspec(line_starts_with_name ? type : newleafnode(ttag, locn),
                                  nptr, locn, indent, TRUE);
            }
            /*}}}*/
        }
      /* We have type in atag, name in 'name' */
      /*{{{  we have reached the IS COLON or RETYPES moment of truth for spec*/
      switch (symb)
        {
          /*{{{  next symbol is IS, RETYPES or COLON => specification*/
          case S_IS: case S_RETYPES: case S_COLON:
          case S_RESHAPES:
            *specflag = TRUE;
          #if 0
            if (list != NULL)
              list = newlistnode(S_LIST, locn, (treenode *)name, list);
            else
          #endif
              list = (treenode *)name;
            return rrestofspec(line_starts_with_name ? type : newleafnode(ttag, locn),
                               list, locn, indent, TRUE);
          /*}}}*/
          default:
          #if 0
            if (line_starts_with_name)
              /*{{{  error - "name name" doesn't parse*/
              {
                *specflag = TRUE;
                /* this supplies the correct error message */
                return rrestofspec(type, (treenode *)name, locn, indent, TRUE);
              }
              /*}}}*/
          #endif
            /*{{{  it was an expression after all*/
            {
              treenode *b;
              /*if (list == NULL)*/
                /*{{{  read the rest of this expression*/
                {
                  /* we have got to: expression = name name
                     now we're checking that the second name isn't part of
                     a larger operand
                  */
                  b = rrestofoperand((treenode *)name, locn);
                  if (b == NULL)
                    { /*goto error;*/  *error_code = 1; return NULL; }

                  if (line_starts_with_name)
                    {
                      b = newlistnode(S_LIST, locn, type,
                           newlistnode(S_LIST, locn, b, NULL));
                    }
                  b = newmopnode(S_EXACT, locn, b, ttag);
                }
                /*}}}*/
            #if 0 /* list is never non-NULL */
              else
                /*{{{  read a type conversion at beginning of list of exps*/
                {
                  b = (treenode *)name;
                #if 0
                  if (line_starts_with_name)
                    {
                      b = newlistnode(S_LIST, locn, type,
                           newlistnode(S_LIST, locn, b, NULL));
                    }
                #endif
                  b = newlistnode(S_LIST, locn,
                       newmopnode(S_EXACT, locn, b, ttag),
                       list);
                }
                /*}}}*/
            #endif
              return(b);
            }
            /*}}}*/
        }
      /*}}}*/
    }
    /*}}}*/

  else
    /*{{{  expression (possibly conversion)*/
    {
      BOOL is_conversion = !line_starts_with_name;
      /*{{{  check for start of another operand*/
      if (line_starts_with_name)
        {
          /*{{{  check for start of next operand*/
          switch(symb)
            {
              /* Anything which might start an operand is permitted here,
                 because we're looking at
                  expression = conversion
                             = name operand
                     or
                  expression = operand
                             = element | name ( expression )
              */

              case S_NAME:
              /*case S_LPAREN:*/ /* ambiguous, decode as fn instance, then decide
                                    later */
              case S_TRUE:    case S_FALSE:
              case S_UINTLIT: case S_UREALLIT: case S_UBYTELIT:
              case S_STRING:  case S_STRINGCONT:
                is_conversion = TRUE;
                break;
              default:
                break;
            }
          /*}}}*/
        }
      /*}}}*/

      if (is_conversion)
        /*{{{  it must be a conversion*/
        {
          treenode *const conversion = rconversion(ttag, type, locn);
          if (conversion == NULL)
            { /*goto error;*/ *error_code = 1; return NULL; }
          return (conversion);
        }
        /*}}}*/
      else
        /*{{{  it was an expression beginning with a name after all*/
        {
          treenode *const a = rrestofoperand(type, locn);
          if (a == NULL)
            { /*goto error;*/  *error_code = 1; return NULL; }
          return (rrestofexp (a, locn));
        }
        /*}}}*/
    }
    /*}}}*/
}
#endif
/*}}}*/
/*{{{  PRIVATE treenode *rspecorexpr (specflag, pindent)*/
/* If (specflag), leaves symb on next line, otherwise leaves symb unchanged */
/*{{{  note on pindent*/
/* Set to the indent of left parenthesis if we have been called
from roperand, negative if called from anywhere else
(rchoice or ralternative). pindent is only used if we find
a VALOF: if it is negative it means that VALOF is not legal in
this context.
*/
/*}}}*/
/*{{{  what we are parsing*/
/* Parse
expression
| specification
  VALOF
| VALOF
and return the parse tree.
*specflag is TRUE  if a specification was found,
FALSE if an expression was found.
if exprlist_permitted is TRUE, we must allow for a list of expressions
  (eg selection list in a CASE)
if fn_permitted is TRUE, we can look for comma separated lists of names
   as fn definitions. We need this falg to be able to distinguish
   between names in a selection list, and named types in function definitions.
*/
/*}}}*/
PRIVATE treenode *rspecorexpr ( BOOL *const specflag, const int pindent,
        const BOOL exprlist_permitted, const BOOL fn_permitted)
{
  SOURCEPOSN locn = flocn;
  const int indent = symbindent;
  *specflag = FALSE;
  switch (symb)
    {
      /*{{{  S_VALOF                              valof*/
      case S_VALOF:
        {
          treenode *valof;
          if ((valof = rvalof ()) == NULL)
            goto error2;
          if (pindent >= 0)
            {
              if (checkindent (pindent)) goto error2;
              while (symb == S_COMMENT)      /* Comments after result statement but*/
                if (checknlindent (pindent)) /* before right bracket               */
                  goto error2;
              return (valof);
            }
          else
            {
              synerr_e (SYN_E_EXPR, locn, S_VALOF);
              goto error2;
            }
        }
      /*}}}*/
    #ifndef OCCAM2_5
      /*{{{  S_BOOL S_BYTE S_INTn S_REALn         expression or specification*/
      case S_BOOL: case S_BYTE: case S_INT:
      case S_INT16: case S_INT32: case S_INT64: case S_REAL32: case S_REAL64:
        {
          const int atag = symb;
          nextsymb ();
          if ((symb == S_TRUNC) || (symb == S_ROUND))
            /*{{{  conversion  so expression*/
            {
              treenode *op, *conversion;
              int ct = symb;
              nextsymb ();
              /*{{{  ignore line break*/
              if (checklinebreak ())
                return NULL;
              /*}}}*/
              if ((op = roperand ()) == NULL)
                goto error;
              conversion = newmopnode(ct, locn, op, atag);
              return (conversion);
            }
            /*}}}*/
          else
            /*{{{  could still be conversion or specification*/
            {
              if (((current_fe_data->fe_lang & FE_LANG_NDL) == 0) &&
                  ((symb == S_COMMA) || (symb == S_FUNCTION) || (symb == S_INLINE)))
                /*{{{  it must be a function definition*/
                {
                  treenode *a = newleafnode(atag, locn);
                  *specflag = TRUE;

                  if (symb == S_COMMA)
                    /*{{{  read in the rest of the specifiers*/
                    {
                      treenode *list;
                      nextsymb ();
                      if (checklinebreak ())
                        goto error2;
                      if ((list = rlist (rspecifier, S_COMMA)) == NULL)
                        goto error;
                      a = newlistnode (S_LIST, locn, a, list);
                    }
                    /*}}}*/
                  return (rfunctiondef (a, locn, indent));
                }
                /*}}}*/
              else
                {
                  if (symb != S_NAME)
                    /*{{{  it must be a conversion*/
                      {
                        treenode *op, *conversion;
                        if ((op = roperand ()) == NULL)
                          goto error;
                        conversion = newmopnode(S_EXACT, locn, op, atag);
                        return (conversion);
                      }
                    /*}}}*/
                  else
                    /*{{{  could still be conversion or specification*/
                    {
                      treenode *list = NULL;
                      wordnode *name = rname ();
                      if (symb == S_COMMA)
                        {
                        #if 0 /* this is never passed as TRUE */
                          if (explist)
                            /*{{{  still don't know*/
                            {
                              nextsymb ();
                              if (checklinebreak ())
                                goto error2;
                              if ((list = rlist (rexp, S_COMMA)) == NULL)
                                goto error;
                              /* we have type in atag, first name in name, expression list in list */
                            }
                            /*}}}*/
                          else
                        #endif
                            /*{{{  it must be a declaration*/
                            {
                              treenode *name_list;
                              treenode *nptr;
                              *specflag = TRUE;
                              nextsymb ();
                              if (checklinebreak ())
                                goto error2;
                              if ((name_list = rlist ((treenode *(*)())rname, S_COMMA)) == NULL)
                                goto error;
                              nptr = newlistnode (S_LIST, locn, (treenode *)name, name_list);
                              return (rrestofspec (newleafnode(atag, locn), nptr, locn, indent, TRUE));
                            }
                            /*}}}*/
                        }
                      /* We have type in atag, name in 'name' */
                      /*{{{  we have reached the IS COLON or RETYPES moment of truth for spec*/
                      switch (symb)
                        {
                          /*{{{  next symbol is IS, RETYPES or COLON => specification*/
                          case S_IS: case S_RETYPES: case S_COLON:
                            *specflag = TRUE;
                            if (list != NULL)
                              list = newlistnode(S_LIST, locn, (treenode *)name, list);
                            else
                              list = (treenode *)name;
                            return (rrestofspec (newleafnode(atag, locn), list, locn, indent, TRUE));
                          /*}}}*/
                          /*{{{  it was an expression after all*/
                          default:
                          {
                            treenode *b;
                            if (list == NULL)
                              /*{{{  read the rest of this expression*/
                              {
                                b = rrestofoperand((treenode *)name, locn);
                                if (b == NULL) goto error;
                                b = newmopnode(S_EXACT, locn, b, atag);
                              }
                              /*}}}*/
                            else
                              {
                                b = newlistnode(S_LIST, locn,
                                     newmopnode(S_EXACT, locn, (treenode *)name, atag),
                                     list);
                              }
                            return(b);
                          }
                          /*}}}*/
                        }
                      /*}}}*/
                    }
                    /*}}}*/
                }
            }
            /*}}}*/
        }

      /*}}}*/
      /*{{{  S_NAME                               specification or expression*/
      case S_NAME:
        /* Could be an abbreviation without a specifier, or an expression */
        {
          treenode *const name = (treenode *)rname ();
          USE_VAR(exprlist_permitted); /* remove unused variable warning */
          USE_VAR(fn_permitted);       /* remove unused variable warning */
          if (symb == S_IS)
            /*{{{  it was an abbreviation*/
            {
              *specflag = TRUE;
              return (rrestofspec (NULL, name, locn, indent, TRUE));
            }
            /*}}}*/
          else
            /*{{{  it was an expression after all*/
            {
              treenode *const a = rrestofoperand(name, locn);
              if (a == NULL)
                goto error;
              return (rrestofexp (a, locn));
            }
            /*}}}*/
        }
      /*}}}*/
    #else
      /*{{{  S_BOOL S_BYTE S_INTn S_REALn, S_NAME expression or specification*/
      case S_BOOL: case S_BYTE: case S_INT:
      case S_INT16: case S_INT32: case S_INT64: case S_REAL32: case S_REAL64:
      case S_NAME:
        {
          int error_code;
          treenode *const a = rtypeetc(specflag, &error_code,
                                       exprlist_permitted, fn_permitted);
          if (a == NULL)
            {
              if (error_code == 1) goto error;
              if (error_code == 2) goto error2;
            }
          return a;
        }
      /*}}}*/
    #endif
      /*{{{  S_LBOX                               specification or expression*/
      case S_LBOX:
        /* Could be start of specifier in a specification,
           or segment in an expression, or constructor in an expression */
        {
          treenode *a;
          nextsymb ();
      #ifdef ARRAYCONSTRUCTOR
          if (symb == S_ARRAYCONSTRUCTOR)
            a = rarrayconstructor();
          else
      #endif
            {
              if ((a = rexp ()) != NULL)
                {
                  if (symb == S_RBOX)
                    /*{{{  constructor with one item, or specifier in specification*/
                    {
                      nextsymb ();
                      switch (symb)
                        {
                          /*{{{  S_LBOX or primitive type specifier in specification*/
                          case S_LBOX:
                          case S_CHAN: case S_PORT: case S_TIMER: case S_BOOL: case S_BYTE: case S_INT:
                          case S_INT16: case S_INT32: case S_INT64: case S_REAL32: case S_REAL64:
                          #ifdef RESOURCES
                          case S_RESOURCE:
                          #endif
                          CASE_CONFIG_TYPE
                          #ifdef OCCAM2_5
                          case S_NAME :
                          #endif
                            {
                              treenode *spec;
                              *specflag = TRUE;
                              if ((spec = rspecifier ()) == NULL)
                                goto error2;
                              a = newtypenode (S_ARRAY, locn, a, spec);
                              if (symb == S_NAME)
                                /*{{{  case S_NAME*/
                                {
                                  wordnode *name = rname ();
                                  treenode *nptr;

                                  if (symb == S_COMMA)
                                    {
                                      treenode *list;
                                      nextsymb ();
                                      if (checklinebreak ())
                                        goto error2;
                                      if ((list = rlist ((treenode *(*)())rname, S_COMMA)) == NULL)
                                        goto error;
                                      nptr = newlistnode (S_LIST, locn, (treenode *)name, list);
                                    }
                                  else
                                    nptr = (treenode *)name;
                                  return rrestofspec (a, nptr, locn, indent, TRUE);
                                }
                                /*}}}*/
                              else
                                /*{{{  otherwise, possible FUNCTION*/
                                {
                                  if ((current_fe_data->fe_lang & FE_LANG_NDL) == 0)
                                    switch(symb)
                                      {
                                        /*{{{  case S_COMMA*/
                                        case S_COMMA:
                                          {
                                            treenode *list;
                                            nextsymb ();
                                            if (checklinebreak ())
                                              goto error2;
                                            if ((list = rlist (rspecifier, S_COMMA)) == NULL)
                                              goto error2;
                                            a = newlistnode (S_LIST, locn, a, list);
                                            return (rfunctiondef(a, locn, indent));
                                          }
                                        /*}}}*/
                                        /*{{{  FUNCTION*/
                                        case S_INLINE: case S_FUNCTION:
                                          return (rfunctiondef(a, locn, indent));
                                        /*}}}*/
                                        default: break;
                                      }
                                  synerr_e (SYN_E_SPEC, locn, symb);
                                  goto error;
                                }
                                /*}}}*/
                            }
                          /*}}}*/
                          default:
                            a = rendofconstructor(newlistnode(S_LIST, locn, a, NULL), locn);
                            break;
                        }
                    }
                    /*}}}*/
                  else if (symb == S_COMMA)
                    a = rconstructor(a);
                #ifdef OCCAM2_5
                  else if ((symb == S_FROM) || (symb == S_FOR))
                #else
                  else if (symb == S_FROM)
                #endif
                    a = rsegment(a);
                  else
                    {
                      synerr_e(SYN_E_EXPR_OR_SPEC, locn, symb);
                      goto error;
                    }
                }
            }
          if (a == NULL)
            goto error;
          if (symb == S_LBOX)
            a = rsubscript (a);
          return (rrestofexp (a, locn));
        }
      /*}}}*/
      default:
        if (mustbespec(symb))
          { *specflag = TRUE; return (rspecification()); }
        if (mustbeexp(symb))
          { return (rexp()); }
        synerr_e (SYN_E_EXPR_OR_SPEC, locn, symb);
      error:
        if (*specflag)
          nextline ();
      error2:
        if (*specflag)
          skiplines (indent);
        return NULL;
    }
}
/*}}}*/

/*{{{  PRIVATE treenode *rspecorexprlist*/
#ifdef OCCAM2_5
PRIVATE BOOL rspecorexprlist_found_spec;
/*{{{  PRIVATEPARAM treenode *sub_rspecorexprlist*/
PRIVATEPARAM treenode *sub_rspecorexprlist()
{
  BOOL specflag;
  treenode *tptr;
  tptr = rspecorexpr(&specflag, -1, FALSE, FALSE);
  if (specflag) rspecorexprlist_found_spec = TRUE;
  return tptr;
}
/*}}}*/
/*{{{  PRIVATE treenode *rspecorexprlist(BOOL *specflag)*/
PRIVATE treenode *rspecorexprlist(BOOL *specflag)
{
  /* parses a list for either a list of exprs (selection list)
     or the lhs of a function definition
  */
  treenode *list;
  const BOOL saved_found_spec = rspecorexprlist_found_spec;
  rspecorexprlist_found_spec  = FALSE;
  list = rlist(sub_rspecorexprlist, S_COMMA);
  *specflag = rspecorexprlist_found_spec;
  rspecorexprlist_found_spec  = saved_found_spec;
  return list;
}
/*}}}*/
#endif
/*}}}*/
/*}}}*/


/* These may return NULL, but will always move onto the next line */
/*{{{  declarations, abbreviations and retypes*/
/* All these functions actually declare items */
/*{{{  PRIVATE treenode *rvalabbr*/
/* Pointer to a tree representing type */
/* Pointer to node representing name */
/* File location of start of abbreviation */
/* Indent of the specifier statement */
PRIVATE treenode *rvalabbr ( treenode *t , wordnode *name , SOURCEPOSN locn , int indent )
{
  treenode *e, *retptr;
  DEBUG_MSG(("rvalabbr... "));

  if ((e = rexp ()) == NULL)
    goto error;
  if (checkfor (S_COLON))
    goto error;
  if (checknlindent (indent))
    return NULL;
  retptr = declare (S_VALABBR, locn, t, name, e);
  return retptr;

error:
  nextline ();
  skiplines (indent);
  if (e != NULL)
    retptr = declare (S_VALABBR, locn, t, name, e);
  return NULL;
}
/*}}}*/
/*{{{  PRIVATE treenode *rvalretype*/

/* Pointer to a tree representing type */
/* Pointer to node representing name */
/* File location of start of retype */
/* Indent of the retypes statement */
PRIVATE treenode *rvalretype (treenode *const t, wordnode *const name,
#ifdef OCCAM2_5
                              const BOOL reshapes,
#endif
                              const SOURCEPOSN locn, const int indent)
{
  treenode *e, *retptr;

  DEBUG_MSG(("rvalretype... "));

  if ((e = rexp ()) == NULL)
    goto error;
  if (checkfor (S_COLON))
    goto error;
  if (checknlindent (indent))
    return NULL;

  retptr = declare (S_VALRETYPE, locn, t, name, e);
#ifdef OCCAM2_5
  SetNVReshapes(DNameOf(retptr), reshapes);
#endif
  return retptr;

error:
  nextline ();
  skiplines (indent);
  if (e != NULL)
    retptr = declare (S_VALRETYPE, locn, t, name, e);
  return NULL;
}
/*}}}*/
/*{{{  PUBLIC treenode *rrestofspec*/
/* Tree structure representing specifier */
/* Tree structure representing name or namelist */
/* May (illegally)be an expression list, so check it */
/* File location of start of specification */
PUBLIC treenode *rrestofspec (treenode *const t, treenode *namelist,
                              const SOURCEPOSN locn, const int indent,
                              const BOOL check_trailing_indent )
{
  treenode *a;

  DEBUG_MSG(("rrestofspec... "));

  switch (symb)
    {
      /*{{{  case S_COLON   declaration*/
      case S_COLON:                                                 /* Declaration */
      /* Lets through unsized arrays, eg. [] INT b :
         These are thrown out by chk */
        {
          treenode *tptr;

          nextsymb ();
          checknewline();
          if (check_trailing_indent && checkindent (indent))
            goto error2;
          tptr = newdeclnode (S_DECL, locn, NULL, NULL, NULL);
          if (TagOf(namelist) == S_LIST)
            {
              a = namelist;
              while (!EndOfList(a))
                /*{{{  declare name on left, move to right*/
                {
                  treenode *nptr = ThisItem(a);
                  if (TagOf(nptr) == S_NAME)
                    NewItem(declname(N_DECL, locn,
                                    (wordnode *)nptr, t, tptr), a);
                  else
                    {
                      synerr_e(SYN_E_SPEC, locn, TagOf(nptr));
                      goto error;
                    }
                  a = NextItem(a);
                }
                /*}}}*/
            }
          else
            namelist = declname(N_DECL, locn,
                               (wordnode *)namelist, t, tptr);
          SetDName(tptr, namelist);
          return (tptr);
        }
      /*}}}*/
      /*{{{  case S_IS case S_RETYPES :*/
      case S_IS: case S_RETYPES:
      #ifdef DECL_EQ
      case S_EQ:
      #endif
      #ifdef OCCAM2_5
      case S_RESHAPES:
      #endif
        {
      #ifdef OCCAM2_5
          const BOOL reshapes = (symb == S_RESHAPES);
      #endif
          const int nodetype = (symb == S_IS ? S_ABBR :
      #ifdef DECL_EQ
                                symb == S_EQ ? S_DECL :
      #endif
                                               S_RETYPE);
          if (TagOf(namelist) == S_NAME)
            {
              treenode *e;
              nextsymb ();
              if (checklinebreak () ||
                  ((e = roperand ()) == NULL) ||
                  checkfor (S_COLON))
                goto error;
              checknewline();
              if (check_trailing_indent && checkindent (indent))
                goto error2;
              e = declare (nodetype, locn, t, (wordnode *)namelist, e);
            #ifdef OCCAM2_5
              SetNVReshapes(DNameOf(e), reshapes);
            #endif
              return e;
            }
          else
            {
              synerr(SYN_TOO_MANY_NAMES, locn);
              goto error;
            }
        }
      /*}}}*/
      default:
        if ((current_fe_data->fe_lang & FE_LANG_NDL) == 0)
          synerr_e(SYN_E_COLON_IS_RETYPES, flocn, symb);
        else
          synerr_e(SYN_E_COLON_IS, flocn, symb); /* bug TS/1466 12/11/91 */
      error:
        nextline ();
      error2:
        skiplines (indent);
        return NULL;
    }
}
/*}}}*/
/*{{{  PUBLIC treenode *rspecnameandbody*/
/* Tree structure representing specifier */
/* File location of start of specification */
PUBLIC treenode *rspecnameandbody ( treenode *const t, const SOURCEPOSN locn,
       const int indent, const BOOL check_trailing_indent)
{
  if (symb == S_NAME)
    /*{{{  S_NAME*/
      {
        wordnode *name = rname();
        treenode *nptr;
        if (symb == S_COMMA)
          {
            treenode *list;
            nextsymb ();
            if (checklinebreak ())
              goto error2;
            if ((list = rlist ((treenode *(*)())rname, S_COMMA)) == NULL)
              goto error;
            nptr = newlistnode(S_LIST, locn, (treenode *)name, list);
          }
        else
          nptr = (treenode *)name;
        return rrestofspec(t, nptr, locn, indent, check_trailing_indent);
      }
    /*}}}*/
  else
    {
      if ((current_fe_data->fe_lang & FE_LANG_NDL) == 0)
        switch(symb)
          {
            /*{{{  case S_COMMA     return*/
            case S_COMMA:
              {
                treenode *list;
                nextsymb ();
                if (checklinebreak ())
                  goto error2;
                if ((list = rlist (rspecifier, S_COMMA)) == NULL)
                  goto error;
                list = newlistnode(S_LIST, locn, t, list);
                return (rfunctiondef(list, locn, indent));
              }
            /*}}}*/
            /*{{{  case S_FUNCTION  return*/
            case S_FUNCTION: case S_INLINE :
              return (rfunctiondef(t, locn, indent));
            /*}}}*/
            default:              break;
          }
      synerr_e(SYN_E_NAME, locn, symb);
    error:
      nextline ();
    error2:
      skiplines (indent);
      return NULL;
    }
}
/*}}}*/
/*}}}*/

/*{{{  protocol definition and instance*/
/* Only protdef actually declares items */
/*{{{  PRIVATE treenode *rpsub ()*/
/* Read a subscripted protocol type */
/* On error, leave symb unchanged */
PRIVATE treenode *rpsub ( void )
{
  SOURCEPOSN locn = flocn;
  DEBUG_MSG(("rpsub... "));

  switch (symb)
    {
      /*{{{  case S_LBOX*/
      case S_LBOX:
        {
          treenode *e, *sub;
          nextsymb ();
          if ((e = rexp ()) == NULL) return NULL;
          if (checkfor (S_RBOX)) return NULL;
          if ((sub = rpsub ()) == NULL) return NULL;
          return (newtypenode (S_ARRAY, locn, e, sub));
        }
      /*}}}*/
      /*{{{  case S_BOOL S_BYTE S_INTn S_REALn*/
      case S_BOOL:case S_BYTE: case S_INT:
      case S_INT16: case S_INT32: case S_INT64: case S_REAL32: case S_REAL64:
        {
          const int s = symb;
          nextsymb ();
          return (newleafnode (s, locn));
        }
      /*}}}*/
      /*{{{  case S_NAME*/
      #ifdef OCCAM2_5
      case S_NAME:
        return((treenode *)rname());
      #endif
      /*}}}*/
      default:
        synerr_e(SYN_E_PTYPE, flocn, symb);
        return NULL;
    }
}
/*}}}*/
/*{{{  PRIVATEPARAM treenode *rsimpleprotocol ()*/
/* On error, leave symb unchanged */
PRIVATEPARAM treenode *rsimpleprotocol ( void )
{
  treenode *s, *t;
  SOURCEPOSN locn = flocn;

  DEBUG_MSG(("rsimpleprotocol... "));
  switch (symb)
    {
      /*{{{  case S_REAL32 S_REAL64 S_BOOL*/
      case S_REAL32: case S_REAL64: case S_BOOL:
      case S_ANY: /* bug 1395 Added 30/09/91 by CON */
        s = newleafnode (symb, locn);
        nextsymb ();
        return (s);
      /*}}}*/
      /*{{{  case S_BYTE S_INT S_INT16 S_INT32 S_INT64*/
      case S_BYTE: case S_INT: case S_INT16: case S_INT32: case S_INT64:
        s = newleafnode (symb, locn);
        nextsymb ();
        if (symb == S_COLON2)
          /*{{{  read in array following*/
          {
            SOURCEPOSN alocn;
            nextsymb ();
            alocn = flocn;
            if (checkfor (S_BOX)) return NULL;
            switch (symb)
              {
                /*{{{  case S_BOOL S_BYTE S_INT S_INTn S_REALn*/
                case S_BOOL: case S_BYTE: case S_INT:
                case S_INT16: case S_INT32: case S_INT64: case S_REAL32: case S_REAL64:
                  t = newleafnode (symb, alocn);
                  nextsymb ();
                  break;
                /*}}}*/
                /*{{{  case S_LBOX*/
                case S_LBOX:
                  if ((t = rpsub ()) == NULL) return NULL;
                  break;
                /*}}}*/
                /*{{{  case S_NAME*/
                #ifdef OCCAM2_5
                case S_NAME:
                  t = ((treenode *)rname());
                  break;
                #endif
                /*}}}*/
                default:
                  synerr_e(SYN_E_PTYPE, flocn, symb);
                  return NULL;
              }
            s = newdopnode (S_COLON2, locn, s, newtypenode (S_ARRAY, alocn, NULL, t), 0);
          }
          /*}}}*/
        return (s);
      /*}}}*/
      /*{{{  case S_LBOX*/
      case S_LBOX:
        return (rpsub ());
      /*}}}*/
      /*{{{  case S_NAME*/
      #ifdef OCCAM2_5
      case S_NAME:
        return((treenode *)rname());
      #endif
      /*}}}*/
      default:
        synerr_e(SYN_E_PTYPE, locn, symb);
        return NULL;
    }
}
/*}}}*/
/*{{{  PRIVATEPARAM treenode *rtaggedprotocol ()*/
/* Terminates with symb at start of next line */
/* Parse
tagged.protocol { tagged.protocol }
where
tagged.protocol = tag { ',' simple.protocol }
*/
PRIVATEPARAM treenode *rtaggedprotocol ( void )
{
  wordnode *name;
  treenode *taglist;
  int indent = symbindent;
  SOURCEPOSN locn;

  linebreakindent = (-1);
  DEBUG_MSG(("rtaggedprotocol... "));
  while (symb == S_COMMENT)
    if (checknlindent (indent))
      goto error2;
  locn = flocn;
  /*{{{  read tagged list*/
  if ((name = rname ()) == NULL) goto error;
  if (symb == S_SEMICOLON)
    {
      nextsymb ();
      if (checklinebreak ()) goto error2;
      taglist = rlist (rsimpleprotocol, S_SEMICOLON);
      if (taglist == NULL) goto error;
    }
  else
    taglist = NULL;
  /*}}}*/
  checknewline ();
  return (declname (N_TAGDEF, locn, name, taglist, NULL));

error:
  nextline ();
error2:
  skiplines (indent);
  return NULL;
}
/*}}}*/
/*{{{  PRIVATE treenode *rprotdef ()*/
/* Terminates with symb at start of next line */
PRIVATE treenode *rprotdef ( void )
{
  const int indent = symbindent;
  const SOURCEPOSN locn = flocn;
  wordnode *name;

  DEBUG_MSG(("rprotdef... "));
  if (checkfor (S_PROTOCOL)) goto error;
  name = rname ();
  ignore (S_COMMENT);
  if (symb == S_NEWLINE)
    /*{{{  we have a tagged protocol*/
    {
      treenode *tlist;                    /* Pointer to list of tagged protocols */
      treenode *retptr;

      /* revised by analogy with RECORD */
      /* nextsymb ();
      if (checkindent (indent + 2)) */
      if (checknlindent (indent + 2))
        goto error2;
      while (symb == S_COMMENT)
        if (checknlindent (indent + 2))
          goto error2;
      if (checkfor (S_CASE))
        goto error;
      checknewline ();

      tlist = rproclist (rtaggedprotocol, indent + 4);

      if (checkindent (indent))
        goto error2;
      /*{{{  check for : but continue if not found*/
      if (checkfor (S_COLON))
        {
          skiplines (indent);
          goto cont;
        }
      /*}}}*/
      if (checknlindent (indent))
        goto error2;

    cont:
      retptr = declare (S_TPROTDEF, locn, tlist, name, NULL);
      return retptr;
    }
    /*}}}*/
  else
    /*{{{  we have a sequential protocol*/
    {
      treenode *p;                             /* Pointer to sequential protocol */

      if (checkfor (S_IS) ||
          checklinebreak () ||
          ((p = rlist (rsimpleprotocol, S_SEMICOLON)) == NULL) ||
          checkfor (S_COLON))
            goto error;
      if (checknlindent (indent))  goto error2;

      return declare (S_SPROTDEF, locn, p, name, NULL);
    }
    /*}}}*/

error:
  nextline ();
error2:
  skiplines (indent);
  if (symb == S_COLON)
    nextline ();
  return NULL;
}
/*}}}*/
/*{{{  PRIVATE treenode *rprotocol ()*/
/* On error, leaves protocol unchanged */
PRIVATE treenode *rprotocol ( void )
{
  DEBUG_MSG(("rprotocol... "));
  switch (symb)
    {
      case S_ANY:
        {
          const SOURCEPOSN locn = flocn;
          nextsymb ();
          return (newleafnode (S_ANY, locn));
        }
    #ifndef OCCAM2_5
      /* this is no longer necessary when rsimpleprotocol is now permitted
         to read a name */
      case S_NAME:
        return((treenode *)rname());
    #endif
      default:
        return (rsimpleprotocol());
    }
}
/*}}}*/
/*}}}*/

/*{{{  PRIVATE treenode *rspecifier_constr ()*/
#ifdef OCCAM2_5
/* On error, leaves symb unchanged */
/* Actually allows PORTs, although they are strictly illegal */
PRIVATE treenode *rspecifier_constr ( void )
{
  const SOURCEPOSN locn = flocn;
  DEBUG_MSG(("rspecifier... "));
  switch (symb)
    {
      /*{{{  BOOL BYTE INT INT16 INT32 INT64 REAL32 REAL64*/
      case S_BOOL: case S_BYTE:
      case S_INT: case S_INT16: case S_INT32: case S_INT64:
      case S_REAL32: case S_REAL64: case S_TIMER:
      #ifdef RESOURCES
      case S_RESOURCE:
      #endif
      CASE_CONFIG_TYPE
        {
          treenode *const s = newleafnode (symb, locn);
          nextsymb ();
          return (s);
        }
      /*}}}*/
      /*{{{  S_CHAN S_PORT*/
      case S_CHAN: case S_PORT:
        {
          treenode *prot;
          const int ss = symb;
          nextsymb ();
          if (checkfor (S_OF)) return NULL;
          if ((prot = rprotocol ()) == NULL) return NULL;
          return (newtypenode (ss, locn, NULL, prot));
        }
      /*}}}*/
      /*{{{  S_LBOX*/
      case S_LBOX:
        {
          treenode *spec;
          treenode *e;
          nextsymb ();
          if ((e = rexp ()) == NULL ||
              checkfor (S_RBOX) ||
              (spec = rspecifier ()) == NULL)
            return NULL;
          return (newtypenode (S_ARRAY, locn, e, spec));
        }
      /*}}}*/
      /*{{{  S_BOX*/
      case S_BOX:
        {
          treenode *spec;
          nextsymb ();
          if ((spec = rspecifier ()) == NULL) return NULL;
          return (newtypenode(S_ARRAY, locn, NULL, spec));
        }
      /*}}}*/
      /*{{{  S_NAME*/
      #ifdef OCCAM2_5
      case S_NAME:
        return (treenode *)rname();
      #endif
      /*}}}*/
      default:
        return NULL;
    };
}
#endif
/*}}}*/

/*{{{  PRIVATE treenode *rtypedecl*/
#ifdef OCCAM2_5
/* Terminates with symb at start of next line */
PRIVATE treenode *rtypedecl (const BOOL datatype, const int indent)
{
  const SOURCEPOSN locn = flocn;
  wordnode *name;

  DEBUG_MSG(("rtypedecl... "));
  if (datatype)
    { if (checkfor(S_DATA)) goto error; }
  if (checkfor (S_TYPE)) goto error;
  name = rname ();
  ignore (S_COMMENT);
  if (!datatype || (symb == S_NEWLINE))
    /*{{{  we have a record style type*/
    {
      treenode *rbody = NULL;
      treenode **rbodyptr = &rbody;
      BOOL packed = FALSE;
      if (checknlindent(indent + 2))
        goto error2;
      while (symb == S_COMMENT)
        if (checknlindent (indent + 2))
          goto error2;
      /*{{{  optional PACKED keyword*/
      if (datatype && (symb == S_PACKED))
        { nextsymb(); packed = TRUE; }
      /*}}}*/
      if (checkfor (S_RECORD))
        goto error;
      checknewline ();

      while (ignorecomments (indent + 4), (symbindent == (indent + 4)))
        {
          SOURCEPOSN this_locn = flocn;
          /**rbodyptr = rspecification ();*/
          treenode *t = rspecifier();
          if (t != NULL)
            t = rspecnameandbody(t, this_locn, indent, FALSE);
          if (t != NULL)
            {
              *rbodyptr = t;
              rbodyptr = DBodyAddr(t);

            }
          else
            { nextsymb(); skiplines(indent+4); }
        }

      if (checkindent (indent))
        goto error2;
      /*{{{  check for : but continue if not found*/
      if (checkfor (S_COLON))
        {
          skiplines (indent);
          goto cont;
        }
      /*}}}*/
      if (checknlindent (indent))
        goto error2;

    cont:
      rbody = newtypenode(S_RECORD, locn, NULL, rbody);
      SetTypeAttr(rbody, TypeAttrOf(rbody)
                       | (datatype ? TypeAttr_datatype : TypeAttr_chantype)
                       | (packed   ? TypeAttr_packed : 0));
      return declare (S_TYPEDECL, locn, rbody, name, NULL);
    }
    /*}}}*/
  else
    /*{{{  we have a normal data type*/
    {
      treenode *p;                             /* Pointer to sequential protocol */

      if (checkfor (S_IS) ||
          checklinebreak () )
            goto error;
      if ((p = rspecifier()) == NULL)
      {
        nextline ();
        skiplines (indent);
        if (symb == S_COLON)
          nextline ();
        return declare (S_TYPEDECL, locn, newleafnode(S_INT, NOPOSN),
                        name, NULL);
      }
      if (checkfor (S_COLON))
      {
        nextline ();
        skiplines (indent);
        if (symb == S_COLON)
          nextline ();
        return declare (S_TYPEDECL, locn, p, name, NULL);
      }
      if (checknlindent (indent))  goto error2;

      return declare (S_TYPEDECL, locn, p, name, NULL);
    }
    /*}}}*/

error:
  nextline ();
error2:
  skiplines (indent);
  if (symb == S_COLON)
    nextline ();
  return NULL;
}
#endif

/*}}}*/

/*{{{  PUBLIC treenode *rinitialspec ()*/
#ifdef INITIAL_DECL
/* On error, leaves symb unchanged */
/* Actually allows PORTs, although they are strictly illegal */
PUBLIC treenode *rinitialspec ( void )
{
  const SOURCEPOSN locn = flocn;
  DEBUG_MSG(("rinitialspec... "));
  switch (symb)
    {
      /* allow udt's, base types, and fixed size arrays */
      /*{{{  BOOL BYTE INT INT16 INT32 INT64 REAL32 REAL64*/
      case S_BOOL: case S_BYTE:
      case S_INT: case S_INT16: case S_INT32: case S_INT64:
      case S_REAL32: case S_REAL64: case S_TIMER:
      CASE_CONFIG_TYPE
        {
          treenode *const s = newleafnode (symb, locn);
          nextsymb ();
          return (s);
        }
      /*}}}*/
      /*{{{  COMMENT S_CHAN S_PORT*/
      /*case S_CHAN: case S_PORT:*/
        /*{*/
          /*treenode *prot;*/
          /*const int ss = symb;*/
          /*nextsymb ();*/
          /*if (checkfor (S_OF)) return NULL;*/
          /*if ((prot = rprotocol ()) == NULL) return NULL;*/
          /*return (newtypenode (ss, locn, NULL, prot));*/
        /*}*/
      /*}}}*/
      /*{{{  S_LBOX*/
      case S_LBOX:
        {
          treenode *spec;
          treenode *e;
          nextsymb ();
          if ((e = rexp ()) == NULL ||
              checkfor (S_RBOX) ||
              (spec = rspecifier ()) == NULL)
            return NULL;
          return (newtypenode (S_ARRAY, locn, e, spec));
        }
      /*}}}*/
      /*{{{  S_BOX*/
      case S_BOX:
        {
          synerr(SYN_INITDECL_ADIM_MISSING, flocn);
          return NULL;
          /*{{{  COMMENT*/
          /*treenode *spec;*/
          /*nextsymb ();*/
          /*if ((spec = rspecifier ()) == NULL) return NULL;*/
          /*return (newtypenode(S_ARRAY, locn, NULL, spec));*/
          /*}}}*/
        }
      /*}}}*/
      #ifdef RESOURCES
      case S_RESOURCE:
        {
          synerr(SYN_INITDECL_RESOURCE, flocn);
          return NULL;
        }
      #endif
      /*{{{  S_NAME*/
      #ifdef OCCAM2_5
      case S_NAME:
        return (treenode *)rname();
      #endif
      /*}}}*/
      default:
        synerr_e(SYN_E_SPECIFIER, flocn, symb);
        return NULL;
    };
}
#endif
/*}}}*/

/*{{{  PUBLIC treenode *rspecifier ()*/
/* On error, leaves symb unchanged */
/* Actually allows PORTs, although they are strictly illegal */
PUBLIC treenode *rspecifier ( void )
{
  const SOURCEPOSN locn = flocn;
  DEBUG_MSG(("rspecifier... "));
  switch (symb)
    {
      /*{{{  BOOL BYTE INT INT16 INT32 INT64 REAL32 REAL64*/
      case S_BOOL: case S_BYTE:
      case S_INT: case S_INT16: case S_INT32: case S_INT64:
      case S_REAL32: case S_REAL64: case S_TIMER:
      #ifdef RESOURCES
      case S_RESOURCE:
      #endif
      CASE_CONFIG_TYPE
        {
          treenode *const s = newleafnode (symb, locn);
          nextsymb ();
          return (s);
        }
      /*}}}*/
      /*{{{  S_CHAN S_PORT*/
      case S_CHAN: case S_PORT:
        {
          treenode *prot;
          const int ss = symb;
          nextsymb ();
          if (checkfor (S_OF)) return NULL;
          if ((prot = rprotocol ()) == NULL) return NULL;
          return (newtypenode (ss, locn, NULL, prot));
        }
      /*}}}*/
      /*{{{  S_LBOX*/
      case S_LBOX:
        {
          treenode *spec;
          treenode *e;
          nextsymb ();
          if ((e = rexp ()) == NULL ||
              checkfor (S_RBOX) ||
              (spec = rspecifier ()) == NULL)
            return NULL;
          return (newtypenode (S_ARRAY, locn, e, spec));
        }
      /*}}}*/
      /*{{{  S_BOX*/
      case S_BOX:
        {
          treenode *spec;
          nextsymb ();
          if ((spec = rspecifier ()) == NULL) return NULL;
          return (newtypenode(S_ARRAY, locn, NULL, spec));
        }
      /*}}}*/
      /*{{{  S_NAME*/
      #ifdef OCCAM2_5
      case S_NAME:
        return (treenode *)rname();
      #endif
      /*}}}*/
      default:
        synerr_e(SYN_E_SPECIFIER, flocn, symb);
        return NULL;
    };
}
/*}}}*/

/* procedure and function definitions */
/* Only rprocdef and rfunctiondef actually declare items */
/*{{{  PRIVATE treenode *rparam(t, valparam)*/
/*****************************************************************************
 *
 *  rparam reads a parameter specification, ie. it parses
 *         [ [ 'VAL' ] specifier ] name
 *         t is the specifier of the previous parameter in the list (if this
 *         specifier is omitted) valparam is the tag of the previous parameter
 *         in the list.
 *
 *****************************************************************************/

/* On error, leaves symb unchanged */
PRIVATE treenode *rparam ( treenode *t , int valparam )
{
  SOURCEPOSN locn = flocn;
  wordnode *name = NULL;
  treenode *nptr;
  /*{{{  check for a new specifier*/
  switch (symb)
    {
      /*{{{  case S_VAL*/
      case S_VAL:                               /* Read in new VAL specifier */
        valparam = N_VALPARAM;
        nextsymb ();
        /*{{{  parse specifier*/
        if ((t = rspecifier ()) == NULL)
          return NULL;
        /*}}}*/
        break;
      /*}}}*/
      /*{{{  case S_LBOX BOOL BYTE INT INT16 INT32 INT64 REAL32 REAL64 CHAN PORT TIMER RESOURCE */
      case S_LBOX: case S_BOX:
      case S_BOOL: case S_BYTE: case S_INT: case S_INT16: case S_INT32: case S_INT64:
      case S_REAL32: case S_REAL64:
      case S_CHAN: case S_PORT:
      case S_TIMER:                          /* Read in new var specifier */
      #ifdef RESOURCES
      case S_RESOURCE:
      #endif
      CASE_CONFIG_TYPE
        valparam = N_PARAM;
        /*{{{  parse specifier*/
        if ((t = rspecifier()) == NULL)
          return NULL;
        /*}}}*/
        break;
      /*}}}*/
      /*{{{  default*/
      default:
        {
          BOOL copy_type_tree = TRUE;
          /*{{{  look for type.name name*/
          #ifdef OCCAM2_5
            /* look for type.name name */
            if ((name = rname()) == NULL) return NULL;
            if (symb == S_NAME)
              {
                t = (treenode *)name;
                name = NULL; /* force the next read of name */
                copy_type_tree = FALSE;
                valparam = N_PARAM;
              }
          #endif
          /*}}}*/
          if (t == NULL) /* We have had no specifier */
            {
              synerr_e(SYN_E_SPECIFIER, locn, symb);
              return NULL;
            }
          else if (copy_type_tree) /* make a new copy of t */
            t = copytree(t, syn_lexlevel);
        }
        break;
      /*}}}*/
    }
  /*}}}*/
  if (name == NULL) /* if not null, we have already read the name! */
    {
      if ((name = rname()) == NULL) return NULL;
    }
  nptr = declname(valparam, locn, name, t, NULL);
  return(newlistnode(S_LIST, locn, nptr, NULL));
}
/*}}}*/
/*{{{  PRIVATE treenode *rparmlist()*/
/*****************************************************************************
 *
 *  rparmlist reads in a formal parameter list,
 *            ie. it parses
 *              [ 'VAL' ] specifier name { ',' [ [ 'VAL' ] specifier ] name }
 *
 *****************************************************************************/
/* On error, leaves symb unchanged */
PRIVATE treenode *rparmlist ( void )
{
  treenode *tptr, *a;
  treenode *t = NULL;
  int valparam = N_PARAM;

  if ((tptr = rparam(t, valparam)) == NULL) return NULL;
  a = tptr;
  while (symb == S_COMMA)
    /*{{{  read another parameter*/
    {
      treenode *b;
      valparam = TagOf(ThisItem(a));
      t = NTypeOf(ThisItem(a));
      nextsymb();
      if (checklinebreak ()) return NULL;
      if ((b = rparam(t, valparam)) == NULL) return NULL;
      NewNextItem(b, a);
      a = b;
    }
    /*}}}*/
  return(tptr);
}
/*}}}*/
/*{{{  PRIVATE treenode *rdescriptorline ()*/
PRIVATE treenode *rdescriptorline ( void )
{
  treenode *n = NULL;
  wordnode *name = rname();
  if (name != NULL)
    {
      if (symb == S_INPUT || symb == S_OUTPUT)
        {
          n = newactionnode(symb, flocn, (treenode *)name, NULL);
          nextsymb();
          checknewline();
        }
      else
        {
          synerr(SYN_BAD_DESCRIPTOR, flocn);
          nextline();
        }
    }
  else
    nextline();
  return n;
}
/*}}}*/
/*{{{  PRIVATE treenode *rdescriptorbody (indent)*/
PRIVATE treenode *rdescriptorbody ( int indent )
{
  treenode *pbody;
  int tag;
  SOURCEPOSN locn;

  if (checknlindent(indent + 2)) goto error2;

  locn = flocn;
  if (symb == S_PRI)
    /*{{{  PRI PAR*/
    {
      tag = S_PRIPAR;
      nextsymb();
      if (checkfor(S_PAR)) goto error2;
    }
    /*}}}*/
  else
    /*{{{  SEQ*/
    {
      if (checkfor(S_SEQ)) goto error2;
      tag = S_SEQ;
    }
    /*}}}*/

  checknewline();
  pbody = rproclist(rdescriptorline, indent + 4);

  checkindent(indent);
  if (checkfor (S_COLON))
    return(NULL);
  pbody = newcnode(tag, locn, pbody);
  return(pbody);

error2:
  skiplines(indent);
  return NULL;
}
/*}}}*/
/*{{{  PRIVATE treenode *rprocdef ()*/

/* Terminates with symb at start of next line */
PRIVATE treenode *rprocdef ( void )
{
  const int indent = symbindent;
  SOURCEPOSN locn = flocn;
  SOURCEPOSN endlocn;
  wordnode *name = NULL; /* initialised to shut up gcc's optimiser */
  treenode *params = NULL, *pbody;
  treenode *retptr;
  BOOL inline_decl = FALSE;
  int parsedheader = FALSE;
  const int oldlexlevel = syn_lexlevel;
  DEBUG_MSG(("rprocdef... "));
  foundroutine = TRUE;
  syn_lexlevel++;

  /*{{{  parse and declare procedure*/
  /*{{{  read proc header*/
  /*{{{  inline ?*/
  if (symb == S_INLINE)
    {
      inline_decl = current_fe_data->fe_allow_inlines;
      nextsymb();
    }
  /*}}}*/
  if (checkfor (S_PROC)           ||
      ((name = rname ()) == NULL) ||
      checkfor (S_LPAREN)         ||
      checklinebreak ()           )
    goto error;
  if (symb != S_RPAREN)
    {
      if ((params = rparmlist()) == NULL)
        goto error;
    }
  if (checkfor (S_RPAREN))
    goto error;
  /*}}}*/
  parsedheader = TRUE;

  /*{{{  read body*/
  if (lexmode == LEX_SOURCE)
    {
      if (checknlindent (indent + 2)) goto error2;
      pbody = rprocess();
      if (checkindent (indent)) goto error2;
      while (symb == S_COMMENT)
        if (checknlindent (indent)) goto error2;
      if (checkfor (S_COLON)) goto error2;
      endlocn = flocn;
    }
  else
    {
      pbody = rdescriptorbody(indent);
      endlocn = NOPOSN;
    }
  if (pbody == NULL) goto error2;
  /*}}}*/

  syn_lexlevel = oldlexlevel;
  retptr = declare(S_PROCDEF, locn, params, name, pbody);
  {
    treenode *const nptr = DNameOf(retptr);
    if (endlocn != NOPOSN)
    {
      DEBUG_MSG (("locn = %8lX, endlocn = %8lX\n", locn, endlocn));
      SetNPEndposn(nptr, endlocn);
    }
    if (separatelycompiled(nptr))
      addtoentrypointlist(nptr);
    if (inline_decl)
      SetTag(nptr, N_INLINEPROCDEF);
  }
  if (checknlindent (indent))
    {
      skiplines(indent);
      return(NULL);
    }
  return retptr;
  /*}}}*/

error:
  nextline ();
error2:
  msg_out_s(SEV_INFO, SYN, SYN_SKIPPING_DEFN, NOPOSN, "PROC");
  skiplines (indent);
  if (symb == S_COLON)
    nextline ();
  syn_lexlevel = oldlexlevel;
  if (parsedheader && params != NULL)
    {
      retptr = declare(S_PROCDEF, locn, params, name, newleafnode(S_SKIP, locn));
      if (separatelycompiled(DNameOf(retptr)))
        addtoentrypointlist(DNameOf(retptr));
    }
  return NULL;
}
/*}}}*/

/*{{{  PRIVATE char *param_type_string ( typetree, length )*/
/* Extracts string if type tree is user defined type, eg "COMPLEX", */
/* or, if the type is a base type (INT, REAL, etc) then it generates */
/* a string from that type instead, returns it and updates length */
#ifdef USER_DEFINED_OPERATORS
PRIVATE char *param_type_string ( treenode *typetree, size_t *length )
{
  char *string_to_add;
  size_t length_of_addition;
  switch (TagOf(typetree)) {
    case S_NAME:
      string_to_add = (char*) WNameOf((wordnode *)typetree);
      length_of_addition = (size_t)WLengthOf((wordnode *)typetree);
      break;
    case S_INT:
      string_to_add = "INT";
      length_of_addition = (size_t)3;
      break;
    case S_INT16:
      string_to_add = "INT16";
      length_of_addition = (size_t)5;
      break;
    case S_INT32:
      string_to_add = "INT32";
      length_of_addition = (size_t)5;
      break;
    case S_INT64:
      string_to_add = "INT64";
      length_of_addition = (size_t)5;
      break;
    case S_BOOL:
      string_to_add = "BOOL";
      length_of_addition = (size_t)4;
      break;
    case S_BYTE:
      string_to_add = "BYTE";
      length_of_addition = (size_t)4;
      break;
    case S_REAL32:
      string_to_add = "REAL32";
      length_of_addition = (size_t)6;
      break;
    case S_REAL64:
      string_to_add = "REAL64";
      length_of_addition = (size_t)6;
      break;
    case S_ARRAY: { size_t templength = 0; int index;
      string_to_add = (char *)malloc(256);
      for (index=0; index<256; string_to_add[index++]=0);
      string_to_add = strncpy(string_to_add,"ARRAY.OF.",(size_t)9);
      templength +=(size_t)9;
      string_to_add = strncat(string_to_add, param_type_string(ARTypeOf(typetree),&length_of_addition),(size_t)length_of_addition);
      length_of_addition+=templength;
      break;
      }
    default:
      printf("Internal Error - unknown base type in param_type_string\n");
      printf("Tag of unknown:%d\n",TagOf(typetree));
      printf("Typetree passed was:\n");
      printtree(stdout,0,typetree);
      printf("contact ofa-bugs@ukc.ac.uk\n");
      exit(1);
      break;
  }
  *length = length_of_addition;
  return(string_to_add);
}
#endif
/*}}}*/

/*{{{  PUBLIC treenode *rfunctiondef (typelist, locn, indent)*/
/* Terminates with symb at start of next line */
/* The return types of the function */
/* The file location of the function header */
/* The indent of the function header */
PUBLIC treenode *rfunctiondef ( treenode *typelist , SOURCEPOSN locn , int indent )
{
  int t;                                            /* Tag used for treenode */
  wordnode *name;                       /* Name of the function */
  treenode *params = NULL, *fbody;       /* Parameter list and function body */
  treenode *retptr;                              /* Parse tree node returned */
  treenode *fntype;                                     /* Funtion type node */
  BOOL inline_decl = FALSE;
  SOURCEPOSN endlocn;
  const int oldlexlevel = syn_lexlevel;
  #ifdef USER_DEFINED_OPERATORS
  BOOL operator_decl = FALSE;
  int num_of_params = 0;
  #endif
  DEBUG_MSG(("rfunctiondef... "));

  foundroutine = TRUE;
  syn_lexlevel++;
  /*{{{  extract function return type as a string*/
  /*}}}*/
  /*{{{  read function header*/
  /*{{{  inline ?*/
  if (symb == S_INLINE)
    {
      inline_decl = current_fe_data->fe_allow_inlines;
      nextsymb();
    }
  /*}}}*/
  if (checkfor (S_FUNCTION)     ||
  #ifdef USER_DEFINED_OPERATORS
      (name = rnameorstring (typelist,&operator_decl, &num_of_params)) == NULL || /* added by Jim 17/10/96, rname b4 */
  #else
      (name = rname ()) == NULL ||
  #endif
      checkfor (S_LPAREN)       ||
      checklinebreak()          )
    goto error;
  if (symb != S_RPAREN)
    /*{{{  parse and declare the parameter list*/
    {
      if ((params = rparmlist()) == NULL)
        goto error;
    }
    /*}}}*/
    #ifdef USER_DEFINED_OPERATORS
    /*{{{  check for correct number of parameters if an overload function*/
    if (operator_decl) {
      int params_declared;
      if (params == NULL)
        params_declared = 0;
      else
      {
        if (RightOf(params)==NULL)
          params_declared = 1;
        else
        {
          if (RightOf(RightOf(params))==NULL)
            params_declared = 2;
          else
            params_declared = 3; /* or more */
        }
      }
      if ((params_declared==0) || (params_declared==3))
      {
        switch(num_of_params)
        {
          case 1:
            synerr_e (SYN_UDO_ONE_PARAM, flocn, symb);
            goto error;
          case 2:
            synerr_e (SYN_UDO_TWO_PARAMS, flocn, symb);
            goto error;
          case -1:
            synerr_e (SYN_UDO_ONE_OR_TWO_PARAMS, flocn, symb);
            goto error;
        }
      }
      else
      {
        if (params_declared==1)
        {
          if (num_of_params==1 || num_of_params==-1)
          {
            /* monadic operators */
            /* BUG FIX 201198: if you call a function with an address of 
               a variable used later on in the parameter sequence, the
               compiler does not guarantee the evaluation order on Linux
                eg.  
                  f(g(&t), t)
               If g updates t then the second t is not necessarily the
               updated t.
             */
            int index=0;
            char *namestring = (char *)malloc(256);
            char *tempstr; /* used in bug fix */
            size_t namelength = 0;
            size_t templength = 0;
            for (index=0; index<256; namestring[index++]=0);
            namestring=strncpy(namestring, WNameOf(name), (size_t)WLengthOf(name));
            namelength+=(size_t)WLengthOf(name);
            tempstr = param_type_string(NTypeOf(LeftOf(params)), &templength);
            namestring=strncat(namestring, tempstr, templength);
            namelength+=templength;
            name = lookupword(namestring, namelength);
          }
          else
          { /* declared one param, but need two */
            synerr_e (SYN_UDO_TWO_PARAMS, flocn, symb);
            goto error;
          }
        }
        else
        {
          if (params_declared==2)
          {
            if ((num_of_params==2) || (num_of_params==-1))
            {/* else code in here to extract name */
              /* code to extract parameter names */
              /* BUG FIX 201198: if you call a function with an address of 
                 a variable used later on in the parameter sequence, the
                 compiler does not guarantee the evaluation order on Linux
                  eg.  
                    f(g(&t), t)
                  If g updates t then the second t is not necessarily the
                 updated t.
               */
              int index=0;
              char *namestring = (char *)malloc(256);
              char *tempstr; /* used in bug fix */
              size_t namelength = 0;
              size_t templength = 0;
              for (index=0; index<256; namestring[index++]=0);
              namestring=strncpy(namestring, WNameOf(name), (size_t)WLengthOf(name));
              namelength+=(size_t)WLengthOf(name);
              tempstr = param_type_string(NTypeOf(LeftOf(params)), &templength);
              namestring=strncat(namestring, tempstr, templength);
              namelength+=templength;
              namestring=strncat(namestring, ".", (size_t)1);
              namelength++;
              tempstr = param_type_string(NTypeOf(LeftOf(RightOf(params))), &templength);
              namestring=strncat(namestring, tempstr, templength);
              namelength+=templength;
              name = lookupword(namestring, namelength);
            }
            else
            {
              synerr_e (SYN_UDO_ONE_PARAM, flocn, symb);
              goto error;
            }
          }
        }
      }
    }

    #endif
    /*}}}*/
  /*{{{  check for )*/
  if (checkfor (S_RPAREN))
    goto error;
  /*}}}*/
  /* convert typelist to a list, if not one already */
  if (TagOf(typelist) != S_LIST)
    typelist = newlistnode(S_LIST, locn, typelist, NULL);
  fntype = newlistnode (S_FNTYPE, locn, typelist, params);
  /*}}}*/
  if (symb == S_IS)
    /*{{{  short function definition*/
    {
      t = S_SFUNCDEF;
      nextsymb ();
      /*{{{  ignore line break*/
      if (checklinebreak ())
        goto error;
      /*}}}*/
      /*{{{  parse expression list*/
      if ((fbody = rlist (rexp, S_COMMA)) == NULL)
        {
          nextline ();
          goto error;
        }
      /*}}}*/
      fbody = newvalofnode (S_VALOF, locn, newleafnode (S_SKIP, locn),fbody);
      /*{{{  check for :*/
      if (checkfor (S_COLON))
        {
          nextline ();
          goto error;
        }
      /*}}}*/
      endlocn = flocn;
    }
    /*}}}*/
  else
    /*{{{  long or descriptor function definition*/
    {
      t = S_LFUNCDEF;
      if (lexmode == LEX_SOURCE)
        /*{{{  long function definition*/
        {
          treenode **fbodyptr = &fbody;
          /*{{{  check nl, indent*/
          if (checknlindent (indent + 2))
            goto error;
          /*}}}*/

          while (ignorecomments (indent + 2),
                     symb != S_VALOF && symbindent == indent + 2)
            /*{{{  parse leading specification*/
            {
              *fbodyptr = rspecification ();
              if (*fbodyptr != NULL)
                fbodyptr = DBodyAddr(*fbodyptr);
            }
            /*}}}*/

          /*{{{  parse VALOF*/
          if ((*fbodyptr = rvalof ()) == NULL)
            goto error;
          /*}}}*/
          /*{{{  check indent*/
          if (checkindent (indent))
            goto error;
          /*}}}*/
          while (symb == S_COMMENT)
            /*{{{  check nl, indent*/
            if (checknlindent (indent))
              goto error;
            /*}}}*/
          /*{{{  check for :*/
          if (checkfor (S_COLON))
            goto error;
          /*}}}*/
          endlocn = flocn;
        }
        /*}}}*/
      else
        /*{{{  descriptor function definition*/
        {
          fbody = rdescriptorbody(indent);
          if (fbody == NULL) goto error;
          endlocn = NOPOSN;
        }
        /*}}}*/
    }
    /*}}}*/
  syn_lexlevel = oldlexlevel;
  retptr = declare (t, locn, fntype, name, fbody);
  {
    treenode *const nptr = DNameOf(retptr);
    if (endlocn != NOPOSN)
      SetNPEndposn(nptr, endlocn);
    if (separatelycompiled(nptr))
      addtoentrypointlist(nptr);
    if (inline_decl)
      SetTag(nptr, N_INLINEFUNCDEF);
  }
  if (checknlindent (indent))
    { skiplines(indent); return(NULL);}
  return retptr;

error:
  /*{{{  discard function definition*/
  msg_out_s(SEV_INFO, SYN, SYN_SKIPPING_DEFN, NOPOSN, "FUNCTION");
  skiplines (indent);
  if (symb == S_COLON)
    nextline ();
  syn_lexlevel = oldlexlevel;
  return NULL;
  /*}}}*/

}
/*}}}*/

/*{{{  PUBLIC BOOL rfile ()*/
PUBLIC BOOL rfile ( void )
{
  BOOL entered = FALSE;
  const int s = symb;
  const int indent = lineindent;
  nextsymb();
  if (symb != S_STRING)
    synerr(SYN_MISSING_FILENAME, flocn);
  else if (s == S_SC)
    synerr_s(SYN_SC_IS_OBSOLETE, flocn, literalv);
  else
    /*{{{  try and open the file*/
    {
      int mode;
      switch (s)
        {
          case S_INCLUDE :
                mode = LEX_SOURCE;  /* #INCLUDE MUST specify the extension */
                break;
          default :
                /* S_SC, S_IMPORT or S_USE, all are descriptor files */
                mode = ((s == S_SC) ? LEX_SC : LEX_LIB);
                current_fe_data->fe_process_filename_fn(literalv, MAXSTRING_SIZE);
                break;
        }
      entered = open_file(literalv, mode, indent);
      if (!entered)
        synerr_s(SYN_FILE_OPEN_ERROR, flocn, literalv);
      else
        {
          if (current_fe_data->fe_information)
            fprintf(current_fe_data->fe_outfile, "%s \"%s\"\n", tagstring(s), literalv);
          nextsymb();
          while (symb == S_NEWLINE)    /* Throw away leading blank lines */
            nextsymb();
          ignorecomments(0);           /* Throw away leading comments */
        }
    }
    /*}}}*/
  if (!entered)
    nextline();
  return entered;
}
/*}}}*/

/*{{{  pragma reading*/
/*{{{  PRIVATEPARAM treenode *rname_treenode*/
PRIVATEPARAM treenode *rname_treenode(void)
{
  return (treenode *)rname();
}
/*}}}*/
/*{{{  PRIVATEPARAM treenode *rpragma_external*/
PRIVATEPARAM treenode *rpragma_external(wordnode *const pragma_name,
                       const pragma_name_tag_t pragma_name_tag, int tag,
                       const int indent)
{
  const SOURCEPOSN locn = flocn;
  treenode *tptr = NULL;
  USE_VAR(tag); /* prevent unused variable warning */
  if ((symb == S_STRING) || (symb == S_STRINGCONT))
    {
      treenode *const str = rstring();
      const BOOL ok = open_file(WNameOf(CTValOf(str)), LEX_EXTERNAL, indent);
      if (ok)
        {
          treenode *const pragma_nptr = declname(N_DECL, locn, pragma_name, NULL, NULL);
          SetNMode(pragma_nptr, pragma_name_tag);
          tptr = newdeclnode(S_PRAGMA, locn, pragma_nptr, str, NULL);

          if (current_fe_data->fe_information)
            fprintf(current_fe_data->fe_outfile, "%s %s \"%s\"\n",
                    tagstring(S_PRAGMA), WNameOf(pragma_name), WNameOf(CTValOf(str)));
          nextsymb();
        }
      else
        nextline();
    }
  else
    {
      synwarn_s(SYN_BAD_PRAGMA_DIRECTIVE, flocn, WNameOf(pragma_name));
      nextline();
    }
  return tptr;
}
/*}}}*/
/*{{{  PRIVATEPARAM treenode *rpragma_string*/
PRIVATEPARAM treenode *rpragma_string(wordnode *const pragma_name,
             const pragma_name_tag_t pragma_name_tag, int tag,
             const int indent)
{
  treenode *tptr = NULL;
  treenode *str  = NULL;
  wordnode *name = pragma_name;
  BOOL ok = TRUE;
  switch (symb)
    {
      case S_STRING:
      case S_STRINGCONT:
        str = rstring();
        nextsymb();
        break;
      /*case S_NEWLINE:*/   /* Now we allow anything, (eg comments) */
      default:
        /*{{{  what to do if there's no string*/
        if (pragma_name == NULL) /* called from rsubpragma */
          {
            synerr_s(SYN_E_STRING_AFTER, flocn, tagstring(tag));
            ok = FALSE;
          }
        else if (pragma_name_tag != pragma_name_linkage)
          {
            synwarn_s(SYN_BAD_PRAGMA_DIRECTIVE, flocn, WNameOf(pragma_name));
            ok = FALSE;
          }
        /*}}}*/
        break;
    }

  if (pragma_name == NULL) /* called from rsubpragma */
    {
      const char *const string = tagstring(tag);
      name = lookupword(string, strlen(string));
    }

  if (ok)
    {
      treenode *const pragma_nptr = declname(N_DECL, flocn, name, NULL, NULL);
      SetNMode(pragma_nptr, pragma_name_tag);
      tptr = newdeclnode(S_PRAGMA, flocn, pragma_nptr, str, NULL);

      /*{{{  #NETWORK may have multiple strings*/
      if (pragma_name_tag == pragma_name_hardware)
        {
          /* more than one string in the list */
          SetDVal(tptr, newlistnode(S_LIST, flocn, str, NULL)); /* turn rhs into a list */
          while (symb == S_STRING)
            {
              str = rstring();
              SetDVal(tptr, appendnode(str, DValOf(tptr)));
              nextsymb();
            }
        }
      /*}}}*/

      /*{{{  information*/
      if (current_fe_data->fe_information)
        {
          FILE *const of = current_fe_data->fe_outfile;
          treenode *temp = DValOf(tptr);
          if (pragma_name == NULL)
            fprintf(of, "%s", tagstring(tag));
          else
            fprintf(of, "%s %s", tagstring(S_PRAGMA), WNameOf(pragma_name));
          if (TagOf(temp) == S_LIST)
            {
              for (; !EndOfList(temp); temp = NextItem(temp))
                fprintf(of, " \"%s\"", WNameOf(CTValOf(ThisItem(temp))));
            }
          else
            fprintf(of, " \"%s\"", WNameOf(CTValOf(temp)));
          fputc('\n', of);
        }
      /*}}}*/
    }
  checknlindent(indent);
  return tptr;
}
/*}}}*/
/*{{{  PRIVATEPARAM treenode *rpragma_translate*/
PRIVATEPARAM treenode *rpragma_translate(wordnode *const pragma_name,
                       const pragma_name_tag_t pragma_name_tag, int tag,
                       const int indent)
{
  treenode *tptr = NULL;
  wordnode *wptr = rname();  /* if ok, this also skips past the name */
  USE_VAR(tag); /* prevent unused variable warning */
  if (wptr != NULL)
    {
      if (symb == S_STRING)
        {
          const BOOL ok = setup_translation(wptr, literalv, literalp, flocn);
          nextsymb(); /* skip over the string */
          if (ok)
            {
              treenode *const pragma_nptr = declname(N_DECL, flocn, pragma_name, NULL, NULL);
              SetNMode(pragma_nptr, pragma_name_tag);
              tptr = newdeclnode(S_PRAGMA, flocn, pragma_nptr,
                      addtofront(newconsttablenode(S_STRING, flocn, wptr, NULL),
                      addtofront(newconsttablenode(S_STRING, flocn, lookupword(literalv, literalp), NULL),
                                 NULL)), NULL);

              if (current_fe_data->fe_information)
                fprintf(current_fe_data->fe_outfile, "%s %s %s \"%s\"\n",
                  tagstring(S_PRAGMA), WNameOf(pragma_name), WNameOf(wptr), literalv);
            }
        }
      else if (symb == S_STRINGCONT)
        synerr_i(SYN_STRING_TOO_LONG, flocn, MAXSTRING_SIZE);
      else
        synwarn_s(SYN_BAD_PRAGMA_DIRECTIVE, flocn, WNameOf(pragma_name));
    }
#if 0
  /*nextline();*/
  checknewline();
  USE_VAR(indent); /* avoid unused variable warning */
#else
  checknlindent(indent);
#endif
  return tptr;
}
/*}}}*/
/*{{{  PRIVATEPARAM treenode *rpragma_namelist*/
PRIVATEPARAM treenode *rpragma_namelist(wordnode *const pragma_name,
                       const pragma_name_tag_t pragma_name_tag, int tag,
                       const int indent)
{
  treenode *tptr = NULL;
#if 0
  treenode *const rhs = (treenode *)rname();  /* if ok, this also skips past the name */
#else
  treenode *const rhs = rlist(rname_treenode, S_COMMA);
#endif
  USE_VAR(tag); /* prevent unused variable warning */
  if (rhs != NULL)
    {
      treenode *const pragma_nptr = declname(N_DECL, flocn, pragma_name, NULL, NULL);
      SetNMode(pragma_nptr, pragma_name_tag);
      tptr = newdeclnode(S_PRAGMA, flocn, pragma_nptr, rhs, NULL);

      if (current_fe_data->fe_information)
        {
        #if 0
          fprintf(current_fe_data->fe_outfile, "%s %s %s\n",
                  tagstring(S_PRAGMA), WNameOf(pragma_name), WNameOf(rhs));
        #else
          treenode *n;
          fprintf(current_fe_data->fe_outfile, "%s %s",
                  tagstring(S_PRAGMA), WNameOf(pragma_name));
          for (n = rhs; !EndOfList(n); n = NextItem(n))
            fprintf(current_fe_data->fe_outfile, "%s %s",
                    n == rhs ? "" : ",", WNameOf((wordnode *)ThisItem(n)));
          fputc('\n', current_fe_data->fe_outfile);
        #endif
        }
    }
  checknlindent(indent);
  return tptr;
}
/*}}}*/
/*{{{  PRIVATEPARAM treenode *rpragma_name*/
PRIVATEPARAM treenode *rpragma_name(wordnode *const pragma_name,
                       const pragma_name_tag_t pragma_name_tag, int tag,
                       const int indent)
{
  treenode *tptr = NULL;
  wordnode *const rhs = rname();  /* if ok, this also skips past the name */
  USE_VAR(tag); /* prevent unused variable warning */
  if (rhs != NULL)
    {
      treenode *const pragma_nptr = declname(N_DECL, flocn, pragma_name, NULL, NULL);
      SetNMode(pragma_nptr, pragma_name_tag);
      tptr = newdeclnode(S_PRAGMA, flocn, pragma_nptr, (treenode *)rhs, NULL);

      if (current_fe_data->fe_information)
        {
          fprintf(current_fe_data->fe_outfile, "%s %s %s\n",
                  tagstring(S_PRAGMA), WNameOf(pragma_name), WNameOf(rhs));
        }
    }
  checknlindent(indent);
  return tptr;
}
/*}}}*/

/*{{{  pragma_list*/
typedef enum
  {
    A  = FE_LANG_ALL,                       /* all languages */
    O  = FE_LANG_OCCAM,                     /* occam compiler */
    C  = FE_LANG_CONFIG2 | FE_LANG_CONFIG3, /* configurer */
    CC = FE_LANG_OCCAM   | C,               /* compiler and configurer */
    CN = FE_LANG_NDL     | C,               /* compiler and configurer */
    C3 = FE_LANG_CONFIG3                    /* Version 3 configurer */
  } lang_t;

PRIVATE const struct
  {
    const char *pragma_name;
    treenode   *(*pragma_fn)(wordnode *const pragma_name, const pragma_name_tag_t pragma_tag, int tag, const int indent);
    lang_t      pragma_lang;
    pragma_name_tag_t pragma_name_tag;
  } pragma_list[] =
  {

  /* No #PRAGMA EXTERNAL when configuring, because that requires linking */
  /* No point in having LINKAGE pragma when configuring, cos we don't link
     after configuring - CON 11/10/90 */

    { "EXTERNAL",      rpragma_external,  O,  pragma_name_external   },
    { "LINKAGE",       rpragma_string,    O,  pragma_name_linkage    },
  #if 0
    { "NOUSAGECHECK",  rpragma_namelist,  O,  pragma_name_shared     }, /* was "SHARED" */
    { "NOTALIASED",    rpragma_namelist,  O,  pragma_name_aliased    },
  #else
    { "SHARED",        rpragma_namelist,  O,  pragma_name_shared     },
    { "PERMITALIASES", rpragma_namelist,  O,  pragma_name_aliased    },
  #endif
    { "ASSUMECONSTANT",rpragma_namelist,  O,  pragma_name_assumeconst},
    { "COMMENT",       rpragma_string,    CC, pragma_name_comment    },
    { "TRANSLATE",     rpragma_translate, CC, pragma_name_translate  },
    { "HARDWARE",      rpragma_string,    C3, pragma_name_hardware   },
    { "NESTEDTIMER",   rpragma_name,      O,  pragma_name_nestedtimer},
    { "NESTEDPLACE",   rpragma_name,      O,  pragma_name_nestedplace},
    { "NESTEDPORT",    rpragma_name,      O,  pragma_name_nestedport },
    { "BADLYBEHAVED",  rpragma_name,      O,  pragma_name_badlybehaved }
  };
/*}}}*/

/*{{{  PRIVATE treenode *rpragma*/
PRIVATE treenode *rpragma (void)
{
  wordnode *wptr;
  int i;
/*  const int indent = lineindent;*/ /* bug 829 20/9/91 */
  const int indent = symbindent;

  nextsymb();         /* skip past the #PRAGMA */
  wptr = rname();     /* if ok, this also skips past the name */
  if (wptr == NULL)
    nextline();       /* we just supplied an error message */
  else
    {
      for (i = 0; i < (sizeof(pragma_list) / sizeof(pragma_list[0])); i++)
        if (((pragma_list[i].pragma_lang & current_fe_data->fe_lang) != 0) &&
            (strcmp(WNameOf(wptr), pragma_list[i].pragma_name) == 0))
          return pragma_list[i].pragma_fn(wptr, pragma_list[i].pragma_name_tag, 0, indent);

      /*synerror(SYN_BAD_PRAGMA_NAME, flocn, (BIT32)symb);*/
      synwarn_s(SYN_BAD_PRAGMA_NAME, flocn, WNameOf(wptr));
      nextline();
    }

  return NULL;
}
/*}}}*/
/*{{{  PRIVATE treenode *rsubpragma*/
PRIVATE treenode *rsubpragma(const pragma_name_tag_t pragma_name_tag)
{
  const int tag = symb;
  const int indent = symbindent;
  nextsymb(); /* skip past the #COMMENT etc */
  return rpragma_string(NULL, pragma_name_tag, tag, indent);
}
/*}}}*/
/*}}}*/

/* Skips to new line on error */
/*{{{  PUBLIC treenode *rspecification ()*/
/* Terminates with symb at start of next line */
PUBLIC treenode *rspecification ( void )
{
  const int indent = symbindent;
  SOURCEPOSN locn = flocn;

  DEBUG_MSG(("rspecification... "));

  linebreakindent = (-1); /* bug TS/1493 27/11/91 */
  switch (symb)
    {
      /*{{{  S_VAL*/
      case S_VAL:
        /* Parse
              'VAL' [specifier] name 'IS' expression ':'
           or 'VAL' specifier name 'RETYPES' expression ':'
        */
      {
        treenode *spec;
        wordnode *name;
        nextsymb ();
        /*{{{  read in specifier, if present*/
        switch (symb)
          {
            case S_BOOL: case S_BYTE:
            case S_INT: case S_INT16: case S_INT32: case S_INT64:
            case S_REAL32: case S_REAL64:
            case S_CHAN: case S_PORT: case S_LBOX:
            case S_TIMER: case S_BOX:
            #ifdef RESOURCES
            case S_RESOURCE:
            #endif
            CASE_CONFIG_TYPE
              /*{{{  parse specifier*/
              if ((spec = rspecifier ()) == NULL)
                goto error;
              /*}}}*/
              break;
            default:
              spec = NULL;
          };
        /*}}}*/
        /*{{{  parse name*/
        if ((name = rname ()) == NULL)
          goto error;
        /*}}}*/
      #ifdef OCCAM2_5
        /*{{{  allow for VAL type.name name IS ...*/
        if ((spec == NULL) && (symb == S_NAME))
          {
            spec = (treenode *)name;
            name = rname();
            if (name == NULL) goto error;
          }
        /*}}}*/
      #endif
        switch(symb)
          {
            /*{{{  parse val abbreviation*/
            case S_IS:
              nextsymb ();
              /*{{{  ignore line break*/
              if (checklinebreak ())
                goto error;
              /*}}}*/
              return (rvalabbr (spec, name, locn, indent));
            /*}}}*/
            /*{{{  parse val retype*/
            #ifdef OCCAM2_5
            /*{{{  RETYPES / RESHAPES*/
            case S_RETYPES:
            case S_RESHAPES:
              {
                const BOOL reshapes = (symb == S_RESHAPES);
                nextsymb ();
                /*{{{  ignore line break*/
                if (checklinebreak ())
                  goto error;
                /*}}}*/
                return rvalretype (spec, name, reshapes, locn, indent);
              }
            /*}}}*/
            #else
            /*{{{  RETYPES*/
            case S_RETYPES:
              nextsymb ();
              /*{{{  ignore line break*/
              if (checklinebreak ())
                goto error;
              /*}}}*/
              return rvalretype (spec, name, locn, indent);
            /*}}}*/
            #endif

            /*}}}*/
            /*{{{  error*/
            default:
              if ((current_fe_data->fe_lang & FE_LANG_NDL) != 0)
                synetoken(S_IS); /* bug TS/1466 12/11/91 */
              else
                synerr_e(SYN_E_IS_OR_RETYPES, locn, symb);
              goto error;
            /*}}}*/
          }
      }
      /*}}}*/
      /*{{{  S_PROC*/
      case S_INLINE: case S_PROC:
        return (rprocdef ());
      /*}}}*/
      /*{{{  S_PROTOCOL*/
      case S_PROTOCOL:
        return (rprotdef ());
      /*}}}*/
      /*{{{  S_BOOL S_BYTE S_INT S_INTn S_REALn S_TIMER S_CHAN S_PORT ...*/
      case S_BOOL: case S_BYTE: case S_INT: case S_INT16:
      case S_INT32: case S_INT64: case S_REAL32: case S_REAL64:
      #ifndef RECORDS_OF_CHANNELS /* Records of channels not yet supported */
      case S_CHAN:
      #endif
      case S_PORT: case S_TIMER:
      case S_LBOX: case S_BOX:
      #ifdef RESOURCES
      case S_RESOURCE:
      #endif
      CASE_CONFIG_TYPE
        {
          /*{{{  COMMENT what we are parsing*/
           /*Parse*/
                 /*specifier name 'IS' element ':'                         |*/
                 /*specifier name 'RETYPES' element ':'                    |*/
                 /*type {1 ',' name } ':'                                  |*/
                 /*type name                                               |*/
                   /*init.body*/
                 /*':'*/
                 /*{1 ',' type } 'FUNCTION' name '(' {0 ',' formal } ')'*/
                   /*function.body*/
                 /*':'                                                     |*/
                 /*{1 ',' type } 'FUNCTION' name '(' {0 ',' formal } ')' 'IS' exprlist ':'*/
          /**/
             /*types and specifiers are treated as equivalent*/
          /**/
          /*}}}*/
          treenode *t;
          /*{{{  parse specifier*/
          if ((t = rspecifier()) == NULL)
            goto error;
          /*}}}*/
          return rspecnameandbody(t, locn, indent, TRUE);
        }
      #ifdef RECORDS_OF_CHANNELS /* Records of channels not yet supported */
      case S_CHAN:
        nextsymb();
        if (symb == S_TYPE)
          return rtypedecl(FALSE, indent);
        else
        {
          /* same as in rspecifier() */
          treenode *t;
          treenode *prot;
          if (checkfor (S_OF)) goto error;
          if ((prot = rprotocol ()) == NULL) goto error;
          t = newtypenode (S_CHAN, locn, NULL, prot);
          return rspecnameandbody(t, locn, indent, TRUE);
        }
      #endif
      /*}}}*/
      /*{{{  S_NAME*/
      case S_NAME:
        {
          wordnode *const name = rname();
          if (symb == S_IS)
            return (rrestofspec (NULL, (treenode *)name, locn, indent, TRUE));
        #ifdef OCCAM2_5
          else if (symb == S_NAME    || symb == S_FUNCTION ||
                   symb == S_INLINE  || symb == S_COMMA         )
            return rspecnameandbody((treenode *)name, locn, indent, TRUE);
        #endif
          else
            {
              synetoken (S_IS);
              goto error;
            }
        }
      /*}}}*/
      /*{{{  S_NETWORK:*/
      #if 1 /*def CONFIG*/
      CASE_CONFIG_SPEC
        return (rconfigdef());
      #endif
      /*}}}*/
      /*{{{  S_PLACE*/
      case S_PLACE:
        {
          treenode *n;
          treenode *e = NULL;
          int tag;
          nextsymb();
          /*{{{  parse name*/
          if ((n = relement()) == NULL)
            goto error;
          /*}}}*/
          if (symb == S_AT)
            /*{{{  PLACE name AT [ WORKSPACE ] expression :*/
            {
              if (TagOf(n) != S_NAME)
                {
                  synerr(SYN_PLACE_NAME, flocn);
                  goto error;
                }
              nextsymb();
              if (symb == S_WORKSPACE)
                {
                  tag = S_WSPLACE;
                  nextsymb();
                }
              else
                tag = S_PLACE;
              if ((e = rexp()) == NULL)
                goto error;
            }
            /*}}}*/
          else if (symb == S_IN)
            /*{{{  PLACE name IN ( WORKSPACE | VECSPACE ) :*/
            {
              if (TagOf(n) != S_NAME)
                {
                  synerr(SYN_PLACE_NAME, flocn);
                  goto error;
                }
              nextsymb();
              if (symb == S_WORKSPACE)
                tag = S_WSPLACE;
              else if (symb == S_VECSPACE)
                tag = S_VSPLACE;
              else
                {
                  synerr_e(SYN_E_WS_OR_VS, flocn, symb);
                  goto error;
                }
              nextsymb();
            }
            /*}}}*/
          else if (((current_fe_data->fe_lang & (FE_LANG_CONFIG2 | FE_LANG_CONFIG3)) != 0) &&
                   ((symb == S_ON) || (symb == S_COMMA)))
            /*{{{  PLACE {1 ',' element} ON element :*/
            {
              treenode *list = NULL;
              if (symb == S_COMMA)
                {
                  nextsymb();
                  if (checklinebreak())
                    goto error2;
                  if ((list = rlist(relement, S_COMMA)) == NULL)
                    goto error;
                }
              n = newlistnode(S_LIST, flocn, n, list);
              if (checkfor(S_ON))
                goto error;
              if ((e = relement()) == NULL)
                goto error;
              tag = S_PLACEON;
            }
            /*}}}*/
          else
            /*{{{  report error*/
            {
              synerr_e(SYN_E_AT_OR_IN, flocn, symb);
              goto error;
            }
            /*}}}*/
          /*{{{  check for :*/
          if (checkfor(S_COLON))
            goto error;
          /*}}}*/
          /*{{{  check nl, indent*/
          if (checknlindent(indent))
            goto error2;
          /*}}}*/
          return newdeclnode(tag, locn, n, e, NULL);
        }
      /*}}}*/
      /*{{{  S_PRAGMA*/
      case S_PRAGMA: /* bug 829 19/9/91 */
        {
          treenode *tptr = rpragma();
          if (tptr == NULL) goto error2;
          return tptr;
        }
      /*}}}*/
      /*{{{  S_HCOMMENT*/
      case S_HCOMMENT: /* bug 829 20/9/91 */
        {
          treenode *const tptr = rsubpragma(pragma_name_comment);
          if (tptr == NULL) goto error2;
          return tptr;
        }
      /*}}}*/
      /*{{{  S_HNETWORK*/
      case S_HNETWORK:
        {
          treenode *const tptr = rsubpragma(pragma_name_hardware);
          if (tptr == NULL) goto error2;
          if ((current_fe_data->fe_lang & FE_LANG_CONFIG3) == 0)
            {
              synwarn_s(SYN_BAD_PRAGMA_NAME, LocnOf(tptr), tagstring(S_HNETWORK));
              return NULL;
            }
          return tptr;
        }
      /*}}}*/
      /*{{{  S_INCLUDE S_SC S_USE etc*/
      case S_INCLUDE: case S_SC: case S_USE: case S_IMPORT:
        rfile ();
        return rspecification();
      /*}}}*/
      /*{{{  S_DATA*/
      #ifdef OCCAM2_5
      case S_DATA:
        return rtypedecl(TRUE, indent);
      #endif
      /*}}}*/
      default:
        synerr_e(SYN_E_SPEC, locn, symb);
      error:
        nextline ();
      error2:
        skiplines (indent);
        return NULL;
    }
}
/*}}}*/
/*{{{  PUBLIC BOOL rleadingspecs(specroot, exp)*/
/* Reads a specifications before a selection, alternative or choice
* return the first expression found.
* specs is left pointing to the root of the spec list.
* readerror is set true if error reading expression
*/
PUBLIC BOOL rleadingspecs (treenode **specroot, treenode **exp,
            const BOOL exprlist_permitted)
{
  treenode *s, **lasts = specroot;
  const int indent = symbindent;
  DEBUG_MSG(("rleadingspecs... "));
  while (TRUE)
    switch (symb)
      {
        /*{{{  S_INCLUDE S_SC S_USE S_IMPORT*/
        case S_INCLUDE: case S_SC: case S_USE: case S_IMPORT:
          if (!rfile ())
            {
              skiplines(indent);
              if (lineindent < indent)
                { *exp = NULL; *lasts = NULL; return FALSE; }
            }
          break;
        /*}}}*/
        /*{{{  primitive type S_NAME S_LBOX type = specification or expression*/
        case S_NAME: case S_LBOX: case S_BOOL: case S_BYTE:
        case S_INT: case S_INT16: case S_INT32: case S_INT64:
        case S_REAL32: case S_REAL64:
          {
            BOOL specflag = FALSE; /* Assume no leading specification */
            const BOOL fn_permitted = ((current_fe_data->fe_lang & FE_LANG_NDL) == 0);
            s = rspecorexpr (&specflag, -1, exprlist_permitted, fn_permitted);
            if (specflag)
              {
                *lasts = s;
                if (*lasts != NULL) lasts = DBodyAddr(*lasts);
              }
            else
              { /* exp or error reading exp */
                *lasts = NULL;
                *exp = s;
                return (s==NULL);
              }
            break;
          }
        /*}}}*/
        default :
          if (mustbespec(symb))
            /*{{{  read spec; break*/
            {
              *lasts = rspecification ();
              if (*lasts != NULL) lasts = DBodyAddr(*lasts);
              break;
            }
            /*}}}*/
          else if (mustbeexp(symb))
            /*{{{  read expression; return*/
            {
              *lasts = NULL;
              *exp = rexp ();
              return (*exp == NULL);
            }
            /*}}}*/
          DEBUG_MSG(("rleadingspecs: returning FALSE... "));
          *lasts = NULL;
          *exp = NULL;
          return FALSE;
      }
}
/*}}}*/

