/* $Id: use2.c,v 1.2 1997/03/25 14:52:40 djb1 Exp $ */

/*
 *	Occam two usage checker
 *	Copyright (C) 1987 Inmos Limited
 *
 *	This program is free software; you can redistribute it and/or modify
 *	it under the terms of the GNU General Public License as published by
 *	the Free Software Foundation; either version 2 of the License, or
 *	(at your option) any later version.
 *
 *	This program is distributed in the hope that it will be useful,
 *	but WITHOUT ANY WARRANTY; without even the implied warranty of
 *	MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *	GNU General Public License for more details.
 *
 *	You should have received a copy of the GNU General Public License
 *	along with this program; if not, write to the Free Software
 *	Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 */

#define DEBUG

/*{{{  include files */
#ifdef HAVE_CONFIG_H
#  include <config.h>
#endif
#include <stdio.h>
#include "feinc.h"
#include "useerror.h"
#include "usehdr.h"
#include "usedef.h"
#include "use1def.h"
#include "use2def.h"
#include "syndef.h" /* syn_lexlevel */
#include "chkdef.h" /* current_fe_handle etc */
/*}}}*/

/*{{{  static variables */
PRIVATE SOURCEPOSN lastloc = NOPOSN;
/*}}}*/

/*{{{  DEBUG_MSG */
#undef DEBUG_MSG
#ifdef DEBUG
  #define DEBUG_MSG(X) { if (current_fe_data->fe_debuguse) printf X; }
#else
  #define DEBUG_MSG(X)
#endif
/*}}}*/

/*{{{  PRIVATE int accesses_overlap (slist1, slist2) */
/*****************************************************************************
 *
 *  Test whether any of the accesses specified in the two lists of accessed
 *  subscripts 'slist1' and 'slist2' overlap.
 *
 *****************************************************************************/
PRIVATE int accesses_overlap ( subscrlist *slist1 , subscrlist *slist2 )
{
  while (slist1 != NULL)
    {
      if (subscr_overlaps(slist2, SLFirstOf(slist1), SLLastOf(slist1)))
        return(TRUE);
      slist1 = SLNextOf(slist1);
    }
  return(FALSE);
}
/*}}}*/
/*{{{  PRIVATE void check_no_overlaps (vlist1, vlist2, loc) */
/*****************************************************************************
 *
 *  Check that accesses do not overlap.  If they do generate the appropriate
 *  error message.
 *  Written variables must not overlap with the same variable being read or
 *  written in the other list.
 *  Channel input usage must not overlap
 *  Channel output usage must not overlap
 *  SHC, 3-Jan-89: Timers are not checked.
 *
 *****************************************************************************/
PRIVATE void check_no_overlaps (varlist *vlist1, varlist *vlist2, SOURCEPOSN loc,
                                const BOOL checking_counted_array,
                                const BOOL in_pri_par) /* INSdi01282 */
{
  for (; vlist1 != NULL; vlist1 = VLNextOf(vlist1))
    {
      treenode *const n = VLNameOf(vlist1);
      const BOOL is_timer = basetype(NTypeOf(n)) == S_TIMER;
    #if 0
      printf("check_no_overlaps: n from vlist1 is %s, is_timer? %d, in_pri_par? %d, shared? %d\n",
             WNameOf(NNameOf(n)), is_timer, in_pri_par, NVSharedOf(n));
    #endif
      if ((!is_timer || in_pri_par) && !NVSharedOf(n))
        /*{{{  do check */
        {
          varlist *list;
          for (list = vlist2; (list != NULL) && (VLNameOf(list) != n); list = VLNextOf(list))
            {
            /*printf("check_no_overlaps: n from vlist2 is %s\n", WNameOf(NNameOf(VLNameOf(list))));*/
              /* skip */
            }
          if (list != NULL)
            {
              /*printf("check_no_overlaps: n from vlist2 is finally %s\n", WNameOf(NNameOf(VLNameOf(list))));*/
              /*const BOOL variableio = current_fe_data->fe_txlib->hasvariableio;*/
              if (is_timer) /* INSdi01282 and hence we must be inside a PRI PAR */
                /*{{{  check for timer use in a PRI PAR */
                {
                  /* TIMERs are not marked with direction use when separately
                     compiled, thus since they are non-VAL parameters, it
                     thinks that they are 'written'.
                     therefore we have to check for
                     'Input'/'Written' against 'Input'/'Written'.
                  */
                  if (accesses_overlap(VLInputOf  (vlist1), VLInputOf  (list)) ||
                      accesses_overlap(VLWrittenOf(vlist1), VLWrittenOf(list)) ||
                      accesses_overlap(VLInputOf  (vlist1), VLWrittenOf(list)) ||
                      accesses_overlap(VLWrittenOf(vlist1), VLInputOf  (list)))
                    usereport(TRUE, USE_TIMER_IN_PRI_PAR, loc, n);
                }
                /*}}}*/
              else
                {
                  /*{{{  Check for written in parallel */
                  if (accesses_overlap(VLWrittenOf(vlist1), VLWrittenOf(list)))
                    {
                      if (checking_counted_array) /* bug 1183 19/3/91 */
                       #if 0
                        usereport(!variableio,
                                   variableio ? USE_COLON2_INPUT : USE_FUTURE_COLON2_INPUT,
                                   loc, n);
                       #else
                        /* bug TS/1172 means that we have to fail on this */
                        usereport(FALSE, USE_COLON2_INPUT, loc, n);
                       #endif
                      else
                        usereport(FALSE, USE_WRITTEN_IN_PAR, loc, n);
                    }
                  /*}}}*/
                  /*{{{  Check for read and written in parallel */
                  if (accesses_overlap(VLWrittenOf(vlist1), VLReadOf(list)) ||
                      accesses_overlap(VLReadOf(vlist1), VLWrittenOf(list))   )
                    {
                      if (checking_counted_array) /* bug 1183 19/3/91 */
                       #if 0
                        usereport(!variableio,
                                   variableio ? USE_COLON2_INPUT : USE_FUTURE_COLON2_INPUT,
                                   loc, n);
                       #else
                        /* bug TS/1172 means that we have to fail on this */
                        usereport(FALSE, USE_COLON2_INPUT, loc, n);
                       #endif
                      else
                        usereport(FALSE, USE_READ_AND_WRITTEN_IN_PAR, loc, n);
                    }
                  /*}}}*/
                  /*{{{  Check channel usages */
                  if (accesses_overlap(VLInputOf(vlist1), VLInputOf(list)))
                    usereport(FALSE, USE_INPUT_IN_PAR, loc, n);
                  if (accesses_overlap(VLOutputOf(vlist1), VLOutputOf(list)))
                    usereport(FALSE, USE_OUTPUT_IN_PAR, loc, n);
                  /*}}}*/
                }
            }
        }
        /*}}}*/
    }
}
/*}}}*/

/*{{{  PRIVATE void applytocountedinputs */
PRIVATE void applytocountedinputs(treenode *tptr, void (*f)(treenode *, void *), void *const voidptr)
/*****************************************************************************
 *  applytocountedinputs applies the supplied function to any counted array
 *                       protocol inputs which are part of the supplied
 *                       input statement.
 *****************************************************************************/
{
  switch(TagOf(tptr))
    {
      default:
        badtag(LocnOf(tptr), TagOf(tptr), "applytocountedinputs");
        break;
      case S_INPUT:
      case S_TAGGED_INPUT:
        for (tptr = RHSOf(tptr); !EndOfList(tptr); tptr = NextItem(tptr))
          if (TagOf(ThisItem(tptr)) == S_COLON2)
            f(ThisItem(tptr), voidptr);
        break;
      case S_CASE_INPUT:
        for (tptr = RHSOf(tptr); !EndOfList(tptr); tptr = NextItem(tptr))
          {
            treenode *variant = skipspecifications(ThisItem(tptr));
            if (TagOf(variant) != S_VARIANT)
              badtag(LocnOf(variant), TagOf(variant), "applytocountedinputs");
            for (variant = VRTaggedListOf(variant);
                 !EndOfList(variant);
                 variant = NextItem(variant))
              if (TagOf(ThisItem(variant)) == S_COLON2)
                f(ThisItem(variant), voidptr);
          }
        break;
    }
}
/*}}}*/
/*{{{  PRIVATE varlist *check_multipleassign */
PRIVATE varlist *check_multipleassign(treenode *tptr, varlist *vars_used,
                                      const SOURCEPOSN locn,
                                      const BOOL checking_counted_array)
{
  varlist *uses_vars = freevarsinexp(tptr, -1, NULL, EXP_WRITTEN, TRUE);
  check_no_overlaps(vars_used, uses_vars, locn, checking_counted_array, FALSE);
  vars_used = merge(uses_vars, -1, vars_used);
  release_varlist(uses_vars);
  return vars_used;
}
/*}}}*/
/*{{{  PRIVATEPARAM void checkcountedinput */
PRIVATE varlist *checkcountedinput_vlist;
PRIVATE int      checkcountedinput_scope;

PRIVATEPARAM void checkcountedinput(treenode *tptr, void *voidptr)
/* Check for 'overlap' of the length and array expression */
{
  varlist *vars_used;
  SOURCEPOSN oldloc = lastloc;
  lastloc = LocnOf(tptr);
  voidptr = voidptr;

  /*fprintf(current_fe_data->fe_outfile, "checkcountedinput: Locn: %lX\n", LocnOf(tptr));*/

  if (TagOf(tptr) != S_COLON2)
    badtag(LocnOf(tptr), TagOf(tptr), "checkcountedinput");
  vars_used = check_multipleassign(LeftOpOf(tptr),  NULL,      lastloc, TRUE);
  vars_used = check_multipleassign(RightOpOf(tptr), vars_used, lastloc, TRUE);

  checkcountedinput_vlist = merge(vars_used, checkcountedinput_scope, checkcountedinput_vlist);
  release_varlist(vars_used);

  lastloc = oldloc;
}
/*}}}*/
/*{{{  PUBLIC varlist *use_freevarsin (n, scope, vlist) */
/*****************************************************************************
 *
 *  use_freevarsin takes a treenode, 'n', a scope level, 'scope', and a list of
 *  free variables, 'vlist', and adds to 'vlist' any variables used in tree
 *  'n' of scope less than 'scope', which aren't already in vlist.
 *  for the special cases not handled by the general freevarsin in use1.c
 *
 *****************************************************************************/

PUBLIC varlist *use_freevarsin ( treenode *n , int scope , varlist *vlist )
{
  DEBUG_MSG(("use_freevarsin: %s\n", usetagstring(n)));
  switch(TagOf(n))
    {
      /*{{{  ASS */
      /* Added here to fix bug 278 in buglist */
      case S_ASS :
        if (TagOf(LHSOf(n)) == S_LIST)
          {
            /* Check for overlap of LHS of multiple assignment */
            varlist *vars_used = NULL;
            SOURCEPOSN oldloc = lastloc;
            lastloc = LocnOf(n);
            for (n = LHSOf(n); !EndOfList(n); n = NextItem(n))
              vars_used = check_multipleassign(ThisItem(n), vars_used, lastloc, FALSE);
            vlist = merge(vars_used, scope, vlist);
            release_varlist(vars_used);
            lastloc = oldloc;
          }
        break;
      /*}}}*/
      /*{{{  INPUT TAGGED_INPUT CASE_INPUT */
      /* Added for bug 1183 19/3/91 */
      case S_INPUT: case S_TAGGED_INPUT: case S_CASE_INPUT:
        {
          varlist *saved_vlist = checkcountedinput_vlist;
          int      saved_scope = checkcountedinput_scope;
          checkcountedinput_vlist = vlist;
          checkcountedinput_scope = scope;
      
          applytocountedinputs(n, checkcountedinput, NULL);
          vlist = checkcountedinput_vlist;
      
          checkcountedinput_scope = saved_scope;
          checkcountedinput_vlist = saved_vlist;
        }
        break;  
      /*}}}*/
      /*{{{  PAR PRIPAR PLACEDPAR */
      case S_PAR: case S_PRIPAR: case S_PLACEDPAR: case S_DO:
        {
          const BOOL in_pri_par = TagOf(n) == S_PRIPAR;
          varlist *vars_used = NULL;
          varlist *uses_vars;
          SOURCEPOSN save_lastloc = lastloc;
          lastloc = LocnOf(n);
          for (n = CBodyOf(n); !EndOfList(n); n = NextItem(n))
            {
              uses_vars = freevarsin(ThisItem(n), -1, NULL, TRUE);
              check_no_overlaps(vars_used, uses_vars, lastloc, FALSE,
                                in_pri_par);
              vars_used = merge(uses_vars, -1, vars_used);
              release_varlist(uses_vars);
            }
          vlist = merge(vars_used, scope, vlist);
          release_varlist(vars_used);
          lastloc = save_lastloc;
        }
        break;
      /*}}}*/
      /*{{{  REPLPAR PRIREPLPAR PLACEDREPLPAR */
      case S_REPLPAR: case S_PRIREPLPAR: case S_PLACEDREPLPAR: case S_REPLDO:
        {
          const BOOL in_pri_par = TagOf(n) == S_PRIREPLPAR;
          const SOURCEPOSN save_lastloc = lastloc;
          if (scope < 0)
            scope = NScopeOf(ReplCNameOf(n));
          vlist = freevarsinexp(ReplCStartExpOf(n),  scope, vlist, EXP_READ, TRUE);
          vlist = freevarsinexp(ReplCLengthExpOf(n), scope, vlist, EXP_READ, TRUE);
          lastloc = LocnOf(n);
          syn_lexlevel++; /* bug 1008 9/10/90 */
          DEBUG_MSG(("use_freevarsin: incrementing syn_lexlevel\n"));
          if (is_evaluable(ReplCStartExpOf(n)) &&
              is_evaluable(ReplCLengthExpOf(n)))
            /*{{{  base and count both known */
            {
              varlist *vars_used = NULL;
              varlist *uses_vars;
              INT32 base = eval_const_treenode(ReplCStartExpOf(n));
              INT32 count = eval_const_treenode(ReplCLengthExpOf(n));
              treenode *name = ReplCNameOf(n);
              n = ReplCBodyOf(n);
              SetNReplKnown(name, TRUE);
              while (count > 0)
                {
                  DEBUG_MSG(("use_freevarsin: repl %s set to %ld\n", WNameOf(NNameOf(name)), base));
                  SetNReplValue(name, base);
                  uses_vars = freevarsin(n, NScopeOf(name), NULL, TRUE);
                  check_no_overlaps(vars_used, uses_vars, lastloc, FALSE,
                                    in_pri_par);
                  vars_used = merge(uses_vars, -1, vars_used);
                  release_varlist(uses_vars);
                  base++;
                  count--;
                }
              vlist = merge(vars_used, scope, vlist);
              release_varlist(vars_used);
              SetNReplKnown(name, FALSE);
            }
            /*}}}*/
          else
            /*{{{  base not known */
            /* replicator base is unknown so no access to free vars */
            {
              varlist *uses_vars;
              SetNReplKnown(ReplCNameOf(n), FALSE);
              n = ReplCBodyOf(n);
              uses_vars = freevarsin(n, -1, NULL, TRUE);
              check_no_overlaps(uses_vars, uses_vars, lastloc, TRUE, /* ie check no free vars */
                                in_pri_par);
              vlist = merge(uses_vars, scope, vlist); /* just in case there were any */
              release_varlist(uses_vars);
            }
            /*}}}*/
          DEBUG_MSG(("use_freevarsin: decrementing syn_lexlevel\n"));
          syn_lexlevel--;
          lastloc = save_lastloc;
        }
        break;
      /*}}}*/
    }
  return vlist;
}
/*}}}*/
/*{{{  PRIVATE BOOL goes_par (n) */
/*{{{  PRIVATEPARAM void find_counted_inputs */
PRIVATEPARAM void find_counted_inputs(treenode *tptr, void *const voidptr)
{
  (void)tptr; /* stop warning message */
  /*fprintf(current_fe_data->fe_outfile, "find_counted_inputs: Locn: %lX\n", LocnOf(tptr));*/
  *(BOOL *)voidptr = TRUE;
}
/*}}}*/
/*{{{  PRIVATEPARAM int do_goes_par (n) */
/*****************************************************************************
 *
 *  Test whether a process goes PAR at all.  PROC and FUNCTION calls
 *  are not expanded.
 *
 *****************************************************************************/
PRIVATEPARAM int do_goes_par ( treenode *n, void *voidptr)
{
  if (!*(BOOL *)voidptr)
    {
      DEBUG_MSG(("do_goes_par: %s\n", usetagstring(n)));
      switch(TagOf(n))
        {
          case S_PAR: case S_PRIPAR: case S_PLACEDPAR:
          case S_REPLPAR: case S_PRIREPLPAR: case S_PLACEDREPLPAR:
          case S_DO: case S_REPLDO:
            *(BOOL *)voidptr = TRUE;
            break;
          case S_ASS :
            /* Added so that replicators are checked inside multi-assign */
            /* (fixing bug 278 in buglist) */
            *(BOOL *)voidptr = (TagOf(LHSOf(n)) == S_LIST);
            break;
          /* Added for bug 1183 19/3/91: */
          case S_INPUT: case S_TAGGED_INPUT: case S_CASE_INPUT:
            {
              BOOL found_counted_input = FALSE;
              applytocountedinputs(n, find_counted_inputs, &found_counted_input);
              *(BOOL *)voidptr = found_counted_input;
            }
            break;
        }
    }
  return (*(BOOL *)voidptr) ? STOP_WALK : CONTINUE_WALK;
}
/*}}}*/
/*{{{  PRIVATE BOOL goes_par (n) */
/*****************************************************************************
 *
 *  Test whether a process goes PAR at all.  PROC and FUNCTION calls
 *  are not expanded.
 *
 *****************************************************************************/
PRIVATE BOOL goes_par ( treenode *const n )
{
  BOOL goes_par_result = FALSE;
  prewalkproctree(n, do_goes_par, &goes_par_result);
  return goes_par_result;
}
/*}}}*/
/*}}}*/
/*{{{  PUBLIC void usagecheck (n) */
/*{{{  PRIVATEPARAM int do_usagecheck (n) */
/*****************************************************************************
 *
 *  Perform usage check of variables and channels inside a routine.
 *  Expand replicators where possible and necessary.
 *
 *****************************************************************************/
PRIVATEPARAM int do_usagecheck ( treenode *n, void *const voidptr)
{
  DEBUG_MSG(("do_usagecheck: %s\n", usetagstring(n)));
  switch(TagOf(n))
    {
      /*{{{  PAR PRIPAR PLACEDPAR + repl versions, plus multiple assign */
      case S_REPLPAR: case S_PRIREPLPAR: case S_PLACEDREPLPAR:
      case S_PAR: case S_PRIPAR: case S_PLACEDPAR: case S_DO: case S_REPLDO:
      case S_ASS: /* Added to allow usage check on multiple assignments (bug 278) */
        {
          varlist *const vars_used = use_freevarsin(n, -1, NULL);
          release_varlist(vars_used);
        }
      return STOP_WALK;
      /*}}}*/
      /*{{{  REPLSEQ REPLIF REPLALT PRIREPLALT */
      case S_REPLSEQ: case S_REPLIF: case S_REPLALT: case S_PRIREPLALT:
        /* N.B. no need to expand replicator if process doesn't go PAR */
        if (is_evaluable(ReplCStartExpOf(n)) &&
            is_evaluable(ReplCLengthExpOf(n)) &&
            goes_par(ReplCBodyOf(n)))
          /*{{{  base & count both known & body goes PAR; return TRUE */
          {
            INT32 base  = eval_const_treenode(ReplCStartExpOf(n));
            INT32 count = eval_const_treenode(ReplCLengthExpOf(n));
            treenode *name = (treenode *) ReplCNameOf(n);
            n = ReplCBodyOf(n);
            SetNReplKnown(name, TRUE);
            while (count > 0)
              {
                SetNReplValue(name, base);
                prewalkproctree(n, do_usagecheck, voidptr);
                base++;
                count--;
              }
            SetNReplKnown(name, FALSE);
            return STOP_WALK;
          }
          /*}}}*/
        else
          /*{{{  otherwise ... */
          {
            SetNReplKnown(ReplCNameOf(n), FALSE);
          }
          /*}}}*/
        break;
      /*}}}*/
      /*{{{  INPUTs */
      case S_INPUT: case S_TAGGED_INPUT: case S_CASE_INPUT: /* bug 1183 19/3/91 */
        {
          varlist *const vars_used = use_freevarsin(n, -1, NULL);
          release_varlist(vars_used);
        }
      break;
      /*}}}*/

    }
  return CONTINUE_WALK;
}
/*}}}*/
/*{{{  PRIVATEPARAM int do_usagecheck_tree (n) */
/*****************************************************************************
 *
 * look for PROCs and FUNCTIONs.
 * call do_usagecheck on each.
 *
 *****************************************************************************/
PRIVATEPARAM int do_usagecheck_tree ( treenode *n, void *const voidptr)
{
  switch(TagOf(n))
    {
      /*{{{  proc def */
      case S_PROCDEF:        /* N.B. Functions can't go PAR */
      CASE_CONFIG_SPEC
        if (!separatelycompiled(DNameOf(n)))
          {
            syn_lexlevel++; /* bug 1008 9/10/90 */
            DEBUG_MSG(("do_usagecheck_tree: incrementing syn_lexlevel\n"));
            prewalkproctree(DValOf(n), do_usagecheck_tree, voidptr);  /* do nested stuff */
            DEBUG_MSG(("do_usagecheck_tree: now checking %s\n", usetagstring(n)));
            prewalkproctree(DValOf(n), do_usagecheck, voidptr);
            DEBUG_MSG(("do_usagecheck_tree: decrementing syn_lexlevel\n"));
            syn_lexlevel--;
          }
        break;
      /*}}}*/
      #if 0
      /*{{{  PROCESSOR */
      case S_PROCESSOR:
        return STOP_WALK;
      /*}}}*/
      #endif
    }
  return CONTINUE_WALK;
}
/*}}}*/
/*{{{  PUBLIC void usagecheck (n) */
/*****************************************************************************
 *
 *  Walk tree looking for PROCs and FUNCTIONs,
 *  then usage check each one.
 *
 *****************************************************************************/
PUBLIC void usagecheck ( treenode *n )
{
  prewalkproctree(n, do_usagecheck_tree, NULL);
}
/*}}}*/
/*}}}*/
