/* $Id: genhdr.h,v 1.1 1996/04/15 10:52:08 djb1 Exp $ */

/*
 *	code generator definitions
 *	Copyright (C) 1987 Inmos Limited
 *
 *	This program is free software; you can redistribute it and/or modify
 *	it under the terms of the GNU General Public License as published by
 *	the Free Software Foundation; either version 2 of the License, or
 *	(at your option) any later version.
 *
 *	This program is distributed in the hope that it will be useful,
 *	but WITHOUT ANY WARRANTY; without even the implied warranty of
 *	MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *	GNU General Public License for more details.
 *
 *	You should have received a copy of the GNU General Public License
 *	along with this program; if not, write to the Free Software
 *	Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 */

/*{{{  special namenode values */
/* Workspace offset of name which doesn't have an offset */
#define NO_SLOT 0x80000000
/*}}}*/

/*{{{  static link, vector space, constant pointer constants */
#define MAX_LEX_LEVELS           254
#define STATICLINK_NOT_USED      0xFF    /* an illegal lexical level */
#define CONSTANTPOINTER_NOT_USED NO_SLOT /* an impossible offset */

/* Fields within an entry in staticlinkoffsets */
#define OFFSET_BITS 0x7fffffff
#define REPL_FLAG   0x80000000

/*}}}*/
/*{{{  expression evaluation constants */
#define MAXREGS 3
#define MAXFPUREGS 3
#define MANY_REGS (-1)
/*}}}*/
/*{{{  expression modes */
#define P_EXP     1   /* operand is an expression */
#define P_TEMP    2   /* operand is an evaluated temporary */
#define P_TEMPPTR 3   /* operand is an evaluated temporary pointer */
#define P_PTR     4   /* operand is a pointer */
/*}}}*/
/*{{{  constants for below workspace data requirements */
#if 0
#define DS_MIN  2   /* All processes need at least two slots */
#define DS_IO   3   /* Processes using channels need three slots */
#define DS_WAIT 5   /* Processes using timers need five slots */
#else
/* if L_process is 1 (TRUE), then add 1 to each of these: */
#define DS_MIN  (int)(2 + L_process) /* All processes need at two(three) slots */
#define DS_IO   (int)(3 + L_process) /* Processes using channels need three(four) slots */
#define DS_WAIT (int)(5 + L_process) /* Processes using timers need five(six) slots */
#endif

#if 0
#define DEFAULT_HIGH_PRI_TRAP_OFFSET 20   /* H1 L-process's high priority trap handler */
                                  /* is at this many words from low pri one */
#else
#define DEFAULT_HIGH_PRI_TRAP_OFFSET -2   /* H1 L-process's high priority trap handler */
                                  /* is a pointer at this many words from low pri one */
#endif

/* Three below workspace slots contain the register parameters */
#define REG_PARAMS MAXREGS
/* One below workspace slot contains the return address */
#define INS_EXTRA  ((INT32)1)
/*}}}*/
/*{{{  constants used when generating replicators */
/* Offsets from replicator workspace position of the components of a
   replicator */
#define REPL_BASE 0
#define REPL_COUNT 1
/*#define REPL_SELECTED 2*/ /* For an ALT replicator, the selected value */
/*}}}*/
/*{{{  constants used when generating repl PAR */
/* Offsets from wptr when setting up repl PAR */
#define MIN_REPLPAR_SLOTS 3  /* Minimum number of slots required to set up */
                             /* a repl PAR: par join, count & tempws  */
#define REPLPAR_WS_TEMP   2  /* Temp. slot for calculating repl wptrs */
#define REPLPAR_VS_TEMP   3  /* Temp. slot for calculating repl vsps  */

/* Special slots in a replicated PAR process' workspace */
/* These are above any variables local to the process */
#define MIN_REPLPAR_SPECIALS 2 /* Minimum number of special slots */
#define REPLPAR_REPLICATOR 0 /* Local value of replicator */
#define REPLPAR_STATICLINK 1 /* Link to workspace of invoking process */
#define REPLPAR_VSP        2 /* Vector space pointer (may not be needed) */
/*}}}*/
/*{{{  constants used inside predefines  */
#define INFINITY (0x7F800000)
/*}}}*/
/*{{{  constants used for T9000 error handling */
/*
(See T9000 specification part 2, section 30.6.2 "Reporting causes of traps",
in a table under the heading "Indication of the particular error which has
occurred"). The value 5 is named "err_interr".
*/
#define ERR_INTERR 5
/*}}}*/
/*{{{  constants used for label handling */
#define NO_LABEL (-1) /* any value which isn't a valid label number */
/*}}}*/
/*{{{  constants used to adjust the behaviour of 'prod' */
/* If a constant is smaller than this, it should be moved to the rhs */
#define TIMES_RHS_IS_FASTER    16

/* If a constant is greater than this, it should be moved to the lhs */
/* the reasoning behind this value is that it is 'half' the width of an INT */
/* These are calculated at run-time according to the type of the operand! */
/*#define TIMES_LHS_IS_FASTER_32 0x10000*/
/*}}}*/

/* This is used to check for links being placed on the top of the
   debugger's links */
#define NO_OF_LINKS 4
