/*
 * KROC translator: 
 *
 * $Source: /extra/play/cvs/kroc/src/target/arm/arm-asm.c,v $
 *
 * $Id: arm-asm.c,v 1.1 1999/09/27 21:24:00 rivimey Exp $
 *
 * arm-asm.c:
 * (C) Copyright 1999 R. Ivimey-Cook <ruthc@sharra.demon.co.uk>
 *
 */

#include <stdio.h>

#include "alpha-asm.h"
#include "asm-utils.h"
#include "stack-utils.h"
#include "regs.h"

char *rgname (int r);		/* generate register name */
FormatIntElementProc  formatdecimal;
FormatIntElementProc  formatimmdec;
FormatIntElementProc  formatldrcon;
FormatIntElementProc  formathex;

FormatStrElementProc   formatldrequ;
FormatStrElementProc   formatstr; /* not in procvec */
FormatStrElementProc   formatqstr;
FormatStrElementProc   formatsymbol;

FormatStrIntElementProc   formatindexbyn;
FormatStrStrElementProc   formatindexbyr;
FormatStrStrIntElementProc formatshftbyn;
FormatStrStrStrElementProc formatshftbyr;

static void Init(void)
{
}

static void Close(void)
{
}

static void Enter(void)
{
}

static void Leave(void)
{
}

static void Mnemonic(char *mnemonic, char *arg1, char *arg2, char *arg3)
{
    NewLine();
    SpaceTo(GetTabSize());
    Output(mnemonic);
    if (arg1) {
	SpaceTo(GetTabSize());
	Output(arg1);
    }
    if (arg2) {
    	Output(",");
	Output(arg2);
    }
    if (arg3) {
    	Output(",");
	Output(arg3);
    }
}

static void Comment(char *str)
{
    OutputTab();
    Output("; ");
    Output(str);
    NewLine();
}

static void GlobalFn(char *name)
{
    NewLine();
    SpaceTo(GetTabSize());
    Output("export");
    Output(name);
    NewLine();
    
    Output(name);
}

static void LocalFn(char *name)
{
    NewLine();
    Output(name);
}

static void FnExit()
{
    Comment("End of fn");
}

static void StartofFile(void)
{
    Comment("This file is automatically generated by Octran");
    NewLine();
    Comment("");
    NewLine();
    Comment((char*)GetCurrentFilename());
    NewLine();
}

static void EndofFile(void)
{
    NewLine();
    op("end");
    NewLine();
}

static void ImportSym(char *name)
{
    NewLine();
    SpaceTo(GetTabSize());
    Output("import");
    OutputTab();
    Output(name);
}

static void ExportSym(char *name)
{
    NewLine();
    SpaceTo(GetTabSize());
    Output("export");
    OutputTab();
    Output(name);
}

static void EmitLab(int r)
{
    NewLine();
    Output(formatdecimal(r));
    Output(" ");
}

static char *RefLab(enum ReferenceType d, int r)
{
    static char b[16];
    char *p = ((d == ForwardReference) ? "F" : "B");

    sprintf(b, "%s%d", p, r);
    return b;
}

static void ConstString(char *str)
{
    NewLine();
    SpaceTo(GetTabSize());
    Output("dcb");
    SpaceTo(GetTabSize());
    Output("\"");
    Output(str);
    Output("\"");
}

static void ConstData(int size, long value)
{
    NewLine();
    SpaceTo(GetTabSize());
    switch(size)
    {
        case 1:
            Output("dcb");
            break;
        case 2:
            Output("dcw");
            break;
        case 4:
            Output("dcd");
            break;
        case 8:
            Output("dcl");
            break;
        default:            
            break;
    }
    OutputTab();
    OutputF("0x%x\n", value);
}


void Equate (char *sym, char *val)			/* define symbol */
{
    NewLine();
    Output(formatsymbol(sym));
    SpaceTo(GetTabSize());
    Output("EQU");
    SpaceTo(GetTabSize());
    Output(val);
}


void Align (int sz)			/* align to address boundary */
{
    /* Align to (l)word address */
    Mnemonic("ALIGN", formatdecimal(sz), 0, 0);
}

void StackShift(void)
{
    comment("Stack shift");
    op2("mov", Creg, Breg); /* Creg := Breg */
    op2("mov", Breg, Areg); /* Breg := Areg */
}


/*	Details depend on target	*/
char *rgname (int r)		/* generate register name */
{
    static char names[4][4] = /* Evaluation stack */
    {"...", "r4", "r5", "r6"};

    if (r < 1)
    {
        fputs ("\n*** Integer stack underflow", stderr);
	return "X";
    }
    return names[r];
}


char *formatdecimal (int op)			/* generate decimal constant operand */
{
    static char out[32];

    sprintf (out, "%d", op);
    return out;
}


char *formatimmdec (int op)			/* generate immediate decimal constant operand */
{
    static char out[32];

    sprintf (out, "#%d", op);
    return out;
}

char *formatldrcon (int n)			/* generate immediate decimal constant operand */
{
    static char out[32];

    sprintf (out, "=%d", n);
    return out;
}

char *formatldrequ (char *s)			/* generate immediate string constant operand */
{
    static char out[256];

    sprintf (out, "=%s", s);
    return out;
}


char *formathex (int op)			/* generate hexadecimal constant operand */
{
    static char out[32];

    sprintf (out, "0x%x", op);
    return out;
}


char *formatstr (char *op)			/* generate string operand */
{
    static char out[256];

    sprintf (out, "%s", op);
    return out;
}


char *formatqstr (char *op)			/* generate quoted string operand */
{
    static char out[256];

    sprintf (out, "\"%s\"", op);
    return out;
}

char *formatsymbol (char *op)			/* generate quoted string operand */
{
    static char out[256];

    sprintf (out, "|%s|", op);
    return out;
}


static char *offset (int c)			/* word offset */
{
    static char out[32];

    if (c >= 0)
        sprintf (out, "%1d", c*BytesPerWord);
    else
        sprintf (out, "-%1d", -c*BytesPerWord);
    return out;
}

char *formatshftbyn (char *rg, char *op, int c)		/* generate shifter operand, const */
{
    static char out[32];

    sprintf (out, "%s %s #%d", rg, op, c);
    return out;
}

char *formatshftbyr (char *rg1, char *op, char *rg2)  /* generate shifter operand, reg */
{
    static char out[32];

    sprintf (out, "%s %s %s", rg1, op, rg2);
    return out;
}

char *formatindexbyn (char *r, int c)		/* generate scaled index operand */
{
    static char out[32];

    if (c == 0)
        sprintf (out, "[%s]", r);
    else
        sprintf (out, "[%s, #%s]", r, offset (c));
    return out;
}

char *formatindexbyr (char *r1, char *r2)		/* generate double index operand */
{
    static char out[32];

    sprintf (out, "[%s, %s]", r1, r2);
    return out;
}


AssemblerProcVec AsmARMASMProcs =
{
    Init,
    Close,
    Enter,
    Leave,
    Mnemonic,
    Comment,
    Equate,
    Align,
    GlobalFn,
    LocalFn,
    FnExit,
    StartofFile,
    EndofFile,
    ImportSym,
    ExportSym,
    ConstData,
    ConstString,
    StackShift,
    EmitLab,
    RefLab,

    {	formatdecimal, formatimmdec, formathex, formatldrcon },
    {	formatsymbol, formatqstr, formatldrequ },
    {	formatindexbyn, 0 },
    {	formatindexbyr, 0 },
    {	formatshftbyn, 0 },
    {	formatshftbyr, 0 }
};

