/*
 * KROC translator: Function definitions and prototypes
 *
 * $Source: /extra/play/cvs/kroc/src/octran.h,v $
 *
 * $Id: octran.h,v 1.2 1999/09/08 17:00:29 rivimey Exp $
 *
 * asm.h
 * (C) Copyright 1999 R. Ivimey-Cook <ruthc@sharra.demon.co.uk>
 *
 */

#ifndef ARM_ASM_H
#define ARM_ASM_H

/* Define the assembler modes and shortcuts */

/*----------------------------------------*/
/* Integer-only argument */
#define AMI_DECIMAL	0	/* Just the number, in decimal */
#define AMI_IMMDECIMAL	1	/* An immediate constant: Hash then the number, in decimal */
#define AMI_HEX		2	/* A Hex constant: Just the number, in hexadecimal */
#define AMI_LDRCON	3	/* LDR op : equals then the number */

#define AMI_NUMITEMS	4


/*----------------------------------------*/
/* String-only argument */
#define AMS_SYMBOL	0	/* A symbol name */
#define AMS_STRING	1	/* A double-quote quoted string */
#define AMS_LDREQ	2	/* LDR op : equals then the symbol name */

#define AMS_NUMITEMS	3


/*----------------------------------------*/
/* String then Integer arguments */
#define AMSI_INDEX	0	/* Scaled index: register name, index value -- e.g. [r0, #3] */

#define AMSI_NUMITEMS	2


/*----------------------------------------*/
/* Two String arguments */
#define AMSS_INDEX	0	/* Scaled index: register name, register name -- e.g. [r0, r1] */

#define AMSS_NUMITEMS	2


/*----------------------------------------*/
/* Two String then Integer arguments */
#define AMSSI_SHIFT	0	/* Shifter op: register name, shifter-op, shift-by number -- e.g. r0 LSR #3 */

#define AMSSI_NUMITEMS	2


/*----------------------------------------*/
/* Three-String arguments */
#define AMSSS_SHIFT	0	/* Shifter op: register name, shifter-op, shift-by register -- e.g. r0 LSR r1 */

#define AMSSS_NUMITEMS	2


/*----------------------------------------*/

#define dec(val)		(currentAsm->asmIntFormats[AMI_DECIMAL])(val)
#define immdec(val)		(char*)(currentAsm->asmIntFormats[AMI_IMMDECIMAL])(val)
#define hex(val)		(currentAsm->asmIntFormats[AMI_HEX])(val)
#define ldrcon(val)		(currentAsm->asmIntFormats[AMI_LDRCON])(val)

#include "asm.h"

static inline char *ldreq(char *val)
{
    FormatStrElementProc *p = currentAsm->asmStrFormats[AMS_LDREQ];
    return p(val);
}

static inline char *string(char *val)
{
    FormatStrElementProc *p = currentAsm->asmStrFormats[AMS_STRING];
    return p(val);
}

static inline char *symbol(char *val)
{
    FormatStrElementProc *p = currentAsm->asmStrFormats[AMS_SYMBOL];
    return p(val);
}

static inline char *shiftByN(char *rg1,char *op,int val)
{
    FormatStrStrIntElementProc *p = currentAsm->asmStrStrIntFormats[AMSSI_SHIFT];
    return p(rg1,op,val);
}

static inline char *shiftByR(char *rg1,char *op,char *rg2)
{
    FormatStrStrStrElementProc *p = currentAsm->asmStrStrStrFormats[AMSSS_SHIFT];
    return p(rg1,op,rg2);
}

static inline char *indexByN(char *rg1,int val)
{
    FormatStrIntElementProc *p = currentAsm->asmStrIntFormats[AMSI_INDEX];
    return p(rg1,val);
}

static inline char *indexByR(char *rg1,char *rg2)
{
    FormatStrStrElementProc *p = currentAsm->asmStrStrFormats[AMSS_INDEX];
    return p(rg1,rg2);
}

#endif /* ARM_ASM_H */

