/*
 *  ASSEMBLY LANGUAGE MACROS (Alpha)
 *
 *
 */

#define INC_Reg(Reg)          mnem("addq"); Reg; dec(1); Reg;  NL;
#define DEC_Reg(Reg)          mnem("subq"); Reg; dec(1); Reg;  NL;
#define INC_Reg2(Src,Dest)    mnem("addq"); Src; dec(1); Dest; NL;
#define DEC_Reg2(Src,Dest)    mnem("subq"); Src; dec(1); Dest; NL;

#define CLR_Reg(Reg)          mnem("clr"); Reg; NL;
#define CALL_RTN(Rtn)         mnem("jsr"); L; str(Rtn); NL;

#define UNSGNED(Reg)          mnem("zap");  Reg; dec(240); Reg; NL;
#define SGNEXT(Reg)           mnem("sll");  Reg; dec(32);  Reg; NL;             \
			      mnem("sra");  Reg; dec(32);  Reg; nl; 

#define SET_EFLAG             mnem("or");  STATUS; dec(1); STATUS; NL;
#define CLR_EFLAG             mnem("and"); STATUS; str("0xfe"); STATUS; NL;
#define TST_EFLAG(Dest)       mnem("and"); STATUS; dec(1); Dest;   NL;

#define SET_HFLAG             mnem("or");  STATUS; dec(2); STATUS; NL;
#define CLR_HFLAG             mnem("and"); STATUS; str("0xfd"); STATUS; NL;
#define TST_HFLAG(Dest)       mnem("and"); STATUS; dec(2); Dest;   NL;

#define FCLR_Reg(Freg)        fmnem("cpys");  ZERO; ZERO; str(Freg);  NL;
#define FMOV_Reg(FSrc,FDest)  fmnem("cpys");  str(FSrc); str(FSrc); str(FDest); NL;
#define FNEG_Reg(FSrc,FDest)  fmnem("cpysn"); str(FSrc); str(FSrc); str(FDest); NL;
#define FABS_Reg(FSrc,FDest)  fmnem("cpys");  ZERO; str(FSrc); str(FDest); NL;

/*  call a general function obeying normal C coding conventions
 */
#define CCALL(Rtn) \
	mnem("lda"); PV; str(Rtn); nl;  \
	mnem("jsr"); RA; indirect(PVreg); str(Rtn); NL;  \
	mnem("ldgp"); GP; indirect0(RAreg); NL;


/*  FPCC_INTO:
 *  -----------
 *
 *    Move floating point (FPUCC) into integer register (Destreg), with conversion.
 *
 */

#define FPCC_INTO(Destreg)    mnem("clr"); Destreg; nl;                       \
			      fmnem("fbeq"); FPUCC; f(9); nl;                 \
			      fmnem("or"); Z; dec(1); Destreg; nl;            \
			      lab(9); nl;                                     



/*  PACK:
 *  ----
 *
 *    Forms a 64 bit result (Dest64) from the two 32 bit values (MS_Reg32, LS_Reg32).
 *
 *      MS_Reg32 - Long word value (forms the top of the destination)
 *      LS_Reg32 - Long word value (forms the bottom of the destination) 
 */
#define PACK(MS_Reg32, LS_Reg32, Dest64)                                      \
									      \
	mnem("sll"); MS_Reg32; dec(32);  T2; NL;                              \
	mnem("zap"); LS_Reg32; dec(240); T3; NL;                              \
	mnem("or");  T2;  T3;  Dest64;   NL;



/*  UNPACK:
 *  ------
 *
 *    Forms two 32 bit values (MS_Reg32, LS_Reg32) from a 64 bit value (Src64).
 *
 *      MS_Reg32 - Long word value (the top of Src64)
 *      LS_Reg32 - Long word value (the bottom of Src64) 
 */
#define UNPACK(MS_Reg32, LS_Reg32, Src64)                                     \
									      \
	mnem("zap");    Src64;      dec(240);  LS_Reg32;  NL;                 \
	mnem("sra");    Src64;      dec(32);   MS_Reg32;  NL;



#define LOAD_BYTE_Zero_Extended(Src_Reg, Dest_Reg)                            \
									      \
	mnem("ldq_u");  T1;   indirect0(Src_Reg);    NL;                       \
	mnem("lda");    T0;   indirect0(Src_Reg);    NL;                       \
	mnem("extbl");  T1;   T0;   str(Dest_Reg);  NL;                 


#define LOAD_BYTE_Sign_Extended(Src_Reg, Dest_Reg)                            \
									      \
	mnem("ldq_u");  T1;              indirect0(Src_Reg);         NL;       \
	mnem("lda");    T0;              indirect0(Src_Reg);         NL;       \
	mnem("addq");   T0;              dec(1);    T0;             NL;       \
	mnem("extqh");  T1;              T0;        str(Dest_Reg);  NL;       \
	mnem("srq");    str(Dest_Reg);   dec(56);   str(Dest_Reg);  NL;



#define SAVE_BYTE(Src_Reg, Dest_Reg)                                          \
									      \
	mnem("lda");    T0;         indirect0(Dest_Reg);   NL;                 \
	mnem("ldq_u");  T1;         indirect0(Dest_Reg);   NL;                 \
	mnem("insbl");  str(Src_Reg);  T0;     T2;        NL;                 \
	mnem("mskbl");  T1;         T0;        T1;        NL;                 \
	mnem("or");     T1;         T2;        T1;        NL;                 \
	mnem("stq_u");  T1;         indirect0(Dest_Reg);   NL;


#define READ_FROM_FPCR(addr)                                                  \
									      \
	mnem("trapb"); NL;                                                    \
	mnem("mf_fpcr"); FT; NL;                                              \
	fmnem("stt"); FT; addr; NL;                                           \
	mnem("trapb"); NL;                                                    \


#define WRITE_TO_FPCR(addr)                                                   \
									      \
	mnem("trapb"); NL;                                                    \
	fmnem("ldt"); FT; addr; NL;                                           \
	mnem("mt_fpcr"); FT; NL;                                              \
	mnem("trapb"); NL;                                                    \





/*   Assigns the symbolic value  (Lable1 - Lable2)  to ... 
 *
 *   The destination register assignment is completed by the calling code.
 *
 *
 */
#define SYMBOLIC_OFFSET(sym1, sym2)                                           \
									      \
	mnem("lda");  A; str(sym1); NL;                                       \
	mnem("lda"); T0; str(sym2); NL;                                       \
	mnem("subq"); A; T0;


