/*
 * KROC ARM translator: register, assembler and FPU definitions
 *
 * $Source: /extra/play/cvs/kroc/src/target/arm/regs.h,v $
 *
 * $Id: regs.h,v 1.1 1999/09/27 21:10:12 rivimey Exp $
 *
 * regs.h:
 * (C) Copyright 1998 R. Ivimey-Cook <ruthc@sharra.demon.co.uk>
 *
 *
 * Derived from alpha/regs.h:
 * (C) Copyright 1995 D.C. Wood <D.C.Wood@ukc.ac.uk>
 * University of Kent at Canterbury
 *   /u3/proj/kroc/develop/src/alpha/RCS/regs.h,v
 *   regs.h,v 1.6 1996/06/26 10:12:37 djb1 Exp
 *
 */

#define BitsPerWord     32              /* size-dependent */
#define BitsPerByte     8

#define BytesPerWord    (BitsPerWord/BitsPerByte)

/*{{{  occam kernel jump table entries*/
/* code segment address of kernel jump table */
#define K_IN        1
#define K_OUT       2
#define K_OUTWORD   3
#define K_OUTBYTE   4
#define K_RUNP      5
#define K_STARTP    6
#define K_ENDP      7
#define K_STOPP     8
#define K_PAUSE     9
#define K_LDTIMER   10
#define K_TIN       11
#define K_ALTWT     12
#define K_TALTWT    13
#define K_ENBC      14
#define K_ENBT      15
#define K_DISS      16
#define K_DISC      17
#define K_DIST      18
#define K_RANGERR   19
#define K_SETERR    20
#define K_NORM      21
#define K_UNPACKSN  22
#define K_POSTNORMSN 23
#define K_ROUNDSN   24
#define K_MOVE      25
#define K_DIV       26
#define K_REM       27
#define K_LDIV      28
#define K_MOVE_ZERO 29
#define K_MOVE_NONZERO 30
#define K_FPSQRT    31

#define W_TIME     (-5<<WSH)
#define W_TLINK    (-4<<WSH)
#define W_STATUS   (-3<<WSH)
#define W_LINK     (-2<<WSH)
#define W_IPTR     (-1<<WSH)

#define MOSTNEG_INT32 0x80000000
#define NOT_PROCESS MOSTNEG_INT32
#define Z_ENABLING  NOT_PROCESS + 1
#define Z_WAITING   NOT_PROCESS + 2
#define Z_READY     NOT_PROCESS + 3
#define Z_TIMESET    NOT_PROCESS + 1
#define Z_TIMENOTSET NOT_PROCESS + 2
/*}}}  */

/*      Registers                       */

/* #include "amacros.h" */

char *rgname();

/* leave r0 - r2 for temps, procedure calls etc */

#define Temp0           "r0"            /* - Temporary */
#define Temp1           "r1"            /* - Temporary */
#define Temp2           "r2"            /* - Temporary */
#define Temp3           "r3"            /* - Temporary */

#define Areg            rgname(QueryStack()) /* Transputer evaluation stack in r4 - r6 */
#define Breg            rgname(QueryStack() - 1)
#define Creg            rgname(QueryStack() - 2)
#define Wptr            "r7"            /* - Workspace pointer */
#define Fptr            "r8"            /* - Front pointer  */
#define Bptr            "r9"            /* - Back pointer   */
#define Status          "r10"           /* - Error status bits */
                                        /*  [r10 also APCS stack limit] */
#define Tptr            "r11"           /* - Time Queue pointer */
#define Vptr            "r12"           /* - Vector pointer */
                                        /*  [r13 == APCS stack ptr] */

#define LRreg           "lr"            /* lr (r14) - Return Address */
#define PCreg           "pc"            /* pc (r15) - program counter */

#define CCreg           "cpsr"          /*  */

#define Trans_TRUE      immdec(1)          /* Transputer Booleans */
#define Trans_FALSE     immdec(0)


#ifdef FLOAT

/*      Floating-point registers        */

#define FPUSH           ++fstackp
#define FPOP            --fstackp

#define FAreg           fname(fstackp)
#define FBreg           fname(fstackp - 1)
#define FCreg           fname(fstackp - 2)
#define FTreg           "$f3"                 /* temp and function results */


#define FA              str(FAreg)
#define FB              str(FBreg)
#define FC              str(FCreg)
#define FT              str(FTreg)

#define FSR             str("$fsr")
#define FPtemp          str("SaveFP")

#endif

