/*
 * KROC ALPHA translator - Instructions concerned with run-time errors
 *                 (including floating point)
 *
 * $Source: /extra/play/cvs/kroc/src/target/alpha/f_error.c,v $
 *
 * $Id: f_error.c,v 1.1 1999/09/27 21:10:12 rivimey Exp $
 *
 * (C) Copyright 1995 D.C. Wood <D.C.Wood@ukc.ac.uk>
 * University of Kent at Canterbury
 *
 * ALPHA target only.
 */

static char f_error_c_rcsid[] = "$Id: f_error.c,v 1.1 1999/09/27 21:10:12 rivimey Exp $";

#include "octran.h"

/*
 *  SETERR:
 *  ------
 *   Ref: (T9000 ISM  Page )
 *
 *     A' <- A
 *     B' <- B
 *     C' <- C
 *     ErrorFlag' <- True
 *     
 */
void SETERR ()
{
	switch (o_checks)
	{
	case CH_NONE:
	case CH_SETFLAG:
		SET_EFLAG
		break;
	case CH_TRAP:
	case CH_HALT:
		GEN_TRAP(T_SETERR);
		break;
	}
}

/*
 *  TESTERR:
 *  -------
 *
 *     A' <- ~ ErrorFlag
 *     B' <- A
 *     C' <- B
 *     ErrorFlag' <- clear
 *     
 *     
 *     
 */
void TESTERR ()
{       PUSH;
	switch (o_checks)
	{
	case CH_NONE:
	case CH_TRAP:
		dummy("testerr"); COMMENT("remember NEVER"); NL;
		cond = C_NEVER;
		break;
	case CH_SETFLAG:
	case CH_HALT:
		TST_EFLAG(A)
		CLR_EFLAG
		cond = C_NE;
	}
}

/*
 *  STOPERR:
 *  -------
 *
 *     if ErrorFlag
 *      then  
 *       A' <- Undefined
 *       B' <- Undefined
 *       C' <- Undefined
 *       W[ ] <- i
 *      else
 *       A' <- A
 *       B' <- B
 *       C' <- C
 *
 *     ErrorFlag' <- ErrorFlag
 *     
 */
void STOPERR ()
{       switch (o_checks)
	{
	case CH_NONE:
	case CH_TRAP:
	case CH_HALT:
		dummy("stoperr"); COMMENT("off"); NL;
		break;
	case CH_SETFLAG:
		TST_EFLAG(CC)
		mnem("beq"); CC; f(1); nl;
		CALL_KRN(K_STOPP);
		lab(1); nl;
	}
}

/*
 *  CLRHALTERR:
 *  ----------
 *
 *     A' <- A
 *     B' <- B
 *     C' <- C
 *
 *     ErrorFlag' <- ErrorFlag
 *     HaltFlag'  <- False
 *     
 *     
 */
void CLRHALTERR ()
{       CLR_HFLAG
}

/*
 *  SETHALTERR:
 *  ----------
 *
 *     A' <- A
 *     B' <- B
 *     C' <- C
 *
 *     ErrorFlag' <- ErrorFlag
 *     HaltFlag'  <- True
 *     
 *     
 *     
 */
void SETHALTERR ()
{       SET_HFLAG
}

/*
 *  TESTHALTERR:
 *  ----------
 *
 *     A' <- HaltFlag
 *     B' <- A
 *     C' <- B
 *
 *     ErrorFlag' <- ErrorFlag
 *     HaltFlag'  <- HaltFlag
 *     
 *     
 *     
 */
void TESTHALTERR ()
{       PUSH;
	TST_HFLAG(A)
	cond = C_NE;
}

/*
 *  FPCHKERR:
 *  --------
 *   Ref: (T9000 ISM  Page )
 *
 *     
 *     
 *     
 */
void FPCHKERR ()
{
	/* assume floating point errors trapped by software interrupt to OS */
	fround (R_N);
}

/*
 *  FPTESTERR:
 *  ---------
 *   Ref: (T9000 ISM  Page )
 *
 *     
 *     
 *     
 */
void FPTESTERR ()
{       PUSH;
	/* assume floating point errors trapped by software interrupt to OS */
	fround (R_N);
}

/*
 *  FPUSETERR:
 *  ---------
 *     
 */
void FPUSETERR ()
{       switch (o_checks)
	{
	case CH_NONE:
	case CH_SETFLAG:
	case CH_HALT:
		dummy("fpuseterr"); COMMENT("off"); NL;
		break;
	case CH_TRAP:
		GEN_TRAP(T_FLOAT);
		COMMENT("force floating-point error"); NL;
		break;
	}
}

/*
 *  FPUCLRERR:
 *  ---------
 *     
 */
void FPUCLRERR ()
{       switch (o_checks)
	{
	case CH_NONE:
	case CH_TRAP:
	case CH_SETFLAG:
	case CH_HALT:
		dummy("fpuclrerr"); COMMENT("off"); NL;
		break;
	}
}
