/*
 * KROC ALPHA translator - Long arithmetic instructions
 *
 * $Source: /extra/play/cvs/kroc/src/target/alpha/f_long.c,v $
 *
 * $Id: f_long.c,v 1.1 1999/09/27 21:10:12 rivimey Exp $
 *
 * (C) Copyright 1995 D.C. Wood <D.C.Wood@ukc.ac.uk>
 * University of Kent at Canterbury
 *
 * ALPHA target only.
 */

static char f_long_c_rcsid[] = "$Id: f_long.c,v 1.1 1999/09/27 21:10:12 rivimey Exp $";

#include "octran.h"


/*
 *  LADD:
 *  ----
 *   Ref (T9000 ISM  Page 350)  [CSL - Areg after Overflow is not fixed-up]
 *
 *     if (sum > MAXINT)
 *       Areg' <- sum - 2^BitsPerWord
 *       ... IntegerOverflow
 *     else if (sum < MININT) 
 *       Areg' <- sum + 2^BitsPerWord
 *       ... IntegerOverflow
 *     else
 *       Areg' <- sum
 *     
 *        where sum = Areg + Breg + Creg[bit 0]
 */
void LADD ()
{
	/* sum = Areg + Breg + Creg[bit 0] */
	mnem("and");  C; dec(1); C; NL;
	mnem("addq"); A; B; A; NL;
	mnem("addq"); A; C; C; NL;

	if (o_checks)
	    {
		mnem("srl"); C; dec(BitsPerWord-1); T1; NL; 
		mnem("srl"); C; dec(BitsPerWord);   T0; NL; 
		mnem("and"); T0; dec(1); T0; NL;
		mnem("and"); T1; dec(1); T1; NL;
		mnem("xor"); T0; T1; T2; nl;
		mnem("beq"); T2; f(1); NL;
		GEN_TRAP(T_OFLOW);
		lab(1); nl;
	    }

	SGNEXT(C)

	pop (2);
}

/*
 *  LDIV:
 *  ----
 *   Ref (T9000 ISM  Page 361)
 *
 *     if (Creg(unsigned) >= Areg(unsigned))
 *       ... IntegerOverflow
 *     else
 *       Areg'(unsigned) <- LONG(unsigned) / Areg(unsigned)
 *       Breg'(unsigned) <- LONG(unsigned) rem Areg(unsigned)
 *
 *         where LONG = (Creg(unsigned) * 2^BitsPerWord) + Breg(unsigned)
 */
void LDIV ()                                    /* muldiv == _SW? */
{
	if (o_checks)
	{   /* if (Creg(unsigned) >= Areg(unsigned))  ...IntegerOverflow */
	    mnem("cmpule"); A; C; T0; NL;
	    mnem("beq"); T0; f(1); NL;
	    GEN_TRAP(T_OFLOW);
	    mnem("br"); f(2); NL;
	}
	/* else */
lab(1); nl;
	/* LONG = (Creg(unsigned) * 2^BitsPerWord) + Breg(unsigned) */
	mnem("sll"); C; dec(BitsPerWord); C; NL;
	UNSGNED(A)
	UNSGNED(B)
	mnem("or");  C; B; T0; NL;

	/* Areg'(unsigned) <- LONG(unsigned) / Areg(unsigned) */
	mnem("divqu"); T0; A; T1; NL;

	/* Breg'(unsigned) <- LONG(unsigned) rem Areg(unsigned) */
	mnem("mulq"); T1; A; B; NL;
	mnem("subq"); T0; B; C; NL;
	mnem("mov"); T1; B; NL;
lab(2); nl;
	ishuff = 1;
	POP;
}

/*
 *  LMUL:
 *  ----
 *   Ref (T9000 ISM  Page 375)
 *
 *       Areg'(unsigned) <- PROD(unsigned) rem 2^BitsPerWord
 *       Breg'(unsigned) <- PROD(unsigned) / 2^BitsPerWord
 *
 *         where PROD = (Breg(unsigned) * Areg(unsigned)) + Creg(unsigned)
 *     
 */
void LMUL ()
{
	/* PROD = (Breg(unsigned) * Areg(unsigned)) + Creg(unsigned) */
	UNSGNED(A)
	UNSGNED(B)
	UNSGNED(C)
	mnem("mulq"); B; A; T0; NL;
	mnem("addq"); T0; C; T0; NL;

	/* Breg'(unsigned) <- PROD(unsigned) / 2^BitsPerWord */
	mnem("srl"); T0; dec(BitsPerWord); C; NL;

	/* Areg'(unsigned) <- PROD(unsigned) rem 2^BitsPerWord */
	/* Carl's method
	mnem("sll"); C; dec(BitsPerWord); B; NL;
	mnem("subq"); T0; B; B; NL;
	*/
	mnem("zapnot"); T0; hex(0xf); B; NL;

	ishuff = 1;
	POP;
}

/*
 *  LSHR:
 *  ----
 *   Ref (T9000 ISM  Page 378)
 *
 *     Areg' <- (LONG >> Areg(unsigned)) rem 2^BitsPerWord 
 *     Breg' <- (LONG >> Areg(unsigned) / 2^BitsPerWord) rem 2^BitsPerWord
 *     
 *       where LONG = (Creg(unsigned) * 2^BitsPerWord)) + Breg(unsigned)
 *     
 */
void LSHR ()
{

	/* PACK 32 bit values from C&B into 64 bit T0 */
	PACK(C,B,T0)

	/* Shift T0 */
	mnem("srl"); T0; A; T0; NL;

	/* Special Case: shift of 64 Places. */
	mnem("and"); A; dec(64); T1; nl;
	mnem("cmovne"); T1; dec(0); T0; nl;

	/* UNPACK 64 bit T0 into 32 bit values in C&B */
	UNPACK(C,B,T0)


	ishuff = 1;
	POP;
}

/*
 *  LSHL:
 *  ----
 *   Ref (T9000 ISM  Page 377)
 *
 *     Areg' <- (LONG << Areg(unsigned)) rem 2^BitsPerWord 
 *     Breg' <- (LONG << Areg(unsigned) / 2^BitsPerWord) rem 2^BitsPerWord
 *     
 *       where LONG = (Creg(unsigned) * 2^BitsPerWord)) + Breg(unsigned)
 *     
 */
void LSHL ()
{
	mnem("beq"); A; f(1); nl;

	/* PACK 32 bit values from C&B into 64 bit T0 */
	PACK(C,B,T0)

	/* Shift T0 */
	mnem("sll"); T0; A; T0; NL;

	/* Special Case: shift of 64 Places. */
	mnem("and"); A; dec(64); T1; nl;
	mnem("cmovne"); T1; dec(0); T0; nl;

	/* UNPACK 64 bit T0 into 32 bit values in C&B */
	UNPACK(C,B,T0)

	lab(1); nl;

	ishuff = 1;
	POP;
}

/*
 *  LSUM:
 *  ----
 *   Ref (T9000 ISM  Page 380)
 *
 *     if (SUM(unsigned) > MostPosUnsigned)
 *        Areg' <- SUM(unsigned) - 2^BitsPerWord
 *        Breg' <- 1
 *     else
 *        Areg' <- SUM(unsigned)
 *        Breg' <- 0
 *     
 *         where  SUM = Areg(unsigned) + Breg(unsigned) + Creg[0]
 */
void LSUM ()
{
	/* SUM = Areg(unsigned) + Breg(unsigned) + Creg(0) */
	/* Areg' <- SUM(unsigned) */
	/* Breg' <- 0 */
	UNSGNED(A)
	UNSGNED(B)
	mnem("and");  C; dec(1); C; NL;
	mnem("addq"); A; C; A; NL;
	mnem("addq"); A; B; B; NL;
	mnem("clr");  C; NL;

	/* if (SUM(unsigned) > MostPosUnsigned) */
	mnem("srl"); B; dec(BitsPerWord); T1; nl;
	mnem("beq"); T1; f(1); NL;
	  /* Areg' <- SUM(unsigned) - 2^BitsPerWord */
	  UNSGNED(B)
	  /* Breg' <- 1 */
	  mnem("or");   Z; dec(1); C; NL;

	lab(1); nl;

	ishuff = 1;
	POP;
}

/*
 *  LSUB:
 *  ----
 *   Ref (T9000 ISM  Page 379)
 *
 *     if (DIFF > MAX)
 *       ...IntegerOverflow
 *     else if (DIFF < MIN)
 *       ...IntegerOverflow
 *     else
 *       Areg' <- DIFF
 *
 *         where  DIFF = (Breg - Areg) - Creg(0)
 */
void LSUB ()
{
	/* DIFF = (Breg - Areg) - Creg(0) */
	mnem("subq"); B; A; T0; NL;
	mnem("and"); C; dec(1); C; nl;
	mnem("subq"); T0; C; C; NL;

	if (o_checks)
	    {
		mnem("srl"); C; dec(31); T1; NL; 
		mnem("srl"); C; dec(32); T0; NL; 
		mnem("and"); T0; dec(1); T0; NL;
		mnem("and"); T1; dec(1); T1; NL;
		mnem("xor"); T0; T1; T2; nl;
		mnem("beq"); T2; f(1); NL;
		GEN_TRAP(T_OFLOW);
		lab(1);
	    }

	pop (2);
}

/*
 *  LDIFF:
 *  ----
 *   Ref (T9000 ISM  Page 360)
 *
 *     if (DIFF(unsigned) <= MAXUINT)
 *       Areg'(unsigned) <- DIFF(unsigned)
 *       Breg' <- 0
 *     else
 *       Areg'(unsigned) <- DIFF(unsigned) + 2^BitsPerWord
 *       Breg' <- 1
 *
 *         where  DIFF = (Breg(unsigned) - Areg(unsigned)) - Creg[0]
 *     
 */
void LDIFF ()
{
	/* DIFF = (Breg(unsigned) - Areg(unsigned)) - Creg[0] */
	UNSGNED(A)
	UNSGNED(B)
	mnem("and");  C; dec(1); C; nl;
	mnem("subq"); B; A; T0; NL;
	mnem("subq"); T0; C; T0; NL

	/* Breg' <- 0 */
	mnem("clr"); C; NL;

	/* if (DIFF(unsigned) <= MAXUINT) */
	mnem("srl"); T0; dec(BitsPerWord); T1; NL;
	mnem("beq"); T1; f(1); NL;

	/* Breg' <- 1 */
	mnem("or"); Z; dec(1); C; NL;

	/* Areg'(unsigned) <- DIFF(unsigned) + 2^BitsPerWord */
	mnem("sll");  C; dec(BitsPerWord); T1; NL;
	mnem("addq"); T0; T1; T0; NL;

lab(1);
	/* Areg' <- Answer */
	mnem("mov"); T0; B; NL;

	ishuff = 1;
	POP;
}
