/*
 * KROC ALPHA translator - Special instructions for the T800
 *
 * $Source: /extra/play/cvs/kroc/src/target/alpha/f_t800.c,v $
 *
 * $Id: f_t800.c,v 1.1 1999/09/27 21:10:12 rivimey Exp $
 *
 * (C) Copyright 1995 D.C. Wood <D.C.Wood@ukc.ac.uk>
 * University of Kent at Canterbury
 *
 * ALPHA target only.
 */

static char f_t800_c_rcsid[] = "$Id: f_t800.c,v 1.1 1999/09/27 21:10:12 rivimey Exp $";

#include "octran.h"


/*
 *  DUP:
 *  ---
 *   Ref: (T9000 ISM  Page 271)
 *
 *    A' <- A
 *    B' <- A
 *    C' <- B
 *
 */
void DUP ()
{       PUSH;
        mnem("mov"); B; A; NL;
}

/*
 *  FMUL:
 *  ----
 *   Ref: (T9000 ISM  Page 281)
 *
 *    if (Areg = MININT) and (Breg = MININT)
 *      ... Integer Overflow
 *    else
 *      A' <- (B * A) / 2^(Bitsperword-1)
 *    
 *    B' <- C
 *
 */
void FMUL ()
{       if (o_hardware)
        {       mnem("mov"); A; P0; NL;
                mnem("mov"); B; P1; NL;
                CALL_RTN(".mul")
                mnem("sra"); R0; dec(BitsPerWord-1); B; NL;
        }
        else
        {
                mnem("mulq"); B; A; T0; NL;
                mnem("sll"); T0; dec(33); T1; NL;
                mnem("cmplt"); T1; Z; T1; NL;     /* rounding bit */
                mnem("sra"); T0; dec(31); B; NL;
                mnem("addl"); B; T1; B; NL;
        }
        POP;
}

/*
 *  CRCWORD:
 *  -------
 *   Ref: (T9000 ISM  Page 251)
 *
 *    Areg' <- temp(BitsPerWord)
 *    Breg' <- Creg
 *    where
 *      temp(0) = Breg
 *      for i = 1 .. BitsPerWord
 *        temp(i) = (temp(i-1) << 1) + Areg[BPW-i] xor (Creg * temp(i-1)[BPW-1])
 *
 */
void CRCWORD ()
/* revised by mdp 25.10.95 */
{       
        mnem("bis"); Z; dec(32); T3; NL
        mnem("sll"); A; dec(32); A; NL;     /* move to top 32 bits */
lab(1);
        mnem("sll"); B; dec(1); T1; NL;        
        mnem("cmplt"); A; Z; T2; NL;        /* 1 if A is negative */
        mnem("or");  T2; T1; T1; nl;
        mnem("sll"); B; dec(32); B; NL;     /* move to top 32 bits */
        mnem("bge"); B; f(2); NL;
        mnem("xor"); C; T1; T1; nl;
lab(2);
        mnem("sll"); A; dec(1); A; NL;
        mnem("mov"); T1; B; nl;
        mnem("subl"); T3; dec(1); T3; NL;
        mnem("bne"); T3; b(1); NL;

        ishuff = 1;
        POP;

}

/*
 *  CRCBYTE:
 *  -------
 *   Ref: (T9000 ISM  Page 250)
 *
 *    Areg' <- temp(8)
 *    Breg' <- Creg
 *    where
 *      temp(0) = Breg
 *      for i = 1 .. 8
 *        temp(i) = (temp(i-1) << 1) + Areg[BPW-i] xor (Creg * temp(i-1)[BPW-1])
 *    
 *
 */
void CRCBYTE ()
/* revised by mdp 25.10.95 */
{
        mnem("bis"); Z; dec(8); T3; NL
        mnem("sll"); A; dec(32); A; NL;     /* move to top 32 bits */
lab(1);
        mnem("sll"); B; dec(1); T1; NL;
        mnem("cmplt"); A; Z; T2; NL;          /* 1 if A is negative */
        mnem("or");  T2; T1; T1; nl;
        mnem("sll"); B; dec(32); B; NL;     /* move to top 32 bits */
        mnem("bge"); B; f(2); NL;
        mnem("xor"); C; T1; T1; nl;
lab(2);
        mnem("sll"); A; dec(1); A; NL;
        mnem("mov"); T1; B; nl;
        mnem("subl"); T3; dec(1); T3; NL;
        mnem("bne"); T3; b(1); NL;

        ishuff = 1;
        POP;
}

/*
 *  BITCNT:
 *  ------
 *   Ref: (T9000 ISM  Page 237)
 *
 *    Areg' <- Breg + ...bits set in Areg
 *    Breg' <- Creg
 *
 */
void BITCNT ()
{
        mnem("sll"); A; dec(32); A; NL;
        mnem("srl"); A; dec(32); A; NL;  /* zero top 32 bits of 64 bit register. */
lab(1);
        mnem("blbc"); A; f(2); NL;
        mnem("addl"); B; dec(1); B; NL;
lab(2);
        mnem("srl"); A; dec(1); A; NL;
        mnem("bne"); A; b(1); NL;

        POP;
}

/*
 *  BITREVWORD:
 *  ----------
 *   Ref: (T9000 ISM  Page 239)
 *
 *    Areg' <- reversed Areg
 *
 */
void BITREVWORD ()
{
        mnem("clr"); T0; NL;
        mnem("bis"); Z; dec(1); T1; NL;
        mnem("sll"); T1; dec(BitsPerWord-1); T1; NL;
lab(1);
        mnem("blbc"); A; f(2); NL
        mnem("or"); T0; T1; T0; nl;
lab(2);
        mnem("srl"); T1; dec(1); T1; NL;
        mnem("srl"); A; dec(1); A; NL;
        mnem("bne"); T1; b(1); NL;
        mnem("mov"); T0; A; NL;
}

/*
 *  BITREVNBITS:
 *  -----------
 *   Ref: (T9000 ISM  Page 238)
 *
 *    Areg'[0..Areg-1] <- reversed Breg[0..Areg-1]
 *    Areg'[Areg..BitsPerWord-1] <- 0
 *    
 *    Breg' <- Creg
 */
void BITREVNBITS ()
{
        mnem("clr"); T0; NL;
        mnem("bis"); Z; dec(1); T1; NL;
        mnem("sll"); T1; A; T1; NL;
        mnem("srl"); T1; dec(1); T1; NL;
lab(1);
        mnem("blbc"); B; f(2); NL
        mnem("or"); T0; T1; T0; nl;
lab(2);
        mnem("srl"); T1; dec(1); T1; NL;
        mnem("srl"); B; dec(1); B; NL;
        mnem("bne"); T1; b(1); NL;
        mnem("mov"); T0; B; NL;

        POP;
}

/*
 *  WSUBDB:
 *  ------
 *   Ref: (T9000 ISM  Page 448)
 *
 *    Areg' <- Areg + (2 * Bytesperword * Breg)
 *    Breg' <- Creg
 */
void WSUBDB ()
{
        /* mnem("sll"); B; dec(3); B; NL;
        mnem("addl"); B; A; B; nl; */
        mnem("s8addl"); B; A; B; nl;
        POP;
}
