/*
 * KROC ALPHA translator: register, assembler and FPU definitions
 *
 * $Source: /extra/play/cvs/kroc/src/target/alpha/regs.h,v $
 *
 * $Id: regs.h,v 1.1 1999/09/27 21:10:12 rivimey Exp $
 *
 * (C) Copyright 1995 D.C. Wood <D.C.Wood@ukc.ac.uk>
 * University of Kent at Canterbury
 */

/*{{{  occam kernel jump table entries*/
/* code segment address of kernel jump table */
#define K_IN        1
#define K_OUT       2
#define K_OUTWORD   3
#define K_OUTBYTE   4
#define K_RUNP      5
#define K_STARTP    6
#define K_ENDP      7
#define K_STOPP     8
#define K_PAUSE     9
#define K_LDTIMER   10
#define K_TIN       11
#define K_ALTWT     12
#define K_TALTWT    13
#define K_ENBC      14
#define K_ENBT      15
#define K_DISS      16
#define K_DISC      17
#define K_DIST      18
#define K_RANGERR   19
#define K_SETERR    20
#define K_NORM      21
#define K_UNPACKSN  22
#define K_POSTNORMSN 23
#define K_ROUNDSN   24
#define K_MOVE      25
#define K_DIV       26
#define K_REM       27
#define K_LDIV      28
#define K_MOVE_ZERO 29
#define K_MOVE_NONZERO 30
#define K_FPSQRT    31

#define W_TIME     (-5<<WSH)
#define W_TLINK    (-4<<WSH)
#define W_STATUS   (-3<<WSH)
#define W_LINK     (-2<<WSH)
#define W_IPTR     (-1<<WSH)

#define NOT_PROCESS MOSTNEG_INT32
#define Z_ENABLING  NOT_PROCESS + 1
#define Z_WAITING   NOT_PROCESS + 2
#define Z_READY     NOT_PROCESS + 3
#define Z_TIMESET    NOT_PROCESS + 1
#define Z_TIMENOTSET NOT_PROCESS + 2
/*}}}  */

/*      Registers                       */

#include "amacros.h"

#define Areg            rgname(istackp) /* Transputer evaluation stack */
#define Breg            rgname(istackp - 1)
#define Creg            rgname(istackp - 2)
#define Wptr            "$13"           /* s4 - Workspace pointer */
#define Fptr            "$10"           /* s1 - Front pointer  */
#define Bptr            "$11"           /* s2 - Back pointer   */
#define Tptr            "$12"           /* s3 - Time Queue pointer  */
#define Vptr            "$8"            /* t7 - Vector pointer */
#define Temp            "$9"            /* s0 - Temporary */
#define Temp0           "$22"           /* t8 - Temporary */
#define Temp1           "$23"           /* t9 - Temporary */
#define Temp2           "$24"           /* t10 - Temporary */
#define Temp3           "$25"           /* t11 - Temporary */
#define RAreg           "$26"           /* ra - Return Address */
#define PVreg           "$27"           /* pv - Procedure value */
#define GPreg           "$29"           /* gp - Global pointer */
#define Zreg            "$31"           /* zero - Zero register */

#define KLINK           "$14"           /* s5 - Kernel link */
#define KTAB            "$15"           /* s6 - Kernel entry jump table address */
#define CCreg           "$0"            /* v0 - (Condition Code register) */
#define Status          "$4"            /* t3 - Error status bits */

#define RA              str(RAreg)
#define PV              str(PVreg)
#define GP              str(GPreg)
#define CC              str(CCreg)
#define A               str(Areg)       /* Transputer registers */
#define B               str(Breg)
#define C               str(Creg)
#define W               str(Wptr)
#define FP              str(Fptr)
#define BP              str(Bptr)
#define R0              str("$0")       /* v0 - Results for C functions */
#define P0              str("$16")      /* a0 - Parameters for C functions */
#define P1              str("$17")      /* a1 - Parameters for C functions */
#define P2              str("$18")      /* a2 - Parameters for C functions */
#define P3              str("$19")      /* a3 - Parameters for C functions */
#define P4              str("$20")      /* a4 - Parameters for C functions */
#define P5              str("$21")      /* a5 - Parameters for C functions */
#define L               str("$14")      /* s5 - Kernel Link (return address) */
#define KT              str(KTAB)       /* s6 - Kernel Entry point table */
#define T0              str(Temp0)      /* Temporaries */
#define T1              str(Temp1)
#define T2              str(Temp2)
#define T3              str(Temp3)
#define MIN             str("$9")       /* s0 - minint register */
#define STATUS          str(Status)     /* t3 - Error Flag etc. */
#define Z               str(Zreg)       /* Source of all zeros */
#define G1              str("$5")       /* t4 - For MOVE2D */
#define G2              str("$6")       /* t5 - For MOVE2D */
#define G3              str("$7")       /* t6 - For MOVE2D */

#define Trans_TRUE      dec(1)          /* Transputer Booleans */
#define Trans_FALSE     dec(0)

#define Fp              "%fp"           /* Frame pointer */


/*      Assembler-dependent stuff       */
#define CCHAR1          '!'             /* Line comment character */
#define CCHAR2          '#'             /* End of Line comment character */
#define CCHAR           ( (b_label[0] || b_symbol[0]) ? CCHAR2 : CCHAR1)   /* introduces comment */
#define DCCHAR          ( (b_label[0]) ? CCHAR2 : CCHAR1)                  /* introduces dummy opcodes */
#define HERE            "."             /* location counter */
#define LINEUP          32              /* output comment */


#ifdef FLOAT

/*      Floating-point registers        */

#define FPUSH           ++fstackp
#define FPOP            --fstackp

#define FAreg           fname(fstackp)
#define FBreg           fname(fstackp - 1)
#define FCreg           fname(fstackp - 2)
#define FTreg           "$f3"                 /* temp and function results */


#define FA              str(FAreg)
#define FB              str(FBreg)
#define FC              str(FCreg)
#define FT              str(FTreg)

/*      %f8 and %f9 currently unemployed        */

#define ZERO            str("$f31")     /* useful constants */
#define TWO             str("$f5")      /* no longer used */
#define HALFD           str("$f4")
#define TWO32D          str("$f6")
#define TWOM32D         str("$f7")
#define TWO52D          str("$f9")
#define HALFS           str("$f4")
#define TWO32S          str("$f6")
#define TWOM32S         str("$f7")
#define TWO23S          str("$f8")
#define FPUCC           str("$f14")

/*      etc.?   */

#define FSR             str("$fsr")
#define FPtemp          str("SaveFP")
#endif
