/*
 * KROC ALPHA translator - Instructions involved in ALT constructs and timeouts
 *
 * $Source: /extra/play/cvs/kroc/src/target/arm/f_alt.c,v $
 *
 * $Id: f_alt.c,v 1.1 1999/09/27 21:10:12 rivimey Exp $
 *
 * (C) Copyright 1995 D.C. Wood <D.C.Wood@ukc.ac.uk>
 * University of Kent at Canterbury
 *
 * ALPHA target only.
 */

static char f_alt_c_rcsid[] = "$Id: f_alt.c,v 1.1 1999/09/27 21:10:12 rivimey Exp $";

#include "arm-asm.h"
#include "octran.h"


/*  ALT:
 *  ---
 *  Ref: (T9000 ISM  Page 232)
 *
 *    word'[Wptr @ pw.state] <- Enabling.p
 *
 */
void ArmALT ()
{
    /* word'[Wptr @ pw.state] <- Enabling.p */
    op2("mov", Temp0, immdec(Enabling_p));
    op2("str", Temp0, indexByN(Wptr, pw_State));
}


/*  ALTWT:
 *  -----
 *  Ref: (T9000 ISM  Page 234)
 *
 *    if (word'[Wptr @ pw.state] != Ready.p)
 *    {
 *       word'[Wptr @ pw.state] <- Waiting.p
 *       ... deschedule process, wait for guard.
 *    }
 *
 *    word'[Wptr @ pw.Temp] <- NoneSelected.o
 *
 */
void ArmALTWT ()
{
    op2("ldr", Temp0, indexByN(Wptr, pw_State));
    op2("cmp", Temp0, immdec(0));
    op1("bne", labF(1));
    CALL_KRN(K_ALTWT);
    emitlab(1);
    op2("mov", Temp0, immdec(NoneSelected_o));
    op2("str", Temp0, indexByN(Wptr, pw_Temp));

    StackReset();
}


/*  ALTEND:
 *  ------
 *  Ref: (T9000 ISM  Page 233)
 *
 *    IptrReg' <- next instruction + word[Wptr @ pw.Temp]
 *
 */
void ArmALTEND ()
{
    op2("ldr", Temp0, indexByN(Wptr, pw_Temp));
    op2("adr", Temp1, labF(1));

    op3("add", PCreg, Temp0, Temp1);
    emitlab(1);
    StackReset();
}


/*  ENBC:
 *  ----
 *  Ref: (T9000 ISM  Page 272)
 *
 *    if (Areg != false)
 *    {
 *       if (word[Breg] = Notprocess.p)    word'[Breg] <- WdescRef
 *       else if (word[Breg] != WdescReg)  word'[Wptr @ pw.State] <- Ready.p
 *    }
 *
 *    Breg' <- Creg
 *
 */
void ArmENBC ()
{
    importsym("$enbc");

    op2("cmp", Areg, "#0");
    op1("blne", symbol("$enbc"));

    StackPop(1);
}


/*  ENBS:
 *  ----
 *  Ref: (T9000 ISM  Page 274)
 *
 *    if (Areg != false)  word'[Wptr @ pw.State] <- Ready.p
 *
 */
void ArmENBS ()
{
    op2("cmp", Areg, "#0");
    op2("movne", Temp0, immdec(Ready_p));
    op2("strne", Temp0, indexByN(Wptr, pw_State));
    /*ishuff += 1; */
}


/*  DISC:
 *  ----
 *  Ref: (T9000 ISM  Page 265)
 *
 *    if (Breg = false)  Areg' <- false
 *    else if (word[Creg] = NotProcess.p)  Areg' <- false
 *    else if (word[Creg] = WdescReg)
 *    {
 *       word'[Creg] <- NotProcess.p
 *       Areg' <- false
 *    }
 *    else if (word[Wptr @ pw.Temp] = NoneSelected.o)
 *    {
 *       word'[Wptr @ pw.Temp] <- Areg
 *       Areg' <- true
 *    }
 *    else  Areg' <- false
 *
 */
void ArmDISC ()
{
    CALL_KRN(K_DISC);
    /*ishuff = 1; */
    StackPop(2);
}


/*  DISS:
 *  ----
 *  Ref: (T9000 ISM  Page 268)
 *
 *    if (Breg = false)  Areg' <- false
 *    else if (word[Wptr @ pw.Temp] != NoneSelected.o)  Areg' <- false
 *    else
 *    {
 *       word'[Wptr @ pw.Temp] <- Areg
 *       Areg' <- true
 *    }
 *
 *    Breg' <- Creg
 *
 */
void ArmDISS ()
{
    CALL_KRN(K_DISS);
    /* ishuff += 1, */
    StackPop(1);
}


/*      Timer stuff     */


/*  TIN:
 *  ---
 *  Ref: (T9000 ISM  Page 439)
 *
 *    if not (ClockReg[Priority] after Areg)
 *    {
 *       word'[Wptr @ pw.State] <- Enabling.p
 *       word'[Wptr @ pw.Time]  <- (Areg + 1)
 *       ...Insert into timer list with time (Areg +1) and start next process
 *    }
 *
 */
void ArmTIN ()
{
    /* shufflecount += shuffle (istackp, 1,   * requires ONE item on stack */
    CALL_KRN(K_TIN);
    StackReset();
}


/*  TALT:
 *  ----
 *  Ref: (T9000 ISM  Page 435)
 *
 *    word'[Wptr @ pw.State]  <- Enabling.p
 *    word'[Wptr @ pw.TLink]  <- TimeNotSet.p
 *
 */
void ArmTALT ()
{
    /* word'[Wptr @ pw.State]  <- Enabling.p */
    /* word'[Wptr @ pw.TLink]  <- TimeNotSet.p */
    op2("mov", Temp0, immdec(Enabling_p));
    op2("mov", Temp1, immdec(TimeNotSet_p));
    op2("str", Temp0, indexByN(Wptr, pw_State));
    op2("str", Temp1, indexByN(Wptr, pw_TLink));
}


/*  TALTWT:
 *  ------
 *  Ref: (T9000 ISM  Page 436)
 *
 *    if (word[Wptr @ pw.State] = Ready.p)  word'[Wptr @ pw.Time] <- ClockReg[Priority]
 *    else if (word[Wptr @ pw.TLink] = TimeNotSet.p)
 *    {
 *       word'[Wptr @ pw.State] <- Waiting.p
 *       ... deschedule and wait for guard.
 *    }
 *    else if (word[Wptr @ pw.TLink] = TimeNotSet.p)
 *    {
 *       if (ClockReg[Priority] after word[Wptr @ pw.Time])
 *       {
 *          word'[Wptr @ pw.State] <- Ready.p
 *          word'[Wptr @ pw.Time]  <- ClockReg[Priority]
 *       }
 *       else
 *       {
 *          word'[Wptr @ pw.Time] <- (word[Wptr @ pw.Time] +1)
 *          ...insert into timer list with alarm time (word[Wptr @ pw.Time] +1)
 *          if (no-guards-ready)
 *          {
 *             word'[Wptr @ pw.State] <- Waiting.p
 *             ...deschedule and wait for guard
 *          }
 *       }
 *    }
 *
 *    word'[Wptr @ pw.Temp]  <- NoneSelected.o
 *
 */
void ArmTALTWT ()
{
    CALL_KRN(K_TALTWT);
    StackReset();
}


/*  ENBT:
 *  ----
 *  Ref: (T9000 ISM  Page 275)
 *
 *    if (Areg != false)
 *    {
 *       if (word[Wptr @ pw.TLink] = TimeNotSet.p)
 *       {
 *          word'[Wptr @ pw.TLink] <- TimeSet.p
 *          word'[Wptr @ pw.Time]  <- Breg
 *       }
 *       else if (word[Wptr @ pw.TLink] = TimeSet.p)
 *       {
 *          if (word[Wptr @ pw.Time] after Breg)
 *             word'[Wptr @ pw.Time] <- Breg
 *       }
 *    }
 *
 *    Breg <- Creg
 *
 */
void ArmENBT ()
{
    import("$enbt");

    op2("cmp", Areg, immdec(0));
    op1("blne", symbol("$enbt"));

    /* ishuff++; */
    StackPop(1);
}


/*  DIST:
 *  ----
 *  Ref: (T9000 ISM  Page 269)
 *
 *    if (Breg = false)  Areg' <- false
 *    else if (word[Wptr @ pw.TLink] = TimeNotSet.p)  Areg' <- false
 *    else if (word[Wptr @ pw.TLink] = TimeSet.p)
 *    {
 *       if not (word[Wptr @ pw.Time] after Creg)  Areg' <- false
 *       else if (word[Wptr @ pw.Temp] = NoneSelected.o)
 *       {
 *          word'[Wptr @ pw.Temp] <- Areg
 *          Areg' <- true
 *       }
 *       else  Areg' <- false
 *    }
 *    else  Areg' <- false
 *
 *    ...remove process from timer list
 *
 */
void ArmDIST ()
{
    CALL_KRN(K_DIST);
    /*ishuff ++; */
    StackPop(2);
}

