/*
 * KROC translator - Arithmetic and logical instructions
 *
 * $Source: /extra/play/cvs/kroc/src/target/arm/f_arith.c,v $
 *
 * $Id: f_arith.c,v 1.1 1999/09/27 21:10:12 rivimey Exp $
 *
 *
 * arm/f_arith.c:
 * (C) Copyright 1998 R. Ivimey-Cook <ruthc@sharra.demon.co.uk>
 *
 * Derived from alpha/f_arith.c:
 * (C) Copyright 1995 D.C. Wood <D.C.Wood@ukc.ac.uk>
 * University of Kent at Canterbury
 *   /u3/proj/kroc/develop/src/alpha/RCS/f_arith.c,v
 *   f_arith.c,v 1.14 1996/06/26 15:32:09 mdp2 Exp
 *
 */

static char f_arith_c_rcsid[] = "$Id: f_arith.c,v 1.1 1999/09/27 21:10:12 rivimey Exp $";

#include "arm-asm.h"
#include "octran.h"

static int imported_sdiv = 0;


/*   DIFF:
 *   ----
 *    Ref: (T9000 ISM  Page 264)
 *      
 *      A' <- B - A
 *      B' <- C
 */
void ArmDIFF ()
{
    op3("sub", Breg, Breg, Areg);
    StackPop(1);
}

/*   ADD:
 *   ---
 *    Ref: (T9000 ISM  Page 230)
 *      
 *      A' <- B +(checked) A
 *      B' <- C
 */
void ArmADD ()
{
    op3("add", Breg, Areg, Breg);
    /* if (strstr(b_comment, " check for replicator overflow") != NULL)
        StackPop(2);
    else */
        StackPop(1);
}

/*   PROD:
 *   ----
 *    Ref: (T9000 ISM  Page 399)
 *      
 *      A' <- A * B
 *      B' <- C
 */
void ArmPROD ()
{
    op3("mul", Breg, Areg, Breg);
    StackPop(1);
}

/*   SUB:
 *   ---
 *    Ref: (T9000 ISM  Page 429)
 *      
 *      A' <- B -(checked) A
 *      B' <- C
 */
void ArmSUB ()
{
    op3("sub", Breg, Breg, Areg);
    StackPop(1);
}

/*   REM:
 *   ---
 *    Ref: (T9000 ISM  Page 402)
 *      
 *      if (A = 0) or ( (A=-1) and (B = MININT) )
 *        ...IntegerOverflow
 *      else
 *        A' <- A rem B
 *      
 *      B' <- C
 */
void ArmREM ()
{
    if (!imported_sdiv)
    {
        import("__rt_sdiv");
        imported_sdiv = 1;
    }
    
    op2("mov", Temp0, Areg);
    op2("mov", Temp1, Breg);
    op1("bl", "__rt_sdiv");
    op2("mov", Breg, Temp1);
    comment("div returns r0 = div, r1 = mod");
    StackPop(1);
}

/*   DIV:
 *   ---
 *    Ref: (T9000 ISM  Page 270)
 *      
 *      if (A = 0) or ( (A=-1) and (B = MININT) )
 *        ...IntegerOverflow
 *      else
 *        A' <- B / A
 *      
 *      B' <- C
 */
void ArmDIV ()
{
    if (!imported_sdiv)
    {
        import("__rt_sdiv");
        imported_sdiv = 1;
    }
    
    op2("mov", Temp0, Areg);
    op2("mov", Temp1, Breg);
    op1("bl", "__rt_sdiv");
    op2("mov", Breg, Temp0);
    comment("div returns r0 = div, r1 = mod");
    
    StackPop(1);
}

/*   NOT:
 *   ---
 *    Ref: (T9000 ISM  Page 393)
 *      
 *      A' <- ~A
 */
void ArmNOT ()
{
    op3("rsb", Areg, Areg, immdec(0));
}

/*   XOR:
 *   ---
 *    Ref: (T9000 ISM  Page 451)
 *      
 *      A' <- B xor C  * really? *
 *      B' <- C
 */
void ArmXOR ()
{
    op3("eor", Breg, Areg, Breg);
    StackPop(1);
}

/*   SHR:
 *   ---
 *    Ref: (T9000 ISM  Page 414)
 *      
 *      if ( 0 <= A <= 31)
 *         A' <- B >> A
 *      else
 *         A' <- 0
 *       
 *      B' <- C 
 */
void ArmSHR ()
{
    char *ra = Breg,
         *rb = Areg;
    char *sh = shiftByR(ra, "lsr", rb);

    op2("mov", Breg, sh);
    StackPop(1);
}

/*   SHL:
 *   ---
 *    Ref: (T9000 ISM  Page 413)
 *      
 *      if ( 0 <= A <= 31)
 *         A' <- B << A
 *      else
 *         A' <- 0
 *       
 *      B' <- C 
 */
void ArmSHL ()
{       
    op2("mov", Breg, shiftByR(Breg, "lsl", Areg) );
    StackPop(1);
}

/*   AND:
 *   ---
 *    Ref: (T9000 ISM  Page 235)
 *      
 *      A' <- B and A
 *      B' <- C 
 */
void ArmAND ()
{
    op3("and", Breg, Breg, Areg);
    StackPop(1);
}

/*   OR:
 *   --
 *    Ref: (T9000 ISM  Page 394)
 *      
 *      A' <- B or A
 *      B' <- C       
 */
void ArmOR ()
{
    op3("orr", Breg, Breg, Areg);
    StackPop(1);
}

/*   SUM:
 *   ---
 *    Ref: (T9000 ISM  Page 430)
 *      
 *      A' <- B + A
 *      B' <- C       
 *      
 */
void ArmSUM ()
{
    op3("add", Breg, Breg, Areg);
    StackPop(1);
}

/*   MUL:
 *   ---
 *    Ref: (T9000 ISM  Page 390)
 *      
 *      A' <- B *(checked) A
 *      B' <- C       
 *      
 */
void ArmMUL ()
{
    op3("mul", Temp0, Breg, Areg);
    op2("mov", Areg, Temp0);
    StackPop(1);
}

