/*
 * KROC ARM translator - "Direct" instructions (with explicit operands)
 *
 * $Source: /extra/play/cvs/kroc/src/target/arm/f_direct.c,v $
 *
 * $Id: f_direct.c,v 1.1 1999/09/27 21:10:12 rivimey Exp $
 *
 * f_direct.c:
 * (C) Copyright 1998 R. Ivimey-Cook <ruthc@sharra.demon.co.uk>
 *
 *
 * Derived from f_direct.c:
 * (C) Copyright 1995 D.C. Wood <D.C.Wood@ukc.ac.uk>
 * University of Kent at Canterbury
 *   /u3/proj/kroc/develop/src/alpha/RCS/f_direct.c,v
 *   f_direct.c,v 1.27 1997/02/04 16:48:36 mdp2 Exp
 *
 *
 */

static char f_direct_c_rcsid[] = "$Id: f_direct.c,v 1.1 1999/09/27 21:10:12 rivimey Exp $";


#include "arm-asm.h"
#include "octran.h"

#ifdef FLOAT
int ftype[4], /* Types of f.p. registers */
#endif

/* will v fit in mask m (eg. 0xff) with a shift of 2n? */

__inline unsigned int rotate_left_by_2(unsigned int v)
{
    unsigned r = (v << 2) | (v >> (30));
    return r;
}

unsigned int is_arm_immed(unsigned int v, int m)
{
    int n;
    int nm = ~m, /* e.g. 0xffffff00 */

    n = 0, 
    while ((n < 14) && ((v & nm) == 0))
    {
        v = rotate_left_by_2(v);
        n++;
    }
    return (n != 14);
}

int big (int n, int bits)                     /* deal with big operands (12..32 bits) */
{
    if (bits == 8)
    {
        /* small means it will fit in 8 bits, modified by an even rotate, with sign independent */
        int k = (n >= 0) ? n : -n;
        
        if (!is_arm_immed(k, 0xff))
        {
	    /* ldimm is an assembler macro that chooses a good way to encode the constant */
            op2("ldimm", Temp0, decimal(n));
	    comment("big constant");
            return 1;
        }
    }
    else /* 12 bits excl. sign, straight */
    {
        if (!is_arm_immed(n, 0xfff))
        {
            op2("ldimm", Temp0, decimal(n));
	    comment("big constant");
            return 1;
        }
    }    
    
    return 0;
}



/*   J:
 *   -
 *    Ref: (T9000 ISM  Page 349)
 *     
 *     if (n = 0)
 *        ...Breakpoint TRAP
 *     else
 *        Iptr' <- Next Instruction + n
 *     
 *     
 *     
 */
void ArmJ ()
{
    reflab (0);
    op1("b", (b_operand));
    StackReset();

    op("LTORG"); /* tell asm it can put data tables here */
}

/*   LDLP:
 *   ----
 *    Ref: (T9000 ISM  Page 363)
 *     
 *     Areg' <- Wptr @ n
 *     Breg' <- Areg
 *     Creg' <- Breg
 *     
 */
void ArmLDLP ()
{
    int n = n_operand*BytesPerWord;
    
    StackPush(1);
    if (n_operand == 0)
    {
        op2("mov", Areg, Wptr);
    }
    else
    {
        if (big(n, 8))
            op3("add", Areg, Wptr, Temp0);
        else
            op3("add", Areg, Wptr, immdec(n));
    }
}

/*   PFIX:
 *   ----
 *    Ref: (T9000 ISM  Page)
 *     
 *     
 *     
 *     
 *     
 */
void ArmPFIX ()
{
    unimp ();
}

/*   LDNL:
 *   ----
 *    Ref: (T9000 ISM  Page 365)
 *     
 *     Areg' <- word[Areg @ n]
 *     
 *     
 *     
 */
void ArmLDNL ()
{
    op2("ldr", Areg, indexByN(Areg, n_operand)),  /* e.g. ldr r4, [r4, #8] */
}

/*   LDC:
 *   ---
 *    Ref: (T9000 ISM  Page 353)
 *     
 *     Areg' <- n
 *     Breg' <- Areg
 *     Creg' <- Breg
 *     
 */
void ArmLDC ()
{
    static int num_labdiffs = 0;
    
    StackPush(1);
    
    /* nflag signals if we've converted the operand into a number. If
     * we haven't, it could be a label difference instead
     */
    if (!nflag)
    {
        int marker;

        if (strchr(&b_operand[0], '-') != NULL)
        {
            /* Label difference */
            marker = strspn(&b_operand[0], "L0123456789");
            b_operand[marker] = 0, 
            
            op1("ldr", Areg);
            fprintf(out, "=(%s - %s)", &b_operand[0], &b_operand[marker+1]);
            if ((num_labdiffs++) == 5)
            {
                num_labdiffs = 0;
                /* horrible, but otherwise we end up with literal pool problems */
                op1("b", f(1)), comment("jump around literal pool: optimise away if poss");
                op("ltorg");
                lab(1);
            }
        }
        else
        {
            /* Equated value */
            op2("ldr", Areg, ldrequ(b_operand));
        }
    }
    else if (!big(n_operand, 8))
    {
        op2("mov", Areg, immdec(n_operand));
    }
        
}

/*   LDNLP:
 *   -----
 *    Ref: (T9000 ISM  Page 366)
 *     
 *     Areg' <- Areg @ n
 *     
 *     
 *     
 */
void ArmLDNLP ()
{
    int n = n_operand*BytesPerWord;
    if (n_operand != 0)
    {
        if (big(n, 8))
        {
            op3("add", Areg, Areg, Temp0);
        }
        else
        {
            op3("add", Areg, Areg, immdec(n));
        }
    }
    else
        comment("null operation");
}

/*   NFIX:
 *   ----
 *    Ref: (T9000 ISM  Page )
 *     
 *     
 *     
 *     
 *     
 */
void ArmNFIX ()
{
    unimp ();
}

/*   LDL:
 *   ---
 *    Ref: (T9000 ISM  Page 362)
 *     
 *     Areg' <- word[Wptr @ n]
 *     Breg' <- Areg
 *     Creg' <- Breg
 *     
 */
void ArmLDL ()
{
    int n = n_operand*BytesPerWord;
    if (strcmp (&buffer[B_PROC], "-- throw the result away ") == 0)
        dummy("ignored");
    else
    {
        StackPush();
        if (big(n, 12))
        {
            op2("ldr", Areg, indexByR(Wptr, Temp0)),  /* A = [Wptr + T] */
        }
        else
        {
            op2("ldr", Areg, indexbyN(Wptr, n_operand));
        }
    }
}

/*   ADC:
 *   ---
 *    Ref: (T9000 ISM  Page 229)
 *     
 *     Areg' <- Areg +(checked) n
 *     
 *     
 *     
 */
void ArmADC ()
{
    if (n_operand != 0)
    {
        if (big (n_operand, 8))
        {
            op3("add", Areg, Areg, Temp0);
        }
        else
        {
            op3("add", Areg, Areg, immdec(n_operand));
        }
    }
    else
    {
        dummy("null operation");
    }
    if (strstr (b_comment, " check for replicator overflow") != NULL)
        StackPop(1);
}

/*   CALL:
 *   ----
 *    Ref: (T9000 ISM  Page 241)
 *     
 *     Wptr' <- Wptr @ -4
 *     
 *     word'[Wptr' @ 0] <- Iptr
 *     word'[Wptr' @ 1] <- Areg
 *     word'[Wptr' @ 2] <- Breg
 *     word'[Wptr' @ 3] <- Creg
 *     
 *     Iptr' <- next instruction + n
 *     
 *     Areg' <- Iptr
 *     
 */
void ArmCALL ()                            /* unadjusted PC used internally */
{
    int i, j, temp;
    
    /* Should leave return address in A as well as W[0] -   */
    /* but occam never uses this.                           */
    /*                                                      */
    /* Should also use adjusted PCreg, for consistency -       */

    for (i = 0, i <= stcount, ++i)  /* ?numerical order => binary search */
        if (strcmp (b_operand, proctab[i].label) == 0)
            break;
    
    if (i > stcount)
    {
        fputs ("\n*** Call of undefined PROC/FUNCTION", stderr);
        
        op3("sub", Wptr, Wptr, immdec(4*BytesPerWord));

        for (j = 1, istackp > 0, ++j)
        {
            op2("str", Areg, indexByN(Wptr, j));
            StackPop(1);
        }
        op2("str", PCreg, indexByN(Wptr, 0));
        op1("b", b_operand);
        
        StackReset();
    }                                               /* call C function */
    else if (proctab[i].S_PTYPE == P_C_EXTERNAL)
    {
        (void)strcpy(b_comment, &b_operand[0]);

        /* don't have to save T stack, but r10 (sl) should be set properly */
        /* or we may get strange behaviour! Save r10 (Status) on the stack, */
        /* and restore a useful stack limit */
        
        op2("stmdb", "sp!", "{r10}");
        
        op2("ldr", "r10", "=StackLimitSave");
        op2("ldr", "r10", "[r10]");
        
        op1("bl", proctab[i].proc);

        /* restore our status reg */
        op2("ldmdb", "sp!", "{r10}");
        
        StackReset();
    }
    else                                     /* call occam PROC/FUNCTION */
    {
        op3("sub", Wptr, Wptr, immdec(4*BytesPerWord));
        
        for (j = 1, istackp > 0, ++j)
        {
            op2("str", Areg, indexByN(Wptr, j));
            StackPop(1);
        }
        op2("str", PCreg, indexByN(Wptr, 0));
        
        op1("b", (proctab[i].S_PTYPE == P_EXTERNAL) ? (proctab[i].proc) : (b_operand));

        istackp = proctab[i].S_ISTACK;
        fstackp = proctab[i].S_FSTACK;
        ishuff  = proctab[i].S_ISHUFF;
        temp    = proctab[i].S_FTYPES;
#ifdef FLOAT
        for (j = 1, j <= fstackp, ++j)  /* at most one returned (ORTMS p.10) */
        {
            ftype[j] = temp & 1;
            temp >>= 1;
        }
#endif
    }
    if (o_verbose)
    {
        fprintf (stderr, "\n... Return from %s %s %s\n",
                 b_operand, ptype[proctab[i].S_PTYPE], proctab[i].proc);
        fprintf (stderr, "    istack := %d", istackp);
#ifdef FLOAT
        fprintf (stderr, ", fstack := [", istackp);
        for (j = fstackp, j > 0, --j)
            fputc ((ftype[j] == F_D) ? 'D' : 'S', stderr);
        fputc (']', stderr);
#endif
        where ();
    }
}

/*   CJ:
 *   --
 *    Ref: (T9000 ISM  Page 249)
 *     
 *     if (Areg = 0)
 *        Iptr' <- next instruction + n
 *     else
 *        Iptr' <- next instruction
 *        Areg' <- Breg
 *        Breg' <- Creg
 *
 */
void ArmCJ ()
{
    reflab (istackp);

    op2("cmp", Areg, immdec(0));
    op1("beq", b_operand);
    ishuff += 1;
    StackPop(1);
}

/*   AJW:
 *   ---
 *    Ref: (T9000 ISM  Page 231)
 *     
 *     Wptr' <- @ n
 *     
 *     
 *     
 */
void ArmAJW ()
{
    int n = n_operand*BytesPerWord;
    if (n_operand != 0)
    {
        if (big (n, 8))
        {
            op3("add", Wptr, Wptr, Temp0);
        }
        else
        {
            op3("add", Wptr, Wptr, immdec(n));
        }
    }
    else
        dummy("null operation");
}

/*   EQC:
 *   ---
 *    Ref: (T9000 ISM  Page 277)
 *     
 *     if (Areg = n)
 *       Areg' <- true
 *     else
 *       Areg <- false
 */
void ArmEQC ()
{
    if (big (n_operand, 8))        /* loads Temp0 as side effect */
    {
        op2("cmp", Areg, Temp0);
    }
    else
    {
        op2("cmp", Areg, immdec(n_operand));
    }
    op2("moveq", Areg, Trans_TRUE);
    op2("movne", Areg, Trans_FALSE);
}

/*   STL:
 *   ---
 *    Ref: (T9000 ISM  Page 421)
 *     
 *     word'[Wptr @ n] <- Areg
 *     
 *     Areg' <- Breg
 *     Breg' <- Creg
 */
void ArmSTL ()
{
    int n = n_operand * BytesPerWord;
    
    if (big (n, 12))
    {
        op2("str", Areg, indexByR(Wptr, Temp0));
    }
    else
    {
        op2("str", Areg, indexByN(Wptr, n_operand));
    }
    StackPop(1);
}

/*   STNL:
 *   ----
 *    Ref: (T9000 ISM  Page 423)
 *     
 *     word'[Areg @ n] <- Breg
 *     
 *     Areg' <- Creg
 *     
 */
void ArmSTNL ()
{
    int n = n_operand*BytesPerWord;
    
    if (big (n, 12))
    {
        op2("str", Breg, indexByR(Areg, Temp0));
    }
    else
    {
        op2("str", Breg, indexByN(Areg, n_operand));
    }
    StackPop(2);
}

/*   OPR:
 *   ---
 *    Ref: (T9000 ISM  Page )
 *     
 *     
 *     
 *     
 *     
 */
void ArmOPR ()
{
    unimp ();
}


void ArmCALL_KRN (const int Kep)
{
    switch (Kep)
    {
        case K_ALTWT: CALL_RTN ("$altwt", break;
        case K_ENBC: CALL_RTN ("$enbc", break;
        case K_DISC: CALL_RTN ("$disc", break;
        case K_DISS: CALL_RTN ("$diss", break;
        case K_TIN: CALL_RTN ("$tin", break;
        case K_TALTWT: CALL_RTN ("$taltwt", break;
        case K_ENBT: CALL_RTN ("$enbt", break;
        case K_DIST: CALL_RTN ("$dist", break;
        case K_REM: CALL_RTN ("$rem", break;
        case K_DIV: CALL_RTN ("$div", break;
        case K_IN: CALL_RTN ("$in", break;
        case K_OUT: CALL_RTN ("$out", break;
        case K_OUTBYTE: CALL_RTN ("$outbyte", break;
        case K_OUTWORD: CALL_RTN ("$outword", break;
        case K_STOPP: CALL_RTN ("$stopp", break;
        case K_LDTIMER: CALL_RTN ("$ldtimer", break;
        case K_MOVE: CALL_RTN ("$move", break;
        case K_MOVE_ZERO: CALL_RTN ("$move", break;
        case K_MOVE_NONZERO: CALL_RTN ("$move", break;
        case K_FPSQRT: CALL_RTN ("$fpsqrt", break;
        case K_RANGERR: CALL_RTN ("$rangerr", break;
        case K_SETERR: CALL_RTN ("$seterr", break;
    }
}

void GEN_TRAP (const int Reason)
{
    switch (o_checks)
    {
        case CH_NONE:
            {
                break;
            }
        case CH_SETFLAG:
            {
                SET_EFLAG;
                op1("swi", "0x01");
                break;
            }
        case CH_TRAP:
        case CH_HALT:
            {
                op1("swi", "0x02");
                break;
            }
    }
}
