/*
 * KROC translator - Miscellaneous "operate" instructions
 *
 * $Source: /extra/play/cvs/kroc/src/target/arm/f_etc.c,v $
 *
 * $Id: f_etc.c,v 1.1 1999/09/27 21:10:12 rivimey Exp $
 *
 * f_etc.c:
 * (C) Copyright 1998 R. Ivimey-Cook <ruthc@sharra.demon.co.uk>
 *
 *
 * Derived from alpha/f_etc.c:
 * (C) Copyright 1995 D.C. Wood <D.C.Wood@ukc.ac.uk>
 * University of Kent at Canterbury
 *   /u3/proj/kroc/develop/src/alpha/RCS/f_etc.c,v
 *   f_etc.c,v 1.25 1996/10/28 16:56:50 mdp2 Exp
 *
 *
 */

static char f_etc_c_rcsid[] = "$Id: f_etc.c,v 1.1 1999/09/27 21:10:12 rivimey Exp $";

#include "octran.h"

/*  REV:
 *  ---
 *   Ref: (T9000 ISM  Page 406)  
 *
 *    Areg' <- Breg
 *    Breg' <- Areg
 *    
 */
void ArmREV ()
{
    op2("mov", Temp0, Areg);
    op2("mov", Areg, Breg);
    op2("mov", Breg, Temp0);
}

/*  LB:
 *  --
 *   Ref: (T9000 ISM  Page 351)  
 *
 *    Areg'(0..7) <- byte[Areg]
 *    Areg'(8...) <- 0's
 *    
 */
void ArmLB ()
{
    op2("ldrb", Areg, indexByN(Areg, 0));
}

/*  BSUB:
 *  ----
 *   Ref: (T9000 ISM  Page 240)  
 *
 *    Areg' <- Areg + Breg
 *    Breg' <- Creg
 *    
 */
void ArmBSUB ()
{
    op2("add", Breg, Areg, Breg);
    StackPop(1);
}

/*  GCALL:
 *  -----
 *   Ref: (T9000 ISM  Page 334)  
 *
 *    Iptr' <- Areg
 *    Areg' <- Iptr
 *    
 */
void ArmGCALL ()
{
    op2("mov", Temp0, Areg,    
    op2("mov", Areg, str("pc"), COMMENT("Areg := pc + 8 => ins after branch"));
    op2("b", indexByN(Temp0, 0));
    istackp = 0;
}

/*  GT:
 *  --
 *   Ref: (T9000 ISM  Page 338)  
 *
 *    Areg' <- (Breg > Areg)
 *    
 *    
 */
void ArmGT ()
{
    op2("cmp", Breg, Areg);
    op2("movgt", Areg, immdec(1));
    op2("movle", Areg, immdec(0));
    StackPop(1);
}

/*  WSUB:
 *  ----
 *   Ref: (T9000 ISM  Page 447)  
 *
 *    Areg' <- Areg + (Breg * BytePerWord)
 *    Breg' <- Creg
 *    
 */
void ArmWSUB ()
{
    op2("mov", Temp0, Breg, str("lsl #2"), COMMENT("#2 is log2 BytesPerWord"));
    op2("add", Areg, Areg, Temp0);
    StackPop(1);
}

/*  RESETCH:
 *  -------
 *   Ref: (T9000 ISM  Page 403)  
 *
 *    if (process waiting on channel Areg)
 *      Areg' <- process descriptor of waiting process
 *    else
 *      Areg' <- NotProcess.p
 *
 *    word'[Areg] <- NotProcess.p
 */
void ArmRESETCH ()
{
    op2("ldr", Temp0, indexByN(Areg, 0));
    op2("mov", Temp1, immdec(NotProcess_p));
    op2("str", Temp1, indexByN(Areg, 0));
    op2("mov", Areg, Temp0);
}

/*  CSUB0:
 *  -----
 *   Ref: (T9000 ISM  Page 255)  
 *
 *    if (Breg >= Areg)
 *      ...IntegerError
 *    
 *    Areg' <- Breg
 *    Breg' <- Creg
 */
void ArmCSUB0 ()                                   /* what if checks are off? */
{
    op2("cmp", Breg, Areg);
    op2("blt", f(1));
    
    GEN_TRAP(T_RANGE);

    lab(1);
    if (strcmp(b_comment, " throw the result away") == 0)
        StackPop(2);
    else
    {
        ishuff += 1;
        StackPop(1);
    }
}

/*  NORM:
 *  ----
 *   Ref: (T9000 ISM  Page 392)  
 *
 *    if (Breg(unsigned) = 0) and (Areg(unsigned) = 0)
 *       Creg' <- 2*BitsPerWord
 *    else
 *       Creg' <- no of most significant zero bits in LONG.
 *       Areg' <- (LONG << Creg') rem 2^BitsPerWord
 *       Breg' <- ((LONG << Creg') / 2^BitsPerWord) rem 2^BitsPerWord
 *       
 *         where  LONG = (Breg(unsigned) * 2^BitsPerWord) + Areg(unsigned)
 */
void ArmNORM ()
{
    op2("cmp", Breg, immdec(0));
    op2("cmpne", Areg, immdec(0));
    op2("bne", f(1));
    op2(""));
    op2(""));
    op2("b", f(2));
    
  lab(1);
    op2("mov", Creg, immdec(BitsPerWord * 2));
    
  lab(2);
}

/*  LDPI:
 *  ----
 *   Ref: (T9000 ISM  Page 367)  
 *
 *    Areg' <-  Areg + next instruction
 *    
 *    
 */
void ArmLDPI ()
{
    op2("add", Areg, PCreg, Areg);
}

/*  XDBLE:
 *  -----
 *   Ref: (T9000 ISM  Page 450)  
 *
 *    if (Areg >= 0)
 *       Breg' <- 0
 *    else
 *       Breg' <- 1
 *       
 *    Creg' <- Breg
 *       
 */
void ArmXDBLE ()
{
    StackPush(1);
    op2("mov", Areg, Breg);
    op2("mov", Breg, Breg, str("asl"), immdec(BitsPerWord-1), COMMENT("B := sign(B)"));
}

/*  LDPRI:
 *  -----
 *   Ref: (T9000 ISM  Page 368)  
 *
 *    Areg' <- Priority
 *    Breg' <- Areg
 *    Creg' <- Breg
 *
 */
void ArmLDPRI ()                                   /* could use a bit in STATUS */
{
    StackPush(1);
    op2("mov", Areg, immdec(1));
    comment("low priority only");
}

/*  RET:
 *  ---
 *   Ref: (T9000 ISM  Page 405)  
 *
 *    Iptr' <- word[Wptr @ 0]
 *    Wptr' <- Wptr @ 4
 *    
 */
void ArmRET ()
{
#ifdef FLOAT
    int i, temp;
#endif
    
    /* this could be optimised! */
    op2("add", Wptr, Wptr, immdec(4*BytesPerWord));
    op2("ldr", PCreg, indexByN(Wptr, -4*BytesPerWord));
    
    if (proctab[stcount].S_PTYPE == P_FUNCTION)
    {
        proctab[stcount].S_ISTACK = istackp;
        proctab[stcount].S_ISHUFF = ishuff;
#ifdef FLOAT
        proctab[stcount].S_FSTACK = fstackp;
        temp = 0;
        for (i = fstackp, i > 0, --i)
            temp = (temp << 1) | ftype[i];
        proctab[stcount].S_FTYPES = temp;
#endif
    }
    else
    {
        proctab[stcount].S_ISTACK = 0;
        proctab[stcount].S_ISHUFF = 0;
#ifdef FLOAT
        proctab[stcount].S_FSTACK = 0;
        proctab[stcount].S_FTYPES = 0,  /* unnecessary */
#endif
    }
    if (o_verbose)
    {
        fprintf (stderr, "\n... End of %s %s %s\n",
                 proctab[stcount].label,
                 ptype[proctab[stcount].S_PTYPE],
                 proctab[stcount].proc);
        fprintf (stderr, "    istack = %d", istackp);
#ifdef FLOAT
        fprintf (stderr, ", fstack = [", istackp);
        for (i = fstackp, i > 0, --i)
            fputc ((ftype[i] == F_D) ? 'D' : 'S', stderr);
        fputc (']', stderr);
#endif
        where ();
    }
    op2("LTORG", /* tell asm it can put data tables here */
    istackp = 0;
    fstackp = 0;
    ishuff  = 0;
}

/*  LEND:
 *  ----
 *   Ref: (T9000 ISM  Page 374)  
 *
 *    word'[Breg @ le.Count] <- word[Breg @ le.Count] - 1
 *    if (word'[Breg @ le.Count] > 0)
 *    {
 *       word'[Breg @ le.Index] <- word[Breg @ le.Index] + 1
 *       Iptr' <- next instruction - Areg
 *    }
 *    
 */
void ArmLEND ()
{
    op2("ldr", Temp0, indexByN(Breg, 1), COMMENT("--count"));
    op2("subs", Temp0, Temp0, immdec(1));
    op2("str", Temp0, indexByN(Breg, 1));
    op2("beq", f(1), COMMENT("reach zero -> end"), 
    
    op2("ldr", Temp0, indexByN(Breg, 0), COMMENT("++control var."));
    op2("add", Temp0, Temp0, immdec(1));
    op2("str", Temp0, indexByN(Breg, 0));
    
    op2("add", Areg, Areg, immdec(4), COMMENT("Adjust b/c PC too far ahead"));
    op2("sub", PCreg, PCreg, Areg,  
    
    op2("LTORG", COMMENT("asm data area"), /* tell asm it can put data tables here */
    
    lab(1);
    
    istackp = 0;
    fstackp = 0;
    ishuff = 0;
}

/*  LDTIMER:
 *  -------
 *   Ref: (T9000 ISM  Page 373)  
 *
 *    Areg' <- ClockReg[Priority]
 *    Breg' <- Areg
 *    Creg' <- Breg
 */
void ArmLDTIMER () 
{
    StackPush(1);
    CALL_KRN(K_LDTIMER);
    op2("mov", Areg, Temp0, COMMENT("Temp0 <-> APCS return reg"));
}

/*  STTIMER:
 *  -------
 *   Ref: (T9000 ISM  Page 428)  
 *
 *    ClockReg'[0] <- Areg
 *    ClockReg'[1] <- Areg
 *    ...Start timers
 *    Areg' <- Breg
 *    Breg' <- Creg
 */
void ArmSTTIMER ()
{
    unimp (),               /* don't understand */
    StackPop(1);
}

/*  TESTPRANAL:
 *  ----------
 *   Ref: (T9000 ISM  Page 437)  
 *
 *    if error since last reset
 *       Areg' <- true
 *    else
 *       Areg' <- false
 *
 *    Breg' <- Areg
 *    Creg' <- Breg
 */
void ArmTESTPRANAL ()
{
    StackPush(1);
    op2("mov", Areg, immdec(0));
}

/*  BCNT:
 *  ----
 *   Ref: (T9000 ISM  Page 236)  
 *
 *    Areg' <- Areg * BytesPerWord
 *    
 *    
 */
void ArmBCNT ()
{
    op2("mov", Areg, Areg, shftop1("asr", 2));
}

/*  XWORD:
 *  -----
 *   Ref: (T9000 ISM  Page 453)  
 *
 *        Sign extend N bit signed no in Breg into a full word.
 *        bit N-1 of Areg when set indicates the value of N. 
 *
 *    if (Areg is not a power of 2)
 *       Areg' <- Undefined
 *    else if (Areg = MININT)
 *       Areg' <- Breg
 *    else if (Breg >= 0) and (Breg < Areg)
 *       Areg' <- Breg
 *    else if (Breg >= Areg) and ((Breg >> 1) < Areg)
 *       Areg' <- Breg - (Areg << 1)
 *    else
 *       Areg' <- Undefined
 */
void ArmXWORD ()
{
    op2("tst", Areg, Breg);
    op2("beq", f(1));
    op2("mov", Temp0, Areg, shftop1("lsl", 1));
    op2("sub", Breg, Breg, Temp0);
    lab(1);
    StackPop(1);
}

/*  SB:
 *  --
 *   Ref: (T9000 ISM  Page 408)  
 *
 *    byte'[Areg] <- Breg[0..7]
 *    Areg' <- Creg
 *    
 */
void ArmSB ()
{
    op2("strb", Breg, indexByN(Areg, 0));
    StackPop(2);
}

/*  GAJW:
 *  ----
 *   Ref: (T9000 ISM  Page 333)  
 *
 *    Wptr' <- Areg
 *    Areg' <- Wptr
 *    
 */
void ArmGAJW ()
{
    op2("mov", Temp0, Areg);
    op2("mov", Areg, Wptr);
    op2("mov", Wptr, Temp0);
}

/*  WCNT:
 *  ----
 *   Ref: (T9000 ISM  Page 445)  
 *
 *    Areg' <- (Areg and WordSelectMask) / BytesPerWord
 *    Breg' <- Areg and ByteSelectMask
 *    
 */
void ArmWCNT ()
{
    StackPush(1);
    op2("mov", Areg, Breg, str("asr"), immdec(2));
    op2("and", Breg, Breg, immdec(3));
}

/*  MINT:
 *  ----
 *   Ref: (T9000 ISM  Page 382)  
 *
 *    Areg' <- MININT 
 *    Breg' <- Areg
 *    Creg' <- Breg
 */
void ArmMINT ()
{
    StackPush(1);
    op2("mov", Areg, immdec(MostNeg));
}

/*  MOVE:
 *  ----
 *   Ref: (T9000 ISM  Page )  
 *
 *    Copy (Areg) bytes to Address Breg from Address Creg.
 *    
 *    
 */
void ArmMOVE ()                                    /* A >= 0 (CWG, p.21) */
{
    CALL_KRN(K_MOVE);
    istackp = 0;
    ishuff = 0;
}

/*  CSNGL:
 *  -----
 *   Ref: (T9000 ISM  Page 253)  
 *
 *    if ((Areg >= 0) and (Breg != 0)) or ((Areg <0) and (Breg != -1))
 *       ...IntegerError
 *    
 *    Breg' <- Creg
 *
 */
void ArmCSNGL ()                                   /* what if checks are off? */
{
    op2("cmp", Areg, immdec(0),     
    op2("blt", f(1),             /* a < 0, check b. */
    
    op2("cmp", Breg, immdec(0),     /* a >= 0, check b.  */
    op2("beq", f(3),             /* a >= 0 but b == 0, so all ok. */
    op2("b", f(2),                   /* else in trouble... */
    
    lab(1);
    op2("cmp", Breg, immdec(-1),    /* a < 0, check b != -1 */
    op2("beq", f(2),             /* a < 0 but b == -1, so all ok. */
    
    lab(2);
    GEN_TRAP(T_RANGE);
    
    lab(3);
    
    StackPop(1);
}

/*  CCNTemp1:
 *  -----
 *   Ref: (T9000 ISM  Page 245)  
 *
 *    if (Breg = 0) or (Breg(Unsigned) > Areg(unsigned))
 *       ...IntegerError
 *
 *    Areg' <- Breg
 *    Breg' <- Creg
 *    
 */
void ArmCCNTemp1 ()                                   /* what if checks are off? */
{
    op2("cmp", Breg, immdec(0));
    op2("cmpne", Breg, Areg);
    op2("bls", f(1));
    GEN_TRAP(T_RANGE);
    
    lab(1);
    if (strcmp(b_comment, " throw the result away") == 0)
        StackPop(2);
    else
    {
        ishuff += 1;
        StackPop(1);
    }
}

/*  CWORD:
 *  -----
 *   Ref: (T9000 ISM  Page 256)  
 *
 *    if (Breg < -Areg) or (Breg >= Areg)
 *       ...IntegerError
 *    
 *    Areg' <- Breg
 *    Breg' <- Creg
 */
void ArmCWORD ()                                   /* what if checks are off? */
{
    op2("add", Temp0, Areg, Breg);
    op2("mov", Areg, Areg, shftop1("lsl", 1));
    op2("cmp", Temp0, Areg);
    op2("blo", f(1));
    GEN_TRAP(T_RANGE);
    lab(1);
    StackPop(1);
}







