/*
 * KROC translator - Long arithmetic instructions
 *
 * $Source: /extra/play/cvs/kroc/src/target/arm/f_long.c,v $
 *
 * $Id: f_long.c,v 1.1 1999/09/27 21:10:12 rivimey Exp $
 *
 * arm/f_long.c:
 * (C) Copyright 1998 R. Ivimey-Cook <ruthc@sharra.demon.co.uk>
 *
 * Derived from alpha/f_arith.c:
 * (C) Copyright 1995 D.C. Wood <D.C.Wood@ukc.ac.uk>
 * University of Kent at Canterbury
 *   /u3/proj/kroc/develop/src/alpha/RCS/f_long.c,v
 *   f_long.c,v 1.15 1996/08/09 13:39:30 mdp2 Exp
 *
 */

static char f_long_c_rcsid[] = "$Id: f_long.c,v 1.1 1999/09/27 21:10:12 rivimey Exp $";

#include "octran.h"


/*
 *  LADD:
 *  ----
 *   Ref (T9000 ISM  Page 350)  [CSL - Areg after Overflow is not fixed-up]
 *
 *     if (sum > MAXINT)
 *       Areg' <- sum - 2^BitsPerWord
 *       ... IntegerOverflow
 *     else if (sum < MININT) 
 *       Areg' <- sum + 2^BitsPerWord
 *       ... IntegerOverflow
 *     else
 *       Areg' <- sum
 *     
 *        where sum = Areg + Breg + Creg[bit 0]
 */
void ArmLADD ()
{
    /* sum = Areg + Breg + Creg[bit 0] */
    op2("and", Creg, Creg, immdec(1));
    op3("add), Areg, Breg, Areg);
    op3("add), Areg, Creg, Creg);

    StackPop(2);
}

/*
 *  LDIV:
 *  ----
 *   Ref (T9000 ISM  Page 361)
 *
 *     if (Creg(unsigned) >= Areg(unsigned))
 *       ... IntegerOverflow
 *     else
 *       Areg'(unsigned) <- LONG(unsigned) / Areg(unsigned)
 *       Breg'(unsigned) <- LONG(unsigned) rem Areg(unsigned)
 *
 *         where LONG = (Creg(unsigned) * 2^BitsPerWord) + Breg(unsigned)
 */
void ArmLDIV ()                                    /* muldiv == _SW? */
{
    unimp();
    StackPop(1);
}

/*
 *  LMUL:
 *  ----
 *   Ref (T9000 ISM  Page 375)
 *
 *       Areg'(unsigned) <- PROD(unsigned) rem 2^BitsPerWord
 *       Breg'(unsigned) <- PROD(unsigned) / 2^BitsPerWord
 *
 *         where PROD = (Breg(unsigned) * Areg(unsigned)) + Creg(unsigned)
 *     
 */
void ArmLMUL ()
{
    /* PROD = (Breg(unsigned) * Areg(unsigned)) + Creg(unsigned) */
    op2("smull", Temp0, Temp1, Breg, Areg);
    op2("adds",  Temp0, Temp0, Creg);
    op2("addcs", Temp1, immdec(1));
    
    /* Breg'(unsigned) <- PROD(unsigned) / 2^BitsPerWord */
    op2("mov", Breg, Temp0,    
    op2("mov", Creg, Temp1);
    
    ishuff = 1;
    StackPop(1);
}

/*
 *  LSHR:
 *  ----
 *   Ref (T9000 ISM  Page 378)
 *
 *     Areg' <- (LONG >> Areg(unsigned)) rem 2^BitsPerWord 
 *     Breg' <- (LONG >> Areg(unsigned) / 2^BitsPerWord) rem 2^BitsPerWord
 *     
 *       where LONG = (Creg(unsigned) * 2^BitsPerWord)) + Breg(unsigned)
 *     
 */
void ArmLSHR ()
{
    op2("movs", Creg, Creg, str("lsr"), Areg);
    op2("mov",  Breg, Breg, str("rrx"));
    /* rrx is only 1 bit... if more, use lsr again */
    op2("subs", Areg, Areg, immdec(1));
    op2("beq", f(1),    
    op2("mov",  Breg, str("lsr"), Areg);
    
    lab(1);
    
    ishuff = 1;
    StackPop(1);
}

/*
 *  LSHL:
 *  ----
 *   Ref (T9000 ISM  Page 377)
 *
 *     Areg' <- (LONG << Areg(unsigned)) rem 2^BitsPerWord 
 *     Breg' <- (LONG << Areg(unsigned) / 2^BitsPerWord) rem 2^BitsPerWord
 *     
 *       where LONG = (Creg(unsigned) * 2^BitsPerWord)) + Breg(unsigned)
 *     
 */
void ArmLSHL ()
{
    
    ishuff = 1;
    StackPop(1);
}

/*
 *  LSUM:
 *  ----
 *   Ref (T9000 ISM  Page 380)
 *
 *     if (SUM(unsigned) > MostPosUnsigned)
 *        Areg' <- SUM(unsigned) - 2^BitsPerWord
 *        Breg' <- 1
 *     else
 *        Areg' <- SUM(unsigned)
 *        Breg' <- 0
 *     
 *         where  SUM = Areg(unsigned) + Breg(unsigned) + Creg[0]
 */
void ArmLSUM ()
{
    op2("and",   Creg, Creg, dec(1));
    op2("adds",  Areg, Creg, Areg);
    op2("adds",  Breg, Areg, Breg);

    op2("movcc", Creg, immdec(0));
    op2("movcs", Creg, immdec(1));
    
    ishuff = 1;
    StackPop(1);
}

/*
 *  LSUB:
 *  ----
 *   Ref (T9000 ISM  Page 379)
 *
 *     if (DIFF > MAX)
 *       ...IntegerOverflow
 *     else if (DIFF < MIN)
 *       ...IntegerOverflow
 *     else
 *       Areg' <- DIFF
 *
 *         where  DIFF = (Breg - Areg) - Creg(0)
 */
void ArmLSUB ()
{
    /* DIFF = (Breg - Areg) - Creg(0) */
    op3("sub), Temp0, Breg, Areg);
    op2("and", Creg, Creg, immdec(1));
    op3("sub), Creg, Temp0, Creg);

    StackPop(2);
}

/*
 *  LDIFF:
 *  ----
 *   Ref (T9000 ISM  Page 360)
 *
 *     if (DIFF(unsigned) <= MAXUINT)
 *       Areg'(unsigned) <- DIFF(unsigned)
 *       Breg' <- 0
 *     else
 *       Areg'(unsigned) <- DIFF(unsigned) + 2^BitsPerWord
 *       Breg' <- 1
 *
 *         where  DIFF = (Breg(unsigned) - Areg(unsigned)) - Creg[0]
 *     
 */
void ArmLDIFF ()
{
    op2("and", Creg, Creg, immdec(1));
    op3("sub), Temp0, Breg, Areg);
    op3("sub), Temp0, Temp0, Creg);
    
    /* Breg' <- 0 */
    op2("mov", Creg, immdec(0));
    
    ishuff = 1;
    StackPop(1);
}
