/*
 * KROC ARM translator - Two-Dimensional block move instructions
 *
 * $Source: /extra/play/cvs/kroc/src/target/arm/f_move2d.c,v $
 *
 * $Id: f_move2d.c,v 1.1 1999/09/27 21:10:12 rivimey Exp $
 *
 *
 * arm/f_move2d.c:
 * (C) Copyright 1998 R. Ivimey-Cook <ruthc@sharra.demon.co.uk>
 *
 * Derived from alpha/f_move2d.c:
 * (C) Copyright 1995 D.C. Wood <D.C.Wood@ukc.ac.uk>
 * University of Kent at Canterbury
 *   /u3/proj/kroc/develop/src/alpha/RCS/f_move2d.c,v
 *   f_move2d.c,v 1.11 1996/01/08 17:44:19 mdp2 Exp
 *
 */

static char f_move2d_c_rcsid[] = "$Id: f_move2d.c,v 1.1 1999/09/27 21:10:12 rivimey Exp $";

#include "octran.h"

/*   MOVE2DINIT
 *   ----------
 *     Ref: (T9000 ISM  Page 387)
 *
 *       BMreg0' <- Areg
 *       BMreg1' <- Breg
 *       BMreg2' <- Creg
 *
 */
void ArmMOVE2DINIT ()                              /* (CWG, p.95) */
{
    op2("mov", T3, Areg, COMMENT("length"));
    op2("mov", Temp1, Breg, COMMENT("destination stride"));
    op2("mov", T2, Creg, COMMENT("source stride"));
    istackp = 0;
    ishuff = 0;
}

typedef enum { MOVE_ALL, MOVE_ZERO, MOVE_NONZERO } move_types;

void MOVE2D (move_types move_case)
{
    op2("mov", Temp0, Breg, COMMENT("dest address"));
    op2("stmfd", str("sp!, {r0}"));
    op2("mov", Temp0, Creg, COMMENT("source address"));
    op2("stmfd", str("sp!, {r0}"));
    op2("mov", Temp0, Areg, COMMENT("column count"));
    op2("stmfd", str("sp!, {r0}"));
    op2("mov", Temp0, dec(move_case), COMMENT("move type"));
    CALL_RTN("move2d");
    
    istackp = 0;
    ishuff = 0;
}




/*   MOVE2DALL
 *   ---------
 *     Ref: (T9000 ISM  Page 386)
 *
 *       if (Overlap)
 *          ... Undefined
 *       else
 *          for y = 0 .. (BMreg0(unsigned)-1)
 *            for x = 0 .. (Areg(unsigned) - 1)
 *              byte' [Breg + (y * BMreg1) + x] <- byte [Creg + (y * BMreg2) + x]
 *       
 */
void ArmMOVE2DALL ()                               /* A >= 0 (CWG, p.80) */
{
    MOVE2D (MOVE_ALL);
}


/*   MOVE2DNONZERO
 *   -------------
 *     Ref: (T9000 ISM  Page 388)
 *
 *       
 *       if (Overlap)
 *          ... Undefined
 *       else
 *          for y = 0 .. (BMreg0(unsigned)-1)
 *            for x = 0 .. (Areg(unsigned) - 1)
 *              if (byte [Creg + (y * BMreg2) + x] <> 0)
 *                byte' [Breg + (y * BMreg1) + x] <- byte [Creg + (y * BMreg2) + x]
 *
 */
void ArmMOVE2DNONZERO ()
{       
    MOVE2D (MOVE_NONZERO);
}

/*   MOVE2DZERO
 *   ----------
 *     Ref: (T9000 ISM  Page 389)
 *
 *       
 *       if (Overlap)
 *          ... Undefined
 *       else
 *          for y = 0 .. (BMreg0(unsigned)-1)
 *            for x = 0 .. (Areg(unsigned) - 1)
 *              if (byte [Creg + (y * BMreg2) + x] = 0)
 *                byte' [Breg + (y * BMreg1) + x] <- byte [Creg + (y * BMreg2) + x]
 *
 */
void ArmMOVE2DZERO ()
{       
    MOVE2D (MOVE_ZERO);
}





