/*
 * KROC ARM translator - Special instructions for the T800
 *
 * $Source: /extra/play/cvs/kroc/src/target/arm/f_t800.c,v $
 *
 * $Id: f_t800.c,v 1.1 1999/09/27 21:10:12 rivimey Exp $
 *
 *
 * arm/f_move2d.c:
 * (C) Copyright 1998 R. Ivimey-Cook <ruthc@sharra.demon.co.uk>
 *
 * Derived from alpha/f_t800.c:
 * (C) Copyright 1995 D.C. Wood <D.C.Wood@ukc.ac.uk>
 * University of Kent at Canterbury
 *   /u3/proj/kroc/develop/src/alpha/RCS/f_t800.c,v
 *   f_t800.c,v 1.10 1996/06/26 15:35:50 mdp2 Exp
 *
 */

static char f_t800_c_rcsid[] = "$Id: f_t800.c,v 1.1 1999/09/27 21:10:12 rivimey Exp $";

#include "octran.h"


/*
 *  DUP:
 *  ---
 *   Ref: (T9000 ISM  Page 271)
 *
 *    A' <- A
 *    B' <- A
 *    C' <- B
 *
 */
void ArmDUP ()
{
    StackPush(1);
    op2("mov", Areg, Breg);
}

/*
 *  FMUL:
 *  ----
 *   Ref: (T9000 ISM  Page 281)
 *
 *    if (Areg = MININT) and (Breg = MININT)
 *      ... Integer Overflow
 *    else
 *      A' <- (B * A) / 2^(Bitsperword-1)
 *    
 *    B' <- C
 *
 */
void ArmFMUL ()
{
    op2("smull", Temp0, Breg, Breg, A;);
    StackPop(1);
}

/*
 *  CRCWORD:
 *  -------
 *   Ref: (T9000 ISM  Page 251)
 *
 *    Areg' <- temp(BitsPerWord)
 *    Breg' <- Creg
 *    where
 *      temp(0) = Breg
 *      for i = 1 .. BitsPerWord
 *        temp(i) = (temp(i-1) << 1) + Areg[BPW-i] xor
 *                  (Creg * temp(i-1)[BPW-1])
 *
 */
void ArmCRCWORD ()
{       
    unimp();
    ishuff = 1;
    StackPop(1);
}

/*
 *  CRCBYTE:
 *  -------
 *   Ref: (T9000 ISM  Page 250)
 *
 *    Areg' <- temp(8)
 *    Breg' <- Creg
 *    where
 *      temp(0) = Breg
 *      for i = 1 .. 8
 *        temp(i) = (temp(i-1) << 1) + Areg[BPW-i] xor
 *                  (Creg * temp(i-1)[BPW-1])
 *    
 *
 */
void ArmCRCBYTE ()
{
    unimp();
    
    ishuff = 1;
    StackPop(1);
}

/*
 *  BITCNT:
 *  ------
 *   Ref: (T9000 ISM  Page 237)
 *
 *    Areg' <- Breg + ...bits set in Areg
 *    Breg' <- Creg
 *
 * Pseudo-code 1:
 *   while (A > 0) begin
 *     A := A >> 1
 *     if (carry set) B := B + 1
 *   end
 *
 * Pseudo-code 2:
 *   while (A > 0) begin
 *     A := (A - 1) and A
 *     B := B + 1
 *   end
 */
void ArmBITCNT ()
{
    op2("cmp", Areg, immdec(0));
    op2("beq", f(2));
lab(1);
    op2("movs", Areg, Areg, str("lsr #1"));
    op2("addcs", Breg, immdec(1),     COMMENT("add 1 to b if bit shifted set"));
    op2("cmp", Areg, immdec(0));
    op2("bne", b(1));
lab(2);
    
    StackPop(1);
}

void ArmBITCNT ()
{
    op2("cmp", Areg, immdec(0));
    op2("beq", f(2));
lab(1);
    op3("sub), Temp0, Areg, immdec(1););
    op2("ands", Areg, Temp0, Areg);
    op3("add), Breg, Breg, immdec(1));
    op2("bne", b(1));
lab(2);
    StackPop(1);
}

/*
 *  BITREVWORD:
 *  ----------
 *   Ref: (T9000 ISM  Page 239)
 *
 *    Areg' <- reversed Areg
 *
 *   Inlined operation: first swap halfwords with a rotate, then
 * swap in turn the bytes, nibbles, bit-pairs & bits. The smaller
 * chunks are performed using:
 *
 *   res =  ( ( A >> shift ) & const ) | (( A & const ) << shift)
 *
 *  for shift = 8, 4, 2, 1, while const is a mask value selecting bits.
 *
 *  Pseudo-code:
 *  
 *  Areg := (Areg << 16) | (Areg >> 16)             -- rotate 16
 *  shift := 8
 *  const := 0xff00ff
 *  do begin
 *    Areg :=  ( ( Areg >> shift ) & const ) | (( Areg & const ) << shift)
 *    shift := shift >> 1
 *    const := const eor (const << shift)
 *  while (shift > 0)
 * 
 * ARM Cycles (approx):   17
 */
void ArmBITREVWORD ()
{    
    op2("mov", Temp0, #ff,     COMMENT("Temp1 = temp reg, A = input value, Temp0 = constant"));
    op2("orr", Temp0, Temp0, 0xff00);
    
    COMMENT("swap halfwords & bytes");
    op2("and", Temp1, Temp0, Areg, ror #24);
    op2("and", Areg, Temp0, Areg, lsl #8);
    op2("orr", Areg, Temp1, Areg);
    op2("eor", Temp0, Temp0, Temp0, lsl #4);

    COMMENT("swap nibbles");
    op2("and", Temp1, Temp0, Areg, lsr #4);
    op2("and", Areg, Temp0, Areg);
    op2("orr", Areg, Temp1, Areg, lsl #4);
    op2("eor", Temp0, Temp0, Temp0, lsl #2);

    COMMENT("swap bit-pairs");
    op2("and", Temp1, Temp0, Areg, lsr #2);
    op2("and", Areg, Temp0, Areg);
    op2("orr", Areg, Temp1, Areg, lsl #2);
    op2("eor", Temp0, Temp0, Temp0, lsl #1);

    COMMENT("swap bits");
    op2("and", Temp1, Temp0, Areg, lsr #1);
    op2("and", Areg, Temp0, Areg);
    op2("orr", Areg, Temp1, Areg, lsl #1);
}

/*
 *  BITREVNBITS:
 *  -----------
 *   Ref: (T9000 ISM  Page 238)
 *
 *    Areg'[0..Areg-1] <- reversed Breg[0..Areg-1]
 *    Areg'[Areg..BitsPerWord-1] <- 0
 *    
 *    Breg' <- Creg
 *
 *   Shift Breg down, extracting the bottom bit on each
 * shift and putting it into a temporary, which is shifted up
 * each time.
 *
 * Pseudo-code:
 *   
 *   temp := 0
 *   while ( Areg > 0 ) begin
 *     temp := (Breg & 1) | (temp << 1)
 *     Breg := Breg >> 1
 *     Areg := Areg - 1
 *   end
 *   Breg = temp
 *
 * ARM Cycles (approx):
 *   Areg <> 0 : 4 + (Areg) * 6
 *   Areg == 0: 5
 *
 */
void ArmBITREVNBITS ()
{
    op2("mov", T2, dec(0));
    op2("cmp", Areg, dec(0));
    op2("beq", f(2));
lab(1);
    op2("and", Temp0, Breg, dec(1));
    op2("orr", T2, Temp0, T2, lsl #1);
    op2("mov", Breg, Breg, lsr #1);
    op2("subs", Areg, Areg, dec(1));
    op2("bne", b(1));
lab(2);
    op2("mov", Breg, T2);
    StackPop(1);
}

/*
 *  WSUBDB:
 *  ------
 *   Ref: (T9000 ISM  Page 448)
 *
 *    Areg' <- Areg + (2 * Bytesperword * Breg)
 *    Breg' <- Creg
 */
void ArmWSUBDB ()
{
    op2("mov", Breg, Breg, str("lsl #3"));
    op3("add), Breg, Areg, Breg);
    StackPop(1);
}
