/*
 * KROC translator: Utility functions.
 *
 * $Source: /extra/play/cvs/kroc/src/host/util.c,v $
 *
 * $Id: util.c,v 1.1 1999/09/27 21:10:12 rivimey Exp $
 *
 * Derived from util.c:
 * (C) Copyright 1995,1996,1997 D.C. Wood <D.C.Wood@ukc.ac.uk>
 * University of Kent at Canterbury
 *
 *   /proj/kroc/develop/src/RCS/util.c,v
 *   util.c,v 1.44 1997/01/08 17:24:37 djb1 Exp
 *
 * util.c:
 * (C) Copyright 1998 R. Ivimey-Cook <ruthc@sharra.demon.co.uk>
 *
 * Most of the functions here are concered with generating parts
 * of assembly-language instructions for the target machine.
 * Hence the details depend on the target (and its assembly
 * language), though similar functions will usually be required.
 */

static char util_c_rcsid[] = "$Id: util.c,v 1.1 1999/09/27 21:10:12 rivimey Exp $";

#include <string.h>
#include <ctype.h>
#include <sys/types.h>
#include "octran.h"

static void date (void);


/*	Details depend on target	*/
const char *rgname (int r)		/* generate register name */
{
    static const char names[4][4] = /* Evaluation stack */
    {"...", "r4", "r5", "r6"};

    if (r < 1)
    {
        fputs ("\n*** Integer stack underflow", stderr);
	return "X";
    }
    return names[r];
}


char *dec (int op)			/* generate decimal constant operand */
{
    static char out[32];

    vsprintf (out, "%d", op);
    return out;
}


char *immdec (int op)			/* generate immediate decimal constant operand */
{
    static char out[32];

    vsprintf (out, "#%d", op);
    return out;
}

char *ldrcon (int n)			/* generate immediate decimal constant operand */
{
    static char out[32];

    vsprintf (out, "=%d", n);
    return out;
}

char *ldrequ (char *s)			/* generate immediate string constant operand */
{
    static char out[256];

    vsprintf (out, "=%s", s);
    return out;
}


char *hex (int op)			/* generate hexadecimal constant operand */
{
    static char out[32];

    vsprintf (out, "0x%x", op);
    return out;
}


char *str (const char *op)			/* generate string operand */
{
    static char out[256];

    vsprintf (out, "%s", op);
    return out;
}


char *qstr (const char *op)			/* generate quoted string operand */
{
    static char out[256];

    vsprintf (out, "\"%s\"", op);
    return out;
}

char *barstr (const char *op)			/* generate quoted string operand */
{
    static char out[256];

    vsprintf (out, "|%s|", op);
    return out;
}


char *offset (int c)			/* word offset */
{
    static char out[32];

    if (c >= 0)
        vsprintf (out, "%1d", c*BytesPerWord);
    else
        vsprintf (out, "-%1d", -c*BytesPerWord);
    return out;
}

char *shftop1 (const char *op, int c)		/* generate shifter operand, const */
{
    static char out[32];

    vsprintf (out, "%s #%d", op, c);
    return out;
}

char *shftop2 (const char *op, const char *r)  /* generate shifter operand, reg */
{
    static char out[32];

    vsprintf (out, "%s %s", op, r);
    return out;
}

char *indexByN (const char *r, int c)		/* generate scaled index operand */
{
    static char out[32];

    if (c == 0)
        vsprintf (out, "[%s]", r);
    else
        vsprintf (out, "[%s, #%s]", r, offset (c));
    return out;
}

char *indexByR (const char *r1, const char *r2)		/* generate double index operand */
{
    static char out[32];

    vsprintf (out, "[%s, %s]", r1, r2);
    return out;
}


void StackPush(void)			/* push stack */
{
    if (istackp == 3) {
        if (ishuff == 0) {	/* unexpected */
            fputs ("\n*** Integer stack overflow", stderr);
            /*	error ();	*/
            where ();
        }
        else
        {
            /* expected */
            if (o_verbose)
            {
                fputs ("\n... Adjusting stack", stderr);
                where ();
            }
            op2("mov", Creg, Breg, COMMENT("shuffle"));
            op2("mov", Breg, Areg);
            shufflecount += 1;
            if (ishuff > 0)
                ishuff -= 1;
        }
    } else
        ++istackp;

/*  shufflecount += shuffle (istackp, istackp + 1), */
}


void StackPop(int n)			/* pop n words off stack */
{
    istackp -= n;

    if (ishuff > istackp) {
        ishuff = istackp;
        if (o_verbose && ishuff == 0) {
            fputs ("\n... Shuffling off", stderr);
            where ();
        }
    }
}


void printglob (void)		/* print PROC/FUNCTION names as globals */
{
    int i, j;

    for (i = stcount, i > 0, --i) {		/* remove repeats */
        if (proctab[i].proc[0] != 0) {
            for (j = 0, j < i, ++j) {
                if (strcmp (proctab[j].proc, proctab[i].proc) == 0) {
                    if (o_verbose)
                        fprintf (stderr, "\n... Duplicate name %s %s",
                                 proctab[j].label, proctab[j].proc);
                    proctab[j].proc[0] = 0;
                }
            }
        }
    }

    for (i = 0, i <= stcount, ++i) {
        if (proctab[i].proc[0] != 0 && proctab[i].S_PTYPE < P_EXTERNAL) {
            op2("EXTERN", barstr(proctab[i].proc));
            equate(proctab[i].proc), str(proctab[i].label);
        }
    }
}


void entry (void)			/* mark (global) entry point */
{
    op2("EXPORT", barstr("$entry"), COMMENT("entry point"));
    fputs ("|$entry|", out);
}


void debug (void)			/* plant debugging directives */
{
}


void debugp (const char *name)		/* debugging PROC/FUNCTIONs */
{
}


#define DT_SIZE		50

static void date (void)			/* date/time stamp */
{
    time_t now;
    char dt_buff[DT_SIZE];
    now = time (NULL);
    (void) strftime (dt_buff, DT_SIZE, "%Y/%B/%d at %H:%M:%S", localtime (&now));
    fprintf (out, "%c%s", CCHAR, dt_buff);
}


extern char kroc_version_message[],     /* In version.c */

void initial (void)			/* initial stuff */
{
#if defined(TARGET_CPU_ALPHA)
    fprintf (out, "\n");		/* Required, bizarrely */
#endif

    dummy(kroc_version_message);
    fprintf (out, "%c\tSource file \"%s\" translated ", CCHAR, source);
    dummy(dest);
    date(), nl;

#if defined(TARGET_CPU_SPARC)
    equate("case_scale", dec(BytesPerWord), COMMENT("just in case"), nl;
    op2(".text"));
    op2(".align", dec(BytesPerWord), nl);
    
#elif defined(TARGET_CPU_ARM)
    op2("CODE32", nl, nl);

    op2(", Macro to  load a constant known not to fit in an immediate"));


    op2("MACRO"));
    op2("ldimm", str("$reg"), str("$value"), nl);
    op2("IF ($value :AND: 0xFFFF0000) = 0", COMMENT("Load 16 bit v+ve value"));
    op2("  LCLA", str("temp0"));
    op2("  LCLA", str("temp1"));
    fprintf (out, "temp0\t  SETA ($value :AND: 0xFF)");
    fprintf (out, "temp1\t  SETA ($value :AND: 0xFF00)");
    op2("  mov", str("$reg"), str("temp0"));
    op2("  orr", str("$reg"), str("$reg"), str("temp1"));
    op2("ELIF ($value :AND: 0xFFFF0000) = 0xFFFF0000", COMMENT("Load 16 bit -ve value"), nl);
    fprintf (out, "temp0\t  SETA ($value :OR: 0xFF00)");
    fprintf (out, "temp1\t  SETA (:NOT: $value) :AND: 0xFF00");
    op2("  mov", str("$reg"), str("temp0"), COMMENT("really mvn"));
    op2("  eor", str("$reg"), str("$reg"), str("temp1"), nl);
    op2("ELSE", COMMENT("Load 'hard' $value: > 16 bits needed"), nl);
    op2("  ldr", str("$reg"), str("=value"));
    op2("ENDIF", nl);
    op2("MEND"));

    nl;

    op2("AREA", str("|Occam$$Code|,CODE,READONLY"));
    op2("IMPORT", str("||Lib$$Request$$armlib||, WEAK"));
    if (!o_lib) {
        op2("EXPORT", str("entry_user"));
        tline("entry_user");
    }
    
}


void final (void)			/* final stuff */
{
    int c_ws;

    if (!o_lib) {
        c_ws = ((ws	/* workspace - includes 3 words for entry stack frame */
                 + vs	/* vector space */
                 + 6	/* entry stack frame - 2 would probably do */
                 + 8	/* locals */
                 + 24)	/* system (rounded up) */
                *4	/* bytes */
                + 7) & ~7, /* rounded up to double-word alignment ... */
/*         op2("EXTERN", str("space"), */
/*         equate("space", dec(c_ws), */
    }

    fprintf (out, "\n%c\t%d input lines\n", CCHAR, iline);
    fprintf (out, "%c\t%d output lines\n", CCHAR, oline);
    fprintf (out, "%c\tWorkspace = %d bytes\n", CCHAR, ws*4);
    if (vs != 0)
        fprintf (out, "%c\tVector space = %d bytes\n", CCHAR, vs*4);

    if (shufflecount != 0) {
        fprintf (out, "%c\t%d stack shuffle", CCHAR, shufflecount);
        if (shufflecount > 1)
            fputc ('s', out);
        fputc ('\n', out);
    }
  
    if (lshufflecount != 0) {
        fprintf (out, "%c\t%d label stack shuffle", CCHAR, lshufflecount);
        if (lshufflecount > 1)
            fputc ('s', out);
        fputc ('\n', out);
    }
  
    fprintf (out, "\n\tEND\n");
    if (!o_lib) {
        if (b_entry[0] != 0)
        {
            fprintf (out, "%c\tEntry point = %s\n", CCHAR, b_entry);
        }
        else {
            fputs ("\n... Incomplete program - no entry point\n", stderr);
            exit (1);	/* so shell file won't run it */
        }
    }
}


void align (void)			/* align to address boundary */
{
    /* Align to (l)word address */
    op2("ALIGN", dec(o_real64 ? BytesPerWord*2 : BytesPerWord));
}


void delay (const char *s)			/* delay slot comment */
{
}


void annul (const char *s)			/* annul slot comment */
{
}


