/*
 * KROC translator: Function definitions and prototypes
 *
 * $Source: /extra/play/cvs/kroc/src/target/ops.h,v $
 *
 * $Id: ops.h,v 1.1 1999/09/27 21:24:00 rivimey Exp $
 *
 * ops.h
 * (C) Copyright 1999 R. Ivimey-Cook <ruthc@sharra.demon.co.uk>
 *
 */

#ifndef OCTRAN_OPS_H
#define OCTRAN_OPS_H

typedef void InstrProc (void);
typedef void InitialiseProc (void);
typedef void CloseProc (void);
typedef void TargetDescriptionProc(char *buf, int size);

/* flags defns */
#define target_has_fpu       0x01  /* FP instruction support (FMUL etc) */
#define target_has_softfloat 0x01  /* Soft-FP instruction support (UNPACKSN etc) */
#define target_has_2dops     0x02  /* MOVE2D et al instruction support */
#define target_has_halfword  0x04  /* halfword ( 16 bit-on-32 bit ) support */
#define target_has_t9ops     0x08  /* t9000 extensions */
#define target_is_bigend     0x10  /* if set, big endian, else little */

/* set unimplemented opcode fnd. to this */
InstrProc Unimplemented;


typedef struct 
{
    char targetName[16];   /* base name of target -- e.g. "ARM" */        
    unsigned flags;
    
    TargetDescriptionProc *DescribeTarget;
    InitialiseProc *Init;
    CloseProc *Close;

    InstrProc *OpADC;
    InstrProc *OpADD;
    InstrProc *OpAJW;
    InstrProc *OpALT;
    InstrProc *OpALTEND;
    InstrProc *OpALTWT;
    InstrProc *OpAND;
    InstrProc *OpBCNT;
    InstrProc *OpBITCNT;
    InstrProc *OpBITREVNBITS;
    InstrProc *OpBITREVWORD;
    InstrProc *OpBSUB;
    InstrProc *OpCALL;
    InstrProc *OpCAUSEERROR;
    InstrProc *OpCB;
    InstrProc *OpCBU;
    InstrProc *OpCCNT1;
    InstrProc *OpCFLERR;
    InstrProc *OpCJ;
    InstrProc *OpCLRHALTERR;
    InstrProc *OpCRCBYTE;
    InstrProc *OpCRCWORD;
    InstrProc *OpCS;
    InstrProc *OpCSNGL;
    InstrProc *OpCSU;
    InstrProc *OpCSUB0;
    InstrProc *OpCWORD;
    InstrProc *OpDIFF;
    InstrProc *OpDISC;
    InstrProc *OpDISS;
    InstrProc *OpDIST;
    InstrProc *OpDIV;
    InstrProc *OpDUP;
    InstrProc *OpENBC;
    InstrProc *OpENBS;
    InstrProc *OpENBT;
    InstrProc *OpENDP;
    InstrProc *OpFMUL;
    InstrProc *OpFPUABS;
    InstrProc *OpFPADD;
    InstrProc *OpFPADDDBSN;
    InstrProc *OpFPB32TOR64;
    InstrProc *OpFPCHKERR;
    InstrProc *OpFPUCHKI32;
    InstrProc *OpFPUCHKI64;
    InstrProc *OpFPDIV;
    InstrProc *OpFPUDIVBY2;
    InstrProc *OpFPDUP;
    InstrProc *OpFPENTRY;
    InstrProc *OpFPEQ;
    InstrProc *OpFPUEXPDEC32;
    InstrProc *OpFPUEXPINC32;
    InstrProc *OpFPGE;
    InstrProc *OpFPGT;
    InstrProc *OpFPI32TOR32;
    InstrProc *OpFPI32TOR64;
    InstrProc *OpFPINT;
    InstrProc *OpFPLDNLADDDB;
    InstrProc *OpFPLDNLADDSN;
    InstrProc *OpFPLDNLDB;
    InstrProc *OpFPLDNLDBI;
    InstrProc *OpFPLDNLMULDB;
    InstrProc *OpFPLDNLMULSN;
    InstrProc *OpFPLDNLSN;
    InstrProc *OpFPLDNLSNI;
    InstrProc *OpFPLDZERODB;
    InstrProc *OpFPLDZEROSN;
    InstrProc *OpFPLG;
    InstrProc *OpFPMUL;
    InstrProc *OpFPUMULBY2;
    InstrProc *OpFPNAN;
    InstrProc *OpFPNOTFINITE;
    InstrProc *OpFPORDERED;
    InstrProc *OpFPUR32TOR64;
    InstrProc *OpFPUR64TOR32;
    InstrProc *OpFPREM;
    InstrProc *OpFPREMFIRST;
    InstrProc *OpFPREMSTEP;
    InstrProc *OpFPREV;
    InstrProc *OpFPURM;
    InstrProc *OpFPURN;
    InstrProc *OpFPURP;
    InstrProc *OpFPRTOI32;
    InstrProc *OpFPURZ;
    InstrProc *OpFPSTNLDB;
    InstrProc *OpFPSTNLI32;
    InstrProc *OpFPSTNLSN;
    InstrProc *OpFPSUB;
    InstrProc *OpFPTESTERR;
    InstrProc *OpFPUCLRERR;
    InstrProc *OpFPUNOROUND;
    InstrProc *OpFPUSETERR;
    InstrProc *OpFPUSQRTFIRST;
    InstrProc *OpFPUSQRTLAST;
    InstrProc *OpFPUSQRTSTEP;
    InstrProc *OpGAJW;
    InstrProc *OpGCALL;
    InstrProc *OpGT;
    InstrProc *OpGTU;
    InstrProc *OpIN;
    InstrProc *OpJ;
    InstrProc *OpLADD;
    InstrProc *OpLB;
    InstrProc *OpLBX;
    InstrProc *OpLDC;
    InstrProc *OpLDCNT;
    InstrProc *OpLDIFF;
    InstrProc *OpLDINF;
    InstrProc *OpLDIV;
    InstrProc *OpLDL;
    InstrProc *OpLDLP;
    InstrProc *OpLDNL;
    InstrProc *OpLDNLP;
    InstrProc *OpLDPI;
    InstrProc *OpLDPRI;
    InstrProc *OpLDTH;
    InstrProc *OpLDTIMER;
    InstrProc *OpLEND;
    InstrProc *OpLMUL;
    InstrProc *OpLS;
    InstrProc *OpLSHL;
    InstrProc *OpLSHR;
    InstrProc *OpLSUB;
    InstrProc *OpLSUM;
    InstrProc *OpLSX;
    InstrProc *OpMINT;
    InstrProc *OpMOVE;
    InstrProc *OpMOVE2DALL;
    InstrProc *OpMOVE2DINIT;
    InstrProc *OpMOVE2DNONZERO;
    InstrProc *OpMOVE2DZERO;
    InstrProc *OpMUL;
    InstrProc *OpNFIX;
    InstrProc *OpNORM;
    InstrProc *OpNOT;
    InstrProc *OpOPR;
    InstrProc *OpOR;
    InstrProc *OpOUT;
    InstrProc *OpOUTBYTE;
    InstrProc *OpOUTWORD;
    InstrProc *OpPFIX;
    InstrProc *OpPOSTNORMSN;
    InstrProc *OpPROD;
    InstrProc *OpREM;
    InstrProc *OpRESETCH;
    InstrProc *OpRET;
    InstrProc *OpREV;
    InstrProc *OpROUNDSN;
    InstrProc *OpRUNP;
    InstrProc *OpSAVEH;
    InstrProc *OpSAVEL;
    InstrProc *OpSB;
    InstrProc *OpSETERR;
    InstrProc *OpSETHALTERR;
    InstrProc *OpSHL;
    InstrProc *OpSHR;
    InstrProc *OpSS;
    InstrProc *OpSSUB;
    InstrProc *OpSTARTP;
    InstrProc *OpSTHB;
    InstrProc *OpSTHF;
    InstrProc *OpSTL;
    InstrProc *OpSTLB;
    InstrProc *OpSTLF;
    InstrProc *OpSTNL;
    InstrProc *OpSTOPERR;
    InstrProc *OpSTOPP;
    InstrProc *OpSTTIMER;
    InstrProc *OpSUB;
    InstrProc *OpSUM;
    InstrProc *OpTALT;
    InstrProc *OpTALTWT;
    InstrProc *OpTESTERR;
    InstrProc *OpTESTHALTERR;
    InstrProc *OpTESTPRANAL;
    InstrProc *OpTIN;
    InstrProc *OpUNPACKSN;
    InstrProc *OpVIN;
    InstrProc *OpVOUT;
    InstrProc *OpWCNT;
    InstrProc *OpWSUB;
    InstrProc *OpWSUBDB;
    InstrProc *OpXDBLE;
    InstrProc *OpXOR;
    InstrProc *OpXSWORD;
    InstrProc *OpXWORD;
} OpcodesProcVec;


#endif /* OCTRAN_OPS_H */
