/*
 * KROC SPARC translator: "Direct" instructions (with explicit operands).
 *
 * $Source: /extra/play/cvs/kroc/src/target/sparc/f_direct.c,v $
 *
 * $Id: f_direct.c,v 1.1 1999/09/27 21:10:13 rivimey Exp $
 *
 * (C) Copyright 1995 D.C. Wood <D.C.Wood@ukc.ac.uk>
 * University of Kent at Canterbury
 *
 * SPARC target only.
 */

static char f_direct_c_rcsid[] = "$Id: f_direct.c,v 1.1 1999/09/27 21:10:13 rivimey Exp $";

#include "octran.h"
#include <string.h>

int big (n)			/* deal with big operands */
int n;				/* beware of side-effect! */
{	if (n < MINO || n > MAXO)
	{	mnem("set"); dec(n); T0; COMMENT("big operand"); nl;
		return True;
	}
	else
		return False;
}

void J ()
{	state_type state;		/* normal jump */
	if (started)
	{	state = reflab (0);
		if (o_resched && state.B_DEF)
		{	mnem("ld"); vindex("sync"); T0; NL;
			mnem("tst"); T0; nl;
			mnem("beq"); str(b_operand); nl;
			mnem("nop"); nl;
			extcall(X_RESCHED); nl;
			ADJUST; nl;
		}
		mnem("ba,a"); str(b_operand); NL;
		istackp = 0;
		fstackp = 0;
		ishuff = 0;
		BREAK;
	}
	else if (o_lib)			/* ignore it */
	{	dummy("ba,a"); str(b_operand); NL; }
	else				/* jump to entry point */
	{	strcpy (b_entry, b_operand);	/* not used */
		entry();
		mnem("ba,a"); str(b_operand); NL;
		BREAK;
	}
}

void LDLP ()
{	int n = n_operand*BytesPerWord;
	PUSH;
	if (big (n))		/* "big" has side-effect, so can't take out common code */
	{	mnem("add"); W; T0; }
	else if (n > 0)
	{	mnem("add"); W; dec(n); }
	else if (n < 0)
	{	mnem("sub"); W; dec(-n); }
	else /* n == 0 */
	{	mnem("mov"); W; }
	A; NL;
}

void PFIX ()
{	if (n_operand == 0)
	{	mnem("nop"); NL; }
	else
		unimp ();
}

void LDNL ()
{	int n = n_operand*BytesPerWord;
	if (big (n))
	{	mnem("ld"); index2(Areg, Temp); }
	else
	{	mnem("ld"); sindex(Areg, n_operand); }
	A; NL;
}

void LDC ()
{	int i, j;
	PUSH;
	if (nflag)		/* numeric operand */
	{	if (n_operand == 0)
			mnem("clr");
		else if (n_operand >= MINO && n_operand <= MAXO)
		{	mnem("mov"); dec(n_operand); }
		else if (n_operand == MostNeg)
		{	mnem("mov"); MIN; }
		else if (n_operand >= MINO - MostNeg)
		{	mnem("sub"); MIN; dec(MostNeg - n_operand ); }
		else if (n_operand <= MAXO + MostNeg)
		{	mnem("add"); MIN; dec(n_operand - MostNeg); }
		else
		{	mnem("set"); dec(n_operand); }
		A; NL;
	}
	else			/* symbolic operand */
	{	if (o_label)
		{	if (strcmp (b_operand, "case_scale") == 0)
			{	mnem("mov"); str(b_operand); A; NL; }
			else
			{	/* L<left>-L<right> */
				strcpy (l_operand, b_operand);
				strcpy (l_comment, b_comment);
				for (i = 0; b_operand[i] != '-'; ++i)
					l_left[i] = b_operand[i];
				l_left[i] = 0;
				for (++i, j = 0; b_operand[i] != 0; ++i, ++j)
					l_right[j] = b_operand[i];
				l_right[j] = 0;
			}
		}
		else
		{	mnem("set"); str(b_operand); A; NL; }
	}
}

void LDNLP ()
{	int n = n_operand*BytesPerWord;
	if (n != 0)
	{	if (big (n))
		{	mnem("add"); A; T0; }
		else if (n > 0)
		{	mnem("inc"); dec(n); }
		else /* n < 0 */
		{	mnem("dec"); dec(-n); }
		A;
	}
	else
		dummy("null operation");
	NL;
}

void NFIX ()
{	unimp ();
}

void LDL ()
{	int n = n_operand*BytesPerWord;
	if (strcmp (b_comment, " throw the result away") == 0)
		dummy("ignored");
	else
	{	PUSH;
		if (big (n))
		{	mnem("ld"); index2(Wptr, Temp); }
		else
		{	mnem("ld"); sindex(Wptr, n_operand); }
		A;
	}
	NL;
}

void ADC ()
{	if (n_operand != 0)
	{	if (big (n_operand))
		{	mnem("addcc"); A; T0; }
		else if (n_operand > 0)
		{	mnem("inccc"); dec(n_operand); }
		else if (n_operand < 0)
		{	mnem("deccc"); dec(-n_operand); }
		A; NL;
		CHECK;
	}
	else
	{	dummy("null operation"); NL; }
	if (strstr (b_comment, " check for replicator overflow") != NULL)
		POP;
}

void CALL ()				/* unadjusted PC used internally */
{	int i, j, temp;
	/* Should leave return address in A as well as W[0] -	*/
	/*	sethi	%hi(b_operand),T			*/
	/*	jmpl	T+%lo(b_operand),A			*/
	/*	st	A,[W]					*/
	/* but occam never uses this.				*/
	/*							*/
	/* Should also use adjusted PC, for consistency -	*/
	/*	call	1f					*/
	/*	inc	12,L					*/
	/* 1:	ba	b_operand				*/
	/*	st	L,[W]					*/
	for (i = 0; i <= stcount; ++i)	/* ?numerical order => binary search */
		if (strcmp (b_operand, proctab[i].label) == 0)
			break;
	if (i > stcount)			/* shouldn't happen */
	{	fputs ("\n*** Call of undefined PROC/FUNCTION", stderr);
		where ();
		mnem("dec"); dec(4*BytesPerWord); W; NL;
		for (j = 1; istackp > 0; ++j)
		{	mnem("st"); A; sindex(Wptr, j); nl;
			POP;
		}
		mnem("call"); str(b_operand); nl;
		mnem("st"); L; sindex(Wptr, 0); DELAY; nl;	/* unadjusted */
		istackp = 0;
		fstackp = 0;
		ishuff = 0;
	}					/* call C function */
	else if (proctab[i].S_PTYPE == P_C_EXTERNAL)
	{	for (j = 1; istackp > 0; ++j)
		{	mnem("st"); A; sindex(Wptr, j - 4); NL;
			POP;
		}
		mnem("call"); str(proctab[i].proc); COMMENT(b_operand); nl;
		mnem("sub"); W; dec(3*BytesPerWord); P0;
			delay("C parameter -> occam parameters"); nl;
		istackp = 0;
		fstackp = 0;
		ishuff = 0;
	}
	else					/* call occam PROC/FUNCTION */
	{	mnem("dec"); dec(4*BytesPerWord); W; NL;
		for (j = 1; istackp > 0; ++j)
		{	mnem("st"); A; sindex(Wptr, j); nl;
			POP;
		}
		mnem("call");
		if (proctab[i].S_PTYPE == P_EXTERNAL)
		{	str(proctab[i].proc); COMMENT(b_operand); nl; }
		else
		{	str(b_operand); COMMENT(proctab[i].proc); nl; }
		mnem("st"); L; sindex(Wptr, 0); DELAY; nl;	/* unadjusted */
		istackp = proctab[i].S_ISTACK;
		fstackp = proctab[i].S_FSTACK;
		ishuff = proctab[i].S_ISHUFF;
		temp = proctab[i].S_FTYPES;
		for (j = 1; j <= fstackp; ++j)	/* at most one returned (ORTMS p.10) */
		{	ftype[j] = temp & 1;
			temp >>= 1;
		}
	}
	if (o_verbose)
	{	fprintf (stderr, "\n... Return from %s %s %s\n",
			b_operand, ptype[proctab[i].S_PTYPE], proctab[i].proc);
		fprintf (stderr, "    istack := %d, fstack := [", istackp);
		for (j = fstackp; j > 0; --j)
			fputc ((ftype[j] == F_D) ? 'D' : 'S', stderr);
		fputc (']', stderr);
		where ();
	}
}

void CJ ()
{	int c = cond;
	(void) reflab (istackp);
	if (c != C_NONE)
	{	cond = C_NONE;
		if (c == C_ALWAYS)
			fputc (CCHAR, out);
		mnema(condtab[INV(c)].branch); str(b_operand); lineup ();
			fprintf (out, "%c+ was %s", CCHAR, condtab[c].name);
		if (c < C_ALWAYS)
		{	NL;
			mnem("mov"); FALSE; A; annul("make sure it's false");
		}
	}
	else
	{	mnem("tst"); A; NL;
		mnem("be"); str(b_operand); nl;
		mnem("nop");		/* we know it's false */
	}
	NL;
	ishuff += 1;
	POP;
}

void AJW ()
{	int n = n_operand*BytesPerWord;
	if (o_real64 && (n_operand&1) != 0)
	{	fputs ("\n*** Workspace alignment error", stderr);
		error();
	}
	if (n != 0)
	{	if (big (n))
		{	mnem("add"); W; T0; }
		else if (n > 0)
		{	mnem("inc"); dec(n); }
		else if (n < 0)
		{	mnem("dec"); dec(-n); }
		W;
	}
	else
		dummy("null operation");
	NL;
}

void EQC ()
{	if (n_operand == 0)
	{	if (cond != C_NONE)	/* idiom for inverting condition */
		{	dummy("invert condition"); NL;
			cond = INV(cond);
			return;		/* nasty */
		}
		else
		{	mnem("tst"); A; }
	}
	else if (n_operand == MostNeg)
	{	mnem("cmp"); A; MIN; }
	else if (big (n_operand))
	{	mnem("cmp"); A; T0; }
	else
	{	mnem("cmp"); A; dec(n_operand); }
	if (o_opt)
	{	COMMENT("remember EQ");
		cond = C_EQ;
	}
	else
	{	NL;
		mnem("be,a"); f(1); nl;
		mnem("mov"); TRUE; A; annul("true"); nl;
		mnem("mov"); FALSE; A; COMMENT("false"); nl;
	lab(1);
	}
	NL;
}

void STL ()
{	int n = n_operand*BytesPerWord;
	if (big (n))
	{	mnem("st"); A; index2(Wptr, Temp); }
	else
	{	mnem("st"); A; sindex(Wptr, n_operand); }
	NL;
	POP;
}

void STNL ()
{	int n = n_operand*BytesPerWord;
	if (big (n))
	{	mnem("st"); B; index2(Areg, Temp); }
	else
	{	mnem("st"); B; sindex(Areg, n_operand); }
	NL;
	pop (2);
}

void OPR ()
{	unimp ();
}
