/*
 * KROC SPARC translator: Support for floating-point instructions.
 *
 * $Source: /extra/play/cvs/kroc/src/target/sparc/float.c,v $
 *
 * $Id: float.c,v 1.1 1999/09/27 21:10:13 rivimey Exp $
 *
 * (C) Copyright 1995 D.C. Wood <D.C.Wood@ukc.ac.uk>
 * University of Kent at Canterbury
 *
 * Mostly dependent on target, but something similar will be needed
 * for anything with floating-point.
 */

static char float_c_rcsid[] = "$Id: float.c,v 1.1 1999/09/27 21:10:13 rivimey Exp $";

#include "octran.h"

void fpush (t)			/* push floating-point stack */
int t;
{	fstackp += 1;
	FAtype = t;
}

const char *fname (r, s)	/* f.p. register name */
int r, s;
{	static const char *fnames[4][2] =
	{	{"...", "..."},
		{"%f0", "%f1"},
		{"%f2", "%f3"},
		{"%f4", "%f5"}
	};
	if (r < 1)
	{	fputs ("\n*** Floating stack underflow", stderr);
		error ();
	}
	if (r > 3)
	{	fputs ("\n*** floating stack overflow", stderr);
		error ();
	}
/*	else	*/
		return fnames[r][s];
}

void fmnem (op)			/* generate f.p. SPARC mnemonic */
const char *op;
{	mnem(op);
	fputc (FS, out);
	++posn;
}

void fround (tround)		/* set f.p. rounding mode */
int tround;
{	static const char *rnm[4][2] =
	{	{	"RN",	"fprn"	},
		{	"RZ",	"fprz"	},
		{	"RP",	"fprp"	},
		{	"RM"	"fprm"	}
	};
	if (sround != tround)
	{	if (o_checks == CH_TRAP)
		{	mnem("ld"); Round(tround); FSR; COMMENT(rnm[tround][1]); nl; }
		else
		{	mnem("st"); FSR; FPtemp; COMMENT(rnm[tround][1]); nl;
							/* have to preserve AEXC */
			mnem("ldub"); FPtemp; T0; nl;
			if (tround == R_N)
			{	mnem("bclr"); hex(R_M << 6); }
			else if (tround == R_M)
			{	mnem("bset"); hex(R_M << 6); }
			else
			{	mnem("btog"); hex((sround ^ tround) << 6); }
			T0; nl;
			mnem("stb"); T0; FPtemp; nl;
			mnem("ld"); FPtemp; FSR; nl;
		}
		mnem("nop"); nl;	/* sometimes required */
		if (o_verbose)
		{	fprintf (stderr, "\n... Floating-point rounding mode %s to %s",
				rnm[sround][1], rnm[tround][1]);
			where ();
		}
		sround = tround;
	}
/*	else							*/
/*	{	dummy("fpround"); COMMENT("unchanged"); nl; }	*/
}

void fscheck (s)		/* check floating-point type */
int s;
{	static const char *typename[2] =
	{	"REAL32", "REAL64" };
	int ss = FAtype;
	if (ss != s)
	{	fputs ("\n*** Inconsistent floating-point types\n", stderr);
		fprintf (stderr, "    %s should be %s", typename[ss], typename[s]);
		error ();
	}
}
