!
! KRoC SPARC assembly-language scheduler.
!
! $Source: /extra/play/cvs/kroc/src/target/sparc/sched.s,v $
!
! $Id: sched.s,v 1.1 1999/09/27 21:10:13 rivimey Exp $
!
! (C) Copyright 1995 D.C. Wood <D.C.Wood@ukc.ac.uk>
!                    P.H. Welch <P.H.Welch@ukc.ac.uk>
!                    et al.
! University of Kent at Canterbury
!
! Contains most of the code concerned with scheduling and process queues,
! though "runp" and "startp" are translated in line.
!
! enter
!       Called from "child"; initializes registers and local stack frame.
!       Calls occam program at $$entry.
! fpconst or $$fpconst
!       Initializes floating-point registers.
! $$scheduler
!       Called by "endp" and used by $$in/$$out etc.
! $$stopp
!       Entry to $$scheduler; called by "stopp", "altwt", and "stoperr".
! $$resched
!       Entry to $$scheduler, optionally called in "lend" and backward "j".
! $$in, $$out, $$outbyte, $$outword
!       Communication over channels, including keyboard and screen.
! $$move
!       Block move; also used by "move2dall", and shared with $$in/$$out.
! $$tin
!       Timeout.
! $$taltwait
!       Timer ALT wait.
! $$dist
!       Disable timer.
!
! Functions called from translated code start with $$;
! names shared with C may start with _ depending on the compiler system.
!
! Instructions (other than "nop"s) in delay slots are commented with !*,
! and those in annulled delay slots with !@.
!
! Since the allocation of registers to the evaluation stack is static here,
! they are called P, Q, and R.
! Comments show the corresponding names A, B, and C.
!
! Originally written by:
!       Kneale J. Rothwell, Graham Shaw, and Keith E. Woollacott
! Optimization, endian swapping, etc. by:
!       David C. Wood
! Hard channels and timer queue stuff by:
!       Peter H. Welch (algorithms) and David C. Wood (coding)

#include "sched.h"

	.text
MAKENAME(sched_s_rcsid):
	.ascii  "$Id: sched.s,v 1.1 1999/09/27 21:10:13 rivimey Exp $\0"

	.data
	.align  8

!       Order must agree with translator

	.global variables
variables:
	.word   $$fsr,0                 !   0 FPtemp
	.word   R_N<<30|TEM             !   2 Round
	.word   R_Z<<30|TEM             !   3
	.word   R_P<<30|TEM             !   4
	.word   R_M<<30|TEM             !   5
	.word   AEXC                    !   6 Aexc
	.word   0                       !   7 Zero

#ifdef KERNEL_RUN

!       Table of entry points for position-independent calls
					! (8+)
	.word   $$fpconst               !   0 - in kernel
	.word   $$outbyte               !   1
	.word   $$outword               !   2
	.word   $$out                   !   3
	.word   $$in                    !   4
	.word   $$move                  !   5
	.word   $$stopp                 !   6
	.word   $$scheduler             !   7
	.word   $$resched               !   8
	.word   $$tin                   !   9
	.word   $$taltwt                !  10
	.word   $$dist                  !  11
	.word   $$vin                   !  12
	.word   $$vout                  !  13

	.word   $$dump                  !  14
	.word   $$fdump                 !  15

	.word   .mul                    !  16 - in system
	.word   .div                    !  17
	.word   .rem                    !  18
	.word   .umul                   !  19

	.word   MAKENAME(rtime)         !  20 - in RTS
	.word   MAKENAME(utime)         !  21

	.word   MAKENAME(remainder)     !  22 - in math library

#endif

!       Shared with "child.c"

!       Up to four bytes packed in one word
!       (Could have eight in a double word)

!       .align  8                       !  For "ldd"
	.global MAKENAME(combined_sync)
MAKENAME(combined_sync):
	.byte   0                       !  0x01000000 (spare)
#if 0
	.global MAKENAME(panic_sync)
MAKENAME(panic_sync):
#endif
	.byte   0                       !  0x00010000
	.global MAKENAME(tim_sync)
MAKENAME(tim_sync):
	.byte   0                       !  0x00000100
	.global MAKENAME(kbd_sync)
MAKENAME(kbd_sync):
	.byte   0                       !  0x00000001

!       Channels for special processes

	.global MAKENAME(kbd_chan)
_kbd_chan:
kbd_chan:
	.word   MostNeg                 !  NotProcess.p
scr_chan:
	.word   scr_workspace
err_chan:
	.word   err_workspace

!       .global _call_ualarm            !  phw
!_call_ualarm:                          !  phw
!       .word   0                       !  phw

!       Workspace for keyboard process (stdin)

	.word   MostNeg                 !  State/Pointer
	.word   MostNeg                 !  Link
	.word   MostNeg                 !  Iptr
kbd_workspace:
	.word   MostNeg                 !  Temp

!       Workspace for screen process (stdout)

	.word   scr_workspace           !  Pointer
	.word   MostNeg                 !  Link
	.word   hard_print              !  Iptr
scr_workspace:
	.word   0                       !  Temp
	.word   scr_chan                !  Channel
	.word   0                       !  Device (stdout)

!       Workspace for error process (stderr)

	.word   err_workspace           !  Pointer
	.word   MostNeg                 !  Link
	.word   hard_print              !  Iptr
err_workspace:
	.word   0                       !  Temp
	.word   err_chan                !  Channel
	.word   0                       !  Device (stderr)

!       Offsets from Vptr

	.global $$sync
	.global ks
	.global ts
	.global kc
	.global sc
	.global ec
	.global kw
	.global sw
	.global ew

$$sync  = MAKENAME(combined_sync) - variables
ks      = MAKENAME(kbd_sync) - variables
ts      = MAKENAME(tim_sync) - variables
#if 0
ps      = MAKENAME(panic_sync) - variables
#endif
kc      = kbd_chan - variables
sc      = scr_chan - variables
ec      = err_chan - variables
kw      = kbd_workspace - variables
sw      = scr_workspace - variables
ew      = err_workspace - variables

!cu     = call_ualarm - variables

	.text
	.align  4

! Process for keyboard.

kbd_process:
	call    MAKENAME(read_the_kbd)  !  In "child.c"
	add     Vptr,kw,Wptr            !* kbd_workspace
	mov     P0,Q
	add     Vptr,kc,R               !  kbd_chan
	call    $$outbyte
	inc     8,L                     !*

	call    MAKENAME(acknowledge)   !  In "child.c"
	nop
	ba,a    $$scheduler
!       ------------------------

! Output from a channel.
!
!       Q (A) - Word or byte to transfer
!       R (B) - Channel address

	.global $$outword
$$outword:
	ba      o.1
	mov     4,P                     !* Count of four
!       ------------------------

	.global $$outbyte
$$outbyte:
	mov     1,P                     !  Count of one
o.1:    st      Q,[Wptr]                !  Save word or byte
	mov     R,Q                     !  Get registers straight
	mov     Wptr,R                  !  R points to byte(s)

!       fall through

!       P (A) - Number of bytes to transfer
!       Q (B) - Channel address
!       R (C) - Address of source data

	.global $$out
$$out:
	ld      [Q],T0                  !  Get other process Wptr
	cmp     T0,MIN                  !  Has no other process communicated?
	be      next
!       nop

	bclr    1,T0                    !* Get rid of priority bit to get Wptr
	ld      [T0+Pointer],T1         !  Get ALT descriptor or destination addr
	add     MIN,Ready_d,T2          !  (Ready_p)
	cmp     T1,T2                   !  Special case?
	ble     1f                      !  Yes
	nop                             !  No - copy data

	st      MIN,[Q]                 !  (NotProcess_p)
	ba      copy                    !  Source address already in R
	mov     T1,Q                    !* Destination address
!       ------------------------

1:      be      next                    !  Ready
	nop

	st      T2,[T0+State]           !  Ready_p
	add     MIN,Waiting_d,T2        !  (Waiting_p)
	cmp     T1,T2                   !  Waiting?
	bne     next                    !  No - must be Enabling (Impossible?)
	nop                             !  Yes

	cmp     Fptr,MIN
	bne,a   1f
	st      T0,[Bptr+Link]          !@
	mov     T0,Fptr
1:      mov     T0,Bptr

next:   st      Wptr,[Q]                !  Workspace into channel
	ba      $$stopp
	st      R,[Wptr+Pointer]        !* Message address into workspace
!       ------------------------

! Process for hard output.

1:      call    MAKENAME(fputc)         !  fputc (character, port)
	ld      [Wptr+Device],P1        !*
2:      mov     1,P                     !  One byte
	ld      [Wptr+Channel],Q        !  From channel
	mov     Wptr,R                  !  Into workspace
	call    $$in
	inc     8,L
hard_print:
#ifdef LITTLE_ENDIAN
	ldub    [Wptr+3],P0             !  Get the character from workspace
#else /* BIG_ENDIAN */
	ldub    [Wptr],P0
#endif
	cmp     P0,Flush                !  Is it a Flush code?
	bne     1b                      !  No
	nop

	call    MAKENAME(fflush)        !  fflush (port)
	ld      [Wptr+Device],P0        !*
	ba,a    2b
!       ------------------------

! Input from a channel.
!
!       P (A) - Number of bytes to transfer; A > 0 (CWG p.21)
!       Q (B) - Channel address (hardware link address)
!       R (C) - Address of destination for data

	.global $$in
$$in:
	ld      [Q],T0                  !  Get other process Wptr
	cmp     T0,MIN                  !  Has other party communicated?
	be      next                    !  De-schedule this process
	nop

	st      MIN,[Q]                 !  Reset channel to NotProcess_p
	mov     R,Q                     !  Destination address
!       ba      copy
	ld      [T0+Pointer],R          !  Fetch source address

!       Shared by "$$in" and "$$out"; the cost is undetectable
!       T0 points to other workspace

copy:   cmp     Fptr,MIN
	bne,a   1f
	st      T0,[Bptr+Link]          !@
	mov     T0,Fptr
1:
#ifdef QUICK

!       The Transputer just puts the OTHER process on the queue,
!       and continues as this process.  For at least some programs,
!       this is slower on the SPARC.

	mov     T0,Bptr
#else

!       We put BOTH processes onto the queue, and then run the next
!       (which may be the other one).  This is "fairer", and sometimes
!       faster.

	st      Wptr,[T0+Link]          !  Chain on current process
	st      L,[Wptr+Iptr]           !  Place Iptr into workspace
	mov     Wptr,Bptr               !  Make it the last process
	mov     Fptr,Wptr               !  Get the first process in queue
	ld      [Wptr+Link],Fptr        !  Point to next process
	ld      [Wptr+Iptr],L           !  Make new process active
#endif

!       IN/OUT must be at least one byte (CWG p.21)
!       MOVE is unlikely to be used for a single byte

	cmp     P,1                     !  Single byte?
	bg,a    m.1                     !  No
	or      R,Q,T0                  !@ Duplicated

#ifdef LITTLE_ENDIAN
	btog    3,R                     !  Copy one byte (e.g. $$outbyte)
#endif
	ldub    [R],T0
#ifdef LITTLE_ENDIAN
	btog    3,Q
#endif
	jmp     L
	stb     T0,[Q]                  !*
!       ------------------------

! Block move:
!
!       P (A) - Number of bytes to copy
!       Q (B) - Destination address
!       R (C) - Source address
!
! Optimized for word-aligned data.
! Loop unrolling should help, and unaligned copies can be done with shifts.
!
! Blocks will frequently be whole arrays, which start on a word address,
! so even byte arrays can be copied in larger units half the time.
! With the new option in the compiler, double words are aligned.

	.global $$move
$$move:
	tst     P                       !  For MOVE and MOVE2DALL
	be      m.end
!       nop

	or      R,Q,T0                  !* Can we copy words?
m.1:    or      T0,P,T0                 !  IN/OUT come in here
	btst    1,T0
	bne,a   m.bytes
	deccc   P                       !@ Duplicated
	btst    2,T0
	bne,a   m.halves
	deccc   2,P                     !@ Duplicated
	btst    4,T0
	bne,a   m.words
	deccc   4,P                     !@ Duplicated

1:      deccc   8,P                     !  Copy double words
	ldd     [R+P],T0
	bg      1b
	std     T0,[Q+P]                !*

!       ba,a    m.end
	jmp     L
	nop
!       ------------------------

1:      deccc   4,P                     !  Copy words
m.words:ld      [R+P],T0
	bg      1b
	st      T0,[Q+P]                !*

!       ba,a    m.end
	jmp     L
	nop
!       ------------------------

#ifdef LITTLE_ENDIAN
1:      deccc   2,P                     !  Copy half-words
m.halves:
	add     R,P,T1
	btog    2,T1                    !  Endian bit-twiddle
	lduh    [T1],T0                 !  Fetch data
	add     Q,P,T1
	btog    2,T1
	bg      1b
	sth     T0,[T1]                 !* Place data
#else /* BIG_ENDIAN */
1:      deccc   2,P                     !  Copy half-words
m.halves:
	lduh    [R+P],T0
	bg      1b
	sth     T0,[Q+P]                !*
#endif

!       ba,a    m.end
	jmp     L
	nop
!       ------------------------

#ifdef LITTLE_ENDIAN
1:      deccc   P                       !  Copy bytes
m.bytes:add     R,P,T1
	btog    3,T1                    !  Endian bit-twiddle
	ldub    [T1],T0                 !  Fetch data
	add     Q,P,T1
	btog    3,T1
	bg      1b
	stb     T0,[T1]                 !* Place data
#else /* BIG_ENDIAN */
1:      deccc   P                       !  Copy bytes
m.bytes:ldub    [R+P],T0
	bg      1b
	stb     T0,[Q+P]                !*
#endif

m.end:  jmp     L
	nop
!       ------------------------

! This function de-schedules the current process.  It does this by simply
! leaving it off the end of the runnable processes queue.  If the process
! queue contains one or more runnable processes, then run the next process;
! otherwise finish.

	.global $$resched
$$resched:                              !  Reschedule on J and LEND
	cmp     Bptr,MIN
	bne,a   1f
	st      Wptr,[Bptr+Link]
	mov     Wptr,Fptr
1:      mov     Wptr,Bptr

!       fall through

	.global $$stopp
$$stopp:                                !  STOPP, ALTWT, STOPERR
	st      L,[Wptr+Iptr]           !  Save Iptr in workspace

!       fall through

	.global $$scheduler
$$scheduler:                            !  ENDP and internal
	ld      [Vptr+$$sync],T0        !  Get kbd_sync, tim_sync, and panic_sync
	tst     T0                      !  Any set?
	beq     s.queue                 !  No
	btst    1,T0                    !  kbd_sync?
	bne,a   kbd_process             !  Yes
	clrb    [Vptr+ks]               !@ Clear kbd_sync
	btst    0x100,T0                !  tim_sync?
	be      panic                   !  No - must be panic

!       be,a    MAKENAME(impossible)
!       mov     0,P0                    !@

	clrb    [Vptr+ts]               !* Clear tim_sync
!       clr     [Vptr+cu]               !  Clear call_ualarm


! SEQ
!   tim ? now
!   removed := FALSE
!   WHILE (Tptr <> NotProcess.p) AND (NOT (Tptr^[Time] AFTER now))
!     SEQ
!       ...  move first process on timer queue to the run queue
!       removed := TRUE
!   IF
!     (Tptr <> NotProcess.p) AND removed
!       ualarm (Tptr^[Time] MINUS now, 0)
!     TRUE
!       SKIP

	call    MAKENAME(rtime)         !  tim ? now
	clr     T0                      !* removed := FALSE
s.loop: cmp     Tptr,MIN                !  WHILE Tptr <> NotProcess_p ...
!       be      panic
	be      s.queue                 !  Short cut
	nop

	ld      [Tptr+Time],T1          !  ... AND NOT (Tptr^[Time] AFTER now)
	cmp     P0,T1                   !  Very careful!
	bneg    1f
	nop

! -- Move first process on timer queue to the run queue
!
! POINTER temp := Tptr:
! SEQ
!   Tptr := Tptr^[TLink]
!   temp^[TLink] := TimeSet.p           -- i.e. time expired
!   temp^[Time] := now                  -- ??
!   IF
!     temp^[State] = Ready.p
!       SKIP                            -- used to trap to impossible(1)
!                                       -- the ALTing process has
!                                       -- already been put on the
!                                       -- run queue by something else
!     TRUE
!       SEQ
!         temp^[State] := Ready.p
!         ...  add on to run queue

	mov     Tptr,T1                 !  temp := Tptr
	ld      [Tptr+TLink],Tptr       !  Remove from timer queue
	add     MIN,TimeSet_d,T2        !  temp^[TLink] := TimeSet.p
	st      T2,[T1+TLink]
	st      P0,[T1+Time]            !  temp^[Time] := now
	add     MIN,Ready_d,T2          !  temp^[State] = Ready.p?
	ld      [T1+State],T3
	cmp     T2,T3
!       be,a    MAKENAME(impossible)
!       mov     1,P0                    !@
	be,a    s.loop
	mov     True,T0                 !* removed := TRUE

	st      T2,[T1+State]           !  temp^[State] := Ready.p
	cmp     Fptr,MIN                !  Process queue empty?
	bne,a   2f
	st      T1,[Bptr+Link]          !@ No
	mov     T1,Fptr
2:      mov     T1,Bptr

	ba      s.loop
	mov     True,T0                 !* removed := TRUE
!       ------------------------

1:
!       tst     T0                      !  IF removed
!       be      s.queue
!       be      panic                   !  phw
!       nop                             !  phw

!       st      MIN,[Vptr+cu]           !  phw

	sub     T1,P0,P0                !* Set alarm
	call    MAKENAME(ualarm)
	clr     P1                      !*

!!      st      MIN,[Tptr+Link]
s.queue:cmp     Fptr,MIN                !  Any processes in queue?
	bne,a   2f                      !  Yes
	mov     Fptr,Wptr               !@ Make the next process current

	ld      [Vptr+kc],T0            !  kbd_chan
	cmp     T0,MIN                  !  Empty?
	bne     1f                      !  No
!       nop

	cmp     Tptr,MIN                !* Yes - Timer queue empty?
	be      MAKENAME(dead)          !  Yes - No more processes to run
	and     STATUS,1,P0             !* Result (not coming back)

1:      call    MAKENAME(safe_pause)    !  No - wait for signal
	nop
!       tst     P0
!       bne     panic
!       nop
	ba,a    $$scheduler
!       ------------------------

2:                                      !  The queue is not empty
	cmp     Fptr,Bptr               !  Only one element?
	bne     1f                      !  No
	ld      [Wptr+Iptr],L           !*
	mov     MIN,Bptr                !  Make Fptr and Bptr null
	jmp     L
	mov     MIN,Fptr                !*
!       ------------------------

1:      jmp     L                       !  The workspace pointer is set up.
	ld      [Fptr+Link],Fptr        !* Just need to pass control to the
!       ------------------------        !  new current process and move
					!  the front pointer on.

! Timeout.
!
!       R (A) - Time.
!
! Only real (not user) time.

	.global $$tin
$$tin:
	mov     L,T0                    !  Save link
	call    MAKENAME(rtime)         !  Real time (in "time.c")
	nop
	cmp     P0,R                    !  After time (or same)?
	bpos    t.end                   !  Yes - return
	nop                             !  No - wait

	st      T0,[Wptr+Iptr]
	add     MIN,Waiting_d,T0        !  T9000 book says "Enabling"
	st      T0,[Wptr+State]
	inc     R                       !  T9000 book again
	st      R,[Wptr+Time]

! POINTER current := Tptr:
! POINTER previous := NotProcess.p:
! SEQ
!   WHILE (current <> NotProcess.p) AND (timeout AFTER current^[Time])
!     SEQ
!       previous := current             -- deduce: previous <> NotProcess.p
!       current := current^[TLink]
!   Wptr^[TLink] := current
!   IF
!     previous <> NotProcess.p          -- iff: current <> Tptr
!       SEQ                             -- insert in middle of queue
!         previous^[TLink] := Wptr
!     TRUE
!       SEQ                             -- insert at head of queue
!         Tptr := Wptr
!         tim ? now
!         ualarm (timeout MINUS now, 0)

	mov     Tptr,T0                 !  current := Tptr
	mov     MIN,T1                  !  previous := NotProcess.p
t.loop: cmp     T0,MIN                  !  WHILE (current <> MIN) ...
	be      1f
	nop

	ld      [T0+Time],T2            !  ... AND (A AFTER current^[Time])
	cmp     T2,R
	bpos    1f
	nop

	mov     T0,T1                   !  previous := current
	ba      t.loop
	ld      [T0+TLink],T0           !* current := current^[TLink]
!       ------------------------

1:      st      T0,[Wptr+TLink]         !  Wptr^[TLink] := current
!!      st      T1,[Wptr+Link]
!!      cmp     T0,MIN
!!      bne,a   1f
!!      st      Wptr,[T0+Link]          !@
1:      cmp     T1,MIN                  !  IF previous <> MIN
	bne,a   $$scheduler
	st      Wptr,[T1+TLink]         !@ previous^[TLink] := Wptr

!       st      MIN,[Vptr+cu]           !  phw

	sub     R,P0,P0                 !  Set alarm
	call    MAKENAME(ualarm)
	clr     P1                      !*

	ba      $$scheduler
	mov     Wptr,Tptr               !* Tptr := Wptr
!       ------------------------

t.end:  jmp     T0                      !  Return
	nop
!       ------------------------

! Timer ALT Wait.

	.global $$taltwt
$$taltwt:

! SEQ
!   Wptr^[Temp] := NoneSelected.o
!   IF
!     Wptr^[State] = Ready.p
!       SEQ                                     -- why?
!         tim ? now
!         Wptr^[Time] := now
!     Wptr^[TLink] = TimeNotSet.p
!       SEQ
!         Wptr^[State] = Waiting.p              -- see altwt
!         ...  wait until Wptr^[State] = Ready.p
!     now AFTER Wptr^[Time]                     -- T9000 Instruction Set p.436
!       SEQ
!         Wptr^[State] := Ready.p
!         Wptr^[Time] := now                    -- why not?
!     TRUE
!       SEQ
!         Wptr^[Time] := Wptr^[Time] + 1
!         ...  insert this process in timer queue
!         IF
!           Wptr^[State] <> Ready.p             -- redundant?
!             SEQ
!               Wptr^[State] := Waiting.p       -- see altwt
!               ...  wait until Wptr^[State] = Ready.p
!           TRUE
!             impossible (1)

	mov     NoneSelected_o,T0       !  Wptr^[Temp] := NoneSelected.o
	st      T0,[Wptr+Temp]
	add     MIN,Ready_d,T0
	ld      [Wptr+State],T1
	cmp     T0,T1
	bne     1f
!       nop

	mov     L,T0                    !*
	call    MAKENAME(rtime)
	nop
	jmp     T0
	st      P0,[Wptr+Time]          !*
!       ------------------------

1:      add     MIN,TimeNotSet_d,T0
	ld      [Wptr+TLink],T1
	cmp     T0,T1
!       be      ta.end
	bne     1f
!       nop

	add     MIN,Waiting_d,T0        !*
	ba      $$stopp
	st      T0,[Wptr+State]         !*
!       ------------------------

1:      mov     L,T3
	call    MAKENAME(rtime)
	ld      [Wptr+Time],R           !* Run out of temporaries; see $$tin
	cmp     R,P0                    !  Even more careful! (was wrong)
	bpos,a  1f
	inc     R                       !@

	add     MIN,Ready_d,T0          !*
	st      T0,[Wptr+State]
	jmp     T3
	st      P0,[Wptr+Time]          !*
!       ------------------------

1:      st      R,[Wptr+Time]

! -- Insert process in timer queue.
!
! POINTER current := Tptr:
! POINTER previous := NotProcess.p:
! SEQ
!   WHILE (current <> NotProcess.p) AND (timeout AFTER current^[Time])
!     SEQ
!       previous := current             -- deduce: previous <> NotProcess.p
!       current := current^[TLink]
!   Wptr^[TLink] := current
!   IF
!     previous <> NotProcess.p          -- iff: current <> Tptr
!       SEQ                             -- insert in middle of queue
!         previous^[TLink] := Wptr
!     TRUE
!       SEQ                             -- insert at head of queue
!         Tptr := Wptr
!         ualarm (timeout MINUS now, 0)

	mov     Tptr,T0                 !  current := Tptr
	mov     MIN,T1                  !  previous := NotProcess.p
ta.loop:cmp     T0,MIN                  !  WHILE (current <> MIN) ...
	be      1f
	nop

	ld      [T0+Time],T2            !  ... AND (R AFTER current^[Time])
	cmp     T2,R
	bpos    1f
	nop

	mov     T0,T1                   !  previous := current
	ba      ta.loop
	ld      [T0+TLink],T0           !* current := current^[TLink]
!       ------------------------

1:      st      T0,[Wptr+TLink]         !  Wptr^[TLink] := current
!!      st      T1,[Wptr+Link]
!!      cmp     T0,MIN
!!      bne,a   1f
!!      st      Wptr,[T0+Link]          !@
1:      cmp     T1,MIN                  !  IF previous <> MIN
	bne,a   1f
	st      Wptr,[T1+TLink]         !@ previous^[TLink] := Wptr

	mov     Wptr,Tptr               !  ELSE Tptr := Wptr
!       st      MIN,[Vptr+cu]           !  phw

	sub     R,P0,P0                 !  Set alarm
	call    MAKENAME(ualarm)
	clr     P1                      !*

1:      add     MIN,Ready_d,T0
	ld      [Wptr+State],T1
	mov     T3,P1                   ! return address to occam
	cmp     T0,T1
	be,a    MAKENAME(impossible)
	mov     2,P0                    !@

	mov     T3,L
ta.end: add     MIN,Waiting_d,T0        !  Wptr^[State] = {T}Waiting.p
	ba      $$stopp
	st      T0,[Wptr+State]         !*
!       ------------------------

! Disable Timer.
!
!       On input:               On output:
!       P (A) - code.offset
!       Q (B) - pre.condition
!       R (C) - timeout         R (A) - result

	.global $$dist
$$dist:

! IF
!   NOT pre.condition
!     result := FALSE
!   Wptr^[TLink] = TimeNotSet.p
!     result := FALSE
!   Wptr^[TLink] = TimeSet.p
!     -- some timer guard has gone off
!     IF
!       NOT (Wptr^[Time] AFTER timeout
!         result := FALSE               -- it wasn't this guard
!       Wptr^[Temp] <> NoneSelected.o
!         result := FALSE               -- an earlier one was selected
!       TRUE
!         SEQ
!           Wptr^[Temp] := code.offset  -- this is the first ready guard
!           result := TRUE
!   TRUE                                -- no timer guard ready; cancel timeout
!     result := FALSE
!     ... remove process from timer queue

	tst     Q                       !  Pre-condition FALSE?
	be      1f                      !  Yes
!       nop

	add     MIN,TimeNotSet_d,T0     !* Timeout cancelled?
	ld      [Wptr+TLink],T1
	cmp     T0,T1
	be      1f                      !  Yes
!       nop

	add     MIN,TimeSet_d,T0        !* Has some timer guard gone off?
	cmp     T0,T1
	bne,a   d.queue                 !  No
	clr     R                       !@ result := FALSE

	ld      [Wptr+Time],T0          !  Was it our guard?
	cmp     T0,R                    !  Very careful!
	bneg    1f                      !  No
	nop

	ld      [Wptr+Temp],T0          !  Was an earlier guard selected?
	cmp     T0,NoneSelected_o
	bne     1f                      !  Yes
	nop

!       It was this guard, and this is the first ready one

	st      P,[Wptr+Temp]           !  Wptr^[Temp] := code.offset
	jmp     L
	mov     True,R                  !* Return TRUE
!       ------------------------

1:      jmp     L
	clr     R                       !* Return FALSE
!       ------------------------

!       Note that the result in R has now been dealt with in all cases.

! -- No timer guard ready; cancel timeout
! -- Remove process from timer queue.
!
! SEQ
!   IF
!     Tptr = Wptr                       -- is it head of timer queue?
!       SEQ
!         Tptr := Tptr^[TLink]          -- remove head of timer queue
!         IF
!           Tptr <> NotProcess.p        -- is timer queue not empty?
!             SEQ
!               -- Tptr^[Link] := NotProcess.p
!               IF
!                 (Tptr^[Time] MINUS now) <= 0
!                   tim_sync := 1       -- time-out has happened!
!                 TRUE
!                   ualarm (Tptr^[Time] MINUS now, 0)
!           TRUE
!             SKIP
!     Tptr = NotProcess.p               -- implies process not on timer queue
!       impossible (3)
!     TRUE
!       -- POINTER previous := Wptr^[Link]:
!       -- next := Wptr^[TLink]
!       -- IF
!       --   next <> NotProcess.p
!       --     next^[Link] := previous
!       --   TRUE
!       --     SKIP
!       -- previous^[TLink] := next
!       POINTER previous := Tptr:       -- search from 2nd item on timer queue
!       POINTER current := Tptr^[TLink]:
!       SEQ
!         WHILE (current <> NotProcess.p) AND (current <> Wptr)
!           SEQ
!             previous := current
!             current := current^[TLink]
!         IF
!           current = NotProcess.p
!             impossible (4)
!           TRUE
!             SKIP
!         -- deduce: current = Wptr
!         next := Wptr^[TLink]
!         previous^[TLink] := next
!   Wptr^[TLink] := TimeNotSet.p        -- i.e. timer cancelled

d.queue:cmp     Tptr,Wptr               !  Head of timer queue?
	bne     2f                      !  No
	nop

	ld      [Tptr+TLink],Tptr       !  Remove head of timer queue
	cmp     Tptr,MIN                !  Timer queue empty?
	be      d.end                   !  Yes
	nop

!!      st      MIN,[Tptr+Link]         !  {Backward link}
	mov     L,T0                    !  Save link
	call    MAKENAME(rtime)
	ld      [Tptr+Time],T1          !*
!       st      MIN,[Vptr+cu]           !  phw

	cmp     P0,T1
	bneg    1f
	sub     T1,P0,P0                !*

	mov     1,T1                    !  Set tim_sync
	stb     T1,[Vptr+ts]
	ba      d.end
	mov     T0,L                    !*
!       ------------------------

1:      call    MAKENAME(ualarm)
	clr     P1                      !*
	ba      d.end
	mov     T0,L                    !*
!       ------------------------

2:      mov     L, P1                   !  return address to occam
	cmp     Tptr,MIN                !  Process not there?
	be,a    MAKENAME(impossible)
	mov     3,P0                    !@

!       Search for Wptr after head of queue

	mov     Tptr,T0                 !  previous := Tptr
	ld      [Tptr+TLink],T1         !  current := Tptr^[TLink]
d.loop: cmp     T1,MIN                  !  WHILE current <> NotProcess.p ...
	be,a    MAKENAME(impossible)
	mov     4,P0                    !@

	cmp     T1,Wptr                 !  ... AND current <> Wptr
	be      1f
	nop

	mov     T1,T0                   !  previous := current
	ba      d.loop
	ld      [T1+TLink],T1           !* current := current^[TLink]
!       ------------------------

!!      ld      [Wptr+Link],T0          !  {Backward link instead of search}

!       We've found it

1:      ld      [Wptr+TLink],T2         !  next := Wptr^[TLink]
!!      cmp     T2,MIN                  !  {next = NotProcess.p?}
!!      bne,a   1f
!!      st      T0,[T2+Link]            !@ {No - next^[Link] := previous}
1:      st      T2,[T0+TLink]           !  previous^[TLink] := next

d.end:  add     MIN,TimeNotSet_d,T0     !  Wptr^[TLink] := TimeNotSet.p
	jmp     L
	st      T0,[Wptr+TLink]         !*
!       ------------------------

$$vin:                                  !  not implemented
$$vout:

! If something seriously impossible happens, read the clock and dump core.

panic:  call    MAKENAME(rtime)         !  phw
	nop                             !  phw
	ld      [Z+1],Z                 !  Force error not trapped
!       ------------------------
