/*
 * KROC C scheduler definitions and function prototypes
 *
 * $Source: /extra/play/cvs/kroc/src/target/unix/rts.h,v $
 *
 * $Id: rts.h,v 1.1 1999/09/27 21:10:13 rivimey Exp $
 *
 * (C) Copyright 1995,1996,1997 D.J. Beckett <D.J.Beckett@ukc.ac.uk>
 * University of Kent at Canterbury
 */

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <signal.h>
#include <errno.h>
#include <unistd.h>
#include <sys/types.h>

#include "config.h"

#include <sys/time.h>
#include <sys/resource.h>
#include <sys/termios.h>
#include <sys/wait.h>
#if defined(HOST_OS_OSF3)
#include <sys/timers.h>
#endif


/* Truth values */

typedef int             bool;

#define FALSE           0
#define TRUE            1

/* Unix system call error codes */

#define ERR             -1

/* keyboard signals */

#define SIGKBD          SIGUSR1 /* phw & dcw */
#if 0
#define SIGPANIC        SIGUSR2 /* phw & dcw */
#endif
/* Signal magic */
typedef void (*SigParam)(int);

/* tty chars */

#define NULL_CHAR       (char)0
#define END_CHAN        (char)255

/* trap codes */

#define _SETERR         16      /* unconditional (SETERR) */
#define _OFLOW          17      /* arithmetic overflow */
#define _RANGE          18      /* range check (CSUB0, CCNT1, CSNGL, CWORD) */
#define _FLOAT          19      /* floating-point (FPCHKERR, FPSETERR) */
#define _UNREC          20      /* unrecognized instruction */
#define _UNIMP          21      /* unimplemented instruction */


/* Globals, defined in rtsmain.c */
#ifndef RTS_MAIN

#ifdef DEBUG
extern bool trace;
#endif

#endif

/* Globals defined in sched.s */
extern int kbd_chan;

extern char kbd_sync;           /* packed in one word (dcw) */
extern char tim_sync;
#if 0
extern char panic_sync;
#endif
extern int combined_sync;

/* KROC C scheduler function prototypes (was proto.h) */

/* cdump.c */
void dump (int n, int iptr, int wptr, int a, int b, int c);
void fdump (int code, int *fa_p, int fb_h, int fb_l, int fc_h, int fc_l);

/* kbdproc.c */
void keyboard_process (int user_pid, int to_user_fd, int from_user_fd, bool is_a_tty);

/* userproc.c */
void occam_exit(int exit_status, int occam_iptr);
void user_bad_exit (void);
void user_good_exit (void);
void user_tim_handler (void);
#if 0
void user_panic_handler (void);
#endif
void user_kbd_handler (void);

#if defined(SIGNAL_TYPE_SYSV)
void user_trap_handler (int sig);
void user_fp_handler (int sig);
#else
void user_trap_handler (int sig, int code, struct sigcontext *scp, char *addr);
void user_fp_handler (int sig, int code, struct sigcontext *scp, char *addr);
#endif

char read_the_kbd (void);
void acknowledge (void);
void user_to_kbd_sync (void);
void impossible (int n, int occam_iptr);
void status (int erfl);
void dead (int erfl);
unsigned int utime (void);
unsigned int rtime (void);
#ifndef HAVE_UALARM
int ualarm(int delay, int interval);
#endif
int safe_pause (void);
void user_process (int kbd_pid, int to_kbd_fd, int from_kbd_fd, bool is_a_tty);

/* rtsmain.c */
extern char * short_cmdline;
extern char * long_cmdline;

/* support.c */
void main (int argc, char **argv);

/* sched.s - occam program entry */
int enter (void);

/* tty.c */
void save_tty_state (void);
void restore_tty_state (void);
