--
--	elf.inc - ELF file format definitions
--	Copyright (C) 1997 M.D. Poole
--
--	This program is free software; you can redistribute it and/or modify
--	it under the terms of the GNU General Public License as published by
--	the Free Software Foundation; either version 2 of the License, or
--	(at your option) any later version.
--
--	This program is distributed in the hope that it will be useful,
--	but WITHOUT ANY WARRANTY; without even the implied warranty of
--	MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
--	GNU General Public License for more details.
--
--	You should have received a copy of the GNU General Public License
--	along with this program; if not, write to the Free Software
--	Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
--

--
--	elf.inc
--	ELF file format definitions
--

-- ELF header
DATA TYPE ELF.EHDR
  PACKED RECORD
    [16]BYTE e.ident:
    INT16 e.type:
    INT16 e.machine:
    INT32 e.version:
    INT32 e.entry:
    INT32 e.phoff:
    INT32 e.shoff:
    INT32 e.flags:
    INT16 e.ehsize:
    INT16 e.phentsize:
    INT16 e.phnum:
    INT16 e.shentsize:
    INT16 e.shnum:
    INT16 e.shstrndx:
:

-- ELF header magic
VAL BYTE ELFMAG0 IS #7F:
VAL BYTE ELFMAG1 IS 'E':
VAL BYTE ELFMAG2 IS 'L':
VAL BYTE ELFMAG3 IS 'F':
VAL BYTE ELFCLASS32 IS 1:
VAL BYTE ELFDATA2LSB IS 1:

-- ELF.EHDR[e.version]
VAL BYTE EV.NONE IS 0:
VAL BYTE EV.CURRENT IS 1:

-- ELF.EHDR[e.type]
VAL INT16 ET.NONE IS 0:
VAL INT16 ET.REL IS 1:
VAL INT16 ET.EXEC IS 2:
VAL INT16 ET.DYN IS 3:
VAL INT16 ET.CORE IS 4:
VAL INT16 ET.LOPROC IS #FF00:
VAL INT16 ET.HIPROC IS #FFFF:

-- ELF.EHDR[e.machine]
VAL INT16 EM.NONE IS 0:
VAL INT16 EM.M32 IS 1:
VAL INT16 EM.SPARC IS 2:
VAL INT16 EM.386 IS 3:
VAL INT16 EM.68K IS 4:
VAL INT16 EM.88K IS 5:
VAL INT16 EM.860 IS 7:
VAL INT16 EM.MIPS IS 8:

-- ELF section header
DATA TYPE ELF.SHDR
  PACKED RECORD
    INT32 sh.name:
    INT32 sh.type:
    INT32 sh.flags:
    INT32 sh.addr:
    INT32 sh.offset:
    INT32 sh.size:
    INT32 sh.link:
    INT32 sh.info:
    INT32 sh.addralign:
    INT32 sh.entsize:
:

-- ELF.SHDR[sh.type]
VAL INT32 SHT.NULL IS 0:
VAL INT32 SHT.PROGBITS IS 1:
VAL INT32 SHT.SYMTAB IS 2:
VAL INT32 SHT.STRTAB IS 3:
VAL INT32 SHT.RELA IS 4:
VAL INT32 SHT.HASH IS 5:
VAL INT32 SHT.DYNAMIC IS 6:
VAL INT32 SHT.NOTE IS 7:
VAL INT32 SHT.NOBITS IS 8:
VAL INT32 SHT.REL IS 9:
VAL INT32 SHT.SHLIB IS 10:
VAL INT32 SHT.DYNSYM IS 11:
VAL INT32 SHT.LOPROC IS #70000000:
VAL INT32 SHT.HIPROC IS #7FFFFFFF:
VAL INT32 SHT.LOUSER IS #80000000:
VAL INT32 SHT.HIUSER IS #FFFFFFFF:

-- ELF.SHDR[sh.flags]
VAL INT32 SHF.WRITE IS 1:
VAL INT32 SHF.ALLOC IS 2:
VAL INT32 SHF.EXECINSTR IS 4:
VAL INT32 SHF.MASKPROC IS #F0000000:

-- Our section indexes
VAL INT16 SH.CODE IS 1:
VAL INT16 SH.REL IS 2:
VAL INT16 SH.DATA IS 3:
VAL INT16 SH.BSS IS 4:
VAL INT16 SH.SHSTR IS 5:
VAL INT16 SH.SYM IS 6:
VAL INT16 SH.SYSTR IS 7:

-- Special section indexes
VAL INT32 SHN.UNDEF IS 0:
VAL INT32 SHN.LORESERVE IS #FF00:
VAL INT32 SHN.LOPROC IS #FF00:
VAL INT32 SHN.HIPROC IS #FF1F:
VAL INT32 SHN.ABS IS #FFF1:
VAL INT32 SHN.COMMON IS #FFF2:
VAL INT32 SHN.HIRESERVE IS #FFFF:

-- Number of sections that will go out
VAL NUM.SECTIONS IS 8:

-- Bits for symbols (and their relocation)
DATA TYPE ELF.SYM
  PACKED RECORD
    INT32 st.name:
    INT32 st.value:
    INT32 st.size:
    BYTE  st.info:
    BYTE  st.other:
    INT16 st.shndx:
:

VAL BYTE STT.SECTION IS 3:
VAL BYTE STB.GLOBAL IS 1:
DATA TYPE ELF.REL
  PACKED RECORD
    INT32 r.offset:
    INT32 r.info:
:
VAL R.386.32 IS 1:
VAL R.386.PC32 IS 2:


